(******************************************************************)
(* UDatatAccessSparse.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Accs  un fichier au format texte)
@author(Ricco)
@created(11/05/2012)

Lecture des fichiers au format "sparse".
Une description est dispo ici --> http://c2inet.sce.ntu.edu.sg/ivor/cvm.html (File format)

>>>
The LibCVM Toolkit supports both sparse and dense data representations. The sparse format is the same as in SVMlight and LIBSVM. Each line represents one training example and is of the form:

    <line> .=. <target> <feature>:<value> <feature>:<value> ... <feature>:<value>

    <target> .=. <label> | <float>
    <feature> .=. <unsigned integer>

    <value> .=. <float>

For the dense format, each line represents one training example and is of the form:

    <line> .=. <value>,<value>, ... <value>,<target>

    <target> .=. <label> | <float>
    <value> .=. <float>
<<<

>>>
R.R. : DEUX REMARQUES IMPORTANTES,
Le point dcimal est forcment le "."
Toutes les valeurs sont numriques
Une ligne commenant par # est un commentaire
<<<
     
}
unit UDatatAccessSparse;

interface

USES
        Classes,
        UDataAccessDefinition;

TYPE
        {classe d'import de fichier texte dans un ensemble de donnes}
        TAccessSparse = class(TAccessAbstract)
        private
        {Position courante dans le fichier}
        FCurPos: Integer;
        {Taille du fichier}
        FFileSize: Integer;
        //ajouter des lignes  l'ensemble de donnes
        procedure   addRows(nbAddRows: integer);
        //ajouter une colonne
        procedure   addCols(nbAddCols: integer);
        protected
        function    coreDownload(prmThread: TThread): boolean; override;
        public
        function    GetProgression: Integer; override;
        end;

implementation

uses
        Sysutils, UFileResources,
        UDatasetDefinition, UDatasetImplementation, UIOFastTextFileAccess,
        UDlgDatasetDownload, UConstConfiguration, ULogFile;

const
        //sparateur de valeur - espace
        VALUE_SEP : char = #32;

        //sparateur colonne - valeur
        CODE_SEP : char = ':';

        //code pour les commentaires
        CODE_COMMENTS : char = '#';

        {dfinit le saut pour la rallocation de mmoire lors d'ajout d'individus}
        DELTA_ADD_EXAMPLES = 5000;

{ TAccessSparse }

procedure TAccessSparse.addCols(nbAddCols: integer);
var i,j: integer;
    curCols: integer;
    att: TAttribute;
begin
 curCols:= LstAtt.Count;
 for j:= 1 to nbAddCols do
  begin
   att:= TAttContinue.Create('v'+inttostr(pred(curCols+j)),LstAtt.Size);
   LstAtt.Add(att);
   //raz
   for i:= 1 to att.Size do
    att.cValue[i]:= 0.0;
  end;
end;

procedure TAccessSparse.addRows(nbAddRows: integer);
var i,j: integer;
    curRows: integer;
    att: TAttribute;
begin
 curRows:= LstAtt.Size;
 LstAtt.Size:= lstAtt.Size + nbAddRows;
 //mettre  zro les cases ajoutes
 for j:= 0 to pred(LstAtt.Count) do
  begin
   att:= lstAtt.Attribute[j];
   for i:= succ(curRows) to (curRows+nbAddRows) do
    att.cValue[i]:= 0.0;
  end;
end;

function TAccessSparse.coreDownload(prmThread: TThread): boolean;
const
 BUF_SIZE_KO = 16;
 SIZE_OF_BUF_TEXTFILE = BUF_SIZE_KO*1024;

var
 dThread: TThreadDownload;
 buf: array [0..pred(SIZE_OF_BUF_TEXTFILE)] of CHAR;//utiliser un buffer de BUF_SIZE Ko
 F: TextFile;
 col,row,p: integer;
 strLine, strValue: string;
 firstVar: TAttribute;
 curVar: TAttribute;
 value: TTypeContinue;
 err: integer;

begin
 firstVar:= nil;

 //rcuprer le thread
 dThread:= prmThread as TThreadDownload;
 //taille du fichier
 FFileSize:= tanagra_getFileSizeByName(FileName);
 FCurPos:= 0;//octets lus...

 row:= 0;

 result:= true;

 TRY
 //accs au fichier
 AssignFile(F,FileName);
 SetTextBuf(F,buf,SIZE_OF_BUF_TEXTFILE);
 Reset(F);

 //commencer la lecture du reste du fichier et l'analyse
 //tant que pas de fin de fichier et non interruption du thread
 while NOT(EOF(F)) and not(assigned(dThread) and dThread.isTerminated) do
  begin
   Readln(F,strLine);
   inc(FCurPos,length(strLine));

   //est-ce un commentaire
   if (strLine[1] = CODE_COMMENTS)
    then CONTINUE;//ne pas traiter la ligne et boucler

   //ajouter une ligne
   inc(row);
   
   //vrifier les lignes dispo
   if (row > LstAtt.Size)
    then self.addRows(DELTA_ADD_EXAMPLES);

   //lire la premire valeur
   p:= pos(VALUE_SEP,strLine);
   if (p > 0)
    then
     begin
      strValue:= copy(strLine,1,pred(p));
      delete(strLine,1,p);
     end
    else
     begin
      strValue:= strLine;
      strLine:= '';
     end;

    //crer le cas chant la premire variable
    if (firstVar = nil)
     then
      begin
       firstVar:= TAttContinue.Create('classe',LstAtt.Size);
       LstAtt.Add(firstVar);
      end;

     //attribuer la valeur
     val(strValue,value,err);//'.' est toujours le point dcimal
     firstVar.cValue[row]:= value;

     //boucler sur la ligne
     while (length(strLine) > 0) do
      begin
       p:= pos(VALUE_SEP, strLine);

       //rcuprer le couple "j:valeur"
       if (p > 0)
        then
         begin
          strValue:= copy(strLine,1,pred(p));
          delete(strLine,1,p);
         end
        else
         begin
          strValue:= strLine;
          strLine:= '';
         end;

         //rcuprer le numro de la colonne
         p:= pos(CODE_SEP,strValue);
         col:= strtoint(copy(strValue,1,pred(p)));

         //colonne dj prsente ?
         if (col >= LstAtt.Count)
          then self.addCols(col-LstAtt.Count+1);

         //pointer sur la variable
         curVar:= LstAtt.Attribute[col];

         //rcuprer la valeur
         delete(strValue,1,p);
         val(strValue,value,err);
         
         //affecter
         curVar.cValue[row]:= value;
      end;

  end;

 LstAtt.Size:= row;

 CloseFile(F);

 EXCEPT
 result:= false;
 END;

end;

function TAccessSparse.GetProgression: Integer;
begin
 if (FFileSize>0)
  then result:= trunc(100.0*FCurPos/(1.0*FFileSize))
  else result:= 0;
end;

end.
