(********************************************************************)
(* UCalcRulesDefinition.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Dfinition des structures de rgles)
@author(Ricco)
@created(12/01/2004)
Simplifi  l'extrme compte tenu de l'exprience accumule ces dernires annes.

New -- 09/08/2006 -- Elargi pour grer les autres cas. L'ide est de pouvoir introduire des conditions complexes
}
unit UCalcRulesDefinition;

interface

USES
        UDatasetDefinition;

TYPE

        {type d'oprateur}
        TTypeConditionOperator = integer;

        {pour le discret: In}
        TTypeConditionDiscrete = (rcdIn);

        {pour le continu : <, <=, =, <>, >=, >}
        TTypeConditionContinue = (rccL, rccLE, rccE, rccNE, rccGE, rccG);

CONST
        STR_RULE_CONDITION_CONTINUE : array[TTypeConditionContinue] of string = ('<','<=','=','<>','>=','>');

TYPE
        {une condition}
        TRuleCondition = class(TObject)
                         private
                         {attribut associ}
                         FAtt: TAttribute;
                         {oprateur associ - diffre selon le type rel de condtion}
                         FOp: TTypeConditionOperator;
                         public
                         {associer la variable et l'oprateur}
                         constructor Create(prmAtt: TAttribute; prmOp: TTypeConditionOperator);
                         {tester si la valeur rpond  la condition - valeur continue}
                         function    TestValue(const prmValue: TTypeContinue): boolean;  overload; virtual; abstract;
                         {tester si la valeur rpond  la condition - valeur discrte}
                         function    TestValue(const prmValue: TTypeDiscrete): boolean; overload; virtual; abstract;
                         {tester un exemple (associ  l'attribut courant donc)}
                         function    TestExample(const example: integer): boolean; virtual; abstract;
                         {fusionner deux conditions}
                         procedure   Merge(prmOther: TRuleCondition); virtual; abstract;
                         {dupliquer la condition - comma au scrabble}
                         function    Duplicate(): TRuleCondition; virtual; abstract;
                         {la description en HTML}
                         function    getHTMLDescription(): string; virtual;
                         {la description en TXT - description courte ou approfondie -- short veut dire galit simple pour les tests sur var. discrtes}
                         function    getTXTDescription(short: boolean = TRUE): string; virtual;
                         //new -- 11/08/2006 -- gnration d'une condition  partir d'une chane de description
                         class function generateFrom(condition: string; lstVar: TLstAttributes): TRuleCondition;
                         {variable en jeu}
                         property    Attribute: TAttribute read FAtt;
                         {oprateur}
                         property    Operator: TTypeConditionOperator read FOp;
                         end;

        {gestionnaire de condition discrte}
        TSetDiscreteValue = set of TTypeDiscrete;

        {une condition pour variable discrte, cf. TTypeConditionDiscrete}
        TRuleCondDiscrete = class(TRuleCondition)
                            private
                            FSetValues: TSetDiscreteValue;
                            public
                            constructor Create(prmAtt: TAttribute; prmOp: TTypeConditionOperator); overload;
                            constructor Create(prmAtt: TAttribute; prmOp: TTypeConditionOperator; prmValue: TTypeDiscrete); overload;
                            function    TestValue(const prmValue: TTypeDiscrete): boolean; override;
                            function    TestValue(const prmValue: TTypeContinue): boolean; override;
                            function    TestExample(const example: integer): boolean; override;
                            {c'est une union}
                            procedure   Merge(prmOther: TRuleCondition); override;
                            function    Duplicate(): TRuleCondition; override;
                            function    getHTMLDescription(): string; override;
                            function    getTXTDescription(short: boolean = TRUE): string; override;
                            //vider
                            procedure   clear();
                            //ajouter une valeur dans le "IN"
                            procedure   addValue(value: TTypeDiscrete);
                            //vrifier s'il est vide
                            function    isEmpty(): boolean;
                            //new 26/11/2011 - oprateur intersection entre deux conditions de mme variable
                            procedure   intersection(other: TRuleCondDiscrete); 
                            //proprits
                            property    SetValues: TSetDiscreteValue read FSetValues;
                            end;

        {une condition pour variable continue, cf. TTypeConditionContinue}
        TRuleCondContinue = class(TRuleCondition)
                            private
                            FThresoldValue: TTypeContinue;
                            public
                            constructor Create(prmAtt: TAttribute; prmOp: TTypeConditionOperator; prmValue: TTypeContinue);
                            function    TestValue(const prmValue: TTypeDiscrete): boolean; override;
                            function    TestValue(const prmValue: TTypeContinue): boolean; override;
                            function    TestExample(const example: integer): boolean; override;
                            {simple mise  jour du seuil}
                            procedure   Merge(prmOther: TRuleCondition); override;
                            function    Duplicate(): TRuleCondition; override;
                            procedure   SetNewThresold(prmValue: TTypeContinue);
                            function    getHTMLDescription(): string; override;
                            function    getTXTDescription(short: boolean = TRUE): string; override;
                            //rcuprer l'oprateur  partir d'une chaine de description
                            class function getCondFromDescription(description: string): TTypeConditionContinue;
                            property    ThresoldValue: TTypeContinue read FThresoldValue;
                            end;

implementation

uses
        Sysutils, UConstConfiguration;

{ TRuleCondition }

constructor TRuleCondition.Create(prmAtt: TAttribute; prmOp: TTypeConditionOperator);
begin
 inherited Create();
 FAtt:= prmAtt;
 FOp:= prmOp;
end;

class function TRuleCondition.generateFrom(condition: string;
  lstVar: TLstAttributes): TRuleCondition;
var nomVar: string;
    sValue: string;
    sComp: string;
    lenComp: integer;
    rComp: TTypeConditionContinue;
    value: TTypeContinue;

    att: TAttribute;
    rc: TRuleCondition;
    p: integer;

    procedure addValueDiscrete(rcd: TRuleCondDiscrete; chaine: string);
    var dValue: TTypeDiscrete;
    begin
     dValue:= att.LstValues.isValueAvailable(chaine);
     if (dValue > 0)
      then rcd.addValue(dValue);
    end;

begin
 result:= nil;
 //tester la forme de la condition
 p:= pos('_IN_',condition);
 if (p > 0)
  then
   //condition discrte
   begin
    nomVar:= copy(condition,1,pred(p));
    //rechercher la variable dans la liste
    att:= lstVar.GetFromName(nomVar);
    if (att <> nil)
     then
      begin
       rc:= TRuleCondDiscrete.Create(att,ord(rcdIn));
       condition:= copy(condition,p + 4,length(condition));
       //virer les crochets "[" et "]"
       condition:= copy(condition,2,length(condition)-2);
       //lire les valeurs
       p:= pos(';',condition);
       while (p > 0) do
        begin
         sValue:= copy(condition,1,pred(p));
         addValueDiscrete(rc as TRuleCondDiscrete,sValue);
         condition:= copy(condition,succ(p),length(condition));
         p:= pos(';',condition);
        end;
       if (length(condition) > 0) then addValueDiscrete(rc as TRuleCondDiscrete,condition);
       //et donc...
       if not((rc as TRuleCondDiscrete).isEmpty())
        then result:= rc
        else rc.Free();
      end;
   end
  else
   //condition continue
   begin
    //rcuprer l'oprateur de comparaison
    //on commence par le plus long pour viter les confusions
    lenComp:= 2;
    p:= pos('<=',condition);
    if (p = 0) then p:= pos('<>',condition);
    if (p = 0) then p:= pos('>=',condition);
    if (p = 0) then lenComp:= 1;
    if (p = 0) then p:= pos('<',condition);
    if (p = 0) then p:= pos('>',condition);
    if (p = 0) then p:= pos('=',condition);
    if (p > 0)
     then
      begin
       //nom de variable
       nomVar:= copy(condition,1,pred(p));
       //rechercher la variable dans la liste
       att:= lstVar.GetFromName(nomVar);
       if (att <> nil)
        then
         begin
          condition:= copy(condition,p,length(condition));
          //oprateur de comparaison
          sComp:= copy(condition,1,lenComp);
          //rcuprer la valeur de comparaison
          sValue:= copy(condition,succ(lenComp),length(condition));
          //la condition
          TRY
          rComp:= TRuleCondContinue.getCondFromDescription(sComp);
          value:= strtofloat(sValue);
          rc:= TRuleCondContinue.Create(att,ord(rComp),value);
          result:= rc; 
          EXCEPT
          result:= nil;
          END;
         end;
      end;
   end;
end;

function TRuleCondition.getHTMLDescription: string;
begin
 result:= FAtt.Name;
end;

function TRuleCondition.getTXTDescription(short: boolean = TRUE): string;
begin
 result:= FAtt.Name;
end;

{ TRuleCondDiscrete }

constructor TRuleCondDiscrete.Create(prmAtt: TAttribute;
  prmOp: TTypeConditionOperator; prmValue: TTypeDiscrete);
begin
 inherited Create(prmAtt,prmOp);
 //singleton au dpart
 FSetValues:= [prmValue];
end;

function TRuleCondDiscrete.TestValue(const prmValue: TTypeDiscrete): boolean;
begin
 case TTypeConditionDiscrete(FOp) of
  rcdIn: result:= prmValue in FSetValues;
  //pour les autres cas non-recenss --> refus systmatique
  else
   begin
    //result:= false;
    raise Exception.Create('Unexpected operator for discrete rule condition');
   end;
 end;
end;

procedure TRuleCondDiscrete.Merge(prmOther: TRuleCondition);
begin
 //c'est bien une union
 FSetValues:= FSetValues + (prmOther as TRuleCondDiscrete).SetValues;
end;

function TRuleCondDiscrete.TestValue(const prmValue: TTypeContinue): boolean;
var tmp: TTypeDiscrete;
begin
 tmp:= trunc(prmValue);
 //bon typage donc pas de risque d'appel rcursif
 result:= self.TestValue(tmp);
end;

function TRuleCondDiscrete.Duplicate: TRuleCondition;
var r: TRuleCondDiscrete;
begin
 r:= TRuleCondDiscrete.Create(Attribute,Operator,1);//le "1" c'est pour la forme, il est vir aprs
 r.FSetValues:= self.FSetValues;
 result:= r;
end;

function TRuleCondDiscrete.getHTMLDescription: string;
var s: string;
    i: TTypeDiscrete;
begin
 s:= inherited getHTMLDescription();
 s:= s+' in [';
 for i:= 1 to Attribute.nbValues do
  if (i in FSetValues)
   then s:= s+Attribute.LstValues.getDescription(i)+',';
 s:= copy(s,1,pred(length(s)));
 s:= s+']';
 result:= s;
end;

function TRuleCondDiscrete.getTXTDescription(short: boolean = TRUE): string;
var s: string;
    i: TTypeDiscrete;
begin
 //WARNING -- si description courte, on considre que le cardinal est gal  1 toujours, a peut ne pas tre le cas dans certaines versions de l'arbre de dcision
 s:= inherited getTXTDescription(short);
 if short
  then
   begin
    s:= s+' = ';
    for i:= 1 to Attribute.nbValues do
     if (i in FSetValues)
      then
       begin
        s:= s+Attribute.LstValues.getDescription(i)+',';
        break;
       end;
   end
  else
   begin
    s:= s+' in [';
    for i:= 1 to Attribute.nbValues do
     if (i in FSetValues)
      then s:= s+Attribute.LstValues.getDescription(i)+',';
    s:= s+']';
   end;
 result:= s;
end;

function TRuleCondDiscrete.TestExample(const example: integer): boolean;
begin
 result:= self.TestValue(self.FAtt.dValue[example]);
end;

procedure TRuleCondDiscrete.clear;
begin
 FSetValues:= [];
end;

constructor TRuleCondDiscrete.Create(prmAtt: TAttribute;
  prmOp: TTypeConditionOperator);
begin
 inherited Create(prmAtt,prmOp);
 //ensemble vide au dpart
 FSetValues:= [];
end;

procedure TRuleCondDiscrete.addValue(value: TTypeDiscrete);
begin
 FSetValues:= FSetValues + [value];
end;

function TRuleCondDiscrete.isEmpty: boolean;
begin
 result:= (FSetValues = []);
end;

procedure TRuleCondDiscrete.intersection(other: TRuleCondDiscrete);
begin
 //uniquement s'il s'agit de la mme variable bien sr...
 if (self.Attribute.Name = other.Attribute.Name)
  then self.FSetValues:= self.FSetValues * other.FSetValues;
end;

{ TRuleCondContinue }

constructor TRuleCondContinue.Create(prmAtt: TAttribute;
  prmOp: TTypeConditionOperator; prmValue: TTypeContinue);
begin
 inherited Create(prmAtt,prmOp);
 FThresoldValue:= prmValue;
end;

function TRuleCondContinue.TestValue(const prmValue: TTypeDiscrete): boolean;
var tmp: TTypeContinue;
begin
 tmp:= prmValue;
 //bon typage
 result:= self.TestValue(tmp);
end;

procedure TRuleCondContinue.Merge(prmOther: TRuleCondition);
var other: TRuleCondContinue;
begin
 other:= prmOther as TRuleCondContinue;
 //<, on prend le min
 if (self.Operator = 1) and (other.Operator = 1) and (other.ThresoldValue<self.ThresoldValue)
  then self.FThresoldValue:= other.ThresoldValue;
 //>=, on prend le max
 if (self.Operator = 2) and (other.Operator = 2) and (other.ThresoldValue>self.ThresoldValue)
  then self.FThresoldValue:= other.ThresoldValue;
end;

function TRuleCondContinue.TestValue(const prmValue: TTypeContinue): boolean;
begin
 case TTypeConditionContinue(Operator) of
  rccL:  result:= (prmValue < FThresoldValue);
  rccLE: result:= (prmValue <= FThresoldValue);
  rccE:  result:= (prmValue = FThresoldValue);
  rccNE: result:= (prmValue <> FThresoldValue);
  rccGE: result:= (prmValue >= FThresoldValue);
  rccG:  result:= (prmValue > FThresoldValue);
  //tous les autres cas non-recenss : false
  else
   begin
    //result:= false;
    raise Exception.Create('Unexpected operator for continuous rule condition');
   end;
 end;
end;

function TRuleCondContinue.Duplicate: TRuleCondition;
begin
 result:= TRuleCondContinue.Create(Attribute,Operator,FThresoldValue);
end;

procedure TRuleCondContinue.SetNewThresold(prmValue: TTypeContinue);
begin
 FThresoldValue:= prmValue;
end;

function TRuleCondContinue.getHTMLDescription: string;
var s: string;
begin
 s:= inherited getHTMLDescription();
 s:= s + ' ' + STR_RULE_CONDITION_CONTINUE[TTypeConditionContinue(Operator)] + ' ';
 s:= s+format(STR_FORMAT_VIEW_STAT_ACCURACY,[FThresoldValue]);
 result:= s;
end;

function TRuleCondContinue.getTXTDescription(short: boolean): string;
var s: string;
begin
 s:= inherited getTXTDescription(short);
 s:= s + ' ' + STR_RULE_CONDITION_CONTINUE[TTypeConditionContinue(Operator)] + ' ';
 s:= s+format(STR_FORMAT_VIEW_STAT_ACCURACY,[FThresoldValue]);
 result:= s;
end;

function TRuleCondContinue.TestExample(const example: integer): boolean;
begin
 result:= self.TestValue(self.FAtt.cValue[example]);
end;

class function TRuleCondContinue.getCondFromDescription(
  description: string): TTypeConditionContinue;
var k: TTypeConditionContinue;
begin
 //par dfaut '<'
 result:= TTypeConditionContinue(0);
 //recherche
 for k:= low(TTypeConditionContinue) to high(TTypeConditionContinue) do
  if (description = STR_RULE_CONDITION_CONTINUE[TTypeConditionContinue(k)])
   then
    begin
     result:= TTypeConditionContinue(k);
     break;
    end;
end;

end.
