(********************************************************************)
(* UCalcStatDesCrossTab.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Tableaux croiss, avec les indicateurs stats associs)
@author(Ricco)
@created(12/01/2004)

03/05/2007 -- Ajout le calcul de la corrlation  la R.R. et S.L. cf. Article PKDD'2000.
09/05/2007 -- Calcul du Tau de Goodman & Kruskal, coef. d'association asymtrique
17/05/2007 -- Calcul du Lambda de Goodman & Kruskal, et du U (Uncertainty Coefficient) de Theil
}
unit UCalcStatDesCrossTab;

interface

USES
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcStatDes,
        UCalcCrossTab,
        UCalcRndGenerator;

TYPE
        {type de statistique envoye par le tableau crois}
        TEnumTypeStatFromCrossTab = (stCxTschuprow,stCxSU,stCxRandom);

CONST
        STR_STAT_NAME_FROM_CROSS_TAB : array[TEnumTypeStatFromCrossTab] of String = ('Tschuprow','Symmetrical uncertainty','Random ranking');

TYPE
        {classe de calcul sur les tableaux croiss}
        TCalcSDCrossTab = class(TCalcStatDes)
                          private
                          {tableau de comptage}
                          FCrossTab: TCrossTab;
                          {l'attribut supplmentaire}
                          FColAtt: TAttribute;
                          {les statistiques - chi2 et ses drivs}
                          FTschuprow,FCramer,FPhi2,FChi2: double;
                          {proba critique chi2}
                          FPrChi2: double;
                          {degr de libert}
                          Fddl: integer;
                          {Symmetrical uncertainty}
                          FSU: double;
                          {Random value attribution}
                          FRandom: double;
                          {Corrlation}
                          FCorrelation, FZCorr, FPrCorr: double;
                          {Tau de Goodman & Kruskall et proba critique associe}
                          FTau, FchiTau, FprTau: double;
                          {U, Chi2U,proba,sigU,sig_H0_U,zScoreU}
                          FU, FChi2U, FprU, FSigmaU: double;
                          {Lambda, Goodman et Kruskal}
                          FLambda, FSigmaLambda: double;
                          //new -- 13/11/2009 -- calcul des degrs de libert (tenir compte des marges vides)
                          procedure   calcDDL();
                          {rcuprer l'attribut en ligne}
                          function    GetRowAtt(): TAttribute;
                          {calculer les indicateus statistiques}
                          procedure   CalcStats();
                          procedure   calc_SU();
                          //new -- 03/05/2007 -- corrlation pour variables catgorielles
                          procedure   calcCorrelation();
                          //new -- 04/05/2007 -- Goodman & Kruskal's Tau (gain normalis du Gini quoi), new formula 09/05/2007
                          procedure   calcTau();
                          //new -- 09/05/2007 -- Theil'U (Uncertainty Coefficient, i.e. gain d'entropie normalis) -- attention, pas de Z-score car on n'a pas l'estimation de la variance sous H0
                          procedure   calcTheilU();
                          //new -- 17/05/2007 -- Lambda de Goodman et Kruskal
                          procedure   calcLambda();
                          protected
                          procedure   BeginUpdate(); override;
                          procedure   AddValue(prmExample: integer); override;
                          procedure   EndUpdate(); override;
                          public
                          {constructeur}
                          constructor Create(prmR,prmC: TAttribute; prmExamples: TExamples = nil; prmRndGenerator: TRndGenerator = nil);
                          destructor  Destroy(); override;
                          function    getHTMLResult(prmOption: integer = -1): string; override;
                          {renvoyer la statistique demande}
                          function    statisticValue(prmTypeStat: TEnumTypeStatFromCrossTab): double;
                          {renvoyer la p-value de la stat demande}
                          function    pValue(prmTypeStat: TEnumTypeStatFromCrossTab): double;
                          {pointeur sur le tableau crois}
                          property   CrossTab: TCrossTab read FCrossTab;
                          {pointeur sur l'attribut en ligne}
                          property   RowAtt: TAttribute read getRowAtt;
                          {pointeur sur l'attribut en colonne}
                          property   ColAtt: TAttribute read FColAtt;
                          {le t-de Tschuprow}
                          property   Tschuprow: double read FTschuprow;
                          {le v de Cramer}
                          property   Cramer: double read FCramer;
                          {le Phi-2}
                          property   Phi2: double read FPhi2;
                          {le chi-2}
                          property   Chi2: double read FChi2;
                          {la proba associe au test du chi-2}
                          property   ProbaChi2: double read FPrChi2;
                          {Symmetrical uncertainty -- new 31/05/2004}
                          property   SU: double read FSU;
                          {Correlation pour variable catgorielle}
                          property   Correlation: double read FCorrelation;
                          {Z-score corrlation}
                          property   ZScoreCorrelation: double read FZCorr;
                          {p-value corrlation}
                          property   prCorrelation: double read FPrCorr;
                          {Random value attribution}
                          property   Random: double read FRandom;
                          {degr de libert}
                          property   DDL: integer read FDDL;
                          {Tau}
                          property   Tau: double read FTau;
                          {Chi-2 du Tau, cf. Light & Margolin}
                          property   chiTau: double read FchiTau;
                          {proba-critique Tau}
                          property   prTau: double read FprTau;
                          {U de Theil, asymtrique}
                          property   U: double read FU;
                          {Chi-2 du U}
                          property   chiU: double read FChi2U;
                          {proba critique}
                          property    prU: double read FprU;
                          {sigma de U}
                          property    sigmaU:double read FSigmaU;
                          {Lambda}
                          property    Lambda: double read FLambda;
                          {Asymp. Sigma Lambda}
                          property    sigmaLambda: double read FSigmaLambda;
                          end;

        {liste de stats tableaux croiss, utile notamment pour les tris}
        TLstCalcStatDesCrossTab = class(TLstCalcStatDes)
                                  public
                                  {reconstruction de la liste de stat des}
                                  procedure   RebuildStatDes(prmLstAtt: TLstAttributes; prmExamples: TExamples); override;
                                  {comparaison pour les stats conditionnelles}
                                  procedure SortStats(); override;
                                  {description HTML des rsultats}
                                  function    getHeaderHTML(): string; override;
                                  end;

        {liste des stats tableaux croiss, mais sous forme de corrlation cette fois-ci}
        TLstCalcSDCrossTabCorrelation = class(TLstCalcStatDesCrossTab)
        public
        {comparaison pour les stats conditionnelles}
        procedure SortStats(); override;
        end;

        {liste sous forme de tau de Goodman & Kruskal}
        TLstCalcSDCrossTabTau = class(TLstCalcStatDesCrossTab)
        public
        {comparaison pour les stats conditionnelles}
        procedure SortStats(); override;
        end;

        {liste sous forme de U de Theil}
        TLstCalcSDCrossTabTheil = class(TLstCalcStatDesCrossTab)
        public
        {comparaison pour les stats conditionnelles}
        procedure SortStats(); override;
        end;

        {liste sous forme de Lambda de Goodman & Kruskal}
        TLstCalcSDCrossTabLambda = class(TLstCalcStatDesCrossTab)
        public
        {comparaison pour les stats conditionnelles}
        procedure SortStats(); override;
        end;

        

implementation

USES
        Sysutils, Classes,
        MATH, FMath, UConstConfiguration, ULogFile;

{ TCalcSDCrossTab }

procedure TCalcSDCrossTab.AddValue(prmExample: integer);
begin
 FCrossTab.AddValue(prmExample);
 inc(FNbExamples);
end;

procedure TCalcSDCrossTab.BeginUpdate;
begin
 FCrossTab.ReInitialization();
 FNbExamples:= 0;
end;

procedure TCalcSDCrossTab.calcCorrelation;
var g11,g12,g21,g22,n: Double;
    k,l: Integer;
    value,denom,r2,v: double;
Begin
 n:= CrossTab.Value[0,0];
 //variance est constante
 v:= 1.0/(-3.0 + n);
 if (v > 0.0) then v:= SQRT(v); //passage  l'cart-type
 //g:= n*(n-1);
 //le tableau n'est plus 2x2 !!!
 g11:= 0; g12:=0; g21:= 0; g22:= 0;
 For k:= 1 To CrossTab.RowCount Do
  For l:= 1 To CrossTab.ColCount Do
   Begin
    //g11:= g11+Tableau[k,l]*(Tableau[k,l]-1.0);
    //modif. tirage avec remise pour r(Y;X)=0 <=> r(Iy,Ix)=0
    //20/05/2004 -- on divise par n pour viter le pbm des gros effectifs
    g11:= g11+(CrossTab.Value[k,l]/n)*(CrossTab.Value[k,l]/n);
    g22:= g22+(CrossTab.Value[k,l]/n)*((n-CrossTab.Value[k,0]-CrossTab.Value[0,l]+CrossTab.Value[k,l]))/n;
    g12:= g12+(CrossTab.Value[k,l]/n)*((CrossTab.Value[k,0]-CrossTab.Value[k,l]))/n;
    g21:= g21+(CrossTab.Value[k,l]/n)*((CrossTab.Value[0,l]-CrossTab.Value[k,l]))/n;
   End;
 //rsultat
 value:= (g11*g22-g12*g21);
 denom:= (g11+g12)*(g21+g22)*(g11+g21)*(g12+g22);
 if (denom>EPSILON_VALUE)
  then FCorrelation:= value/SQRT(denom)
  else FCorrelation:= 0.0;
 //z-score et proba critique
 r2:= SQR(FCorrelation);
 if (r2 < 1.0) and (v > 0.0)
  then
   begin
    FZcorr:= 0.5 * ln ((1.0 + FCorrelation)/(1.0 - FCorrelation));
    FZcorr:= FZcorr / v; //z-score
    FPrCorr:= PNorm(FZcorr);
   end
  else
   begin
    FZcorr:= 0.0;
    FPrCorr:= 1.0;
   end;
end;

procedure TCalcSDCrossTab.calcDDL;
var nbLig,nbCol,k: integer;
begin
 //principe --> si une marge est vide, il ne faut pas en tenir compte
 nbLig:= 0;
 for k:= 1 to CrossTab.RowCount do if (CrossTab.Value[k,0] > 0) then inc(nbLig);
 nbCol:= 0;
 for k:= 1 to CrossTab.ColCount do if (CrossTab.Value[0,k] > 0) then inc(nbCol);
 //puis les degrs de libert
 Fddl:= (nbLig - 1) * (nbCol - 1);
end;

procedure TCalcSDCrossTab.calcLambda;
 //chercher un max dans une colonne, et renvoie sa valeur et le numro de ligne
 procedure getMaxRow(column: integer; var idRow: integer; var value: integer);
 var k: integer;
 begin
  idRow:= -1;
  value:= -1;
  for k:= 1 to CrossTab.RowCount do
   if (CrossTab.Value[k,column] > value)
    then
     begin
      idRow:= k;
      value:= CrossTab.Value[k,column];
     end;
 end;
var n,sAll,sMax : double;
    maxR: double;
    iMax,i,j: integer;
    value: integer;
begin
 n:= 1.0*CrossTab.Value[0,0];
 getMaxRow(0,iMax,value); maxR:= 1.0 * value;
 sAll:= 0.0; sMax:= 0.0;
 for j:= 1 to CrossTab.ColCount do
  begin
   getMaxRow(j,i,value);
   sAll:= sAll + 1.0 * value;
   if (i = iMax) then sMax:= sMax + 1.0 * value;
  end;
 FLambda:= 0.0;
 FSigmaLambda:= 0.0;
 if (n > maxR)
  then
   begin
    FLambda:= (sAll - maxR)/(n - maxR);
    FSigmaLambda:= (n - sAll) * (sAll + maxR - 2.0 * sMax) / Math.Power(n - maxR,3.0);
    if (FSigmaLambda > 0.0) then FSigmaLambda:= sqrt(FSigmaLambda);
   end;
end;

procedure TCalcSDCrossTab.CalcStats;
var i,j: integer;
    d2,v: double;
begin
 //Fddl:= pred(RowAtt.nbValues)*pred(ColAtt.nbValues);
 //new -- 13/11/2009 -- calcul des degrs de libert tenant compte des marges vides...
 self.calcDDL();
 //suite...
 d2:= 0.0;
 for i:= 1 to CrossTab.RowCount do
  begin
   for j:= 1 to CrossTab.ColCount do
    begin
     v:= 1.0*CrossTab.Value[i,0]*CrossTab.Value[0,j]/(1.0*CrossTab.Value[0,0]);
     if (v>0)
      then d2:= d2+1.0*SQR(1.0*CrossTab.Value[i,j]-v)/v;
    end;
  end;
 //rcuprer les infos
 FChi2:= d2;
 FPhi2:= d2/(1.0*CrossTab.Value[0,0]);
 FTschuprow:= 0.0;
 if (Fddl>0)
  then FTschuprow:= sqrt(d2/(1.0*CrossTab.Value[0,0]*sqrt(1.0*Fddl)));
 FCramer:= 0.0;
 if (MIN(pred(RowAtt.nbValues),pred(ColAtt.nbValues))>0)
  then FCramer:= sqrt(d2/(1.0*CrossTab.Value[0,0]*MIN(pred(RowAtt.nbValues),pred(ColAtt.nbValues))));
 FPrChi2:= 1.0;
 if (Fddl>0)
  then FPrChi2:= PKHI2(Fddl,FChi2);
 //nouveaux calculs
 self.calc_SU();
 self.calcCorrelation();
 self.calcTau();
 self.calcTheilU();
 self.calcLambda();
end;

procedure TCalcSDCrossTab.calcTau;
var WSS, TSS, n: double;
    i,j: integer;
begin
 n:= CrossTab.Value[0,0];
 //on pourrait passer par un gain de gini normalis
 //--> a priori, la formule suivante devrait donner les mmes rsultats, cf.Ritschard et al., Cahier Dp. Economtrie, pp.5
 TSS:= 0.0;
 for i:= 1 to CrossTab.RowCount do TSS:= TSS + Math.Power(1.0*CrossTab.Value[i,0]/n,2.0);
 WSS:= 0.0;
 for i:= 1 to CrossTab.RowCount do
  for j:= 1 to CrossTab.ColCount do
   if (CrossTab.Value[0,j] > 0.0) then WSS:= WSS + Math.Power(CrossTab.Value[i,j],2.0)/(n * CrossTab.Value[0,j]);
 //stat.
 FTau:= 0.0;
 FchiTau:= 0.0;
 FprTau:= 1.0;
 //gain normalis
 if (TSS < 1.0)
  then
   begin
    FTau:= (WSS-TSS)/(1.0-TSS);
    FchiTau:= (n - 1.0) * (-1.0 + CrossTab.RowCount) * FTau;
    FprTau:= PKHI2(Fddl,FchiTau);
   end;
 (* --> very complicated, et la formule n'est pas trs assure,  revoir....
 //calcul du ZScore de Tau
 FZTau:= 0.0;
 FSigmaTau:= 0.0;
 delta:= 1.0 - TSS;
 if (delta > 0.0)
  then
   begin
    conj:= 1.0 - WSS;
    phibar:= 0.0;
    //calcul de Phi
    setLength(phi,CrossTab.RowCount+1,CrossTab.ColCount+1);
    for j:= 1 to CrossTab.ColCount do
     begin
      if (CrossTab.Value[0,j] > 0.0)
       then
        begin
         tmp:= 0.0;
         for iprim:= 1 to CrossTab.RowCount do tmp:= tmp + power(CrossTab.Value[iprim,j]/CrossTab.Value[0,j],2.0);
         for i:= 1 to CrossTab.RowCount do
          begin
           phi[i,j]:= 2.0*CrossTab.FullFreq[i,0]*conj + delta*(2.0*CrossTab.Value[i,j]/CrossTab.Value[0,j]-tmp);//(eq.37)
           phibar:= phibar + CrossTab.FullFreq[i,j] * phi[i,j];
          end;
        end;
     end;
    //calcul de la variance
    variance:= 0.0;
    for i:= 1 to CrossTab.RowCount do
     for j:= 1 to CrossTab.ColCount do
      variance:= variance + CrossTab.FullFreq[i,j] * power(phi[i,j]-phibar,2.0);
    variance:= (1.0/(n*power(delta,4.0)))*variance; //(eq.35)
    finalize(phi);//plus besoin du tableau phi ici...
    //le z
    if (variance > 0.0)
     then
      begin
       FSigmaTau:= SQRT(variance);
       FZTau:= FTau/FSigmaTau;
      end;
   end;
 *)
end;

procedure TCalcSDCrossTab.calcTheilU;
var i,j: integer;
    sX,sY,sXY,value: double;
    variance: double;
begin
 //FU, FChi2U, FprU, FSigU, FSigHU, FZScoreU: double;
 FU:= 0.0;
 FChi2U:= 0.0;
 FPrU:= 1.0;
 FSigmaU:= 0.0;
 //>>> hypothse de base ici, le tableau ne peut pas tre vide
 if (CrossTab.Value[0,0] = 0) then Exit;
 //sinon, on reste et zoo...
 //marge X (colonne)
 sX:= 0.0;
 for j:= 1 to CrossTab.ColCount do
  begin
   value:= CrossTab.Value[0,j]/CrossTab.Value[0,0];
   if (value > 0) then sX:= sX-value*log2(value);
  end;
 //marge Y
 sY:= 0.0;
 for i:= 1 to CrossTab.RowCount do
  begin
   value:= CrossTab.Value[i,0]/CrossTab.Value[0,0];
   if (value > 0) then sY:= sY-value*log2(value);
  end;
 //croise
 sXY:= 0.0;
 for i:= 1 to CrossTab.RowCount do
  for j:= 1 to CrossTab.ColCount do
   begin
    value:= CrossTab.FullFreq[i,j];
    if (value > 0) then sXY:= sXY-value*log2(value);
   end;
 //Uncertainty Coefficients
 if (sY > 0.0) then FU:= (sY + sX - sXY)/sY;
 //Chi-2
 FChi2U:= 2.0 * CrossTab.Value[0,0] * LN(2.0) * sY * FU;//Thse R.R. (report de Wehenkel d'une part ; White et Liu d'autre part ; Ouvrage Saporta)
 if (FChi2U > 0.0) then FprU:= PKHI2(Fddl,FChi2U) else FprU:= 1.0;
 //variance
 variance:= 0.0;
 for i:= 1 to CrossTab.RowCount do
  for j:= 1 to CrossTab.ColCount do
   begin
    if (CrossTab.Value[i,j] > 0) and (CrossTab.Value[i,0] > 0)
     then variance:= variance + CrossTab.Value[i,j]*SQR(sY*log2(CrossTab.Value[i,j]/CrossTab.Value[0,j])+(sX-sXY)*log2(CrossTab.Value[i,0]/CrossTab.Value[0,0]));
   end;
 if (sY > 0.0) and (variance > 0.0) then variance:= 1.0/(SQR(CrossTab.Value[0,0])*power(sY,4.0))*variance;
 //new -- 02/03/2008 -- correction bug
 //FSigmaU:= sqrt(variance);
 if (variance > 0.0) then FSigmaU:= sqrt(variance) else FSigmaU:= 0.0;
 //logfile
 //TraceLog.WriteToLogFile(format('[THEIL-U] --> %.6f, %.6f, %.6f, %.6f',[FU,FChi2U,FprU,FSigmaU]));
end;


procedure TCalcSDCrossTab.calc_SU();
var i,j: integer;
    S,sX,sY,sXY,value: double;
begin
 //>>> hypothse de base ici, le tableau ne peut pas tre vide
 if (CrossTab.Value[0,0]=0)
  then FSU:= 0
  else
   begin
     //marge X (colonne)
     sX:= 0.0;
     for j:= 1 to CrossTab.ColCount do
      begin
       value:= CrossTab.Value[0,j]/CrossTab.Value[0,0];
       if (value>0)
        then sX:= sX-value*log2(value);
      end;
     //marge Y
     sY:= 0.0;
     for i:= 1 to CrossTab.RowCount do
      begin
       value:= CrossTab.Value[i,0]/CrossTab.Value[0,0];
       if (value>0)
        then sY:= sY-value*log2(value);
      end;
     //croise
     sXY:= 0.0;
     for i:= 1 to CrossTab.RowCount do
      begin
       for j:= 1 to CrossTab.ColCount do
        begin
         value:= CrossTab.FullFreq[i,j];
         if (value>0) then sXY:= sXY-value*log2(value);
        end;
      end;
     //corrlation SU (Symmetrical uncertainty)
     if ((sX+sY)>0)
      then
       begin
        S:= 2.0*(sX+sY-sXY)/(sX+sY);
        FSU:= S;
        //TraceLog.WriteToLogFile(format('SU --> %.6f',[SU]));
       end
      else FSU:= 0.0;
   end;
end;

constructor TCalcSDCrossTab.Create(prmR, prmC: TAttribute;
  prmExamples: TExamples = nil; prmRndGenerator: TRndGenerator = nil);
begin
 FColAtt:= prmC;
 FCrossTab:= TCrossTab.create(prmR,prmC);
 if assigned(prmRndGenerator)
  then FRandom:= prmRndGenerator.RanMar()
  else FRandom:= random;//de toute manire on ne s'en sert pas dans ce cas...
 inherited Create(prmR,prmExamples);
end;

destructor TCalcSDCrossTab.Destroy;
begin
 FCrossTab.Free();
 inherited Destroy();
end;

procedure TCalcSDCrossTab.EndUpdate;
begin
 CalcStats();
end;

function TCalcSDCrossTab.getHTMLResult(prmOption: integer): string;
var s: string;
begin
 s:= format('<TD>%s</TD><TD>%s</TD>',[self.RowAtt.Name,self.ColAtt.Name]);
 //les indicateurs stats
 s:= s+'<TD valign=top>'+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_BLUE+
       '<TH width=80>Stat</TH><TH width=60>Value</TH></TR>';
 //degr de libert
 s:= s+format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>d.f.</TD><TD align=right>%d</TD></TR>',[self.DDL]);
 //t de tschuprow
 s:= s+format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>Tschuprow''s t</TD><TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD></TR>',[self.Tschuprow]);
 //v de Cramer
 s:= s+format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>Cramer''s v</TD><TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD></TR>',[self.Cramer]);
 //phi-2
 s:= s+format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>Phi</TD><TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD></TR>',[self.Phi2]);
 //chi-2 et proba
 s:= s+format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>Chi (p-value)</TD><TD align=right>%.2f (%.4f)</TD></TR>',[self.Chi2,self.ProbaChi2]);
 //la corrlation -- 03/05/2007
 //s:= s + format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>r (p-value)</TD><TD align=right>%.4f (%.4f)</TD></TR>',[self.Correlation,self.prCorrelation]);
 //Lambda de Goodman & Kruskal -- 17/05/2007
 s:= s + format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>Lambda</TD><TD align=right>%.6f</TD></TR>',[FLambda]);
 //le tau de Goodman & Kruskal -- 04/05/2007
 s:= s + format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>Tau (p-value)</TD><TD align=right>%.4f (%.4f)</TD></TR>',[FTau,FprTau]);
 //U de Theil (asymtrique) -- 09/05/2007
 s:= s + format(HTML_TABLE_COLOR_DATA_BLUE+'<TD>U(R/C) (p-value)</TD><TD align=right>%.4f (%.4f)</TD></TR>',[FU,FprU]);
 //fin des indicateurs
 s:= s+'</table></TD>';
 
 //si l'option est non utilise, on envoie la sauce standard
 if (prmOption >= 0)
  then s:= s+'<TD valign=top>'+self.CrossTab.getHTMLResult(prmOption,'')+'</TD>';

 result:= s;
end;

function TCalcSDCrossTab.GetRowAtt: TAttribute;
begin
 result:= Attribute;
end;

function TCalcSDCrossTab.pValue(
  prmTypeStat: TEnumTypeStatFromCrossTab): double;
begin
 //toujours celui du chi-2 pour l'instant
 result:= self.ProbaChi2;
end;

function TCalcSDCrossTab.statisticValue(
  prmTypeStat: TEnumTypeStatFromCrossTab): double;
begin
 case prmTypeStat of
  stCxSU: result:= self.SU;
  stCxRandom: result:= self.Random
  else result:= self.Tschuprow;
 end;
end;

{ TLstCalcStatDesCrossTab }

function TLstCalcStatDesCrossTab.getHeaderHTML: string;
begin
 result:=  HTML_TABLE_COLOR_HEADER_GRAY+
           '<TH>Row (Y)</TH><TH>Column (X)</TH><TH>Statistical indicator</TH><TH>Cross-tab</TH>';
end;

procedure TLstCalcStatDesCrossTab.RebuildStatDes(prmLstAtt: TLstAttributes;
  prmExamples: TExamples);
var stat: TCalcSDCrossTab;
    i,j: integer;
    attR,attC: TAttribute;    
begin
 //pour chaque attribut de target
 for i:= 0 to prmLstAtt.Count-2 do
  begin
   attR:= prmLstAtt.Attribute[i];
   if attR.isCategory(caDiscrete)
    then
     begin
      //pour chaque attribut de input
      for j:= succ(i) to pred(prmLstAtt.Count) do
       begin
        attC:= prmLstAtt.Attribute[j];
        if attC.isCategory(caDiscrete) and (attR<>attC)
         then
          begin
           stat:= TCalcSDCrossTab.Create(attR,attC,prmExamples);
           self.AddStat(stat);
          end;
       end;
     end;
  end;
end;

{trier selon le nom des attributs en ligne}
function ListSortCompareRowAttName(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 if (st1.RowAtt.Name<st2.RowAtt.Name)
  then result:= -1
  else
   if (st1.RowAtt.Name>st2.RowAtt.Name)
    then result:= +1
    else result:= 0;
end;

{trier selon le nom des attributs en colonne}
function ListSortCompareColumnAttName(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 if (st1.ColAtt.Name<st2.ColAtt.Name)
  then result:= -1
  else
   if (st1.ColAtt.Name>st2.ColAtt.Name)
    then result:= +1
    else result:= 0;
end;

{tri descendant sur le t de Tschuprow}
function ListSortCompareTschuprow(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 //tri invers !!!
 if (st1.Tschuprow<st2.Tschuprow)
  then result:= +1
  else
   if (st1.Tschuprow>st2.Tschuprow)
    then result:= -1
    else result:= 0;
end;

{tri descendant sur le SU}
function ListSortCompareSU(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 //tri invers !!!
 if (st1.SU<st2.SU)
  then result:= +1
  else
   if (st1.SU>st2.SU)
    then result:= -1
    else result:= 0;
end;

{tri alatoire pour la slection alatoire}
function ListSortCompareRandom(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 //tri invers !!!
 if (st1.Random<st2.Random)
  then result:= +1
  else
   if (st1.Random>st2.Random)
    then result:= -1
    else result:= 0;
end;

{tri descendant pour la corrlation}
function ListSortCompareCorrelation(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 //tri invers !!!
 if (st1.Correlation<st2.Correlation)
  then result:= +1
  else
   if (st1.Correlation>st2.Correlation)
    then result:= -1
    else result:= 0;
end;

{tri descendant pour la valeur absolue de la corrlation}
function ListSortCompareAbsCorrelation(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 //tri invers !!!
 if (abs(st1.Correlation) < abs(st2.Correlation))
  then result:= +1
  else
   if (abs(st1.Correlation) > abs(st2.Correlation))
    then result:= -1
    else result:= 0;
end;

{tri descendant pour la corrlation}
function ListSortCompareTau(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 //tri invers !!!
 if (st1.Tau<st2.Tau)
  then result:= +1
  else
   if (st1.Tau>st2.Tau)
    then result:= -1
    else result:= 0;
end;

{tri descendant pour la corrlation}
function ListSortCompareTheil(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 //tri invers !!!
 if (st1.U<st2.U)
  then result:= +1
  else
   if (st1.U>st2.U)
    then result:= -1
    else result:= 0;
end;

{tri descendant pour la corrlation}
function ListSortCompareLambda(item1,item2: pointer): integer;
var st1,st2: TCalcSDCrossTab;
begin
 st1:= TCalcSDCrossTab(item1);
 st2:= TCalcSDCrossTab(item2);
 //tri invers !!!
 if (st1.Lambda<st2.Lambda)
  then result:= +1
  else
   if (st1.Lambda>st2.Lambda)
    then result:= -1
    else result:= 0;
end;

procedure TLstCalcStatDesCrossTab.SortStats;
var funcCompare: TListSortCompare;
begin
 if (CompareMode>=0)
  then
   begin

    case CompareMode of
     0: funcCompare:= ListSortCompareRowAttName;
     1: funcCompare:= ListSortCompareColumnAttName;
     2: funcCompare:= ListSortCompareTschuprow;
     3: funcCompare:= ListSortCompareSU;
     4: funcCompare:= ListSortCompareRandom;
     else
      funcCompare:= NIL;
    end;

    if assigned(funcCompare)
     then LstStat.Sort(funcCompare);
   end;
end;

{ TLstCalcSDCrossTabCorrelation }

procedure TLstCalcSDCrossTabCorrelation.SortStats;
var funcCompare: TListSortCompare;
begin
 if (CompareMode>=0)
  then
   begin

    case CompareMode of
     0: funcCompare:= ListSortCompareRowAttName;
     1: funcCompare:= ListSortCompareColumnAttName;
     2: funcCompare:= ListSortCompareCorrelation;
     3: funcCompare:= ListSortCompareAbsCorrelation;
     else
      funcCompare:= NIL;
    end;

    if assigned(funcCompare)
     then LstStat.Sort(funcCompare);
   end;
end;

{ TLstCalcSDCrossTabTau }

procedure TLstCalcSDCrossTabTau.SortStats;
var funcCompare: TListSortCompare;
begin
 if (CompareMode>=0)
  then
   begin

    case CompareMode of
     0: funcCompare:= ListSortCompareRowAttName;
     1: funcCompare:= ListSortCompareColumnAttName;
     2: funcCompare:= ListSortCompareTau;
     else
      funcCompare:= NIL;
    end;

    if assigned(funcCompare)
     then LstStat.Sort(funcCompare);
   end;
end;

{ TLstCalcSDCrossTabTheil }

procedure TLstCalcSDCrossTabTheil.SortStats;
var funcCompare: TListSortCompare;
begin
 if (CompareMode>=0)
  then
   begin

    case CompareMode of
     0: funcCompare:= ListSortCompareRowAttName;
     1: funcCompare:= ListSortCompareColumnAttName;
     2: funcCompare:= ListSortCompareTheil;
     else
      funcCompare:= NIL;
    end;

    if assigned(funcCompare)
     then LstStat.Sort(funcCompare);
   end;
end;

{ TLstCalcSDCrossTabLambda }

procedure TLstCalcSDCrossTabLambda.SortStats;
var funcCompare: TListSortCompare;
begin
 if (CompareMode>=0)
  then
   begin

    case CompareMode of
     0: funcCompare:= ListSortCompareRowAttName;
     1: funcCompare:= ListSortCompareColumnAttName;
     2: funcCompare:= ListSortCompareLambda;
     else
      funcCompare:= NIL;
    end;

    if assigned(funcCompare)
     then LstStat.Sort(funcCompare);
   end;
end;

end.
