(******************************************************************)
(* UCalcStatDesManova.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Description conditionnelle d'une liste de variables continues  partir d'une variable discrte)
@author(Ricco)
@created(07/08/2005)

Cette classe de base sera utilise pour l'analyse de variance multi-dimensionnelle. Son principal
intrt est qu'il permet de calculer le Lambda de Wilks et les stats associes,  utiliser donc
dans les classes de calcul tel que l'analyse discriminante, etc.

A voir aussi s'il est possible d'intgrer les tests non-paramtriques.

}
unit UCalcStatDesManova;

interface


USES
    UCalcStatDes,
    UDatasetDefinition,
    UDatasetImplementation,
    UDatasetExamples,
    Matrices,
    Contnrs;    

TYPE
    //***************************************************
    //classe abstraite -- dfinit les paramtres  passer
    //***************************************************
    TCalcSDCondMultivariate = class(TCalcStatDes)
    protected
    //new -- 17/04/2010 -- dimension des matrices, pour viter les ambiguts autour de FDescriptors.Count
    FDimension: integer;
    //la liste des attributs  dcrire
    FDescriptors: TLstAttributes;
    //stats conditionnelles (par rapport  la variable groupe) --> liste de liste de stats
    FStatGroups: TObjectList;
    //stat globales sur les descripteurs
    FStatGlobal: TLstCalcStatDesContinuous;
    //tableau des effectifs par groupe
    FTabSizeGroups: array of integer;
    //calculer les stats descriptives marginales
    procedure   computeMultivariateStatDes(prmExamples: TExamples); virtual;
    //calculer les stats descriptives conditionnelles -- le paramtre est une liste de TExamples
    procedure   computeMultivariateCondStatDes(lstExamples: TObjectList); virtual;
    //calculer les statistiques de test
    procedure   computeStatisticalTests(); virtual; abstract;
    public
    //dcrire les stats-conditionnelles en HTML
    function    getHTMLMultivariateStatDes(): string; virtual;
    //dcrire les statistiques de tests associes  la classe
    function    getHTMLStatisticalTests(): string; virtual; abstract;
    //passer les paramtres, et initier les calculs ventuellement
    constructor create(prmD: TAttribute; prmLstAtts: TLstAttributes; prmExamples: TExamples = nil);
    //vider les structures internes
    destructor  destroy(); override;
    //lancer les calculs
    procedure   RefreshStat(prmExamples: TExamples); override;
    //dcrire les rsultats
    function    getHTMLResult(prmOption: integer = -1): string; override;
    //accs  une stat de groupe
    function    getStatGroup(k: integer): TLstCalcStatDesContinuous;
    end;

    //classe de classe
    TClassCalcSDCondMultivariate = class of TCalcSDCondMultivariate;

    //******
    //Manova
    //******
    TCalcSDCondManova = class(TCalcSDCondMultivariate)
    private
    //statistique de Wilks
    FWilks: double;
    //transformation de Bartlett et la p-value associe
    FBartlett,FPValueBartlett: double;
    //ddl. Bartlett
    FDDLBartlett: integer;
    //transformation de Rao
    FRao, FPValueRao: double;
    //ddl. Rao --> on peut ne pas avoir des valeurs entires ! donc on a ici les valeurs tronques
    FDDLNum,FDDLDenom: integer;
    //Stat. de Box pour l'homogneit des matrices de var-covar --> cf. http://www.rrz.uni-hamburg.de/RRZ/Software/SPSS/Algorith.115/app14_boxs_m.pdf (SPSS)
    //FBoxM: double;
    protected
    //matrice de variance-covariance totale
    FT: PMatrix;
    //matrice de variance-covariance intra (sans la pondration finale)
    //a rend l'outil plus gnrique : c'est 1/(n-k) si LDA, 1/n si Wilks, etc.
    FWBrut: PMatrix;
    //calculer la mvcv totale
    procedure   computeMultivariateStatDes(prmExamples: TExamples); override;
    //calculer la mcvc intra
    procedure   computeMultivariateCondStatDes(lstExamples: TObjectList); override;
    //les transformations associes du Wilks
    procedure   computeStatisticalTests(); override;
    public
    //dcrire les statistiques de tests associes  la classe
    function    getHTMLStatisticalTests(): string; override;
    destructor  destroy(); override;
    //calculer le Lambda de Wilks  partir d'un couple de matrices (T et W ) -- attention, les matrices sont chamboules
    function    computeWilks(T,W: PMatrix; size: integer): double;
    //properties
    property    MatT: PMatrix read FT;
    property    MatWBrut: PMatrix read FWBrut;
    property    Wilks: double read FWilks;
    end;

    //**********************************
    //T2 de Hotelling - Homoscdasticit
    //**********************************
    TCalcSDCondT2 = class(TCalcSDCondMultivariate)
    protected
    //cart entre les moyennes, b.basse et b.haute de l'intervalle de confiance de l'cart
    Ecarts,LBound,UBound: array of double;
    //pooled covariance matrix
    FSp: PMatrix;
    //calcul de FSp
    procedure   computeMultivariateCondStatDes(lstExamples: TObjectList); override;
    //calculer la statistique et sa tranformation en Fisher
    procedure   computeStatisticalTests(); override;
    public
    //statistiques
    T2, F, PValue: double;
    //degrs de libert
    Df1, Df2: integer;
    //destructor des matrices locales
    destructor destroy(); override;
    function   getHTMLStatisticalTests(): string; override;
    function   getHTMLMultivariateStatDes(): string; override;
    end;

    //************************************
    //T2 de Hotelling - Htroscdasticit
    //************************************
    TCalcSDCondT2Heteroscedastic = class(TCalcSDCondMultivariate)
    protected
    //cart entre les moyennes, b.basse et b.haute de l'intervalle de confiance de l'cart
    Ecarts,LBound,UBound: array of double;
    //matrice  inverser pour la forme quadratique
    FSt: PMatrix;
    //calcul de FSt
    procedure   computeMultivariateCondStatDes(lstExamples: TObjectList); override;
    //calculer la statistique et sa tranformation en Fisher
    procedure   computeStatisticalTests(); override;
    public
    T2, PValue: double;
    Df: integer;
    destructor destroy(); override;
    function   getHTMLStatisticalTests(): string; override;
    function   getHTMLMultivariateStatDes(): string; override;
    end;

    //*******************************************************************************
    //Test de comparaison de matrices de var-covar - Box's M (ou Bartlett Multivari)
    //*******************************************************************************
    TCalcSDCondBoxM = class(TCalcSDCondMultivariate)
    protected
    //calcul des dterminants
    procedure   computeMultivariateCondStatDes(lstExamples: TObjectList); override;
    //calculer la statistique et sa tranformation en Fisher
    procedure   computeStatisticalTests(); override;
    public
    //statistiques
    T,PValue: double;
    //degrs de libert
    Df: integer;
    function   getHTMLStatisticalTests(): string; override;
    function   getHTMLMultivariateStatDes(): string; override;
    end;

    //liste de statistiques MANOVA -- pas d'option de tri
    TLstCalcSDCondManova = class(TLstCalcStatDes)
    public
    function    getHeaderHTML(): string; override;
    end;

    //liste de T2
    TLstCalcSDCondT2 = class(TLstCalcSDCondManova)
    public
    function    getHeaderHTML(): string; override;
    end;

    //liste de T2 Htroscdastique
    TLstCalcSDCondT2Heteroscedastic = class(TLstCalcSDCondT2)
    end;

    //liste de Box's M
    TLstCalcSDCondBoxM = class(TLstCalcSDCondManova)
    public
    function    getHeaderHTML(): string; override;
    end;

    //classe de classe
    TClassLstCalcSDCondManova = class of TLstCalcSDCondManova;

implementation

uses
    Sysutils,
    UConstConfiguration,
    UCalcMatrixToAttributes, FMath, LOISTAT,
    Math, ULogFile;

CONST
    //niveau de confiance pour les intervalles
    CONF_LEVEL_CI : double = 0.95;

{ TCalcSDCondMultivariate }

procedure TCalcSDCondMultivariate.computeMultivariateCondStatDes(
  lstExamples: TObjectList);
var k: integer;
    stats: TLstCalcStatDesContinuous;
begin
 //pour chaque groupe
 for k:= 0 to pred(lstExamples.Count) do
  begin
   //rcuprer les effectifs par groupes
   FTabSizeGroups[k]:= (lstExamples.Items[k] as TExamples).Size;
   //calculer les stats descriptives par groupes
   stats:= TLstCalcStatDesContinuous.Create(FDescriptors,lstExamples.Items[k] as TExamples);
   FStatGroups.Add(stats);
  end;
end;

procedure TCalcSDCondMultivariate.computeMultivariateStatDes(
  prmExamples: TExamples);
begin
 FStatGlobal.RefreshStat(prmExamples);
end;

constructor TCalcSDCondMultivariate.create(prmD: TAttribute;
  prmLstAtts: TLstAttributes; prmExamples: TExamples);
begin
 inherited Create(prmD);
 TRY
 //dimension des matrices
 FDimension:= prmLstAtts.Count;
 //liste des descripteurs
 FDescriptors:= prmLstAtts;
 //quelques initialisations
 FStatGroups:= TObjectList.Create(TRUE);
 FStatGlobal:= TLstCalcStatDesContinuous.Create(prmLstAtts,NIL);
 setLength(FTabSizeGroups,prmD.nbValues);
 //calculs ventuellement
 if assigned(prmExamples)
  then RefreshStat(prmExamples);
 EXCEPT
 TraceLog.WriteToLogFile('[SDCondMultivariate] Error during creation (constructor)');
 END;
end;

destructor TCalcSDCondMultivariate.destroy;
begin
 TRY
 Finalize(FTabSizeGroups);
 if assigned(FStatGlobal) then FreeAndNil(FStatGlobal);
 if assigned(FStatGroups) then FreeAndNil(FStatGroups);
 inherited;
 EXCEPT
 on e: Exception do
  TraceLog.WriteToLogFile(Format('[%s] destroy, error = %s',[self.ClassName,e.Message]));
 END;
end;

function TCalcSDCondMultivariate.getHTMLMultivariateStatDes: string;
var s: string;
    k,j: integer;
begin
 //dcrire les variables conditionnellement au groupe d'appartenance
 s:= HTML_HEADER_TABLE_RESULT;
 //en-tte avec le nom des modalits (reprsentant les groupes)
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=70>Group</TH>';
 for k:= 1 to Attribute.nbValues do
  s:= s + format('<TH width=70>%s</TH>',[Attribute.LstValues.getDescription(k)]);
  s:= s + format('<TH width=70 %s>ALL</TH>',[HTML_BGCOLOR_HEADER_GRAY]);
 s:= s + '</TR>';
 //effectifs par groupes
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + '<TH>Group Size</TH>';
 for k:= 1 to Attribute.nbValues do
  s:= s + format('<TD align=right>%d</TD>',[FTabSizeGroups[pred(k)]]);
 s:= s + format('<TD align="right">%d</TD>',[NbExamples]);
 s:= s + '</TR>';
 //description des variables dans chaque groupe
 for j:= 0 to pred(FDescriptors.Count) do
  begin
   //nom de la variable
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDescriptors.Attribute[j].Name]);
   //stat conditionnelles
   for k:= 1 to Attribute.nbValues do
    s:= s + format('<TD align="right">%.4f</TD>',[((FStatGroups.Items[pred(k)] as TLstCalcStatDesContinuous).LstStat.Items[j] as TCalcStatDesContinuous).Average]);
   s:= s + format('<TD align="right" %s>%.4f</TD>',[HTML_BGCOLOR_HEADER_GRAY,(FStatGlobal.LstStat.Items[j] as TCalcStatDesContinuous).Average]);
   s:= s + '</TR>';
  end;
 s:= s+'</table>';
 //
 result:= s;
end;

function TCalcSDCondMultivariate.getHTMLResult(prmOption: integer): string;
var s: string;
begin
 s:= format('<TD valign="top" %s>',[HTML_BGCOLOR_DATA_BLUE]) + getHTMLMultivariateStatDes() + '</TD>';
 s:= s + format('<TD valign="top" %s>',[HTML_BGCOLOR_DATA_BLUE]) + getHTMLStatisticalTests() + '</TD>';
 result:= s;
end;

function TCalcSDCondMultivariate.getStatGroup(
  k: integer): TLstCalcStatDesContinuous;
begin
 result:= FStatGroups.Items[k] as TLstCalcStatDesContinuous;
end;

procedure TCalcSDCondMultivariate.RefreshStat(prmExamples: TExamples);
var lst: TObjectList;
begin
 FNbExamples:= prmExamples.Size;
 //caluler les stats marginales
 computeMultivariateStatDes(prmExamples);
 //calculer les stats conditionnelles
 lst:= prmExamples.DispatchExamples(Attribute);
 computeMultivariateCondStatDes(lst);
 lst.Free();
 //calculer les statistiques de test
 computeStatisticalTests();
end;

{ TCalcSDCondManova }

procedure TCalcSDCondManova.computeMultivariateCondStatDes(
  lstExamples: TObjectList);
var mvcv: PMatrix;
    k,i,j: integer;
    ex: TExamples;
    stats: TLstCalcStatDesContinuous;
    vNum: double;
begin
 //stats
 inherited computeMultivariateCondStatDes(lstExamples);
 //matrice de variance covariance intra (sans la pondration finale)
 dimMatrix(FWBrut,FDescriptors.Count,FDescriptors.Count);
 //pour chaque modalit de l'attribut classe
 for k:= 1 to Attribute.nbValues do
  begin
   ex:= lstExamples.Items[pred(k)] as TExamples;
   stats:= FStatGroups.Items[pred(k)] as TLstCalcStatDesContinuous;
   mvcv:= BuildMatVCV(ex,FDescriptors,vcvNormCentered,stats);
   //addition
   vNum:= 1.0*ex.Size;
   for i:= 1 to FDescriptors.Count do
    for j:= 1 to FDescriptors.Count do
     begin
      //dcomposition du calcul car plus tard peut-tre la stat BOX-M pourrait tre programme (galit des mat. var-covar)
      mvcv^[i]^[j]:= vNum * mvcv^[i]^[j];
      FWBrut^[i]^[j]:= FWBrut^[i]^[j] + mvcv^[i]^[j];
     end;
   // optimiser plus tard, c'est dramatique toutes ces allocations/dsallocations inutiles
   delMatrix(mvcv,FDescriptors.Count,FDescriptors.Count);
  end;
end;

procedure TCalcSDCondManova.computeMultivariateStatDes(
  prmExamples: TExamples);
begin
 //stats. descriptives par variables
 inherited computeMultivariateStatDes(prmExamples);
 //matrice
 FT:= BuildMatVCV(prmExamples,FDescriptors,vcvNormCentered,FStatGlobal);
end;

procedure TCalcSDCondManova.computeStatisticalTests;
var matT,matW: PMatrix;
    i,j: integer;
    vPonderation: double;    
    //Tenenhaus, p.251
    n,p,k,r,t,u,num,denom: double;
begin
 //initialisations -- transformation de Bartlett et la p-value associe
 FBartlett:= 0.0;
 FPValueBartlett:= 1.0;
 FDDLBartlett:= 0;
 FRao:= 0.0;
 FPValueRao:= 1.0;
 FDDLNum:= 0;
 FDDLDenom:= 0;
 //copier les matrices et corriger ventuiellement en re-pondrant correctement
 DimMatrix(matT,FDescriptors.Count,FDescriptors.Count);
 CopyMatrix(matT,FT,1,1,FDescriptors.Count,FDescriptors.Count);
 DimMatrix(matW,FDescriptors.Count,FDescriptors.Count);
 CopyMatrix(matW,FWBrut,1,1,FDescriptors.Count,FDescriptors.Count);
 vPonderation:= 1.0 / (1.0 * NbExamples); 
 for i:= 1 to FDescriptors.Count do
  for j:= 1 to FDescriptors.Count do
   matW^[i]^[j]:= vPonderation * matW^[i]^[j];
 //Lambda de Wilks
 FWilks:= self.computeWilks(matT,matW,FDescriptors.Count);
 //la suite est calculable ?
 if (FWilks < 1.0)
  then
   begin
    TraceLog.WriteToLogFile(format('[WILKS] Wilks = %.8f',[FWilks]));
    //****************
    //calcul des stats
    //****************
    n:= 1.0 * NbExamples;
    p:= 1.0 * FDescriptors.Count;
    k:= 1.0 * Attribute.nbValues;
    //transformation de Bartlett --> Chi-2 (Tenenhaus, p.251)
    //new -- 02/02/2007 -- attention, si la discrimination est parfaite, FWilks peut tre gal  zro
    if (FWilks > 0.0)
     then FBartlett:= -1.0 * (n - 1.0 - 0.5 * (p + k)) * LN(FWilks)
     else FBartlett:= 1e308;
    FDDLBartlett:= FDescriptors.Count * (Attribute.nbValues - 1);
    FPValueBartlett:= PKhi2(FDDLBartlett,FBartlett);
    //transformation de Rao --> Fisher (Tenenhaus, p.251)
    num:= p * (k - 1.0);
    FDDLNum:= TRUNC(num);
    t:= power(p,2.0) + power(k-1.0,2.0) - 5.0;
    if (t > 0.0)
     then t:= SQRT((power(p,2.0) * power(k-1,2.0) - 4.0)/t)
     else t:= 1.0;
    r:= (num - 2.0) /  2.0;
    u:= (2.0 * n - p - k - 2.0) / 2.0; 
    denom:= u * t - r;
    FDDLDenom:= TRUNC(denom);
    FRao:= power(FWilks,1.0/t);
    //idem -- new -- 02/02/2007 -- attention  la discrimination parfaite
    if (FRao > 0.0)
     then FRao:= ((1.0 - FRao)/FRao) * (denom / num)
     else FRao:= 1e308;
    FPValueRao:= PSnedecor(FDDLNum,FDDLDenom,FRao);
   end;
 delMatrix(matT,FDescriptors.Count,FDescriptors.Count);
 delMatrix(matW,FDescriptors.Count,FDescriptors.Count);
end;

function TCalcSDCondManova.computeWilks(T,W: PMatrix; size: integer): double;
var dT,dW: double;
begin
 result:= 1.0;
 //dterminant de la mvcv totale
 dT:= det(T,1,size);
 if (dT > 0.0)
  then
   begin
    //det de la mvcv intra
    dW:= det(W,1,size);
    //and then...
    result:= dW / dT;
   end;
end;

destructor TCalcSDCondManova.destroy;
begin
 TRY
 if assigned(FT) then Matrices.DelMatrix(FT,FDimension,FDimension);
 if assigned(FWBrut) then Matrices.DelMatrix(FWBrut,FDimension,FDimension);
 inherited destroy();
 EXCEPT
 on e: Exception do
  TraceLog.WriteToLogFile(Format('[%s] destroy, error = %s',[self.ClassName,e.Message]));
 END;
end;

function TCalcSDCondManova.getHTMLStatisticalTests: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Stat</TH><TH width=70>Value</TH><TH width=70>p-value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Wilks'' Lambda</TD><TD align="right">%.4f</TD><TD align="center">-</TD></TR>',[FWilks]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Bartlett -- C(%d)</TD><TD align="right">%.4f</TD><TD align="right" %s>%.4f</TD></TR>',[FDDLBartlett,FBartlett,colorCodeForPValue(FPValueBartlett),FPValueBartlett]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Rao -- F(%d, %d)</TD><TD align="right">%.4f</TD><TD align="right" %s>%.4f</TD></TR>',[FDDLNum,FDDLDenom,FRao,colorCodeForPValue(FPValueRao),FPValueRao]);
 s:= s + '</table>';
 result:= s;
end;

{ TLstCalcSDCondManova }

function TLstCalcSDCondManova.getHeaderHTML: string;
begin
 result:= HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Descriptive stat. (Mean)</TH><TH>Tests results</TH></TR>';
end;

{ TCalcSDCondT2 }

procedure TCalcSDCondT2.computeMultivariateCondStatDes(
  lstExamples: TObjectList);
var mvcv: PMatrix;
    k,i,j: integer;
    ex: TExamples;
    stats: TLstCalcStatDesContinuous;
    vNum,allNum: double;
begin
 //stats
 inherited computeMultivariateCondStatDes(lstExamples);
 //matrice de variance covariance intra (pooled covariance matrix)
 dimMatrix(FSp,FDescriptors.Count,FDescriptors.Count);
 //pour chaque modalit de l'attribut classe
 allNum:= 0;
 for k:= 1 to Attribute.nbValues do //"Attribute.nbValues" doit tre  2 normalement
  begin
   ex:= lstExamples.Items[pred(k)] as TExamples;
   stats:= FStatGroups.Items[pred(k)] as TLstCalcStatDesContinuous;
   mvcv:= BuildMatVCV(ex,FDescriptors,vcvNormCentered,stats);
   //addition
   vNum:= 1.0*ex.Size;
   allNum:= allNum + vNum;
   for i:= 1 to FDescriptors.Count do
    for j:= 1 to FDescriptors.Count do
     begin
      mvcv^[i]^[j]:= vNum * mvcv^[i]^[j];
      FSp^[i]^[j]:= FSp^[i]^[j] + mvcv^[i]^[j];
     end;
   // optimiser plus tard, c'est dramatique toutes ces allocations/dsallocations inutiles
   delMatrix(mvcv,FDescriptors.Count,FDescriptors.Count);
  end;
 //diviser maintenant par (n1 +n2 - 2) pour la pooled...
 allNum:= allNum - 2.0;
 for i:= 1 to FDescriptors.Count do
  for j:= 1 to FDescriptors.Count do
   FSp^[i]^[j]:= FSp^[i]^[j] / allNum;
end;

procedure TCalcSDCondT2.computeStatisticalTests;
var invSp: PMatrix;
    k,j: integer;
    sum: double;
    n1,n2,p: double;
    fquantile: double;
begin
 T2:= 0.0;
 F:= 0.0;
 PValue:= 1.0;
 //
 setLength(ecarts,FDescriptors.Count+1);
 setLength(LBound,FDescriptors.Count+1);
 setLength(UBound,FDescriptors.Count+1);
 //inverser la matrice FSp
 dimMatrix(invSp,FDescriptors.Count,FDescriptors.Count);
 if (Matrices.InvMat(FSp,1,FDescriptors.Count,invSp) = MAT_OK)
  then
   begin
    n1:= 1.0 * ((FStatGroups.Items[0] as TLstCalcStatDesContinuous)).NbExamples;
    n2:= 1.0 * ((FStatGroups.Items[1] as TLstCalcStatDesContinuous)).NbExamples;
    p:= FDescriptors.Count;
    fquantile:= ILS_FISCHER(CONF_LEVEL_CI,p,n1+n2-p-1);
    for j:= 1 to FDescriptors.Count do
     begin
      Ecarts[j]:= ((FStatGroups.Items[0] as TLstCalcStatDesContinuous).Stat(pred(j)) as TCalcStatDesContinuous).Average -
                  ((FStatGroups.Items[1] as TLstCalcStatDesContinuous).Stat(pred(j)) as TCalcStatDesContinuous).Average;
      //TraceLog.WriteToLogFile(format('[T2] ecart %d = %.5f',[j,ecarts[j]]));
      //bornes de l'intervalle de confiance  95%
      LBound[j]:= Ecarts[j] - sqrt((p*(n1+n2-2.0))/(n1+n2-p-1)*fquantile) * sqrt((1.0/n1+1.0/n2)*FSp^[j]^[j]);
      UBound[j]:= Ecarts[j] + sqrt((p*(n1+n2-2.0))/(n1+n2-p-1)*fquantile) * sqrt((1.0/n1+1.0/n2)*FSp^[j]^[j]);
     end;
    for j:= 1 to FDescriptors.Count do
     begin
      sum:= 0.0;
      for k:= 1 to FDescriptors.Count do sum:= sum + ecarts[k]*invSp^[k]^[j];
      T2:= T2 + sum * ecarts[j];
     end;
    T2:= (n1*n2)/(n1+n2)*T2;
    TraceLog.WriteToLogFile(format('[T2] n1, n2, T2 = %.0f, %.0f, %.6f',[n1,n2,T2]));
    F:= (n1+n2-p-1.0)/(p*(n1+n2-2.0))*T2;
    Df1:= trunc(p);
    Df2:= trunc(n1+n2-p-1);
    PValue:= PSnedecor(Df1,Df2,F);
   end;
 //destroy
 delMatrix(invSp,FDescriptors.Count,FDescriptors.Count);
end;

destructor TCalcSDCondT2.destroy;
begin
 if (FSp <> nil) then delMatrix(FSp,FDimension,FDimension);
 setLength(Ecarts,0);
 setLength(LBound,0);
 setLength(UBound,0);
 inherited;
end;

function TCalcSDCondT2.getHTMLMultivariateStatDes: string;
var s: string;
    k,j: integer;
    colStr: string;
begin
 //dcrire les variables conditionnellement au groupe d'appartenance
 s:= HTML_HEADER_TABLE_RESULT;
 //en-tte avec le nom des modalits (reprsentant les groupes)
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=70>Group</TH>';
 for k:= 1 to Attribute.nbValues do
  s:= s + format('<TH width=70>%s</TH>',[Attribute.LstValues.getDescription(k)]);
 s:= s + format('<TH colspan=3 %s>95%s Simultaneous Conf.Intervals</TH>',[HTML_BGCOLOR_HEADER_GRAY,'%']);
 s:= s + '</TR>';
 //effectifs par groupes
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + '<TH>Group Size</TH>';
 for k:= 1 to Attribute.nbValues do
  s:= s + format('<TD align=right>%d</TD>',[FTabSizeGroups[pred(k)]]);
 s:= s + format('<TD align="center" width=70>%s</TD><TD align="center" width=70>%s</TD><TD align="center" width=70>%s</TD>',['Difference','Lower.Bound','Upper.Bound']);
 s:= s + '</TR>';
 //description des variables dans chaque groupe
 for j:= 0 to pred(FDescriptors.Count) do
  begin
   //nom de la variable
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDescriptors.Attribute[j].Name]);
   //stat conditionnelles
   for k:= 1 to Attribute.nbValues do
    s:= s + format('<TD align="right">%.4f</TD>',[((FStatGroups.Items[pred(k)] as TLstCalcStatDesContinuous).LstStat.Items[j] as TCalcStatDesContinuous).Average]);
   if (LBound[j+1]*UBound[j+1] > 0) then colStr:= HTML_BGCOLOR_LIGHT_RED else colStr:= HTML_BGCOLOR_HEADER_GRAY;
   s:= s + format('<TD align="right" %s>%.4f</TD><TD align="right" %s>%.4f</TD><TD align="right" %s>%.4f</TD>',[HTML_BGCOLOR_HEADER_GRAY,Ecarts[j+1],colStr,LBound[j+1],colStr,UBound[j+1]]);
   s:= s + '</TR>';
  end;
 s:= s+'</table>';
 //
 result:= s;
end;

function TCalcSDCondT2.getHTMLStatisticalTests: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Stat</TH><TH width=70>Value</TH><TH width=70>p-value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>T2</TD><TD align="right">%.4f</TD><TD align="center">-</TD></TR>',[T2]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>F(%d, %d)</TD><TD align="right">%.4f</TD><TD align="right" %s>%.4f</TD></TR>',[Df1,Df2,F,colorCodeForPValue(PValue),PValue]);
 s:= s + '</table>';
 result:= s;
end;

{ TLstCalcSDCondT2 }

function TLstCalcSDCondT2.getHeaderHTML: string;
begin
 result:= HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Descriptive stat. (Mean)</TH><TH>Tests results</TH></TR>';
end;

{ TCalcSDCondT2Heteroscedastic }

procedure TCalcSDCondT2Heteroscedastic.computeMultivariateCondStatDes(
  lstExamples: TObjectList);
var mvcv: PMatrix;
    k,i,j: integer;
    ex: TExamples;
    stats: TLstCalcStatDesContinuous;
    vNum: double;
begin
 //stats
 inherited computeMultivariateCondStatDes(lstExamples);
 //matrice de variance covariance intra (pooled covariance matrix)
 dimMatrix(FSt,FDescriptors.Count,FDescriptors.Count);
 //pour le calcul des intervalles de confiance des carts
 setLength(LBound,FDescriptors.Count+1);
 setLength(UBound,FDescriptors.Count+1);
 //pour chaque modalit de l'attribut classe
 for k:= 1 to Attribute.nbValues do //"Attribute.nbValues" doit tre  2 normalement
  begin
   ex:= lstExamples.Items[pred(k)] as TExamples;
   stats:= FStatGroups.Items[pred(k)] as TLstCalcStatDesContinuous;
   mvcv:= BuildMatVCV(ex,FDescriptors,vcvNormCentered,stats);
   //addition
   vNum:= 1.0*ex.Size;
   for i:= 1 to FDescriptors.Count do
    for j:= 1 to FDescriptors.Count do
     begin
      //mvcv est dj divis par n, il reste donc  diviser par n-1
      mvcv^[i]^[j]:= mvcv^[i]^[j]/(-1.0 + vNum);
      FSt^[i]^[j]:= FSt^[i]^[j] + mvcv^[i]^[j];
     end;
   //calcul des carts-types pour les intervalles de confiance simultans
   for j:= 1 to FDescriptors.Count do LBound[j]:= LBound[j] + mvcv^[j]^[j];
   for j:= 1 to FDescriptors.Count do UBound[j]:= UBound[j] + mvcv^[j]^[j];
   // optimiser plus tard, c'est dramatique toutes ces allocations/dsallocations inutiles
   delMatrix(mvcv,FDescriptors.Count,FDescriptors.Count);
  end;
end;

procedure TCalcSDCondT2Heteroscedastic.computeStatisticalTests;
var invSt: PMatrix;
    k,j: integer;
    sum: double;
    quantile: double;
begin
 T2:= 0.0;
 PValue:= 1.0;
 quantile:= ILS_CHI2(CONF_LEVEL_CI,FDescriptors.Count);
 //inverser la matrice FSt
 dimMatrix(invSt,FDescriptors.Count,FDescriptors.Count);
 if (Matrices.InvMat(FSt,1,FDescriptors.Count,invSt) = MAT_OK)
  then
   begin
    setLength(Ecarts,FDescriptors.Count+1);
    for j:= 1 to FDescriptors.Count do
     begin
      Ecarts[j]:= ((FStatGroups.Items[0] as TLstCalcStatDesContinuous).Stat(pred(j)) as TCalcStatDesContinuous).Average -
                  ((FStatGroups.Items[1] as TLstCalcStatDesContinuous).Stat(pred(j)) as TCalcStatDesContinuous).Average;
      //TraceLog.WriteToLogFile(format('[T2] ecart %d = %.5f',[j,ecarts[j]]));
      LBound[j]:= Ecarts[j] - sqrt(quantile*LBound[j]);
      UBound[j]:= Ecarts[j] + sqrt(quantile*UBound[j]);
     end;
    for j:= 1 to FDescriptors.Count do
     begin
      sum:= 0.0;
      for k:= 1 to FDescriptors.Count do sum:= sum + ecarts[k]*invSt^[k]^[j];
      T2:= T2 + sum * Ecarts[j];
     end;
    Df:= FDescriptors.Count;
    TraceLog.WriteToLogFile(format('[T2] T2 = %.8f',[T2]));
    PValue:= PKhi2(Df,T2);
   end;
 //destroy
 delMatrix(invSt,FDescriptors.Count,FDescriptors.Count);
end;

destructor TCalcSDCondT2Heteroscedastic.destroy;
begin
 if assigned(FSt) then delMatrix(FSt,FDimension,FDimension);
 setLength(Ecarts,0);
 setLength(LBound,0);
 setLength(UBound,0); 
 inherited;
end;

function TCalcSDCondT2Heteroscedastic.getHTMLMultivariateStatDes: string;
var s: string;
    k,j: integer;
    colStr: string;
begin
 //dcrire les variables conditionnellement au groupe d'appartenance
 s:= HTML_HEADER_TABLE_RESULT;
 //en-tte avec le nom des modalits (reprsentant les groupes)
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=70>Group</TH>';
 for k:= 1 to Attribute.nbValues do
  s:= s + format('<TH width=70>%s</TH>',[Attribute.LstValues.getDescription(k)]);
 s:= s + format('<TH colspan=3 %s>95%s Simultaneous Conf.Intervals</TH>',[HTML_BGCOLOR_HEADER_GRAY,'%']);
 s:= s + '</TR>';
 //effectifs par groupes
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + '<TH>Group Size</TH>';
 for k:= 1 to Attribute.nbValues do
  s:= s + format('<TD align=right>%d</TD>',[FTabSizeGroups[pred(k)]]);
 s:= s + format('<TD align="center" width=70>%s</TD><TD align="center" width=70>%s</TD><TD align="center" width=70>%s</TD>',['Difference','Lower.Bound','Upper.Bound']);
 s:= s + '</TR>';
 //description des variables dans chaque groupe
 for j:= 0 to pred(FDescriptors.Count) do
  begin
   //nom de la variable
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDescriptors.Attribute[j].Name]);
   //stat conditionnelles
   for k:= 1 to Attribute.nbValues do
    s:= s + format('<TD align="right">%.4f</TD>',[((FStatGroups.Items[pred(k)] as TLstCalcStatDesContinuous).LstStat.Items[j] as TCalcStatDesContinuous).Average]);
   if (LBound[j+1]*UBound[j+1] > 0) then colStr:= HTML_BGCOLOR_LIGHT_RED else colStr:= HTML_BGCOLOR_HEADER_GRAY;
   s:= s + format('<TD align="right" %s>%.4f</TD><TD align="right" %s>%.4f</TD><TD align="right" %s>%.4f</TD>',[HTML_BGCOLOR_HEADER_GRAY,Ecarts[j+1],colStr,LBound[j+1],colStr,UBound[j+1]]);
   s:= s + '</TR>';
  end;
 s:= s+'</table>';
 //
 result:= s;
end;

function TCalcSDCondT2Heteroscedastic.getHTMLStatisticalTests: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Stat</TH><TH width=70>Value</TH><TH width=70>p-value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>T2</TD><TD align="right">%.4f</TD><TD align="center">-</TD></TR>',[T2]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>CHI-2 (%d)</TD><TD align="right">%.4f</TD><TD align="right" %s>%.4f</TD></TR>',[Df,T2,colorCodeForPValue(PValue),PValue]);
 s:= s + '</table>';
 result:= s;
end;

{ TCalcSDCondBoxM }

procedure TCalcSDCondBoxM.computeMultivariateCondStatDes(
  lstExamples: TObjectList);
var mvcv, Sp: PMatrix;
    k,i,j: integer;
    ex: TExamples;
    stats: TLstCalcStatDesContinuous;
    vNum,allNum: double;
    det,snk,c,p: double;
begin
 T:= 0.0;
 //stats
 inherited computeMultivariateCondStatDes(lstExamples);
 //matrice de variance covariance intra (pooled covariance matrix)
 dimMatrix(Sp,FDescriptors.Count,FDescriptors.Count);
 //pour chaque modalit de l'attribut classe
 allNum:= 0.0; snk:= 0.0;
 for k:= 1 to Attribute.nbValues do
  begin
   ex:= lstExamples.Items[pred(k)] as TExamples;
   stats:= FStatGroups.Items[pred(k)] as TLstCalcStatDesContinuous;
   mvcv:= BuildMatVCV(ex,FDescriptors,vcvNormCentered,stats);
   //addition
   vNum:= 1.0*ex.Size;
   snk:= snk + 1.0 / (-1.0 + vNum);
   allNum:= allNum + vNum;
   for i:= 1 to FDescriptors.Count do
    for j:= 1 to FDescriptors.Count do
     begin
      mvcv^[i]^[j]:= vNum * mvcv^[i]^[j];
      Sp^[i]^[j]:= Sp^[i]^[j] + mvcv^[i]^[j];
      mvcv^[i]^[j]:= mvcv^[i]^[j] / (-1.0 + vNum);
     end;
   //calculer le dterminant
   det:= Matrices.Det(mvcv,1,FDescriptors.Count);
   T:= T + (-1.0 + vNum)*LN(det);
   // optimiser plus tard, c'est dramatique toutes ces allocations/dsallocations inutiles
   delMatrix(mvcv,FDescriptors.Count,FDescriptors.Count);
  end;
 //diviser maintenant par (n - K) pour la pooled...
 allNum:= allNum - 1.0 * Attribute.nbValues;
 for i:= 1 to FDescriptors.Count do
  for j:= 1 to FDescriptors.Count do
   Sp^[i]^[j]:= Sp^[i]^[j] / allNum;
 //last
 det:= Matrices.Det(Sp,1,FDescriptors.Count);
 T:= allNum * LN(det) - T;
 //TraceLog.WriteToLogFile(format('[Box M] T before correction = %.6f',[T]));
 //facteur de correction
 p:= 1.0 * FDescriptors.Count;
 c:= (2.0 * Math.Power(p,2.0) + 3.0 * p - 1.0) / (6.0 * (-1.0 + Attribute.nbValues) * (1.0 + p));
 c:= c * (snk - 1.0 / allNum);
 c:= 1.0 - c;
 //TraceLog.WriteToLogFile(format('[Box M] correction coefficient = %.6f',[c]));
 //and then...
 T:= c * T;
 TraceLog.WriteToLogFile(format('[Box M] T after correction = %.6f',[T]));
end;

procedure TCalcSDCondBoxM.computeStatisticalTests;
var nu,p: double;
begin
 p:= 1.0 * FDescriptors.Count;
 nu:= 0.5 * (-1.0 + Attribute.nbValues) * p * (1.0 + p);
 Df:= round(nu);
 PValue:= PKhi2(Df,T);
end;

function TCalcSDCondBoxM.getHTMLMultivariateStatDes: string;
var s: string;
    k,j: integer;
begin
 //dcrire les variables conditionnellement au groupe d'appartenance
 s:= HTML_HEADER_TABLE_RESULT;
 //en-tte avec le nom des modalits (reprsentant les groupes)
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=70>Group</TH>';
 for k:= 1 to Attribute.nbValues do
  s:= s + format('<TH width=70>%s</TH>',[Attribute.LstValues.getDescription(k)]);
  s:= s + format('<TH width=70 %s>ALL</TH>',[HTML_BGCOLOR_HEADER_GRAY]);
 s:= s + '</TR>';
 //effectifs par groupes
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + '<TH>Group Size</TH>';
 for k:= 1 to Attribute.nbValues do
  s:= s + format('<TD align=right>%d</TD>',[FTabSizeGroups[pred(k)]]);
 s:= s + format('<TD align="right">%d</TD>',[NbExamples]);
 s:= s + '</TR>';
 //description des variables dans chaque groupe
 for j:= 0 to pred(FDescriptors.Count) do
  begin
   //nom de la variable
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDescriptors.Attribute[j].Name]);
   //stat conditionnelles
   for k:= 1 to Attribute.nbValues do
    s:= s + format('<TD align="right">%.4f</TD>',[((FStatGroups.Items[pred(k)] as TLstCalcStatDesContinuous).LstStat.Items[j] as TCalcStatDesContinuous).StdDev]);
   s:= s + format('<TD align="right" %s>%.4f</TD>',[HTML_BGCOLOR_HEADER_GRAY,(FStatGlobal.LstStat.Items[j] as TCalcStatDesContinuous).StdDev]);
   s:= s + '</TR>';
  end;
 s:= s+'</table>';
 //
 result:= s;
end;

function TCalcSDCondBoxM.getHTMLStatisticalTests: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Stat</TH><TH width=70>Value</TH><TH width=70>p-value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>T [CHI-2 (%d)]</TD><TD align="right">%.4f</TD><TD align="right" %s>%.4f</TD></TR>',[Df,T,colorCodeForPValue(PValue),PValue]);
 s:= s + '</table>';
 result:= s;
end;

{ TLstCalcSDCondBoxM }

function TLstCalcSDCondBoxM.getHeaderHTML: string;
begin
 result:= HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Descriptive stat. (Std.Dev)</TH><TH>Tests results</TH></TR>';
end;

end.
