(******************************************************************************)
(* UCalcMatrixAdditionalFunctions.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(******************************************************************************)

{
@abstract(Ajout de nouvelles fonctions dans la bibliothque de matrice de J. Debord)
@author(Ricco)
@created(12/01/2004)
L'ide est de rajouter qqs fonctions issues de Numerical Recipes, notamment tout ce qui touche
 la recherche de vecteurs et valeurs propres.
}
unit UCalcMatrixAdditionalFunctions;

interface

USES
        FMath, Matrices, Eigen;

{la procdure qui encapsule le tout, la matrice M doit tre symtrique,
EVal et EVec sont cres dans la procdure}
procedure GetEigensFromSymetricMatrix(M: PMatrix; n: integer; var EVal: PVector; var EVec: PMatrix);

//produit de deux matrices C = A x B
procedure MatrixProd(A: PMatrix; rowA, colA: integer; B: PMatrix; rowB, colB: integer; var C: PMatrix);

//inversion de matrice symtrique (carre)  partir de la SVD -- ne plante pas quand la matrice est singulire
function  SVDInverse(A: PMatrix; UBound: integer): boolean;

//new -- 19/02/2007 -- premire valeur et vecteur propre par la mthode de la puissance itre, renvoie la valeur propre
function  getFirstEigSymetricMatrix(M: PMatrix; n: integer; var EVec: PVector): double;

implementation

uses
        SysUtils, Math, ULogFile;

{rduire une matrice symtrique (!!!) en une forme tri-diagonale, prparation de la recherche de valeurs propres:
-> A est matrice originelle qui est manipule et devient la matrice orthogonale Q,
-> n est la dimension de la matrice A,
-> D renvoie les lments diagonaux de la matrice tri-diagonale,
-> E reprsente les lments non-diagonaux}        
procedure TRED2(var A: PMatrix; n: integer; var D,E: PVector);

VAR
   l,k,j,i: integer;
   scale,hh,h,g,f: float;

   FUNCTION sign(a,b: float): float;
   BEGIN
      IF (b < 0) THEN sign := -abs(a) ELSE sign := abs(a)
   END;

BEGIN
   IF (n > 1) THEN BEGIN
      FOR i := n DOWNTO 2 DO BEGIN
         l := i-1;
         h := 0.0;
         scale := 0.0;
         IF (l > 1) THEN BEGIN
            FOR k := 1 TO l DO BEGIN
               scale := scale+abs(a^[i]^[k])
            END;
            IF (scale = 0.0) THEN BEGIN
               e^[i] := a^[i]^[l]
            END ELSE BEGIN
               FOR k := 1 TO l DO BEGIN
                  a^[i]^[k] := a^[i]^[k]/scale;
                  h := h+sqr(a^[i]^[k])
               END;
               f := a^[i]^[l];
               g := -sign(sqrt(h),f);
               e^[i] := scale*g;
               h := h-f*g;
               a^[i]^[l] := f-g;
               f := 0.0;
               FOR j := 1 TO l DO BEGIN
            (* Next statement can be omitted if eigenvectors not wanted *)
                  a^[j]^[i] := a^[i]^[j]/h;
                  g := 0.0;
                  FOR k := 1 TO j DO BEGIN
                     g := g+a^[j]^[k]*a^[i]^[k]
                  END;
                  IF (l > j) THEN FOR k := j+1 TO l DO g := g+a^[k]^[j]*a^[i]^[k];
                  e^[j] := g/h;
                  f := f+e^[j]*a^[i]^[j]
               END;
               hh := f/(h+h);
               FOR j := 1 TO l DO BEGIN
                  f := a^[i]^[j];
                  g := e^[j]-hh*f;
                  e^[j] := g;
                  FOR k := 1 TO j DO a^[j]^[k] := a^[j]^[k]-f*e^[k]-g*a^[i]^[k]
               END
            END
         END ELSE BEGIN
            e^[i] := a^[i]^[l]
         END;
         d^[i] := h
      END
   END;
   (* Next statement can be omitted if eigenvectors not wanted *)
   d^[1] := 0.0;
   e^[1] := 0.0;
   FOR i := 1 TO n DO BEGIN
   (* Contents of this loop can be omitted if eigenvectors not wanted,
      except for statement d[i] := a[i,i]; *)
      l := i-1;
      IF (d^[i] <> 0.0) THEN BEGIN
         FOR j := 1 TO l DO BEGIN
            g := 0.0;
            FOR k := 1 TO l DO BEGIN
               g := g+a^[i]^[k]*a^[k]^[j]
            END;
            FOR k := 1 TO l DO BEGIN
               a^[k]^[j] := a^[k]^[j]-g*a^[k]^[i]
            END
         END
      END;
      d^[i] := a^[i]^[i];
      a^[i]^[i] := 1.0;
      IF (l >= 1) THEN BEGIN
         FOR j := 1 TO l DO BEGIN
            a^[i]^[j] := 0.0;
            a^[j]^[i] := 0.0
         END
      END
   END
END;


{calcule les valeurs et vecteurs propres d'une matrice tri-diagonale, c'est le complment idal des deux procdures ci-dessus:
-> D est le vecteur renvoy par TRED2, il est transform et contient en sortie les valeurs propres,
-> E est un vecteur renvoy par TRED2, il ne veut plus rien dire en sortie
-> n est la dimension de la matrice,
-> Z est la matrice orthogonale A renvoye par TRED2, il contient en sortie les vecteurs propres dans l'ordre des valeurs propres}
PROCEDURE TQLI(VAR D,E: PVector; n: integer; VAR Z: PMatrix);
LABEL 1,2;
VAR
   m,l,iter,i,k: integer;
   s,r,p,g,f,dd,c,b: float;

   FUNCTION sign(a,b: float): float;
   BEGIN
      IF (b < 0) THEN sign := -abs(a) ELSE sign := abs(a)
   END;
   
BEGIN
   IF  (n > 1)  THEN BEGIN
      FOR i := 2 TO n DO BEGIN
         e^[i-1] := e^[i]
      END;
      e^[n] := 0.0;
      FOR l := 1 TO n DO BEGIN
         iter := 0;
1:         FOR m := l TO n-1 DO BEGIN
            dd := abs(d^[m])+abs(d^[m+1]);
            IF  (abs(e^[m])+dd = dd) THEN  GOTO 2
         END;
         m := n;
2:         IF (m <> l) THEN BEGIN
            IF (iter = 30) THEN BEGIN
               //writeln('pause in routine TQLI');
               //writeln('too many iterations'); readln
               raise Exception.Create('TQLI --> Too many iterations');
            END;
            iter := iter+1;
            g := (d^[l+1]-d^[l])/(2.0*e^[l]);
            r := sqrt(sqr(g)+1.0);
            g := d^[m]-d^[l]+e^[l]/(g+sign(r,g));
            s := 1.0;
            c := 1.0;
            p := 0.0;
            FOR i := m-1 DOWNTO l DO BEGIN
               f := s*e^[i];
               b := c*e^[i];
               IF (abs(f) >= abs(g)) THEN BEGIN
                  c := g/f;
                  r := sqrt(sqr(c)+1.0);
                  e^[i+1] := f*r;
                  s := 1.0/r;
                  c := c*s
               END ELSE BEGIN
                  s := f/g;
                  r := sqrt(sqr(s)+1.0);
                  e^[i+1] := g*r;
                  c := 1.0/r;
                  s := s*c
               END;
               g := d^[i+1]-p;
               r := (d^[i]-g)*s+2.0*c*b;
               p := s*r;
               d^[i+1] := g+p;
               g := c*r-b;
            (* Next loop can be omitted if eigenvectors not wanted *)
               FOR k := 1 TO n DO BEGIN
                  f := z^[k]^[i+1];
                  z^[k]^[i+1] := s*z^[k]^[i]+c*f;
                  z^[k]^[i] := c*z^[k]^[i]-s*f
               END
            END;
            d^[l] := d^[l]-p;
            e^[l] := g;
            e^[m] := 0.0;
            GOTO 1
         END
      END
   END
END;

{trier la matrice de vecteurs propres selon les valeurs dcroissantes des valeurs propres}
PROCEDURE EIGSRT(VAR D: PVector; VAR V: PMatrix; n: integer);
VAR
   k,j,i: integer;
   p: float;
BEGIN
   FOR i := 1 TO n-1 DO BEGIN
      k := i;
      p := d^[i];
      FOR j := i+1 TO n DO BEGIN
         IF (d^[j] >= p) THEN BEGIN
            k := j;
            p := d^[j]
         END
      END;
      IF (k <> i) THEN BEGIN
         d^[k] := d^[i];
         d^[i] := p;
         FOR j := 1 TO n DO BEGIN
            p := v^[j]^[i];
            v^[j]^[i] := v^[j]^[k];
            v^[j]^[k] := p
         END
      END
   END
END;

{multiplication par un scalaire d'une matrice symtrique}
procedure SCALARMULT(scal: float; var M: PMatrix; n: integer);
var i,j: integer;
begin
 for i:= 1 to n do
  for j:= 1 to n do
   M^[i]^[j]:= scal*M^[i]^[j];
end;

{construire le tout}
procedure GetEigensFromSymetricMatrix(M: PMatrix; n: integer; var EVal: PVector; var EVec: PMatrix);
var A: PMatrix;
    D,E: PVector;

    //B: PMatrix;
begin
 //allocations et copie
 DimMatrix(A,n,n);
 CopyMatrix(A,M,1,1,n,n);

 DimVector(D,n);
 DimVector(E,n);
 //first step
 TRED2(A,n,D,E);
 //second step
 TQLI(D,E,n,A);
 //destructions
 DelVector(E,n);
 //trier les rsultats selon les valeurs dcroissantes des Val. Propres
 EIGSRT(D,A,n);
 //multiplier toutes les valeurs des vecteurs propres par (-1)
 //je ne sais pas pourquoi la mthode ici oblige  cette correction pour tre en accord avec les logiciels franais
 SCALARMULT(-1.0,A,n);
 //restitutions
 EVal:= D;
 EVec:= A;

 (*
 --> autre mthode pour le calcul des valeurs, vecteurs propres....
 DimMatrix(B,n,n);
 DimVector(D,n);
 JACOBI(A,1,n,1000,0.0000001,B,D);
 EVec:= B;
 EVal:= D;
 *)
end;

//produit de deux matrices C = A x B
procedure MatrixProd(A: PMatrix; rowA, colA: integer; B: PMatrix; rowB, colB: integer; var C: PMatrix);
var i,j,k:Integer;
    v: float;
Begin
 If (rowB = colA)
  Then
   Begin
    Matrices.DimMatrix(C,rowA,colB);
    For i:=1 to rowA Do
     For j:=1 to colB Do
      Begin
      v:=0;
      For k:=1 to rowB Do
       v:= v + A^[i]^[k] * B^[k]^[j];
      C^[i]^[j]:= v;
      End;
   End
  Else Raise Exception.Create('Paramtres de matrices incompatibles');
End;

//------------------------------------------------------------------------------

const
    MAX_ITER_MATINV = 50;
    MATINV_ACCURACY_TO_ZERO = 1.0e-12;

FUNCTION sign(a,b: double): double;
BEGIN
      IF (b >= 0.0) THEN sign := abs(a) ELSE sign := -abs(a)
END;

PROCEDURE matinv(VAR a, vtimesw, v, w: PMatrix; n: integer);
(* adapted from the singular value decomposition of a matrix *)
(* a is a symetric matrix with the inverse returned in a *)

LABEL 1,2,3;

VAR
//   vtimesw, v, ainverse : matrix;
//   w : vector;
   ainverse : PMatrix;
   m,mp,np,nm,l,k,j,its,i: integer;
   z,y,x,scale,s,h,g,f,c,anorm: double;
   rv1: PVector;

BEGIN
   //setlength(rv1,n);
   //setlength(ainverse,n,n);

   DimVector(rv1,n);
   DimMatrix(ainverse,n,n);

   m := n;
   mp := n;
   np := n;
   g := 0.0;
   scale := 0.0;
   anorm := 0.0;
   FOR i := 1 to n DO BEGIN
      l := i+1;
      rv1[i] := scale*g;
      g := 0.0;
      s := 0.0;
      scale := 0.0;
      IF (i <= m) THEN BEGIN
         FOR k := i to m DO BEGIN
            scale := scale+abs(a[k,i])
         END;
         IF (scale <> 0.0) THEN BEGIN
            FOR k := i to m DO BEGIN
               a[k,i] := a[k,i]/scale;
               s := s+a[k,i]*a[k,i]
            END;
            f := a[i,i];
            g := -sign(sqrt(s),f);
            h := f*g-s;
            a[i,i] := f-g;
            IF (i <> n-1) THEN BEGIN
               FOR j := l to n DO BEGIN
                  s := 0.0;
                  FOR k := i to m DO BEGIN
                     s := s+a[k,i]*a[k,j]
                  END;
                  f := s/h;
                  FOR k := i to m DO BEGIN
                     a[k,j] := a[k,j]+
                        f*a[k,i]
                  END
               END
            END;
            FOR k := i to m DO BEGIN
               a[k,i] := scale*a[k,i]
            END
         END
      END;
      w[i,i] := scale*g;
      g := 0.0;
      s := 0.0;
      scale := 0.0;
      IF ((i <= m) AND (i <> n)) THEN BEGIN
         FOR k := l to n DO BEGIN
            scale := scale+abs(a[i,k])
         END;
         IF (scale <> 0.0) THEN BEGIN
            FOR k := l to n DO BEGIN
               a[i,k] := a[i,k]/scale;
               s := s+a[i,k]*a[i,k]
            END;
            f := a[i,l];
            g := -sign(sqrt(s),f);
            h := f*g-s;
            a[i,l] := f-g;
            FOR k := l to n DO BEGIN
               rv1[k] := a[i,k]/h
            END;
            IF (i <> m) THEN BEGIN
               FOR j := l to m DO BEGIN
                  s := 0.0;
                  FOR k := l to n DO BEGIN
                     s := s+a[j,k]*a[i,k]
                  END;
                  FOR k := l to n DO BEGIN
                     a[j,k] := a[j,k]
                        +s*rv1[k]
                  END
               END
            END;
            FOR k := l to n DO BEGIN
               a[i,k] := scale*a[i,k]
            END
         END
      END;
      anorm := math.max(anorm,(abs(w[i,i])+abs(rv1[i])))
   END;
   FOR i := n DOWNTO 1 DO BEGIN
      IF (i < n) THEN BEGIN
         IF (g <> 0.0) THEN BEGIN
            FOR j := l to n DO BEGIN
               v[j,i] := (a[i,j]/a[i,l])/g
            END;
            FOR j := l to n DO BEGIN
               s := 0.0;
               FOR k := l to n DO BEGIN
                  s := s+a[i,k]*v[k,j]
               END;
               FOR k := l to n DO BEGIN
                  v[k,j] := v[k,j]+s*v[k,i]
               END
            END
         END;
         FOR j := l to n DO BEGIN
            v[i,j] := 0.0;
            v[j,i] := 0.0
         END
      END;
      v[i,i] := 1.0;
      g := rv1[i];
      l := i
   END;
   FOR i := n DOWNTO 1 DO BEGIN
      l := i+1;
      g := w[i,i];
      IF (i < n) THEN BEGIN
         FOR j := l to n DO BEGIN
            a[i,j] := 0.0
         END
      END;
      IF (g <> 0.0) THEN BEGIN
         g := 1.0/g;
         IF (i <> n) THEN BEGIN
            FOR j := l to n DO BEGIN
               s := 0.0;
               FOR k := l to m DO BEGIN
                  s := s+a[k,i]*a[k,j]
               END;
               f := (s/a[i,i])*g;
               FOR k := i to m DO BEGIN
                  a[k,j] := a[k,j]+f*a[k,i]
               END
            END
         END;
         FOR j := i to m DO BEGIN
            a[j,i] := a[j,i]*g
         END
      END ELSE BEGIN
         FOR j := i to m DO BEGIN
            a[j,i] := 0.0
         END
      END;
      a[i,i] := a[i,i]+1.0
   END;
   FOR k := n DOWNTO 1 DO BEGIN
      FOR its := 1 to MAX_ITER_MATINV DO BEGIN
         FOR l := k DOWNTO 1 DO BEGIN
            nm := l-1;
            IF ((abs(rv1[l])+anorm) = anorm) THEN GOTO 2;
            IF ((abs(w[nm,nm])+anorm) = anorm) THEN GOTO 1
         END;
1:
//         c := 0.0;
         s := 1.0;
         FOR i := l to k DO BEGIN
            f := s*rv1[i];
            IF ((abs(f)+anorm) <> anorm) THEN BEGIN
               g := w[i,i];
               h := sqrt(f*f+g*g);
               w[i,i] := h;
               h := 1.0/h;
               c := (g*h);
               s := -(f*h);
               FOR j := 1 to m DO BEGIN
                  y := a[j,nm];
                  z := a[j,i];
                  a[j,nm] := (y*c)+(z*s);
                  a[j,i] := -(y*s)+(z*c)
               END
            END
         END;
2:         z := w[k,k];
         IF (l = k) THEN BEGIN
            IF (z < 0.0) THEN BEGIN
               w[k,k] := -z;
               FOR j := 1 to n DO BEGIN
               v[j,k] := -v[j,k]
            END
         END;
         GOTO 3
         END;
         IF (its = MAX_ITER_MATINV) THEN BEGIN
            TraceLog.WriteToLogFile(format('[WARNING] No convergence in %d SVDCMP iterations',[MAX_ITER_MATINV]));
            Exception.Create(format('[WARNING] No convergence in %d SVDCMP iterations',[MAX_ITER_MATINV]));
         END;
         x := w[l,l];
         nm := k-1;
         y := w[nm,nm];
         g := rv1[nm];
         h := rv1[k];
         f := ((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
         g := sqrt(f*f+1.0);
         f := ((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x;
         c := 1.0;
         s := 1.0;
         FOR j := l to nm DO BEGIN
            i := j+1;
            g := rv1[i];
            y := w[i,i];
            h := s*g;
            g := c*g;
            z := sqrt(f*f+h*h);
            rv1[j] := z;
            c := f/z;
            s := h/z;
            f := (x*c)+(g*s);
            g := -(x*s)+(g*c);
            h := y*s;
            y := y*c;
            FOR nm := 1 to n DO BEGIN
               x := v[nm,j];
               z := v[nm,i];
               v[nm,j] := (x*c)+(z*s);
               v[nm,i] := -(x*s)+(z*c)
            END;
            z := sqrt(f*f+h*h);
            w[j,j] := z;
            IF (z <> 0.0) THEN BEGIN
               z := 1.0/z;
               c := f*z;
               s := h*z
            END;
            f := (c*g)+(s*y);
            x := -(s*g)+(c*y);
            FOR nm := 1 to m DO BEGIN
               y := a[nm,j];
               z := a[nm,i];
               a[nm,j] := (y*c)+(z*s);
               a[nm,i] := -(y*s)+(z*c)
            END
         END;
         rv1[l] := 0.0;
         rv1[k] := f;
         w[k,k] := x
      END;
3:   END;
{     mat_print(m,a,'U matrix');
     mat_print(n,v,'V matrix');
     writeln(lst,'Diagonal values of W inverse matrix');
      for i := 1 to n do
         write(lst,1/w[i]:6:3);
     writeln(lst);  }
     for i := 1 to n do
         for j := 1 to n do
         begin
              //if w[j,j] < 1.0e-6 then vtimesw[i,j] := 0
              if (w[j,j] < MATINV_ACCURACY_TO_ZERO) then vtimesw[i,j] := 0.0
              else vtimesw[i,j] := v[i,j] * (1.0 / w[j,j] );
         end;
{     mat_print(n,vtimesw,'V matrix times w inverse ');  }
     for i := 1 to m do
         for j := 1 to n do
         begin
             ainverse[i,j] := 0.0;
             for k := 1 to m do
             begin
                  ainverse[i,j] := ainverse[i,j] + vtimesw[i,k] * a[j,k]
             end;
         end;
{     mat_print(n,ainverse,'Inverse Matrix'); }
     for i := 1 to n do
         for j := 1 to n do
             a[i,j] := ainverse[i,j];
     {
     ainverse := nil;
     rv1 := nil;
     }
     DelMatrix(ainverse,n,n);
     DelVector(rv1,n);
END;
//-------------------------------------------------------------------


//inversion de matrice symtrique (carre)  partir de la SVD -- ne plante pas quand la matrice est singulire
function  SVDInverse(A: PMatrix; UBound: integer): boolean;
var
   v, w, vtimesw : PMatrix;
begin
 DimMatrix(V,UBound,UBound);
 DimMatrix(W,UBound,UBound);
 DimMatrix(VTimesW,UBound,UBound);
 //
 TRY
 matinv(a,vtimesw,v,w,UBound);
 DelMatrix(V,UBound,UBound);
 DelMatrix(W,UBound,UBound);
 delMatrix(VTimesW,UBound,UBound);
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

//norme d'un vecteur
function getNorme(n: integer; EVec: PVector): single;
var sum: single;
    k: integer;
begin
 sum:= 0.0;
 for k:= 1 to n do sum:= sum + EVec^[k] * EVec^[k];
 result:= SQRT(sum);
end;

const
    TOLERANCE_FOR_PUISSANCE_ITEREE : single = 1.0e-5;

//new -- 19/02/2007 -- premire valeur et vecteur propre par la mthode de la puissance itre, renvoie la valeur propre
function getFirstEigSymetricMatrix(M: PMatrix; n: integer; var EVec: PVector): double;
var tmp: PVector;
    j,k: integer;
    prevNorm,norm: single;
    value, sum: single;
begin
 //initialiser la matrice temporaire
 dimVector(tmp,n);
 for k:= 1 to n do tmp^[k]:= 1.0;
 norm:= getNorme(n,tmp);
 //initialiser le vecteur de rponse
 dimVector(EVec,n);
 REPEAT
  prevNorm:= norm;
  //produit
  for k:= 1 to n do
   begin
    sum:= 0.0;
    for j:= 1 to n do sum:= sum + M^[k]^[j] * tmp^[j];
    EVec^[k]:= sum;
   end;
  //norme
  norm:= getNorme(n,EVec);
  //normaliser
  if (norm > 0)
   then for k:= 1 to n do EVec^[k]:= EVEc^[k] / norm;
  //recopier le vecteur
  copyVector(tmp,EVEc,1,n);
 UNTIL (ABS(prevNorm-norm) < TOLERANCE_FOR_PUISSANCE_ITEREE);
 delVector(tmp,n);
 result:= norm;
end;

end.
