(***************************************************************************)
(* UCalcDistributionEvaluation.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(***************************************************************************)

{
@abstract(Fonctions pour valuer des distributions)
@author(Ricco)
@created(21/04/2005)

Le tout est de standardiser au possible la prsentation afin de maximiser
la rutilisabilit sans rentrer dans des trucs trop complexes ou contraignants...

}

unit UCalcDistributionEvaluation;

interface

USES
   UCalcDistribution,
   UDatasetDefinition;

TYPE
   //type abstrait de paramtrage des fonctions d'valuation
   TPrmFuncDistEval = class
                      end;

   //type de fonction d'valuation d'une distribution -- avec ventuellement des paramtres
   TFuncEvalDistribution = function (Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;

   //type de fonction pour valuer une rgle,  laquelle on a attribu une conclusion donc
   TFuncEvalRule = function(Root,Current: TTabFrequence; idConculsion: TTypeDiscrete; parameters: TPrmFuncDistEval): double; 

   //*************************************************************
   //* mesures de qualit, fonctions de type TFuncEvalDistribution
   //*************************************************************

   //Shannon -- entropie d'un noeud courant
   Function Shannon(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
   //J-Measure -- comparaison de la racine et du noeud courant
   Function JMeasure(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
   //calcule en p-value la diffrence entre 2 profils de distributions -- pour CN2 (rf. Kalbfleish -- 1979)
   Function CompareDistribCHI2(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
   //new -- 04/11/2009 -- test d'indpendance du KHI-2, pour voir si l'adjonction d'une condition est licite
   Function TestKhi2(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
   //Valeur-Test sur la confiance - Loi hypergometrique
   Function VTConfidence(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
   //Valeur-Test sur les contre-exemples - Loi de Poisson --> Indice d'implication
   Function VTMisclassification(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
   //J-Measure asymtrique
   Function AsymmetricJMeasure(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;

implementation

uses
   Math, FMATH, UCalcSpvRuleInduction;

//Shannon -- entropie d'un noeud courant
Function Shannon(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
var i: integer;
    value: double;
    sum: double;
begin
 sum:= 0.0;
 if (Current.Value[0]>0)
  then
   begin
    for i:= 1 to Current.Size do
     begin
      value:= Current.Frequence[i];
      if (value>0)
       then sum:= sum+value*Math.log2(value);
     end;
    //pas de pondration, on utilise directement l'entropie
   end;
 result:= -1.0 * sum;
end;

//J-Measure -- comparaison de la racine et du noeud courant
Function JMeasure(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
Var Somme: Double;
    Rc,Rr: Double;
    i: Integer;
Begin
 If (Root.Value[0]>0)
  then
   begin
     Somme:= 0;
     For i:= 1 To Current.Size Do
      If (Current.Value[0]>0)
       Then
        Begin
         Rc:= Current.Frequence[i];
         Rr:= Root.Frequence[i];
         // Modified by GongYu 2007-8-12 12:22:30
         //avoid div 0 err
         if (Rc<>0) and (Rr<>0)
          then Somme:= Somme+Rc*Math.log2(Rc/Rr);
        End;
     //ngatif parce qu'on part dans la minimisation
     result:= -1.0*(Current.Value[0]/Root.Value[0])*Somme;
   end
  //aucun calcul  faire...
  else result:= 0.0;
End;

//calcule en p-value la diffrence entre 2 profils de distributions -- pour CN2
Function CompareDistribCHI2(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
Var Chi2: Double;
    i: Integer;
    Value: Double;
Begin
  //tableaux vides ?
  If (Root.Value[0]=0) OR (Current.Value[0]=0)
   Then result:= 1.0 //pas de comparaison possible
   Else
    Begin
     Chi2:= 0;
     For i:= 1 To Root.Size Do
      Begin
       Value:= Root.Value[i]+Current.Value[i];
       If (Value>0)
        Then Chi2:= Chi2+Power(Root.Value[i]/Root.Value[0]-Current.Value[i]/Current.Value[0],2)/Value;
      End;
     Chi2:= Root.Value[0]*Current.Value[0]*Chi2;
     //calcul du risque critique
     Result:= PKHI2(pred(Root.Size),CHI2);
    End;
End;

//new -- 04/11/2009 -- test d'indpendance du KHI-2, pour voir si l'adjonction d'une condition est licite
Function TestKhi2(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
Var Chi2: Double;
    i: Integer;
    n,e,o: Double;
Begin
  //tableaux vides ?
  If (Root.Value[0]=0) OR (Current.Value[0]=0)
   Then result:= 1.0 //pas de comparaison possible
   Else
    Begin
     //effectif total
     n:= Root.Value[0];
     //sommer
     Chi2:= 0;
     For i:= 1 To Root.Size Do
      Begin
       //premire colonne
       e:= Root.Value[i] * Current.Value[0] / n;
       o:= Current.Value[i];
       if (e > 0) then Chi2:= Chi2 + Math.Power(o-e,2)/e;
       //deuxime colonne
       e:= Root.Value[i] * (n - Current.Value[0]) / n;
       o:= Root.value[i] - Current.Value[i];
       if (e > 0) then Chi2:= Chi2 + Math.Power(o-e,2)/e;
      End;
     //calcul du risque critique
     Result:= PKHI2(pred(Root.Size),CHI2);
    End;
End;

//Valeur-Test sur la confiance - Loi hypergometrique
Function VTConfidence(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
var classe: TTypeDiscrete;
    nac,na,nc,n: double;
    num,denom: double;
begin
 classe:= (Parameters as TPrmFuncRuleInduction).ClassRef;
 //rcupration des valeurs
 nac:= Current.Value[classe];
 na:= Current.Value[0];
 nc:= Root.Value[classe];
 n:= Root.Value[0];
 //calcul
 result:= 0.0;
 if (n > 1.0)
  then
   begin
    num:= nac - na * nc / n;
    denom:= (na * (n - na) / n) * (nc * (n - nc) / n);
    denom:= denom / (n - 1.0);
    denom:= sqrt(denom);
    if (denom > 0) then result:= num / denom;
    //et la bidouille pour qu'on parte dans le sens d'une minimisation
    result:= -1.0 * result;
   end;
end;

//Valeur-Test sur les contre-exemples - Loi de Poisson --> Indice d'implication
Function VTMisclassification(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
var classe: TTypeDiscrete;
    nac,na,nc,n: double;
    e,num,denom: double;
begin
 classe:= (Parameters as TPrmFuncRuleInduction).ClassRef;
 //rcupration des valeurs
 nac:= Current.Value[classe];
 na:= Current.Value[0];
 nc:= Root.Value[classe];
 n:= Root.Value[0];
 //calcul de l'indice d'implication
 e:= (n-nc) * na / n;
 num:= (na-nac) - e;
 denom:= sqrt(e);
 if (denom > 0) then result:= num / denom else result:= 0.0;
end;

//J-Measure asymtrique
Function AsymmetricJMeasure(Root,Current: TTabFrequence; parameters: TPrmFuncDistEval): double;
var classe: TTypeDiscrete;
    nac,na,nc,n: double;
    K: double;
    pc_a,pc,pa: double;
    weight,cross: double;
begin
 classe:= (Parameters as TPrmFuncRuleInduction).ClassRef;
 K:= (Parameters as TPrmFuncRuleInduction).Weight;
 //rcupration des valeurs
 nac:= Current.Value[classe];
 na:= Current.Value[0];
 nc:= Root.Value[classe];
 n:= Root.Value[0];
 //calculs...
 pc_a:= (nac + 1.0)/(na + 2.0);//laplacien
 pc:= (nc + 1.0)/(n + 2.0);
 cross:= pc_a * Math.Log2(pc_a/pc) - (1.0 - pc_a) * Math.Log2((1.0 - pc_a)/(1.0 - pc));
 pa:= na / n;
 weight:= Math.Power(pa,1.0/K);
 result:= weight * cross;
 //pour partir dans le sens d'une minimisation de la mesure
 result:= -1.0 * result;
end;

end.

