(**********************************************************************)
(* UCalcAssocStrucOutlier.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Dtection des individus atypiques  partir des rgles d'association)
@author(Ricco)
@created(13/06/2007)

L'ide est simple : dtecter les rgles comportant un nombre "trs faible" de contre-exemples.

Pourquoi ?

Si pas de contre-exemples => dpendance fonctionnelle
Si quelques contre-exemples => rgle statistique (avec un marge d'erreur dans la prcision)

Si trs peu de contre-exemples => ne sont-ce pas des anomalies ?
}
unit UCalcAssocStrucOutlier;

interface

USES
    UCalcAssocStructure;

TYPE
    //structure de rgles (infos internes)
    TAssocRuleStructureOutlier = class(TAssocRuleStructure)
    protected
    ///rcuprer les infos
    procedure   RecupInfos(calcRule: TCalcRuleFromItemset); override;
    public
    ///les infos sur les effectifs
    n, nA, nC, nAC, nANotC: integer;
    ///description de la rgle
    function    DescriptionHTML(): string; override;
    end;

    //gestion de la liste des rgles
    TLstAssocRulesOutlier = class(TLstAssocRules)
    public
    function    DescriptionHTML(): string; override;
    end;

implementation

USES
    Sysutils, UStringAddBuffered, UConstConfiguration;

{ TAssocRuleStructureOutlier }

function TAssocRuleStructureOutlier.DescriptionHTML: string;
var s: string;
begin
 s:= self.ShortDescriptionHTML();
 //infos
 s:= s + format('<TD align=right>%d</TD><TD align=right>%d</TD><TD align=right>%d</TD>',[nA,nAC,nANotC]);
 s:= s + format('<TD align=right>(%.3f ; %.3f ; %.3f)</TD>',[FLift,FSupport,FConfiance]);
 //s:= s+'</TR>';
 //renvoyer le tout
 result:= s;
end;

procedure TAssocRuleStructureOutlier.RecupInfos(
  calcRule: TCalcRuleFromItemset);
begin
 inherited;
 n:= trunc(calcRule.BaseSize);
 nA:= calcRule.BitAnte.Count;
 nC:= calcRule.BitCons.Count;
 nAC:= calcRule.curItemset.Support;
 nANotC:= nA - nAC; 
end;

{ TLstAssocRulesOutlier }

function TLstAssocRulesOutlier.DescriptionHTML: string;
var buf: TBufString;
    i: integer;
    tmp: string;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT);
 buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH colspan=7>Number of rules : %d</TH></TR>',[self.count]));
 buf.AddStr(format(HTML_TABLE_COLOR_HEADER_GRAY+'<TD>N</TD><TD %s>Antecedent</TD><TD %s>Consequent</TD><TD align=right>n(A)</TD><TD align=right>n(AC)</TD><TD align=right>n(A)-n(AC)</TD><TD align=right>(Lift. ; Supp. ; Conf.)</TD></TR>',[HTML_BGCOLOR_HEADER_BLUE,HTML_BGCOLOR_HEADER_GREEN]));
 for i:= 0 to pred(self.count) do
  begin
   tmp:= format('%s<TD>%d</TD>%s</TR>',[HTML_TABLE_COLOR_DATA_GRAY,succ(i),self.getRule(i).DescriptionHTML]);
   buf.AddStr(tmp);
  end;
 buf.AddStr('</table>');
 //suite normale
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free;
end;


end.
