(**********************************************************************)
(* UCompAssocFreqItemsets.pas - Copyright (c) 2011 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Gnration des itemsets frquents - Appel de apriori.exe de Christian BORGELT)
@author(Ricco)
@created(19/09/2011)

Le dispositif est totalement identique  celui de UCompAssocRulePrefixTreeCB.pas

Seul les paramtres d'appels sont modifis.

}
unit UCompAssocFreqItemsets;

interface

USES
    Forms, Classes, IniFiles,
    UCompDefinition,
    UOperatorDefinition,
    UCompAssociationRuleDefinition,
    UDatasetExamples,
    UCalcAssocStructure,
    UCompAssocRuleAPriori,
    DosCommand,
    UCompAssocRulePrefixTreeCB;

TYPE
    //gnrateur
    TGenCompAssocFreqItemsets = class(TGenCompAssocRulePT)
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLCompAssocFreqItemsets = class(TMLCompAssocRulePT)
    protected
    function    getClassOperator: TClassOperator; override;
    end;

    //oprateur
    TOpAssocFreqItemsets = class(TOpAssocAPrioriPT)
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    function    getClassAssocRule(): TClassCalcAssocRule; override;
    function    getNewForm(): TForm; override;
    function    getPrmCmdLine(): string; override;
    end;

    TEnumTypeItemset = (tiFrequent, tiClosed, tiMaximal, tiGenerators);

CONST
    PRM_TYPE_ITEMSET : array[TEnumTypeItemset] of char = ('s','c','m','g');
    DESC_TYPE_ITEMSET : array[TEnumTypeItemset] of string = ('frequent','closed','maximal','generators');

TYPE
    //paramtrage
    TOpPrmAssocFreqItemsets = class(TOpPrmAssocAPrioriPT)
    typeItemset: TEnumTypeItemset;//type d'itemset  produire
    minLength: integer;//longueur min des itemsets  produire, dfaut = 2
    maxSupport: double;//support max. des rgles, dfaut 1.0
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    function    getHTMLParameters(): string; override;
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    end;

implementation

uses
    Sysutils, UFrmOpViewAssocFreqItemsets, UFrmBaseOperator,
  UDlgOpPrmAssocFreqItemsets;

{ TGenCompAssocFreqItemsets }

function TGenCompAssocFreqItemsets.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompAssocFreqItemsets;
end;

{ TMLCompAssocFreqItemsets }

function TMLCompAssocFreqItemsets.getClassOperator: TClassOperator;
begin
 result:= TOpAssocFreqItemsets;
end;

{ TOpAssocFreqItemsets }

function TOpAssocFreqItemsets.getClassAssocRule: TClassCalcAssocRule;
begin
 result:= nil;
end;

function TOpAssocFreqItemsets.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmAssocFreqItemsets;
end;

function TOpAssocFreqItemsets.getNewForm: TForm;
begin
 result:= TfrmOpViewAssocFreqItemsets.CreateFromOperator(self);
end;

function TOpAssocFreqItemsets.getPrmCmdLine: string;
var cmdLine: string;
    prm: TOpPrmAssocFreqItemsets;
begin
 prm:= prmOp as TOpPrmAssocFreqItemsets;
 cmdLine:= '';
 //-t# pour le type d'itemset, -y pour la suppression des itemsets non-frquents
 cmdLine:= cmdLine + format(' -t%s -y',[PRM_TYPE_ITEMSET[prm.typeItemset]]);
 //-m# pour le card. min, -n# pour le cardinal max,
 cmdLine:= cmdLine + format(' -m%d -n%d',[prm.MinLength,prm.MaxRuleLength]);
 //-s# pour le support min, -S# pour le support max.
 cmdLine:= cmdLine + format(' -s%d -S%d',[trunc(100.0*prm.MinSupport),trunc(100.0*prm.MaxSupport)]);
 //-k# pour sparateur d'items
 cmdLine:= cmdLine + ' -k" /\ "';
 //-v"  (xxx)" pour formatter les sorties
 cmdLine:= cmdLine + ' -v" (%7s)"'; //support itemset
 //
 result:= cmdLine;
end;

{ TOpPrmAssocFreqItemsets }

function TOpPrmAssocFreqItemsets.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmAssocFreqItemsets.CreateFromOpPrm(self);
end;

function TOpPrmAssocFreqItemsets.getHTMLParameters: string;
begin
 result:= '';
end;

procedure TOpPrmAssocFreqItemsets.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 typeItemset:= TEnumTypeItemset(prmINI.ReadInteger(prmSection,'target_type',ord(tiFrequent)));
 minLength:= prmINI.ReadInteger(prmSection,'min_length',minLength);
 maxSupport:= prmINI.ReadFloat(prmSection,'max_support',maxSupport);
end;

procedure TOpPrmAssocFreqItemsets.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(typeItemset,sizeof(typeItemset));
 prmStream.ReadBuffer(minLength,sizeof(minLength));
 prmStream.ReadBuffer(maxSupport,sizeof(maxSupport));
end;

procedure TOpPrmAssocFreqItemsets.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'target_type',ord(typeItemset));
 prmINI.WriteInteger(prmSection,'min_length',minLength);
 prmINI.WriteFloat(prmSection,'max_support',maxSupport);
end;

procedure TOpPrmAssocFreqItemsets.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(typeItemset,sizeof(typeItemset));
 prmStream.WriteBuffer(minLength,sizeof(minLength));
 prmStream.WriteBuffer(maxSupport,sizeof(maxSupport));
end;

procedure TOpPrmAssocFreqItemsets.SetDefaultParameters;
begin
 inherited;
 OutRule:= 'outfreq.rul';
 typeItemset:= tiFrequent;
 minLength:= 2;
 maxSupport:= 1.0;
end;

initialization
 RegisterClass(TGenCompAssocFreqItemsets);
end.
