unit UFrmOpViewAssocFreqItemsets;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, UFrmBaseOperator, Menus, ActnList, ComCtrls, ExtCtrls, StdCtrls,
  UCompAssocFreqItemsets, Contnrs, VirtualTrees;

type
  TfrmOpViewAssocFreqItemsets = class(TFrmBaseOperator)
    tabItemsets: TTabSheet;
    Panel1: TPanel;
    memoConsole: TMemo;
    Panel3: TPanel;
    Splitter1: TSplitter;
    PanelVTView: TPanel;
    panelRules: TPanel;
    procedure FormDestroy(Sender: TObject);
  private
    { Dclarations prives }
    OpAssoc: TOpAssocFreqItemsets;
    lstItemsets: TObjectList;
    vtView: TVirtualDrawTree;
    procedure readItemsets(Sender: TObject; ExitCode: LongWord);
    procedure parseItemsets(prmFile: string);
    procedure prepareVirtualTreeView();
    procedure vtPaint(Sender: TBaseVirtualTree;
    TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
    CellRect: TRect);
    procedure vtHeaderClick(Sender: TVTHeader; Column: TColumnIndex; Button: TMouseButton; Shift: TShiftState; X,
    Y: Integer);
  public
    { Dclarations publiques }
    procedure   PrepareView(); override;
  end;

var
  frmOpViewAssocFreqItemsets: TfrmOpViewAssocFreqItemsets;

implementation

{$R *.dfm}

uses
    DosCommand, Math, ULogFile;

TYPE
    TItemset = class
    number: integer;
    description: string;
    support: double;
    constructor create(prmLine: string; num: integer);
    end;

{ TfrmOpViewAssocFreqItemsets }

procedure TfrmOpViewAssocFreqItemsets.parseItemsets(prmFile: string);
var F: TextFile;
    sLine: string;
    rule: TItemset;
begin
 //vider la liste actuelle, si elle existe
 lstItemsets.Clear();
 TRY
 //lire les rgles
 AssignFile(F,prmFile);
 Reset(F);
 while not(eof(f)) do
  begin
   readln(F,sLine);
   //crer la rgle
   rule:= TItemset.create(sLine,lstItemsets.Count);
   //ajouter dans la liste
   lstItemsets.Add(rule);
  end;
 FINALLY
 CloseFile(F);
 END;
 //affichage
 self.panelRules.Caption:= format('ITEMSETS [#%d itemsets loaded]',[lstItemsets.Count]);
 //synchroniser la treeview avec la liste
 //plus simplement, en une seul fois
 vtView.RootNodeCount:= lstItemsets.Count;
end;

procedure TfrmOpViewAssocFreqItemsets.PrepareView;
begin
 inherited;
 //liste des itemsets
 lstItemsets:= TObjectList.Create(TRUE);
 //prparer le virtual treeview
 self.prepareVirtualTreeView();
 //se brancher sur l'oprateur
 OpAssoc:= Operator as TOpAssocFreqItemsets;
 //vrifier s'il a dj t excut ou pas...
 if assigned(OpAssoc) and (OpAssoc.DosCommand.ThreadStatus <> dtsSuccess)
  then
   begin
    //alors brancher les derniers lments
    OpAssoc.DosCommand.OnTerminated:= self.readItemsets;
    OpAssoc.DosCommand.OutputLines:= self.memoConsole.Lines;
    //lancer rellement l'excution
    OpAssoc.DosCommand.Execute();
   end
  else readItemsets(NIL,0);
end;

procedure TfrmOpViewAssocFreqItemsets.prepareVirtualTreeView;
var vtc: TVirtualTreeColumn;
begin
 //crer
 vtView:= TVirtualDrawTree.Create(panelVTView);
 vtView.Parent:= panelVTView;
 vtView.Align:= alClient;
 //brancher les gestionnaires d'vnements 
 vtView.OnAfterCellPaint:= self.vtPaint;
 vtView.OnHeaderClick:= self.vtHeaderClick;
 vtView.Header.Style:=hsFlatButtons; //add by GongYu,look better
 //configurer, crer les colonnes
 vtc:= vtView.Header.Columns.Add();
 vtc.Text:= 'N';
 vtc.Width:= 30;
 vtc:= vtView.Header.Columns.Add();
 vtc.Text:= 'Description';
 vtc.Width:= 550;
 vtc:= vtView.Header.Columns.Add();
 vtc.Text:= 'Support';
 vtView.Header.Options:= vtView.Header.Options + [hoVisible];
 //options d'affichage
 vtView.TreeOptions.PaintOptions:= vtView.TreeOptions.PaintOptions - [toShowTreeLines] + [toShowHorzGridLines];
 vtView.TreeOptions.SelectionOptions:= vtView.TreeOptions.SelectionOptions + [toFullRowSelect];
 vtView.Colors.FocusedSelectionBorderColor:= clInactiveCaptionText;
 vtView.Colors.FocusedSelectionColor:= clInfoBk;
end;

procedure TfrmOpViewAssocFreqItemsets.readItemsets(Sender: TObject; ExitCode: LongWord);
var sRules: string;
begin
 if assigned(OpAssoc)
  then
   begin
     sRules:= (OpAssoc.PrmOp as TOpPrmAssocFreqItemsets).OutRule;
     if FileExists(sRules)
      then self.parseItemsets(sRules)
      else self.PanelRules.Caption:= 'Itemsets [-- NO ITEMSETS --]';
   end
  else self.PanelRules.Caption:= 'Itemsets [-- NO CURRENT OPERATOR AVAILABLE --]';
end;

function SortOnNumber(Item1, Item2: Pointer): Integer;
begin
 if (TItemset(Item1).number > TItemset(Item2).number)
  then result:= +1
  else if (TItemset(Item1).number < TItemset(Item2).number)
       then result:= -1
       else result:= 0;
end;

function SortOnDescription(Item1, Item2: Pointer): Integer;
begin
 if (TItemset(Item1).description > TItemset(Item2).description)
  then result:= +1
  else if (TItemset(Item1).description < TItemset(Item2).description)
       then result:= -1
       else result:= 0;
end;

function SortOnSupport(Item1, Item2: Pointer): Integer;
begin
 if (TItemset(Item1).support > TItemset(Item2).support)
  then result:= -1
  else if (TItemset(Item1).support < TItemset(Item2).support)
       then result:= +1
       else result:= 0;
end;

procedure TfrmOpViewAssocFreqItemsets.vtHeaderClick(Sender: TVTHeader;
  Column: TColumnIndex; Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
 case column of
 1: lstItemsets.Sort(sortOnDescription);
 2: lstItemsets.Sort(sortOnSupport);
 else lstItemsets.Sort(sortOnNumber);
 end;
 //mettre un signe distinctif pour le tri
 sender.SortColumn:= Column;
 if (Column < 2)
  then sender.SortDirection:= sdAscending
  else sender.SortDirection:= sdDescending;
 sender.Options:= sender.Options + [hoShowSortGlyphs];
 //mj l'affichage, bien videmment
 self.vtView.Refresh();
end;

procedure TfrmOpViewAssocFreqItemsets.vtPaint(Sender: TBaseVirtualTree;
  TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
  CellRect: TRect);
var itemset: TItemset;
    fullWidth, width, height: integer;
    value: double;
begin
 if assigned(lstItemsets)
  then
   begin
    if (Node.Index < cardinal(lstItemsets.Count))
     then
      begin
       itemset:= lstItemsets.Items[Node.Index] as TItemset;
       TargetCanvas.Font.Color:= clBlack;
       case Column of
       0: TargetCanvas.TextOut(CellRect.Left+2,CellRect.Top+3,inttostr(itemset.number));
       1: TargetCanvas.TextOut(CellRect.Left+2,CellRect.Top+3,itemset.description);
       2: begin
           value:= itemset.support;
           fullWidth:= CellRect.Right - CellRect.Left - 2;
           width:= TRUNC(1.0*value*fullWidth/100.0);
           height:= CellRect.Bottom - CellRect.Top - 2;
           TargetCanvas.Brush.Style:= bsSolid;
           TargetCanvas.Brush.Color:= $00CBFFCB;
           TargetCanvas.Pen.Color:= TargetCanvas.Brush.Color;
           TargetCanvas.Rectangle(CellRect.Left+1,CellRect.Top+1,CellRect.Left+1+width,CellRect.Top+1+height);
           TargetCanvas.Brush.Style:= bsClear;
           TargetCanvas.TextOut(CellRect.Left+2,CellRect.Top+3,format('%.1f',[itemset.support]));
          end;
       end;
      end;
   end;
end;

{ TItemset }

constructor TItemset.create(prmLine: string; num: integer);
var p: integer;
    s: string;
    err: integer;
    value: string;
    supp: double;
begin
 inherited Create();
 number:= succ(num);
 //dbut parsing...
 s:= prmLine;
 //description de l'itemset
 p:= pos('(',s);
 description:= copy(s,1,p-2);
 delete(s,1,p);
 //antecedent
 p:= pos(')',s);
 value:= copy(s,1,p-1);
 TRY
 val(value,supp,err);
 support:= 100.0 * supp;
 EXCEPT
 support:= 0.0;
 END;
end;

procedure TfrmOpViewAssocFreqItemsets.FormDestroy(Sender: TObject);
begin
 lstItemsets.Free();
 inherited;
end;

end.
