(******************************************************************)
(* UCompClusteringHAC.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(CAH - Stratgie mixte dcrite dans Lebart et al. (2000) - Mthode WARD)
@author(Ricco)
@created(07/03/2004)

new -- 16/06/2006 -- On gre maintenant les deux approches, la CAH mixte et la CAH standard
Dans le premier cas, on prend comme point de dpart les pr-clusters fournie par une mthode quelconque.
Dans le second cas, le point de dpart est l'observation.

new -- 17/06/2006 -- Dessin du dendrogramme
}
unit UCompClusteringHAC;

interface

USES
        Contnrs, Forms, Classes, IniFiles,
        UCompDefinition,
        UOperatorDefinition,
        UCompClusteringDefinition,
        UDatasetDefinition, UDatasetExamples,
        UCalcStatDes;

TYPE
       {gnrateur de composant}
        TGenClusHAC = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant CAH}
        TMLCompClusHAC = class(TMLCompClustering)
        protected
        function    getClassOperator: TClassOperator; override;
        function    getGenericAttName(): string; override;
        end;

        {oprateur CAH}
        TOpClusHAC = class(TOperatorClusteringContinue)
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    getClassCalcClustering(): TClassCalcClustering; override;
        function    CheckAttributes(): boolean; override;
        function    CheckOthers(): boolean; override;
        function    getNewForm(): TForm; override;
        //new -- 15/11/2012 - mixer la description de l'arbre et de la variable de sortie
        function    getHTMLClustersDescriptions(): string; override;
        end;

        {paramtre oprateur CAH}
        TOpPrmHAC = class(TOpPrmClustering)
        protected
        {Mode de dtermination de la coupure optimale : 0->automatique, 1->manuel}
        FModeBestSelection: integer;
        {si choix manuel, nombre de clusters}
        FNbClusters: integer;
        {normalisation pour le calcul des distances}
        FNormalization: integer;
        {montrer la structure hirarchique}
        FShowTreeStructure: boolean;
        {montrer les carts dans les indices de niveau}
        FShowIndexSelection: boolean;
        function    CreateDlgParameters(): TForm; override;
        procedure   SetDefaultParameters(); override;
        public
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        function    getHTMLParameters(): string; override;
        property    ModeBestSelection: integer read FModeBestSelection write FModeBestSelection;
        property    NbClusters: integer read FNbClusters write FNbClusters;
        property    Normalization: integer read FNormalization write FNormalization;
        property    ShowTreeStructure: boolean read FShowTreeStructure write FShowTreeStructure;
        property    ShowIndexSelection: boolean read FShowIndexSelection write FShowIndexSelection;
        end;

        {dclaration forward}
        TCalcHAC = class;

        {Un sommet de CAH}
        TNodeHAC = class(TObject)
        private
        {oprateur de calcul}
        FCalc: TCalcHAC;
        {pre}
        FDad: TNodeHAC;
        {enfants gauche et droite}
        FChildL,FChildR: TNodeHAC;
        {les informations locales}
        FStats: TLstCalcStatDesContinuous;
        {numro de sommet pour suivre les fusions}
        FNumNode: integer;
        {Distance d'agrgation}
        FDist: double;
        {coordonne sur l'axe des X, pour le dessin du dendrogramme}
        FXPos: double;
        {new -- 18/07/2006 -- valeur de l'tiquette si attribut de rfrence utilise}
        FValueAttRef: TTypeDiscrete;
        {new -- 18/07/2006 -- indique si le noeud est une solution}
        FIsCluster: boolean;
        public
        {construire un noeud}
        constructor  Create(prmCalc: TCalcHAC; valueAttRef: TTypeDiscrete = 0);
        {dtruire}
        destructor   Destroy(); override;
        {mj des donnes courantes}
        procedure    Refresh(prmExamples:  TExamples);
        {calculer les distances d'agrgation avec un autre noeud}
        function     MergeDistance(other: TNodeHAC): double;
        {calculer la distance avec un individu}
        function     CalcDistance(example: integer): double;
        {tester si le noeud est une feuille}
        function     isLeaf(): boolean;
        {rcuprer le label qui est l'effectif du sommet en ralit}
        {new -- 12/10/2006 -- et l'tiquette du cluster de base si c'est une feuille}
        function     getLabel(): string;
        {new -- 12/10/2006 -- recursif pour rcuprer les labels sur les feuilles d'un noeud}
        procedure    getLeavesLabels(var s: string);
        //new -- 15/11/2012 -- obtenir le nombre d'observations
        function     getNbInstances(): integer; 
        //proprits
        property     XPos: double read FXPos;
        property     YPos: double read FDist;
        property     ChildL: TNodeHAC read FChildL;
        property     ChildR: TNodeHAC read FChildR;
        property     isCluster: boolean read FIsCluster;
        end;

        {Structure complte de CAH}
        TSetNodeHAC = class(TObject)
        private
        {Calculateur HAC}
        FCalcHAC: TCalcHAC;
        {racine de l'arbre}
        FRootNode: TNodeHAC;
        {tous les noeuds - liste propritaire}
        FLstAllNodes: TObjectList;
        {les noeuds  fusionner - les noeuds orphelins - liste non propritaire}
        FLstMergeNodes: TObjectList;
        {les feuilles de l'arbre - liste non propritaire}
        FLstLeaves: TObjectList;
        {Indicateur global du dernier numro de noeud ajout}
        FLastNumNode: integer;
        {Tableau de suivi des fusions ascendantes}
        FLstNewNodes: TObjectList;
        protected
        {ajouter un noeud dans la liste initiale}
        procedure   AddLeaf(node: TNodeHAC);
        public
        {cration de la structure}
        constructor Create(prmCalc: TCalcHAC);
        {dtruire la structure}
        destructor  Destroy(); override;
        {initialiser les noeuds}
        procedure   InitializeNodes(prmExamples: TExamples);
        {fusionner deux noeuds et l'ajouter dans la structure}
        procedure   MergeNodes(nodeL,nodeR: TNodeHAC; prmDist: double);
        {calculer les coordonnes des X pour les feuilles}
        procedure   setLeavesCoords(node: TNodeHAC; var curCoord: double);
        {calculer les coordonnes des noeuds intrieurs de l'arbre}
        procedure   setNodesCoords(node: TNodeHAC);
        {rcuprer le nombre de noeuds}
        function    getCountNodes(): integer;
        {rcuprer le noeud numro index}
        function    getNode(index: integer): TNodeHAC;
        //proprits
        property    RootNode: TNodeHAC read FRootNode;
        end;

        {tableau dynamique pour la pondration}
        TTabHACAttWeight = array of double;

        {oprateur de calcul CAH}
        TCalcHAC = class(TCalcClusteringContinue)
        private
        {inertie totale}
        FTSS: double;
        {Indice du palier max.}
        FIdMaxPalier: integer;
        {Attribut dfinissant les noeuds de 1er niveau}
        FAttRef: TAttribute;
        {Tableau de pondration par la variance}
        FVariances: TTabHACAttWeight;
        {tableau des hauteurs des noeuds de l'arbre}
        FTabIndex: array of double;
        {tableau des diffrences x-imes pour la dtection de palliers}
        FTabDifIndex: array of double;
        {structure d'arbre}
        FSetNodeHAC:TSetNodeHAC;
        {les noeuds de la solution adopte}
        FLstNodesSolution: TObjectList;
        {initialiser les noeuds}
        procedure   InitializeNodes(prmExamples: TExamples);
        {calculer les coordonnes pour le dessin du dendrogramme}
        procedure   computeNodeCoords();
        protected
        function    getHTMLClustering(): string; override;
        function    GetHTMLEvaluation(): string; override;
        procedure   detectPalier();
        procedure   buildBestNodes();
        public
        destructor  Destroy(); override;
        procedure   BuildClusters(prmExamples: TExamples); override;
        function    SetClusterExample(example: integer): TTypeDiscrete; override;
        procedure   FillClusAttDef(); override;
        property    AttRef: TAttribute read FAttRef;
        property    SetNodeHAC: TSetNodeHAC read FSetNodeHAC;
        end;

implementation

uses
        SysUtils, UConstConfiguration, ULogFile, UDlgOpClusHAC,
        UDlgBaseOperatorParameter, UFrmOpClusHAC, Math, StrUtils;

const
        //nombre de feuilles min. pour commencer une CAH mixte (5 min. pour pouvoir choisir entre plusieurs solutions)
        MIN_LEAF_HAC = 5;

        //dcalage de coordonnes des feuilles
        DECALAGE_COORD_LEAF : double = 10.0;

        

{ TGenClusHAC }

procedure TGenClusHAC.GenCompInitializations;
begin
 FMLComp:= mlcClustering;
end;

function TGenClusHAC.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompClusHAC;
end;

{ TMLCompClusHAC }

function TMLCompClusHAC.getClassOperator: TClassOperator;
begin
 result:= TOpClusHAC;
end;

function TMLCompClusHAC.getGenericAttName: string;
begin
 result:= 'HAC';
end;

{ TOpClusHAC }

function TOpClusHAC.CheckAttributes: boolean;
var ok: boolean;
begin
 (*
 result:= inherited CheckAttributes()
          and (Targets.Count=1)
          and (Targets.isAllCategory(caDiscrete));
 *)
 //new -- 16/06/2006 -- on va grer aussi le cas de la CAH standard
 //tester le cas standard, descripteurs tous continus
 ok:= inherited CheckAttributes();
 //vrifier s'il y a des targets
 if ok and (Targets.Count > 0)
  //si oui, il doit tre unique et discret
  then ok:= (Targets.Count=1) and (Targets.isAllCategory(caDiscrete));
 //
 result:= ok;
end;

function TOpClusHAC.CheckOthers: boolean;
var ok: boolean;
begin
 ok:= true;
 //vrifier le nombre min. de noeuds feuilles dans le cas de CAH mixte
 if (self.Targets.Count = 1)
  then ok:= (self.Targets.Attribute[0].nbValues >= MIN_LEAF_HAC);
 //ok ?
 result:= ok;
end;

function TOpClusHAC.getClassCalcClustering: TClassCalcClustering;
begin
 result:= TCalcHAC;
end;

function TOpClusHAC.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmHAC;
end;

function TOpClusHAC.getHTMLClustersDescriptions: string;
var s: string;
    k: integer;
    calc: TCalcHAC;
begin
 calc:= CalcClustering as TCalcHAC;
 //rsultat gnrique, bilan sur les clusters gnrs
 s:= '<P><H3>Clustering results</H3>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TD>Clusters</TD><TH>From the<br>dendrogram</TH><TH>After one-pass<br>relocation</TH></TR>';
 for k:= 1 to StatCluster.TabFreq.Size do
  s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>cluster n%d</TD><TD align=right>%d</TD><TD align=right>%d</TD></TR>',
      [k,(calc.FLstNodesSolution.Items[pred(k)] as TNodeHAC).getNbInstances(),StatCluster.TabFreq.Value[k]]); 
 s:= s+'</table>';
 result:= s;
end;


function TOpClusHAC.getNewForm: TForm;
begin
 result:= TFrmOpViewClusHAC.CreateFromOperator(self);
end;

{ TCalcHAC }

procedure TCalcHAC.buildBestNodes;
var i: integer;
    node: TNodeHAC;
    tmp: string;
    j: integer;
begin
 //ide est trs simple - on part de la racine, et on descend en se calant sur la hauteur
 //on ajoute et on extrait itrativement les sommets jusqu' ce que le nombre de solutions soit gal  celui demand !
 FLstNodesSolution:= TObjectList.Create(FALSE);
 FLstNodesSolution.Add(FSetNodeHAC.FRootNode);
 for i:= 1 to FIdMaxPalier+1 do
  begin
   node:= FSetNodeHAC.FLstNewNodes.Items[FSetNodeHAC.FLstNewNodes.Count-i] as TNodeHAC;
   FLstNodesSolution.Extract(node);
   FLstNodesSolution.Add(node.FChildL);
   FLstNodesSolution.Add(node.FChildR);
  end;
 //new -- indiquer au noeud qu'il est un cluster
 for i:= 0 to pred(FLstNodesSolution.Count) do
  (FLstNodesSolution.Items[i] as TNodeHAC).FIsCluster:= TRUE;
 //temporaire - test - voir les moyennes conditionnelles
 for i:= 0 to pred(FLstNodesSolution.Count) do
  begin
   tmp:= '';
   for j:= 0 to pred((FLstNodesSolution.Items[i] as TNodeHAC).FStats.Count) do
    tmp:= tmp + format('%.7f ',[((FLstNodesSolution.Items[i] as TNodeHAC).FStats.Stat(j) as TCalcStatDesContinuous).Average]);
   TraceLog.WriteToLogFile(tmp);
  end;
end;

procedure TCalcHAC.BuildClusters(prmExamples: TExamples);
var i,j,i_min,j_min: integer;
    nodeI,nodeJ: TNodeHAC;
    d,min_d: double;
begin
 inherited BuildClusters(prmExamples);
 //initialiser les noeuds de la structure
 TRY
 self.InitializeNodes(prmExamples);
 //fusionner en remontant jusqu' la racine de l'arbre
 while (FSetNodeHAC.FLstMergeNodes.Count > 1) do
  begin
   i_min:= -1;
   j_min:= -1;
   min_d:= math.MaxDouble;
   //tester le croisement
   for i:= 0 to FSetNodeHAC.FLstMergeNodes.Count-2 do
    begin
     nodeI:= FSetNodeHAC.FLstMergeNodes.Items[i] as TNodeHAC;
     for j:= succ(i) to FSetNodeHAC.FLstMergeNodes.Count-1 do
      begin
       nodeJ:= FSetNodeHAC.FLstMergeNodes.Items[j] as TNodeHAC;
       d:= nodeI.MergeDistance(nodeJ)/(1.0*prmExamples.Size);
       if (d < min_d)
        then
         begin
          i_min:= i;
          j_min:= j;
          min_d:= d;
         end;
      end;
    end;
   //rcuprer les plus proches et fusionner
   nodeI:= FSetNodeHAC.FLstMergeNodes.Items[i_min] as TNodeHAC;
   nodeJ:= FSetNodeHAC.FLstMergeNodes.Items[j_min] as TNodeHAC;
   FSetNodeHAC.MergeNodes(nodeI,nodeJ,min_d);
  end;
 //le premier sommet est la racine du dendrogramme
 FSetNodeHAC.FRootNode:= FSetNodeHAC.FLstMergeNodes.Items[0] as TNodeHAC;
 TraceLog.WriteToLogFile(format('CAH -> examples sur la racine : %d, distance agregation = %.4f',[FSetNodeHAC.FRootNode.FStats.NbExamples,FSetNodeHAC.FRootNode.FDist]));
 //structure arbre ok - rechercher maintenant le meilleur cluster
 self.detectPalier();
 //forcer la solution si demande par l'utilisateur
 if ((self.PrmCalc as TOpPrmHAC).FModeBestSelection = 1)
  then self.FIdMaxPalier:= (self.PrmCalc as TOpPrmHAC).FNbClusters-2;
 //dtecter les noeuds du clustering dfinitif
 self.buildBestNodes();
 //new -- 17/06/2006 -- calculer les coordonnes des noeuds pour dessiner le dendrogramme
 self.computeNodeCoords();
 EXCEPT
  on e: exception do
   begin
    TraceLog.WriteToLogFile(format('[HAC] :: BuildClusters = %s',[e.Message]));
    raise Exception.Create(e.Message);
   end;
 END;
end;

procedure TCalcHAC.computeNodeCoords;
var curCoord: double;
begin
 curCoord:= 0.0;
 //numroter les feuilles
 self.FSetNodeHAC.setLeavesCoords(self.SetNodeHAC.FRootNode,curCoord);
 //numroter les noeuds
 self.FSetNodeHAC.setNodesCoords(self.SetNodeHAC.FRootNode);
end;

destructor TCalcHAC.Destroy;
begin
 SetLength(FTabDifIndex,0);
 SetLength(FVariances,0);
 if assigned(FLstNodesSolution)
  then FLstNodesSolution.Free;
 if assigned(FSetNodeHAC)
  then FSetNodeHAC.Free;
 inherited;
end;

procedure TCalcHAC.detectPalier;
var i: integer;
    node: TNodeHAC;
    max: double;
begin
 setLength(FTabDifIndex,FSetNodeHAC.FLstNewNodes.Count+1);
 setLength(FTabIndex,FSetNodeHAC.FLstNewNodes.Count+1);
 //au premier sommet, pas de partitionnement
 FTabIndex[0]:= 0;
 FTabDifIndex[0]:= 0;
 //tableau des distances d'agrgation
 for i:= 0 to pred(FSetNodeHAC.FLstNewNodes.Count) do
  begin
   node:= FSetNodeHAC.FLstNewNodes.Items[pred(FSetNodeHAC.FLstNewNodes.Count)-i] as TNodeHAC;
   FTabIndex[i+1]:= node.FDist;
  end;
 //calcul des diffrences 1res
 for i:= 1 to high(FTabDifIndex)-1 do
  FTabDifIndex[i]:= FTabIndex[i]-FTabIndex[succ(i)];
 //ne pas oublier le premier saut
 FTabDifIndex[high(FtabDifIndex)]:= (FSetNodeHAC.FLstNewNodes.Items[0] as TNodeHAC).FDist;
 //on peut dtecter le palier
 FIdMaxPalier:= -1;
 max:= -1.0 * math.MaxDouble;
 //for i:= low(FTabDifIndex) to high(FTabDifIndex)-1 do
 for i:= 1 to high(FTabDifIndex)-1 do
  begin
   //if (i > 0) and (FTabDifIndex[i] > max)
   if (i > 1) and (FTabDifIndex[i] > max) //new -- 12/02/2008 -- zapper la solution  2 clusters, trop vidente
    then
     begin
      max:= FTabDifIndex[i];
      FIdMaxPalier:= i;
     end;
  end;
 //new -- 10/02/2008 -- attention, correction du dcalage dans la recherche de la hauteur max.
 //sinon, la solution est errone
 FIdMaxPalier:= FIdMaxPalier - 1;
 TraceLog.WriteToLogFile(Format('[HAC] id max palier = %d',[FIdMaxPalier])); 
end;

procedure TCalcHAC.FillClusAttDef;
var k: integer;
begin
 FAttClus.LstValues.clear;
 for k:= 1 to FIdMaxPalier+2 do
  FAttClus.LstValues.getValue('c_hac_'+IntToStr(k));
end;

function TCalcHAC.getHTMLClustering: string;
var s: string;
    sIdMax: string;
    cumul: double;
    i: integer;
    node: TNodeHAC;
begin
 s:= '';
 //affichage de la structure de l'arbre ?
 if (self.PrmCalc as TOpPrmHAC).ShowTreeStructure
  then
   begin
    s:= s+'<P><H3>Tree structure</H3><BR>';
    s:= s+HTML_HEADER_TABLE_RESULT;
    s:= s+format('%s <th colspan="4">First level clusters</th></tr>',[HTML_TABLE_COLOR_HEADER_GRAY]);
    if assigned(FAttRef)
     then s:= s+format('%s <td>Attribute ref.</td><td align="right" colspan="3">%s</td></tr>',[HTML_TABLE_COLOR_DATA_GRAY,FAttRef.Name])
     else s:= s+format('%s <td>Attribute ref.</td><td align="right" colspan="3">%s</td></tr>',[HTML_TABLE_COLOR_DATA_GRAY,'-']);
    s:= s+format('%s <td align="left">Tree leaves</td><td align="right" colspan="3">%d</td></tr>',[HTML_TABLE_COLOR_DATA_GRAY,FSetNodeHAC.FLstLeaves.Count]);
    s:= s+format('%s <th>Node</th><th>Size</th><th>Childs</th><th>Index</th></tr>',[HTML_TABLE_COLOR_HEADER_GRAY]);
    for i:= 0 to pred(FSetNodeHAC.FLstLeaves.Count) do
     begin
      node:= FSetNodeHAC.FLstLeaves.Items[i] as TNodeHAC;
      s:= s+format('%s <td align="center">%d</td><td align="right">%d</td><td align="center">-</td><td align="center">-</td></tr>',[HTML_TABLE_COLOR_DATA_GRAY,node.FNumNode,node.FStats.NbExamples]);
     end;
    for i:= 0 to pred(FSetNodeHAC.FLstNewNodes.Count) do
     begin
      node:= FSetNodeHAC.FLstNewNodes.Items[i] as TNodeHAC;
      s:= s+format('%s <td align="center">%d</td><td align="right">%d</td><td align="right">(%d,%d)</td><td align="right">%8.4f</td></tr>',[HTML_TABLE_COLOR_DATA_GRAY,node.FNumNode,node.FStats.NbExamples,node.FChildL.FNumNode,node.FChildR.FNumNode,node.FDist]);
     end;
    s:= s+'</table>';
   end;
 //affichage de la slection du meilleur cluster ?
 if (self.PrmCalc as TOpPrmHAC).ShowIndexSelection
  then
   begin
    s:= s+'<P><H3>Best cluster selection</H3><BR>';
    s:= s+HTML_HEADER_TABLE_RESULT;
    s:= s+format('%s <th width=60>Clusters</th><th width=60>BSS ratio</th><th width=60>Gap</th></tr>',[HTML_TABLE_COLOR_HEADER_GRAY]);
    cumul:= 0.0;
    //for i:= 0 to FSetNodeHAC.FLstNewNodes.Count-2 do
    for i:= 0 to FSetNodeHAC.FLstNewNodes.Count do
     begin
      //indiquer la bonne solution
      if (i = FIdMaxPalier + 1)
       then sIdMax:= HTML_TABLE_COLOR_DATA_GREEN
       else sIdMax:= HTML_TABLE_COLOR_DATA_GRAY;
      //cumul des inerties expliques
      cumul:= cumul+FTabIndex[i];
      //31/01/2011 -- vrification cohrence somme indices de niveau = TSS
      //TraceLog.WriteToLogFile(format('[HAC] FTabIndex = %.4f ; Cumul = %.4f; TSS = %.4f',[FTabIndex[i],cumul,FTSS]));
      //envoyer la sauce
      s:= s+format('%s <td>%d</td><td align="right">%6.4f</td><td align="right">%6.4f</td></tr>',
                   [sIdMax,i+1,cumul/FTSS,FTabDifIndex[i]]);
      //15/11/2012 - vrif.
      if (sIdMax = HTML_TABLE_COLOR_DATA_GREEN)
       then TraceLog.WriteToLogFile(format('[HAC] WSS Optimal = %.4f',[FTSS-cumul]));
     end;
    s:= s+'</table>';
   end;
 //zoo...
 result:= s;
end;

function TCalcHAC.GetHTMLEvaluation: string;
begin
 result:= inherited getHTMLEvaluation();
end;

procedure TCalcHAC.InitializeNodes(prmExamples: TExamples);
var j: integer;
begin

 if (self.Targets.Count>0)
  then FAttRef:= self.Targets.Attribute[0] as TAttribute
  else FAttRef:= nil;

 //new -- 16/06/2006 -- on traite tous les cas maintenant
 if TRUE
  then 
   begin
    setLength(FVariances,self.Inputs.Count);
    //calculer la variance utilise pour la normalisation
    if ((self.PrmCalc as TOpPrmHAC).FNormalization = 1)
     then
      begin
       For j:= 0 to pred(self.Inputs.Count) do
        //FVariances[j]:= TCalcStatDesContinuous(self.StatsInputs.Stat(j)).Variance;
        //new -- 03/02/2011 -- on normalise avec la variance population (1/n) -- cf. calcul TSS plus bas quand on ne normalise pas
        FVariances[j]:= TCalcStatDesContinuous(self.StatsInputs.Stat(j)).VariancePop;
       //donnes normalise, l'inertie totale est gale au nombre de variables
       FTSS:= self.Inputs.Count;
      end
     else
      begin
       //non normalises - inertie totale = somme des variances avec un distance euclidienne simple
       FTSS:= 0.0;
       for j:= 0 to pred(self.Inputs.Count) do
        begin
         FVariances[j]:= 1.0;
         //FTSS:= FTSS+TCalcStatDesContinuous(self.StatsInputs.Stat(j)).Variance;
         //new - 31/01/2011 - attention, sinon la somme des indices de niveau ne fait pas le TSS (perceptible sur les petits effectifs)
         FTSS:= FTSS + TCalcStatDesContinuous(self.StatsInputs.Stat(j)).VariancePop;
        end;
      end;
   end;
 //construire le 1er niveau de la structure d'arbre
 FSetNodeHAC:= TSetNodeHAC.Create(self);
 FSetNodeHAC.InitializeNodes(prmExamples);
 FSetNodeHAC.FLastNumNode:= FSetNodeHAC.FLstLeaves.Count;
end;

function TCalcHAC.SetClusterExample(example: integer): TTypeDiscrete;
var i,i_min: integer;
    d,min_d: double;
    nodeRef,node: TNodeHAC;
    value: TTypeDiscrete;
    trouve: boolean;
begin
 i_min:= -1;
 //new -- 18/07/2006 -- modifier la procdure d'affectation selon que l'on ait des clusters de rf. en entre
 if (self.FAttRef = nil)
  then
   //procdure normale -- recherche du centre de gravit le plus proche
   begin
    min_d:= Math.MaxDouble;
    //trouver le noeud le plus proche
    for i:= 0 to pred(self.FLstNodesSolution.Count) do
     begin
      node:= self.FLstNodesSolution.Items[i] as TNodeHAC;
      d:= node.CalcDistance(example);
      if (d<min_d)
       then
        begin
         min_d:= d;
         i_min:= i;
        end;
     end;
   end
  else
   //chercher la feuille avec la mme tiquette
   begin
    value:= FAttRef.dValue[example];
    //puisqu'on est sr que les clusters de base sont crs dans l'ordre des modalits
    nodeRef:= self.FSetNodeHAC.FLstLeaves[pred(value)] as TNodeHAC;
    //
    if (nodeRef <> nil)
     then
      begin
       node:= nodeRef;
       trouve:= false;
       //remonter jusqu'au cluster qui est dans la solution
       while (node <> self.FSetNodeHAC.FRootNode) and not(trouve) do
        begin
         trouve:= node.FIsCluster;
         if not(trouve)
          then node:= node.FDad;
        end;
       //recherche une seule fois
       if trouve
        then i_min:= self.FLstNodesSolution.IndexOf(node);
      end;
   end;
 result:= succ(i_min);
end;

{ TSetNodeHAC }

procedure TSetNodeHAC.AddLeaf(node: TNodeHAC);
begin
 FLstAllNodes.Add(node);
 FLstMergeNodes.Add(node);
 FLstLeaves.Add(node);
 node.FNumNode:= FLstLeaves.Count;
 node.FDist:= 0.0;
end;

constructor TSetNodeHAC.Create(prmCalc: TCalcHAC);
begin
 inherited Create();
 FCalcHAC:= prmCalc;
 //les listes internes
 FLstAllNodes:= TObjectList.Create(TRUE);
 FLstMergeNodes:= TObjectList.Create(FALSE);
 FLstLeaves:= TObjectList.Create(FALSE);
 FLstNewNodes:= TObjectList.Create(FALSE);
end;

destructor TSetNodeHAC.Destroy;
begin
 FLstNewNodes.Free;
 FLstLeaves.Free;
 FLstMergeNodes.Free;
 FLstAllNodes.Free;
 inherited Destroy();
end;

function TSetNodeHAC.getCountNodes: integer;
begin
 result:= self.FLstAllNodes.Count;
end;

function TSetNodeHAC.getNode(index: integer): TNodeHAC;
begin
 result:= self.FLstAllNodes[index] as TNodeHAC;
end;

procedure TSetNodeHAC.InitializeNodes(prmExamples: TExamples);
var dispEx: TObjectList;
    ex: TExamples;
    j: integer;
    node: TNodeHAC;
begin
 //lister les individus par cluster de 1er niveau
 //new -- 16/06/2006 -- grer aussi la CAH standard
 //new -- 18/07/2006 -- attention, on affecte aussi des tiquettes aux noeuds, s'il y a un attribut de rfrence 
 if (FCalcHAC.AttRef <> nil)
  then
   begin
    dispEx:= prmExamples.DispatchExamples(FCalcHAC.AttRef);
    for j:= 0 to pred(dispEx.Count) do
     begin
      ex:= dispEx.Items[j] as TExamples;
      if (ex.Size>0)
       then
        begin
         //crer
         node:= TNodeHAC.Create(FCalcHAC,succ(j));
         node.Refresh(ex);
         //ajouter
         self.AddLeaf(node);
        end
       //new -- 26/05/2011 -- si le noeud est vide, on ne le rajoute pas et on provoque une erreur -- comme avant quoi...
       //sinon, on s'expose au problme de l'affectation sur l'chantillon test...
       else raise Exception.Create(format('some leaves are empty for HAC -- target attribute = %s',[FCalcHAC.AttRef.Name]));
     end;
    //dtruire
    dispEx.Free();
   end
  else
   begin
    for j:= 1 to prmExamples.Size do
     begin
      ex:= TExamples.Create(1);
      ex.Number[1]:= prmExamples.Number[j];
      node:= TNodeHAC.Create(FCalcHAC);
      node.Refresh(ex);
      //ajouter le noeud -- oubli -- bug. 01/08/2006
      self.AddLeaf(node);
      ex.free;// add by GongYu avoid memory leak
     end;
   end;
end;

procedure TSetNodeHAC.MergeNodes(nodeL, nodeR: TNodeHAC; prmDist: double);
var node: TNodeHAC;
    j: integer;
    st,stL,stR: TCalcStatDesContinuous;
    min,max,sum,sqsum: double;
    nb: integer;
begin
 node:= TNodeHAC.Create(self.FCalcHAC);
 //Numro du dernier sommet ajout et sa hauteur
 inc(self.FLastNumNode);
 node.FNumNode:= self.FLastNumNode;
 node.FDist:= prmDist;
 //affecter les donnes simples pour chaque variable j
 for j:= 0 to pred(node.FStats.Count) do
  begin
   //brancher
   st:= node.FStats.Stat(j) as TCalcStatDesContinuous;
   stL:= nodeL.FStats.Stat(j) as TCalcStatDesContinuous;
   stR:= nodeR.FStats.Stat(j) as TCalcStatDesContinuous;
   //calculer les stats locales
   if (stL.Min<stR.Min)
    then min:= stL.Min
    else min:= stR.Min;
   if (stL.Max>stR.Max)
    then max:= stL.Max
    else max:= stR.Max;
   sum:= stL.Sum+stR.Sum;
   sqsum:= stL.SumSquare+stR.SumSquare;
   nb:= stL.NbExamples+stR.NbExamples;
   st.SetValues(min,max,sum,sqsum,nb);
  end;
 //mj des listes internes et des branchements
 node.FChildL:= nodeL;
 node.FChildR:= nodeR;
 nodeL.FDad:= node;
 nodeR.FDad:= node;
 self.FLstAllNodes.Add(node);
 self.FLstNewNodes.Add(node);
 self.FLstMergeNodes.Add(node);
 self.FLstMergeNodes.Remove(nodeL);
 self.FLstMergeNodes.Remove(nodeR);
end;

procedure TSetNodeHAC.setLeavesCoords(node: TNodeHAC; var curCoord: double);
begin
 if not(node.isLeaf())
  then
   begin
    self.setLeavesCoords(node.FChildL,curCoord);
    self.setLeavesCoords(node.FChildR,curCoord);
   end
  else
   begin
    node.FXPos:= curCoord;
    curCoord:= curCoord + DECALAGE_COORD_LEAF;
   end;
end;

procedure TSetNodeHAC.setNodesCoords(node: TNodeHAC);
begin
 //doit-on en calculer les coordonnes ?
 if (node.FXPos < 0.0)
  then
   begin
    //tester si fils gauche n'est pas bien plac
    if (node.FChildL.FXPos < 0.0)
     then setNodesCoords(node.FChildL);
    //tester  droite
    if (node.FChildR.FXPos < 0.0)
     then setNodesCoords(node.FChildR);
    //se positionner au milieu de ses enfants
    node.FXPos:= 0.5 * (node.FChildL.FXPos + node.FChildR.FXPos);
   end;
end;

{ TNodeHAC }

function TNodeHAC.CalcDistance(example: integer): double;
var s,d: double;
    j: integer;
begin
 s:= 0.0;
 for j:= 0 to pred(self.FStats.Count) do
  begin
   d:= SQR(self.FCalc.Inputs.Attribute[j].cValue[example]-(self.FStats.Stat(j) as TCalcStatDesContinuous).Average);
   d:= d/self.FCalc.FVariances[j];
   s:= s+d;
  end;
 result:= s;
end;

constructor TNodeHAC.Create(prmCalc: TCalcHAC; valueAttRef: TTypeDiscrete = 0);
begin
 inherited Create();
 FCalc:= prmCalc;
 FDad:= NIL;
 FChildL:= NIL;
 FChildR:= NIL;
 FXPos:= -1.0;
 FValueAttRef:= valueAttRef;
 FIsCluster:= false;
 FStats:= TLstCalcStatDesContinuous.Create(FCalc.Inputs,NIL);
end;

destructor TNodeHAC.Destroy;
begin
 FStats.Free;
 inherited Destroy();
end;

function TNodeHAC.getLabel: string;
var s: string;
    l: integer;
begin
 //result:= Format('%d examples (%.1f%s)',[self.FStats.NbExamples,(100.0*self.FStats.NbExamples)/(1.0*self.FCalc.SetNodeHAC.RootNode.FStats.NbExamples),'%']);
 s:= Format('%d examples (%.1f%s)',[self.FStats.NbExamples,(100.0*self.FStats.NbExamples)/(1.0*self.FCalc.SetNodeHAC.RootNode.FStats.NbExamples),'%']);
 //new -- 12/10/2006 -- ajouter son tiquette, la valeur de la feuille si on a un attribut de rfrence
 if (self.FCalc.FAttRef <> nil)
  then
   begin
    l:= length(s);
    s:= s + #13#10;
    s:= s + DupeString('-',l);
    self.getLeavesLabels(s);
   end;
 //
 result:= s;
end;

procedure TNodeHAC.getLeavesLabels(var s: string);
begin
 if (self.FCalc.AttRef <> nil)
  then
   begin
    if not(self.isLeaf())
     then
      begin
       self.FChildL.getLeavesLabels(s);
       self.FChildR.getLeavesLabels(s);
      end
     else
      begin
       s:= s + #13#10;
       s:= s + self.FCalc.FAttRef.LstValues.getDescription(self.FValueAttRef);
      end;
   end;
end;

function TNodeHAC.getNbInstances: integer;
begin
 TRY
 result:= self.FStats.NbExamples;
 EXCEPT
 result:= 0;
 END;
end;

function TNodeHAC.isLeaf: boolean;
begin
 result:= (self.FChildL = nil) and (self.FChildR = nil);
end;

function TNodeHAC.MergeDistance(other: TNodeHAC): double;
var s,d, v: double;
    j: integer;
begin
 s:= 0.0;
 //distance brute
 for j:= 0 to pred(self.FStats.Count) do
  begin
   d:= SQR((self.FStats.Stat(j) as TCalcStatDesContinuous).Average-(other.FStats.Stat(j) as TCalcStatDesContinuous).Average);
   //d:= d / self.FCalc.FVariances[j];
   //new -- 07/11/2011 - scurisation
   v:= self.FCalc.FVariances[j];
   if (v > 0.0)
    then d:= d / v
    else d:= 0.0;//dsactivation de la colonne si la variance est nulle
   //suite...
   s:= s + d;
  end;
 //pondration par le poids des noeuds
 result:= (1.0*self.FStats.NbExamples*other.FStats.NbExamples)/(1.0*(self.FStats.NbExamples+other.FStats.NbExamples))*s;
end;

procedure TNodeHAC.Refresh(prmExamples: TExamples);
begin
 FStats.RefreshStat(prmExamples);
end;

{ TOpPrmHAC }

function TOpPrmHAC.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmClusHAC.CreateFromOpPrm(self);
end;

const
    STR_HAC_CLUS_DET : array[0..1] of string = ('Automatic','User defined');
    STR_HAC_NORMALIZATION : array[0..1] of string = ('None','Standardization');

function TOpPrmHAC.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2># clusters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Detection</TD><TD align=right>%s</TD></TR>',[STR_HAC_CLUS_DET[FModeBestSelection]]);
 if (FModeBestSelection = 1) then s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD># clusters</TD><TD align=right>%d</TD></TR>',[FNbClusters]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Data transformation</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Transformation</TD><TD align=right>%s</TD></TR>',[STR_HAC_NORMALIZATION[FModeBestSelection]]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Visualization</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Index selection</TD><TD align=right>%d</TD></TR>',[ord(FShowIndexSelection)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Tree structure</TD><TD align=right>%d</TD></TR>',[ord(FShowTreeStructure)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Anova per variable</TD><TD align=right>%d</TD></TR>',[ord(self.ShowAnovaDetails)]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmHAC.LoadFromINI(prmSection: string; prmINI: TMemIniFile);
begin
 inherited;
 FModeBestSelection:= prmINI.ReadInteger(prmSection,'clus_selection',FModeBestSelection);
 FNbClusters:= prmINI.ReadInteger(prmSection,'user_nb_clus',FNbClusters);
 FNormalization:= prmINI.ReadInteger(prmSection,'normalization',FNormalization);
 FShowTreeStructure:= prmINI.ReadBool(prmSection,'show_tree_structure',FShowTreeStructure);
 FShowIndexSelection:= prmINI.ReadBool(prmSection,'show_index_selection',FShowIndexSelection);
end;

procedure TOpPrmHAC.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(FModeBestSelection,sizeof(FModeBestSelection));
 prmStream.ReadBuffer(FNbClusters,sizeof(FNbClusters));
 prmStream.ReadBuffer(FNormalization,sizeof(FNormalization));
 prmStream.ReadBuffer(FShowTreeStructure,sizeof(FShowTreeStructure));
 prmStream.ReadBuffer(FShowIndexSelection,sizeof(FShowIndexSelection));
end;

procedure TOpPrmHAC.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'clus_selection',FModeBestSelection);
 prmINI.WriteInteger(prmSection,'user_nb_clus',FNbClusters);
 prmINI.WriteInteger(prmSection,'normalization',FNormalization);
 prmINI.WriteBool(prmSection,'show_tree_structure',FShowTreeStructure);
 prmINI.WriteBool(prmSection,'show_index_selection',FShowIndexSelection);
end;

procedure TOpPrmHAC.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(FModeBestSelection,sizeof(FModeBestSelection));
 prmStream.WriteBuffer(FNbClusters,sizeof(FNbClusters));
 prmStream.WriteBuffer(FNormalization,sizeof(FNormalization));
 prmStream.WriteBuffer(FShowTreeStructure,sizeof(FShowTreeStructure));
 prmStream.WriteBuffer(FShowIndexSelection,sizeof(FShowIndexSelection));
end;

procedure TOpPrmHAC.SetDefaultParameters;
begin
 inherited SetDefaultParameters();
 FModeBestSelection:= 0;
 FNbClusters:= 3;
 FNormalization:= 1;
 FShowTreeStructure:= false;
 FShowIndexSelection:= true;
 FShowAnova:= false;
end;

initialization
 RegisterClass(TGenClusHAC);
end.
