(*******************************************************************************)
(* UCompClusteringKMeansStrengthen.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(*******************************************************************************)

{
@abstract(Implmentation des K-Means -- Consolidation d'un pr-clustering existant)
@author(Ricco)
@created(14/11/2012)

Cet algo cherche  consolider une segmentation existante par la mthode des K-Means (centres mobiles).

Par rapport  l'algo standard, il a besoin d'une variable TARGET indiquant le regroupement initial.

1) Le nombre de groupes est donc dfini  l'avance et n'est pas paramtrable.
2) Il n'y a qu'un seul essai ici.
3) Le nombre d'itrations reste modifiable en revanche.

Dans les sorties, un croisement entre le regroupement initial et final sera affich. idem, on affichera le BSS avant et aprs consolidation.

}

unit UCompClusteringKMeansStrengthen;

interface

USES
        Forms,Classes,
        IniFiles,
        UCompDefinition,
        UOperatorDefinition,
        UCompClusteringDefinition,
        UDatasetDefinition,
        UDatasetExamples,
        UCalcStatDes,
        UCompClusteringKMeans,
        UCalcStatDesCrossTab,
        contnrs;

TYPE
        {gnrateur de rgression multiple}
        TGenClusKMeansStrengthen = class(TMLGenComp)
         protected
         procedure   GenCompInitializations(); override;
         public
         function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant K-Means}
        TMLCompClusKMeansStrengthen = class(TMLCompClustering)
         protected
         function    getClassOperator: TClassOperator; override;
         function    getGenericAttName(): string; override;
        end;

        {oprateur K-Means}
        TOpClusKMeansStrengthen = class(TOperatorClusteringContinue)
         protected
         function    getClassParameter: TClassOperatorParameter; override;
         function    getClassCalcClustering(): TClassCalcClustering; override;
         function    getHTMLClustersDescriptions(): string; override;
         function    CheckAttributes(): boolean; override;
        end;

        {paramtre oprateur K-Means}
        TOpPrmClusKMeansStrengthen = class(TOpPrmClusKMeans)
         protected
         function    CreateDlgParameters(): TForm; override;
         procedure   SetDefaultParameters(); override;
        end;

        {classe de calcul}
        TCalcClusKMeansStrengthen = class(TCalcClusKMeans)
        protected
        //attribut de rfrence dfinissant le regroupement initial
        FAttRef: TAttribute;
        //WSS initial
        FWSSInit: double;
        //tableau crois mesurant la cohrence des 2 partitions
        FCroisement: TCalcSDCrossTab;
        //initialization des clusters avec l'attribut de rfrence
        procedure   InitializeClusters(prmExamples: TExamples); override;
        public
        //cration des variables de sortie
        procedure   FillClusAttDef(); override;
        //evaluation de la partition en la comparant  la partition initiale
        procedure   EvaluateClustering(prmExamples: TExamples); override;
        //rien ici puisqu'il n'y a qu'un seul essai
        function    getHTMLClustering(): string; override;
        //destructeur
        destructor  Destroy; override;
        end;


implementation

uses
        Sysutils, UConstConfiguration, UDlgOpClusKMeansStrengthen, UStringsResources,
        UCalcRndGenerator, ULogFile, Math, UCalcStatDesConditionnalDesc;

{ TGenClusKMeansStrengthen }

procedure TGenClusKMeansStrengthen.GenCompInitializations;
begin
 FMLComp:= mlcClustering;
end;

function TGenClusKMeansStrengthen.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompClusKMeansStrengthen;
end;

{ TMLCompClusKMeansStrengthen }

function TMLCompClusKMeansStrengthen.getClassOperator: TClassOperator;
begin
 result:= TOpClusKMeansStrengthen;
end;

function TMLCompClusKMeansStrengthen.getGenericAttName: string;
begin
 result:= 'KMS';
end;

{ TOpClusKMeansStrengthen }

function TOpClusKMeansStrengthen.CheckAttributes: boolean;
begin
 result:= inherited CheckAttributes();
 result:= result and ((Targets.Count=1) and Targets.isAllCategory(caDiscrete));
end;

function TOpClusKMeansStrengthen.getClassCalcClustering: TClassCalcClustering;
begin
 result:= TCalcClusKMeansStrengthen;
end;

function TOpClusKMeansStrengthen.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmClusKMeansStrengthen; 
end;

//temporaire....
function TOpClusKMeansStrengthen.getHTMLClustersDescriptions: string;
var s: string;
    calc: TCalcClusKMeansStrengthen;
    optima: TStrucClustersKMeans;
    k: integer;
begin
 calc:= self.CalcClustering as TCalcClusKMeansStrengthen;
 optima:= calc.LstClustersSolutions.Items[pred(calc.IdMaxRatio)] as TStrucClustersKMeans;
 //valuation globale
 s:= '<P><H3>Global evaluation</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_GRAY
       + format('<TD>Within Sum of Squares</TD><TD align=right>%.4f</TD></TR>',[optima.getWss()]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY
       + format('<TD>Total Sum of Squares</TD><TD align=right>%.4f</TD></TR>',[calc.GlobalInertia]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY
       + format('<TD>R-Square</TD><TD align=right>%.4f</TD></TR>',[optima.ExplainedRatio]);
 s:= s + '</table>';
 //description des clusters
 s:= s + '<P><H3>Cluster size and WSS</H3>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_DATA_GRAY;
 s:= s+format('<TD>Clusters</TD><TH align=right>%d</TH><TH></TH></TR>',[StatCluster.TabFreq.Size]);
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Cluster</TH><TH>Description</TH><TH width=60>Size</TH><TH>WSS</TH></TR>';
 for k:= 1 to StatCluster.TabFreq.Size do
  s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>cluster n%d</TD><TD align=center>%s</TD><TD align=right>%d</TD><TD align=right>%.4f</TD></TR>',[k,StatCluster.Attribute.LstValues.GetDescription(k),StatCluster.TabFreq.Value[k],optima.Cluster[pred(k)].WSS]);
 s:= s+'</table>';
 //affichage du croisement avant et aprs optimisation
 s:= s + '<P><H3>Clusters comparison before and after relocation</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>WSS Comparison</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD width=80>Before</TD><TD align="right" width=80>%.4f</TD></TR>',[calc.FWSSInit]);
 s:= s + HTML_TABLE_COLOR_DATA_LAVENDER + format('<TD>After</TD><TD align="right">%.4f</TD></TR>',[optima.getWss()]); 
 s:= s + '</table><br>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD>'+calc.FCroisement.CrossTab.getHTMLResult(-1,'CROSSTAB: Row = Before vs. Column = After')+'</TD></TR>';
 s:= s + '</table>';
 //and then...
 result:= s;
end;

{ TOpPrmClusKMeansStrengthen }

function TOpPrmClusKMeansStrengthen.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmClusKMeansStrengthen.CreateFromOpPrm(self);
end;

procedure TOpPrmClusKMeansStrengthen.SetDefaultParameters;
begin
 inherited SetDefaultParameters();
 //FNbClusters:= 3; c'est la variable TARGET qui va dcider
 NbTrials:= 1;//un seul essai !!! non modifiable
 Normalization:= 1;
 AvgComputation:= 1;
 MaxIter:= 10;
end;

{ TCalcClusKMeansStrengthen }

destructor TCalcClusKMeansStrengthen.Destroy;
begin
 if assigned(FCroisement) then FreeAndNil(FCroisement);
 inherited Destroy();
end;

procedure TCalcClusKMeansStrengthen.EvaluateClustering(
  prmExamples: TExamples);
begin
 inherited EvaluateClustering(prmExamples);
 FCroisement:= TCalcSDCrossTab.Create(self.FAttRef,self.FAttClus,prmExamples,nil);
end;

procedure TCalcClusKMeansStrengthen.FillClusAttDef;
var k: integer;
begin
 FattClus.LstValues.clear;
 for k:= 1 to (PrmCalc as TOpPrmClusKMeans).NbClusters do
  //le fonctionnement particulier de getValue permet cette manip.
  FAttClus.LstValues.getValue('c_kms_'+IntToStr(k));
end;

function TCalcClusKMeansStrengthen.getHTMLClustering: string;
begin
 result:= '';
end;

procedure TCalcClusKMeansStrengthen.InitializeClusters(
  prmExamples: TExamples);
var prm: TOpPrmClusKMeansStrengthen;
    stats: TLstStatDesCondANOVA;
    statRef: TCalcStatDesDiscrete;
    clus: TOneClusterKMeans;
    k,j,i: integer;
    wssClus: double;
begin
 //forcment, il y est grce  CheckAttributes()
 FAttRef:= self.Targets.Attribute[0] as TAttribute;
 statRef:= TCalcStatDesDiscrete.Create(FAttRef,prmExamples);
 //calculer les stats. descriptives conditionnelles
 stats:= TLstStatDesCondANOVA.Create(nil,nil);
 for j:= 0 to pred(self.Inputs.Count) do
  stats.AddStat(TCalcSDCondDescANOVA.Create(self.Inputs.Attribute[j],FAttRef,prmExamples));
 //fixer le nombre de clusters
 prm:= PrmCalc as TOpPrmClusKMeansStrengthen;
 prm.NbClusters:= FAttRef.nbValues;
 //cration de la premire partition
 SetLength(FIdClusters,succ(prmExamples.Size));//tous  zro systmatiquement
 FCurClusters:= TStrucClustersKMeans.Create();
 FWSSInit:= 0.0;
 for k:= 1 to prm.NbClusters do
  begin
   //crer le cluster
   clus:= TOneClusterKMeans.Create(Inputs,prm,StatsInputs,-1);
   FCurClusters.AddCluster(clus);
   //fixer son centre de gravit  partir des moyennes conditionnelles de l'attribut de rfrence
   wssClus:= 0.0;
   for j:= 0 to pred(self.Inputs.Count) do
    begin
     clus.CurAvg[j]:= (stats.stat(j) as TCalcSDCondDescANOVA).StatCond[pred(k)].Average;
     wssClus:= wssClus + (stats.stat(j) as TCalcSDCondDescANOVA).StatCond[pred(k)].TSS / clus.Variances[j];
    end;
   FWSSInit:= FWSSInit + wssClus;
  end;
 TraceLog.WriteToLogFile(format('WSS init = %4f',[FWSSInit]));
 //destruction des objets temporaires
 statRef.Free();
 stats.Free();
 //tableau des inerties intra
 setLength(FWSS,prm.NbClusters);
 //affecter les individus aux clusters - ds la premire itration, le regroupement peut tre diffrent de la partition initiale
 self.BeginUpdate();
 for i:= 1 to prmExamples.Size do self.AffectExample(i,prmExamples.Number[i],TRUE);
 self.EndUpdate();
end;
(*
var k,idExample,i: integer;
    prm: TOpPrmClusKMeans;
    clus: TOneClusterKMeans;
begin
 prm:= PrmCalc as TOpPrmClusKMeans;
 SetLength(FIdClusters,succ(prmExamples.Size));//tous  zro systmatiquement
 FCurClusters:= TStrucClustersKMeans.Create();
 for k:= 1 to prm.NbClusters do
  begin
   //vu le mode de gnration de valeur alatoire, un doublon peut difficilement survenir
   idExample:= prmExamples.Number[succ(FRndGenClustering.IRanMarRange(prmExamples.Size))];
   //crer le cluster
   clus:= TOneClusterKMeans.Create(Inputs,prm,StatsInputs,idExample);
   FCurClusters.AddCluster(clus);
  end;
 //tableau des inerties intra
 setLength(FWSS,prm.NbClusters);
 //new -- 25/10/2008 -- affecter les individus aux clusters
 self.BeginUpdate();
 for i:= 1 to prmExamples.Size do self.AffectExample(i,prmExamples.Number[i],TRUE);
 self.EndUpdate();
end;
*)

initialization
 Classes.RegisterClass(TGenClusKMeansStrengthen);
end.
