(*******************************************************************)
(* UCompClusteringTreeCART.pas - Copyright (c) 2005 Ricco RAKOTOMALALA *)
(*******************************************************************)

{
@abstract(Arbre de Classification -- Extension du schma CART [Regression Tree] au cas endogne-multivarie)
@author(Ricco)
@created(09/09/2005)

Utilisation du cadre CART pour construire un arbre de classification. L'inspiration est la
mthode divisive monothtique de Lechevalier et Chavent (1999), la rsolution thorique est
inspire du chapitre 8 de Breiman et al.(1984).

Composant.

new -- 01/12/2009 -- la rgle de dtection base sur la pente est abandonne
>> au profit d'une rgle de dtection de "coude" trs simple... ::> pour version 1.4.34
}

unit UCompClusteringTreeCART;

interface

USES
        Contnrs, Forms, Classes, IniFiles,
        UCompDefinition,
        UOperatorDefinition,
        UCompClusteringDefinition,
        UDatasetDefinition, UDatasetExamples,
        UCalcStatDes, UCompClusteringTree;

TYPE
        {gnrateur de composant}
        TGenClusTreeCART = class(TGenClusTree)
                      public
                      function    GetClassMLComponent: TClassMLComponent; override;
                      end;

        {composant CT}
        TMLCompClusTreeCART = class(TMLCompClusTree)
                         protected
                         function    getClassOperator: TClassOperator; override;
                         function    getGenericAttName(): string; override;
                         end;

        {oprateur CT}
        TOpClusTreeCART = class(TOpClusTree)
                     protected
                     function    getClassParameter: TClassOperatorParameter; override;
                     function    getClassCalcClustering(): TClassCalcClustering; override;
                     function    getNewForm(): TForm; override;
                     end;

        {paramtre oprateur CT}
        TOpPrmClusTreeCART = class(TOpPrmClusTree)
                    protected
                    {taille du pruning set, en pourcentage de la taille de la base}
                    FPruningSetSize : integer;
                    {x-SE rule}
                    //FSERule: double;
                    //new -- 11/09/2005 -- critre d'arrt "loi du coude" -- pente de la courbe
                    //new -- 01/12/2009 -- deprecated, n'est plus utilis... mais on le laisse pour la compatibilit des fichiers...
                    //FThresholdSlope: double;
                    //new -- 05/01/2010 -- remplac par le paramtre delta - a ne pose pas de problmes dans le ".bdm"
                    FDelta: double;
                    {montrer ou pas toutes les squences d'arbres si > SEUIL_NB_SEQUENCES}
                    FShowAllTreeSeq: boolean;
                    function    CreateDlgParameters(): TForm; override;
                    procedure   SetDefaultParameters(); override;
                    public
                    //
                    procedure   LoadFromStream(prmStream: TStream); override;
                    procedure   SaveToStream(prmStream: TStream); override;
                    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
                    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
                    function    getHTMLParameters(): string; override;
                    //
                    property    PruningSetSize: integer read FPruningSetSize write FPruningSetSize;
                    //property    SERule: double read FSERule write FSERule;
                    property    Delta: double read FDelta write FDelta;
                    property    ShowAllTreeSeq: boolean read FShowAllTreeSeq write FShowAllTreeSeq;
                    end;


implementation

uses
     Sysutils,
     UCalcClusteringTreeCART, UDlgOpPrmClusteringTreeCART,
     UConstConfiguration, UFrmOpClusTree;

{ TGenClusTreeCART }

function TGenClusTreeCART.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompClusTreeCART;
end;

{ TMLCompClusTreeCART }

function TMLCompClusTreeCART.getClassOperator: TClassOperator;
begin
 result:= TOpClusTreeCART;
end;

function TMLCompClusTreeCART.getGenericAttName: string;
begin
 result:= 'CTP';
end;

{ TOpClusTreeCART }

function TOpClusTreeCART.getClassCalcClustering: TClassCalcClustering;
begin
 result:= TCalcClusTreeCTP;
end;

function TOpClusTreeCART.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmClusTreeCART;
end;

function TOpClusTreeCART.getNewForm: TForm;
begin
 result:= TfrmOpViewClusteringTree.CreateFromOperator(self);
end;

{ TOpPrmClusTreeCART }

function TOpPrmClusTreeCART.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmClusteringTreeCART.CreateFromOpPrm(self);
end;

function TOpPrmClusTreeCART.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Tree Parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Rnd generator</TD><TD width=40 align="right">%d</TD></TR>',[ord(ModeRndGenerator)]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Max Number of Clusters</TD><TD align="right">%d</TD></TR>',[FMaxNbClusters]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Distance normalization</TD><TD align="right">%d</TD></TR>',[FNormalization]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Min. size for split</TD><TD align="right">%d</TD></TR>',[FSizeBeforeSplit]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Min. size of leaves</TD><TD align="right">%d</TD></TR>',[FLeavesSizeAfterSplit]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Max. depth</TD><TD align="right">%d</TD></TR>',[FMaxDepth]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Goodness threshold</TD><TD align="right">%.2f</TD></TR>',[FGoodnessThreshold]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Pruning set size</TD><TD align="right">%d %s</TD></TR>',[FPruningSetSize,'%']);
 //s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Slope threshold</TD><TD align="right">%.3f</TD></TR>',[FThresholdSlope]);
 //new -- 05/01/2010
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Delta</TD><TD align="right">%.4f</TD></TR>',[FDelta]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Show all tree sequence</TD><TD align="right">%d</TD></TR>',[ord(FShowAllTreeSeq)]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpPrmClusTreeCART.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 FPruningSetSize:= prmINI.ReadInteger(prmSection,'pruning_set_size',FPruningSetSize);
 FDelta:= prmINI.ReadFloat(prmSection,'delta',FDelta);
 FShowAllTreeSeq:= prmINI.ReadBool(prmSection,'show_all_sequence',FShowAllTreeSeq);
end;

procedure TOpPrmClusTreeCART.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(FPruningSetSize,sizeof(FPruningSetSize));
 prmStream.ReadBuffer(FDelta,sizeof(FDelta));
 prmStream.ReadBuffer(FShowAllTreeSeq,sizeof(FShowAllTreeSeq));
 //new -- 05/01/2010 -- contrle pour le chargement du paramtre delta,  la place du "ThresoldSlope"
 if (FDelta < 0) then FDelta:= 0.05;
end;

procedure TOpPrmClusTreeCART.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'pruning_set_size',FPruningSetSize);
 prmINI.WriteFloat(prmSection,'delta',FDelta);
 prmINI.WriteBool(prmSection,'show_all_sequence',FShowAllTreeSeq);
end;

procedure TOpPrmClusTreeCART.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(FPruningSetSize,sizeof(FPruningSetSize));
 prmStream.WriteBuffer(FDelta,sizeof(FDelta));
 prmStream.WriteBuffer(FShowAllTreeSeq,sizeof(FShowAllTreeSeq));
end;

procedure TOpPrmClusTreeCART.SetDefaultParameters;
begin
 inherited SetDefaultParameters();
 //modifier les critres par dfaut de l'anctre
 //nombre de clusters maximum
 FMaxNbClusters:= 50;
 //paramtres ARBRE
 FSizeBeforeSplit:= 5;
 FLeavesSizeAfterSplit:= 2;
 //profondeur max.
 FMaxDepth:= 10;
 //gain min. en pourcentage de l'inertie totale --> construction hurdling
 FGoodnessThreshold:= 0.0;
 //critres de post-pruning
 FPruningSetSize:= 33;
 //FSERule:= 3.0;
 //new -- 11/09/2005 -- seuil empirique, les autres valeurs possibles sont -0.01 (permissif) et -0.03 (trs restrictif)
 //FThresholdSlope:= -0.015;
 //new -- 05/01/2010 -- paramtre delta
 FDelta:= 0.05;
 //affichage
 FShowAllTreeSeq:= FALSE;//new -- 08/03/2008 -- FALSE maitenant qu'on a le graphique des inerties
end;

initialization
 RegisterClass(TGenClusTreeCART);
end.
