(****************************************************************************)
(* UCompSDCatAssocPartialTheilU.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(****************************************************************************)

{
@abstract(Composant association sur variables catgorielles. U de Theil)
@author(Ricco)
@created(17/05/2007)

Partial U de Theil (Asymmetric Uncertainty Coefficient).
cf. essentiellement Ritschard et al. (Rapport de Recherche, Univ. Genve 1994)

}

unit UCompSDCatAssocPartialTheilU;

interface

USES
        Forms,Classes,
        IniFiles,
        UCompDefinition,
        UCompSDDefinition,
        UOperatorDefinition,
        UCalcStatDesCorrelation,
        UCompSDLinearCorrelation;

TYPE
        {Gnrateur de univariate stat cont desc}
        TMLGenCompSDCatPartialTheilU = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant}
        TMLCompSDCatPartialTheilU = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {l'oprateur associ}
        TOpSDCatPartialTheilU = class(TOpSD)
        protected
        procedure   InitializeListStat(); override;
        function    getClassParameter: TClassOperatorParameter; override;
        procedure   RebuildStatDes(); override;
        function    CheckAttributes(): boolean; override;
        public
        function    getHTMLResultsSummary(): string; override;
        end;

        {paramtrage d'oprateur tableaux croiss}
        TOpPrmSDCatPartialTheilU = class(TOpPrmSD)
        protected
        function    CreateDlgParameters(): TForm; override;
        end;

implementation

uses
        Sysutils, UDatasetDefinition,
  UDatasetImplementation, UDlgOpPrmSDLinearCorrelation, UStringsResources,
  UConstConfiguration, UCalcStatDesCrossTab,
  UDlgOpPrmSDCategoricalAssociation, UDlgOpPrmSDCatAssocTheil,
  UCalcStatDesCrossTab3D, UDlgOpPrmSDCatAssocPartialTheil;


{ TMLGenCompSDCatPartialTheilU }

procedure TMLGenCompSDCatPartialTheilU.GenCompInitializations;
begin
 FMLComp:= mlcNonParametricStatistics;
end;

function TMLGenCompSDCatPartialTheilU.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDCatPartialTheilU; 
end;

{ TMLCompSDCatPartialTheilU }

function TMLCompSDCatPartialTheilU.getClassOperator: TClassOperator;
begin
 result:= TOpSDCatPartialTheilU; 
end;

{ TOpSDCatPartialTheilU }

function TOpSDCatPartialTheilU.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asTarget].Count > 0) and (self.WorkData.LstAtts[asTarget].isAllCategory(caDiscrete));
 result:= result and (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caDiscrete));
 result:= result and (self.WorkData.LstAtts[asIllus].Count > 0) and (self.WorkData.LstAtts[asIllus].isAllCategory(caDiscrete));
end;

function TOpSDCatPartialTheilU.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDCatPartialTheilU;
end;

function TOpSDCatPartialTheilU.getHTMLResultsSummary: string;
begin
 result:= '<H3>Partial Uncertainty Coefficient</H3>' + FLstStat.getHTMLResults(-1);
end;

procedure TOpSDCatPartialTheilU.InitializeListStat;
begin
  FLstStat:= TLstCalcStatDesCrossTab3D.Create(nil,nil);
end;

procedure TOpSDCatPartialTheilU.RebuildStatDes;
var stat: TCalcCrossTab3D;
    A,B,E: TAttribute;
    i,j,k: integer;
begin
 for i:= 0 to pred(workdata.LstAtts[asTarget].Count) do
  begin
   A:= workdata.LstAtts[asTarget].Attribute[i];
   //le type est ok ?
   if A.isCategory(caDiscrete)
    then
     begin
      //les variables  dcrire
      for j:= 0 to pred(WorkData.LstAtts[asInput].Count) do
       begin
        B:= WorkData.LstAtts[asInput].Attribute[j];
        //petit test quand mme
        if B.isCategory(caDiscrete) and (A <> B)
         then
          begin
           for k:= 0 to pred(WorkData.LstAtts[asIllus].Count) do
            begin
             E:= WorkData.LstAtts[asIllus].Attribute[k];
             if E.isCategory(caDiscrete) and (E <> A) and (E <> B)
              then
               begin
                stat:= TCalcCrossTab3D.Create(A,B,E,nil);
                self.LstStat.AddStat(stat);
               end;
            end;
          end;
       end;
     end;
  end;
end;

{ TOpPrmSDCatPartialTheilU }

function TOpPrmSDCatPartialTheilU.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSDCatAssocPartialTheil.CreateFromOpPrm(self);
end;

initialization
 RegisterClass(TMLGenCompSDCatPartialTheilU);
end.
