(*****************************************************************************)
(* UCompSDCategoricalAssociation.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*****************************************************************************)

{
@abstract(Composant association sur variables catgorielles)
@author(Ricco)
@created(06/05/2007)

Calcul du coefficient d'association Tau de Goodman & Kruskal sur variables catgorielles (1954)
Calcul de la proba critique (p-value) avec la transformation du CHI-2 de  Light & Margolin (1971)

On devrait avoir aussi le z-score avec la variance asymtotique, cf. Olszack et Ritschard (1995)

}
unit UCompSDCategoricalAssociation;

interface

USES
        Forms,Classes,
        IniFiles,
        UCompDefinition,
        UCompSDDefinition,
        UOperatorDefinition,
        UCalcStatDesCorrelation,
        UCompSDLinearCorrelation;

TYPE
        {Gnrateur de univariate stat cont desc}
        TMLGenCompSDCatTau = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant}
        TMLCompSDCatTau = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        function    GetLogResultDescription(): string; override;
        end;

        {l'oprateur associ}
        TOpSDCatTau = class(TOpSD)
        protected
        procedure   InitializeListStat(); override;
        procedure   RebuildStatDes(); override;
        function    CheckAttributes(): boolean; override;
        function    getClassParameter: TClassOperatorParameter; override;
        public
        function    getHTMLResultsSummary(): string; override;
        end;

        {paramtrage d'oprateur tableaux croiss}
        TOpPrmSDCatTau = class(TOpPrmSD)
        protected
        function    CreateDlgParameters(): TForm; override;
        end;

implementation

uses
        Sysutils, UDatasetDefinition,
  UDatasetImplementation, UDlgOpPrmSDLinearCorrelation, UStringsResources,
  UConstConfiguration, UCalcStatDesCrossTab,
  UDlgOpPrmSDCategoricalAssociation;


{ TMLCompSDCatTau }

function TMLCompSDCatTau.getClassOperator: TClassOperator;
begin
 result:= TOpSDCatTau;
end;

function TMLCompSDCatTau.GetLogResultDescription: string;
begin
 result:= format('%d Goodman-Kruskal tau computed on %d examples',[(Operator as TOpSD).LstStat.Count,(Operator as TOpSD).Workdata.Examples.Size]);
end;

{ TMLGenCompSDCatTau }

procedure TMLGenCompSDCatTau.GenCompInitializations;
begin
 FMLComp:= mlcNonParametricStatistics;
end;

function TMLGenCompSDCatTau.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDCatTau;
end;

{ TOpSDCatTau }

function TOpSDCatTau.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asTarget].Count > 0) and (self.WorkData.LstAtts[asTarget].isAllCategory(caDiscrete));
 result:= result and (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caDiscrete));
end;

function TOpSDCatTau.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDCatTau; 
end;

function TOpSDCatTau.getHTMLResultsSummary: string;
var s: string;
    stat: TCalcSDCrossTab;
    j: integer;
begin
 s:= '<H3>Goodman & Kruskal''s Tau for nominal attributes</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Y</TH><TH width=80>X</TH><TH width=60>Tau</TH><TH width=60>Chi</TH><TH width=60>d.f.</TH><TH width=60>p-value</TH>' + '</TR>';
 for j:= 0 to pred(self.LstStat.Count) do
  begin
   stat:= self.LstStat.Stat(j) as TCalcSDCrossTab;
   //affichage de la ligne
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD>%s</TD><TD>%s</TD><TD align=right>%7.6f</TD><TD align=right>%.2f</TD><TD align=right>%d</TD><TD align=right>%.4f</TD>',
                  [stat.RowAtt.Name,stat.ColAtt.Name,stat.Tau,stat.chiTau,stat.DDL,stat.prTau]); 
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

procedure TOpSDCatTau.InitializeListStat;
begin
 FLstStat:= TLstCalcSDCrossTabTau.Create(nil,nil);
end;

procedure TOpSDCatTau.RebuildStatDes;
var stat: TCalcSDCrossTab;
    attI,attJ: TAttribute;
    i,j: integer;
begin
 //seul le schma TARGET vs. INPUT est licite !!!
 //pour chaque variable de description
 for i:= 0 to pred(workdata.LstAtts[asTarget].Count) do
  begin
   attI:= workdata.LstAtts[asTarget].Attribute[i];
   //le type est ok ?
   if attI.isCategory(caDiscrete)
    then
     begin
      //les variables  dcrire
      for j:= 0 to pred(WorkData.LstAtts[asInput].Count) do
       begin
        attJ:= WorkData.LstAtts[asInput].Attribute[j];
        //petit test quand mme
        if attJ.isCategory(caDiscrete) and (attI<>attJ)
         then
          begin
           stat:= TCalcSDCrossTab.Create(attI,attJ,nil);
           self.LstStat.AddStat(stat);
          end;
       end;
     end;
  end;
end;

{ TOpPrmSDCatTau }

function TOpPrmSDCatTau.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSDCategoricalAssociation.CreateFromOpPrm(self);
end;

initialization
 RegisterClass(TMLGenCompSDCatTau);
end.
