(*****************************************************************************)
(* UCompSDSemiPartialCorrelation.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(*****************************************************************************)

{
@abstract(Corrlation semi-partielle)
@author(Ricco)
@created(12/06/2008)

Calcul des corrlation semi-partielles.
Voir rf --> http://eric.univ-lyon2.fr/~ricco/cours/cours/Analyse_de_Correlation.pdf 

}
unit UCompSDSemiPartialCorrelation;

interface

USES
    Forms,Classes,
    IniFiles, Contnrs,
    UCompDefinition,
    UCompSDDefinition,
    UOperatorDefinition,
    UCalcStatDesCorrelation,
    UDatasetExamples,UDatasetImplementation,UDatasetDefinition,
    MATRIXMM;

TYPE
    {Gnrateur de univariate stat cont desc}
    TMLGenCompSDSemiPartialCorrelation = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    {le composant}
    TMLCompSDSemiPartialCorrelation = class(TMLCompSD)
    protected
    function    getClassOperator: TClassOperator; override;
    function    GetLogResultDescription(): string; override;
    end;

    {l'oprateur associ}
    TOpSDSemiPartialCorrelation = class(TOpSD)
    protected
    //liste des variables intermdiaires
    FLstTarget, FLstInput, FLstIllus: TObjectList;
    //liste des rsidus
    FResTarget, FResInput: TObjectList;
    //tableau pour des manips faciles
    Attributes: array[TEnumAttStatus] of TObjectList;
    Residuals: array[TEnumAttStatus] of TObjectList;
    //matrice [INV(X'X)]X'
    FX,FXPXM1XP: PManipMatMM;
    //matrice des indicateurs stats (5 colonnes, autant de lignes que de rsultats)
    //r, t, p-value, b.basse, b.haute
    FTabStats: array of array of double;
    procedure   prepareDataset(examples: TExamples);
    procedure   computeResiduals(examples: TExamples);
    procedure   computeTargetResiduals(examples: TExamples); virtual;//ce qui fait le semi-partiel ici -- simple recopie,  surcharger
    procedure   computeInputResiduals(examples: TExamples);
    function    computeResidual(examples: TExamples; dependent: TAttribute): TAttribute;
    procedure   computeProjectionMatrix(examples: TExamples);
    procedure   computePostStat();
    procedure   setLstOwner(own: boolean);
    procedure   clearAttributes();
    procedure   recupAttributes();
    procedure   InitializeListStat(); override;
    function    getClassParameter: TClassOperatorParameter; override;
    function    CoreExecute(): boolean; override;
    function    CheckAttributes(): boolean; override;
    function    getCorrelationTitleHtml(): string; virtual;// surcharger
    public
    function    getHTMLResultsSummary(): string; override;
    destructor  destroy(); override;
    end;

    {paramtrage d'oprateur tableaux croiss}
    TOpPrmSDSemiPartialCorrelation = class(TOpPrmSD)
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    //type de corrlation : 0 - Pearson (dfaut) ; 1 - Spearman
    TypeCorr: integer;
    //tri : 0 - non ; 1 - selon le r ; 2 - selon le r^2
    Sorting: integer;
    //sig.level pour les intervalle de confiance : dfaut 0.05
    SigLevel: double;
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    end;


implementation

uses
    Sysutils, Math, FMath,
    UCalcStatDesRankForNonParametricStat, UConstConfiguration,
    UDlgOpPrmSDSemiPartialCorrelation;

const
    STR_CORRELATION : array[0..1] of string =('Pearson','Spearman');    

{ TMLGenCompSDSemiPartialCorrelation }

procedure TMLGenCompSDSemiPartialCorrelation.GenCompInitializations;
begin
 FMLComp:= mlcDescriptiveStat;
end;

function TMLGenCompSDSemiPartialCorrelation.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDSemiPartialCorrelation;
end;

{ TMLCompSDSemiPartialCorrelation }

function TMLCompSDSemiPartialCorrelation.getClassOperator: TClassOperator;
begin
 result:= TOpSDSemiPartialCorrelation;
end;

function TMLCompSDSemiPartialCorrelation.GetLogResultDescription: string;
begin
 result:= format('[SEMI-PARTIAL Correlation] %d correlations computed',[(Operator as TOpSDSemiPartialCorrelation).LstStat.Count]);
end;

{ TOpSDSemiPartialCorrelation }

function TOpSDSemiPartialCorrelation.CheckAttributes: boolean;
var ok: boolean;
begin
 //au moins 1 target continu
 ok:= (self.WorkData.LstAtts[asTarget].Count>0) and (self.WorkData.LstAtts[asTarget].isAllCategory(caContinue));
 //au moins 1 input continu
 ok:= ok and (self.WorkData.LstAtts[asInput].Count>0) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue));
 //et... grande nouveaut - au moins 1 illustratif continu
 ok:= ok and (self.WorkData.LstAtts[asIllus].Count>0) and (self.WorkData.LstAtts[asIllus].isAllCategory(caContinue));
 //and then...
 result:= ok;
end;

procedure TOpSDSemiPartialCorrelation.clearAttributes;
var st: TEnumAttStatus;
begin
 for st:= asTarget to asIllus do Attributes[st].Clear();
end;

procedure TOpSDSemiPartialCorrelation.computeInputResiduals(
  examples: TExamples);
var j: integer;
begin
 FResInput.Clear();
 for j:= 0 to pred(FLstInput.Count) do
  FResInput.Add(self.computeResidual(examples,FLstInput.Items[j] as TAttribute));
end;

procedure TOpSDSemiPartialCorrelation.computePostStat;
var k: integer;
    z,alpha,ddl,t,r2,pvalue,bb,bh,u,stderr: double;
    stat: TCalcSDCorrelation;

begin
 alpha:= (self.PrmOp as TOpPrmSDSemiPartialCorrelation).SigLevel;
 u:= InvNorm(1.0-alpha/2.0);
 setLength(FTabStats,LstStat.Count,5);
 ddl:= self.WorkData.Examples.Size - 2 - self.FLstIllus.Count;//trs important !!!
 stderr:= sqrt(1.0/ddl);
 for k:= 0 to pred(LstStat.Count) do
  begin
   stat:= LstStat.Stat(k) as TCalcSDCorrelation;
   r2:= Math.Power(stat.R,2.0);
   t:= 0.0;
   if (r2 < 1.0) then t:= stat.r/sqrt((1.0-r2)/ddl);
   pvalue:= PStudent(round(ddl),abs(t));
   z:= 0.0;
   if (r2 < 1.0) then z:= Math.ArcTanh(stat.r);
   bb:= z - u * stderr; bb:= Math.Tanh(bb);
   bh:= z + u * stderr; bh:= Math.Tanh(bh);
   FTabStats[k,0]:= r2;
   FTabStats[k,1]:= t;
   FTabStats[k,2]:= pvalue;
   FTabStats[k,3]:= bb;
   FTabStats[k,4]:= bh;
  end;
end;

procedure TOpSDSemiPartialCorrelation.computeProjectionMatrix(
  examples: TExamples);
var XP,XPX,XPXM1: PManipMatMM;
    i,j,example,NbJ: integer;
begin
 NbJ:= FLstIllus.Count;
 //construire la matrice de donnes
 FX:= PManipMatMM.Create(examples.Size,NbJ+1);
 for i:= 1 to examples.Size do
  begin
   example:= examples.Number[i];
   for j:= 1 to NbJ do
    FX[i,j]:= (FLstIllus.Items[pred(j)] as TAttribute).cValue[example];
   FX[i,NbJ+1]:= 1.0;
  end;
 //transposition, multiplier, inverser et re-multiplier --> XPXM1XP
 TranspositionMM(FX,XP);
 MultiplicationMM(XP,FX,XPX);
 if (InversionMM(XPX,XPXM1)<>ErrMM_NoError)
  then Raise Exception.Create('SINGULAR MATRIX');
 MultiplicationMM(XPXM1,XP,FXPXM1XP);
 //dtruire les matrices intermdiaires qui ne servent plus
 if assigned(XP) then XP.Free();
 if assigned(XPX) then XPX.Free();
 if assigned(XPXM1) then XPXM1.Free();
end;

function TOpSDSemiPartialCorrelation.computeResidual(
  examples: TExamples; dependent: TAttribute): TAttribute;
var Y,YC: PManipMatMM;
    coef: PManipMatMM;
    i: integer;
    output: TAttribute;
begin
 Y:= PManipMatMM.Create(examples.size,1);
 for i:= 1 to examples.Size do Y[i,1]:= dependent.cValue[examples.Number[i]];
 //calculer les coefs.
 MultiplicationMM(FXPXM1XP,Y,coef);
 //projection
 MultiplicationMM(FX,coef,YC);
 output:= TAttContinue.Create(dependent.Name,dependent.Size);
 for i:= 1 to examples.Size do output.cValue[examples.Number[i]]:= Y[i,1]-YC[i,1];
 if assigned(Y) then Y.Free();
 if assigned(YC) then YC.Free();
 if assigned(coef) then coef.Free();
 result:= output;
end;

procedure TOpSDSemiPartialCorrelation.computeResiduals(
  examples: TExamples);
begin
 //crer les rangs le cas chant, sinon recopie simple
 self.prepareDataset(examples);
 //prparer les matrices de calcul (X'X)^(-1)X'
 if assigned(FX) then FX.Free();
 if assigned(FXPXM1XP) then FXPXM1XP.Free();
 self.computeProjectionMatrix(examples);
 //crr les listes de rsidus
 self.computeInputResiduals(examples);
 self.computeTargetResiduals(examples);
 //vider ces gros trucs l
 if assigned(FX) then FreeAndNil(FX);
 if assigned(FXPXM1XP) then FreeAndNil(FXPXM1XP);
end;

procedure TOpSDSemiPartialCorrelation.computeTargetResiduals(
  examples: TExamples);
var j: integer;
begin
 FResTarget.Clear();
 FResTarget.OwnsObjects:= false;
 for j:= 0 to pred(FLstTarget.Count) do
  FResTarget.Add(FLstTarget.Items[j]);
end;

function TOpSDSemiPartialCorrelation.CoreExecute: boolean;
var i,j: integer;
    target,input: TAttribute;
    stat: TCalcSDCorrelation;
begin
 TRY
 //prparer les variables intermdiaires ou recopier (selon Pearson - Spearman)
 computeResiduals(self.WorkData.Examples);
 //allons-y alors
 for i:= 0 to pred(FResTarget.Count) do
  begin
   target:= FResTarget.Items[i] as TAttribute;
   for j:= 0 to pred(FResInput.Count) do
    begin
     input:= FResInput.Items[j] as TAttribute;
     //si c'est la mme variable, a ne sert  rien
     if (target.Name <> input.Name)
      then
       begin
        stat:= TCalcSDCorrelation.Create(target,input,self.WorkData.Examples);
        LstStat.AddStat(stat);
       end;
    end;
  end;
 //faut-il trier ?
 if ((self.PrmOp as TOpPrmSDSemiPartialCorrelation).Sorting > 0)
  then
   begin
    LstStat.CompareMode:= (self.PrmOp as TOpPrmSDSemiPartialCorrelation).Sorting + 1;
    LstStat.SortStats();
   end;
 //calculer les autres indicateurs et intervalle de confiance
 self.computePostStat();
 //
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

destructor TOpSDSemiPartialCorrelation.destroy;
var st: TEnumAttStatus;
begin
 SetLength(FTabStats,0,0);
 //vider les objets de la classe
 for st:= asTarget to asIllus do if assigned(Attributes[st]) then Attributes[st].Free();
 for st:= asTarget to asInput do if assigned(Residuals[st]) then Residuals[st].Free();
 if assigned(FX) then FX.Free();
 if assigned(FXPXM1XP) then FXPXM1XP.Free();
 //et les autres...
 inherited;
end;

function TOpSDSemiPartialCorrelation.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDSemiPartialCorrelation;
end;

function TOpSDSemiPartialCorrelation.getCorrelationTitleHtml: string;
begin
 result:= '<H4>Semi-partial correlation (part correlation)</H4>';
end;

function TOpSDSemiPartialCorrelation.getHTMLResultsSummary: string;
var stat: TCalcSDCorrelation;
    k: integer;
    s: string;
begin
 s:= '<H4>Control variables</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_BLUE + '<TH colspan=2>Variable<TH></TR>';
 for k:= 0 to pred(self.FLstIllus.Count) do
  s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD align=center width=20>%d</TD><TD align=right width=100>%s</TD></TR>',[k+1,(self.FLstIllus.Items[k] as TAttribute).Name]);
 s:= s + '</table>';
 s:= s + self.getCorrelationTitleHtml();
 s:= s + HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=7>-</TH><TH colspan=2>Conf.Interval</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + '<TH width=20>N</TH><TH width=60>Att.Y</TH><TH width=60>Att.X</TH><TH width=60>r</TH><TH width=60>r</TH><TH width=60>t</TH><TH width=60>p-value</TH><TH width=60>Lower.Limit</TH><TH width=60>Upper.Limit</TH></TR>';
 for k:= 0 to pred(self.LstStat.Count) do
  begin
   stat:= self.LstStat.Stat(k) as TCalcSDCorrelation;
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD align=center>%d</TD><TD>%s</TD><TD>%s</TD><TD align=right>%.5f</TD><TD align=right>%.5f</TD><TD align=right>%.5f</TD>'+
                  '<TD align=right>%.5f</TD><TD align=right>%.5f</TD><TD align=right>%.5f</TD>',
                  [k+1,stat.YAtt.Name,stat.XAtt.Name,stat.R,FTabStats[k,0],FTabStats[k,1],
                  FTabStats[k,2],FTabStats[k,3],FTabStats[k,4]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

procedure TOpSDSemiPartialCorrelation.InitializeListStat;
begin
 //l'astuce est l - on maintient une liste standard de corrlation
 //mais les donnes seront prpars de manire adquate au pralable...
 FLstStat:= TLstCalcSDCorrelation.Create(nil,nil);
 Attributes[asAll]:= nil;
 FLstTarget:= TObjectList.Create(true); Attributes[asTarget]:= FLstTarget;
 FLstInput:= TObjectList.Create(true); Attributes[asInput]:= FLstInput;
 FLstIllus:= TObjectList.Create(true); Attributes[asIllus]:= FLstIllus;
 Residuals[asAll]:= nil; Residuals[asIllus]:= nil;
 FResTarget:= TObjectList.Create(true); Residuals[asTarget]:= FResTarget;
 FResInput:= TObjectList.Create(true); Residuals[asInput]:= FResInput;
end;

procedure TOpSDSemiPartialCorrelation.prepareDataset(examples: TExamples);
var st: TEnumAttStatus;
    rank: TRankComputed;
    ex: TExamples;
    source,output: TAttribute;
    j,i,example: integer;
begin
 self.clearAttributes();//vider (voire librer) les prcdentes
 case (self.PrmOp as TOpPrmSDSemiPartialCorrelation).TypeCorr of
 //Spearman
 1 : begin
      self.setLstOwner(true);
      ex:= TExamples.Create(examples.Size);
      //transformer les donnes en rang
      for st:= asTarget to asIllus do
       begin
        for j:= 0 to pred(self.WorkData.LstAtts[st].Count) do
         begin
          source:= self.WorkData.LstAtts[st].Attribute[j];
          ex.Copy(examples);
          ex.QuickSortBy(source);
          rank:= TRankComputed.create(source);
          rank.computeRank(ex);
          //crer la nouvelle variable avec les valeurs de rang (rang moyens)
          output:= TAttContinue.Create(source.Name,source.Size);
          for i:= 1 to examples.Size do
           begin
            example:= examples.Number[i];
            output.cValue[example]:= rank.getRank(example);
           end;
          rank.Free();
          Attributes[st].Add(output);
         end;
       end;
      ex.Free();
     end;
 else
 //Pearson
   begin
    self.setLstOwner(false);
    self.recupAttributes();
   end;
  end;
end;

procedure TOpSDSemiPartialCorrelation.recupAttributes;
var j: integer;
    st: TEnumAttStatus;
begin
 for st:= asTarget to asIllus do
  for j:= 0 to pred(self.WorkData.LstAtts[st].Count) do
   Attributes[st].Add(self.WorkData.LstAtts[st].Attribute[j]);
end;

procedure TOpSDSemiPartialCorrelation.setLstOwner(own: boolean);
begin
 FLstTarget.OwnsObjects:= own;
 FLstInput.OwnsObjects:= own;
 FLstIllus.OwnsObjects:= own;
end;

{ TOpPrmSDSemiPartialCorrelation }

function TOpPrmSDSemiPartialCorrelation.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSDSemiPartialCorrelation.CreateFromOpPrm(self); 
end;

function TOpPrmSDSemiPartialCorrelation.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Parameters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Correlation</TD><TD>%s</TD></TR>',[STR_CORRELATION[TypeCorr]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Sort results</TD><TD align=right>%d</TD></TR>',[sorting]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Sig.Level</TD><TD align=right>%.4f</TD></TR>',[SigLevel]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmSDSemiPartialCorrelation.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 TypeCorr:= prmINI.ReadInteger(prmSection,'type_corr',TypeCorr);
 sorting:= prmINI.ReadInteger(prmSection,'sorting',sorting);
 SigLevel:= prmINI.ReadFloat(prmSection,'sig_level',SigLevel);
end;

procedure TOpPrmSDSemiPartialCorrelation.LoadFromStream(
  prmStream: TStream);
begin
 prmStream.ReadBuffer(TypeCorr,sizeof(TypeCorr));
 prmStream.ReadBuffer(Sorting,sizeof(Sorting));
 prmStream.ReadBuffer(SigLevel,sizeof(SigLevel));
end;

procedure TOpPrmSDSemiPartialCorrelation.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'type_corr',TypeCorr);
 prmINI.WriteInteger(prmSection,'sorting',sorting);
 prmINI.WriteFloat(prmSection,'sig_level',SigLevel);
end;

procedure TOpPrmSDSemiPartialCorrelation.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(TypeCorr,sizeof(TypeCorr));
 prmStream.WriteBuffer(Sorting,sizeof(Sorting));
 prmStream.WriteBuffer(SigLevel,sizeof(SigLevel));
end;

procedure TOpPrmSDSemiPartialCorrelation.SetDefaultParameters;
begin
 TypeCorr:= 0; //Pearson
 Sorting:= 0; //pas de tri
 SigLevel:= 0.05;//(1 - SigLevel) = Confidence Level
end;

initialization
 RegisterClass(TMLGenCompSDSemiPartialCorrelation);
end.
