(***************************************************************************)
(* UCalcSDCondDescKolmogorov2S.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(***************************************************************************)

{
@abstract(Test de Kolmogorov-Smirnov pour la comparaison de 2 populations)
@author(Ricco)
@created(17/08/2008)

Classe de calcul.
Voir le support de cours "Comparaison de populations - Tests non paramtriques"

}
unit UCalcSDCondDescKolmogorov2S;

interface

USES
    UDatasetExamples,
    UCalcStatDesConditionnalDesc;

TYPE
    TCalcSDCondDescKS2SampleTest = class(TCalcSDCondDesc)
    protected
    //description des stastiques de test
    function    getHTMLStatistic(): string;
    //calcul des statistiques de test
    procedure   calcStatistics();
    public
    //rsultats intermdiaires
    DPlus,DMoins: double;
    XPlus,XMoins: double;
    EDFPlus,EDFMoins: double;
    //cramer von mises
    CVM: double;
    //statistique de KS
    KSStat,KSAsymp,pKS: double;
    //Kuiper
    Kuiper,KuiperAsymp,pKuiper: double;
    //affichage des rsultats
    function    getHTMLResult(prmOption: integer = -1): string; override;
    //calcul proprement dit
    procedure   RefreshStat(prmExamples: TExamples); override;
    end;

    TLstCalcSDCondDescKS2SampleTest = class(TLstCalcStatDesCond)
    public
    function    getHeaderHTML(): string; override;
    end;

implementation

USES
    Math,
    Sysutils,
    UDatasetDefinition,
    UConstConfiguration, ULogFile;

CONST
    STR_REF_EDF : array[0..1] of string = ('+','-');
    PREC_P_VALUE_KS : integer = 3;

{ TLstCalcSDCondDescKS2SampleTest }

function TLstCalcSDCondDescKS2SampleTest.getHeaderHTML: string;
begin
 result:= HTML_TABLE_COLOR_HEADER_GRAY+
          '<TH>Attribute_Y</TH><TH>Attribute_X</TH><TH>Description</TH><TH>Statistical test</TH></TR>';
end;

{ TCalcSDCondDescKS2SampleTest }

procedure TCalcSDCondDescKS2SampleTest.calcStatistics;
var temp: double;
    j: integer;
    lambda,q: double;
begin
 temp:= SQRT((1.0 * self.StatCond[0].NbExamples * self.StatCond[1].NbExamples)/(1.0 * self.StatCond[0].NbExamples + 1.0 * self.StatCond[1].NbExamples));
 //Kolmogorov
 KSStat:= Math.Max(DPlus,DMoins);
 KSASymp:= temp * KSStat;
 pKS:= 0.0;
 for j:= 1 to PREC_P_VALUE_KS do
  pKS:= pKS + Math.Power(-1.0,1.0+j) * EXP(-2.0 * Math.Power(1.0*j,2.0) * Math.Power(KSAsymp,2.0));
 pKS:= 2.0 * pKS;
 //Kuiper
 Kuiper:= DPlus + DMoins;
 KuiperAsymp:= temp * Kuiper;
 lambda:= KuiperAsymp;
 q:= 0.0;
 for j:= 1 to PREC_P_VALUE_KS do
  q:= q + (4.0*sqr(j)*sqr(lambda)-1.0)*exp(-2.0*sqr(j)*sqr(lambda));
 q:= 2.0 *q;
 pKuiper:= q;
end;

function TCalcSDCondDescKS2SampleTest.getHTMLResult(
  prmOption: integer): string;
var s: string;
    k: integer;
begin
 //description sur une variable
 s:= format('<TD>%s</TD><TD>%s</TD>',[self.Attribute.Name,self.AttDescription.Name]);
 //les stat descriptives conditionelles
 s:= s+'<TD valign="top">'+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_BLUE+
       '<TH width=70>Value</TH><TH width=70>Examples</TH><TH width=70>Average</TH><TH width=70>Std-dev</TH></TR>';
 //pour chaque modalit de la variable de description
 for k:= 0 to pred(self.AttDescription.nbValues) do
  begin
   s:= s+HTML_TABLE_COLOR_DATA_BLUE+
         format('<TH>%s (%s)</TH><TD align=right>%d</td><td align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</td>'+'<td align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</td>',
                [self.AttDescription.LstValues.GetDescription(succ(k)), STR_REF_EDF[k],StatCond[k].NbExamples,StatCond[k].Average,StatCond[k].StdDev])+'</tr>';
  end;
 //stat globales (sur tous les individus)
 s:= s+HTML_TABLE_COLOR_DATA_GREEN+
       format('<TH>All</TH><TD align=right>%d</td><td align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</td>'+'<td align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY+'</td>',
              [self.NbExamples,self.StatGlobal.Average,self.StatGlobal.StdDev])+'</tr>';
 s:= s+'</table></td>';
 //s'occuper de la statistique de test maintenant

 s:= s +'<TD valign=top>';
 s:= s + self.getHTMLStatistic();
 s:= s + '</TD>';

 //renvoyer le tout
 result:= s;
end;

function TCalcSDCondDescKS2SampleTest.getHTMLStatistic: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 //dtail des carts
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>-</TH><TH>Value</TH><TH>Y coord.</TH><TH>EDF coord.</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD width=70>D+</TD><TD align=right width=60>%.6f</TD><TD align=right width=60>%.6g</TD><TD align=right width=60>%.6f</TD></TR>',[DPlus,XPlus,EDFPlus]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD width=70>D-</TD><TD align=right width=60>%.6f</TD><TD align=right width=60>%.6g</TD><TD align=right width=60>%.6f</TD></TR>',[DMoins,XMoins,EDFMoins]);
 //statistiques calcules
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH colspan=4>Statistics</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>&nbsp;</TH><TH>Value</TH><TH>Asymp.Value</TH><TH>p-value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD width=70>KS Stat. (D)</TD><TD align=right width=60>%.6f</TD><TD align=right width=60>%.6f</TD><TD align=right width=60>%.6f</TD></TR>',[KSStat,KSAsymp,pKS]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD width=70>Kuiper''s V</TD><TD align=right width=60>%.6f</TD><TD align=right width=60>%.6f</TD><TD align=right width=60>%.6f</TD></TR>',[Kuiper,KuiperAsymp,pKuiper]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD width=70 colspan=2>Cramer - von Mises</TD><TD align=right width=60>%.6f</TD><TD align=center width=60>-</TD></TR>',[CVM]);
 //
 s:= s + '</table>';
 result:= s;
end;

//***** structure temporaire
TYPE
    TDataCoord = record
    x: TTypeContinue;
    t: double;
    edf: array [1..2] of double;
    end;
//**************************

procedure TCalcSDCondDescKS2SampleTest.RefreshStat(prmExamples: TExamples);
var table: array of TDataCoord;
    sorted: TExamples;
    example,numValue,i,k: integer;
    value: TTypeContinue;
    e: double;
begin
 inherited refreshStat(prmExamples);
 //construire la liste trie
 sorted:= TExamples.Create(prmExamples.Size);
 sorted.Copy(prmExamples);
 sorted.QuickSortBy(self.Attribute);
 //laborer le tableau des EDF
 setlength(table,sorted.Size);
 //initaliser avec la premire valeur
 numValue:= 0;
 table[0].x:= self.Attribute.cValue[sorted.Number[1]];
 table[0].edf[self.AttDescription.dValue[sorted.Number[1]]]:= 1.0;
 table[0].t:= 1.0;
 //remplir les valeurs -- attention, cela suppose que le tableau edf est initialis  zro par dfaut ( vrifier si bug)
 for i:= 2 to sorted.Size do
  begin
   example:= sorted.Number[i];
   value:= self.Attribute.cValue[example];
   if (value > table[numValue].x)
    then
     begin
      inc(numValue);
      table[numValue].x:= value;
      table[numValue].t:= 0.0;
      table[numValue].edf[1]:= 0.0; table[numValue].edf[2]:= 0.0; 
     end;
   k:= self.AttDescription.dValue[example];
   table[numValue].edf[k]:= table[numValue].edf[k] + 1.0;
   table[numValue].t:= table[numValue].t + 1.0;  
  end;
 inc(numValue);//pour le dernier coup
 //retailler le tableau
 setLength(table,numValue);
 //repasser en frquences relatives cumules
 table[0].edf[1]:= table[0].edf[1] / (1.0 * self.StatCond[0].NbExamples);
 table[0].edf[2]:= table[0].edf[2] / (1.0 * self.StatCond[1].NbExamples);
 for i:= 1 to high(table) do
  begin
   table[i].edf[1]:= table[i-1].edf[1] + table[i].edf[1] / (1.0 * self.StatCond[0].NbExamples);
   table[i].edf[2]:= table[i-1].edf[2] + table[i].edf[2] / (1.0 * self.StatCond[1].NbExamples);
  end;
 //calculer les ecarts
 XPlus:= table[0].x;
 XMoins:= table[0].x;
 DPlus:= -1.0 * Math.MaxSingle;
 DMoins:= -1.0 * Math.MaxSingle;
 EDFPlus:= 0.0;
 EDFMoins:= 0.0;
 CVM:= 0.0;
 //chercher les carts
 for i:= low(table) to high(table) do
  begin
   e:= table[i].edf[1] - table[i].edf[2];
   //TraceLog.WriteToLogFile(format('[KS2Sample] edf1 = %.5f, edf2 = %.5f, e = %.5f',[table[i].edf[1],table[i].edf[2],e]));
   if (e >= 0) and (e > DPlus)
    then
     begin
      DPlus:= e;
      XPlus:= table[i].x;
      EDFPlus:= table[i].edf[1];
     end;
   //et l'inverse
   e:= -1.0 * e;
   if (e >= 0) and (e > DMoins)
    then
     begin
      DMoins:= e;
      XMoins:= table[i].x;
      EDFMoins:= table[i].edf[2];
     end;
   //cramer - von mises
   CVM:= CVM + table[i].t * SQR(e);
  end;
 //puis
 CVM:= (1.0 * self.StatCond[0].NbExamples * self.StatCond[1].NbExamples)/SQR(self.StatCond[0].NbExamples + self.StatCond[1].NbExamples) * CVM;
 //vider tout a
 sorted.Free();
 setLength(table,0);
 //statistique de test
 self.calcStatistics();
end;

end.
