(***********************************************************************************)
(* UCompSDAnsariBradleyScaleDifference.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(***********************************************************************************)

{
@abstract(Comparaison de populations, mais avec une approche non paramtrique, celle de Ansari-Bradley)
@author(Ricco)
@created(17/08/2008)

}

unit UCompSDAnsariBradleyScaleDifference;

interface

USES
        Forms, Classes,
        UCompDefinition,
        UCompManageDataset,
        UCompSDDefinition,
        UCompSDCondDescDefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcStatDesConditionnalDesc,
        UCalcStatDesRankForNonParametricStat; 

TYPE

        {gnrateur de composant one-way anova}
        TMLGenCompSDABScale = class(TMLGenCompNonParOneWay)
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant one anova}
        TMLCompSDABScale = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur de calcul }
        TOpSDABScale = class(TOpSDCondDescDefinition)
        protected
        function    getClassStat(): TClassCalcSDCondDesc; override;
        function    getClassLstStat(): TClassLstCalcStatDesCond; override;
        function    getClassParameter: TClassOperatorParameter; override;
        end;

        {paramtrage de l'oprateur de calcul}
        TOpPrmSDABScale = class(TOpPrmSDCondDescDefinition)
        end;

        {classe de calcul statistique}
        TCalcSDNonParamABScale = class(TCalcSDNonParamCondDescScores)
        protected
        function    getClassRankScoreComputation(): TClassRankComputed; override;
        end;

        {classe de calcul des scores}
        TScoresABScale = class(TRankComputed)
        protected
        procedure   computeScoresFromRank(lstSorted: TExamples); override;
        end;

implementation

USES
    FMath;

{ TMLGenCompSDABScale }

function TMLGenCompSDABScale.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDABScale;
end;

{ TMLCompSDABScale }

function TMLCompSDABScale.getClassOperator: TClassOperator;
begin
 result:= TOpSDABScale;
end;

{ TOpSDABScale }

function TOpSDABScale.getClassLstStat: TClassLstCalcStatDesCond;
begin
 result:= TLstCalcStatDesCondScores;
end;

function TOpSDABScale.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDABScale;
end;

function TOpSDABScale.getClassStat: TClassCalcSDCondDesc;
begin
 result:= TCalcSDNonParamABScale;
end;

{ TCalcSDNonParamABScale }

function TCalcSDNonParamABScale.getClassRankScoreComputation: TClassRankComputed;
begin
 result:= TScoresABScale;
end;

{ TScoresABScale }

procedure TScoresABScale.computeScoresFromRank(lstSorted: TExamples);
var i,example: integer;
    r,moyenne,nbExamples: double;
begin
 nbExamples:= 1.0 * lstSorted.Size;
 moyenne:= (1.0 + nbExamples)/2.0;
 for i:= 1 to lstSorted.Size do
  begin
   example:= lstSorted.Number[i];
   r:= FRankValue[example];
   FRankValue[example]:= moyenne - abs(r - moyenne);
  end;
end;

initialization
 RegisterClass(TMLGenCompSDABScale);
end.

