(***********************************************************)
(* UCompSDFYTH.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(***********************************************************)

{
@abstract(Comparaison de populations, mais avec une approche non paramtrique, celle de Fisher-Yates-Terry-Hoeffding)
@author(Ricco)
@created(17/08/2008)

}

unit UCompSDFYTH;

interface

USES
        Forms, Classes,
        UCompDefinition,
        UCompManageDataset,
        UCompSDDefinition,
        UCompSDCondDescDefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcStatDesConditionnalDesc,
        UCalcStatDesRankForNonParametricStat; 

TYPE

        {gnrateur de composant one-way anova}
        TMLGenCompSDFYTH = class(TMLGenCompNonParOneWay)
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant one anova}
        TMLCompSDFYTH = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur de calcul }
        TOpSDFYTH = class(TOpSDCondDescDefinition)
        protected
        function    getClassStat(): TClassCalcSDCondDesc; override;
        function    getClassLstStat(): TClassLstCalcStatDesCond; override;
        function    getClassParameter: TClassOperatorParameter; override;
        end;

        {paramtrage de l'oprateur de calcul}
        TOpPrmSDFYTH = class(TOpPrmSDCondDescDefinition)
        end;

        {classe de calcul statistique}
        TCalcSDNonParamFYTH = class(TCalcSDNonParamCondDescScores)
        protected
        function    getClassRankScoreComputation(): TClassRankComputed; override;
        end;

        {classe de calcul des scores}
        TScoresFYTH = class(TRankComputed)
        protected
        procedure   computeScoresFromRank(lstSorted: TExamples); override;
        end;

implementation

USES
    FMath;

{ TMLGenCompSDFYTH }

function TMLGenCompSDFYTH.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDFYTH;
end;

{ TMLCompSDFYTH }

function TMLCompSDFYTH.getClassOperator: TClassOperator;
begin
 result:= TOpSDFYTH;
end;

{ TOpSDFYTH }

function TOpSDFYTH.getClassLstStat: TClassLstCalcStatDesCond;
begin
 result:= TLstCalcStatDesCondScores;
end;

function TOpSDFYTH.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDFYTH;
end;

function TOpSDFYTH.getClassStat: TClassCalcSDCondDesc;
begin
 result:= TCalcSDNonParamFYTH;
end;

{ TCalcSDNonParamFYTH }

function TCalcSDNonParamFYTH.getClassRankScoreComputation: TClassRankComputed;
begin
 result:= TScoresFYTH;
end;

{ TScoresFYTH }

procedure TScoresFYTH.computeScoresFromRank(lstSorted: TExamples);
var i,example: integer;
    r,denom,nbExamples: double;
begin
 nbExamples:= 1.0 * lstSorted.Size;
 denom:= 0.25 + nbExamples;
 for i:= 1 to lstSorted.Size do
  begin
   example:= lstSorted.Number[i];
   r:= FRankValue[example];
   FRankValue[example]:= invNorm((r - 0.375) / denom);
  end;
end;

initialization
 RegisterClass(TMLGenCompSDFYTH);
end.
