(****************************************************************************)
(* UCompSDKlotzScaleDifferences.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(****************************************************************************)

{
@abstract(Comparaison de populations, mais avec une approche non paramtrique, celle de Klotz)
@author(Ricco)
@created(17/08/2008)

}

unit UCompSDKlotzScaleDifferences;

interface

USES
        Forms, Classes,
        UCompDefinition,
        UCompManageDataset,
        UCompSDDefinition,
        UCompSDCondDescDefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcStatDesConditionnalDesc,
        UCalcStatDesRankForNonParametricStat; 

TYPE

        {gnrateur de composant one-way anova}
        TMLGenCompSDKlotzScale = class(TMLGenCompNonParOneWay)
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant one anova}
        TMLCompSDKlotzScale = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur de calcul }
        TOpSDKlotzScale = class(TOpSDCondDescDefinition)
        protected
        function    getClassStat(): TClassCalcSDCondDesc; override;
        function    getClassLstStat(): TClassLstCalcStatDesCond; override;
        function    getClassParameter: TClassOperatorParameter; override;
        end;

        {paramtrage de l'oprateur de calcul}
        TOpPrmSDKlotzScale = class(TOpPrmSDCondDescDefinition)
        end;

        {classe de calcul statistique}
        TCalcSDNonParamKlotzScale = class(TCalcSDNonParamCondDescScores)
        protected
        function    getClassRankScoreComputation(): TClassRankComputed; override;
        end;

        {classe de calcul des scores}
        TScoresKlotzScale = class(TRankComputed)
        protected
        procedure   computeScoresFromRank(lstSorted: TExamples); override;
        end;

implementation

USES
    FMath;

{ TMLGenCompSDKlotzScale }

function TMLGenCompSDKlotzScale.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDKlotzScale;
end;

{ TMLCompSDKlotzScale }

function TMLCompSDKlotzScale.getClassOperator: TClassOperator;
begin
 result:= TOpSDKlotzScale;
end;

{ TOpSDKlotzScale }

function TOpSDKlotzScale.getClassLstStat: TClassLstCalcStatDesCond;
begin
 result:= TLstCalcStatDesCondScores;
end;

function TOpSDKlotzScale.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDKlotzScale;
end;

function TOpSDKlotzScale.getClassStat: TClassCalcSDCondDesc;
begin
 result:= TCalcSDNonParamKlotzScale;
end;

{ TCalcSDNonParamKlotzScale }

function TCalcSDNonParamKlotzScale.getClassRankScoreComputation: TClassRankComputed;
begin
 result:= TScoresKlotzScale;
end;

{ TScoresKlotzScale }

procedure TScoresKlotzScale.computeScoresFromRank(lstSorted: TExamples);
var i,example: integer;
    r,denom,nbExamples: double;
begin
 nbExamples:= 1.0 * lstSorted.Size;
 denom:= 1.0 + nbExamples;
 for i:= 1 to lstSorted.Size do
  begin
   example:= lstSorted.Number[i];
   r:= FRankValue[example];
   FRankValue[example]:= SQR(invNorm(r / denom));
  end;
end;


initialization
 RegisterClass(TMLGenCompSDKlotzScale);
end.

