(***************************************************************************)
(* UCompSDMoodScaleDifferences.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(***************************************************************************)

{
@abstract(Comparaison de populations, mais avec une approche non paramtrique, celle de Mood)
@author(Ricco)
@created(17/08/2008)

}

unit UCompSDMoodScaleDifferences;

interface

USES
        Forms, Classes,
        UCompDefinition,
        UCompManageDataset,
        UCompSDDefinition,
        UCompSDCondDescDefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcStatDesConditionnalDesc,
        UCalcStatDesRankForNonParametricStat; 

TYPE

        {gnrateur de composant one-way anova}
        TMLGenCompSDMoodScale = class(TMLGenCompNonParOneWay)
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant one anova}
        TMLCompSDMoodScale = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur de calcul }
        TOpSDMoodScale = class(TOpSDCondDescDefinition)
        protected
        function    getClassStat(): TClassCalcSDCondDesc; override;
        function    getClassLstStat(): TClassLstCalcStatDesCond; override;
        function    getClassParameter: TClassOperatorParameter; override;
        end;

        {paramtrage de l'oprateur de calcul}
        TOpPrmSDMoodScale = class(TOpPrmSDCondDescDefinition)
        end;

        {classe de calcul statistique}
        TCalcSDNonParamMoodScale = class(TCalcSDNonParamCondDescScores)
        protected
        function    getClassRankScoreComputation(): TClassRankComputed; override;
        end;

        {classe de calcul des scores}
        TScoresMoodScale = class(TRankComputed)
        protected
        procedure   computeScoresFromRank(lstSorted: TExamples); override;
        end;

implementation

USES
    FMath;

{ TMLGenCompSDMoodScale }

function TMLGenCompSDMoodScale.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDMoodScale;
end;

{ TMLCompSDMoodScale }

function TMLCompSDMoodScale.getClassOperator: TClassOperator;
begin
 result:= TOpSDMoodScale;
end;

{ TOpSDMoodScale }

function TOpSDMoodScale.getClassLstStat: TClassLstCalcStatDesCond;
begin
 result:= TLstCalcStatDesCondScores;
end;

function TOpSDMoodScale.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDMoodScale;
end;

function TOpSDMoodScale.getClassStat: TClassCalcSDCondDesc;
begin
 result:= TCalcSDNonParamMoodScale;
end;

{ TCalcSDNonParamMoodScale }

function TCalcSDNonParamMoodScale.getClassRankScoreComputation: TClassRankComputed;
begin
 result:= TScoresMoodScale;
end;

{ TScoresMoodScale }

procedure TScoresMoodScale.computeScoresFromRank(lstSorted: TExamples);
var i,example: integer;
    r,moyenne,nbExamples: double;
begin
 nbExamples:= 1.0 * lstSorted.Size;
 moyenne:= (1.0 + nbExamples)/2.0;
 for i:= 1 to lstSorted.Size do
  begin
   example:= lstSorted.Number[i];
   r:= FRankValue[example];
   FRankValue[example]:= SQR(r - moyenne);
  end;
end;

initialization
 RegisterClass(TMLGenCompSDMoodScale);
end.

