(********************************************************************)
(* UCompSDVanDerWaerden.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Comparaison de populations, mais avec une approche non paramtrique, celle de Van der Waerder)
@author(Ricco)
@created(17/08/2008)

}
unit UCompSDVanDerWaerden;

interface

USES
        Forms, Classes,
        UCompDefinition,
        UCompManageDataset,
        UCompSDDefinition,
        UCompSDCondDescDefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcStatDesConditionnalDesc,
        UCalcStatDesRankForNonParametricStat; 

TYPE

        {gnrateur de composant one-way anova}
        TMLGenCompSDVDW = class(TMLGenCompNonParOneWay)
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant one anova}
        TMLCompSDVDW = class(TMLCompSD)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur de calcul }
        TOpSDVDW = class(TOpSDCondDescDefinition)
        protected
        function    getClassStat(): TClassCalcSDCondDesc; override;
        function    getClassLstStat(): TClassLstCalcStatDesCond; override;
        function    getClassParameter: TClassOperatorParameter; override;
        end;

        {paramtrage de l'oprateur de calcul}
        TOpPrmSDVDW = class(TOpPrmSDCondDescDefinition)
        end;

        {classe de calcul statistique}
        TCalcSDNonParamVDW = class(TCalcSDNonParamCondDescScores)
        protected
        function    getClassRankScoreComputation(): TClassRankComputed; override;
        end;

        {classe de calcul des scores}
        TScoresVDW = class(TRankComputed)
        protected
        procedure   computeScoresFromRank(lstSorted: TExamples); override;
        end;

implementation

USES
    FMath;

{ TMLGenCompSDVDW }

function TMLGenCompSDVDW.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSDVDW;
end;

{ TMLCompSDVDW }

function TMLCompSDVDW.getClassOperator: TClassOperator;
begin
 result:= TOpSDVDW;
end;

{ TOpSDVDW }

function TOpSDVDW.getClassLstStat: TClassLstCalcStatDesCond;
begin
 result:= TLstCalcStatDesCondScores;
end;

function TOpSDVDW.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSDVDW;
end;

function TOpSDVDW.getClassStat: TClassCalcSDCondDesc;
begin
 result:= TCalcSDNonParamVDW;
end;

{ TCalcSDNonParamVDW }

function TCalcSDNonParamVDW.getClassRankScoreComputation: TClassRankComputed;
begin
 result:= TScoresVDW;
end;

{ TScoresVDW }

procedure TScoresVDW.computeScoresFromRank(lstSorted: TExamples);
var i,example: integer;
    r,denom,nbExamples: double;
begin
 nbExamples:= 1.0 * lstSorted.Size;
 denom:= 1.0 + nbExamples;
 for i:= 1 to lstSorted.Size do
  begin
   example:= lstSorted.Number[i];
   r:= FRankValue[example];
   FRankValue[example]:= invNorm(r / denom);
  end;
end;

initialization
 RegisterClass(TMLGenCompSDVDW);
end.
