(*****************************************************************************)
(* UCalcStructOrdinalAssociation.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(*****************************************************************************)

{
@abstract(Structure pour calculer les dpendances entre variables ordinales)
@author(Ricco)
@created(26/09/2007)

On utilise pleinement le support de cours crit dans le mme tempo
--> http://eric.univ-lyon2.fr/~ricco/cours/cours/Dependance_Variables_Qualitatives.pdf 
}
unit UCalcStructOrdinalAssociation;

interface

USES
    UDatasetDefinition,
    UDatasetExamples;

CONST
    //fixer le max. de valeurs diffrentes grable
    MAX_DISTINCT_ORDINAL_VALUES = 2048;

TYPE

    ///structure de vecteur unique des valeurs
    TVecValues = array[1..MAX_DISTINCT_ORDINAL_VALUES] of TTypeContinue;
    ///pointeur sur la structure
    PVecValues = ^TVecValues;

    ///classe de gestion des vecteur des valeurs
    TVecDistinctValues = class
    private
    //variable associe
    FVariable: TAttribute;
    //vecteur interne
    FVecValues: PVecValues;
    //nombre de valeurs en interne
    FNbValues: integer;
    //calculer les valeurs diffrentes
    procedure   computeDistinctValues();
    //trier selon les valeurs
    procedure   sortValues();
    public
    //constructeur
    constructor create(att: TAttribute);
    //dtruire
    destructor  destroy(); override;
    //rcuprer l'index d'une valeur
    function    getIndex(value: TTypeContinue): integer;
    //rcuprer la valeur corresp.  l'index
    function    getValue(index: integer): TTypeContinue;
    //le nombre de valeurs
    property    nbValues: integer read FNbValues;
    //la variable associatie
    property    Attribute: TAttribute read FVariable;
    end;

    ///type de comptage
    TCountValue = double;

    ///structure de tableau crois pour variables ordinales
    TTabOrdinal = array of array of TCountValue;

    ///classe table de contingence pour variables ordinales
    TCrossOrdinalVariables = class
    private
    //lments en ligne et en colonne
    FRow, FColumn: TVecDistinctValues;
    //nombre de lignes et de colonnes
    FNbRow, FNbColumn: integer;
    //effectifs
    FValues: TTabOrdinal;
    //les C_{lc} -- paires concordantes
    FClc: TTabOrdinal;
    //les D_{lc} -- paires discordantes
    FDlc: TTabOrdinal;
    //nombre de paires concordantes - P
    FNbConcPairs: double;
    //nombre de paires discordantes - Q
    FNbDiscPairs: double;
    //nombre de paires sans ex-aequo sur Y et sur X
    FDRow, FDColumn: double;
    //calculer les tableaux intermdiaires
    procedure   computeIntermediateTables();
    //calculer les paires
    procedure   computePairs();
    //obtenir la marge totale
    function    getNbExamples(): double;
    public
    //constructeur
    constructor create(row,column: TVecDistinctValues; examples : TExamples = nil);
    //detruire
    destructor  destroy(); override;
    //comptabiliser les tableaux
    procedure   refreshTables(examples: TExamples);
    //obtenir la description HTML du tableau des effectifs
    function    getHTMLDescription(): string;
    //paires ...
    property    P: double read FNbConcPairs;
    property    Q: double read FNbDiscPairs;
    property    DY: double read FDRow;
    property    DX: double read FDColumn;
    property    n: double read getNbExamples;
    //tableaux
    property    Values: TTabOrdinal read FValues;
    property    Clc: TTabOrdinal read FClc;
    property    Dlc: TTabOrdinal read FDlc;
    end;
    

implementation

USES
    Sysutils, Math,
    UConstConfiguration;

{ TVecDistinctValues }

procedure TVecDistinctValues.computeDistinctValues;
var i,k: integer;
    value: TTypeContinue;
begin
 //passer en revue toutes les observations pour avoir une description complte
 for i:= 1 to FVariable.Size do
  begin
   value:= FVariable.cValue[i];
   k:= self.getIndex(value);
   if (k < 0)
    then
     //il faut l'ajouter
     begin
      inc(FNbValues);
      //new -- 26/01/2009
      if (FVecValues = nil)
       then FVecValues:= AllocMem(FNbValues*sizeof(TTypeContinue))
       else ReAllocMem(FVecValues, FNbValues*sizeof(TTypeContinue));
      FVecValues^[FNbValues]:= value;
     end;
  end;
 //trier maintenant les valeurs par ordre croissant
 if (FNbValues > 1) then self.sortValues();
end;

constructor TVecDistinctValues.create(att: TAttribute);
begin
 inherited create();
 FVariable:= att;
 FVecValues:= nil;
 FNbValues:= 0;
 //calculer les valeurs distinctes
 self.computeDistinctValues();
end;

destructor TVecDistinctValues.destroy;
begin
 if (FVecValues <> nil) then ReAllocMem(FVecValues,0);
 inherited;
end;

function TVecDistinctValues.getIndex(value: TTypeContinue): integer;
var i: integer;
begin
 result:= -1;
 //recherche squentielle trs simple mais bon...
 for i:= 1 to FNbValues do
  begin
   if (FVecValues^[i] = value)
    then
     begin
      result:= i;
      break;
     end;
  end;
end;

function TVecDistinctValues.getValue(index: integer): TTypeContinue;
begin
 if (index >= 1) and (index <= FNbValues)
  then result:= FVecValues^[index]
  else result:= -math.MaxSingle;
end;

procedure TVecDistinctValues.sortValues;
LABEL 99;
VAR
   l,j,ir,i,n: integer;
   rra: TTypeContinue;
BEGIN
   //taille du tableau
   n:= self.FNbValues;

   l := (n DIV 2)+1;
   ir := n;
   WHILE true DO BEGIN
      IF (l > 1) THEN BEGIN
         l := l-1;
         rra := FVecValues^[l]
      END ELSE BEGIN
         rra := FVecValues^[ir];
         FVecValues^[ir] := FVecValues^[1];
         ir := ir-1;
         IF (ir = 1) THEN BEGIN
            FVecValues^[1] := rra;
            GOTO 99
         END
      END;
      i := l;
      j := l+l;
      WHILE (j <= ir) DO BEGIN
         IF (j < ir) THEN
            IF (FVecValues^[j] < FVecValues^[j+1]) THEN j := j+1;
         IF (rra < FVecValues^[j]) THEN BEGIN
            FVecValues^[i] := FVecValues^[j];
            i := j;
            j := j+j
         END ELSE
            j := ir+1
      END;
      FVecValues^[i] := rra
   END;
99:
END;

{ TCrossOrdinalVariables }

procedure TCrossOrdinalVariables.computeIntermediateTables;
var l,c: integer;
    h,k: integer;
    s: TCountValue;
begin
 //calcul pour C_{lc}
 for l:= 1 to FNbRow do
  for c:= 1 to FNbColumn do
   begin
    s:= 0.0;
    //coin haut gauche
    for h:= 1 to pred(l) do for k:= 1 to pred(c) do s:= s + FValues[h,k];
    //coin bas droite
    for h:= succ(l) to FNbRow do for k:= succ(c) to FNbColumn do s:= s + FValues[h,k];
    //and then...
    FClc[l,c]:= s;
   end;
 //calcul pour D_{lc}
 for l:= 1 to FNbRow do
  for c:= 1 to FNbColumn do
   begin
    s:= 0.0;
    //coin bas gauche
    for h:= succ(l) to FNbRow do for k:= 1 to pred(c) do s:= s + FValues[h,k];
    //coin bas droite
    for h:= 1 to pred(l) do for k:= succ(c) to FNbColumn do s:= s + FValues[h,k];
    //and then...
    FDlc[l,c]:= s;
   end;
end;

procedure TCrossOrdinalVariables.computePairs;
var l,c: integer;
    pc,pd,s,n: TCountValue;
begin
 //paires concordantes et discordantes
 pc:= 0.0;
 pd:= 0.0;
 for l:= 1 to FNbRow do
  for c:= 1 to FNbColumn do
   begin
    pc:= pc + FValues[l,c] * FClc[l,c];
    pd:= pd + FValues[l,c] * FDlc[l,c];
   end;
 //and then...
 FNbConcPairs:= pc;
 FNbDiscPairs:= pd;
 //pour les ex-aequos
 n:= FValues[0,0];
 s:= 0.0;
 for l:= 1 to FNbRow do s:= s + FValues[l,0] * FValues[l,0];
 FDrow:= n * n - s;
 s:= 0.0;
 for c:= 1 to FNbColumn do s:= s + FValues[0,c] * FValues[0,c];
 FDColumn:= n * n - s;
end;

constructor TCrossOrdinalVariables.create(row, column: TVecDistinctValues; examples : TExamples = nil);
begin
 inherited create();
 FRow:= row;
 FColumn:= column;
 setlength(FValues,FRow.nbValues+1,FColumn.nbValues+1);
 setlength(FClc,FRow.nbValues+1,FColumn.nbValues+1);
 setlength(FDlc,FRow.nbValues+1,FColumn.nbValues+1);
 FNbRow:= FRow.nbValues;
 FNbColumn:= FColumn.nbValues;
 if (examples <> nil) then self.refreshTables(examples);
end;

destructor TCrossOrdinalVariables.destroy;
begin
 Finalize(FValues);
 Finalize(FClc);
 Finalize(FDlc);
 inherited;
end;

function TCrossOrdinalVariables.getHTMLDescription: string;
var s: string;
    l,c: integer;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_WHITE + format('<TD align="center" colspan =%d>Conc.pairs = %.0f ; Disc.pairs = %.0f</TD></TR>',[FColumn.nbValues+2,P,Q]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Y vs. X</TH>';
 for c:= 1 to FColumn.nbValues do
  s:= s +format('<TH width=60>%.1f</TH>',[FColumn.getValue(c)]);
 s:= s + format('<TH %s width=60>Sum</TH></TR>',[HTML_BGCOLOR_HEADER_GRAY]);
 //pour chaque ligne du tableau
 for l:= 1 to FRow.nbValues do
  begin
   s:= s + format('<TR><TH %s>%.1f</TH>',[HTML_BGCOLOR_HEADER_GRAY,FRow.getValue(l)]);
   for c:= 1 to FColumn.nbValues do s:= s + format('<TD %s align=right>%.0f</TD>',[HTML_BGCOLOR_DATA_WHITE,FValues[l,c]]);
   s:= s + format('<TD %s align=right>%.0f</TD>',[HTML_BGCOLOR_DATA_WHITE,FValues[l,0]]);
  end;
 //la marge colonne
 s:= s + format('<TR><TH %s width=60>Sum</TH>',[HTML_BGCOLOR_HEADER_GRAY]);
 for c:= 1 to FColumn.nbValues do s:= s + format('<TD %s align=right>%.0f</TD>',[HTML_BGCOLOR_DATA_WHITE,FValues[0,c]]);
 s:= s + format('<TD %s align=right>%.0f</TD>',[HTML_BGCOLOR_DATA_WHITE,FValues[0,0]]);
 s:= s + '</TR>';

 s:= s + '</table>';
 result:= s;
end;

function TCrossOrdinalVariables.getNbExamples: double;
begin
 result:= FValues[0,0];
end;

procedure TCrossOrdinalVariables.refreshTables(examples: TExamples);
var i,l,c,example: integer;
    vl,vc: TTypeContinue;

    procedure clearTables(tab: TTabOrdinal);
    var i,j: integer;
    begin
     for i:= 0 to FRow.nbValues do
      for j:= 0 to FColumn.nbValues do
       tab[i,j]:= 0.0;
    end;

begin
 //vider
 (*
 FillChar(FValues[0,0],(FRow.nbValues+1)*(FColumn.nbValues+1)*sizeof(TCountValue),0);
 FillChar(FClc[0,0],(FRow.nbValues+1)*(FColumn.nbValues+1)*sizeof(TCountValue),0);
 FillChar(FDlc[0,0],(FRow.nbValues+1)*(FColumn.nbValues+1)*sizeof(TCountValue),0);
 *)
 clearTables(FValues);
 clearTables(FClc);
 clearTables(FDlc);
 //remplir pour chaque exemple
 for i:= 1 to examples.size do
  begin
   example:= examples.Number[i];
   vl:= FRow.Attribute.cValue[example]; l:= FRow.getIndex(vl);
   vc:= FColumn.Attribute.cValue[example]; c:= FColumn.getIndex(vc);
   FValues[l,c]:= FValues[l,c] + 1.0;
   FValues[l,0]:= FValues[l,0] + 1.0;
   FValues[0,c]:= FValues[0,c] + 1.0;
   FValues[0,0]:= FValues[0,0] + 1.0;
  end;
 //calculer maintenant les tableaux intermdiaires C_{lc} et D_{lc}
 self.computeIntermediateTables();
 //puis les paires et les ex-aequos
 self.computePairs();
end;

end.
