(****************************************************************************)
(* UCompOutlierDetectUnivariate.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(****************************************************************************)

{
@abstract(Dtection des points atypiques pour les variables continues. Mthodes univaries)
@author(Ricco)
@created(22/06/2007)

Un test global, dit Test de Grubbs (cf. http://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm)
pour dceler l'existence d'un point aberrant.

Deux approches possibles pour dtecter les points
(1) rgle de l'cart-type (la rgle des 3 sigmas est la plus connue)
(2) rgle de l'intervalle inter-quartile (cf. http://www.itl.nist.gov/div898/handbook/prc/section1/prc16.htm)
}
unit UCompOutlierDetectUnivariate;

interface

USES
  Forms, Classes, Inifiles,  
  UCompDefinition,
  UOperatorDefinition, UCompManageDataset,
  UCalcStatDesContinuousPlus;

TYPE
  {Gnrateur de univariate stat cont desc}
  TMLGenCompODUnivariate = class(TMLGenComp)
  protected
  procedure   GenCompInitializations(); override;
  public
  function    GetClassMLComponent: TClassMLComponent; override;
  end;

  {le composant}
  TMLCompODUnivariate = class(TMLCompLocalData)
  protected
  //dsactivation de la gestion standard de sortie de exemples -> ainsi on peut filtrer
  procedure  RefreshOutput(); override;
  //classe de calcul
  function    getClassOperator: TClassOperator; override;
  end;

  TEnumOD = (odLBound, odUBound, odCount);

  //bornes pour la dtection
  TTabUnivOD = array of array[TEnumOD] of double;

  {l'oprateur associ}
  TOpODUnivariate = class(TOpLocalData)
  protected
  //stat. descriptive avances
  stats : TLstCalcStatDesContinuousPlus;
  //seuil pour le point critique de la statistique de Grubbs
  GrubbsSeuil: double;
  //bornes et statistiques calcules pour chaque variable
  Grubbs : array of double;//statistique de Grubbs, test bilatral
  Sigma: TTabUnivOD;//borne basse et haute de la rgle des "m" sigmas
  InnerFence : TTabUnivOD;//bornes pour le test faible de l'intervalle inter-quartile
  OuterFence : TTabUnivOD;//bornes pour le test fort de l'intervalle inter-quartile
  Detected : array of array of byte; //indicateur pour chaque individu, dtect ou pas
  NbDetectedExamples: integer;//nombre d'individus dtects
  //calculer les seuils
  procedure   computeSeuilGrubbs();
  //
  function    getClassParameter: TClassOperatorParameter; override;
  function    CheckAttributes(): boolean; override;
  function    CoreExecute(): boolean; override;
  //montrer les individus suspects en HTML
  function    showExamplesHTML(): string;
  public
  destructor  destroy(); override;
  function    getHTMLResultsSummary(): string; override; 
  end;

  {le paramtrage}
  TOpPrmODUnivariate = class(TOperatorParameter)
  protected
  function    CreateDlgParameters(): TForm; override;
  procedure   SetDefaultParameters(); override;
  public
  //p-value pour le test de Grubbs
  pValue: double;
  //multiple de l'cart-type pour la rgle de l'cart-type
  mSigma: double;
  //montrer le dtail sur les individus dtects ?
  isShowDetectedExamples: boolean;
  //faut-il utiliser les fences ?
  isUseFences: boolean;
  //filtrage des outliers ?
  isRemoveOutliers: boolean;
  //quel(s) critre(s) ?
  cSigma, cInnerFence, cOuterFence: boolean;
  function    getHTMLParameters(): string; override;
  procedure   LoadFromStream(prmStream: TStream); override;
  procedure   SaveToStream(prmStream: TStream); override;
  procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
  procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
  end;


implementation

uses
    Sysutils, Math,
    UDatasetImplementation, UDatasetDefinition, LOISTAT, ULogFile,
    UCalcStatDes, UStringAddBuffered, UConstConfiguration,
    UDlgOpPrmODUnivariate, UDatasetExamples;

{ TMLGenCompODUnivariate }

procedure TMLGenCompODUnivariate.GenCompInitializations;
begin
 FMLComp:= mlcDescriptiveStat;
end;

function TMLGenCompODUnivariate.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompODUnivariate; 
end;

{ TMLCompODUnivariate }

function TMLCompODUnivariate.getClassOperator: TClassOperator;
begin
 result:= TOpODUnivariate; 
end;

procedure TMLCompODUnivariate.RefreshOutput;
begin
 //rcuprer l'entre tout btement, et ajouter les variables locales
 if assigned(self.Predecessor) and self.Predecessor.InheritsFrom(TMLCompLocalData)
  then FOutputMLData.CopyWithoutExamples((self.Predecessor as TMLCompLocalData).OutputData);
end;

{ TOpODUnivariate }

function TOpODUnivariate.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue) = true);
end;

procedure TOpODUnivariate.computeSeuilGrubbs;
var alpha, n, t: double;
begin
 n:= 1.0 * self.WorkData.Examples.Size;
 alpha:= (self.PrmOp as TOpPrmODUnivariate).pValue;
 //corriger le risque critique
 alpha:= alpha / (2.0 * n);
 t:= ILS_STUDENT(1.0 - alpha,n - 2.0);
 //TraceLog.WriteToLogFile(format('[OD Univariate], t de student pour proba = %.4f, est egal a %.4f',[1.0-alpha,t]));
 //cf. http://www.itl.nist.gov/div898/handbook/eda/section3/eda35h.htm
 GrubbsSeuil:= (n - 1.0)/sqrt(n) * sqrt(power(t,2.0) / (n - 2.0 + power(t,2.0)));
end;

function TOpODUnivariate.CoreExecute: boolean;
var i,j,nbVar,example: integer;
    stat: TCalcStatDesContPlus;
    G: double;
    m: double;
    IQ, Q1, Q2: double;
    att: TAttribute;
    value: double;
    toRemove: boolean;
    nbToRemove: integer;
    sortie: TExamples;
    prm: TOpPrmODUnivariate;
begin
 prm:= self.PrmOp as TOpPrmODUnivariate;
 m:= prm.mSigma;
 //calcul du seuil de Grubbs
 self.computeSeuilGrubbs();
 //calcul des statistiques descriptives
 if assigned(stats) then freeAndNil(stats);
 stats:= TLstCalcStatDesContinuousPlus.Create(workdata.LstAtts[asInput],workdata.Examples);
 //prparer les tableaux
 nbVar:= self.WorkData.LstAtts[asInput].Count;
 setLength(Grubbs,nbVar);
 setLength(Sigma,nbVar);
 setLength(InnerFence,nbVar);
 setLength(OuterFence,nbVar);
 //calculer la statistique de Grubbs et les bornes
 for j:= 0 to pred(nbVar) do
  begin
   stat:= stats.stat(j) as TCalcStatDesContPlus;
   //Grubbs
   G:= MAX(ABS(stat.Max - Stat.Average),ABS(stat.Average - stat.Min));
   if (stat.StdDev > 0) then G:= G / stat.StdDev;
   Grubbs[j]:= G;
   //bornes rgle de l'cart-type
   Sigma[j][odLBound]:= stat.Average - m * stat.StdDev;
   Sigma[j][odUBound]:= stat.Average + m * stat.StdDev;
   Sigma[j][odCount] := 0;
   //intervalle inter-quartile, cf. http://www.itl.nist.gov/div898/handbook/prc/section1/prc16.htm
   IQ:= stat.InterQuartileRange;
   Q1:= stat.LowerQuartile;
   Q2:= stat.UpperQuartile;
   //bornes "Inner"
   InnerFence[j][odLBound]:= Q1 - 1.5 * IQ;
   InnerFence[j][odUBound]:= Q2 + 1.5 * IQ;
   InnerFence[j][odCount] := 0;
   //brones "Outer"
   OuterFence[j][odLBound]:= Q1 - 3.0 * IQ;
   OuterFence[j][odUBound]:= Q2 + 3.0 * IQ;
   OuterFence[j][odCount] := 0;
  end;
 //****************************
 //comptage et retrait ventuel
 //****************************
 sortie:= self.OutputData.Examples;
 sortie.BeginAdd();
 nbToRemove:= 0;
 //dtection
 NbDetectedExamples:= 0;
 setLength(Detected,succ(self.WorkData.Examples.Size),succ(nbVar));
 for i:= 1 to self.WorkData.Examples.Size do
  begin
   example:= self.WorkData.Examples.Number[i];
   toRemove:= false;
   for j:= 0 to pred(nbVar) do
    begin
     Detected[i][j]:= 0;
     att:= self.WorkData.LstAtts[asInput].Attribute[j];
     value:= att.cValue[example];
     //dtecter
     if (value < Sigma[j][odLBound]) or (value > Sigma[j][odUBound]) then
      begin
       Sigma[j][odCount]:= Sigma[j][odCount] + 1.0;
       inc(Detected[i][j]);
       if (prm.isRemoveOutliers and prm.cSigma) then toRemove:= true;
      end;
     if (prm.isUseFences = true)
      then
       begin
        if (value < InnerFence[j][odLBound]) or (value > InnerFence[j][odUBound])
         then
          begin
           InnerFence[j][odCount]:= InnerFence[j][odCount] + 1.0; inc(Detected[i][j]);
           if (prm.isRemoveOutliers and prm.cInnerFence) then toRemove:= true;
          end;
        if (value < OuterFence[j][odLBound]) or (value > OuterFence[j][odUBound])
         then
          begin
           OuterFence[j][odCount]:= OuterFence[j][odCount] + 1.0; inc(Detected[i][j]);
           if (prm.isRemoveOutliers and prm.cOuterFence) then toRemove:= true;
          end;
       end;
    end;
   //comptabiliser pour l'individu i
   Detected[i][nbVar]:= 0;
   for j:= 0 to pred(nbVar) do Detected[i][nbVar]:= Detected[i][nbVar] + ord(Detected[i][j] > 0);
   NbDetectedExamples:= NbDetectedExamples + ord(Detected[i][nbVar] > 0);
   //et si on doit grer la suppression
   if (prm.isRemoveOutliers = true)
    then if not(toRemove) then sortie.AddExample(example) else inc(nbToRemove)
    else sortie.AddExample(example);
  end;
 sortie.EndAdd();
 TraceLog.WriteToLogFile(format('[Univariate outlier detection] Removed examples : %d, keeped examples = %d',[nbToRemove,sortie.Size]));
 //on est mal s'il n'y a plus personne en sortie
 result:= (sortie.Size > 0);
end;

destructor TOpODUnivariate.destroy;
begin
 if assigned(stats) then freeAndNil(stats);
 setLength(Grubbs,0);
 setLength(Sigma,0);
 setLength(InnerFence,0);
 setLength(OuterFence,0);
 if (detected <> nil) then finalize(detected);
 inherited;
end;

function TOpODUnivariate.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmODUnivariate; 
end;

function TOpODUnivariate.getHTMLResultsSummary: string;
var s: string;
    j: integer;
    od: TEnumOD; 
begin
 s:= '<H3>Univariate Outliers Detection</H3>';
 //former le tableau de rsultats et lister chaque variable
 s:= s + '<H4>Detailed results for each variable</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Variable</TH><TH>Grubbs Stat.</TH><TH colspan=3>Sigma rule</TH><TH colspan=3>Inner Fence rule</TH><TH colspan=3>Outer Fence Rule</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TH>-</TH><TH>Cut : %.4f</TH><TH>L.B</TH><TH>U.B</TH><TH>Detected</TH><TH>L.B</TH><TH>U.B</TH><TH>Detected</TH><TH>L.B</TH><TH>U.B</TH><TH>Detected</TH></TR>',[GrubbsSeuil]);
 for j:= 0 to pred(workdata.LstAtts[asInput].Count) do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[self.Workdata.LstAtts[asInput].Attribute[j].Name]);
   s:= s + format('<TD align=right>%.4f</TD>',[Grubbs[j]]);
   for od:= odLBound to odUBound do s:= s + format('<TD align=right>%.4f</TD>',[Sigma[j][od]]);
   s:= s + format('<TD align=right>%.0f</TD>',[Sigma[j][odCount]]);
   for od:= odLBound to odUBound do s:= s + format('<TD align=right>%.4f</TD>',[InnerFence[j][od]]);
   s:= s + format('<TD align=right>%.0f</TD>',[InnerFence[j][odCount]]);
   for od:= odLBound to odUBound do s:= s + format('<TD align=right>%.4f</TD>',[OuterFence[j][od]]);
   s:= s + format('<TD align=right>%.0f</TD>',[OuterFence[j][odCount]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 //montrer les individus ?
 if (self.PrmOp as TOpPrmODUnivariate).isShowDetectedExamples then s:= s + self.showExamplesHTML();
 //des individus ont-ils t enlevs
 s:= s + format('<H4>Removed outliers (if option activated) : %d</H4>',[self.WorkData.Examples.Size - self.OutputData.Examples.Size]);
 //and then...
 result:= s;
end;

function TOpODUnivariate.showExamplesHTML: string;
var buf: TBufString;
    i,example,nbVar,j: integer;
    s: string;
begin

 nbVar:= workdata.LstAtts[asInput].Count;
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr('<H4>Detected Outliers</H4>');
 buf.AddStr(format('<b># outliers : %d</b><P>',[NbDetectedExamples]));
 buf.AddStr(HTML_HEADER_TABLE_RESULT);
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY + '<TH>n example</TH><TH># detection</TH><TH>Variable(s)</TH></TR>');
 for i:= 1 to self.WorkData.Examples.Size do
  begin
   if (Detected[i][nbVar] > 0)
    then
     begin
      example:= self.WorkData.Examples.Number[i];
      //former la chane des variables
      s:= '';
      for j:= 0 to pred(nbVar) do
       if (Detected[i][j] > 0) then s:= s + workdata.LstAtts[asInput].Attribute[j].Name + ' ; ';
      s:= copy(s,1,length(s)-3);
      //on n'affiche pas le dtail par variable pour l'instant
      buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TD>%d</TD><TD align=center>%d</TD><TD>%s</TD>',[example,Detected[i][nbVar],s])+'</TR>');
     end;
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

{ TOpPrmODUnivariate }

function TOpPrmODUnivariate.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmODUnivariate.CreateFromOpPrm(self);
end;

function TOpPrmODUnivariate.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Parameters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Grubbs'' p-value</TD><TD align=right>%.4f</TD></TR>',[pValue]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Multiple of sigma</TD><TD align=right>%.1f</TD></TR>',[mSigma]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Use Fences</TD><TD align=center>%d</TD></TR>',[ord(isUseFences)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Show outliers</TD><TD align=center>%d</TD></TR>',[ord(isShowDetectedExamples)]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Filtering Parameters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Remove outliers</TD><TD align=center>%d</TD></TR>',[ord(isRemoveOutliers)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Use sigma</TD><TD align=center>%d</TD></TR>',[ord(cSigma)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Use inner fence</TD><TD align=center>%d</TD></TR>',[ord(cInnerFence)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Use outer fence</TD><TD align=center>%d</TD></TR>',[ord(cOuterFence)]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmODUnivariate.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 pValue:= prmINI.ReadFloat(prmSection,'pvalue',pValue);
 mSigma:= prmINI.ReadFloat(prmSection,'mSigma',mSigma);
 isShowDetectedExamples:= prmINI.ReadBool(prmSection,'show_detected',isShowDetectedExamples);
 isUseFences:= prmINI.ReadBool(prmSection,'use_fences',isUseFences);
 isRemoveOutliers:= prmINI.ReadBool(prmSection,'isRemoveOutliers',isRemoveOutliers);
 cSigma:= prmINI.ReadBool(prmSection,'cSigma',cSigma);
 cInnerFence:= prmINI.ReadBool(prmSection,'cInnerFence',cInnerFence);
 cOuterFence:= prmINI.ReadBool(prmSection,'cOuterFence',cOuterFence);
end;

procedure TOpPrmODUnivariate.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(pValue,sizeof(pValue));
 prmStream.ReadBuffer(mSigma,sizeof(mSigma));
 prmStream.ReadBuffer(isShowDetectedExamples,sizeof(isShowDetectedExamples));
 prmStream.ReadBuffer(isUseFences,sizeof(isUseFences));
 prmStream.ReadBuffer(isRemoveOutliers,sizeof(isRemoveOutliers));
 prmStream.ReadBuffer(cSigma,sizeof(cSigma));
 prmStream.ReadBuffer(cInnerFence,sizeof(cInnerFence));
 prmStream.ReadBuffer(cOuterFence,sizeof(cOuterFence));
end;

procedure TOpPrmODUnivariate.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteFloat(prmSection,'pvalue',pValue);
 prmINI.WriteFloat(prmSection,'mSigma',mSigma);
 prmINI.WriteBool(prmSection,'show_detected',isShowDetectedExamples);
 prmINI.WriteBool(prmSection,'use_fences',isUseFences);
 prmINI.WriteBool(prmSection,'isRemoveOutliers',isRemoveOutliers);
 prmINI.WriteBool(prmSection,'cSigma',cSigma);
 prmINI.WriteBool(prmSection,'cInnerFence',cInnerFence);
 prmINI.WriteBool(prmSection,'cOuterFence',cOuterFence);
end;

procedure TOpPrmODUnivariate.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(pValue,sizeof(pValue));
 prmStream.WriteBuffer(mSigma,sizeof(mSigma));
 prmStream.WriteBuffer(isShowDetectedExamples,sizeof(isShowDetectedExamples));
 prmStream.WriteBuffer(isUseFences,sizeof(isUseFences));
 prmStream.WriteBuffer(isRemoveOutliers,sizeof(isRemoveOutliers));
 prmStream.WriteBuffer(cSigma,sizeof(cSigma));
 prmStream.WriteBuffer(cInnerFence,sizeof(cInnerFence));
 prmStream.WriteBuffer(cOuterFence,sizeof(cOuterFence));
end;

procedure TOpPrmODUnivariate.SetDefaultParameters;
begin
 //p-value
 pValue:= 0.05;
 //multiple de sigma
 mSigma:= 3.0;
 //montrer les dtects
 isShowDetectedExamples:= false;
 //utiliser les fences ?
 isUseFences:= true;
 //supprimer les individus ?
 isRemoveOutliers:= true;
 //quel critre
 cSigma:= false;
 cInnerFence:= false;
 cOuterFence:= true;
end;

initialization
 RegisterClass(TMLGenCompODUnivariate);
end.
