(********************************************************************)
(* UCompFactAnalysisPCA.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Analyse en Composantes Principales)
@author(Ricco)
@created(12/01/2004)

new -- 30/06/2011 -- introduire l'ACP non norme

}
unit UCompFactAnalysisPCA;

interface

USES
        Forms,Classes,
        IniFiles,
        Matrices,
        UCalcStatDes,
        UCompDefinition,
        UCompFADefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation;

TYPE
        {gnrateur de composant ACP}
        TMLGenCompFactPCA = class(TMLGenComp)
                            protected
                            procedure   GenCompInitializations(); override;
                            public
                            function    GetClassMLComponent: TClassMLComponent; override;
                            end;

        {composant ACP}
        TMLCompFactPCA = class(TMLCompFactAnalysis)
                         protected
                         function    getClassOperator: TClassOperator; override;
                         end;

        //index des variables tries
        TPCAIndexVariables = array of integer;

        {oprateur ACP, on travaille sur les INPUT}
        TOpFactPCA = class(TOpFactAnalysis)
                     protected
                     {Trace de la matrice  diagonaliser}
                     FTrace: double;
                     {Nombre d'axes  traiter}
                     FNbAxis: integer;
                     {StatDes}
                     FLstStats: TLstCalcStatDesContinuous;
                     {dimension des matrices}
                     FDimAnalysis: integer;
                     {matrice de corrlation}
                     FCorrMat: PMatrix;
                     {matrice, corrlation des variables avec les axes}
                     FLoadings: PMatrix;
                     {vecteur des valeurs propres}
                     //FEgVal: PVector;
                     {matrice des vecteurs propres}
                     FEgVec: PMatrix;
                     //new - 07/09/2012 - factor scores (coef. de projection)
                     FFactorScores: PMatrix;
                     //new - 06/06/2012 - indicateurs pour le test de Bartlett
                     FDetR, FBartlett, FDf, FPValue: double;
                     //new - 06/06/2012 - indicateurs pour l'indice KMO, global et par variable
                     FKmo: double;
                     FTabKmo: array of double;
                     //matrice des corrlations partielles
                     FPartialR: PMatrix;
                     //inverse of the correlation matrix
                     FInvMatCorr: PMatrix;
                     //Prior communality estimate
                     FPriorCommunality: array of double;
                     {calculer les stats descriptives}
                     procedure   CalcStats();
                     {construire la matrice de corrlation}
                     procedure   BuildVCV(); virtual;
                     {calculer les valeurs et vecteurs propres}
                     procedure   BuildEigen(); virtual;
                     {calculer les corrlations des variables avec les axes}
                     procedure   CalcLoadings(); virtual;
                     //new -- 07/09/2012 -- calculer les coefficients de projection
                     procedure   CalcFactorScores(); virtual;
                     //new -- 06/06/2012 -- calculer le test de Bartlett et l'indice KMO
                     procedure   calcBartlett(); virtual;
                     procedure   calcKMO(); virtual;
                     procedure   calcBartlettKMO;
                     {dtruire les matrices et vecteurs}
                     procedure   DelPrivateMatrix(); override;
                     function    getClassParameter: TClassOperatorParameter; override;
                     function    CheckAttributes(): boolean; override;
                     //function    CoreExecute(): boolean; override;
                     function    getNewForm(): TForm; override;
                     function    getDimEigVal(): integer; override;
                     //trier la liste des variables selon les loadings, sur plusieurs colonnes
                     function    sortOnLoadings(loadings: PMatrix; nbAxis: integer): TPCAIndexVariables;
                     //obtenir un index standard des variables
                     function    getVariablesIndex(): TPCAIndexVariables;
                     //new - 23/09/2012 - obtenir un prfixe pour les noms de variables en projection
                     function    getPrexifNameForProjection(): string; virtual;
                     public
                     constructor Create(AOwner: TObject); override;
                     constructor CreateWithDataset(AOwner: TObject; prmDataset: TMLDataset); override;
                     destructor  Destroy; override;
                     function    getHTMLResultsSummary(): string; override;
                     function    getHTMLLoadings(loadings: PMatrix; nbFactors: integer; sorted: boolean; loadingsForSorting: PMatrix = nil): string; virtual;
                     function    getHTMLRegCoef(factorCoefs: PMatrix; loadings: PMatrix; nbFactors: integer): string; virtual;
                     //new - 06/06/2012 - tester la significativit des axes
                     function    getHTMLEigenValues(): string;
                     function    getHTMLSignificance(): string; virtual;
                     //new - 07/06/2012 - statistiques optionnelles (Bartlett et Kmo)
                     function    getHTMLBartlett(): string; virtual;
                     function    getHTMLKmo(): string; virtual;
                     //new - 07/06/2012 - les matrices de calcul (corrlation ou covariance)
                     function    getHTMLCorrMatrices(): string;
                     //new - 13/09/2012 - matrice reproduite et rsidu
                     function    getHTMLRepResCorrMatrice(loadings: PMatrix; nbAxis: integer; sorted: boolean): string;
                     {effectuer les projections sur les axes factoriels}
                     procedure   SetProjections(atts: TLstAttributes; coefs: PMatrix; nbFactors: integer; prefix: string = 'PCA'; number: integer = 1);
                     //new -- 04/02/2007 -- calculer les cos2 et ctr
                     procedure   setCos2Ctr(atts: TLstAttributes; nbFactors: integer; prefix: string = 'PCA'; number : integer = 1);
                     //new - 08/06/2012 - dcomposer les calculs en 3 tapes
                     function    calcEigen(): boolean; override;
                     function    calcAdditional(): boolean; override;
                     function    calcProjections(): boolean; override;
                     //new - 08/06/2012 - obtenir une instance d'analyse factorielle
                     function    getInstanceFA(): TClassOpFactAnalysis; override;
                     //new - 07/09/2012 - obtenir les coefficients aprs rotation (varimax ou quartimax)
                     function    getRotatedCoefs(newLoadings, newEgVec: PMatrix; nbFactors: integer): PMatrix; virtual;
                     //obtenir un index des variables, ventuellement tri selon les loadings...
                     function    getIndexVariables(sorted: boolean; refLoadings: PMatrix; nbAxis: integer): TPCAIndexVariables;
                     //proprits
                     property    Trace: double read FTrace;
                     property    Loadings: PMatrix read FLoadings;
                     property    EgVec: PMatrix read FEgVec;
                     property    DimAnalysis: integer read FDimAnalysis;
                     property    NbAxis: integer read FNbAxis;
                     property    MatCorr: PMatrix read FCorrMat;
                     //property    EigVal: PVector read FEgVal;
                     end;

        {paramtre oprateur ACP}
        TOpPrmFactPCA = class(TOpPrmFactAnalysis)
                        private
                        {nombre d'axes  crer, si -1 on les gnre tous}
                        FNbAxisCreated: integer;
                        //new -- 04/02/2007 -- projeter les COS et CTR par individu, df. false
                        FCos2Ctr: boolean;
                        //new -- 30/06/2011 -- acp non norme ou norme
                        FStandardize: boolean;
                        //new - 07/06/2012 - calculer le test de Bartlett et l'indice kmo
                        FBartlettKmo: boolean;
                        //new - 07/06/2012 - montrer la matrice des corrlations partielles
                        FShowCorr: boolean;
                        //new - 13/09/2012 - montrer la matrice de corrlation reconstruite et le rsidu
                        FShowRepCorr: boolean;
                        //new - 15/09/2012
                        FSortVarOnLoadings: boolean;
                        protected
                        function    CreateDlgParameters(): TForm; override;
                        procedure   SetDefaultParameters(); override;
                        public
                        function    getHTMLParameters(): string; override;
                        procedure   LoadFromStream(prmStream: TStream); override;
                        procedure   SaveToStream(prmStream: TStream); override;
                        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
                        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
                        procedure   getFactSettings(prmSource: TOpPrmFactAnalysis); override;
                        property    NbAxisCreated: integer read FNbAxisCreated write FNbAxisCreated;
                        property    Cos2Ctr: boolean read FCos2Ctr write FCos2Ctr;
                        property    Standardize: boolean read FStandardize write FStandardize;
                        property    BartlettKmo: boolean read FBartlettKmo write FBartlettKmo;
                        property    ShowCorr: boolean read FShowCorr write FShowCorr;
                        property    ShowRepCorr: boolean read FShowRepCorr write FShowRepCorr;
                        property    SortVarOnLoadings: boolean read FSortVarOnLoadings write FSortVarOnLoadings;
                        end;


        //code de couleur selon la valeur absolue
        function codeCouleurHTMLLoadings(value: double): string;

implementation

USES
        SysUtils,Math,
        Contnrs,
        Windows,
        UCalcMatrixToAttributes,
        UCalcMatrixAdditionalFunctions, UConstConfiguration,
        UDlgOpPrmPCA, UStringsResources, UStringAddBuffered, ULogFile,
        UCalcFactRotateFactors, UFrmOpViewFactPCA, FMath, UDatasetExamples;

CONST
        SEUIL_DET_MAT_CORR : double = Math.MinSingle;

        SEUIL_COR_HIGHLIGTH_LOW : double = 0.5;
        SEUIL_COR_HIGHLIGTH_HIGH : double = 0.7;

        SEUIL_RESIDUAL_COR_HIGHLIGTH : double = 0.05;


//canceled
(*
function codeCouleurHTMLLoadings(value: double): string;
var s: string;
begin
 value:= abs(value);
 s:= HTML_BGCOLOR_DATA_GRAY;
 if (value > SEUIL_COR_HIGHLIGTH_LOW) then s:= HTML_BGCOLOR_LIGHT_RED;
 if (value > SEUIL_COR_HIGHLIGTH_HIGH) then s:= HTML_BGCOLOR_DATA_RED;
 result:= s;
end;
*)

//code de couleur selon la valeur absolue
function codeCouleurHTMLLoadings(value: double): string;
var s: string;
begin
 s:= HTML_BGCOLOR_DATA_GRAY;
 //distinguer les valeurs positives de valeurs negatives
 if (value < -SEUIL_COR_HIGHLIGTH_LOW) then s:= FACT_COLOR_HIGHLIGHT[cthNegLow];
 if (value < -SEUIL_COR_HIGHLIGTH_HIGH) then s:= FACT_COLOR_HIGHLIGHT[cthNegHigh];
 if (value > +SEUIL_COR_HIGHLIGTH_LOW) then s:= FACT_COLOR_HIGHLIGHT[cthPosLow];
 if (value > +SEUIL_COR_HIGHLIGTH_HIGH) then s:= FACT_COLOR_HIGHLIGHT[cthPosHigh];
 //
 result:= s;
end;

{ TMLGenCompFactPCA }

procedure TMLGenCompFactPCA.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
 //FMLNumIcon:= 10;
 //FMLCompName:= str_comp_name_fact_pca;
 //FMLBitmapFileName:= 'MLFactorialPCA.bmp';
end;

function TMLGenCompFactPCA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactPCA;
end;

{ TMLCompFactPCA }

function TMLCompFactPCA.getClassOperator: TClassOperator;
begin
 result:= TOpFactPCA;
end;

{ TOpFactPCA }

procedure TOpFactPCA.BuildEigen;
begin
 GetEigensFromSymetricMatrix(FCorrMat,FDimAnalysis,FEigVal,FEgVec);
end;

procedure TOpFactPCA.SetProjections(atts: TLstAttributes; coefs: PMatrix; nbFactors: integer; prefix: string = 'PCA'; number: integer = 1);
var i,j,k: integer;
    newAtt: TAttContinue;
    //att: TAttribute;
    //stat: TCalcStatDesContinuous;
    attSize: integer;
    s: TTypeContinue;
    //new -- 18/03/2006 -- acclrer le temps de traitement
    //en crant des tableaux temporaires
    avg,std: array of double;
begin
 atts.Clear;
 attSize:= Dataset.LstAtts[asAll].Size;
 //new -- 18/03/2006 -- crer les tableaux temporaires, une fois pour toutes
 setlength(avg,succ(FDimAnalysis));
 for j:= 1 to FDimAnalysis do
  avg[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).Average;
 setlength(std,succ(FDimAnalysis));
 for j:= 1 to FDimAnalysis do
  //new -- 30/06/2011
  //if (abs(TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop) < EPSILON_STD_DEV_FOR_PCA)
  if (abs(TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop) < EPSILON_STD_DEV_FOR_PCA) OR ((self.PrmOp as TOpPrmFactPCA).Standardize = false)
   then std[j]:= 1.0
   //else std[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDev;
   //new -- 06/07/2006 -- on utilise l'cart-type de la population pour les calculs (cf. matrice de corrlation)
   else std[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop;
 //>> suite...
 //ok on peut lancer les calculs
 for k:= 1 to nbFactors do
  begin
   newAtt:= TAttContinue.Create(format('%s_%d_Axis_%d',[prefix,number,k]),attSize);
   //pour chaque individu i
   for i:= 1 to attSize do
    begin
     s:= 0.0;
     for j:= 1 to FDimAnalysis do
      begin
       //att:= Dataset.LstAtts[asInput].Attribute[pred(j)];
       //stat:= TCalcStatDesContinuous(FLstStats.Stat(pred(j)));
       //new -- pas de division par zro please
       (*
       if (abs(stat.StdDev) < EPSILON_STD_DEV_FOR_PCA)
        then s:= s + FEgVec^[j]^[k]*(att.cValue[i]-stat.Average)/1.0
        else s:= s + FEgVec^[j]^[k]*(att.cValue[i]-stat.Average)/stat.StdDev;
       *)
       //new -- 18/03/2006 -- a devrait amliorer (oui !!! division par 8 du temps de calcul...)
       s:= s + coefs^[j]^[k]*(Dataset.LstAtts[asInput].Attribute[pred(j)].cValue[i] - avg[j]) / std[j];
      end;
     newAtt.cValue[i]:= s;
    end;
   atts.Add(newAtt);
  end;
 //>>
 finalize(avg);
 finalize(std);
end;

procedure TOpFactPCA.BuildVCV;
var i: integer;
    vcvCalc: TEnumTypeCVCNormalisation;
begin
 //new -- 30/06/2011
 if ((self.PrmOp as TOpPrmFactPCA).Standardize = true) then vcvCalc:= vcvNormNormalized else vcvCalc:= vcvNormCentered;
 //FCorrMat:= BuildMatVCV(Dataset.Examples,Dataset.lstAtts[asInput],vcvNormNormalized,FLstStats);
 FCorrMat:= BuildMatVCV(Dataset.Examples,Dataset.lstAtts[asInput],vcvCalc,FLstStats);
 //...
 FTrace:= 0.0;
 for i:= 1 to self.FDimAnalysis do
  FTrace:= FTrace + FCorrMat^[i]^[i];
 //new - communality est 1.0 tout le temps si matrice de corrlation utilise
 if ((self.PrmOp as TOpPrmFactPCA).Standardize = true)
  then
   begin
    setLength(FPriorCommunality,FDimAnalysis+1);
    for i:= 1 to FDimAnalysis do FPriorCommunality[i]:= 1.0;
   end;
end;

(*
function TOpFactPCA.CoreExecute: boolean;
var tps: cardinal;
begin
 result:= TRUE;
 TRY
 //dtruire les matrices internes
 self.DelPrivateMatrix();
 //calculs
 tps:= GetTickCount();
 self.CalcStats();
 TraceLog.WriteToLogFile(format('[PCA] ok calc stat = %d ms.',[GetTickCount()-tps]));
 tps:= GetTickCount();
 self.BuildVCV();
 TraceLog.WriteToLogFile(format('[PCA] ok build vcv = %d ms.',[GetTickCount()-tps]));
 tps:= GetTickCount();
 self.BuildEigen();
 TraceLog.WriteToLogFile(format('[PCA] ok compute eigens = %d ms.',[GetTickCount()-tps]));
 //nombre d'axes cr = min de demand ou nombre de variables
 //FNbAxis:= MIN((self.PrmOp as TOpPrmFactPCA).NbAxisCreated,FDimAnalysis);
 //new -- 27/02/2011 -- attention, il faut grer le cas o NbAxisCreated < 0, il faut gnrer tous les axes
 if ((self.PrmOp as TOpPrmFactPCA).NbAxisCreated < 0) then FNbAxis:= FDimAnalysis else FNbAxis:= MIN((self.PrmOp as TOpPrmFactPCA).NbAxisCreated,FDimAnalysis);
 tps:= GetTickCount();
 self.CalcLoadings();
 TraceLog.WriteToLogFile(format('[PCA] ok compute loadings = %d ms.',[GetTickCount()-tps]));
 //new - 06/06/2012 - introduire test de Bartlett et indice KMO - si acp norme
 if ((self.PrmOp as TOpPrmFactPCA).Standardize = true)
  then
   begin
    tps:= GetTickCount();
    self.calcBartlettKMO();
    TraceLog.WriteToLogFile(format('[PCA] ok Bartlett and KMO index = %d ms.',[GetTickCount()-tps]));
   end;
 tps:= GetTickCount();
 self.SetProjections(self.GenAtts,FEgVec,FNbAxis,'PCA',(MLOwner as TMLComponent).Number);
 TraceLog.WriteToLogFile(format('[PCA] ok projection = %d ms.',[GetTickCount()-tps]));
 //new -- 04/02/2007 -- le cas cheant, calculer les contributions et les cos2  partir des projections
 if ((self.PrmOp as TOpPrmFactPCA).Cos2Ctr = true)
  then self.setCos2Ctr(self.GenAtts,FNbAxis,'PCA',(MLOwner as TMLComponent).Number);
 EXCEPT
 self.DelPrivateMatrix();
 result:= FALSE;
 END;
end;
*)

procedure TOpFactPCA.DelPrivateMatrix;
begin
 inherited;//new - 08/06/2012 - pour dtruire le tableau des vecteurs propres
 TRY
 //dtruire les matrices internes
 if assigned(FCorrMat) then delMatrix(FCorrMat,FDimAnalysis,FDimAnalysis);
 if assigned(FInvMatCorr) then delMatrix(FInvMatCorr,FDimAnalysis,FDimAnalysis);
 if assigned(FEgVec) then delMatrix(FEgVec,FDimAnalysis,FDimAnalysis);
 //if assigned(FEgVal) then delVector(FEgVal,FDimAnalysis);
 if assigned(FLoadings) then delMatrix(FLoadings,FDimAnalysis,FNbAxis);
 if assigned(FFactorScores) then delMatrix(FFactorScores,FDimAnalysis,FDimAnalysis);
 if assigned(FPriorCommunality) then Finalize(FPriorCommunality);
 FLstStats.FreeAll;
 //new - 06/06/2012 - pour l'indice Kmo
 if (self.FTabKmo <> nil) then Finalize(FTabKmo);
 if (FPartialR <> nil) then DelMatrix(FPartialR,1,FDimAnalysis);
 EXCEPT
 END;
end;

function TOpFactPCA.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactPCA;
end;

function TOpFactPCA.getHTMLResultsSummary: string;
var s: string;
    buf: TBufString;
begin
 //buffriser l'ajout -- 23/06/2004
 buf:= TBufString.Create();
 buf.BeginUpdate();
 //#ToDo1 - affichage des rsultats cf. SPAD et STATISTICA
 //afficher les stats des.
 //afficher la matrice de corrlation
 //afficher les valeurs propres
 //la trace de la matrice

 //tableau des valeurs propres
 s:= '<P><H3>Eigen values</H3>';
 buf.AddStr(s+self.getHTMLEigenValues());

 //new - 06/06/2012 - tester les valeurs propres
 if ((self.PrmOp as TOpPrmFactPCA).Standardize = true)
  then buf.AddStr(self.getHTMLSignificance());

 //new - 07/06/2012 - statistiques optionnelles, Bartlett et Kmo (uniquement pour ACP norme)
 if ((self.PrmOp as TOpPrmFactPCA).BartlettKmo = true) and ((self.PrmOp as TOpPrmFactPCA).Standardize = true)
  then
   begin
    s:= '<BR><BR><H3>Optional Statistics</H3>';
    buf.AddStr(s+self.getHTMLBartlett());
    buf.AddStr(self.getHTMLKmo());
   end;

 //**
 //afficher la corrlation des attributs avec les axes
 //**
 s:= '<BR><BR><H3>Factor Loadings [Communality Estimates]</H3>';
 buf.AddStr(s+self.getHTMLLoadings(FLoadings,FNbAxis,(self.PrmOp as TOpPrmFactPCA).SortVarOnLoadings,nil));

 //**
 //new -- 13/02/2005 -- afficher les vecteurs propres
 //**
 s:= '<BR><BR><H3>Factor Score Coefficients</H3>';
 buf.AddStr(s+self.getHTMLRegCoef(FFactorScores,FLoadings,FNbAxis));//buf.AddStr(s+self.getHTMLRegCoef(FEgVec,FNbAxis));

 //new - 07/06/2012 - afficher les matrices de calcul (corrlation ou covariance)
 if ((self.PrmOp as TOpPrmFactPCA).ShowCorr = true)
  then
   begin
    s:= '<BR><BR><H3>Matrices</H3>';
    buf.AddStr(s+self.getHTMLCorrMatrices());
   end;

 //new - 13/09/2012 - matrice des corrlations reproduites et rsiduelles
 if ((self.PrmOp as TOpPrmFactPCA).Standardize = true) and ((self.PrmOp as TOpPrmFactPCA).ShowRepCorr = true)
  then
   begin
    s:= '<BR><BR><H3>Original, reproduced and residual correlations</H3>';
    buf.AddStr(s+self.getHTMLRepResCorrMatrice(self.FLoadings,self.FNbAxis,(self.PrmOp as TOpPrmFactPCA).SortVarOnLoadings));
   end;

 //envoyer la sauce
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

constructor TOpFactPCA.Create(AOwner: TObject);
begin
 inherited Create(AOwner);
 FLstStats:= TLstCalcStatDesContinuous.Create(NIL,NIL);
end;

procedure TOpFactPCA.CalcStats;
begin
 FDimAnalysis:= Dataset.LstAtts[asInput].Count;
 FLstStats.RebuildStatDes(Dataset.LstAtts[asInput],Dataset.Examples);
end;

destructor TOpFactPCA.Destroy;
begin
 inherited Destroy;
 FLstStats.Free;
end;

function TOpFactPCA.CheckAttributes: boolean;
begin
 result:= (self.Dataset.LstAtts[asInput].Count>0) and (self.Dataset.LstAtts[asInput].isAllCategory(caContinue));
end;

procedure TOpFactPCA.CalcLoadings;
var i,j: integer;
    value: double;
begin
 dimMatrix(FLoadings,FDimAnalysis,FNbAxis);
 for i:= 1 to FDimAnalysis do
  begin
   for j:= 1 to FNbAxis do
    begin
      value:= sqrt(EigVal[j])*FEgVec^[i]^[j];
      FLoadings[i,j]:= value;
    end;
  end;
end;

(*
function TOpFactPCA.getHTMLLoadings(loadings: PMatrix; nbFactors: integer): string;
var buf: TBufString;
    i,j: integer;
    sum,value: double;
    vExpl: array of double;
    strColor: string;
    index: TSortedIndexVariables;
begin

 index:= self.sortOnLoadings(loadings,nbFactors);

 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Attribute</TH>');
 setLength(vExpl,succ(nbFactors));
 for j:= 1 to nbFactors do
  begin
   buf.AddStr(format('<TH align=center colspan=2>Axis_%d</TH>',[j]));
   //en profiter pour calculer la variance explique
   vExpl[j]:= 0.0;
   for i:= 1 to FDimAnalysis do
    vExpl[j]:= vExpl[j] + loadings[i,j] * loadings[i,j];
  end;
 buf.AddStr('</TR>');
 
 //rajouter la seconde ligne d'en-tte
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>-</TH>');
 for j:= 1 to nbFactors do
  buf.AddStr('<TD align=center width=60>Corr.</TD><TD align=center width=60>% (Tot. %)</TD>');
 //les valeurs
 for i:= 1 to FDimAnalysis do
  begin
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[Dataset.LstAtts[asInput].Attribute[pred(i)].Name]));
   //new -- 06/07/2006 -- calcul des communalities
   sum:= 0.0;
   for j:= 1 to nbFactors do
    begin
      TRY
      value:= loadings[i,j];
      //new -- 30/06/2011
      if ((self.PrmOp as TOpPrmFactPCA).Standardize = false) AND (abs(TCalcStatDesContinuous(FLstStats.Stat(pred(i))).StdDevPop) >= EPSILON_STD_DEV_FOR_PCA)
       then value:= value / TCalcStatDesContinuous(FLstStats.Stat(pred(i))).StdDevPop;
      //
      sum:= sum + value * value;
      //mettre en valeur si % de la variable > % de l'axe
      strColor:= codeCouleurHTMLLoadings(value);
      //la ligne  afficher
      buf.AddStr(format('<TD align=right width=60 %s>%.5f</TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[strColor,value,100.0*value*value,'%',100.0*sum,'%']));
      EXCEPT
      END;
    end;
   buf.AddStr('</TR>');
  end;
 //pourcentage expliqu par l'axe
 buf.AddStr(HTML_TABLE_COLOR_DATA_BLUE+format('<TH>%s</TH>',['Var. Expl.']));
 sum:= 0.0;
 for j:= 1 to nbFactors do
  begin
   sum:= sum + vExpl[j];
   buf.AddStr(format('<TD align=right width=60>%.5f</TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[vExpl[j],100.0*vExpl[j]/FTrace,'%',100.0*sum/FTrace,'%']));
  end;
 buf.AddStr('</TR>');
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
 Finalize(vExpl);
 Finalize(index);
end;
*)

//new - 14/09/2012 - introduire une option de tri selon les loadings
function TOpFactPCA.getHTMLLoadings(loadings: PMatrix; nbFactors: integer; sorted: boolean; loadingsForSorting: PMatrix = nil): string;
var buf: TBufString;
    i,j: integer;
    sum,value: double;
    vExpl: array of double;
    strColor: string;
    index: TPCAIndexVariables;
begin

 if (loadingsForSorting = nil)
  then index:= self.getIndexVariables(sorted,loadings,nbFactors)
  else index:= self.getIndexVariables(sorted,loadingsForSorting,nbFactors);

 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Attribute</TH>');
 
 setLength(vExpl,succ(nbFactors));
 for j:= 1 to nbFactors do
  begin
   buf.AddStr(format('<TH align=center colspan=2>Axis_%d</TH>',[j]));
   //en profiter pour calculer la variance explique
   vExpl[j]:= 0.0;
   for i:= 1 to FDimAnalysis do
    vExpl[j]:= vExpl[j] + loadings[i,j] * loadings[i,j];
  end;
 buf.AddStr('</TR>');
 
 //rajouter la seconde ligne d'en-tte
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>-</TH>');
 for j:= 1 to nbFactors do
  buf.AddStr('<TD align=center width=60>Corr.</TD><TD align=center width=60>% (Tot. %)</TD>');
 //les valeurs
 for i:= 1 to FDimAnalysis do
  begin
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[Dataset.LstAtts[asInput].Attribute[pred(index[i])].Name]));
   //new -- 06/07/2006 -- calcul des communalities
   sum:= 0.0;
   for j:= 1 to nbFactors do
    begin
      TRY
      value:= loadings[index[i],j];
      //new -- 30/06/2011
      if ((self.PrmOp as TOpPrmFactPCA).Standardize = false) AND (abs(TCalcStatDesContinuous(FLstStats.Stat(pred(index[i]))).StdDevPop) >= EPSILON_STD_DEV_FOR_PCA)
       then value:= value / TCalcStatDesContinuous(FLstStats.Stat(pred(index[i]))).StdDevPop;
      //
      sum:= sum + value * value;
      //mettre en valeur si % de la variable > % de l'axe
      strColor:= codeCouleurHTMLLoadings(value);
      //la ligne  afficher
      buf.AddStr(format('<TD align=right width=60 %s>%.5f</TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[strColor,value,100.0*value*value,'%',100.0*sum,'%']));
      EXCEPT
      END;
    end;
   buf.AddStr('</TR>');
  end;
 //pourcentage expliqu par l'axe
 buf.AddStr(HTML_TABLE_COLOR_DATA_BLUE+format('<TH>%s</TH>',['Var. Expl.']));
 sum:= 0.0;
 for j:= 1 to nbFactors do
  begin
   sum:= sum + vExpl[j];
   buf.AddStr(format('<TD align=right width=60>%.5f</TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[vExpl[j],100.0*vExpl[j]/FTrace,'%',100.0*sum/FTrace,'%']));
  end;
 buf.AddStr('</TR>');
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();

 Finalize(vExpl);
 Finalize(index);
end;

function TOpFactPCA.getHTMLRegCoef(factorCoefs: PMatrix; loadings: PMatrix; 
  nbFactors: integer): string;
var  buf: TBufString;
     s: string;
     i,j: integer;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 //en-tte pour chaque axe factoriel demand
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>Attribute</TH>');
 //new -- 06/07/2006 -- ajouter la moyenne et l'cart-type
 s:= format('<TH width=80 align=right %s>Mean</TH><TH width=80 align=right %s>Std-dev</TH>',[HTML_BGCOLOR_HEADER_GREEN,HTML_BGCOLOR_HEADER_GREEN]);
 for j:= 1 to nbFactors do
  s:= s+format('<TH width=80 align=right>Axis_%d</TH>',[j]);
 s:= s+'</TR>';
 //le contenu
 for i:= 1 to FDimAnalysis do
  begin
   s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[Dataset.LstAtts[asInput].Attribute[pred(i)].Name]);
   //new -- 06/07/2006 -- ajouter la moyenne et l'cart-type
   s:= s + format('<TD align=right %s>%.7f</TD>',[HTML_BGCOLOR_DATA_GREEN,TCalcStatDesContinuous(FLstStats.Stat(pred(i))).Average]);
   //s:= s + format('<TD align=right %s>%.6f</TD>',[HTML_BGCOLOR_DATA_GREEN,TCalcStatDesContinuous(FLstStats.Stat(pred(i))).StdDevPop]);
   //new -- 30/06/2011
   if ((self.PrmOp as TOpPrmFactPCA).Standardize = false)
    then s:= s + format('<TD align=right %s>%.1f</TD>',[HTML_BGCOLOR_DATA_GREEN,1.0])
    else s:= s + format('<TD align=right %s>%.7f</TD>',[HTML_BGCOLOR_DATA_GREEN,TCalcStatDesContinuous(FLstStats.Stat(pred(i))).StdDevPop]);
   //suite...
   for j:= 1 to nbFactors do
    begin
      TRY
      s:= s+format('<TD align=right>%.7f</TD>',[factorCoefs^[i]^[j]]);
      EXCEPT
      END;
    end;
   s:= s+'</TR>';
  end;
 s:= s+'</table>';
 //dans le buffer...
 buf.AddStr(s);
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

procedure TOpFactPCA.setCos2Ctr(atts: TLstAttributes; nbFactors: integer;
  prefix: string; number: integer);
var lst: TObjectList;
    newAtt: TAttribute;
    factor: TAttribute;
    attSize: integer;
    nbInstances: integer;
    k,i,j: integer;
    poids,d: single;
    avg, std: array of single;
    dist: array of single;
begin
 //new -- 07/02/2012 -- il faut prendre le nombre d'instances dans la slection courante (les individus slectionns) !!!
 //et non pas le nombre de ligne dans la base entire cf. e-mail Laffond du 07/02/2012
 nbInstances:= Dataset.Examples.Size;
 //new -- 07/02/2012 -- poids doit tre bas sur les individus slectionns
 //poids:= 100.0/(1.0*attSize);//poids en % -- erreur si seule une partie de la base est slectionne
 poids:= 100.0/(1.0 * nbInstances);
 //mais on a besoin quand mme de crer la nouvelle variable sur toute la base
 attSize:= Dataset.LstAtts[asAll].Size;
 //suite...
 lst:= TObjectList.Create(false);//non propritaire
 //calcul des CTR et COS  partir des atts... (coordonnes sur les axes factoriels)
 //**************
 //calcul des CTR
 //**************
 for k:= 1 to nbFactors do
  begin
   factor:= atts.Attribute[pred(k)];
   newAtt:= TAttContinue.Create(format('%s_%d_CTR_%d',[prefix,number,k]),attSize);
   lst.Add(newAtt);
   //calcul Saporta --> pp.175, CTR en % puisque poids en %
   if (EigVal[k] > 0.0)
    then
     begin
      for i:= 1 to attSize do newAtt.cValue[i]:= poids * MATH.Power(factor.cValue[i],2.0) / EigVal[k];
     end
    else for i:= 1 to attSize do newAtt.cValue[i]:= 0.0;
  end;
 //***************
 //calcul des COS
 //***************
 //crer les tableaux temporaires, une fois pour toutes
 setlength(avg,succ(FDimAnalysis));
 for j:= 1 to FDimAnalysis do avg[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).Average;
 setlength(std,succ(FDimAnalysis));
 for j:= 1 to FDimAnalysis do
  //new -- 30/06/2011
  //if (abs(TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop) < EPSILON_STD_DEV_FOR_PCA)
  if (abs(TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop) < EPSILON_STD_DEV_FOR_PCA) OR ((self.PrmOp as TOpPrmFactPCA).Standardize = false)
   then std[j]:= 1.0
   else std[j]:= TCalcStatDesContinuous(FLstStats.Stat(pred(j))).StdDevPop;//new -- 06/07/2006 -- on utilise l'cart-type de la population pour les calculs (cf. matrice de corrlation)
 //distance au centre de gravit des individus
 setLength(dist,succ(attSize));
 for i:= 1 to attSize do
  begin
   d:= 0.0;
   for j:= 1 to FDimAnalysis do
    d:= d + MATH.POWER((Dataset.LstAtts[asInput].Attribute[pred(j)].cValue[i] - avg[j]) / std[j], 2.0);
   dist[i]:= d;
  end;
 //fin des tableaux temporaires avg, std
 finalize(avg);
 finalize(std);
 //calculer le COS2
 for k:= 1 to nbFactors do
  begin
   factor:= atts.Attribute[pred(k)];
   newAtt:= TAttContinue.Create(format('%s_%d_COS2_%d',[prefix,number,k]),attSize);
   lst.Add(newAtt);
   for i:= 1 to attSize do
    begin
     if (dist[i] > 0.0)
      then newAtt.cValue[i]:= MATH.Power(factor.cValue[i],2.0)/dist[i]
      else newAtt.cValue[i]:= 0.0;
    end;
  end;
 finalize(dist);
 //flush...
 for j:= 0 to pred(lst.Count) do
  atts.Add(lst[j] as TAttribute);
 //>>
 lst.Free();
end;

//tableau des couleurs pour la significativit
const
    COLOR_SIG_EIGVAL : array[0..3] of string = (HTML_BGCOLOR_DATA_GRAY,HTML_BGCOLOR_LIGHT_RED,HTML_BGCOLOR_DATA_RED,HTML_BGCOLOR_HEAVY_RED);

function TOpFactPCA.getHTMLSignificance: string;
var s: string;
    i: integer;
    kaiser, karlis, p, n: double;
    nbHits: array of integer;
    bs: array of double;
    okBs: boolean;
begin
 s:= '<BR><BR><H3>Significance of Principal Components</H3>';
 //seuils simples, kaiser et karlis
 kaiser:= 1.0;
 p:= 1.0*self.FDimAnalysis;
 n:= 1.0*Dataset.Examples.Size;
 if (n > 1) then karlis:= 1.0+2.0*SQRT((p-1.0)/(n-1.0)) else karlis:= 0.0;
 //seuil broken-stick
 setLength(bs,self.DimAnalysis+1);
 bs[FDimAnalysis]:= 1.0/(1.0*FDimAnalysis);
 for i:= FDimAnalysis-1 downto 1 do
  bs[i]:= (1.0/(1.0*i))+bs[i+1];
 //compteur de hit
 setLength(nbHits,self.FDimAnalysis+1);
 okBs:= true;
 for i:= 1 to FDimAnalysis do
  begin
   if (EigVal[i] > kaiser) then inc(nbHits[i]);
   if (EigVal[i] > karlis) then inc(nbHits[i]);
   okBs:= okBs and (EigVal[i] > bs[i]);//06/09/2012 - comme a, invalid ds qu'il y a une inversion
   if (okBs = true) then inc(nbHits[i]);
  end;
 //affichage des seuils globaux
 s:= s + '<br>'+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Global critical values</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + format('<TD>%s</TD><TD align=center>%.6g</TD></TR>',['Kaiser-Guttman',kaiser]);
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + format('<TD>%s</TD><TD align=center>%.6g</TD></TR>',['Karlis-Saporta-Spinaki',karlis]);
 s:= s + '</table>';

 s:= s + '<br><H5>Eigenvalue table - Test for significance</H5>';

 //tableau
 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=3>Eigenvalues - Significance</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Axis</TH><TH width=80>Eigenvalue</TH><TH width=80>Broken-stick<br>critical values</TH></TR>';
 for i:= 1 to self.FDimAnalysis do
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align=center>%d</TD><TD %s align="center">%.6f</TD><TD align="center">%.6f</TD></TR>',[i,COLOR_SIG_EIGVAL[nbHits[i]],EigVal[i],bs[i]]);
 s:= s + '</table>';
 
 //librer
 finalize(bs);
 finalize(nbHits);

 //
 result:= s;
end;

function TOpFactPCA.getHTMLEigenValues: string;
var s: string;
    sum: double;
    tot: double;
    i: integer;
    buf: TBufString;
    diff: string;
begin
 //buffriser l'ajout -- 23/06/2004
 buf:= TBufString.Create();
 buf.BeginUpdate();

 s:= HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY + '<TD width=100><b>Matrix trace</b></TD>'+format('<TD align=right width=100><b>%.6f</b></TD></TR>',[FTrace]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TD><b>Average</b></TD>'+format('<TD align=right><b>%.6f</b></TD></TR>',[FTrace/(1.0*FDimAnalysis)]);
 s:= s + '</table><br><br>';
 //s:= format('<H5>Matrix trace = %.6f</H5>',[FTrace]);
 //s:= s + format('<h5>Average = %.6f</h5>',[FTrace/FDimAnalysis]);
 s:= s+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH>Axis</TH><TH width=80>Eigen value</TH><TH width=80>Difference</TH><TH width=80>Proportion (%)</TH><TH width="200">Histogram</TH><TH width=80>Cumulative (%)</TH></TR>';
 buf.AddStr(s);
 //tot:= 1.0*FDimAnalysis;
 //new -- 30/06/2011 -- inclure l'acp non norme
 tot:= 1.0 * FTrace;
 //
 sum:= 0.0;
 for i:= 1 to FDimAnalysis do
  begin
   sum:= sum+EigVal[i];
   if (tot > 0)
    then
     begin
      if (i < FDimAnalysis) then diff:= format('<TD align=right>%.6f</TD>',[EigVal[i]-EigVal[i+1]]) else diff:= '<TD align=center>-</TD>'; 
      s:= HTML_TABLE_COLOR_DATA_GRAY+
         format('<TH>%d</TH><TD align=right>%.6f</TD>%s<TD align=right>%.2f %s</TD><TD>%s</TD><TD align=right>%.2f %s</TD></TR>',
                [i,EigVal[i],diff,100.0*EigVal[i]/tot,'%',getHtmlHistogram(TRUNC(100.0*EigVal[i]/tot)),100.0*sum/tot,'%'])+'</TR>';
      buf.AddStr(s);
     end;
  end;
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY + format('<TH>Tot.</TH><TD align=right>%.6f</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD></TR>',[sum]));
 buf.AddStr('</table>');

 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TOpFactPCA.getNewForm: TForm;
begin
 result:= TFrmOpViewFactPCA.CreateFromOperator(self);
end;

procedure TOpFactPCA.calcBartlettKMO;
begin
 self.calcBartlett();
 //voyons maintenant s'il est possible de calculer l'indice KMO
 FKmo:= 0.0;
 if (self.FDetR > SEUIL_DET_MAT_CORR)
  then self.calcKMO();
end;

procedure TOpFactPCA.calcBartlett;
var i: integer;
    p,n: double;
begin
 //calcul du dterminant
 FDetR:= 1.0;
 i:= 1;
 while (FDetR > Math.MinSingle) and (i <= FDimAnalysis) and (EigVal[i] > 0.0) do
  begin
   FDetR:= FDetR * EigVal[i];
   inc(i);
  end;
 if (i < FDimAnalysis) or (FDetR < Math.MinSingle) then FDetR:= 0.0;
 //calcul de la statistique de Bartlett
 FBartlett:= 0.0;
 p:= 1.0*self.Dataset.LstAtts[asInput].Count;
 FDf:= (p*(p-1.0))/2.0;
 FPValue:= 1.0;
 if (FDetR > 0.0)
  then
   begin
    n:= 1.0*self.Dataset.Examples.Size;
    FBartlett:= -(n - (2 * p + 11.0)/6.0) * ln(FDetR);
    FPValue:= PKhi2(trunc(FDf),FBartlett);
   end;
end;

procedure TOpFactPCA.calcKMO;
var i,j: integer;
    value,num,denom: double;
begin
 DimMatrix(FPartialR,FDimAnalysis,FDimAnalysis);
 //inversion de la matrice de corrlation
 //InvMat(MatCorr,1,FDimAnalysis,FPartialR);
 //new - 07/09/2012 - on passe par une matrice intermdiaire inverse des corrlations
 if (assigned(FInvMatCorr) = false)
  then
   begin
    DimMatrix(FInvMatCorr,FDimAnalysis,FDimAnalysis);
    InvMat(MatCorr,1,FDimAnalysis,FInvMatCorr);
   end;
 //du coup on ne fait que copier la matrice inverse
 CopyMatrix(FPartialR,FInvMatCorr,1,1,FDimAnalysis,FDimAnalysis);
 //suite...
 //calculer la matrice des corrlations partielles
 for i:= 1 to self.FDimAnalysis-1 do
  for j:= i+1 to self.FDimAnalysis do
   begin
    value:= FPartialR^[i]^[i] * FPartialR^[j]^[j];
    if (value > 0.0)
     then FPartialR^[i]^[j]:= -1.0 * FPartialR^[i]^[j] / sqrt(value)
     else FPartialR^[i]^[j]:= 0.0;
    FPartialR^[j]^[i]:= FPartialR^[i]^[j];
   end;
 //ne pas oublier la diagonale
 for i:= 1 to FDimAnalysis do FPartialR^[i]^[i]:= 1.0;
 //indice KMO
 FKmo:= 0.0;
 num:= 0.0;
 denom:= 0.0;
 for i:= 1 to FDimAnalysis do
  for j:= 1 to FDimAnalysis do
   begin
    if (j <> i)
     then
      begin
       num:= num + SQR(MatCorr^[i]^[j]);
       denom:= denom + SQR(MatCorr^[i]^[j]) + SQR(FPartialR^[i]^[j]);
      end;
   end;
 if (denom > 0.0)
  then FKmo:= num / denom;
 //indice KMO par variable
 setlength(FTabKmo,FDimAnalysis+1);
 for j:= 1 to FDimAnalysis do
  begin
   FTabKmo[j]:= 0.0;
   num:= 0.0;
   denom:= 0.0;
   for i:= 1 to FDimAnalysis do
    if (i <> j)
     then
      begin
       num:= num + SQR(MatCorr^[i]^[j]);
       denom:= denom + SQR(MatCorr^[i]^[j]) + SQR(FPartialR^[i]^[j]);
      end;
    if (denom > 0.0)
     then FTabKmo[j]:= num / denom;
  end;
end;

function colorCodeCorrelation(corr: double): string;
begin
 result:= codeCouleurHTMLLoadings(corr); 
end;

function TOpFactPCA.getHTMLCorrMatrices: string;
var buf: TBufString;
    s: string;
    i,j: integer;
    index: TPCAIndexVariables;
    value: double;
    strColor: string;
begin

 index:= self.getIndexVariables((self.PrmOp as TOpPrmFactPCA).SortVarOnLoadings,self.Loadings,self.FNbAxis);

 buf:= TBufString.Create();
 buf.BeginUpdate();
 if ((self.PrmOp as TOpPrmFactPCA).Standardize = true)
  then s:= '<br><h5>Correlations</h5>'
  else s:= '<br><h5>Covariances</h5>';
 buf.AddStr(s);
 s:= '<br>'+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_DATA_GRAY+'<TH></TH>';
 for j:= 1 to FDimAnalysis do
  s:= s + '<TH width=60>'+self.Dataset.LstAtts[asInput].Attribute[index[j]-1].Name+'</TH>';
 buf.AddStr(s+'</TR>');
 for i:= 1 to FDimAnalysis do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY+'<TH>'+self.Dataset.LstAtts[asInput].Attribute[index[i]-1].Name+'</TH>';
   for j:= 1 to FDimAnalysis do
    begin
     value:= FCorrMat^[index[i]]^[index[j]];
     strColor:= colorCodeCorrelation(value);
     s:= s + format('<TD align=center %s>%.5f</TD>',[strColor,value]);
    end;
   buf.AddStr(s+'</TR>');
  end;
 buf.AddStr('</table>');

 //est-ce qu'il y a lieu d'afficher la matrice des corrlations partielles
 if ((self.PrmOp as TOpPrmFactPCA).Standardize = true) and (FPartialR <> nil)
  then
   begin
    s:= '<br><br><h5>Partial Correlations Controlling all other Variables</h5>';
    buf.AddStr(s);
    s:= '<br>'+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_DATA_GRAY+'<TH></TH>';
    for j:= 1 to FDimAnalysis do
     s:= s + '<TH width=60>'+self.Dataset.LstAtts[asInput].Attribute[index[j]-1].Name+'</TH>';
    buf.AddStr(s+'</TR>');
    for i:= 1 to FDimAnalysis do
     begin
      s:= HTML_TABLE_COLOR_DATA_GRAY+'<TH>'+self.Dataset.LstAtts[asInput].Attribute[index[i]-1].Name+'</TH>';
      for j:= 1 to FDimAnalysis do
       begin
        value:= FPartialR^[index[i]]^[index[j]];
        strColor:= colorCodeCorrelation(value);
        s:= s + format('<TD align=center %s>%.5f</TD>',[strColor,value]);
       end;
      buf.AddStr(s+'</TR>');
     end;
    buf.AddStr('</table>');
   end;
 //
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();

 Finalize(index);

end;

function TOpFactPCA.getDimEigVal: integer;
begin
 result:= self.FDimAnalysis;
end;

function TOpFactPCA.calcAdditional: boolean;
//var tps: cardinal;
begin
 result:= true;
 TRY
 //tps:= GetTickCount();
 self.CalcLoadings();
 //new - 07/09/2012 - calculer les coefficients de projection
 self.CalcFactorScores();
 //TraceLog.WriteToLogFile(format('[PCA] ok compute loadings = %d ms.',[GetTickCount()-tps]));
 //new - 06/06/2012 - introduire test de Bartlett et indice KMO - si acp norme
 if ((self.PrmOp as TOpPrmFactPCA).Standardize = true)
  then
   begin
    //tps:= GetTickCount();
    self.calcBartlettKMO();
    //TraceLog.WriteToLogFile(format('[PCA] ok Bartlett and KMO index = %d ms.',[GetTickCount()-tps]));
   end;
 EXCEPT
 result:= false;
 END;
end;

function TOpFactPCA.calcEigen: boolean;
//var tps: cardinal;
begin
 result:= TRUE;
 TRY
 //dtruire les matrices internes
 self.DelPrivateMatrix();
 //calculs
 //tps:= GetTickCount();
 self.CalcStats();
 //nombre d'axes cr = min de demand ou nombre de variables
 //FNbAxis:= MIN((self.PrmOp as TOpPrmFactPCA).NbAxisCreated,FDimAnalysis);
 //new -- 27/02/2011 -- attention, il faut grer le cas o NbAxisCreated < 0, il faut gnrer tous les axes
 if ((self.PrmOp as TOpPrmFactPCA).NbAxisCreated < 0) then FNbAxis:= FDimAnalysis else FNbAxis:= MATH.MIN((self.PrmOp as TOpPrmFactPCA).NbAxisCreated,FDimAnalysis);
 //TraceLog.WriteToLogFile(format('[PCA] ok calc stat = %d ms.',[GetTickCount()-tps]));
 //tps:= GetTickCount();
 self.BuildVCV();
 //TraceLog.WriteToLogFile(format('[PCA] ok build vcv = %d ms.',[GetTickCount()-tps]));
 //tps:= GetTickCount();
 self.BuildEigen();
 //TraceLog.WriteToLogFile(format('[PCA] ok compute eigens = %d ms.',[GetTickCount()-tps]));
 EXCEPT
 result:= false;
 END;
end;

function TOpFactPCA.calcProjections: boolean;
var tps: cardinal;
begin
 result:= true;
 TRY
 tps:= GetTickCount();
 //self.SetProjections(self.GenAtts,FEgVec,FNbAxis,'PCA',(MLOwner as TMLComponent).Number);
 //new -- 07/09/2012 -- utiliser la variable FactorScores
 //self.SetProjections(self.GenAtts,FFactorScores,FNbAxis,'PCA',(MLOwner as TMLComponent).Number);
 self.SetProjections(self.GenAtts,FFactorScores,FNbAxis,self.getPrexifNameForProjection(),(MLOwner as TMLComponent).Number);
 TraceLog.WriteToLogFile(format('[PCA] ok projection = %d ms.',[GetTickCount()-tps]));
 //new -- 04/02/2007 -- le cas cheant, calculer les contributions et les cos2  partir des projections
 if ((self.PrmOp as TOpPrmFactPCA).Cos2Ctr = true)
  then self.setCos2Ctr(self.GenAtts,FNbAxis,'PCA',(MLOwner as TMLComponent).Number);
 EXCEPT
 result:= FALSE;
 END;
end;

constructor TOpFactPCA.CreateWithDataset(AOwner: TObject;
  prmDataset: TMLDataset);
begin
 inherited CreateWithDataset(AOwner,prmDataset);
 FLstStats:= TLstCalcStatDesContinuous.Create(NIL,NIL);
end;

function TOpFactPCA.getInstanceFA: TClassOpFactAnalysis;
begin
 result:= TOpFactPCA;
end;

function TOpFactPCA.getHTMLBartlett: string;
var s: string;
begin
 if (self.FDetR > 0.0)
  then
   begin
    s:= '<br><H5>Bartlett''s test of sphericity</h5>';
    s:= s + '<br>'+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Bartlett''s test</TH></TR>';
    s:= s + HTML_TABLE_COLOR_DATA_GRAY+format('<TD>|CORR.MATRIX|</TD><TD align=right>%.7g</TD></TR>',[FDetR]);
    s:= s + HTML_TABLE_COLOR_DATA_GRAY+format('<TD>CHISQ</TD><TD align=right>%.7g</TD></TR>',[FBartlett]);
    s:= s + HTML_TABLE_COLOR_DATA_GRAY+format('<TD>d.f.</TD><TD align=right>%.0f</TD></TR>',[FDf]);
    s:= s + HTML_TABLE_COLOR_DATA_GRAY+format('<TD>p-value</TD><TD align=right>%.7g</TD></TR>',[FPValue]);
    s:= s + '</table>';
    result:= s;
   end
  else result:= format('|CORR.MATRIX| < %.2e, Bartlett''s Test of Sphericity cannot be computed<br>',[SEUIL_DET_MAT_CORR]);
end;


function TOpFactPCA.getHTMLKmo: string;
var s: string;
    i: integer;
    nbCol: integer;
begin
 if (self.FDetR > 0.0)
  then
   begin
    s:= '<br><br><H5>Kaiser''s Measure of Sampling Adequacy (MSA)</h5>';
    nbCol:= min(5,self.FDimAnalysis);
    s:= s + '<br>'+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_DATA_GREEN+format('<TH colspan=%d>Overall MSA = %.7g</TH></TR>',[2*nbCol,FKmo]);
    s:= s + HTML_TABLE_COLOR_DATA_GRAY;
    for i:= 1 to FDimAnalysis do
     begin
      //passage  la ligne ?
      if (i mod nbCol = 1)
       then
        begin
         s:= s + '</TR>';
         if (i < FDimAnalysis) then s:= s + HTML_TABLE_COLOR_DATA_GRAY;
        end;
      //remplir
      s:= s + format('<TH align=center width=80 %s>%s</TH><TD align=center width=80>%.7g</TD>',[HTML_BGCOLOR_HEADER_GRAY,Dataset.lstAtts[asInput].Attribute[i-1].Name,FTabKmo[i]]);
     end;
    s:= s + '</table>';
    result:= s;
   end
  else result:= format('|CORR.MATRIX| < %.2e, Kaiser-Mayer-Olkin (KMO) measure of sampling adequacy cannot be computed',[SEUIL_DET_MAT_CORR]);
end;

procedure TOpFactPCA.CalcFactorScores;
begin
 DimMatrix(FFactorScores,FDimAnalysis,FDimAnalysis);
 CopyMatrix(FFactorScores,FEgVec,1,1,FDimAnalysis,FDimAnalysis);
end;

function TOpFactPCA.getRotatedCoefs(newLoadings, newEgVec: PMatrix;
  nbFactors: integer): PMatrix;
var ok: boolean;
    i,j: integer;
    sum: double;
    fScores: PMatrix;
begin
 fScores:= nil;
 ok:= true;
 //matrice de corrlation interne indisponible ?
 if (assigned(FInvMatCorr) = false)
  then
   begin
    Matrices.DimMatrix(FInvMatCorr,self.DimAnalysis,self.DimAnalysis);
    //new -- 22/11/2008 -- l'inverse SVD n'est pas prcise, surttout quand la matrice est inversible
    //modifier la procdure --> tenter d'abord l'inversion normale...
    TRY
     ok:= (Matrices.InvMat(self.MatCorr,1,self.DimAnalysis,FInvMatCorr) = MAT_OK);
    EXCEPT
     ok:= FALSE;
    END;
    //si plantage, on passe  la SVDInverse...
    if not(ok)
     then
      begin
       //ok:= (Matrices.InvMat(FOpACP.MatCorr,1,FOpACP.DimAnalysis,invCorr) = Matrices.MAT_OK);
       Matrices.CopyMatrix(FInvMatCorr,self.MatCorr,1,1,self.DimAnalysis,self.DimAnalysis);
       ok:= UCalcMatrixAdditionalFunctions.SVDInverse(FInvMatCorr,self.DimAnalysis);
      end;
   end;
 //suite normale...
 if (ok = true)
  then
   begin
    try
    //produit matriciel inverse de la matrice de corrlation avec les nouveaux loadings
    UCalcMatrixAdditionalFunctions.MatrixProd(FInvMatCorr,DimAnalysis,DimAnalysis,newLoadings,DimAnalysis,nbFactors,newEgVec);

    //new -- 09/02/2012
    //multiplier par la racine du vecteur propre de chaque facteur pour que variance.pop du facteur = somme des carrs des corrlations des variables avec le facteur
    //ainsi la variance de la colonne projete sera bien gale  la valeur propre (et la somme des carrs des corrlations des variables avec le facteur)
    //cf. e-mail Gilbert Laffond du 09/02/2012

    DimMatrix(fScores,FDimAnalysis,nbFactors);

    for j:= 1 to nbFactors do
     begin
      //calculer la somme des carrs des loadings = valeur propre de l'axe
      sum:= 0.0;
      for i:= 1 to DimAnalysis do
       sum:= sum + newLoadings[i,j] * newLoadings[i,j];
      //mettre  la racine carre
      if (sum > 0.0) then sum:= sqrt(sum) else sum:= 0.0;
      //normaliser en multipliant la racine carre de la valeur propre associe  l'axe
      for i:= 1 to DimAnalysis do
       fScores[i,j]:= newEgVec[i,j] * sum;
     end;

    except
    ok:= false;
    end;
   end;

 if not(ok) and (fScores<>nil) then DelMatrix(fScores,FDimAnalysis,nbFactors);

 result:= fScores;
end;

function colorCodeCorrResidual(orig,res: double): string;
var s: string;
begin
 s:= HTML_BGCOLOR_DATA_GRAY;
 if (abs(orig) >= SEUIL_COR_HIGHLIGTH_LOW) AND (abs(res) <= SEUIL_RESIDUAL_COR_HIGHLIGTH)
  then s:= 'BGCOLOR="#FFEFB6"';
 result:= s;
end;

//on considre que les conditions sont runies pour calculer les corrlations reproduites
function TOpFactPCA.getHTMLRepResCorrMatrice(loadings: PMatrix; nbAxis: integer; sorted: boolean): string;
var buf: TBufString;
    s: string;
    i,j,k: integer;
    cor,res,orig: double;
    strColor: string;
    index: TPCAIndexVariables; 
begin

 index:= self.getIndexVariables(sorted,loadings,nbAxis);

 buf:= TBufString.Create();
 buf.BeginUpdate();
 s:= '<br>'+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH></TH>';
 for j:= 1 to FDimAnalysis do
  s:= s + '<TH width=60>'+self.Dataset.LstAtts[asInput].Attribute[index[j]-1].Name+'</TH>';
 buf.AddStr(s+'</TR>');
 for i:= 1 to FDimAnalysis do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY+'<TH>'+self.Dataset.LstAtts[asInput].Attribute[index[i]-1].Name+'</TH>';
   for j:= 1 to FDimAnalysis do
    begin
     if (i = j)
      then s:= s + '<TD align=center>-</TD>'
      else
       begin
        cor:= 0.0;
        for k:= 1 to nbAxis do cor:= cor + loadings[index[i],k]*loadings[index[j],k];
        orig:= FCorrMat^[index[i]]^[index[j]];
        res:= orig - cor;
        strColor:= colorCodeCorrResidual(orig,res);
        s:= s + format('<TD align=center %s><font color="#5F8CE0">%.4f</font><br><font color="green">%.4f</font><br><font color="gray">(%.4f)</font></TD>',[strColor,orig,cor,res]);
       end;
    end;
   buf.AddStr(s+'</TR>');
  end;
 buf.AddStr('</table>');
 //
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();

 Finalize(index);
end;

function TOpFactPCA.sortOnLoadings(loadings: PMatrix;
  nbAxis: integer): TPCAIndexVariables;
var k,i,nbVar,prem: integer;
    aIndex: TExamples;
    aRef: TAttribute;
    aOut: TPCAIndexVariables;
begin
 nbVar:= self.Dataset.LstAtts[asInput].Count;
 setLength(aOut,nbVar+1);
 TRY
  //crer la colonne d'index
  aIndex:= TExamples.Create(nbVar);
  aIndex.fillValues();
  //crer la colonne de rfrence
  aRef:= TAttContinue.Create('ref',nbVar);
  TRY
  prem:= 1;
  for k:= 1 to nbAxis do
   begin
    //2 variables ou plus, on peut trier
    if (aIndex.Size > 1)
     then
      begin
       for i:= 1 to nbVar do
        aRef.cValue[i]:= abs(loadings[i,k]);
       //attention le tri est croissant, il faut piocher  la fin de la liste
       aIndex.HeapSortBy(aRef);
       //rcuprer les valeurs
       //toutes si c'est les dernires
       if (k = nbAxis)
        then
         begin
          for i:= 1 to aIndex.Size do
           aOut[prem+i-1]:= aIndex[aIndex.Size-i+1];
         end
        else
         begin
          for i:= 1 to aIndex.Size do
           begin
            if (aRef.cValue[aIndex[aIndex.Size-i+1]] >= 0.5)
             then
              begin
               aOut[prem+i-1]:= aIndex[aIndex.Size-i+1];
               //si on a affect le dernier
               if (i = aIndex.Size)
                then
                 begin
                  aIndex.Size:= 0;
                  break;
                 end;
              end
             else
              begin
               prem:= prem+i-1;
               aIndex.Size:= nbVar-prem+1;
               break;
              end;
           end;
         end;
      end
     else
      begin
       //s'il n'y en a qu'un, on le met en dernier
       if (aIndex.Size = 1) then aOut[nbVar]:= aIndex[1];
       //sortir de for k
       BREAK;
      end;
   end;
  FINALLY
  aRef.Free();
  aIndex.Free();
  END;
 EXCEPT
 //option la plus basique
 for i:= 1 to nbVar do aOut[i]:= i;
 END;
 result:= aOut;
end;

function TOpFactPCA.getVariablesIndex: TPCAIndexVariables;
var nbVar,i: integer;
    aOut: TPCAIndexVariables;
begin
 nbVar:= self.Dataset.LstAtts[asInput].Count;
 setLength(aOut,nbVar+1);
 for i:= 1 to nbVar do aOut[i]:= i;
 result:= aOut;
end;

function TOpFactPCA.getIndexVariables(sorted: boolean;
  refLoadings: PMatrix; nbAxis: integer): TPCAIndexVariables;
begin
 if (sorted = false) or (refLoadings = nil)
  then result:= self.getVariablesIndex()
  else result:= self.sortOnLoadings(refLoadings,nbAxis);
end;

function TOpFactPCA.getPrexifNameForProjection: string;
begin
 result:= 'PCA';
end;

{ TOpPrmFactPCA }

function TOpPrmFactPCA.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmPCA.CreateFromOpPrm(self);
end;

procedure TOpPrmFactPCA.getFactSettings(prmSource: TOpPrmFactAnalysis);
var source: TOpPrmFactPCA;
begin
 source:= prmSource as TOpPrmFactPCA;
 //
 FNbAxisCreated:= source.FNbAxisCreated;
 FCos2Ctr:= source.FCos2Ctr;
 FStandardize:= source.FStandardize;
 FBartlettKmo:= source.FBartlettKmo;
 FShowCorr:= source.FShowCorr;
end;

function TOpPrmFactPCA.getHTMLParameters: string;
var s: string;
begin
 s:= '<P><B>Number of asked factors : </B>';
 if (FNbAxisCreated<0)
  then s:= s+'all'
  else s:= s+IntToStr(FNbAxisCreated);
 //new -- 04/02/2007
 s:= s + '<br><B>Compute COS2 and CTR : </B>' + IntToStr(ord(FCos2Ctr));
 //new -- 30/06/2011
 s:= s + '<br><B>Standardizing attributes : </B>' + IntToStr(ord(FStandardize));
 s:= s + '<br><B>Bartlett''s test and MSA (KMO indices) : </B>' + IntToStr(ord(FBartlettKmo));
 s:= s + '<br><B>Correlations and partial correlations : </B>' + IntToStr(ord(FShowCorr));
 s:= s + '<br><B>Reproduced correlations : </B>' + IntToStr(ord(FShowRepCorr));
 s:= s + '<br><B>Sort variables according to loadings : </B>' + IntToStr(ord(FSortVarOnLoadings));
 result:= s;
end;

procedure TOpPrmFactPCA.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FNbAxisCreated:= prmINI.ReadInteger(prmSection,'nb_axis',FNbAxisCreated);
 FCos2Ctr:= prmINI.ReadBool(prmSection,'cos2_ctr',FCos2Ctr);
 FStandardize:= prmINI.ReadBool(prmSection,'standardize',FStandardize);
 FBartlettKmo:= prmINI.ReadBool(prmSection,'bartlett',FBartlettKmo);
 FShowCorr:= prmINI.ReadBool(prmSection,'showcorr',FShowCorr);
 FShowRepCorr:= prmINI.ReadBool(prmSection,'show_rep_corr',FShowRepCorr);
 FSortVarOnLoadings:= prmINI.ReadBool(prmSection,'sorted',FSortVarOnLoadings);
end;

procedure TOpPrmFactPCA.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FNBAxisCreated,sizeof(FNBAxisCreated));
 prmStream.ReadBuffer(FCos2Ctr,sizeof(FCos2Ctr));
 prmStream.ReadBuffer(FStandardize,sizeof(FStandardize));
 prmStream.ReadBuffer(FBartlettKmo,sizeof(FBartlettKmo));
 prmStream.ReadBuffer(FShowCorr,sizeof(FShowCorr));
 prmStream.ReadBuffer(FShowRepCorr,sizeof(FShowRepCorr));
 prmStream.ReadBuffer(FSortVarOnLoadings,sizeof(FSortVarOnLoadings));
end;

procedure TOpPrmFactPCA.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_axis',FNbAxisCreated);
 prmINI.WriteBool(prmSection,'cos2_ctr',FCos2Ctr);
 prmINI.WriteBool(prmSection,'standardize',FStandardize);
 prmINI.WriteBool(prmSection,'bartlett',FBartlettKmo);
 prmINI.WriteBool(prmSection,'showcorr',FShowCorr);
 prmINI.WriteBool(prmSection,'show_rep_corr',FShowRepCorr);
 prmINI.WriteBool(prmSection,'sorted',FSortVarOnLoadings);
end;

procedure TOpPrmFactPCA.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FNBAxisCreated,sizeof(FNBAxisCreated));
 prmStream.WriteBuffer(FCos2Ctr,sizeof(FCos2Ctr));
 prmStream.WriteBuffer(FStandardize,sizeof(FStandardize));
 prmStream.WriteBuffer(FBartlettKmo,sizeof(FBartlettKmo));
 prmStream.WriteBuffer(FShowCorr,sizeof(FShowCorr));
 prmStream.WriteBuffer(FShowRepCorr,sizeof(FShowRepCorr));
 prmStream.WriteBuffer(FSortVarOnLoadings,sizeof(FSortVarOnLoadings));
end;

procedure TOpPrmFactPCA.SetDefaultParameters;
begin
 FNbAxisCreated:= 5;
 FCos2Ctr:= false;
 FStandardize:= true;
 FBartlettKmo:= false;
 FShowCorr:= false;
 FShowRepCorr:= false;
 FSortVarOnLoadings:= true;
end;

initialization
 CLASSES.RegisterClass(TMLGenCompFactPCA);
end.
