(**********************************************************************)
(* UCompFactRotateFactors.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Rotation des facteurs d'une analyse en composantes principales)
@author(Ricco)
@created(08/07/2006)
}

unit UCompFactRotateFactors;

interface

USES
    UCompDefinition,
    UCompManageDataset,
    UOperatorDefinition,
    UCompFactAnalysisPCA,
    UCalcFactRotateFactors,
    Classes, Forms, Inifiles,
    Matrices;

TYPE
    {gnrateur de composant}
    TMLGenCompFactRotation = class(TMLGenComp)
      protected
      procedure   GenCompInitializations(); override;
      public
      function    GetClassMLComponent: TClassMLComponent; override;
    end;

    {composant}
    TMLCompFactRotation = class(TMLCompLocalData)
      protected
      function    getClassOperator: TClassOperator; override;
      public
      //ne pourra tre connect que sous une analyse en composantes principales et ses drivs
      class function isConnectableTo(destination: TMLComponent): boolean; override;
    end;

    {oprateur}
    TOpFactRotation = class(TOpLocalData)
      private
      {oprateur d'ACP  traiter}
      FOpACP: TOpFactPCA;
      {nouveau loadings}
      FNewLoadings: PMatrix;
      {nouveaux vecteurs propres}
      FNewEgVec: PMatrix;
      {nouveau coefs pour les axes}
      FNewFactorScores: PMatrix;
      {nombre de facteurs  traiter}
      FNbFactors: integer;
      {dimension du tableau de donnes de l'ACP}
      FDimAnalysis: integer;
      protected
      function    getClassParameter: TClassOperatorParameter; override;
      function    CoreExecute(): boolean; override;
      //recalculer les nouveaux coefficients de rgression (ou les nouveaux vecteurs propres... c'est la mme chose)
      function    computeNewRegCoefficients(): boolean;
      //raliser la projection avec les nouveaux coefs.
      function    setProjections(): boolean;
      public
      //en profiter pour rcuprer l'oprateur d'ACP 
      constructor Create(AOwner: TObject); override;
      destructor  Destroy(); override;
      function    getHTMLResultsSummary(): string; override;
    end;

    {paramtrage}
    TOpFactRotationParameter = class(TOperatorParameter)
      private
      {nombre de facteurs  traiter}
      FNbFactors: integer;
      {type de rotation  mettre en place}
      FRotation: TEnumRotateFactors;
      //new - 14/09/2012 - montrer la matrice de corrlation reconstruite et le rsidu
      FShowRepCorr: boolean;
      //new - 15/09/2012 - trier selon les loadings
      FSortVarOnLoadings: boolean;
      protected
      function    CreateDlgParameters(): TForm; override;
      procedure   SetDefaultParameters(); override;
      public
      function    getHTMLParameters(): string; override;
      procedure   LoadFromStream(prmStream: TStream); override;
      procedure   SaveToStream(prmStream: TStream); override;
      procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
      procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
      property    NbFactors: integer read FNbFactors write FNbFactors;
      property    Rotation: TEnumRotateFactors read FRotation write FRotation;
      property    ShowRepCorr: boolean read FShowRepCorr write FShowRepCorr;
      property    SortVarOnLoadings: boolean read FSortVarOnLoadings write FSortVarOnLoadings;
    end;

implementation

uses
    UCalcMatrixAdditionalFunctions, UConstConfiguration,
    Sysutils, UDlgOpPrmRotateFactors;

{ TMLGenCompFactRotation }

procedure TMLGenCompFactRotation.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactRotation.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactRotation; 
end;

{ TMLCompFactRotation }

function TMLCompFactRotation.getClassOperator: TClassOperator;
begin
 result:= TOpFactRotation; 
end;

class function TMLCompFactRotation.isConnectableTo(
  destination: TMLComponent): boolean;
begin
 result:= destination.inheritsFrom(TMLCompFactPCA);
end;

{ TOpFactRotation }

(*
function TOpFactRotation.computeNewRegCoefficients(): boolean;
var ok: boolean;
    invCorr: PMatrix;
    i,j: integer;
    sum: double;
begin
 //inverser la matrice de corrlation
 Matrices.DimMatrix(invCorr,FOpACP.DimAnalysis,FOpACP.DimAnalysis);
 //new -- 22/11/2008 -- l'inverse SVD n'est pas prcise, surttout quand la matrice est inversible
 //modifier la procdure --> tenter d'abord l'inversion normale...
 TRY
  ok:= (Matrices.InvMat(FOpACP.MatCorr,1,FOpACP.DimAnalysis,invCorr) = MAT_OK);
 EXCEPT
  ok:= FALSE;
 END;
 //si plantage, on passe  la SVDInverse...
 if not(ok)
  then
   begin
    //ok:= (Matrices.InvMat(FOpACP.MatCorr,1,FOpACP.DimAnalysis,invCorr) = Matrices.MAT_OK);
    Matrices.CopyMatrix(invCorr,FOpACP.MatCorr,1,1,FOpACP.DimAnalysis,FOpACP.DimAnalysis);
    ok:= UCalcMatrixAdditionalFunctions.SVDInverse(invCorr,FOpACP.DimAnalysis);
   end;
 //suite normale...
 if ok
  then
   begin
    try
    //produit matriciel inverse de la matrice de corrlation avec les nouveaux loadings
    UCalcMatrixAdditionalFunctions.MatrixProd(invCorr,FOpACP.DimAnalysis,FOpACP.DimAnalysis,FNewLoadings,FOpACP.DimAnalysis,FNbFactors,FNewEgVec);

    //new -- 09/02/2012 -- on va utiliser un autre facteur de normalisation cf. plus bas

    //normer les vecteurs ???  la diffrence des logiciels U.S.
    //calculer la norme
    {
    for j:= 1 to FNbFactors do
     begin
      sum:= 0.0;
      for i:= 1 to FOpACP.DimAnalysis do
       sum:= sum + FNewEgVec[i,j] * FNewEgVec[i,j];
      //passer  la racine carre
      sum:= SQRT(sum);
      //normer maintenant
      if (sum > 0)
       then
        begin
         for i:= 1 to FOpACP.DimAnalysis do
          FNewEgVec[i,j]:= FNewEgVec[i,j] / sum;
        end;
     end;
     }
    //


    //new -- 09/02/2012
    //multiplier par la racine du vecteur propre de chaque facteur pour que variance.pop du facteur = somme des carrs des corrlations des variables avec le facteur
    //ainsi la variance de la colonne projete sera bien gale  la valeur propre (et la somme des carrs des corrlations des variables avec le facteur)
    //cf. e-mail Gilbert Laffond du 09/02/2012

    for j:= 1 to FNbFactors do
     begin
      //calculer la somme des carrs des loadings = valeur propre de l'axe
      sum:= 0.0;
      for i:= 1 to FOpACP.DimAnalysis do
       sum:= sum + FNewLoadings[i,j] * FNewLoadings[i,j];
      //mettre  la racine carre
      if (sum > 0.0) then sum:= sqrt(sum) else sum:= 0.0;
      //normaliser en multipliant la racine carre de la valeur propre associe  l'axe
      for i:= 1 to FOpACP.DimAnalysis do
       FNewEgVec[i,j]:= FNewEgVec[i,j] * sum;
     end;

    except
    ok:= false;
    end;
   end;
 Matrices.delMatrix(invCorr,FOpACP.DimAnalysis,FOpACP.DimAnalysis);
 result:= ok;
end;
*)

function TOpFactRotation.computeNewRegCoefficients: boolean;
begin
 FNewFactorScores:= FOpACP.getRotatedCoefs(FNewLoadings,FNewEgVec,FNbFactors);
 result:= (FNewFactorScores <> nil);
end;

function TOpFactRotation.CoreExecute: boolean;
var ok: boolean;
begin
 //si c'est matrices sont assignes, on doit bien sr utiliser l'ancienne valeur de NbFactors pour les librer
 //si NbFactors = 0, ces matrices n'ont pas pu tre assignes, donc pas de soucis potentiels...
 if assigned(FNewLoadings) then DelMatrix(FNewLoadings,FDimAnalysis,FNbFactors);
 if assigned(FNewEgVec) then DelMatrix(FNewEgVec,FDimAnalysis,FNbFactors);
 TRY
 //rcuprer les nouvelles valeurs (ventuellement) de NbFactors et de FDimAnalysis
 FNbFactors:= (self.PrmOp as TOpFactRotationParameter).FNbFactors;
 FDimAnalysis:= FOpACP.DimAnalysis;
 //copier la matrice
 Matrices.DimMatrix(FNewLoadings,FOpACP.DimAnalysis,FNbFactors);
 Matrices.CopyMatrix(FNewLoadings,FOpACP.Loadings,1,1,FOpACP.DimAnalysis,FNbFactors);
 //effectuer la rotation selon le paramtrage demand
 case (self.PrmOp as TOpFactRotationParameter).FRotation of
 rfQuartimax: QUARTIMAX(FNewLoadings,FOpACP.DimAnalysis,FNbFactors)
 else VARIMAX(FNewLoadings,FOpACP.DimAnalysis,FNbFactors);
 end;
 //recalculer les nouveaux coefficients de projection
 ok:= self.computeNewRegCoefficients();
 //la projection -- si c'est pas ok dans la ligne du dessus, la projection n'est pas tente
 ok:= ok and self.setProjections(); 
 //
 result:= ok;
 EXCEPT
 result:= false;
 END;
end;

constructor TOpFactRotation.Create(AOwner: TObject);
begin
 inherited create(AOwner);
 FOpACP:= (MLOwner as TMLComponent).Predecessor.Operator as TOpFactPCA;
end;

destructor TOpFactRotation.Destroy;
begin
 if assigned(FNewLoadings) then DelMatrix(FNewLoadings,FOpACP.DimAnalysis,FNbFactors);
 if assigned(FNewEgVec) then DelMatrix(FNewEgVec,FOpACP.DimAnalysis,FNbFactors);
 if assigned(FNewFactorScores) then DelMatrix(FNewFactorScores,FOpACP.DimAnalysis,FNbFactors);
 inherited destroy();
end;

function TOpFactRotation.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpFactRotationParameter;
end;

function TOpFactRotation.getHTMLResultsSummary: string;
var s: string;
begin
 s:= '<H3>Rotated Factor Loadings</H3>';
 s:= s + FOpACP.getHTMLLoadings(FNewLoadings,FNbFactors,(self.PrmOp as TOpFactRotationParameter).SortVarOnLoadings);
 s:= s + '<H4>vs. Unrotated Factor Loadings</H4>';
 s:= s + FOpACP.getHTMLLoadings(FOpACP.Loadings,FNbFactors,(self.PrmOp as TOpFactRotationParameter).SortVarOnLoadings,FNewLoadings);
 s:= s + '<H3>Factor Scores</H3>';
 //s:= s + FOpACP.getHTMLRegCoef(FNewEgVec,FNbFactors);
 s:= s + FOpACP.getHTMLRegCoef(FNewFactorScores,FNewLoadings,FNbFactors);
 //new - 14/09/2012 - matrice reproduite des corrlations
 if ((self.PrmOp as TOpFactRotationParameter).ShowRepCorr = true)
  then
   begin
    s:= s + '<H3>Original, reproduced and residual correlations</H3>';
    s:= s + FOpACP.getHTMLRepResCorrMatrice(FNewLoadings,self.FNbFactors,(self.PrmOp as TOpFactRotationParameter).SortVarOnLoadings);
   end;
 result:= s;
end;

function TOpFactRotation.setProjections(): boolean;
begin
 try
 //FOpACP.SetProjections(GenAtts,FNewEgVec,FNbFactors,'FR',(MLOwner as TMLComponent).Number);
 //on envoie la matrice de facteurs !!!
 FOpACP.SetProjections(GenAtts,FNewFactorScores,FNbFactors,'FR',(MLOwner as TMLComponent).Number);
 result:= true;
 except
 result:= false;
 end;
end;

{ TOpFactRotationParameter }

function TOpFactRotationParameter.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmFactRotateFactors.CreateFromOpPrm(self);
end;

function TOpFactRotationParameter.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Factors rotation</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD>Method</TD>'+ format('<TD>%s</TD>',[STR_FACTOR_ROTATION[FRotation]]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD># factors</TD>'+ format('<TD>%d</TD>',[FNbFactors]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD>Reproduced correlations</TD>'+ format('<TD>%d</TD>',[ord(FShowRepCorr)]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD>Sort variables according to loadings</TD>'+ format('<TD>%d</TD>',[ord(FSortVarOnLoadings)]) + '</TR>';
 s:= s + '</table>';
 result:= s;
end;

procedure TOpFactRotationParameter.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FNbFactors:= prmINI.ReadInteger(prmSection,'nb_factors',FNbFactors);
 FRotation:= TEnumRotateFactors(prmINI.ReadInteger(prmSection,'method',ord(FRotation)));
 FShowRepCorr:= prmINI.ReadBool(prmSection,'show_rep_corr',FShowRepCorr);
 FSortVarOnLoadings:= prmINI.ReadBool(prmSection,'sorted',FSortVarOnLoadings);
end;

procedure TOpFactRotationParameter.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FNbFactors,sizeof(FNbFactors));
 prmStream.ReadBuffer(FRotation,sizeof(FRotation));
 prmStream.ReadBuffer(FShowRepCorr,sizeof(FShowRepCorr));
 prmStream.ReadBuffer(FSortVarOnLoadings,sizeof(FSortVarOnLoadings));
end;

procedure TOpFactRotationParameter.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_factors',FNbFactors);
 prmINI.WriteInteger(prmSection,'method',ord(FRotation));
 prmINI.WriteBool(prmSection,'show_rep_corr',FShowRepCorr);
 prmINI.WriteBool(prmSection,'sorted',FSortVarOnLoadings);
end;

procedure TOpFactRotationParameter.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FNbFactors,sizeof(FNbFactors));
 prmStream.WriteBuffer(FRotation,sizeof(FRotation));
 prmStream.WriteBuffer(FShowRepCorr,sizeof(FShowRepCorr));
 prmStream.WriteBuffer(FSortVarOnLoadings,sizeof(FSortVarOnLoadings));
end;

procedure TOpFactRotationParameter.SetDefaultParameters;
begin
 FNbFactors:= 2;
 FRotation:= rfVarimax;
 FShowRepCorr:= false;
 FSortVarOnLoadings:= true;
end;

initialization
 RegisterClass(TMLGenCompFactRotation);
end.
