(********************************************************************)
(* UCompFactAnalysisAFC.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Analyse Factorielles des Correspondances)
@author(Ricco)
@created(09/02/2007)

Calculs Lebart et al., pp. 67  107
}
unit UCompFactAnalysisAFC;

interface

USES
  Forms,Classes,Contnrs,
  IniFiles,
  Matrices,
  UCalcStatDes,
  UCompDefinition,
  UCompFADefinition,
  UOperatorDefinition,
  UDatasetDefinition,
  UConstConfiguration;

TYPE
  //gnrateur
  TMLGenCompFactAFC = class(TMLGenComp)
   protected
   procedure   GenCompInitializations(); override;
   public
   function    GetClassMLComponent: TClassMLComponent; override;
  end;

  //composant
  TMLCompFactAFC = class(TMLCompFactAnalysis)
   protected
   function    getClassOperator: TClassOperator; override;
  end;

  //description d'une liste de modalite
  TLstModalite = class;

  //oprateur
  TOpFactAFC = class(TOpFactAnalysis)
   private
   //nb de valeurs propres  analyser
   FEgSize: integer;
   //matrices des donnes, des frquences relatives
   MatD, MatF: PMatrix;
   //la matrice  diagonaliser
   MatA: PMatrix;
   //vecteur des valeurs propres
   //FEgVal: PVector;
   //matrice des vecteurs propres
   FEgVec: PMatrix;
   //La trace -- inertie totale et le CHI-2
   FTrace, FChi2: double;
   //matrice des coordonnes factorielles, CTR et cos2 pour les lignes
   FLigCtr, FLigCos2: PMatrix;
   //matrice des coordonnes factorielles, CTR et cos2 pour les colonnes
   FColCtr, FColCos2: PMatrix;
   //liste des contributions
   FLstContrib: TObjectList;
   //les rsultats pour une meilleure organisation de l'affichage
   FResLig, FResCol: TLstModalite;
   //initialiser les matrices
   procedure   initMatrix();
   //diagonalisation
   procedure   diagonalize();
   //calculer les projections lignes
   procedure   computeLigProj();
   //calculer les projetcions colonnes,  partir des projections lignes
   procedure   computeColProj();
   //calculer les contributions
   procedure   computeContributions();
   //le test d'independance
   function    getHTMLChi2(): string;
   //les contributions
   function    getHTMLContributions(): string;
   //score coefs. pour les lignes supplmentaires - dfinies sur les colonnes donc
   function    getHTMLScoresSupRows(): string;
   //score coefs. pour les colonnes supplmentaires - dfinies sur les lignes donc
   function    getHTMLScoresSupColumns(): string;
   protected
   procedure   DelPrivateMatrix(); override;
   function    getClassParameter: TClassOperatorParameter; override;
   function    CheckAttributes(): boolean; override;
   //function    CoreExecute(): boolean; override;
   function    getNewForm(): TForm; override;
   function    getDimEigVal(): integer; override;
   //rcupration des rsultats
   procedure   recupResults();
   public
   //nombre d'axes rellement calcul
   FNbAxis: integer;
   //nombre max. d'axes potentiel
   FMaxNbAxis: integer;
   //nombre de lignes et de colonnes
   FNbLig, FNbCol: integer;
   //coordonnes lignes et colonnes
   FLigCoord, FColCoord: PMatrix;
   //distance au ceentre de gravit
   FDistoLig, FDistoCol: array of double;
   //les valeurs propres en HTML
   function    getHTMLEgValues(): string;
   //les projections des lignes
   function    getHTMLLignes(): string;
   //les projections colonnes
   function    getHTMLColonnes(): string;
   //rapport HTML complet
   function    getHTMLResultsSummary(): string; override;
   //new - 08/06/2012 - dcomposer les calculs en 3 tapes
   function    calcEigen(): boolean; override;
   //new - 18/11/2012 - rcupration des rsultats
   function    calcAdditional(): boolean; override;
   //projection pour obtenir les coordonnes factorielles
   function    calcProjections(): boolean; override;
   //new - 08/06/2012 - obtenir une instance d'analyse factorielle
   function    getInstanceFA(): TClassOpFactAnalysis; override;
   //proprits
   property    Trace: double read FTrace;
   property    MatFreq: PMatrix read MatF;
   property    LigCtr: PMatrix read FLigCtr;
   property    LigCos2: PMatrix read FLigCos2;
   property    ColCtr: PMatrix read FColCtr;
   property    ColCos2: PMatrix read FColCos2;
  end;

  //paramtrage
  TOpPrmFactAFC = class(TOpPrmFactAnalysis)
   protected
   function    CreateDlgParameters(): TForm; override;
   procedure   SetDefaultParameters(); override;
   public
   //nombre d'axes -- df. 5
   AskNbAxis: integer;
   //calculer et afficher les contributions au CHI-2 -- df. false
   IsComputeContrib: boolean;
   //afficher les coefficients des facteurs
   ShowScoreCoefs: boolean;
   //trier les valeurs selon les contributions - df. true
   SortValuesOnCtr: boolean;
   procedure   getFactSettings(prmSource: TOpPrmFactAnalysis); override;
   function    getHTMLParameters(): string; override;
   procedure   LoadFromStream(prmStream: TStream); override;
   procedure   SaveToStream(prmStream: TStream); override;
   procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
   procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
  end;

  //**********************************************
  // pour une meilleure organisation des rsultats
  //**********************************************

  //description des rsultats d'une modalit
  TModalite = class
   public
   //nombre d'axes
   NbAxis: integer;
   //description
   Description: string;
   //coordonnes
   Coord: array of double;
   //contributions
   Ctr: array of double;
   //cos2
   Cos2: array of double;
   //poids
   Weight: double;
   //carr de la distance  l'origine
   SqDist: double;
   //inertie
   Inertia: double;
   //constructeur
   constructor Create(prmDescription: string; prmNbAxis: integer);
   //destructeur
   destructor  destroy(); override;
   //description HTML
   function    getHTML(seuilCtr,seuilCos2: double; ligneBGColor : string = HTML_BGCOLOR_DATA_GREEN): string;
   //function    getHTML(seuilCos2: double; ligneBGColor : string = HTML_BGCOLOR_DATA_GREEN): string;
  end;

  //description pour une liste de modalit
  TLstModalite = class
   public
   //nombre d'axes
   NbAxis: integer;
   //nombre max. d'axes potentiels
   MaxNbAxis: integer;
   //liste des objets
   LstModalite: TObjectList;
   //constructeur
   constructor create(prmNbAxis,prmMaxNbAxis: integer);
   //destructeur
   destructor  destroy(); override;
   //ajouter un lment
   procedure   add(moda: TModalite);
   //nombre d'lments
   function    count(): integer;
   //accs  un lment
   function    get(index: integer): TModalite;
   //description d'une liste de modalits
   function    getHTML(seuilCos2: double = -1; ligneBGColor : string = HTML_BGCOLOR_DATA_GREEN): string;
   //trier de manire dcroissante sur le max. des CTR - par axe...
   procedure   SortOnCtr(seuil: double);
  end;


implementation

USES
    UDatasetImplementation,
    UCalcMatrixAdditionalFunctions, UStringAddBuffered,
    Math, FMath, Sysutils, UFrmOpViewFactAnalysisAFC,
    UDlgOpPrmFactAnalysisAFC;

TYPE
  //structure pour calcul des contributions
  TContribCHI2 = class
  //identifiant de ligne et de colonne
  sLig, sCol: string;
  //valeur et valeur espre pour la case
  value,expected: double;
  //contribution
  contrib: double;
  //en pourcentage
  percent: double;
  //new -- 07/12/2012 -- rsidu standardis
  stdResidual: double;
  end;

CONST
    //valeur  partir de laquelle on considre que c'est zro
    ZERO_VALUE_AFC : double = 1e-6;

var
  //indique le facteur de rfrence pour le tri des variables/modalits
  idFactor: integer;    


{ TMLGenCompFactAFC }

procedure TMLGenCompFactAFC.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactAFC.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactAFC; 
end;

{ TMLCompFactAFC }

function TMLCompFactAFC.getClassOperator: TClassOperator;
begin
 result:= TOpFactAFC;
end;

{ TOpFactAFC }

function TOpFactAFC.calcAdditional: boolean;
begin
 result:= true;
 TRY
 //none...
 EXCEPT
 result:= false;
 END;
end;

function TOpFactAFC.calcEigen: boolean;
begin
 result:= true;
 TRY
 //supprimer les matrices prcdentes
 self.DelPrivateMatrix();
 //nouvelle dimension des matrices
 FNbLig:= Dataset.Examples.Size;
 FNbCol:= Dataset.LstAtts[asInput].Count;
 //calculer le nombre d'axes  traiter
 FEgSize:= MATH.Min(FNbLig,FNbCol);//puisqu'on travaille dans l'espace des colonnes toujours
 FMaxNbAxis:= MATH.Min(FNbLig,FNbCol)-1;
 FNbAxis:= MATH.Min(FMaxNbAxis,(self.PrmOp as TOpPrmFactAFC).AskNbAxis);
 //copier les matrices de donnes et la matrice  diagonaliser
 self.initMatrix();
 //calcul des valeurs et vecteurs propres
 self.diagonalize();
 EXCEPT
 result:= false;
 END;
end;

function TOpFactAFC.calcProjections: boolean;
begin
 result:= true;
 TRY
 //calculer les projections lignes
 self.computeLigProj();
 //et les projections colonnes
 self.computeColProj();
 //ventuellement, calcul des contributions
 if ((self.PrmOp as TOpPrmFactAFC).IsComputeContrib = true)
  then self.computeContributions();
 //rcuprer les rsultats dans les structures particulires pour l'affichage
 self.recupResults();
 EXCEPT
 result:= false;
 END;
end;

function TOpFactAFC.CheckAttributes: boolean;
var ok: boolean;
begin
 //un target discret pour indicateur de lignes du tableau de contingence
 ok:= (self.Dataset.LstAtts[asTarget].Count = 1) and (self.Dataset.LstAtts[asTarget].isAllCategory(caDiscrete) = true);
 //puis plusieurs input, tous continus
 ok:= ok and ((self.Dataset.LstAtts[asInput].Count > 1) and (self.Dataset.LstAtts[asInput].isAllCategory(caContinue) = true));
 //and then...
 result:= ok;
end;

procedure TOpFactAFC.computeColProj;
var a,j,i: integer;
    sum: double;
    //dist: array of single;
begin
 //les coordonnes, Lebart et al., p.85 -- relations de transitions
 dimMatrix(FColCoord,FNbCol,FNbAxis);
 for a:= 1 to FNbAxis do
  begin
   for j:= 1 to FNbCol do
    begin
     sum:= 0.0;
     for i:= 1 to FNbLig do
       sum:= sum + MatF^[i]^[j]/MatF^[0]^[j]*FLigCoord^[i]^[a];
     //vrif. avant de diviser
     if (abs(EigVal[a+1]) > ZERO_VALUE_AFC)
      then FColCoord^[j]^[a]:= sum/SQRT(EigVal[a+1])
      else FColCoord^[j]^[a]:= 0.0;
    end;
  end;
 //contributions
 dimMatrix(FColCtr,FNbCol,FNbAxis);
 for a:= 1 to FNbAxis do
  for j:= 1 to FNbCol do
   if (abs(EigVal[a+1]) > ZERO_VALUE_AFC)
    then FColCtr^[j]^[a]:= 100.0 * MatF^[0]^[j] * SQR(FColCoord^[j]^[a]) / EigVal[a+1]
    else FColCtr^[j]^[a]:= 0.0;
 //distance
 setLength(FDistoCol,succ(FNbCol));
 for j:= 1 to FNbCol do
  for i:= 1 to FNbLig do
    FDistoCol[j]:= FDistoCol[j] + 1.0/MatF^[i]^[0]*SQR(MatF^[i]^[j]/matF^[0]^[j]-MatF^[i]^[0]);
 //cos
 dimMatrix(FColCos2,FNbCol,FNbAxis);
 for a:= 1 to FNbAxis do
  for j:= 1 to FNbCol do
   if (abs(FDistoCol[j]) > ZERO_VALUE_AFC)
    then FColCos2^[j]^[a]:= SQR(FColCoord^[j]^[a])/FDistoCol[j]
    else FColCos2^[j]^[a]:= 0.0;
 //
 //finalize(dist);
end;

//critre de tri, descendant
function triContrib(Item1, Item2: Pointer): integer;
var c1, c2: TContribChi2;
begin
 c1:= TContribCHI2(Item1);
 c2:= TContribCHI2(Item2);
 if (c1.contrib < c2.contrib)
  then result:= +1
  else if (c1.contrib > c2.contrib)
    then result:= -1
    else result:= 0;
end;

procedure TOpFactAFC.computeContributions;
var i,j: integer;
    sI,sJ: string;
    c: TContribCHI2;
begin
 FLstContrib:= TObjectList.Create(TRUE);//propritaire
 //pour chaque case du tableau de contingence
 for i:= 1 to FNbLig do
  begin
   sI:= self.Dataset.LstAtts[asTarget].Attribute[0].sValue[self.Dataset.Examples.Number[i]];
   for j:= 1 to FNbCol do
    begin
     sJ:= self.Dataset.LstAtts[asInput].Attribute[pred(j)].Name;
     //
     c:= TContribCHI2.Create();
     c.sLig:= sI;
     c.sCol:= sJ;
     c.value:= MatD^[i]^[j];
     c.expected:= MatD^[i]^[0]*MatD^[0]^[j]/MatD^[0]^[0];
     if (abs(c.expected) > ZERO_VALUE_AFC)
      then c.contrib:= SQR(c.value - c.expected)/c.expected
      else c.contrib:= 0.0;
     c.percent:= 100.0*c.contrib/FChi2;
     //new - 07/12/2012 - calcul des rsidus standardiss
     c.stdResidual:= 0;
     if (abs(c.expected) > ZERO_VALUE_AFC)
      then c.stdResidual:= (c.value - c.expected)/sqrt(c.expected);
     //
     FLstContrib.Add(c);
    end;
  end;
 //trier par ordre d'importance dcroissante
 FLstContrib.Sort(triContrib);
end;

procedure TOpFactAFC.computeLigProj;
var i,j,a: integer;
    sum: double;
    //dist: array of single;
begin
 //coordonnes
 dimMatrix(FLigCoord,FNbLig,FNbAxis);
 for a:= 1 to FNbAxis do
  begin
   for i:= 1 to FNbLig do
    begin
     sum:= 0.0;
     for j:= 1 to FNbCol do
      sum:= sum + MatF^[i]^[j] / (MatF^[i]^[0] * MatF^[0]^[j]) * FEgVec^[j]^[a+1];
     FLigCoord^[i]^[a]:= sum;
    end;
  end;
 //contributions
 dimMatrix(FLigCtr,FNbLig,FNbAxis);
 for a:= 1 to FNbAxis do
  for i:= 1 to FNbLig do
   if (EigVal[a+1] > ZERO_VALUE_AFC)
    then FLigCtr^[i]^[a]:= 100.0 * MatF^[i]^[0] * SQR(FLigCoord^[i]^[a]) / EigVal[a+1]
    else FLigCtr^[i]^[a]:= 0.0;
 //calculer la distance du point au centre de gravit
 setLength(FDistoLig,succ(FNbLig));
 for i:= 1 to FNbLig do
  for j:= 1 to FNbCol do
   FDistoLig[i]:= FDistoLig[i] + 1.0/MatF^[0]^[j]*SQR(MatF^[i]^[j]/MatF^[i]^[0]-MatF^[0]^[j]);
 //cos
 dimMatrix(FLigCos2,FNbLig,FNbAxis);
 for a:= 1 to FNbAxis do
  for i:= 1 to FNbLig do
   if (FDistoLig[i] > ZERO_VALUE_AFC)
    then FLigCos2^[i]^[a]:= SQR(FLigCoord^[i]^[a])/FDistoLig[i]
    else FLigCos2^[i]^[a]:= 0.0;
 //finalize
 //Finalize(dist);
end;

(*
function TOpFactAFC.CoreExecute: boolean;
begin
 result:= true;
 TRY
 //supprimer les matrices prcdentes
 self.DelPrivateMatrix();
 //nouvelle dimension des matrices
 FNbLig:= self.WorkData.Examples.Size;
 FNbCol:= self.WorkData.LstAtts[asInput].Count;
 //calculer le nombre d'axes  traiter
 FEgSize:= MATH.Min(FNbLig,FNbCol);//puisqu'on travaille dans l'espace des colonnes toujours
 FNbAxis:= MATH.Min(MATH.Min(FNbLig,FNbCol)-1,(self.PrmOp as TOpPrmFactAFC).AskNbAxis);
 //copier les matrices de donnes et la matrice  diagonaliser
 self.initMatrix();
 //calcul des valeurs et vecteurs propres
 self.diagonalize();
 //calculer les projections lignes
 self.computeLigProj();
 //et les projections colonnes
 self.computeColProj();
 //ventuellement, calcul des contributions
 if ((self.PrmOp as TOpPrmFactAFC).IsComputeContrib = true)
  then self.computeContributions();
 EXCEPT
 result:= false;
 END;
end;
*)

procedure TOpFactAFC.DelPrivateMatrix;
begin
 inherited;
 if assigned(MatD) then delMatrix(MatD,FNbLig,FNbCol);
 if assigned(MatF) then delMatrix(MatF,FNbLig,FNbCol);
 if assigned(MatA) then delMatrix(MatA,FNbCol,FNbCol);
 if assigned(FEgVec) then delMatrix(FEgVec,FNbCol,FNbCol);
 //if assigned(FEgVal) then delVector(FEgVal,FNbCol);
 //pour les lignes
 if assigned(FLigCoord) then delMatrix(FLigCoord,FNbLig,FNbAxis);
 if assigned(FLigCtr) then delMatrix(FLigCtr,FNbLig,FNbAxis);
 if assigned(FLigCos2) then delMatrix(FLigCos2,FNbLig,FNbAxis);
 //pour les colonnes
 if assigned(FColCoord) then delMatrix(FColCoord,FNbCol,FNbAxis);
 if assigned(FColCtr) then delMatrix(FColCtr,FNbCol,FNbAxis);
 if assigned(FColCos2) then delMatrix(FColCos2,FNbCol,FNbAxis);
 //pour la liste des contributions
 if assigned(FLstContrib) then FreeAndNil(FLstContrib);
 //pour les distances aux centres de gravit
 if assigned(FDistoLig) then Finalize(FDistoLig);
 if assigned(FDistoCol) then Finalize(FDistoCol);
 //pour l'affichage des rsultats
 if assigned(FResLig) then FreeAndNil(FResLig);
 if assigned(FResCol) then FreeAndNil(FResCol);
end;

procedure TOpFactAFC.diagonalize;
var j,i: integer;
begin
 GetEigensFromSymetricMatrix(MatA,FNbCol,FEigVal,FEgVec);
 //calcul de la trace
 FTrace:= 0.0;
 for j:= 2 to FEgSize do FTrace:= FTrace + EigVal[j];
 //et du chi-2 : trace * effectif total
 FChi2:= FTrace * MatD^[0]^[0];
 //correction des vecteurs propres. Lebart et al., pp.102
 for j:= 1 to FNbCol do //cf. la formule u = Dp^(-1/2) * w (le signe semble invers ? c'est +1/2 ???) ou c'est moi qui interprte mal ???
   for i:= 1 to FNbCol do    //hum, il semble que ce soit une multiplication par la racine carre de la marge, et non une division ???
    FEgVec^[i]^[j]:= FEgVec^[i]^[j] * SQRT(MatF^[0]^[i]); //en tous cas, avec cette criture du programme, le rsultat est correct, c'est l'essentiel
end;

function TOpFactAFC.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactAFC;
end;

function TOpFactAFC.getDimEigVal: integer;
begin
 result:= FNbCol;
end;

function TOpFactAFC.getHTMLChi2: string;
var p: double;
    ddl: integer;
    s: string;
begin
 ddl:= (FNbLig-1) * (FNbCol-1);
 p:= PKhi2(ddl,FChi2);
 s:= '<P><H3>CHI-SQUARE statistic</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD>Trace</TD><TD>%.4f</TD></TR>',[FTrace]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD>Chi</TD><TD>%.2f</TD></TR>',[FChi2]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD>d.f</TD><TD>%d</TD></TR>',[ddl]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD>p-value</TD><TD>%.4f</TD></TR>',[p]);
 s:= s + '</table>';
 result:= s; 
end;

function TOpFactAFC.getHTMLColonnes: string;
var s: string;
    j,a: integer;
    seuilCos2, seuilCtr: double;
    couleur: string;
begin
 //calculer les seuils pour l'affichage
 seuilCos2:= 1.0 / (-1.0 + self.FEgSize) - ZERO_VALUE_AFC;//pour les problmes de troncature
 seuilCtr:= 100.0 / (1.0 * FNbCol) - ZERO_VALUE_AFC;
 //
 s:= '';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH colspan=4>%s</TH>',['Column Characterization']) +
         format('<TH colspan=%d>Coord.</TH>',[FNbAxis]) + format('<TH colspan=%d>Contributions (%s)</TH>',[FNbAxis,'%']) + format('<TH colspan=%d>COS</TH>',[FNbAxis]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=150>Values</TH><TH width="60">Weight</TH>';
 s:= s + '<TH width="60">Sq. Dist.</TH><TH width="60">Inertia</TH>';
 //en-tte des coordones, etc.
 for j:= 1 to FNbAxis do s:= s + format('<TD width="60" align="center">coord %d</TD>',[j]);
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">ctr %d</TD>',[j]);
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">cos %d</TD>',[j]);
 s:= s + '</TR>';
 //valeurs
 for j:= 1 to FNbCol do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD align="left" %s>%s</TD>',[HTML_BGCOLOR_DATA_BLUE,Dataset.lstAtts[asInput].Attribute[pred(j)].Name]);
   //poids, distances au centre de gravit (Sq. Dist) et inerties (inertia)
   s:= s + format('<TD align="right" %s><font color="teal">%.5f</font></TD>',[HTML_BGCOLOR_DATA_LAVENDER,MatF^[0]^[j]]);
   s:= s + format('<TD align="right" %s><font color="teal">%.5f</font></TD>',[HTML_BGCOLOR_DATA_LAVENDER,FDistoCol[j]]);
   s:= s + format('<TD align="right" %s><font color="teal">%.5f</font></TD>',[HTML_BGCOLOR_DATA_LAVENDER,MatF^[0]^[j]*FDistoCol[j]]);
   for a:= 1 to FNbAxis do
    begin
     couleur:= HTML_BGCOLOR_DATA_GRAY;
     if (FColCtr^[j]^[a] >= seuilCtr) and (FColCos2^[j]^[a] >= seuilCos2)
      then
       begin
        if (FColCoord^[j]^[a] < 0)
         then couleur:= FACT_COLOR_HIGHLIGHT[cthNegLow]
         else couleur:= FACT_COLOR_HIGHLIGHT[cthPosLow];
       end;    
     s:= s + format('<TD align="right" %s>%.5f</TD>',[couleur,FColCoord^[j]^[a]]);
    end;
   for a:= 1 to FNbAxis do s:= s + format('<TD align="right">%.1f</TD>',[FColCtr^[j]^[a]]);
   for a:= 1 to FNbAxis do s:= s + format('<TD align="right">%.3f</TD>',[FColCos2^[j]^[a]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactAFC.getHTMLContributions: string;
var seuil: double;
    i: integer;
    bf: TBufString;
    c: TContribCHI2;
    color,signe: string;
begin
 if assigned(FLstContrib) and (FLstContrib.Count > 0)
  then
   begin
    //le seuil pour ne pas tout afficher
    seuil:= FChi2/(1.0*FNbLig*FNbCol);
    //zoo..
    bf:= TBufString.Create();
    bf.BeginUpdate();
    //
    bf.AddStr('<H3>CHI-2 contributions</H3>');
    bf.AddStr(HTML_HEADER_TABLE_RESULT);
    bf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Id</TH><TH>Row</TH><TH>Column</TH><TH>Value</TH><TH>Expected</TH><TH>Std.Resid.</TH><TH colspan=2>Contrib.</TH><TH>%</TH></TR>');
    i:= 0;
    c:= FLstContrib.Items[i] as TContribCHI2;
    while (c<>nil) and (c.contrib > seuil) do
     begin
      //affichage
      bf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TD align="right">%d</TD><TD>%s</TD><TD>%s</TD><TD align="right">%.0f</TD><TD align="right">%.1f</TD><TD align="right">%.2f</TD>',[i+1,c.sLig,c.sCol,c.value,c.expected,c.StdResidual]));
      if (c.value < c.expected)
       then
        begin
         color:= HTML_BGCOLOR_LIGHT_RED;
         signe:= '(-)';
        end
       else
        begin
         color:= HTML_BGCOLOR_LIGHT_GREEN;
         signe:= '(+)';
        end;
      bf.AddStr(format('<TD %s><b>%s</b></TD><TD %s align="right"> %.2f</TD><TD align="right">%.2f</TD>',[color,signe,color,c.contrib,c.percent]));
      bf.AddStr('</TR>');
      //next...
      inc(i);
      if (i < FLstContrib.Count)
       then c:= FLstContrib.Items[i] as TContribCHI2
       else c:= nil;
     end;
    bf.AddStr('</table>');
    //
    bf.EndUpdate();
    result:= bf.BufS;
    bf.Free();
   end
  else result:= '';
end;

function TOpFactAFC.getHTMLEgValues: string;
var tot, sum: double;
    s,couleur: string;
    i: integer;
    seuilEig: double;
begin
 //seuil - rgle du Kaiser, cf. aussi -> http://marketing-bulletin.massey.ac.nz/V14/MB_V14_T2_Bendixen.pdf
 if (FEgSize > 1) then seuilEig:= FTrace/(-1.0 + FEgSize) else seuilEig:= 0.0;
 //
 s:= format('<b>Matrix trace = %.4f</b>',[FTrace]);
 s:= s + format('<br><b>SQRT(Matrix trace) = %.4f</b><br>',[SQRT(FTrace)]);
 s:= s+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH>Axis</TH><TH>Eigen value</TH><TH>% explained</TH><TH width="200">Histogram</TH><TH>% cumulated</TH></TR>';
 tot:= FTrace;
 sum:= 0.0;
 for i:= 2 to FEgSize do
  begin
   sum:= sum + EigVal[i];
   if (tot>0)
    then
     begin
      if (EigVal[i] >= SeuilEig) then couleur:= HTML_BGCOLOR_DATA_LAVENDER else couleur:= HTML_BGCOLOR_DATA_GRAY;
      s:= s + HTML_TABLE_COLOR_DATA_GRAY+
          format('<TH>%d</TH><TD align=right %s>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD><TD align=right>%.2f%s</TD><TD>%s</TD><TD align=right>%.2f%s</TD></TR>',
                [i-1,couleur,FEigVal[i],100.0*EigVal[i]/tot,'%',getHtmlHistogram(TRUNC(100.0*EigVal[i]/tot)),100.0*sum/tot,'%'])+'</TR>';
     end;
  end;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH>Tot.</TH><TD align=right>%.6f</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD></TR>',[sum]);
 s:= s + '</table>';
 result:= s;
end;

function TOpFactAFC.getHTMLLignes: string;
var s: string;
    i,j: integer;
    seuilCos2, seuilCtr: double;
    couleur: string;
begin
 //calculer les seuils pour l'affichage
 seuilCos2:= 1.0 / (-1.0 + self.FEgSize) - ZERO_VALUE_AFC;
 seuilCtr:= 100.0 / (1.0 * FNbLig) - ZERO_VALUE_AFC;
 //
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH colspan=4>%s</TH>',['Row Characterization']) +
         format('<TH colspan=%d>Coord.</TH>',[FNbAxis]) + format('<TH colspan=%d>Contributions (%s)</TH>',[FNbAxis,'%']) + format('<TH colspan=%d>COS</TH>',[FNbAxis]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Values</TH><TH width="60">Weight</TH>';
 //mettre les distances au centre de gravit (Sq. Dist) et les inerties (inertia)
 s:= s + '<TH width="60">Sq. Dist.</TH><TH width="60">Inertia</TH>';
 //en-tte des coordones, etc.
 for j:= 1 to FNbAxis do s:= s + format('<TD width="60" align="center">coord %d</TD>',[j]);
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">ctr %d</TD>',[j]);
 for j:= 1 to FNbAxis do s:= s + format('<TD width="40" align="center">cos %d</TD>',[j]);
 s:= s + '</TR>';
 //valeurs
 for i:= 1 to FNbLig do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD width=150 align="left" %s>%s</TD>',[HTML_BGCOLOR_DATA_GREEN,Dataset.lstAtts[asTarget].Attribute[0].sValue[Dataset.examples.Number[i]]]);
   //poids, distances au centre de gravit (Sq. Dist) et inerties (inertia)
   s:= s + format('<TD align="right" %s><font color="teal">%.5f</font></TD>',[HTML_BGCOLOR_DATA_LAVENDER,MatF^[i]^[0]]);
   s:= s + format('<TD align="right" %s><font color="teal">%.5f</font></TD>',[HTML_BGCOLOR_DATA_LAVENDER,FDistoLig[i]]);
   s:= s + format('<TD align="right" %s><font color="teal">%.5f</font></TD>',[HTML_BGCOLOR_DATA_LAVENDER,MatF^[i]^[0]*FDistoLig[i]]);
   //coordonnes,
   for j:= 1 to FNbAxis do
    begin
     couleur:= HTML_BGCOLOR_DATA_GRAY;
     if (FLigCtr^[i]^[j] >= seuilCtr) and (FLigCos2^[i]^[j] >= seuilCos2)
      then
       begin
        if (FLigCoord^[i]^[j] < 0)
         then couleur:= FACT_COLOR_HIGHLIGHT[cthNegLow]
         else couleur:= FACT_COLOR_HIGHLIGHT[cthPosLow];
       end;
     s:= s + format('<TD align="right" %s>%.5f</TD>',[couleur,FLigCoord^[i]^[j]]);
    end;
   for j:= 1 to FNbAxis do s:= s + format('<TD align="right">%.2f</TD>',[FLigCtr^[i]^[j]]);
   for j:= 1 to FNbAxis do s:= s + format('<TD align="right">%.3f</TD>',[FLigCos2^[i]^[j]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactAFC.getHTMLResultsSummary: string;
var buf: TBufString;
    seuilCos2: double;
begin
 seuilCos2:= 1.0 / (-1.0 + self.FEgSize) - ZERO_VALUE_AFC;
 //buffriser l'ajout -- 23/06/2004
 buf:= TBufString.Create();
 buf.BeginUpdate();
 //test d'indpendance
 buf.AddStr(self.getHTMLChi2());
 //afficher les valeurs propres et la trace de la matrice
 buf.AddStr('<P><H3>Eigen values</H3>' + self.getHTMLEgValues());
 //les lignes
 //buf.AddStr('<H3>Rows analysis</H3>' + self.getHTMLLignes());
 buf.AddStr('<H3>Rows analysis</H3>' + Self.FResLig.getHTML(seuilCos2));
 //les colonnes
 //buf.AddStr('<H3>Columns analysis</H3>' + self.getHTMLColonnes());
 buf.AddStr('<H3>Columns analysis</H3>' + self.FResCol.getHTML(seuilCos2,HTML_BGCOLOR_LIGHT_BLUE));
 //les contributions ?
 if ((self.PrmOp as TOpPrmFactAFC).IsComputeContrib = true)
  then buf.AddStr(self.getHTMLContributions());
 //les coefficients des facteurs ?
 if ((self.PrmOp as TOpPrmFactAFC).ShowScoreCoefs = true)
  then
   begin
    buf.AddStr('<P><H3>Factor score coefficients for supplementary points</H3>');
    buf.AddStr(self.getHTMLScoresSupRows());
    buf.AddStr(self.getHTMLScoresSupColumns());
   end;
 //and then...
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TOpFactAFC.getHTMLScoresSupColumns: string;
var s: string;
    i,j: integer;
    value: double;
begin
 s:= '<BR><BR><font size="3">Factor score coefficients for supplementary column</font>';
 s:= s + '<br><font color="gray">From row values (relative frequency)</font>';
 s:= s + '<P>' + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Row value</TH>';
 for j:= 1 to FNbAxis do
  s:= s + format('<TH width=60>Factor %d</TH>',[j]);
 s:= s + '</TR>';
 //valeurs
 for i:= 1 to FNbLig do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD width=150 align="left" %s>%s</TD>',[HTML_BGCOLOR_DATA_GREEN,Dataset.lstAtts[asTarget].Attribute[0].sValue[Dataset.examples.Number[i]]]);
   //coordonnes,
   for j:= 1 to FNbAxis do
    begin
     if (FEigVal[j+1] > 0.0)
      then value:= FLigCoord^[i]^[j] / SQRT(FEigVal[j+1])
      else value:= 0.0;
     s:= s + format('<TD align="right">%.6f</TD>',[value]);
    end;
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactAFC.getHTMLScoresSupRows: string;
var s: string;
    j,a: integer;
    value: double;
begin
 s:= '<font size="3">Factor score coefficients for supplementary row</font>';
 s:= s + '<br><font color="gray">From column values (relative frequency)</font>';
 s:= s + '<P>' + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Column value</TH>';
 for j:= 1 to FNbAxis do
  s:= s + format('<TH width=60>Factor %d</TH>',[j]);
 s:= s + '</TR>';
 //valeurs
 for j:= 1 to FNbCol do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD width=150 align="left" %s>%s</TD>',[HTML_BGCOLOR_DATA_BLUE,Dataset.lstAtts[asInput].Attribute[pred(j)].Name]);
   //poids, distances au centre de gravit (Sq. Dist) et inerties (inertia)
   for a:= 1 to FNbAxis do
    begin
     if (FEigVal[a+1] > 0.0)
      then value:= FColCoord^[j]^[a] / SQRT(FEigVal[a+1])
      else value:= 0.0;
     s:= s + format('<TD align="right">%.6f</TD>',[value]);
    end;
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactAFC.getInstanceFA: TClassOpFactAnalysis;
begin
 result:= TOpFactAFC;
end;

function TOpFactAFC.getNewForm: TForm;
begin
 result:= TFrmOpViewFactAnalysisAFC.CreateFromOperator(self);
end;

procedure TOpFactAFC.initMatrix;
var sum: single;
    i,j,jprim,example: integer;
begin
 //matrice des donnes -- les marges sont mises  zro automatiquement, ouf...
 dimMatrix(MatD,FNbLig,FNbCol);
 for i:= 1 to FNbLig do
  begin
   example:= self.Dataset.Examples.Number[i];
   for j:= 1 to FNbCol do
    begin
     MatD^[i]^[j]:= self.Dataset.LstAtts[asInput].Attribute[pred(j)].cValue[example];
     //les marges
     matD^[i]^[0]:= matD^[i]^[0] + MatD^[i]^[j];
     matD^[0]^[j]:= matD^[0]^[j] + MatD^[i]^[j];
     matD^[0]^[0]:= matD^[0]^[0] + MatD^[i]^[j];
    end;
  end;
 //matrice des frquences, y compris les frquences marginales
 dimMatrix(MatF,FNbLig,FNbCol);
 for i:= 0 to FNbLig do
  for j:= 0 to FNbCol do
   MatF^[i]^[j]:= MatD^[i]^[j] / matD^[0]^[0];
 //matrice de calcul -- de taille col x col -  diagonaliser
 dimMatrix(matA,FNbCol,FNbCol);
 for j:= 1 to FNbCol do
  for jprim:= 1 to FNbCol do
   begin
    sum:= 0.0;
    for i:= 1 to FNbLig do
     sum:= sum + (MatF^[i]^[j] * MatF^[i]^[jprim]) / (MatF^[i]^[0] * SQRT(MatF^[0]^[j] * MatF^[0]^[jprim]));
    MatA^[j]^[jprim]:= sum;
   end;
end;

procedure TOpFactAFC.recupResults;
var k,j: integer;
    moda: TModalite;
    seuilCtr: double;
begin
 //rcupration des valeurs lignes
 FResLig:= TLstModalite.create(self.FNbAxis,self.FMaxNbAxis);
 for k:= 1 to self.FNbLig do
  begin
   moda:= TModalite.Create(Self.Dataset.LstAtts[asTarget].Attribute[0].sValue[k],Self.FNbAxis);
   moda.Weight:= Self.MatFreq[k][0];
   moda.SqDist:= Self.FDistoLig[k];
   moda.Inertia:= moda.Weight * moda.SqDist;
   for j:= 1 to moda.NbAxis do
    begin
     moda.Coord[j-1]:= Self.FLigCoord[k][j];
     moda.Ctr[j-1]:= Self.LigCtr[k][j];
     moda.Cos2[j-1]:= Self.LigCos2[k][j];
    end;
   FResLig.add(moda);
  end;
 //rcupration des valeurs colonnes
 FResCol:= TLstModalite.create(self.FNbAxis,self.FMaxNbAxis);
 for k:= 1 to self.FNbCol do
  begin
   moda:= TModalite.Create(Self.Dataset.LstAtts[asInput].Attribute[k-1].Name,Self.FNbAxis);
   moda.Weight:= Self.MatFreq[0][k];
   moda.SqDist:= Self.FDistoCol[k];
   moda.Inertia:= moda.Weight * moda.SqDist;
   for j:= 1 to moda.NbAxis do
    begin
     moda.Coord[j-1]:= Self.FColCoord[k][j];
     moda.Ctr[j-1]:= Self.FColCtr[k][j];
     moda.Cos2[j-1]:= Self.FColCos2[k][j];
    end;
   FResCol.add(moda);
  end;
 //faut-il trier les rsultats ?
 if ((self.PrmOp as TOpPrmFactAFC).SortValuesOnCtr = true)
  then
   begin
    seuilCtr:= 100.0/(1.0 * Self.FNbLig) - ZERO_VALUE_AFC;
    FResLig.SortOnCtr(seuilCtr);
    seuilCtr:= 100.0/(1.0 * Self.FNbCol) - ZERO_VALUE_AFC;
    FResCol.SortOnCtr(seuilCtr);
   end;
end;

{ TOpPrmFactAFC }

function TOpPrmFactAFC.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFactAnalysisAFC.CreateFromOpPrm(self);
end;

procedure TOpPrmFactAFC.getFactSettings(prmSource: TOpPrmFactAnalysis);
var source: TOpPrmFactAFC;
begin
 source:= prmSource as TOpPrmFactAFC;
 AskNbAxis:= source.AskNbAxis;
 IsComputeContrib:= source.IsComputeContrib;
 ShowScoreCoefs:= source.ShowScoreCoefs;
 SortValuesOnCtr:= source.SortValuesOnCtr;
end;

function TOpPrmFactAFC.getHTMLParameters: string;
var s: string;
begin
 s:= format('<b>Nb axis : %d</b><br>',[AskNbAxis]);
 s:= s + format('<b>Contrib. : %d</b><br>',[ord(IsComputeContrib)]);
 s:= s + format('<b>Fact. Score Coefs. : %d</b><br>',[ord(ShowScoreCoefs)]);
 s:= s + format('<b>Sort values on CTR : %d</b><br>',[ord(SortValuesOnCtr)]);
 result:= s;
end;

procedure TOpPrmFactAFC.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 AskNbAxis:= prmINI.ReadInteger(prmSection,'nb_axis',AskNbAxis);
 IsComputeContrib:= prmINI.ReadBool(prmSection,'contrib',IsComputeContrib);
 ShowScoreCoefs:= prmINI.ReadBool(prmSection,'scores',ShowScoreCoefs);
 SortValuesOnCtr:= prmINI.ReadBool(prmSection,'sort_values',SortValuesOnCtr);
end;

procedure TOpPrmFactAFC.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(AskNbAxis,sizeof(AskNbAxis));
 prmStream.ReadBuffer(IsComputeContrib,sizeof(IsComputeContrib));
 prmStream.ReadBuffer(ShowScoreCoefs,sizeof(ShowScoreCoefs));
 prmStream.ReadBuffer(SortValuesOnCtr,sizeof(SortValuesOnCtr));
end;

procedure TOpPrmFactAFC.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_axis',AskNbAxis);
 prmINI.WriteBool(prmSection,'contrib',IsComputeContrib);
 prmINI.WriteBool(prmSection,'scores',ShowScoreCoefs);
 prmINI.WriteBool(prmSection,'sort_values',SortValuesOnCtr);
end;

procedure TOpPrmFactAFC.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(AskNbAxis,sizeof(AskNbAxis));
 prmStream.WriteBuffer(IsComputeContrib,sizeof(IsComputeContrib));
 prmStream.WriteBuffer(ShowScoreCoefs,sizeof(ShowScoreCoefs));
 prmStream.WriteBuffer(SortValuesOnCtr,sizeof(SortValuesOnCtr));
end;

procedure TOpPrmFactAFC.SetDefaultParameters;
begin
 //nombre d'axes -- df. 5
 AskNbAxis:= 5;
 //calculer et afficher les contributions au CHI-2 -- df. false
 IsComputeContrib:= false;
 //afficher les coefficients de facteurs (pour les individus supplmentaires) - df. false
 self.ShowScoreCoefs:= true;
 //trier
 self.SortValuesOnCtr:= true;
end;

{ TModalite }

constructor TModalite.Create(prmDescription: string; prmNbAxis: integer);
begin
 inherited create();
 self.Description:= prmDescription;
 self.NbAxis:= prmNbAxis;
 setLength(self.Coord,prmNbAxis);
 setLength(self.Ctr,prmNbAxis);
 setLength(self.Cos2,prmNbAxis);
end;

destructor TModalite.destroy;
begin
 Finalize(Coord);
 Finalize(Ctr);
 Finalize(Cos2);
 inherited;
end;

//function TModalite.getHTML(seuilCos2: double; ligneBGColor : string = HTML_BGCOLOR_DATA_GREEN): string;
function TModalite.getHTML(seuilCtr, seuilCos2: double; ligneBGColor : string = HTML_BGCOLOR_DATA_GREEN): string;
var s: string;
    j: integer;
    couleur: string;
    cumsum: double;
begin
 s:= HTML_TABLE_COLOR_DATA_GRAY;
 s:= s + format('<TD width=100 align="left" %s>%s</TD>',[ligneBGColor,Description]);
 //poids, distances au centre de gravit (Sq. Dist) et inerties (inertia)
 s:= s + format('<TD align="right"><font color="teal">%.5f</font></TD>',[Weight]);
 s:= s + format('<TD align="right"><font color="teal">%.5f</font></TD>',[SqDist]);
 s:= s + format('<TD align="right"><font color="teal">%.5f</font></TD>',[Inertia]);
 //coordonnes,
 for j:= 1 to NbAxis do
  begin
   couleur:= HTML_BGCOLOR_DATA_GRAY;
   //if (Ctr[j-1] >= 100.0 * Weight) and (Cos2[j-1] >= seuilCos2)
   if (Ctr[j-1] >= seuilCtr) and (Cos2[j-1] >= seuilCos2)
    then
     begin
      if (Coord[j-1] < 0)
       then couleur:= FACT_COLOR_HIGHLIGHT[cthNegLow]
       else couleur:= FACT_COLOR_HIGHLIGHT[cthPosLow];
     end;
   s:= s + format('<TD align="right" %s>%.5f</TD>',[couleur,Coord[j-1]]);
  end;
 for j:= 1 to NbAxis do s:= s + format('<TD align="right">%.2f</TD>',[Ctr[j-1]]);
 cumsum:= 0.0;
 for j:= 1 to NbAxis do
  begin
   cumsum:= cumsum + Cos2[j-1];
   s:= s + format('<TD align="right">%.2f &nbsp; (%.2f)</TD>',[Cos2[j-1],cumsum]);
  end;
 s:= s + '</TR>';
 result:= s;
end;

{ TLstModalite }

procedure TLstModalite.add(moda: TModalite);
begin
 LstModalite.Add(moda);
end;

function TLstModalite.count: integer;
begin
 result:= LstModalite.Count;
end;

constructor TLstModalite.create(prmNbAxis,prmMaxNbAxis: integer);
begin
 inherited create();
 NbAxis:= prmNbAxis;
 MaxNbAxis:= prmMaxNbAxis;
 LstModalite:= TObjectList.Create(true);
end;

destructor TLstModalite.destroy;
begin
 LstModalite.Free();
 inherited;
end;

function TLstModalite.get(index: integer): TModalite;
begin
 result:= LstModalite.Items[index] as TModalite;
end;

function TLstModalite.getHTML(seuilCos2: double = -1; ligneBGColor : string = HTML_BGCOLOR_DATA_GREEN): string;
var s: string;
    i,j: integer;
    seuilCtr: double;
begin
 //calculer les seuils
 seuilCtr:= 100.0 / (1.0 * self.count) - ZERO_VALUE_AFC;
 //if (seuilCos2 <= 0) then seuilCos2:= 1.0 / (1.0 * self.NbAxis) - ZERO_VALUE_AFC;
 if (seuilCos2 <= 0) then seuilCos2:= 1.0 / (1.0 * self.MaxNbAxis) - ZERO_VALUE_AFC;
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH colspan=4>%s</TH>',['Characterization']) +
         format('<TH colspan=%d>Coord.</TH>',[NbAxis]) + format('<TH colspan=%d>Contributions (%s)</TH>',[NbAxis,'%']) + format('<TH colspan=%d>COS</TH>',[NbAxis]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width="150">Values</TH><TH width="50">Weight</TH>';
 //mettre les distances au centre de gravit (Sq. Dist) et les inerties (inertia)
 s:= s + '<TH width="50">Sq. Dist.</TH><TH width="50">Inertia</TH>';
 //en-tte des coordones, etc.
 for j:= 1 to NbAxis do s:= s + format('<TD width="50" align="center">coord %d</TD>',[j]);
 for j:= 1 to NbAxis do s:= s + format('<TD width="50" align="center">ctr %d</TD>',[j]);
 for j:= 1 to NbAxis do s:= s + format('<TD width="70" align="center">cos %d</TD>',[j]);
 s:= s + '</TR>';
 for i:= 0 to pred(self.count) do
  s:= s + self.get(i).getHTML(seuilCtr,seuilCos2,ligneBGColor);
  //s:= s + self.get(i).getHTML(seuilCos2,ligneBGColor);
 s:= s + '</table>';
 result:= s;
end;

//comparaison de deux modalits
function CompareModalite(Item1, Item2: Pointer): integer;
var p1, p2: TModalite;
    value1, value2: double;
begin
 p1:= TModalite(Item1);
 p2:= TModalite(Item2);
 //c'est ici que a va se jouer
 value1:= p1.Ctr[idFactor];
 value2:= p2.Ctr[idFactor];
 //comparaison - tri croissant
 if (value1 < value2)
  then result:= -1
  else
   begin
    if (value1 > value2)
     then result:= +1
     else result:= 0;
   end;
end;


procedure TLstModalite.SortOnCtr(seuil: double);
var output: TObjectList;

  //dpiler tant que non vide et suprieur  seuil
  procedure depiler(limit: double);
  var i,initNb: integer;
      ctrMax: double;
  begin
   initNb:= self.count();
   for i:= pred(initNb) downto 0 do
    begin
     ctrMax:= self.get(i).Ctr[idFactor];
     if (ctrMax >= limit)
      then
       begin
        output.add(self.get(i));
        self.LstModalite.Delete(i);
       end
      else break;
    end;
  end;

begin
 //modifier le statut de la liste interne
 LstModalite.OwnsObjects:= false;
 //crer la liste de rsultats
 output:= TObjectList.Create(false);
 //lancer les tris successifs
 idFactor:= 0;
 while (idFactor < self.NbAxis) and (self.count() > 1) do
  begin
   self.LstModalite.Sort(CompareModalite);
   depiler(seuil);
   inc(idFactor);   
  end;
 //s'il en reste, dpiler compltement
 if (self.count() > 0)
  then
   begin
    //prendre en rfrence le dernier facteur
    idFactor:= pred(self.NbAxis);
    depiler(-1);//-1 pour qu'il prenne tout ce qui reste
   end;
 //assigner et elle redevient propritaire
 LstModalite.Assign(output);
 LstModalite.OwnsObjects:= true;
 output.Free();
end;


initialization
 RegisterClass(TMLGenCompFactAFC);
end.
