(********************************************************************)
(* UCompFactAnalysisMCA.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Analyse (factorielle) des correspondances multiples)
@author(Ricco)
@created(12/01/2004)

New -- 04/02/2007 -- Scinder l'affichage en 2 parties pour une meilleure lisibilit.
Rajouter la contribution cumule dans l'affichage, qui est passe en %.
}
unit UCompFactAnalysisMCA;

interface

USES
        Forms, Classes,
        IniFiles, Contnrs,
        UCompDefinition,
        UCompFADefinition,
        UOperatorDefinition,
        UCalcStatDes,
        UCalcStatDesCrossTab,
        Matrices,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UConstConfiguration;

TYPE
        {gnrateur de composant ACM}
        TMLGenCompFactMCA = class(TMLGenComp)
                            protected
                            procedure   GenCompInitializations(); override;
                            public
                            function    GetClassMLComponent: TClassMLComponent; override;
                            end;

        {composant ACM}
        TMLCompFactMCA = class(TMLCompFactAnalysis)
                         protected
                         function    getClassOperator: TClassOperator; override;
                         function    GetLogResultDescription(): string; override;
                         end;

        //classe de caractrisation des rsultats
        TCaracFactorsLstVar = class;

        {oprateur ACM, on travaille sur les target}
        TOpFactMCA = class(TOpFactAnalysis)
                     private
                     {statistiques univaries sur les attributs}
                     FLstStatDiscrete: TLstCalcStatDesDiscrete;
                     {statistiques bivaries sur les couples d'attributs}
                     FLstStatCrossTab: TLstCalcStatDesCrossTab;
                     {Vecteur des distributions marginales des modalits}
                     FVecDistValues: PVector;
                     {vecteur des valeurs propres}
                     //FEigVals: PVector;
                     {matrice des vecteurs propres}
                     FEigVecs: PMatrix;
                     {dimension globale, addition dn nombre des valeurs de l'ensemble des modalits}
                     FDimAllValues: integer;
                     {nombre de valeurs propres}
                     FNbEigens: integer;
                     {nombre d'axes de l'tude}
                     FNbAxis: integer;
                     {Liste des rsultats sur les variables actives}
                     FLstCaracActive: TCaracFactorsLstVar;
                     //liste des caractrisations illustratives qualitatives
                     FLstCaracIllusDiscrete: TCaracFactorsLstVar;
                     //liste des caractrisations illustratives quantitatives
                     FLstCaracIllusContinuous: TCaracFactorsLstVar;
                     {calculs stats desc prparation du tableau de BURT}
                     procedure   CalcStats();
                     {construire les val-vec propres}
                     procedure   CalcEigens();
                     {calculer les coordonnes des modalits sur les axes}
                     function    BuildCoordValues(): PMatrix;
                     {calculer les cos des modalits,  partir des coordonnes}
                     function    BuildCosValues(prmCoord: PMatrix): PMatrix;
                     {calculer les contributions des modalits}
                     function    BuildContribValues(prmCoord: PMatrix): PMatrix;
                     {effectuer les projections sur les axes des individus}
                     procedure   SetProjections();
                     //new -- 05/02/2007 -- calculer les cos2 et ctr
                     procedure   setCos2Ctr();
                     protected
                     procedure   DelPrivateMatrix(); override;
                     function    getClassParameter: TClassOperatorParameter; override;
                     //function    CoreExecute(): boolean; override;
                     function    CheckAttributes(): boolean; override;
                     //new - 08/06/2012 - dimension du tableau des valeurs propres
                     function    getDimEigVal(): integer; override;
                     //new - 18/09/2012 - obtenir les factor scores
                     function    getHTMLFactorScores(): string;
                     //new - 08/11/2012 - obtenir le tableau des valeurs propres
                     function    getHTMLEigenValues(): string;
                     //new - 10/11/2012 - tableau des coordonnes et valeur-test - variables actives
                     function    getHTMLCoordVTestActive(): string;
                     //new - 10/11/2012 - tableau des cos2 et des contributions - variables actives
                     function    getHTMLCos2CtrActive(): string;
                     //new - 10/11/2012 - ancienne prsentation des tableaux de rsultats
                     function    getHTMLCaracActive(): string;
                     //new - 10/11/2012 - nouvelle prsentation des tableaux de rsultats
                     function    getHTMLCaracActiveAlternative(): string;
                     //new - 11/11/2012 - description du problme  rsoudre
                     function    getHTMLProblemDescription(): string;
                     //new - 20/12/2012 - fentre d'affichage spcifique
                     function    getNewForm(): TForm; override;
                     public
                     constructor Create(AOwner: TObject); override;
                     constructor CreateWithDataset(AOwner: TObject; prmDataset: TMLDataset); override;
                     destructor  Destroy; override;
                     function    getHTMLResultsSummary(): string; override;
                     //new - 08/06/2012 - dcomposer les calculs en 3 tapes
                     function    calcEigen(): boolean; override;
                     function    calcAdditional(): boolean; override;
                     function    calcProjections(): boolean; override;
                     //new - 10/11/2012 - en profiter pour effectuer les calculs sur les variables supplmentaires
                     function    calcPostProjections(): boolean; override;
                     //new - 08/06/2012 - obtenir une instance d'analyse factorielle
                     function    getInstanceFA(): TClassOpFactAnalysis; override;
                     //proprits
                     property    NbAxis: integer read FNbAxis;
                     property    LstCarActive: TCaracFactorsLstVar read FLstCaracActive;
                     property    LstCarIllus: TCaracFactorsLstVar read FLstCaracIllusDiscrete; 
                     end;

        {paramtre oprateur ACM}
        TOpPrmFactMCA = class(TOpPrmFactAnalysis)
                        private
                        {nombre d'axes  afficher}
                        FNbAxis: integer;
                        {seuil de la valeur test}
                        FThresoldValue: double;
                        //calculer les cos2 et CTR -- df. false
                        FCos2Ctr: boolean;
                        //alternative presentation
                        FAlternativeReport: boolean;
                        //trier les variables
                        FSortResult: boolean;
                        protected
                        function    CreateDlgParameters(): TForm; override;
                        procedure   SetDefaultParameters(); override;
                        public
                        procedure   getFactSettings(prmSource: TOpPrmFactAnalysis); override;
                        procedure   LoadFromStream(prmStream: TStream); override;
                        procedure   SaveToStream(prmStream: TStream); override;
                        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
                        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
                        function    getHTMLParameters(): string; override;
                        property    NbAxis: integer read FNbAxis write FNbAxis;
                        property    ThresoldValue: double read FThresoldValue write FThresoldValue;
                        property    Cos2Ctr: boolean read FCos2Ctr write FCos2Ctr;
                        property    AlternativeReport: boolean read FAlternativeReport write FAlternativeReport;
                        property    SortResult: boolean read FSortResult write FSortResult;
                        end;

        //**********************************************************************************
        //new - 08/11/2012 - nouvelles classes pour une meilleure organisation des rsultats
        //**********************************************************************************

        //new - 08/11/2012
        //caractrisation d'un facteur avec une modalit de variable
        TCaracModaFactor = record
            //coordonnes
            coord: double;
            //valeur test
            vtest: double;
            //cosinus 2
            cos2 : double;
            //contribution
            ctr  : double
        end;

        //new - 08/11/2012
        //caractrisation des facteurs avec une modalit
        TCaracModalite = record
            //frequence relative (n_j/n)
            frequence: double;
            //poids de la modalit [n_j/(n*p)] - Saporta p.231
            mass: double;
            //distance au carr par rapport  l'origine - Saporta p.232
            disto: double;
            //informations relatives aux facteurs
            factors: array of TCaracModaFactor;
        end;

        //caractrisation des facteurs avec une variable
        TCaracFactorsVariable = class
            public
            //nombre de facteurs
            NbAxis: integer;
            //variable
            LaVar: TAttribute;
            //contribution globale de la variable pour chaque axe
            CTRFactors: array of double;
            //description pour chaque modalit
            LstCarac: array of TCaracModalite;
            //constructeur
            constructor create(sourceVar: TAttribute; nbFactors: integer);
            //destructeur
            destructor  destroy(); override;
            //obtenir les infos pour une modalit
            function get(numModa: integer): TCaracModalite;
            //obtenir le nombre de categorie
            function getCount(): integer;
            //tableau des descriptions globales
            function    getHTMLOverall(): string;
            //tableau des coordonnes et valeur-test
            function    getHTMLCoordVTest(seuilVTest, seuilCos2: double): string;
            //obtenir les cos2 et ctr
            function    getHTMLCos2Ctr(): string;
            //tableau des coordonnes,valeur-test, cos2 et ctr
            function    getHTMLCoordVTestCos2Ctr(seuilVTest, seuilCos2: double): string;
            //tableau des coordonnes,valeur-test, cos2
            function getHTMLCoordVTestCos2(seuilVTest: double; colorBackTR: string = HTML_TABLE_COLOR_DATA_GRAY): string;
            //tableau des coordonnes,valeur-test, cos2 - variable continue
            function getHTMLCoordVTestCos2Continuous(seuilVTest: double; colorBackTR: string = HTML_TABLE_COLOR_DATA_GRAY): string;
            //caractrisation  partir d'une variable catgorielle
            class function caracFromDiscrete(att: TAttribute; examples: TExamples; op: TOpFactMCA): TCaracFactorsVariable;
            //caractrisation  partir d'une variable quantitative
            class function caracFromContinuous(att: TAttribute; examples: TExamples; op: TOpFactMCA): TCaracFactorsVariable;
        end;

        //caractrisation des facteurs avec des variables
        TCaracFactorsLstVar = class
            public
            //nombre d'axes
            NbAxis: integer;
            //liste des variables avec leurs caractrisations
            LstCaracVar: TObjectList;
            //constructeur
            constructor create(aNbAxis: integer; owner: boolean);
            //destructeur
            destructor  destroy(); override;
            //ajouter un attribut de caractrisation
            function   addVar(prmAtt: TAttribute; prmNbAxis: integer): TCaracFactorsVariable;
            //ajouter un objet de caractrisation
            procedure  addCarac(carac: TCaracFactorsVariable);
            //nombre d'lments
            function    getCount(): integer;
            //accder  la variable i
            function    getCaracVar(index: integer): TCaracFactorsVariable;
            //retirer la caractrisation - la suppression dpend de "owner" de la liste interne
            procedure   remove(index: integer);
            //recopie
            function    recopie(owner: boolean): TCaracFactorsLstVar;
            //trier les lements selon un critre
            function    sortOnCTR(owner: boolean; seuil: double): TCaracFactorsLstVar;
            //tableau des coordonnes et valeur-test
            function    getHTMLCoordVTest(seuilVTest,seuilCos2: double): string;
            //obtenir les cos2 et ctr
            function    getHTMLCos2Ctr(): string;
            //tableau des coordonnes,valeur-test, cos2 et ctr
            function    getHTMLCoordVTestCos2Ctr(seuilVTest,seuilCos2: double): string;
            //tableau des coordonnes,valeur-test, cos2
            function    getHTMLCoordVTestCos2Supp(seuilVTest: double): string;
            //tableau des coordonnes,valeur-test, cos2 (variables continues)
            function    getHTMLCoordVTestCos2SuppContinuous(seuilVTest: double): string;
        end;
        
implementation

uses
        SysUtils, Math, UCalcMatrixToAttributes,
        UCalcDistribution, UCalcMatrixAdditionalFunctions,
        UDlgOpPrmMCA, ULogFile, UStringAddBuffered, UStringsResources,
        UCalcStatDesConditionnalDesc, UCalcStatDesCorrelation,
  UFrmOpViewFactACM;

const
        SEUIL_FOR_CTR_SORTING : double = 1.0;

        HTML_COLOR_CODE_OVERALL = HTML_BGCOLOR_DATA_GRAY; //'BGCOLOR="#FFFFE0"';
        HTML_TABLE_COLOR_CODE_OVERALL = '<TR ' + HTML_COLOR_CODE_OVERALL + '>';


{ TMLGenCompFactCMA }

procedure TMLGenCompFactMCA.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
 //FMLNumIcon:= 11;
 //FMLCompName:= str_comp_name_fact_mca;
 //FMLBitmapFileName:= 'MLFactorialMCA.bmp';
end;

function TMLGenCompFactMCA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactMCA;
end;

{ TMLCompFactMCA }

function TMLCompFactMCA.getClassOperator: TClassOperator;
begin
 result:= TOpFactMCA;
end;

function TMLCompFactMCA.GetLogResultDescription: string;
var data: TMLDataset;
begin
 data:= (Operator as TOpFactMCA).Dataset;
 result:= Format('%d attributes, and %d examples',[data.LstAtts[asInput].Count,data.Examples.Size]);
end;

{ TOpPrmFactMCA }

function TOpPrmFactMCA.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmMCA.CreateFromOpPrm(self);
end;

procedure TOpPrmFactMCA.getFactSettings(prmSource: TOpPrmFactAnalysis);
var source: TOpPrmFactMCA;
begin
 source:= prmSource as TOpPrmFactMCA;
 FNbAxis:= source.FNbAxis;
 FThresoldValue:= source.FThresoldValue;
 FCos2Ctr:= source.FCos2Ctr;
end;

function TOpPrmFactMCA.getHTMLParameters: string;
var s,sPrm: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>MCA parameters</TH></TR>';
 if (self.NbAxis<0)
  then sPrm:= 'all'
  else sPrm:= 'limited';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Generate axis</TD><TD align="right">%s</TD></TR>',[sPrm]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Number of axis</TD><TD align="right">%d</TD></TR>',[self.NbAxis]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Thresold value</TD><TD align="right">%.2f</TD></TR>',[self.ThresoldValue]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Alternative report</TD><TD align="right">%d</TD></TR>',[ord(FAlternativeReport)]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Sort variables according to ctr.</TD><TD align="right">%d</TD></TR>',[ord(FSortResult)]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpPrmFactMCA.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FNbAxis:= prmINI.ReadInteger(prmSection,'nb_axis',FNbAxis);
 FThresoldValue:= prmINI.ReadFloat(prmSection,'thresold_value',FThresoldValue);
 FCos2Ctr:= prmINI.ReadBool(prmSection,'cos2_ctr',FCos2Ctr);
 FAlternativeReport:= prmINI.ReadBool(prmSection,'alt_report',FAlternativeReport);
 FSortResult:= prmINI.ReadBool(prmSection,'sort_result',FSortResult);
end;

procedure TOpPrmFactMCA.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FNBAxis,sizeof(FNBAxis));
 prmStream.ReadBuffer(FThresoldValue,sizeof(FThresoldValue));
 prmStream.ReadBuffer(FCos2Ctr,sizeof(FCos2Ctr));
 prmStream.ReadBuffer(FAlternativeReport,sizeof(FAlternativeReport));
 prmStream.ReadBuffer(FSortResult,sizeof(FSortResult));
end;

procedure TOpPrmFactMCA.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_axis',FNbAxis);
 prmINI.WriteFloat(prmSection,'thresold_value',FThresoldValue);
 prmINI.WriteBool(prmSection,'cos2_ctr',FCos2Ctr);
 prmINI.WriteBool(prmSection,'alt_report',FAlternativeReport);
 prmINI.WriteBool(prmSection,'sort_result',FSortResult);
end;

procedure TOpPrmFactMCA.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FNBAxis,sizeof(FNBAxis));
 prmStream.WriteBuffer(FThresoldValue,sizeof(FThresoldValue));
 prmStream.WriteBuffer(FCos2Ctr,sizeof(FCos2Ctr));
 prmStream.WriteBuffer(FAlternativeReport,sizeof(FAlternativeReport));
 prmStream.WriteBuffer(FSortResult,sizeof(FSortResult));
end;

procedure TOpPrmFactMCA.SetDefaultParameters;
begin
 FNbAxis:= 5;
 FThresoldValue:= 3.0;
 FCos2Ctr:= false;
 FAlternativeReport:= false;
 FSortResult:= true;
end;

{ TOpFactMCA }

function TOpFactMCA.BuildContribValues(prmCoord: PMatrix): PMatrix;
var i,j: integer;
    n,m: double;
    mat: PMatrix;
begin
 //transtypage pour calculs
 n:= 1.0*Dataset.Examples.Size;
 m:= 1.0*Dataset.LstAtts[asInput].Count;
 //matrice des contributions
 dimMatrix(mat,FDimAllValues,FNbEigens);
 for i:= 1 to FDimAllValues do
  for j:= 1 to FNbEigens do
   mat^[i]^[j]:= FVecDistValues^[i]*SQR(prmCoord^[i]^[j])/(n*m*EigVal[j]);
 //renvoyer le tout
 result:= mat;
end;

function TOpFactMCA.BuildCoordValues: PMatrix;
var i,j: integer;
    mat: PMatrix;
    n,m: double;

begin
 //transtypage
 n:= 1.0*Dataset.Examples.Size;
 m:= 1.0*Dataset.LstAtts[asInput].Count;
 //matrice des coordonnes
 dimMatrix(mat,FDimAllValues,FNbEigens);
 for i:= 1 to FDimAllValues do
  for j:= 1 to FNbEigens do
   begin
   //coordModa^[i]^[j]:= sqrt(1.0*ex.Size*lstAtt.Count/MATBURT^[i]^[i]*goodEVal^[j])*goodEVec^[i]^[j];//test...
   //TraceLog.WriteToLogFile(format('vectdist>>%.4f | eigvals >> %.4f | eigvec >> %.4f',[FVecDistValues^[i],FEigVals^[j],FEigVecs^[i]^[j]]));
   //new -- 03/04/2008 -- idem, protger un cran supplmentaire
   //if (FEigVals^[j]>0)
   if (FVecDistValues^[i]*EigVal[j] > 0.0)
   //suite...
    then mat^[i]^[j]:= SQRT(1.0*n*m/FVecDistValues^[i]*EigVal[j])*FEigVecs^[i]^[j]
    else mat^[i]^[j]:= 0.0;
   end;
 //renvoyer le tout
 result:= mat;
end;

function TOpFactMCA.BuildCosValues(prmCoord: PMatrix): PMatrix;
var i,j: integer;
    matCos: PMatrix;
    n: double;
begin
 n:= 1.0*Dataset.Examples.Size;
 
 dimMatrix(matCos,FDimAllValues,FNbEigens);
 for i:= 1 to FDimAllValues do
  for j:= 1 to FNbEigens do
   if (FVecDistValues^[i]>0) and (n/FVecDistValues^[i]-1.0<>0)
    then matCos^[i]^[j]:= SQR(prmCoord^[i]^[j]/sqrt(1.0*n/FVecDistValues^[i]-1.0))
    else matCos^[i]^[j]:= 0.0;

 result:= matCos;
end;

procedure TOpFactMCA.CalcEigens;
var MatBURT,MatACM: PMatrix;
    tmpEVal: PVector;
    tmpEVec: PMatrix;
    i,j: integer;
begin
 //tableau de BURT
 MatBURT:= BuildBurtMat(FDimAllValues,Dataset.Examples,Dataset.LstAtts[asInput],FLstStatDiscrete,FLstStatCrossTab);
 //tableau  diagonaliser
 MatACM := BuildACMMat(FDimAllValues,Dataset.LstAtts[asInput].Count,Dataset.Examples.Size,MatBURT);
 //giclons la matrice de BURT
 delMatrix(MatBURT,FDimAllValues,FDimAllValues);
 //valeurs et vecteurs propres temporaires, avant giclation des lments triviaux
 GetEigensFromSymetricMatrix(MatACM,FDimAllValues,tmpEVal,tmpEVec);
 //giclons la matrice diagonalise
 delMatrix(MatACM,FDimAllValues,FDimAllValues);
 //rcuprer les valeurs propres intrssantes, et les vecteurs propres associes
 dimVector(FEigVal,FNbEigens);
 dimMatrix(FEigVecs,FDimAllValues,FNbEigens);
 for j:= 1 to FNbEigens do
  begin
   EigVal[j]:= tmpEVal^[succ(j)];
   for i:= 1 to FDimAllValues do
    FEigVecs^[i]^[j]:= tmpEVec^[i]^[succ(j)];
  end;
 //dtruire les eigens temporaires
 delVector(tmpEVal,FDimAllValues);
 delMatrix(tmpEVec,FDimAllValues,FDimAllValues);
end;

procedure TOpFactMCA.CalcStats;
var i,i_cumul,j: integer;
    stat: TCalcStatDesDiscrete;
begin
 FLstStatDiscrete.RebuildStatDes(Dataset.LstAtts[asInput],Dataset.Examples);
 FLstStatCrossTab.RebuildStatDes(Dataset.LstAtts[asInput],Dataset.Examples);
 //calculer galement qqs indicateurs ncessaires
 //le vecteur des distributions marginales, la dimension et la liste des modalits
 FDimAllValues:= 0;
 for i:= 0 to pred(FLstStatDiscrete.Count) do
  begin
   stat:= TCalcStatDesDiscrete(FLstStatDiscrete.Stat(i));
   inc(FDimAllValues,stat.Attribute.nbValues);
  end;
 DimVector(FVecDistValues,FDimAllValues);//ok, FDimValues contient la bonne valeur
 i_cumul:= 0;
 for i:= 1 to FLstStatDiscrete.Count do
  begin
   stat:= TCalcStatDesDiscrete(FLstStatDiscrete.Stat(pred(i)));
   for j:= 1 to stat.TabFreq.Size do
    begin
     inc(i_cumul);
     FVecDistValues^[i_cumul]:= stat.TabFreq.Value[j];
    end;
  end;
 //nombre de vecteurs propres  rcuprer rellement
 FNbEigens:= FDimAllValues-Dataset.LstAtts[asInput].Count;
end;

function TOpFactMCA.CheckAttributes: boolean;
begin
 result:= (self.Dataset.LstAtts[asInput].Count>0) and (self.Dataset.LstAtts[asInput].isAllCategory(caDiscrete));
end;

(*
function TOpFactMCA.CoreExecute: boolean;
begin
 result:= TRUE;
 TRY
 //TraceLog.WriteToLogFile('[ACM] begin calc stat');
 self.CalcStats();
 //TraceLog.WriteToLogFile('[ACM] begin calc eigens');
 self.CalcEigens();
 TraceLog.WriteToLogFile('[ACM] begin set projections');
 self.SetProjections();
 //new -- 05/02/2007 -- ventuellement, calculer les COS2 et CTR
 if ((self.PrmOp as TOpPrmFactMCA).Cos2Ctr = true)
  then self.setCos2Ctr();
 EXCEPT
 self.DelPrivateMatrix();
 result:= FALSE;
 END;
end;
*)

constructor TOpFactMCA.Create(AOwner: TObject);
begin
 inherited Create(AOwner);
 FLstStatDiscrete:= TLstCalcStatDesDiscrete.Create(NIL,NIL);
 FLstStatCrossTab:= TLstCalcStatDesCrossTab.Create(NIL,NIL);
end;

procedure TOpFactMCA.DelPrivateMatrix;
begin
 inherited;
 FLstStatDiscrete.FreeAll;
 FLstStatCrossTab.FreeAll;
 delVector(FVecDistValues,FDimAllValues);
 //delVector(FEigVals,FNbEigens);
 delMatrix(FEigVecs,FDimAllValues,FNbEigens);
end;

destructor TOpFactMCA.Destroy;
begin
 inherited Destroy;
 FLstStatDiscrete.Free;
 FLstStatCrossTab.Free;
 if assigned(self.FLstCaracActive) then FLstCaracActive.Free();
 if assigned(FLstCaracIllusDiscrete) then FLstCaracIllusDiscrete.Free();
 if assigned(FLstCaracIllusContinuous) then FLstCaracIllusContinuous.Free();
end;

function TOpFactMCA.getDimEigVal: integer;
begin
 result:= FNbEigens;
end;

function TOpFactMCA.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactMCA; 
end;

function TOpFactMCA.getHTMLResultsSummary: string;
var bs: TBufString;
    seuilVTest: double;
begin
 //seuil pour V-TEST
 seuilVTest:= Math.MaxDouble;//dsactiver pour les points supplmentaires... pour l'instant...
 //empiler les descriptions
 bs:= TBufString.Create();
 bs.BeginUpdate();
 //description gnrale
 bs.AddStr(self.getHTMLProblemDescription());
 //valeurs propres
 bs.AddStr(self.getHTMLEigenValues());
 //caractrisation
 if ((self.PrmOp as TOpPrmFactMCA).FAlternativeReport = true)
  then bs.AddStr(self.getHTMLCaracActiveAlternative())
  else bs.AddStr(self.getHTMLCaracActive());
 //variables supplmentaires quali
 if (self.FLstCaracIllusDiscrete.getCount() > 0)
  then bs.AddStr('<P><H3>Supplementary variables (discrete)</H3>'+self.FLstCaracIllusDiscrete.getHTMLCoordVTestCos2Supp(seuilVTest)); 
 //variables supplmentaires quanti
 if (self.FLstCaracIllusContinuous.getCount() > 0)
  then bs.AddStr('<P><H3>Supplementary variables (continuous)</H3>'+self.FLstCaracIllusContinuous.getHTMLCoordVTestCos2SuppContinuous(seuilVTest));
 //factor scores
 bs.AddStr(self.getHTMLFactorScores());
 //end...
 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free();
end;

procedure TOpFactMCA.setCos2Ctr;
var poids: single;
    coef: single;
    nbAxis, attSize: integer;
    lst: TObjectList;
    i,j,k,j_cumul,j_dim: integer;
    att,newAtt, factor: TAttribute;
    dist: array of single;
    d: single;
begin
 nbAxis:= GenAtts.Count;
 coef:= 1.0*Dataset.Examples.Size/(1.0*Dataset.LstAtts[asInput].Count);
 attSize:= Dataset.LstAtts[asInput].Size;
 poids:= 100.0/(1.0 * attSize);
 //
 lst:= TObjectList.Create(false);
 
 //**************************************
 //** calcul des CTR ********************
 //**************************************
 for k:= 1 to nbAxis do
  begin
   factor:= GenAtts.Attribute[k-1];
   newAtt:= TAttContinue.Create(format('MCA_%d_CTR_%d',[(MLOwner as TMLCompFactMCA).Number,k]),attSize);
   lst.Add(newAtt);
   if (EigVal[k] > 0.0)
    then
     begin
      for i:= 1 to attSize do newAtt.cValue[i]:= poids * MATH.Power(factor.cValue[i],2.0) / EigVal[k];
     end
    else for i:= 1 to attSize do newAtt.cValue[i]:= 0.0;
  end;

 //**************************************
 //** calcul des COS *******************
 //**************************************

 //calcul de la distance au centre de gravit de chaque individu
 setLength(dist,succ(attSize));
 for i:= 1 to attSize do
  begin
   d:= 0.0;
   //l'objectif toujours est de ne pas construire explicitement de tableau disjonctif complet
   j_cumul:= 0;
   for j:= 1 to Dataset.LstAtts[asInput].Count do
    begin
     att:= Dataset.LstAtts[asInput].Attribute[pred(j)];
     j_dim:= j_cumul+att.dValue[i];//d'o l'intrt d'avoir des modalits numrots 1,2,3...
     //new -- 03/04/2008 -- protection +
     if (FVecDistValues^[j_dim] > 0.0) then
      d:= d + (1.0/FVecDistValues^[j_dim]);
     inc(j_cumul,att.nbValues);
    end;
   d:= coef * d - 1.0;
   dist[i]:= d;
  end;

 //calcul du cos2 proprement dit
 for k:= 1 to nbAxis do
  begin
   factor:= GenAtts.Attribute[k-1];
   newAtt:= TAttContinue.Create(format('MCA_%d_COS2_%d',[(MLOwner as TMLCompFactMCA).Number,k]),attSize);
   lst.Add(newAtt);
   for i:= 1 to attSize do
    begin
     if (dist[i] > 0.0)
      then newAtt.cValue[i]:=  MATH.Power(factor.cValue[i],2.0)/dist[i]
      else newAtt.cValue[i]:= 0.0;
    end;
  end;

 finalize(dist);

 //flush...
 for j:= 0 to pred(lst.Count) do
  GenAtts.Add(lst.Items[j] as TAttribute);
 //done...
 lst.Free();
end;

procedure TOpFactMCA.SetProjections;
var coef,v: double;
    i,j,k,j_cumul,j_dim: integer;
    s: double;
    nbAxis,attSize: integer;
    newAtt,att: TAttribute;
begin
 coef:= 1.0*sqrt(1.0*Dataset.Examples.Size/(1.0*Dataset.LstAtts[asInput].Count));
 (*
 if ((PrmOp as TOpPrmFactMCA).NbAxis<0)
  then nbAxis:= FNbEigens
  else nbAxis:= Min(FNbEigens,(PrmOp as TOpPrmFactMCA).NbAxis);
 *)
 nbAxis:= self.FNbAxis;
 attSize:= Dataset.LstAtts[asInput].Size;
 GenAtts.Clear;
 //pour chaque axe
 for k:= 1 to nbAxis do
  begin
   newAtt:= TAttContinue.Create(format('MCA_%d_Axis_%d',[(MLOwner as TMLCompFactMCA).Number,k]),attSize);
   //pour chaque individu
   for i:= 1 to attSize do
    begin
     s:= 0.0;
     //l'objectif toujours est de ne pas construire explicitement de tableau disjonctif complet
     j_cumul:= 0;
     for j:= 1 to Dataset.LstAtts[asInput].Count do
      begin
       att:= Dataset.LstAtts[asInput].Attribute[pred(j)];
       j_dim:= j_cumul+att.dValue[i];//d'o l'intrt d'avoir des modalits numrots 1,2,3...
       //new -- 03/04/2008 -- la marge peut tre nulle dans certains cas particuliers
       v:= 0.0;
       if (FVecDistValues^[j_dim] > 0.0) then
       //suite...
       v:= FEigVecs^[j_dim]^[k]/sqrt(FVecDistValues^[j_dim]);
       s:= s+v;
       inc(j_cumul,att.nbValues);
      end;
     s:= coef*s;
     newAtt.cValue[i]:= s;
    end;
   GenAtts.Add(newAtt);
  end;
end;

function TOpFactMCA.calcAdditional: boolean;
var coordValues: PMatrix;
    cosValues: PMatrix;
    contribValues: PMatrix;
    caracVar: TCaracFactorsVariable;
    caracModalite: TCaracModalite; 
    i,k,j,numDim: integer;
    att: TAttribute;
    n,testValue,p: double;
    tmpLst: TCaracFactorsLstVar;
begin
 result:= true;
 TRY
 //nombre d'axes retenus
 if ((PrmOp as TOpPrmFactMCA).NbAxis <= 0)
  then FNbAxis:= FNbEigens
  else FNbAxis:= Min(FNbEigens,(PrmOp as TOpPrmFactMCA).NbAxis);
 //effectif de calcul
 n:= 1.0*Dataset.Examples.Size;
 //nombre de variables
 p:= 1.0*Dataset.LstAtts[asInput].Count;
 //les indicateurs de qualit de reprsentation des modalits sur les axes
 coordValues:= self.BuildCoordValues();
 cosValues:= self.BuildCosValues(coordValues);
 contribValues:= self.BuildContribValues(coordValues);
 //les transfrer dans la structure de rsultats pour chaque variable - la liste n'est pas propritaire
 tmpLst:= TCaracFactorsLstVar.create(FNbAxis,false);
 numDim:= 0;
 for i:= 1 to self.Dataset.LstAtts[asInput].Count do
  begin
   att:= self.Dataset.LstAtts[asInput].Attribute[pred(i)];
   caracVar:= tmpLst.addVar(att,FNbAxis);
   for k:= 1 to att.nbValues do
    begin
     //pour avoir accs  la rfrence directe !!! sinon Delphi cre une copie locale
     caracVar.LstCarac[pred(k)].frequence:= FVecDistValues[numDim+k]/n;
     caracVar.LstCarac[pred(k)].mass:= FVecDistValues[numDim+k]/(n*p);//Saporta p. 231
     if (FVecDistValues[numDim+k] > 0.0) then  caracVar.LstCarac[pred(k)].disto:= n/FVecDistValues[numDim+k]-1.0;//Saporta p.232
     //pour simplifier l'criture - le champs factor est bien un pointeur
     caracModalite:= caracVar.get(k);
     for j:= 1 to FNbAxis do
      begin
       caracModalite.factors[pred(j)].coord:= coordValues[numDim+k][j];
       caracModalite.factors[pred(j)].cos2:= cosValues[numDim+k][j];
       caracModalite.factors[pred(j)].ctr:= 100.0*contribValues^[numDim+k]^[j];
       caracVar.CTRFactors[pred(j)]:= caracVar.CTRFactors[pred(j)] + caracModalite.factors[pred(j)].ctr;
       //calculer la valeur test
       try
       testValue:= 1.0*coordValues^[numDim+k]^[j]*SQRT(FVecDistValues^[numDim+k]*(n-1.0)/(n-FVecDistValues^[numDim+k]));
       Except
       testValue:= 0.0;
       end;
       //TraceLog.WriteToLogFile(Format('%0.5f',[testValue]));
       caracModalite.factors[pred(j)].vtest:= testValue;
      end;
    end;
   inc(numDim,att.nbValues);
  end;
 //dtruire les matrices intermdiaires
 delMatrix(coordValues,FDimAllValues,FNbEigens);
 delMatrix(cosValues,FDimAllValues,FNbEigens);
 delMatrix(contribValues,FDimAllValues,FNbEigens);
 //trier ou pas
 if ((self.PrmOp as TOpPrmFactMCA).FSortResult = true)
  //par dfaut, le seuil est fix par rapport au nombre de variables
  then self.FLstCaracActive:= tmpLst.sortOnCTR(true,(100.0 * SEUIL_FOR_CTR_SORTING)/(1.0*self.Dataset.LstAtts[asInput].Count))
  else self.FLstCaracActive:= tmpLst.recopie(true);
 tmpLst.Free();
 EXCEPT
 result:= false;
 END;
end;

function TOpFactMCA.calcEigen: boolean;
begin
 result:= TRUE;
 TRY
 //TraceLog.WriteToLogFile('[ACM] begin calc stat');
 self.CalcStats();
 //TraceLog.WriteToLogFile('[ACM] begin calc eigens');
 self.CalcEigens();
 EXCEPT
 result:= false;
 END;
end;

function TOpFactMCA.calcProjections: boolean;
begin
 result:= TRUE;
 TRY
 TraceLog.WriteToLogFile('[ACM] begin set projections');
 self.SetProjections();
 //new -- 05/02/2007 -- ventuellement, calculer les COS2 et CTR
 if ((self.PrmOp as TOpPrmFactMCA).Cos2Ctr = true)
  then self.setCos2Ctr();
 EXCEPT
 self.DelPrivateMatrix();
 result:= FALSE;
 END;
end;

constructor TOpFactMCA.CreateWithDataset(AOwner: TObject;
  prmDataset: TMLDataset);
begin
 inherited CreateWithDataset(AOwner, prmDataset);
 FLstStatDiscrete:= TLstCalcStatDesDiscrete.Create(NIL,NIL);
 FLstStatCrossTab:= TLstCalcStatDesCrossTab.Create(NIL,NIL);
end;

function TOpFactMCA.getInstanceFA: TClassOpFactAnalysis;
begin
 result:= TOpFactMCA;
end;

function TOpFactMCA.getHTMLFactorScores: string;
var bs: TBufString;
    s: string;
    i,j,k,j_dim: integer;
    att: TAttribute;
    value,coef: double;
begin
 coef:= 1.0*sqrt((1.0*Dataset.Examples.Size)/(1.0*Dataset.LstAtts[asInput].Count));

 TRY
 bs:= TBufString.Create();
 bs.BeginUpdate();
 //construction du tableau
 s:= '<BR><BR><font size="4">Factor Score Coefficients</font><br>';
 s:= s + '<font color="gray">Applied to indicator matrix i.e. columns are dummy variables</font>';
 //en-tte du tableau
 s:= s + HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=150>Attribute = Value</TH>';
 for k:= 1 to FNbAxis do
  s:= s + format('<TH width=80>Axis_%d</TH>',[k]);
 s:= s + '</TR>';
 bs.AddStr(s);
 //les coefficients
 j_dim:= 0;
 for j:= 1 to Dataset.LstAtts[asInput].Count do
  begin
   att:= Dataset.LstAtts[asInput].Attribute[pred(j)];
   for i:= 1 to att.nbValues do
    begin
     inc(j_dim);
     s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TD align=left><b>%s = %s</b></T>',[att.Name,att.LstValues.getDescription(i)]);
     bs.AddStr(s);
     for k:= 1 to FNbAxis do
      begin
       if (FVecDistValues^[j_dim] > 0)
        then value:= FEigVecs^[j_dim]^[k]/sqrt(FVecDistValues^[j_dim])
        else value:= 0.0;
       bs.AddStr(format('<TD align=right>%.7f</TD>',[coef*value]));
      end;
     bs.AddStr('</TR>');
    end;
  end;
 bs.AddStr('</table>');
 //renvoyer la sauce
 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free;
 EXCEPT
 result:= '';
 END;
 
end;

function TOpFactMCA.getHTMLEigenValues: string;
var s: string;
    tot,sum: double;
    i: integer;
    seuil: double;
    //new - 08/09/2013 - outils pour correction de Benzecri
    p: double;
    nbSupSeuil: integer;
    tabLambdaPrim,tabPercentPrim: array of double;
    SumLambdaPrim,sumPercent: double;
begin
 //seuil = moyenne des valeurs propres
 p:= 1.0*Dataset.LstAtts[asInput].Count;
 seuil:= 1.0/(1.0*p);
 //dtecter le nombre de facteurs dont le lambda est suprieur au seuil
 nbSupSeuil:= 0;
 for i:= 1 to FNbEigens do
  if (EigVal[i] >= seuil) then inc(nbSupSeuil);
 //calculer les lambda_prim
 setLength(tabLambdaPrim,nbSupSeuil);
 for i:= 0 to pred(nbSupSeuil) do
  tabLambdaPrim[i]:= SQR((p/(p-1.0))*(EigVal[succ(i)]-1/p));
 //somme des lambda_prim
 sumLambdaPrim:= Math.Sum(tabLambdaPrim);
 //lambda_prim en pourcentage
 setLength(tabPercentPrim,nbSupSeuil);
 if (sumLambdaPrim > 0)
  then for i:= 0 to pred(nbSupSeuil) do
        tabPercentPrim[i]:= tabLambdaPrim[i]/sumLambdaPrim;
 //construction du tableau
 s:= '<P><H3>Eigen values</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY+ format('<TH colspan=5>Original eigenvalues</TH><TH colspan=3 %s>Benzecri correction</TH></TR>',[HTML_BGCOLOR_DATA_BRUN]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH>Axis</TH><TH>Eigenvalue</TH><TH>% explained</TH><TH width="200">Histogram</TH><TH>% cumulated</TH>'+
       format('<TH>Eigenvalue%s</TH><TH>(%s)</TH><TH>cumsum (%s)</TH>',['&rsquo;','%','%'])+
       '</TR>';
 //inertie totale
 if (Dataset.LstAtts[asInput].Count>1)
  then tot:= 1.0*FDimAllValues/(1.0*Dataset.LstAtts[asInput].Count)-1.0
  else tot:= 0.0;
 sum:= 0.0;
 sumPercent:= 0.0;
 for i:= 1 to FNbEigens do
  begin
   sum:= sum+EigVal[i];
   if (tot > 0.0)
    then
     begin
     if (EigVal[i] >= seuil)
      then s:= s + HTML_TABLE_COLOR_DATA_LAVENDER
      else s:= s + HTML_TABLE_COLOR_DATA_GRAY;
     s:= s + format('<TH>%d</TH><TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD><TD align=right>%.2f %s</TD><TD>%s</TD><TD align=right>%.2f %s</TD>',
                      [i,EigVal[i],100.0*EigVal[i]/tot,'%',getHtmlHistogram(TRUNC(100.0*EigVal[i]/tot)),100.0*sum/tot,'%']);
     if (EigVal[i] >= seuil)
      then
       begin
        sumPercent:= sumPercent + tabPercentPrim[pred(i)];
        s:= s + format('<TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'<TD align=right>%.2f %s</TD><TD align=right>%.2f %s</TD>',[tabLambdaPrim[pred(i)],100*tabPercentPrim[pred(i)],'%',100*sumPercent,'%']);
       end
      else s:= s + '<TD colspan=3 align=center>-</TD>';
     s:= s +'</TR>';
     end;
  end;
 s:= s+'</table>';
 result:= s;
 //libration de la mmoire
 setLength(tabLambdaPrim,0);
 setLength(tabPercentPrim,0);
end;

(*
function TOpFactMCA.getHTMLCoordVTest: string;
var seuilVTest, SeuilCtr: double;
    s: string;
    bs: TBufString;
    i,j,k: integer;
    caracVar: TCaracFactorsVariable;
begin
 //seuils sur la valeur test et la contribution
 seuilVTest:= (PrmOp as TOpPrmFactMCA).ThresoldValue;
 seuilCtr:= 100.0 / (1.0 * self.FDimAllValues);//contribue plus que la moyenne des contributions

 //remplissage
 bs:= TBufString.Create();
 bs.BeginUpdate();

 //en-tte
 s:= '<P><H3>Coordinates and test-values (active variables)</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Values</TH>'+format('<TH colspan=%d>Coordinate</TH><TH colspan=%d>Test-Value</TH></TR>',[FNbAxis,FNbAxis]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=150>Attribute = Value</TH>';
 for j:= 1 to FNbAxis do s:= s + '<TD>'+format('coord_%d',[j])+'</TD>';
 for j:= 1 to FNbAxis do s:= s+'<TD>'+format('v-test_%d',[j])+'</TD>';
 s:= s + '</TR>';
 bs.AddStr(s);

 //listing pour chaque variable de caractrisation
 for i:= 1 to self.FLstCaracActive.getCount() do
  begin
   //variable de caractrisation
   caracVar:= self.FLstCaracActive.getCaracVar(pred(i));

   //liste des modalits par variable
   s:= HTML_TABLE_COLOR_DATA_GRAY+'<TD>';
   s:= s+HTML_HEADER_TABLE_RESULT;
   for j:= 1 to caracVar.LaVar.nbValues do
    s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD align=left width=150>%s = %s</TD></TR>',[caracVar.LaVar.Name,caracVar.LaVar.LstValues.getDescription(j)]);
   s:= s+'</table>';
   s:= s+'</TD>';

   //coordonnes pour chaque axe
   for j:= 1 to FNbAxis do
    begin
     s:= s+'<td>';
     s:= s+HTML_HEADER_TABLE_RESULT;

     for k:= 1 to caracVar.getCount() do //att.nbValues do
      begin
       s:= s+HTML_TABLE_COLOR_DATA_GRAY;

       //mise en vidence si valeur test de la modalit > seuil.vtest et contribution de la variable > seuil.ctr
       if ((ABS(caracVar.get(k)[pred(j)].vtest) > seuilVTest) and (caracVar.get(k)[pred(j)].ctr > seuilCtr))
        then s:= s+format('<td align=right width=50 %s>%.5f</td>',[HTML_BGCOLOR_LIGHT_RED,caracVar.get(k)[pred(j)].coord])
        else s:= s+format('<td align=right width=50 %s>%.5f</td>',[HTML_BGCOLOR_DATA_GRAY,caracVar.get(k)[pred(j)].coord]);

       //passage  la ligne suivante
       s:= s+'</tr>';
      end;
     s:= s+'</table>';
     //passage  l'axe suivant
     s:= s+'</td>';
    end;

   //valeur test -- doublon dans les calculs mais a simplifie la tche
   for j:= 1 to FNbAxis do
    begin
     s:= s+'<td>';
     s:= s+HTML_HEADER_TABLE_RESULT;

     //lister les lments  afficher : adresse -->> [numDim+k,j]
     for k:= 1 to caracVar.getCount() do //att.nbValues do
      begin
       s:= s + HTML_TABLE_COLOR_DATA_GRAY;
       //calculer la valeur test
       s:= s + format('<td align=right %s width=40>%.3f</td>',[HTML_BGCOLOR_DATA_GRAY,caracVar.get(k)[pred(j)].vtest]);
       //passage  la ligne suivante
       s:= s+'</tr>';
      end;
     s:= s+'</table>';
     //passage  l'axe suivant
     s:= s+'</td>';
    end;

   //fin d'une variable
   s:= s+'</tr>';
   bs.AddStr(s);
  end;
 bs.AddStr('</table>');


 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free();
end;
*)

(*
function TOpFactMCA.getHTMLCos2Ctr: string;
var s: string;
    bs: TBufString;
    i,j,k: integer;
    caracVar: TCaracFactorsVariable;
begin

 //remplissage
 bs:= TBufString.Create();
 bs.BeginUpdate();

 s:= '<P><H3>Cos and contributions (active variables)</H3>';

 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + '<TH>Values</TH>' +  format('<TH colspan=%d>Cos</TH>',[FNbAxis])+format('<TH colspan=%d>CTR (%s)</TH>',[FNbAxis,'%']) + '</TR>';

 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>Attribute = Value</TH>';

 for j:= 1 to FNbAxis do
  s:= s+'<TD>'+format('cos2_%d',[j])+'</TD>';

 for j:= 1 to FNbAxis do
  s:= s+'<TD>'+format('ctr_%d',[j])+'</TD>';  
  
 s:= s+'</TR>';

 bs.AddStr(s);

 //listing pour chaque variable de caractrisation
 for i:= 1 to self.FLstCaracActive.getCount() do
  begin
   //variable de caractrisation
   caracVar:= self.FLstCaracActive.getCaracVar(pred(i));

   //liste des modalits par variable
   s:= HTML_TABLE_COLOR_DATA_GRAY+'<TD>';
   s:= s+HTML_HEADER_TABLE_RESULT;

   for k:= 1 to caracVar.getCount() do
    s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD align=left width=150>%s = %s</TD></TR>',[caracVar.LaVar.Name,caracVar.LaVar.LstValues.getDescription(k)]);
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD align="center">-</TD></TR>';//pour le cumul
   s:= s+'</table>';
   s:= s+'</TD>';

   //cos pour chaque axe
   for j:= 1 to FNbAxis do
    begin
     s:= s+'<td>';
     s:= s+HTML_HEADER_TABLE_RESULT;

     for k:= 1 to caracVar.getCount() do
      begin
       s:= s+HTML_TABLE_COLOR_DATA_GRAY;
       //la coordonne
       s:= s+format('<td align=right width=40>%.3f</td>',[caracVar.get(k)[pred(j)].cos2]);
       //passage  la ligne suivante
       s:= s+'</tr>';
      end;

     //dernire colonne des marges des cos2
     if (j = FNbAxis)
      then s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<td align="center" %s>%s</td></tr>',[HTML_BGCOLOR_HEADER_GRAY,'Tot. ctr.'])
      else s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD align="center">-</td></tr>';
     s:= s + '</table>';
     //passage  l'axe suivant
     s:= s+'</td>';
    end;

   //contribution pour chaque axe
   for j:= 1 to FNbAxis do
    begin
     s:= s+'<td>';
     s:= s+HTML_HEADER_TABLE_RESULT;

     //lister les lments  afficher
     for k:= 1 to caracVar.getCount() do
      begin
       s:= s+HTML_TABLE_COLOR_DATA_GRAY;
       s:= s+format('<td align=right width=40>%.3f</td>',[caracVar.get(k)[pred(j)].ctr]);
       s:= s+'</tr>';
      end;

     s:= s+format('<td align=right width=40 %s>%.3f</td></tr>',[HTML_BGCOLOR_HEADER_GRAY,caracVar.CTRFactors[pred(j)]]);

     s:= s+'</table>';
     //passage  l'axe suivant
     s:= s+'</td>';
    end;

   //fin d'une variable
   s:= s+'</tr>';
   bs.AddStr(s);
  end;
 bs.AddStr('</table>');

 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free();

end;
*)

function TOpFactMCA.getHTMLCos2CtrActive: string;
begin
 result:= '<P><H3>Cos and contributions</H3>' + self.FLstCaracActive.getHTMLCos2Ctr();
end;

function TOpFactMCA.getHTMLCoordVTestActive: string;
var seuilVTest,seuilCos2: double;
begin
 //seuils sur la valeur test et la contribution
 seuilVTest:= (PrmOp as TOpPrmFactMCA).ThresoldValue;
 seuilCos2:= 1.0/(1.0*Self.FNbEigens);
 result:= '<P><H3>Coordinates and test-values</H3>' + self.FLstCaracActive.getHTMLCoordVTest(seuilVTest,seuilCos2);
end;

function TOpFactMCA.getHTMLCaracActiveAlternative(): string;
var seuilVTest,seuilCos2: double;
begin
 //seuils sur la valeur test et la contribution
 seuilVTest:= (PrmOp as TOpPrmFactMCA).ThresoldValue;
 seuilCos2:= 1.0/(1.0*Self.FNbEigens);
 result:= '<P><H3>Factors characterization (active variables)</H3>' + self.FLstCaracActive.getHTMLCoordVTestCos2Ctr(seuilVTest,seuilCos2);
end;

function TOpFactMCA.getHTMLCaracActive: string;
var s: string;
begin
 s:= '<P><H3>Factors characterization (active variables)</H3>';
 s:= s + self.getHTMLCoordVTestActive();
 s:= s + self.getHTMLCos2CtrActive();
 result:= s;
end;

function TOpFactMCA.calcPostProjections: boolean;
var j: integer;
    att: TAttribute;
begin
 result:= true;
 TRY
 FLstCaracIllusDiscrete:= TCaracFactorsLstVar.create(self.FNbAxis,true);
 FLstCaracIllusContinuous:= TCaracFactorsLstVar.create(self.FNbAxis,true);
 if (self.Dataset.LstAtts[asIllus].Count > 0)
  then
   begin
    //passer en revue les variables illustratives
    for j:= 0 to pred(self.Dataset.LstAtts[asIllus].Count) do
     begin
      att:= self.Dataset.LstAtts[asIllus].Attribute[j];
      if (att.isCategory(caContinue) = true)
       then FLstCaracIllusContinuous.addCarac(TCaracFactorsVariable.caracFromContinuous(att,Dataset.Examples,self))
       else FLstCaracIllusDiscrete.addCarac(TCaracFactorsVariable.caracFromDiscrete(att,Dataset.Examples,self));
     end;
   end;
 EXCEPT
 result:= false;
 END;
end;

function TOpFactMCA.getHTMLProblemDescription: string;
var s: string;
begin
 s:= '<P><H3>Problem statement</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH align="left"># of instances</TH><TD align=right>%d</TD>',[Dataset.Examples.Size])+'</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH align="left"># of variables</TH><TD align=right>%d</TD>',[Dataset.LstAtts[asInput].Count])+'</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH align="left"># of variable values</TH><TD align=right>%d</TD>',[FDimAllValues])+'</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH align="left">Max # of factors</TH><TD align=right>%d</TD>',[FNbEigens])+'</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH align="left"># of factors extracted</TH><TD align=right>%d</TD>',[FNbAxis])+'</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + format('<TH align="left">Total inertia</TH><TD align=right width="80">%.6f</TD>',[(1.0*FDimAllValues)/(1.0*Dataset.LstAtts[asInput].Count)-1.0])+'</TR>';
 s:= s + '</table>';
 result:= s;
end;

function TOpFactMCA.getNewForm: TForm;
begin
 result:= TfrmOpViewFactACM.CreateFromOperator(self);
end;

{ TCaracFactorsVariable }

class function TCaracFactorsVariable.caracFromContinuous(att: TAttribute;
  examples: TExamples; op: TOpFactMCA): TCaracFactorsVariable;
var stat: TCalcSDCorrelation;
    output: TCaracFactorsVariable;
    j: integer;
    axis: TAttribute;
begin
 output:= TCaracFactorsVariable.create(att,op.FNbAxis);
 for j:= 1 to op.FNbAxis do
  begin
   //axe de rfrence
   axis:= op.GenAtts.Attribute[pred(j)];
   //calculer la corrlation
   stat:= TCalcSDCorrelation.Create(axis,att,examples);
   //coordonnes = corrlation
   output.get(1).factors[pred(j)].coord:= stat.R;
   //cos^2 = corr^2
   output.get(1).factors[pred(j)].cos2:= SQR(stat.R);
   //v.test = t de Student
   //output.get(1).factors[pred(j)].vtest:= stat.T;
   //-> Z de Fisher standardis
   output.get(1).factors[pred(j)].vtest:= stat.ZStd;
   //vider
   stat.Free();
  end;
 result:= output;
end;

class function TCaracFactorsVariable.caracFromDiscrete(att: TAttribute;
  examples: TExamples; op: TOpFactMCA): TCaracFactorsVariable;
var stat: TCalcSDCondDescANOVA;
    output: TCaracFactorsVariable;
    k,j: integer;
    axis: TAttribute;
    n, frequence: double;
begin
 n:= 1.0 * examples.Size;
 output:= TCaracFactorsVariable.create(att,op.FNbAxis);
 for j:= 1 to op.FNbAxis do
  begin
   //axe de rfrence
   axis:= op.GenAtts.Attribute[pred(j)];
   //calcul des stats conditionnelles
   stat:= TCalcSDCondDescANOVA.Create(axis,att,examples);
   //recopie des valeurs
   for k:= 1 to output.getCount() do
    begin
     frequence:= (1.0*stat.StatCond[pred(k)].NbExamples) / (1.0*stat.StatGlobal.NbExamples);
     //coordonne
     if (op.FEigVal[k] > 0.0)
      then output.get(k).factors[pred(j)].coord:= stat.StatCond[pred(k)].Average / sqrt(op.FEigVal[j]);
     //valeur test
     if (stat.StatGlobal.VariancePop > 0) and (n > 1.0)
      then output.get(k).factors[pred(j)].vtest:= (stat.StatCond[pred(k)].Average - stat.StatGlobal.Average)
            /sqrt(stat.StatGlobal.VariancePop/(1.0*stat.StatCond[pred(k)].NbExamples)*(n - 1.0*stat.StatCond[pred(k)].NbExamples)/(n - 1.0));
     //cos - Tenenhaus p. 259
     if (frequence > 0)
      then output.get(k).factors[pred(j)].cos2:= SQR(output.get(k).factors[pred(j)].coord)/(1.0/frequence - 1.0);
    end;
   stat.Free();
  end;
 result:= output;
end;

constructor TCaracFactorsVariable.create(sourceVar: TAttribute; nbFactors: integer);
var i: integer;
begin
 inherited create();
 setLength(self.CTRFactors,nbFactors);
 self.LaVar:= sourceVar;
 if (sourceVar.isCategory(caContinue) = true)
  then setLength(self.LstCarac,1)
  else setLength(self.LstCarac,sourceVar.nbValues);
 for i:= 0 to pred(length(self.LstCarac)) do
  setLength(self.LstCarac[i].factors,nbFactors);
 self.NbAxis:= NbFactors;
end;

destructor TCaracFactorsVariable.destroy;
var i: integer;
begin
 for i:= 0 to pred(length(self.LstCarac)) do Finalize(self.LstCarac[i].factors);
 Finalize(self.LstCarac);
 inherited;
end;

function TCaracFactorsVariable.get(numModa: integer): TCaracModalite;
begin
 if (self.LaVar.isCategory(caContinue) = true)
  then result:= self.LstCarac[0]
  else result:= self.LstCarac[pred(numModa)];
end;


function TCaracFactorsVariable.getCount: integer;
begin
 result:= length(LstCarac);
end;

function TCaracFactorsVariable.getHTMLCoordVTest(seuilVTest, seuilCos2: double): string;
var s,color: string;
    j,k: integer;
begin

 //liste des modalits par variable
 s:= HTML_TABLE_COLOR_DATA_GRAY+'<TD>';

 s:= s + HTML_HEADER_TABLE_RESULT;
 for j:= 1 to self.LaVar.nbValues do
  s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD align=left width=150>%s = %s</TD></TR>',[self.LaVar.Name,self.LaVar.LstValues.getDescription(j)]);
 s:= s+'</table>';
 s:= s+'</TD>';

 //description globale des variables
 s:= s + self.getHTMLOverall();

 //coordonnes pour chaque axe
 for j:= 1 to NbAxis do
  begin
   //coordonnes et v.test
   s:= s+'<td>';
   s:= s+HTML_HEADER_TABLE_RESULT;
   for k:= 1 to self.getCount() do //att.nbValues do
    begin
     s:= s+HTML_TABLE_COLOR_DATA_GRAY;
     //mise en vidence si valeur test de la modalit > seuil.vtest et contribution de la variable > seuil.ctr
     if ((ABS(self.get(k).factors[pred(j)].vtest) >= seuilVTest) and (self.get(k).factors[pred(j)].cos2 >= seuilCos2) and (self.get(k).factors[pred(j)].ctr >= 100.0*self.get(k).mass))
      then
       begin
        //tenir compte du signe
        if (self.get(k).factors[pred(j)].coord < 0)
         then color:= FACT_COLOR_HIGHLIGHT[cthNegHigh]
         else color:= FACT_COLOR_HIGHLIGHT[cthPosHigh];
        s:= s+format('<td align=right width=50 %s>%.5f</td>',[color,self.get(k).factors[pred(j)].coord])
       end
      else s:= s+format('<td align=right width=50 %s>%.5f</td>',[HTML_BGCOLOR_DATA_GRAY,self.get(k).factors[pred(j)].coord]);
     //passage  la ligne suivante
     s:= s+'</tr>';
    end;
   s:= s+'</table>';
   //passage  l'axe suivant
   s:= s+'</td>';
  end;

 //valeur test -- doublon dans les calculs mais a simplifie la tche
 for j:= 1 to NbAxis do
  begin
   s:= s+'<td>';
   s:= s+HTML_HEADER_TABLE_RESULT;

   //lister les lments  afficher : adresse -->> [numDim+k,j]
   for k:= 1 to self.getCount() do //att.nbValues do
    begin
     s:= s + HTML_TABLE_COLOR_DATA_GRAY;
     //calculer la valeur test
     s:= s + format('<td align=right %s width=40>%.3f</td>',[HTML_BGCOLOR_DATA_GRAY,self.get(k).factors[pred(j)].vtest]);
     //passage  la ligne suivante
     s:= s+'</tr>';
    end;
   s:= s+'</table>';
   //passage  l'axe suivant
   s:= s+'</td>';
  end;

 //fin d'une variable
 s:= s+'</tr>';
 result:= s;
end;

function TCaracFactorsVariable.getHTMLCoordVTestCos2Ctr(seuilVTest,seuilCos2: double): string;
var s: string;
    j,k: integer;
    couleur: string;
    sumCos: double;
begin
 //liste des modalits par variable
 s:= '';

 for k:= 1 to self.getCount() do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY+format('<TD align=left width=150 %s>%s = %s</TD>',[HTML_BGCOLOR_DATA_BLUE,self.LaVar.Name,self.LaVar.LstValues.getDescription(k)]);

   //overall
   s:= s + format('<td align="right" %s><font color="teal">%.4f</font></td>',[HTML_COLOR_CODE_OVERALL,self.get(k).mass]);
   s:= s + format('<td align="right" %s><font color="teal">%.4f</font></td>',[HTML_COLOR_CODE_OVERALL,self.get(k).disto]);
   s:= s + format('<td align="right" %s><font color="teal">%.4f</font></td>',[HTML_COLOR_CODE_OVERALL,self.get(k).mass*self.get(k).disto]);

   //description de chaque axe
   sumCos:= 0.0;
   for j:= 1 to NbAxis do
    begin
       //mise en vidence
       if (abs(self.get(k).factors[pred(j)].vtest) >= seuilVTest) and (self.get(k).factors[pred(j)].cos2 >= seuilCos2) and (self.get(k).factors[pred(j)].ctr >= 100.0*self.get(k).mass)
        then
         begin
          if (self.get(k).factors[pred(j)].coord < 0)
           then couleur:= FACT_COLOR_HIGHLIGHT[cthNegHigh]
           else couleur:= FACT_COLOR_HIGHLIGHT[cthPosHigh];
         end
        else couleur:= HTML_BGCOLOR_DATA_GRAY;
       //la coordonne
       s:= s + format('<td align=right %s><font color="blue">%.5f</font></td>',[couleur,self.get(k).factors[pred(j)].coord]);
       //les autres informations
       sumCos:= sumCos + self.get(k).factors[pred(j)].cos2; 
       s:= s + format('<td align=right>%.3f</td><td align=right>%.3f &nbsp (%.2f)&nbsp;</td><td align=right>%.2f</td>',
           [self.get(k).factors[pred(j)].vtest,self.get(k).factors[pred(j)].cos2,sumCos,self.get(k).factors[pred(j)].ctr]);
    end;

   //passage  la modalit suivante
   s:= s+'</tr>';
  end;

 //description de la marge
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TD align="center">-</TD><TD align="center">-</TD><TD align="center">-</TD><TD align="center">-</TD>';
 for j:= 1 to NbAxis do
  begin
   s:= s + '<TD align="center">-</TD><TD align="center">-</TD><TD align="center">Tot.ctr.</TD>';
   s:= s + format('<TD align="right">%.2f</TD>',[self.CTRFactors[pred(j)]]);
  end;

 s:= s + '</TR>';

 result:= s;
end;

function TCaracFactorsVariable.getHTMLCoordVTestCos2(seuilVTest: double; colorBackTR: string = HTML_TABLE_COLOR_DATA_GRAY): string;
var s: string;
    j,k: integer;
    couleur: string;
begin
 //liste des modalits par variable
 s:= '';

 for k:= 1 to self.getCount() do
  begin
   s:= s + '<TR ' + colorBackTR + '>' + format('<TD align=left width=150 %s>%s = %s</TD>',[HTML_BGCOLOR_DATA_BLUE,self.LaVar.Name,self.LaVar.LstValues.getDescription(k)]);

   //description de chaque axe
   for j:= 1 to NbAxis do
    begin
       //mise en vidence
       if (abs(self.get(k).factors[pred(j)].vtest) >= seuilVTest)
        then
         begin
          if (self.get(k).factors[pred(j)].coord < 0)
           then couleur:= FACT_COLOR_HIGHLIGHT[cthNegHigh]
           else couleur:= FACT_COLOR_HIGHLIGHT[cthPosHigh];
         end
        else couleur:= colorBackTR;
       //la coordonne
       s:= s + format('<td align=right %s><font color="blue">%.5f</font></td>',[couleur,self.get(k).factors[pred(j)].coord]);
       //les autres informations
       s:= s + format('<td align=right>%.3f</td><td align=right>%.4f</td>',
           [self.get(k).factors[pred(j)].vtest,self.get(k).factors[pred(j)].cos2]);
    end;

   //passage  la modalit suivante
   s:= s+'</tr>';
  end;

 //s:= s + '</TR>';

 result:= s;
end;

function TCaracFactorsVariable.getHTMLCos2Ctr: string;
var s: string;
    j,k: integer;
    tabSum: array of double;
begin
 //liste des modalits par variable
 s:= HTML_TABLE_COLOR_DATA_GRAY+'<TD>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 for k:= 1 to self.getCount() do
  s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD align=left width=150>%s = %s</TD></TR>',[self.LaVar.Name,self.LaVar.LstValues.getDescription(k)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD align="center">-</TD></TR>';//pour le cumul
 s:= s+'</table>';
 s:= s+'</TD>';

 //pour les cumuls de cos2
 setLength(tabSum,self.getCount());

 //cos pour chaque axe
 for j:= 1 to NbAxis do
  begin
   s:= s+'<td>';
   s:= s+HTML_HEADER_TABLE_RESULT;

   for k:= 1 to self.getCount() do
    begin
     s:= s+HTML_TABLE_COLOR_DATA_GRAY;
     //la coordonne
     //s:= s+format('<td align=right width=40>%.4f</td>',[self.get(k).factors[pred(j)].cos2]);
     tabSum[pred(k)]:= tabSum[pred(k)] + self.get(k).factors[pred(j)].cos2;
     s:= s+format('<td align=right>%.3f %s (%.2f)</td>',[self.get(k).factors[pred(j)].cos2,'&nbsp;',tabSum[pred(k)]]);
     //passage  la ligne suivante
     s:= s+'</tr>';
    end;

   //dernire colonne des marges des cos2
   if (j = NbAxis)
    then s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<td align="center" %s>%s</td></tr>',[HTML_BGCOLOR_HEADER_GRAY,'Tot. ctr.'])
    else s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD align="center">-</td></tr>';
   s:= s + '</table>';
   //passage  l'axe suivant
   s:= s+'</td>';
  end;

 finalize(tabSum);

 //contribution pour chaque axe
 for j:= 1 to NbAxis do
  begin
   s:= s+'<td>';
   s:= s+HTML_HEADER_TABLE_RESULT;

   //lister les lments  afficher
   for k:= 1 to self.getCount() do
    begin
     s:= s+HTML_TABLE_COLOR_DATA_GRAY;
     s:= s+format('<td align=right width=40>%.3f</td>',[self.get(k).factors[pred(j)].ctr]);
     s:= s+'</tr>';
    end;

   s:= s+format('<td align=right width=40 %s>%.3f</td></tr>',[HTML_BGCOLOR_HEADER_GRAY,self.CTRFactors[pred(j)]]);

   s:= s+'</table>';
   //passage  l'axe suivant
   s:= s+'</td>';
  end;
 s:= s + '</TR>';

 result:= s;
end;

function TCaracFactorsVariable.getHTMLCoordVTestCos2Continuous(
  seuilVTest: double; colorBackTR: string): string;
var s: string;
    j: integer;
    couleur: string;
begin
 //liste des modalits par variable
 s:= '';


 s:= s + '<TR ' + colorBackTR + '>' + format('<TD align=left width=150 %s>%s</TD>',[HTML_BGCOLOR_DATA_BLUE,self.LaVar.Name]);

 //description de chaque axe
 for j:= 1 to NbAxis do
  begin
     //mise en vidence
     if (abs(self.get(1).factors[pred(j)].vtest) >= seuilVTest)
      then
         begin
          if (self.get(1).factors[pred(j)].coord < 0)
           then couleur:= FACT_COLOR_HIGHLIGHT[cthNegHigh]
           else couleur:= FACT_COLOR_HIGHLIGHT[cthPosHigh];
         end
      else couleur:= colorBackTR;
     //la coordonne
     s:= s + format('<td align=right %s><font color="blue">%.5f</font></td>',[couleur,self.get(1).factors[pred(j)].coord]);
     //les autres informations
     s:= s + format('<td align=right>%.3f</td><td align=right>%.4f</td>',
         [self.get(1).factors[pred(j)].vtest,self.get(1).factors[pred(j)].cos2]);
  end;

 //passage  ligne
 s:= s+'</tr>';

 result:= s;
end;

function TCaracFactorsVariable.getHTMLOverall: string;
var s: string;
    k: integer;
begin

 //mass
 s:= format('<td align="right" %s>',[HTML_BGCOLOR_DATA_GRAY]);
 s:= s+HTML_HEADER_TABLE_RESULT;
 for k:= 1 to self.getCount() do //att.nbValues do
  begin
   s:= s + HTML_TABLE_COLOR_CODE_OVERALL;
   s:= s + format('<td align="right" width="50"><font color="teal">%.4f</font></td>',[self.get(k).mass]);
   s:= s + '</TR>';
  end;
 s:= s+'</table>';
 s:= s+'</td>';

 //distance^2
 s:= s+format('<td align="right" %s>',[HTML_BGCOLOR_DATA_GRAY]);
 s:= s+HTML_HEADER_TABLE_RESULT;
 for k:= 1 to self.getCount() do //att.nbValues do
  begin
   s:= s+HTML_TABLE_COLOR_CODE_OVERALL;
   s:= s + format('<td align="right" width="50"><font color="teal">%.4f</font></td>',[self.get(k).disto]);
   s:= s + '</TR>';
  end;
 s:= s+'</table>';
 s:= s+'</td>';

 //inertie
 s:= s+format('<td align="right" %s>',[HTML_BGCOLOR_DATA_GRAY]);
 s:= s+HTML_HEADER_TABLE_RESULT;
 for k:= 1 to self.getCount() do //att.nbValues do
  begin
   s:= s + HTML_TABLE_COLOR_CODE_OVERALL;
   s:= s + format('<td align="right" width="50"><font color="teal">%.4f</font></td>',[self.get(k).mass*self.get(k).disto]);
   s:= s + '</TR>';
  end;
 s:= s+'</table>';
 s:= s+'</td>';

 result:= s;

end;

{ TCaracFactorsLstVar }

function TCaracFactorsLstVar.addVar(prmAtt: TAttribute; prmNbAxis: integer): TCaracFactorsVariable;
var tmp: TCaracFactorsVariable;
begin
 tmp:= TCaracFactorsVariable.create(prmAtt,prmNbAxis);
 self.LstCaracVar.Add(tmp);
 result:= tmp;
end;

procedure TCaracFactorsLstVar.addCarac(carac: TCaracFactorsVariable);
begin
 self.LstCaracVar.Add(carac);
end;

constructor TCaracFactorsLstVar.create(aNbAxis: integer; owner: boolean);
begin
 inherited create();
 self.LstCaracVar:= TObjectList.Create(owner);
 NbAxis:= aNbAxis;
end;

destructor TCaracFactorsLstVar.destroy;
begin
  self.LstCaracVar.Free();
  inherited;
end;

function TCaracFactorsLstVar.getCaracVar(index: integer): TCaracFactorsVariable;
begin
 result:= self.LstCaracVar.Items[index] as TCaracFactorsVariable;
end;

function TCaracFactorsLstVar.getCount: integer;
begin
 result:= self.LstCaracVar.Count;
end;

function TCaracFactorsLstVar.recopie(owner: boolean): TCaracFactorsLstVar;
var i: integer;
    output: TCaracFactorsLstVar;
begin
 output:= TCaracFactorsLstVar.create(self.NbAxis,owner);
 for i:= 0 to pred(self.getCount()) do
  output.addCarac(self.getCaracVar(i));
 result:= output;
end;

function TCaracFactorsLstVar.getHTMLCoordVTest(seuilVTest,seuilCos2: double): string;
var s: string;
    bs: TBufString;
    i,j: integer;
    caracVar: TCaracFactorsVariable;
begin

 //remplissage
 bs:= TBufString.Create();
 bs.BeginUpdate();

 //en-tte
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Values</TH>';
 s:= s + '<TH colspan=3>Overall</TH>';
 s:= s + format('<TH colspan=%d>Coordinate</TH><TH colspan=%d>Test-Value</TH></TR>',[NbAxis,NbAxis]);
 
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=150>Attribute = Value</TH>';
 s:= s + '<TH width=50>Mass</TH><TH width=50>Sq.Dist</TH><TH width=50>Inertia</TH>';
 
 for j:= 1 to NbAxis do s:= s + '<TD>'+format('coord_%d',[j])+'</TD>';
 for j:= 1 to NbAxis do s:= s+'<TD>'+format('v-test_%d',[j])+'</TD>';
 s:= s + '</TR>';
 bs.AddStr(s);

 //listing pour chaque variable de caractrisation
 for i:= 1 to self.getCount() do
  begin
   //variable de caractrisation
   caracVar:= self.getCaracVar(pred(i));
   //liste des modalits par variable
   bs.AddStr(caracVar.getHTMLCoordVTest(seuilVTest,seuilCos2));
  end;

 bs.AddStr('</table>');
 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free();
end;

function TCaracFactorsLstVar.getHTMLCos2Ctr: string;
var s: string;
    bs: TBufString;
    i,j: integer;
    caracVar: TCaracFactorsVariable;
begin

 //remplissage
 bs:= TBufString.Create();
 bs.BeginUpdate();

 s:= HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + '<TH>Values</TH>' +  format('<TH colspan=%d>Cos</TH>',[NbAxis])+format('<TH colspan=%d>CTR (%s)</TH>',[NbAxis,'%']) + '</TR>';

 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>Attribute = Value</TH>';

 for j:= 1 to NbAxis do
  begin
   s:= s+'<TD align=center width=80>'+format('cos2_%d (tot.)',[j])+'</TD>';
  end;

 for j:= 1 to NbAxis do
  s:= s+'<TD>'+format('ctr_%d',[j])+'</TD>';  
  
 s:= s+'</TR>';

 bs.AddStr(s);

 //listing pour chaque variable de caractrisation
 for i:= 1 to self.getCount() do
  begin
   //variable de caractrisation
   caracVar:= self.getCaracVar(pred(i));
   bs.AddStr(caracVar.getHTMLCos2Ctr());
  end;
 bs.AddStr('</table>');

 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free();

end;

function TCaracFactorsLstVar.getHTMLCoordVTestCos2Ctr(seuilVTest,seuilCos2: double): string;
var s: string;
    bs: TBufString;
    i,j: integer;
    caracVar: TCaracFactorsVariable;
begin

 //remplissage
 bs:= TBufString.Create();
 bs.BeginUpdate();

 //en-tte
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Values</TH>';
 s:= s + '<TH colspan=3>Overall</TH>';
 for j:= 1 to NbAxis do
  s:= s + format('<TH colspan = 4>Factor %d</TH>',[j]);
 s:= s + '</TR>';

 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=150>Attribute = Value</TH>';
 s:= s + '<TH width=50>Mass</TH><TH width=50>Sq.Dist</TH><TH width=50>Inertia</TH>';
 
 for j:= 1 to NbAxis do
  s:= s + '<TH width=50><font color="blue">coord</font></TH><TH width=50>v.test</TH><TH align=center width=80>cos2 &nbsp; (tot.)</TH><TH width=50>ctr (%)</TH>';
 s:= s + '</TR>';

 bs.AddStr(s);

 //listing pour chaque variable de caractrisation
 for i:= 1 to self.getCount() do
  begin
   //variable de caractrisation
   caracVar:= self.getCaracVar(pred(i));
   //liste des modalits par variable
   bs.AddStr(caracVar.getHTMLCoordVTestCos2Ctr(seuilVTest,seuilCos2));
  end;

 bs.AddStr('</table>');
 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free();
end;

//variable indiquant le facteur de rfrence
var idAxis: integer;

//comparaison de deux variables
function CompareCarac(Item1, Item2: Pointer): integer;
var p1, p2: TCaracFactorsVariable;
    value1, value2: double;
begin
 p1:= TCaracFactorsVariable(Item1);
 p2:= TCaracFactorsVariable(Item2);
 //pour l'instant, on s'en tient  la contribution sur le premier axe
 //c'est ici que a va se jouer
 value1:= p1.CTRFactors[idAxis];
 value2:= p2.CTRFactors[idAxis];
 //comparaison - tri croissant
 if (value1 < value2)
  then result:= -1
  else
   begin
    if (value1 > value2)
     then result:= +1
     else result:= 0;
   end;
end;

function TCaracFactorsLstVar.sortOnCTR(owner: boolean; seuil: double): TCaracFactorsLstVar;
var output: TCaracFactorsLstVar;

  //dpiler tant que non vide et suprieur  seuil
  procedure depiler(limit: double);
  var i,initNb: integer;
  begin
   initNb:= self.getCount();
   for i:= pred(initNb) downto 0 do
    begin
     if (self.getCaracVar(i).CTRFactors[idAxis] >= limit)
      then
       begin
        output.addCarac(self.getCaracVar(i));
        self.remove(i);
       end
      else break;
    end;
  end;

begin
 output:= TCaracFactorsLstVar.create(self.NbAxis,owner);
 idAxis:= 0;
 while (idAxis < self.NbAxis) and (self.getCount() > 1) do
  begin
   self.LstCaracVar.Sort(compareCarac);
   depiler(seuil);
   inc(idAxis);
  end;
 //s'il en reste, dpiler compltement
 if (self.getCount() > 0)
  then
   begin
    //prendre en rfrence le dernier facteur pour viter les dbordements...
    idAxis:= pred(self.NbAxis);
    depiler(-1);//-1 pour qu'il prenne tout ce qui reste
   end;
 //s'il en reste on recopie
 result:= output;
end;

procedure TCaracFactorsLstVar.remove(index: integer);
begin
 self.LstCaracVar.Delete(index);
end;

const
 COLORS: array[0..1] of string = (HTML_BGCOLOR_DATA_LAVENDER,HTML_BGCOLOR_DATA_GRAY);

function TCaracFactorsLstVar.getHTMLCoordVTestCos2Supp(
  seuilVTest: double): string;
var s: string;
    bs: TBufString;
    i,j: integer;
    caracVar: TCaracFactorsVariable;
begin

 //remplissage
 bs:= TBufString.Create();
 bs.BeginUpdate();

 //en-tte
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Values</TH>';
 for j:= 1 to NbAxis do
  s:= s + format('<TH colspan = 3>Factor %d</TH>',[j]);
 s:= s + '</TR>';

 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=150>Attribute = Value</TH>';
 for j:= 1 to NbAxis do
  s:= s + '<TH width=50><font color="blue">coord</font></TH><TH width=50>v.test</TH><TH width=50>cos2</TH>';
 s:= s + '</TR>';

 bs.AddStr(s);

 //listing pour chaque variable de caractrisation
 for i:= 1 to self.getCount() do
  begin
   //variable de caractrisation
   caracVar:= self.getCaracVar(pred(i));
   //liste des modalits par variable
   bs.AddStr(caracVar.getHTMLCoordVTestCos2(seuilVTest,COLORS[i mod 2]));  //astuce pour avoir l'alternance de couleurs
  end;

 bs.AddStr('</table>');
 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free();
end;

function TCaracFactorsLstVar.getHTMLCoordVTestCos2SuppContinuous(
  seuilVTest: double): string;
var s: string;
    bs: TBufString;
    i,j: integer;
    caracVar: TCaracFactorsVariable;
begin

 //remplissage
 bs:= TBufString.Create();
 bs.BeginUpdate();

 //en-tte
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>-</TH>';
 for j:= 1 to NbAxis do
  s:= s + format('<TH colspan = 3>Factor %d</TH>',[j]);
 s:= s + '</TR>';

 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=150>Attribute</TH>';
 for j:= 1 to NbAxis do
  s:= s + '<TH width=50><font color="blue">coord</font></TH><TH width=50>v.test</TH><TH width=50>cos2</TH>';
 s:= s + '</TR>';

 bs.AddStr(s);

 //listing pour chaque variable de caractrisation
 for i:= 1 to self.getCount() do
  begin
   //variable de caractrisation
   caracVar:= self.getCaracVar(pred(i));
   //liste des modalits par variable
   bs.AddStr(caracVar.getHTMLCoordVTestCos2Continuous(seuilVTest,COLORS[i mod 2]));  //astuce pour avoir l'alternance de couleurs
  end;

 bs.AddStr('</table>');
 bs.EndUpdate();
 result:= bs.BufS;
 bs.Free();
end;

initialization
 RegisterClass(TMLGenCompFactMCA);
end.
