unit UFrmOpViewFactACM;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, UFrmBaseOperator, Menus, ActnList, ComCtrls, ExtCtrls, TeEngine,
  Series, TeeProcs, Chart, StdCtrls, UCompFactAnalysisMCA;

type
  TfrmOpViewFactACM = class(TFrmBaseOperator)
    tabChart: TTabSheet;
    panelChart: TPanel;
    cmbX2: TComboBox;
    cmbX1: TComboBox;
    chart: TChart;
    SerieActive: TPointSeries;
    SerieIllus: TPointSeries;
    N1: TMenuItem;
    mnCopyChart: TMenuItem;
    N2: TMenuItem;
    mnZoomIn: TMenuItem;
    mnZoomOut: TMenuItem;
    N3: TMenuItem;
    mnJitter: TMenuItem;
    Panel1: TPanel;
    tbCTR: TTrackBar;
    Panel2: TPanel;
    tbCos2: TTrackBar;
    lbCtr: TLabel;
    lbCos2: TLabel;
    chkSupplementary: TCheckBox;
    procedure pgctrlFondChange(Sender: TObject);
    procedure cmbX1Change(Sender: TObject);
    procedure mnZoomInClick(Sender: TObject);
    procedure mnZoomOutClick(Sender: TObject);
    procedure mnCopyChartClick(Sender: TObject);
    procedure mnJitterClick(Sender: TObject);
    procedure tbCos2Change(Sender: TObject);
    procedure chkSupplementaryClick(Sender: TObject);
  private
    { Dclarations prives }
    //new - le min et max en abcisse et ordonne
    xMin, xMax, yMin, yMax: double;
    //rafrachir la vue
    procedure  refreshView(jitter: boolean = false);
    //insrer la valeur
    procedure  setValue(points: TPointSeries; x,y: double; desc: string; jitter: boolean = false);
    //reset les min et les max
    procedure  resetMinMax();
    //obtenir les min et max des abcisses et ordonnes
    procedure  getMinMaxAbsOrd();
  public
    { Dclarations publiques }
    //oprateur
    FOpMCA: TOpFactMCA;
    { Dclarations publiques }
    procedure   PrepareView(); override;
    function    getAdditionalInfoForReport(): string; override;
    procedure   getOperator(); virtual;    
  end;

var
  frmOpViewFactACM: TfrmOpViewFactACM;

implementation

{$R *.dfm}

uses
    UDatasetImplementation, Math;

const
    JITTER_FACTOR : double = 0.06;

{ TfrmOpViewFactACM }

function TfrmOpViewFactACM.getAdditionalInfoForReport: string;
var s: string;
begin
 //titre
 s:= '<H3>Multiple Correspondance Analysis</H3>';
 s:= s + self.addChartToReport(chart);
 result:= s;
end;

procedure TfrmOpViewFactACM.getOperator;
begin
 FOpMCA:= self.Operator as TOpFactMCA;
end;

procedure TfrmOpViewFactACM.PrepareView;
var a: integer;
begin
 inherited;
 //rcuprer l'oprateur
 self.getOperator();
 //prpaer les axes
 self.getMinMaxAbsOrd();
 //mj des combo
 for a:= 1 to FOpMCA.NbAxis do
  begin
   self.cmbX1.Items.Add(format('Axis_%d',[a]));
   self.cmbX2.Items.Add(format('Axis_%d',[a]));
  end;
 //slectionner
 self.cmbX1.ItemIndex:= 0;
 self.cmbX2.ItemIndex:= 0;
 //vrifier si var. illustratives prsentes
 if (FOpMCA.LstCarIllus.getCount() = 0)
  then
   begin
    self.chkSupplementary.State:= cbUnchecked;
    self.chkSupplementary.Enabled:= false; 
   end;
 //jitter -> false
 self.refreshView(false);
end;

procedure TfrmOpViewFactACM.refreshView(jitter: boolean);
var x1,x2: integer;
    i,k: integer;
    carac: TCaracFactorsVariable;
    seuilCtr, seuilCos2: double;
    ok: boolean;
begin
 //reser le min et le max si pas de jitter demand
 if (jitter = false) then self.resetMinMax();
 //numro des facteurs  afficher
 x1:= self.cmbX1.ItemIndex;
 x2:= self.cmbX2.ItemIndex;
 //rcuprer les seuils
 seuilCtr:= 1.0 * self.tbCTR.Position;
 seuilCos2:= (1.0 * self.tbCos2.Position)/100.0;
 //insrer les valeurs
 if (x1 >= 0) and (x2 >= 0)
  then
   begin
    //monter la srie des variables actives
    self.SerieActive.Clear();
    for i:= 0 to pred(FOpMCA.LstCarActive.getCount()) do
     begin
      carac:= FOpMCA.LstCarActive.getCaracVar(i);
      for k:= 1 to carac.getCount() do
       begin
        ok:= (carac.get(k).factors[x1].ctr >= seuilCtr) or (carac.get(k).factors[x2].ctr >= seuilCtr);
        ok:= ok AND ((carac.get(k).factors[x1].cos2 >= seuilCos2) or (carac.get(k).factors[x2].cos2 >= seuilCos2));
        if (ok = true)
         then self.setValue(SerieActive,carac.get(k).factors[x1].coord,carac.get(k).factors[x2].coord,format('%s=%s',[carac.LaVar.Name,carac.LaVar.LstValues.getDescription(k)]),jitter);
       end;
     end;
    //monter la srie des variables illustratives
    self.SerieIllus.Clear();
    if (self.chkSupplementary.Checked = true)
     then
      begin
       for i:= 0 to pred(FOpMCA.LstCarIllus.getCount()) do
        begin
         carac:= FOpMCA.LstCarIllus.getCaracVar(i);
         for k:= 1 to carac.getCount() do
          begin
           ok:= (carac.get(k).factors[x1].cos2 >= seuilCos2) or (carac.get(k).factors[x2].cos2 >= seuilCos2);
           if (ok = true)
            then self.setValue(SerieIllus,carac.get(k).factors[x1].coord,carac.get(k).factors[x2].coord,format('%s=%s',[carac.LaVar.Name,carac.LaVar.LstValues.getDescription(k)]),jitter);
          end;
        end;
      end;
   end;
end;

procedure TfrmOpViewFactACM.resetMinMax;
begin
 xMin:= Math.MaxDouble;
 yMin:= Math.MaxDouble;
 xMax:= -1.0 * xMin;
 yMax:= -1.0 * yMin;
end;

procedure TfrmOpViewFactACM.setValue(points: TPointSeries; x, y: double;
  desc: string; jitter: boolean);
var newx, newy: double;
begin
 if (jitter = false)
  then
   begin
    newx:= x;
    newy:= y;
    xMin:= Min(x,xMin);
    yMin:= Min(y,yMin);
    xMax:= Max(x,xMax);
    yMax:= Max(y,yMax);
   end
  else
   begin
    newx:= x + (0.5 - random)*JITTER_FACTOR*(self.xMax - self.xMin);
    newy:= y + (0.5 - random)*JITTER_FACTOR*(self.yMax - self.yMin);
   end;
 points.AddXY(newx,newy,desc);
end;

procedure TfrmOpViewFactACM.pgctrlFondChange(Sender: TObject);
begin
 if (self.pgctrlFond.ActivePage = self.TabChart)
  then
   begin
    self.actCopy.Enabled:= false;
    self.mnCopyChart.Enabled:= true;
    self.mnZoomIn.Enabled:= true;
    self.mnZoomOut.Enabled:= true;
    self.mnJitter.Enabled:= true;
   end
  else
   begin
    self.actCopy.Enabled:= true;
    self.mnCopyChart.Enabled:= false;
    self.mnZoomIn.Enabled:= false;
    self.mnZoomOut.Enabled:= false;
    self.mnJitter.Enabled:= false;
   end;
end;

procedure TfrmOpViewFactACM.cmbX1Change(Sender: TObject);
begin
 self.refreshView();
end;

procedure TfrmOpViewFactACM.mnZoomInClick(Sender: TObject);
begin
 self.SerieIllus.Marks.Font.Size:= self.SerieIllus.Marks.Font.Size + 1;
 self.SerieActive.Marks.Font.Size:= self.SerieActive.Marks.Font.Size + 1;
end;

procedure TfrmOpViewFactACM.mnZoomOutClick(Sender: TObject);
begin
 self.SerieIllus.Marks.Font.Size:= self.SerieIllus.Marks.Font.Size - 1;
 self.SerieActive.Marks.Font.Size:= self.SerieActive.Marks.Font.Size - 1;
end;

procedure TfrmOpViewFactACM.mnCopyChartClick(Sender: TObject);
begin
 self.chart.CopyToClipboardMetafile(false);
end;

procedure TfrmOpViewFactACM.mnJitterClick(Sender: TObject);
begin
 self.refreshView(true);
end;

procedure TfrmOpViewFactACM.tbCos2Change(Sender: TObject);
begin
 //afficher les nouveaux seuils
 self.lbCtr.Caption:= IntToStr(self.tbCTR.Position)+' %';
 self.lbCos2.Caption:= format('%.2f',[(1.0*self.tbCos2.Position)/100.0]);
 //mj de l'affichage
 self.refreshView();
end;

procedure TfrmOpViewFactACM.chkSupplementaryClick(Sender: TObject);
begin
 self.refreshView();
end;

procedure TfrmOpViewFactACM.getMinMaxAbsOrd;
var min, max: double;
    i,j,k: integer;
    carac: TCaracFactorsVariable;
begin
 min:= Math.MaxDouble;
 max:= -1.0*min;
 //actives
 for i:= 0 to pred(FOpMCA.LstCarActive.getCount()) do
  begin
   carac:= FOpMCA.LstCarActive.getCaracVar(i);
   for k:= 1 to carac.getCount() do
    for j:= 0 to pred(carac.NbAxis) do
     begin
      if (carac.get(k).factors[j].coord < min)
       then min:= carac.get(k).factors[j].coord;
      if (carac.get(k).factors[j].coord > max)
       then max:= carac.get(k).factors[j].coord;
     end;
  end;
 //illustratives
 for i:= 0 to pred(FOpMCA.LstCarIllus.getCount()) do
  begin
   carac:= FOpMCA.LstCarIllus.getCaracVar(i);
   for k:= 1 to carac.getCount() do
    for j:= 0 to pred(carac.NbAxis) do
     begin
      if (carac.get(k).factors[j].coord < min)
       then min:= carac.get(k).factors[j].coord;
      if (carac.get(k).factors[j].coord > max)
       then max:= carac.get(k).factors[j].coord;
     end;
  end;
 //mj des bornes des axes
 self.chart.LeftAxis.Automatic:= false;
 self.chart.LeftAxis.Minimum:= min;
 self.chart.LeftAxis.Maximum:= max;
 self.chart.BottomAxis.Automatic:= false;
 self.chart.BottomAxis.Minimum:= min;
 self.chart.BottomAxis.Maximum:= max;
end;

end.
