(*********************************************************************)
(* UCompFactAnalysisAFDM.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(*********************************************************************)

{
@abstract(Analyse Factorielle de Donnes Mixtes)
@author(Ricco)
@created(23/04/2012)

Analyse factorielle pour tableau de donnes contenant  la fois des descripteurs qunatitatifs et qualitatifs.

Source : J. Pags, "Analyse Factorielle de donnes mixtes", Revue de Statistique Applique, Tome 52, n 4, pp. 93-111, 2004.
http://archive.numdam.org/ARCHIVE/RSA/RSA_2004__52_4/RSA_2004__52_4_93_0/RSA_2004__52_4_93_0.pdf

Cale sur l'implmentation de l'AFDM dans le package FactoMineR.
Mais les sorties sont organises de manire diffrente pour une meilleure lisibilit.
}
unit UCompFactAnalysisAFDM;

interface

USES
        Forms,Classes,Contnrs,
        IniFiles,
        Matrices,
        UCalcStatDes,
        UCompDefinition,
        UCompFADefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation;

TYPE
        {gnrateur}
        TMLGenCompFactAFDM = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant}
        TMLCompFactAFDM = class(TMLCompFactAnalysis)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {attribut transform}
        TAttTrans = class
        source: TAttribute;
        center, scale: double;
        constructor create(att: TAttribute; mean, std: double);
        function    getValue(i: integer): double; virtual; abstract;
        end;

        {attribut transform continu}
        TAttTransCont = class(TAttTrans)
        function    getValue(i: integer): double; override;
        end;

        {attribut transform discret}
        TAttTransDisc = class(TAttTrans)
        refValue: TTypeDiscrete;
        constructor create(att: TAttribute; mean, std: double; ref: TTypeDiscrete);
        function    getValue(i: integer): double; override;
        end;

        {oprateur AFDM, on travaille sur les INPUT}
        TOpFactAFDM = class(TOpFactAnalysis)
        private
        //trace
        FTrace: double;
        //liste des attributs pour les calculs (nombre = FDimension, sinon il y a un souci)
        FLstVar: TObjectList;
        //sous-listes des variables
        FLstVarCont, FLstVarDisc: TObjectList;
        //statistiques descriptives - mixtes
        FStats: TLstCalcStatDes;
        //matrice des vecteurs propres
        FEgVec: PMatrix;
        //nombre d'axes  traiter
        FNbAxis: integer;
        //dimension des matrices (incluant les indicatrices)
        FDimension: integer;
        //nombre maximum d'axes (FDimension - Nb. var. catgorielles)
        FMaxDimAxis: integer;
        //nombre de variables (Count = NbVarCont + NbVarDisc)
        FNbVar, FNbVarCont, FNbVarDisc: integer;
        //effectif
        N: double;
        //Seuils pour le highlighting
        FSeuilEigenVal, FSeuilCos2, FSeuilContrib: double;
        //tableau des COORD - toutes les variables
        FCOORD: array of array of double;
        //tableau pour variables quanti - corrlations
        FCORR: array of array of double;
        //tableau pour variables quali
        FDISC: TObjectList;
        //calcul des statistiques
        function calcStats(): boolean;
        //cration des variables centres et rduites, et des indicatrices pondres
        function createVars(): boolean;
        //calcul effectif des val./vec. propres
        function calcCorEigen(): boolean;
        //tableau des valeurs propres
        function getHTMLEigen(): string;
        //tableau des carrs des corrlations (avec communalities)
        function getHTMLSquaredCorrelation(): string;
        //tableau des corrlations
        function getHTMLCorrelation(): string;
        //tableau de caractrisation des variables discrtes
        function getHTMLCaracDisc(): string;
        //description des vecteur propres
        function getHTMLFactorScores(): string;
        //mettre en surbrillance ( partir de coord) ? (i: variable ; j: axe)
        function isHighlighted(i,j: integer): boolean;
        //effectuer les projections sur les axes factoriels
        procedure   SetProjections(atts: TLstAttributes; coefs: PMatrix; nbFactors: integer; prefix: string = 'AFDM'; number: integer = 1);
        //cos2 et CTR
        procedure   setCos2Ctr(atts: TLstAttributes; nbFactors: integer; prefix: string = 'AFDM'; number : integer = 1);
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    CheckAttributes(): boolean; override;
        function    getDimEigVal(): integer; override;
        procedure   DelPrivateMatrix(); override;
        public
        //calculs prparatoires, puis de la matrice de corr, puis des valeurs/vecteurs propres
        function    calcEigen(): boolean; override;
        function    calcAdditional(): boolean; override;
        function    calcProjections(): boolean; override;
        function    calcPostProjections(): boolean; override;
        function    getInstanceFA(): TClassOpFactAnalysis; override;
        //affichage
        function    getHTMLResultsSummary(): string; override;
        end;

        TOpPrmFactAFDM = class(TOpPrmFactAnalysis)
        //calculer les COS2 et CTR des individus (false par dfaut)
        Cos2Ctr: boolean;
        //nombre d'axes  gnrer (si <= 0, tous)
        NbAxis: integer;
        protected
        function    CreateDlgParameters(): TForm; override;
        procedure   SetDefaultParameters(); override;
        public
        function    getHTMLParameters(): string; override;
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   getFactSettings(prmSource: TOpPrmFactAnalysis); override;
        end;


implementation

uses
    Sysutils, Math, UDatasetExamples, UCalcMatrixAdditionalFunctions,
    UStringAddBuffered, UConstConfiguration, UCalcStatDesCorrelation,
    UCalcStatDesConditionnalDesc, UDlgOpPrmFactAFDM;

const
    MIN_SCALE : double = 1e-12;

type
    //caractrisation var. discrte
    TCaracDisc = class
    EigVal: double;
    Stat: TCalcSDCondDescANOVA;
    TabCTR: array of double;
    TabVTest: array of double;//valeurs test - 14/09/2013
    constructor create(paramStat: TCalcSDCondDescANOVA; paramEig: double);
    destructor  destroy(); override;
    end;

    TCollCaracDisc = class
    LstCD: TObjectList;
    constructor create();
    destructor  destroy(); override;
    procedure   add(prmCD: TCaracDisc);
    function    getCD(k: integer): TCaracDisc;
    end;

{ TMLGenCompFactAFDM }

procedure TMLGenCompFactAFDM.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactAFDM.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactAFDM; 
end;

{ TMLCompFactAFDM }

function TMLCompFactAFDM.getClassOperator: TClassOperator;
begin
 result:= TOpFactAFDM;
end;

{ TOpFactAFDM }

function TOpFactAFDM.calcAdditional: boolean;
begin
 //rien en particulier
 result:= true;
end;

function TOpFactAFDM.calcCorEigen: boolean;
var cor: PMatrix;
    i,j,k,id: integer;
    sum: double;
    attI, attJ: TAttTrans;
begin
 result:= true;
 try
 //calcul de la matrice des corrlations
 DimMatrix(cor,self.FDimension,self.FDimension);
 for i:= 1 to FDimension do
  begin
   attI:= FLstVar[pred(i)] as TAttTrans;
   //triangle sup. (incluant la diagonale)
   for j:= i to FDimension do
    begin
     attJ:= FLstVar[pred(j)] as TAttTrans;
     sum:= 0.0;
     for k:= 1 to Dataset.Examples.Size do
      begin
       id:= Dataset.Examples.Number[k];
       sum:= sum + attI.getValue(id)*attJ.getValue(id);
      end;
     sum:= sum/self.N;
     cor^[i]^[j]:= sum;
     //la matrice est symtrique
     cor^[j]^[i]:= sum;
    end;
  end;
 //calculer la trace
 FTrace:= 0.0;
 for i:= 1 to FDimension do
  FTrace:= FTrace + cor^[i]^[i];
 //valeurs propres
 GetEigensFromSymetricMatrix(cor,FDimension,FEigVal,FEgVec);
 //librer la matrice de corrlation
 delMatrix(cor,FDimension,FDimension);
 except
 result:= false;
 end;
end;

function TOpFactAFDM.calcEigen: boolean;
var ok: boolean;
begin
 TRY
 N:= 1.0*self.Dataset.Examples.Size;
 self.DelPrivateMatrix();
 ok:= self.calcStats();
 ok:= ok and self.createVars();
 ok:= ok and self.calcCorEigen();
 result:= ok;
 EXCEPT
 result:= false;
 END;
end;

function TOpFactAFDM.calcPostProjections: boolean;
var k,j,iC: integer;
    statCor: TCalcSDCorrelation;
    statAnova: TCalcSDCondDescANOVA;
    att: TAttribute;
    oneDISC: TCollCaracDisc;
begin
 result:= true;
 try
 //COORD, CORR
 setLength(FCOORD,FNbVar,FNbAxis);
 setLength(FCORR,FNbVarCont,FNbAxis);
 FDISC:= TObjectList.Create(TRUE);
 //pour chaque axe
 for k:= 0 to pred(FNbAxis) do
  begin
   iC:= 0;
   oneDISC:= TCollCaracDisc.create();
   FDISC.Add(oneDISC);
   //pour chaque variable
   for j:= 0 to pred(Dataset.LstAtts[asInput].Count) do
    begin
     att:= Dataset.LstAtts[asInput].Attribute[j];
     if (att.isCategory(caContinue) = TRUE)
      then
       //travail sur la corrlation
       begin
        statCor:= TCalcSDCorrelation.Create(self.GenAtts.Attribute[k],att,Dataset.Examples);
        FCOORD[j,k]:= math.Power(statCor.R,2.0);
        FCORR[iC,k]:= statCor.R;
        inc(iC);
        statCor.Free();
       end
      else
       //travail sur l'anova
       begin
        statAnova:= TCalcSDCondDescANOVA.Create(self.GenAtts.Attribute[k],att,Dataset.Examples);
        if (statAnova.TSS > 0.0)
         then FCOORD[j,k]:= statAnova.BSS / statAnova.TSS
         else FCOORD[j,k]:= 0.0;
        //statAnova.Free();
        oneDISC.add(TCaracDisc.create(statAnova,FEigVal^[succ(k)]));
       end;
    end;
  end;
 except
 result:= false;
 end;
end;

function TOpFactAFDM.calcProjections: boolean;
begin
 result:= true;
 try
 self.SetProjections(self.GenAtts,FEgVec,FNbAxis,'AFDM',(MLOwner as TMLComponent).Number);
 if ((self.PrmOp as TOpPrmFactAFDM).Cos2Ctr = true)
  then self.setCos2Ctr(self.GenAtts,FNbAxis,'AFDM',(MLOwner as TMLComponent).Number);
 except
 result:= false;
 end;
end;

function TOpFactAFDM.calcStats: boolean;
begin
 result:= true;
 try
 //calcul  la vole de toutes les stats. desc.
 FStats:= TLstCalcStatDes.Create(Dataset.LstAtts[asInput],Dataset.Examples);
 except
 result:= false;
 end;
end;

function TOpFactAFDM.CheckAttributes: boolean;
begin
 //il faut au moins 2 variables...
 result:= (self.Dataset.LstAtts[asInput].Count > 1);
end;

function TOpFactAFDM.createVars: boolean;
var j,k: integer;
    att: TAttTrans;
    attVar: TAttribute;
    statCont: TCalcStatDesContinuous;
    statDisc: TCalcStatDesDiscrete;
    pk: double;
begin
 result:= true;
 try
 FNbVarCont:= 0;
 FNbVarDisc:= 0;
 FLstVar:= TObjectList.Create(true);
 FLstVarCont:= TObjectList.Create(FALSE);
 FLstVarDisc:= TObjectList.Create(FALSE);
 for j:= 0 to pred(Dataset.LstAtts[asInput].Count) do
  begin
   attVar:= Dataset.LstAtts[asInput].Attribute[j];
   if (attVar.isCategory(caContinue) = true)
    then
     begin
      inc(FNbVarCont);
      FLstVarCont.Add(attVar);
      statCont:= FStats.Stat(j) as TCalcStatDesContinuous;
      att:= TAttTransCont.create(attVar,statCont.Average,statCont.StdDevPop);
      FLstVar.Add(att);
     end
    else
     begin
      inc(FNbVarDisc);
      FLstVarDisc.Add(attVar);
      statDisc:= FStats.Stat(j) as TCalcStatDesDiscrete;
      for k:= 1 to statDisc.TabFreq.Size do
       begin
        pk:= statDisc.TabFreq.Frequence[k];
        if (pk > 0.0)
         then att:= TAttTransDisc.create(attVar,pk,sqrt(pk),k)
         else att:= TAttTransDisc.create(attVar,0.0,0.0,k);//pour prvenir la sqrt(0.0)
        FLstVar.Add(att);
       end;
     end;
  end;
 FNbVar:= FNbVarCont + FNbVarDisc;
 FDimension:= FLstVar.Count;
 FMaxDimAxis:= FDimension - FNbVarDisc;
 //dtecter le nombre d'axes  produire
 if ((self.PrmOp as TOpPrmFactAFDM).NbAxis < 1) then FNbAxis:= FMaxDimAxis else FNbAxis:= Math.MIN((self.PrmOp as TOpPrmFactAFDM).NbAxis,FMaxDimAxis);
 //les seuils pour l'affichage
 FSeuilEigenVal:= 1.0+1.65*SQRT((-1.0+FMaxDimAxis)/(-1.0+self.N));//Saporta et al. (page 172) - "1.65" pour IC  90%
 FSeuilCos2:= 1.0/(1.0 * FMaxDimAxis);
 FSeuilContrib:= 1.0/(1.0 * FNbVar);
 except
 result:= false;
 end;
end;

procedure TOpFactAFDM.DelPrivateMatrix;
begin
 inherited;
 if assigned(FEgVec) then delMatrix(FEgVec,FDimension, FDimension);
 if assigned(FStats) then FreeAndNil(FStats);
 if assigned(FLstVar) then FreeAndNil(FLstVar);
 if assigned(FLstVarCont) then FreeAndNil(FLstVarCont);
 if assigned(FLstVarDisc) then FreeAndNil(FLstVarDisc);
 if assigned(FCOORD) then Finalize(FCOORD);
 if assigned(FCORR) then Finalize(FCORR);
 if assigned(FDISC) then FreeAndNil(FDISC);
end;

function TOpFactAFDM.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactAFDM;
end;

function TOpFactAFDM.getDimEigVal: integer;
begin
 result:= FDimension;
end;

function TOpFactAFDM.getHTMLCaracDisc: string;
var buf: TBufString;
    i,j,k,iD: integer;
    att: TAttribute;
    strColor: string;
    okHighLight: boolean;
    carac: TCaracDisc;
    seuilCTR: double;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Attribute</TH>');
 for j:= 1 to FNbAxis do
  buf.AddStr(format('<TH align=center colspan=3>Axis_%d</TH>',[j]));
 buf.AddStr('</TR>');
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>-</TH>');
 for j:= 1 to FNbAxis do
  buf.AddStr('<TH align=center>Mean</TH><TH align=center>CTR (%)</TH><TH align=center>v.test</TH>');
 buf.AddStr('</TR>');
 iD:= -1;
 //les valeurs
 for i:= 1 to FNbVar do
  begin
   att:= Dataset.LstAtts[asInput].Attribute[pred(i)];
   if (att.isCategory(caDiscrete) = true)
    then
     begin
      inc(iD);
      buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH width=100>%s</TH>',[att.Name]));
      //liste des modalits
      buf.AddStr('<TD width=100>');
      buf.AddStr(HTML_HEADER_TABLE_RESULT);
      for k:= 1 to att.nbValues do  buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+'<TD>'+att.LstValues.getDescription(k)+'</TD></TR>');
      buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TD width=100>Tot.</TD></TR>');
      buf.AddStr('</table>');
      buf.AddStr('</TD>');
      //les valeurs
      for j:= 1 to FNbAxis do
       begin
        //mettre en valeur...les valeurs ?
        okHighlight:= self.isHighlighted(i,j);
        //variable i, axe j
        carac:= (FDISC.Items[pred(j)] as TCollCaracDisc).getCD(iD);
        seuilCTR:= carac.TabCTR[0]/(1.0*att.nbValues);
        //tableau des moyennes conditionnelles
        buf.AddStr('<TD width=60>');
        buf.AddStr(HTML_HEADER_TABLE_RESULT);
        for k:= 1 to att.nbValues do
         begin
          strColor:= 'black';
          if (okHighlight = true) and ((att.nbValues = 2) or (carac.TabCTR[k] > seuilCTR))
           then strColor:= 'blue';
          buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TD align=right><font color="%s">%.4f</font></TD></TR>',[strColor,carac.Stat.StatCond[pred(k)].Average]));
         end;
        buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TD width=60 align=center>-</TD></TR>');
        buf.AddStr('</table>');
        buf.AddStr('</TD>');
        //table des contributions
        buf.AddStr('<TD width=60>');
        buf.AddStr(HTML_HEADER_TABLE_RESULT);
        for k:= 1 to att.nbValues do
         begin
          strColor:= 'black';
          if (okHighlight = true) and ((att.nbValues = 2) or (carac.TabCTR[k] > seuilCTR))
           then strColor:= 'blue';
          buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TD align=right><font color="%s">%.2f</font></TD></TR>',[strColor,carac.TabCTR[k]]));
         end;
        buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+format('<TD width=50 align=right>%.2f</TD></TR>',[carac.TabCTR[0]]));
        buf.AddStr('</table>');
        buf.AddStr('</TD>');
        //new - table des valeurs test - 14/09/2013
        buf.AddStr('<TD width=60>');
        buf.AddStr(HTML_HEADER_TABLE_RESULT);
        for k:= 1 to att.nbValues do
         begin
          strColor:= 'black';
          if (okHighlight = true) and ((att.nbValues = 2) or (carac.TabCTR[k] > seuilCTR))
           then strColor:= 'blue';
          buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TD align=right><font color="%s">%.3f</font></TD></TR>',[strColor,carac.TabVTest[k]]));
         end;
        buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+format('<TD width=50 align=center>%s</TD></TR>',['-']));
        buf.AddStr('</table>');
        buf.AddStr('</TD>');
       end;
      buf.AddStr('</TR>');
     end;
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TOpFactAFDM.getHTMLCorrelation: string;
var buf: TBufString;
    i,iC,j: integer;
    att: TAttribute;
    strColor: string;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Attribute</TH>');
 for j:= 1 to FNbAxis do
  buf.AddStr(format('<TH align=center>Axis_%d</TH>',[j]));
 buf.AddStr('</TR>');
 iC:= -1;
 //les valeurs
 for i:= 1 to FNbVar do
  begin
   att:= Dataset.LstAtts[asInput].Attribute[pred(i)];
   if (att.isCategory(caContinue) = true)
    then
     begin
      inc(iC);
      buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH width=100>%s</TH>',[att.Name]));
      for j:= 1 to FNbAxis do
       begin
        //mettre en valeur si...
        if (self.isHighlighted(i,j) = true)
         then strColor:= 'blue'
         else strColor:= 'black';
        buf.AddStr(format('<TD align=right width=60><font color="%s">%.6f</font></TD></TD>',[strColor,FCORR[iC,pred(j)]]));
       end;
      buf.AddStr('</TR>');
     end;
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;


function TOpFactAFDM.getHTMLEigen: string;
var s: string;
    sum: double;
    tot: double;
    i: integer;
    buf: TBufString;
    strColor: string;
begin
 //buffriser l'ajout -- 23/06/2004
 buf:= TBufString.Create();
 buf.BeginUpdate();
 s:= format('<H5>Matrix trace = %.2f</H5>',[FTrace]);
 s:= s+HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH>Axis</TH><TH>Eigen value</TH><TH>% explained</TH><TH width="200">Histogram</TH><TH>% cumulated</TH></TR>';
 buf.AddStr(s);
 tot:= 1.0 * FTrace;
 //
 sum:= 0.0;
 for i:= 1 to self.FMaxDimAxis do
  begin
   sum:= sum+EigVal[i];
   if (tot > 0)
    then
     begin
      if (EigVal[i] > FSeuilEigenVal) then strColor:= 'blue' else strColor:= 'black';
      s:= HTML_TABLE_COLOR_DATA_GRAY+
         format('<TH>%d</TH><TD align=right><font color="%s">'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</font></TD><TD align=right>%.2f%s</TD><TD>%s</TD><TD align=right>%.2f%s</TD></TR>',
                [i,strColor,EigVal[i],100.0*EigVal[i]/tot,'%',getHtmlHistogram(TRUNC(100.0*EigVal[i]/tot)),100.0*sum/tot,'%'])+'</TR>';
      buf.AddStr(s);
     end;
  end;
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY + format('<TH>Tot.</TH><TD align=right>%.6f</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD></TR>',[sum]));
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TOpFactAFDM.getHTMLFactorScores: string;
var i,j: integer;
    buf: TBufString;
    att: TAttTrans;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Attribute</TH><TH width=60>Center</TH><TH width=60>Scale</TH>');
 for j:= 1 to FNbAxis do
  buf.AddStr(format('<TH align=center width=60>Axis_%d</TH>',[j]));
 buf.AddStr('</TR>');
 for i:= 0 to FLstVar.Count-1 do
  begin
   att:= FLstVar.Items[i] as TAttTrans;
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY);
   if (att.source.isCategory(caContinue) = true)
    then buf.AddStr(format('<TD align=center>%s</TD>',[att.Source.Name]))
    else buf.AddStr(format('<TD align=center>%s = %s</TD>',[att.Source.Name,att.Source.LstValues.getDescription((att as TAttTransDisc).refValue)]));
   //paramtres de centrage-rduction
   buf.AddStr(format('<TD align=right %s>%.6f</TD><TD align=right %s>%.6f</TD>',[HTML_BGCOLOR_LIGHT_GREEN,att.center,HTML_BGCOLOR_LIGHT_GREEN,att.scale]));
   //axes
   for j:= 1 to FNbAxis do
    buf.AddStr(format('<TD align=right>%.6f</TD>',[FEgVec^[i+1]^[j]]));
   buf.AddStr('</TR>');
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TOpFactAFDM.getHTMLResultsSummary: string;
var s: string;
begin
 s:= '<P><H3>Eigen values</H3>';
 s:= s + self.getHTMLEigen();
 s:= s + '<BR><P><H3>Squared Correlation (Communalities)</H3>';
 s:= s + self.getHTMLSquaredCorrelation();
 s:= s + '<BR><P><H3>Continuous Attributes - Correlation (Factor Loadings)</H3>';
 s:= s + self.getHTMLCorrelation();
 s:= s + '<BR><P><H3>Discrete Attributes - Conditional means and contributions</H3>';
 s:= s + self.getHTMLCaracDisc();
 s:= s + '<BR><P><H3>Eigen vectors - Factor Scores</H3>';
 s:= s + self.getHTMLFactorScores();
 result:= s;
end;

function TOpFactAFDM.getHTMLSquaredCorrelation: string;
var buf: TBufString;
    i,j: integer;
    sum,value: double;
    strColor: string;
    att: TAttribute;
    coord,cos2: double;
    asterisque: string;
    ctr: double;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Attribute</TH>');
 for j:= 1 to FNbAxis do
  buf.AddStr(format('<TH align=center colspan=3>Axis_%d</TH>',[j]));
 buf.AddStr('</TR>');
 //rajouter la seconde ligne d'en-tte
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>-</TH>');
 for j:= 1 to FNbAxis do
  buf.AddStr('<TD align=center width=60>Coord.</TD><TD align=center width=60>CTR (%)</TD><TD align=center width=60>QLT % (Tot. %)</TD>');
 //les valeurs
 for i:= 1 to FNbVar do
  begin
   att:= Dataset.LstAtts[asInput].Attribute[pred(i)];
   if (att.isCategory(caContinue) = true)
    then asterisque:= '*'
    else asterisque:= '**';
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s (%s)</TH>',[att.Name,asterisque]));
   //new -- 06/07/2006 -- calcul des communalities
   sum:= 0.0;
   for j:= 1 to FNbAxis do
    begin
      TRY
      coord:= FCOORD[i-1,j-1];
      value:= coord;
      //contribution
      if (FEigVal^[j] > 0)
       then ctr:= value/FEigVal^[j]
       else ctr:= 0.0;
      //relativiser si catgorielle
      if (att.isCategory(caDiscrete) = true) then value:= value/(-1.0 + att.nbValues);
      //cos
      cos2:= value;
      //somme en ligne
      sum:= sum + cos2;
      //mettre en valeur si...
      if (self.isHighlighted(i,j) = true)
       then strColor:= 'blue'
       else strColor:= 'black';
      //la ligne  afficher
      buf.AddStr(format('<TD align=right width=60><font color="%s">%.6f</font></TD><TD align=right width=50><i>%0.1f %s&nbsp;</i></TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[strColor,coord,100.0*ctr,'%',100.0*cos2,'%',100.0*sum,'%']));
      EXCEPT
      END;
    end;
   buf.AddStr('</TR>');
  end;
 //pourcentage expliqu par l'axe
 buf.AddStr(HTML_TABLE_COLOR_DATA_BLUE+format('<TH>%s</TH>',['Var. Expl.']));
 sum:= 0.0;
 for j:= 1 to FNbAxis do
  begin
   sum:= sum + FEigVal^[j];
   buf.AddStr(format('<TD align=right width=60>%.6f</TD><TD align=center><i>%s</i></TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[FEigVal^[j],'-',100.0*FEigVal^[j]/FTrace,'%',100.0*sum/FTrace,'%']));
  end;
 buf.AddStr('</TR>');
 buf.AddStr('</table>');
 buf.AddStr('<br>(*) Square of correlation coefficient');
 buf.AddStr('<br>(**) Correlation ratio');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TOpFactAFDM.getInstanceFA: TClassOpFactAnalysis;
begin
 result:= TOpFactAFDM;
end;

function TOpFactAFDM.isHighlighted(i, j: integer): boolean;
var att: TAttribute;
    eigval,coord,cos2,contrib: double;
begin
 eigval:= FEigVal^[j];
 result:= (eigval > FSeuilEigenVal);//si true, il est suprieur  0 forcment
 if (result = true)
  then
   begin
    att:= Dataset.LstAtts[asInput].Attribute[pred(i)];
    coord:= FCOORD[i-1,j-1];
    if (att.isCategory(caDiscrete) = true)
     then cos2:= coord/(-1.0+att.nbValues)
     else cos2:= coord;
    contrib:= coord/eigval;
    result:= (cos2 > FSeuilCos2) and (contrib > FSeuilContrib);
   end;
end;

procedure TOpFactAFDM.setCos2Ctr(atts: TLstAttributes; nbFactors: integer;
  prefix: string; number: integer);
var lst: TObjectList;
    newAtt: TAttribute;
    factor: TAttribute;
    attSize: integer;
    nbInstances: integer;
    k,i,j: integer;
    poids,d: double;
    dist: array of double;
begin
 nbInstances:= Dataset.Examples.Size;
 poids:= 100.0/(1.0 * nbInstances);
 //mais on a besoin quand mme de crer la nouvelle variable sur toute la base
 attSize:= Dataset.LstAtts[asAll].Size;
 //suite...
 lst:= TObjectList.Create(false);//non propritaire
 //calcul des CTR et COS  partir des atts... (coordonnes sur les axes factoriels)
 //**************
 //calcul des CTR
 //**************
 for k:= 1 to nbFactors do
  begin
   factor:= atts.Attribute[pred(k)];
   newAtt:= TAttContinue.Create(format('%s_%d_CTR_%d',[prefix,number,k]),attSize);
   lst.Add(newAtt);
   //calcul Saporta --> pp.175, CTR en % puisque poids en %
   if (EigVal[k] > 0.0)
    then
     begin
      for i:= 1 to attSize do newAtt.cValue[i]:= poids * MATH.Power(factor.cValue[i],2.0) / EigVal[k];
     end
    else for i:= 1 to attSize do newAtt.cValue[i]:= 0.0;
  end;
 //***************
 //calcul des COS
 //***************
 //distance au centre de gravit des individus
 setLength(dist,succ(attSize));
 for i:= 1 to attSize do
  begin
   d:= 0.0;
   for j:= 1 to FDimension do
    d:= d + MATH.POWER((FLstVar.Items[pred(j)] as TAttTrans).getValue(i), 2.0);
   dist[i]:= d;
  end;
 //calculer le COS2
 for k:= 1 to nbFactors do
  begin
   factor:= atts.Attribute[pred(k)];
   newAtt:= TAttContinue.Create(format('%s_%d_COS2_%d',[prefix,number,k]),attSize);
   lst.Add(newAtt);
   for i:= 1 to attSize do
    begin
     if (dist[i] > 0.0)
      then newAtt.cValue[i]:= MATH.Power(factor.cValue[i],2.0)/dist[i]
      else newAtt.cValue[i]:= 0.0;
    end;
  end;
 finalize(dist);
 //flush...
 for j:= 0 to pred(lst.Count) do
  atts.Add(lst[j] as TAttribute);
 //>>
 lst.Free();
end;

procedure TOpFactAFDM.SetProjections(atts: TLstAttributes; coefs: PMatrix;
  nbFactors: integer; prefix: string; number: integer);
var i,j,k: integer;
    newAtt: TAttContinue;
    att: TAttTrans;
    s: TTypeContinue;
    attSize: integer;
begin
 atts.Clear;
 attSize:= Dataset.LstAtts[asAll].Size;
 //ok on peut lancer les calculs
 for k:= 1 to nbFactors do
  begin
   newAtt:= TAttContinue.Create(format('%s_%d_Axis_%d',[prefix,number,k]),attSize);
   //pour chaque individu i
   for i:= 1 to attSize do
    begin
     s:= 0.0;
     for j:= 1 to FDimension do
      begin
       att:= FLstVar.Items[pred(j)] as TAttTrans;
       s:= s + coefs^[j]^[k]*att.getValue(i);
      end;
     newAtt.cValue[i]:= s;
    end;
   atts.Add(newAtt);
  end;
end;

{ TAttTrans }

constructor TAttTrans.create(att: TAttribute; mean, std: double);
begin
 inherited create();
 source:= att;
 center:= mean;
 scale:= std;
 //scuriser la division
 if (scale < MIN_SCALE) then scale:= 1.0;
end;

{ TAttTransCont }

function TAttTransCont.getValue(i: integer): double;
begin
 //centrer et rduire
 result:= (source.cValue[i] - center)/scale;
end;

{ TAttTransDisc }

constructor TAttTransDisc.create(att: TAttribute; mean, std: double;
  ref: TTypeDiscrete);
begin
 inherited create(att, mean, std);
 refValue:= ref;
end;

function TAttTransDisc.getValue(i: integer): double;
var value: double;
begin
 if (source.dValue[i] = refValue)
  then value:= 1.0
  else value:= 0.0;
 result:= (value - center)/scale;
end;

{ TOpPrmFactAFDM }

function TOpPrmFactAFDM.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFactAFDM.CreateFromOpPrm(self);
end;

procedure TOpPrmFactAFDM.getFactSettings(prmSource: TOpPrmFactAnalysis);
begin
 self.NbAxis:= (prmSource as TOpPrmFactAFDM).NbAxis;
end;

function TOpPrmFactAFDM.getHTMLParameters: string;
var s: string;
begin
 s:= format('# factors = %d<br>',[NbAxis]);
 s:= s + format('Generate COS2 and CTR for instances : %d',[ord(Cos2Ctr)]);
 result:= s;
end;

procedure TOpPrmFactAFDM.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 NbAxis:= prmINI.ReadInteger(prmSection,'nb_axis',NbAxis);
 Cos2Ctr:= prmINI.ReadBool(prmSection,'cos2_ctr',Cos2Ctr);
end;

procedure TOpPrmFactAFDM.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(NbAxis,sizeof(NbAxis));
 prmStream.ReadBuffer(Cos2Ctr,sizeof(Cos2Ctr));
end;

procedure TOpPrmFactAFDM.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_axis',NbAxis);
 prmINI.WriteBool(prmSection,'cos2_ctr',Cos2Ctr);
end;

procedure TOpPrmFactAFDM.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(NbAxis,sizeof(NbAxis));
 prmStream.WriteBuffer(Cos2Ctr,sizeof(Cos2Ctr));
end;

procedure TOpPrmFactAFDM.SetDefaultParameters;
begin
 Cos2Ctr:= false;
 NbAxis:= 5;
end;

{ TCaracDisc }

constructor TCaracDisc.create(paramStat: TCalcSDCondDescANOVA; paramEig: double);
var k: integer;
    value,sum: double;
begin
 inherited create();
 Stat:= paramStat;
 EigVal:= paramEig;
 setLength(TabCTR,stat.AttDescription.nbValues+1);
 setLength(TabVTest,stat.AttDescription.nbValues+1);
 sum:= 0.0;
 for k:= 1 to stat.AttDescription.nbValues do
  begin
   value:= 0.0;
   if (stat.TSS > 0) and (EigVal > 0)
    then value:= ((1.0*stat.StatCond[pred(k)].NbExamples*SQR(stat.StatCond[pred(k)].Average))/Stat.TSS)/EigVal*100.0;
   sum:= sum + value;
   TabCTR[k]:= value;
   //valeur test - 14/09/2013
   if (stat.StatCond[pred(k)].NbExamples > 0) and (stat.NbExamples > 1)
    then TabVTest[k]:= stat.StatCond[pred(k)].Average/SQRT((1.0*stat.NbExamples-stat.StatCond[pred(k)].NbExamples)/(-1.0 + stat.NbExamples)*(EigVal/(1.0*stat.StatCond[pred(k)].NbExamples)));
  end;
 TabCTR[0]:= sum;
end;

destructor TCaracDisc.destroy;
begin
 Finalize(TabCTR);
 Finalize(TabVTest);
 Stat.Free();
 inherited;
end;

{ TCollCaracDisc }

procedure TCollCaracDisc.add(prmCD: TCaracDisc);
begin
 LstCD.Add(prmCD);
end;

constructor TCollCaracDisc.create;
begin
 inherited create();
 LstCD:= TObjectList.Create(TRUE);
end;

destructor TCollCaracDisc.destroy;
begin
 LstCD.Free();
 inherited;
end;

function TCollCaracDisc.getCD(k: integer): TCaracDisc;
begin
 result:= LstCD.Items[k] as TCaracDisc;
end;

initialization
 CLASSES.RegisterClass(TMLGenCompFactAFDM);
end.
