(****************************************************************************)
(* UCompFactDiscCorrespAnalysis.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(****************************************************************************)

{
@abstract(Discriminant Correspondence Analysis)
@author(Ricco)
@created(24/11/2012)

Analyse Discriminante pour variables descriptives qualitatives base sur l'analyse des correspondances.

Une traduction possible : "Analyse des correspondances discriminante".

Rfrence : Herv Abdi, "Discriminant Correspondence Analysis", in Encyclopedia of Measurement and Statistics, 2007.
http://www.utd.edu/~herve/Abdi-DCA2007-pretty.pdf

}

unit UCompFactDiscCorrespAnalysis;

interface

USES
  Classes, Forms, Contnrs,
  IniFiles,
  UCompDefinition,
  UCompFADefinition,
  UOperatorDefinition,
  UCompManageDataset,
  UCalcFactDiscriminantAnalysis,
  UCompFactAnalysisAFC,
  UDatasetDefinition, UDatasetImplementation,
  UCalcStatDesCrossTab, UCalcStatDesConditionnalDesc;

TYPE

    {gnrateur de composant}
    TMLGenCompFactDiscCorrespAnalysis = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    {composant DiscAnalysis}
    TMLCompFactDiscCorrespAnalysis = class(TMLCompFactAnalysis)
    protected
    function    getClassOperator: TClassOperator; override;
    end;

    //description d'une liste de variable
    TLstVariable = class;

    {oprateur DiscAnalysis}
    TOpFactDiscCorrespAnalysis = class(TOpLocalData)
    private
    //variable cible
    FTarget: TAttribute;
    //oprateur de calcul AFC pour tableau de contingence
    FCalc: TOpFactAFC;
    //donnes
    FData: TMLDataset;
    //donnes utilises pour l'AFC
    FDataAfc: TMLDataset;
    //coefficients des fonctions canoniques
    FCoefs: array of array of double;
    //liste des tableaux croiss
    FLstCrossTab: TLstCalcStatDesCrossTab;
    //Liste des anova pour les axes gnrs
    FLstAnova: TLstStatDesCondANOVA;
    //rsultats pour target
    FResTarget: TLstModalite;
    //rsultats pour inputs (avec 's')
    FResInputs: TLstVariable;
    //nombre max de facteurs
    FMaxNbFactors: integer;
    //dtruire les objets locaux
    procedure   destroyLocalObjects();
    //calculer les croisements entre cible et descripteurs
    function    calcCrossTabs(): boolean;
    //concatener les tableaux de contingence
    function    concatenateCrossTabs(): boolean;
    //prparer les donnes pour l'AFC
    function    prepareLocalData(): boolean;
    //lancer l'AFC
    function   launchAFC(): boolean;
    //calculer les coefficients des fonctions canoniques
    function   calcCanonicalFunctions(): boolean;
    //calculer les ANOVA pour chaque facteur -> pour obtenir les rapports de corrlation
    function   calcCorrelationRatio(): boolean;
    //qqs post-traitements, ex. tri des tableaux croiss, etc.
    function    postProcess(): boolean;
    protected
    function    CheckAttributes(): boolean; override;
    function    CoreExecute(): boolean; override;
    function    getClassParameter: TClassOperatorParameter; override;
    function    setProjection(): boolean;
    //rcupration des rsultats  partir de l'objet AFC
    function    getResultsFromAFC(): boolean;
    //dification du rapport
    function    getHTMLOverallEvaluation(): string;
    function    getHTMLRoots(): string;
    function    getHTMLCanonicalFunctions(): string;
    function    getHTMLSqDistCentroids(): string;
    function    getHTMLCrossTabs(): string;
    function    getNewForm(): TForm; override;
    public
    destructor  destroy(); override;
    function    getHTMLResultsSummary(): string; override;
    //proprits
    property    Calc: TOpFactAFC read FCalc;
    end;

    {paramtrage de l'oprateur DiscAnalysis}
    TOpPrmFactDiscCorrespAnalysis = class(TOpPrmFactAnalysis)
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    //affichage des tableaux de contingence
    ShowCrossTabs: boolean;
    //trier les variables selon les contributions
    SortValues: boolean;
    //fixer le nombre de facteurs  construire - NON, cela fausserait certains rsultats, ex. distance entre centres de classes
    //NbFactors: integer;
    procedure   getFactSettings(prmSource: TOpPrmFactAnalysis); override;
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    end;

    //**********************************************
    // pour une meilleure organisation des rsultats
    //**********************************************

    //description des rsultats pour une variable (avec sa liste interne de modalits)
    TVariable = class
     public
     //nom de variable
     NomVar: string;
     //nb d'axes
     Nbaxis: integer;
     //nombre de modalits
     NbModalites: integer;
     //tableau des descriptions
     Descriptions: array of string;
     //tableau des poids
     Weight: array of double;
     //tableau des sq.dist
     SqDist: array of double;
     //tableau des inerties
     Inertia: array of double;
     //matrice des coordonnes : dim 1 = axe, dim 2 = modalites
     Coord: array of array of double;
     //ctr
     Ctr: array of array of double;
     //cos2
     Cos2: array of array of double;
     //constructeur
     constructor create(prmNomVar: string; prmNbAxis: integer; prmNbModa: integer);
     //destructeur
     destructor  destroy(); override;
     //obtenir le max de contribution pour un axe
     function    getMaxCtr(axe: integer): double;
     //description HTML
     function    getHTML(seuilCtr,seuilCos2: double; couleurBG: string): string;
    end;

    //liste des variables
    TLstVariable = class
     public
     //nombre d'axes rellement utilis
     NbAxis: integer;
     //nombre max d'axes potentiels
     MaxNbAxis: integer;
     //liste interne
     LstVariable: TObjectList;
     //constructeur
     constructor create(prmNbAxis,prmMaxNbAxis: integer);
     //destructeur
     destructor  destroy(); override;
     //ajouter
     procedure   add(variable: TVariable);
     //compter
     function    count(): integer;
     //accder
     function    get(index: integer): TVariable;
     //description d'une liste de variables
     function    getHTML(nbDummies: integer): string;
     //trier de manire dcroissante sur le max. des CTR - par axe...
     procedure   SortOnMaxCtr(seuil: double);
    end;

implementation

uses
  Sysutils, Math,
  UDatasetExamples,
  ULogFile, UConstConfiguration, UCalcMatrixAdditionalFunctions,
  UFrmOpViewFactDiscCorresp, UDlgOpPrmFactDCA;

const
  ZERO_FACT_DCA_POUR_SEUIL : double = 1e-4;//correction du seuil pour viter les erreurs de troncature

var
  //indique le facteur de rfrence pour le tri des variables/modalits
  idFactor: integer;

{ TOpPrmFactDiscCorrespAnalysis }

function TOpPrmFactDiscCorrespAnalysis.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmFactDCA.CreateFromOpPrm(self);
end;

procedure TOpPrmFactDiscCorrespAnalysis.getFactSettings(
  prmSource: TOpPrmFactAnalysis);
begin
 ShowCrossTabs:= (prmSource as TOpPrmFactDiscCorrespAnalysis).ShowCrossTabs;
 SortValues:= (prmSource as TOpPrmFactDiscCorrespAnalysis).SortValues;
 //NbFactors:= (prmSource as TOpPrmFactDiscCorrespAnalysis).NbFactors;
end;

function TOpPrmFactDiscCorrespAnalysis.getHTMLParameters: string;
begin
 result:= format('Show all crosstabs = %d<br>',[ord(ShowCrossTabs)]);
 result:= result + format('Sort results according to CTR = %d',[ord(SortValues)]);
 //result:= result + format('Number of factors = %d',[NbFactors]);
end;

procedure TOpPrmFactDiscCorrespAnalysis.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 ShowCrossTabs:= prmINI.ReadBool(prmSection,'show_crosstabs',ShowCrossTabs);
 SortValues:= prmINI.ReadBool(prmSection,'sort_results',SortValues);
 //NbFactors:= prmINI.ReadInteger(prmSection,'nb_factors',NbFactors);
end;

procedure TOpPrmFactDiscCorrespAnalysis.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(ShowCrossTabs,sizeof(ShowCrossTabs));
 prmStream.ReadBuffer(SortValues,sizeof(SortValues));
 //prmStream.ReadBuffer(NbFactors,sizeof(NbFactors));
end;

procedure TOpPrmFactDiscCorrespAnalysis.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteBool(prmSection,'show_crosstabs',ShowCrossTabs);
 prmINI.WriteBool(prmSection,'sort_results',SortValues);
 //prmINI.WriteInteger(prmSection,'nb_factors',NbFactors);
end;

procedure TOpPrmFactDiscCorrespAnalysis.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(ShowCrossTabs,sizeof(ShowCrossTabs));
 prmStream.WriteBuffer(SortValues,sizeof(SortValues));
 //prmStream.WriteBuffer(NbFactors,sizeof(NbFactors));
end;

procedure TOpPrmFactDiscCorrespAnalysis.SetDefaultParameters;
begin
 ShowCrossTabs:= false;
 SortValues:= true;
end;

{ TMLGenCompFactDiscCorrespAnalysis }

procedure TMLGenCompFactDiscCorrespAnalysis.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactDiscCorrespAnalysis.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactDiscCorrespAnalysis;
end;

{ TMLCompFactDiscCorrespAnalysis }

function TMLCompFactDiscCorrespAnalysis.getClassOperator: TClassOperator;
begin
 result:= TOpFactDiscCorrespAnalysis;
end;

{ TOpFactDiscCorrespAnalysis }

function TOpFactDiscCorrespAnalysis.calcCanonicalFunctions: boolean;
var k,j: integer;
    value,p: double;
begin
 result:= true;
 TRY
 p:= 1.0 * Self.FData.LstAtts[asInput].Count;
 SetLength(FCoefs,FCalc.FNbCol,FCalc.FNbAxis);
 for k:= 1 to FCalc.FNbAxis do
  begin
   For j:= 1 to FCalc.FNbCol do
    begin
     value:= 0;
     if (FCalc.EigVal[k+1] > 0)
      then value:= (FCalc.FColCoord[j,k] / p) / SQRT(FCalc.EigVal[k+1]);
     FCoefs[j-1,k-1]:= value;
    end;
  end;
 EXCEPT
 result:= false;
 END;
end;

function TOpFactDiscCorrespAnalysis.calcCorrelationRatio: boolean;
var j: integer;
begin
 result:= true;
 TRY
 FLstAnova:= TLstStatDesCondANOVA.Create(nil,nil);
 for j:= 0 to pred(GenAtts.Count) do
  FLstAnova.AddStat(TCalcSDCondDescANOVA.Create(GenAtts.Attribute[j],FTarget,FData.Examples));
 EXCEPT
 result:= false;
 END;
end;

function TOpFactDiscCorrespAnalysis.calcCrossTabs(): boolean;
var target: TAttribute;
    j: integer;
    ok: boolean;
begin
 ok:= true;
 TRY
 target:= FData.LstAtts[asTarget].Attribute[0];
 FLstCrossTab:= TLstCalcStatDesCrossTab.Create(nil,nil);
 for j:= 0 to pred(FData.LstAtts[asInput].Count) do
  FLstCrossTab.AddStat(TCalcSDCrossTab.Create(target,FData.LstAtts[asInput].Attribute[j],FData.Examples));
 EXCEPT
 ok:= false;
 END;
 result:= ok;
end;

function TOpFactDiscCorrespAnalysis.CheckAttributes: boolean;
var ok: boolean;
begin
 //tous discrets pour INPUT
 ok:= (self.WorkData.LstAtts[asInput].Count>0) and (self.WorkData.LstAtts[asInput].isAllCategory(caDiscrete));
 //un target discret
 ok:= ok and ((self.WorkData.LstAtts[asTarget].Count = 1) and (self.WorkData.LstAtts[asTarget].isAllCategory(caDiscrete)));
 //and then...
 result:= ok; 
end;

function TOpFactDiscCorrespAnalysis.concatenateCrossTabs: boolean;
var target,input,att: TAttribute;
    i,j,k: integer;
    calc: TCalcSDCrossTab;
begin
 result:= true;
 TRY
 FDataAfc:= TMLDataset.CreateFromNil(FTarget.nbValues);
 //variable cible
 target:= TAttDiscrete.Create(FTarget.Name,FTarget.nbValues);
 for i:= 1 to target.Size do
  begin
   target.LstValues.getValue(FTarget.LstValues.getDescription(i));
   target.dValue[i]:= i;
  end;
 FDataAfc.LstAtts[asTarget].Add(target);
 //les variables input
 for j:= 0 to pred(self.FData.LstAtts[asInput].Count) do
  begin
   att:= self.FData.LstAtts[asInput].Attribute[j];
   calc:= self.FLstCrossTab.Stat(j) as TCalcSDCrossTab;
   for k:= 1 to att.nbValues do
    begin
     input:= TAttContinue.Create(att.Name+' = '+att.LstValues.getDescription(k),FTarget.nbValues);
     for i:= 1 to input.Size do
      input.dValue[i]:= calc.CrossTab.Value[i,k];
     FDataAfc.LstAtts[asInput].Add(input);
    end;
  end;
 EXCEPT
 result:= false;
 END;
end;

function TOpFactDiscCorrespAnalysis.CoreExecute: boolean;
var ok: boolean;
begin
 TRY
 self.destroyLocalObjects();
 ok:= self.prepareLocalData();
 ok:= ok and self.launchAFC();
 ok:= ok and self.calcCanonicalFunctions();
 ok:= ok and self.setProjection();
 ok:= ok and self.calcCorrelationRatio();
 ok:= ok and self.postProcess();
 ok:= ok and self.getResultsFromAFC();
 EXCEPT
 ok:= false;
 END;
 result:= ok;
end;

destructor TOpFactDiscCorrespAnalysis.destroy;
begin
 self.destroyLocalObjects();
 inherited;
end;

procedure TOpFactDiscCorrespAnalysis.destroyLocalObjects;
begin
 if assigned(FCalc) then FCalc.Free();
 if assigned(FDataAfc) then FDataAfc.Free();
 if assigned(FLstCrossTab) then FLstCrossTab.Free();
 if assigned(FCoefs) then Finalize(FCoefs);
 if assigned(FLstAnova) then FLstAnova.Free();
 if assigned(FResTarget) then FResTarget.Free();
 if assigned(FResInputs) then FResInputs.Free();
end;

function TOpFactDiscCorrespAnalysis.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactDiscCorrespAnalysis;
end;

function TOpFactDiscCorrespAnalysis.getHTMLCanonicalFunctions: string;
var j,k: integer;
    s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=150>Attribute.Value</TH>';
 for k:= 1 to FCalc.FNbAxis do
  s:= s + format('<TH width=80>Factor %d</TH>',[k]);
 s:= s + '</TR>';
 for j:= 1 to FCalc.FNbCol do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + Format('<TD align="left" %s>%s</TD>',[HTML_BGCOLOR_DATA_BLUE,FCalc.Dataset.LstAtts[asInput].Attribute[j-1].Name]);
   for k:= 1 to FCalc.FNbAxis do
    s:= s + Format('<TD align="right">%.7f</TD>',[FCoefs[j-1,k-1]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactDiscCorrespAnalysis.getHTMLCrossTabs: string;
begin
 result:= FLstCrossTab.getHTMLResults(0);
end;

function TOpFactDiscCorrespAnalysis.getHTMLOverallEvaluation: string;
var s: string;
    total,R2: double;
    j: integer;
begin
 //variation totale
 total:= 0.0;
 for j:= 0 to pred(FLstAnova.Count) do
  total:= total + (FLstAnova.Stat(j) as TCalcSDCondDescANOVA).StatGlobal.VariancePop;
 //affichage
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Source of variation</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Between groups</TD><TD align=right width=80><font color="lime">%.4f</font></TD>',[FCalc.Trace])+ '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Within groups</TD><TD align=right>%.4f</TD>',[Total-FCalc.Trace])+ '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD><b>Total</b></TD><TD align=right>%.4f</TD>',[Total])+ '</TR>';
 R2:= 0; if (total > 0) then R2:= FCalc.Trace/total;
 s:= s + HTML_TABLE_COLOR_DATA_LIGHT_RED + format('<TD>R-Squared</b></TD><TD align="right"><font color="blue"><b>%.4f<b></font></TD>',[R2])+ '</TR>';
 s:= s + '</table>';
 result:= s;
end;

function TOpFactDiscCorrespAnalysis.getHTMLResultsSummary: string;
var s: string;
begin
 result:= 'Error during learning process';
 if assigned(FCalc)
  then
   begin
    s:= '';
    //dcomposition de la variation -  valider !!!
    //s:= s + '<P><H3>Overall evaluation</H3>';
    //s:= s + self.getHTMLOverallEvaluation();
    //valeurs propres, en profiter pour afficher les rapports de corrlation (cf. CDA)
    //?\ contrairement  l'analyse discriminante, on ne peut pas tester "les 'q' derniers facteurs sont-ils significatifs ?"
    //puisque le nuage n'est pas gaussien... dommage, dommage...
    s:= s + '<H3>Roots - Eigen values</H3>';
    s:= s + self.getHTMLRoots();
    //***************************
    //caractrisation des groupes
    //***************************
    s:= s + '<BR><BR><H3>Group characterization</H3>';
    //coordonnes factorielles des groupes  discriminer
    //s:= s + '<H5>Group centroids on canonical variables</H5>' + FCalc.getHTMLLignes();
    s:= s + '<H5>Group centroids on canonical variables</H5>' + Self.FResTarget.getHTML();
    //carr des distances entres groupes
    s:= s + '<BR><BR><H5>Squared distance between group centroids</H5>';
    s:= s + self.getHTMLSqDistCentroids();
    //*********************
    //Rle des descripteurs
    //*********************
    //structure des facteurs
    s:= s + '<BR><BR><H3>Canonical Structure</H3>';
    //s:= s + FCalc.getHTMLColonnes();
    s:= s + self.FResInputs.getHTML(FCalc.FNbCol);
    //fonction canonique
    s:= s + '<BR><BR><font size="4">Canonical Coefficients</font><br>';
    s:= s + '<font color="gray">Applied to the indicator matrix i.e. columns are dummy variables</font>';
    s:= s + self.getHTMLCanonicalFunctions();
    //affichage des tableaux de croisement si demand
    if ((self.PrmOp as TOpPrmFactDiscCorrespAnalysis).ShowCrossTabs = true)
     then
      begin
       s:= s + '<BR><BR><H3>Crosstabs between target and inputs</H3>';
       s:= s + self.getHTMLCrossTabs();
      end;
    //and then...
    result:= s;
   end;
end;

function TOpFactDiscCorrespAnalysis.getHTMLRoots: string;
var s: string;
    i: integer;
    tot,sum: double;
    R2: double;
    stat: TCalcSDCondDescANOVA;
begin
 tot:= FCalc.Trace;
 s:= '';
 s:= s + format('<font size = "2">Matrix trace = %.5f</font>',[tot]);
 //canonical corr, sq. canonical corr, valeurs propres, proportion, cumulative
 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+
       '<TH rowspan=2>Factor</TH><TH width=80 rowspan=2>Canonical Correlation R</TH><TH width=80 rowspan=2>Squared R</TH><th colspan=3>Explained (between) variation</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=80>Eigen value</TH><TH width=80>Proportion (%)</TH><TH width=80>Cumulative (%)</TH></TR>';
 sum:= 0.0;
 for i:= 2 to FCalc.FNbAxis+1 do
  begin
   stat:= FLstAnova.Stat(i-2) as TCalcSDCondDescANOVA;
   //carr du rapport de corrlation
   R2:= 0.0;
   //if (stat.StatGlobal.VariancePop > 0.0) then R2:= FCalc.EigVal[i] / stat.StatGlobal.VariancePop;
   //autre solution uniquement base sur les caractristiques des axes factoriels
   if (stat.TSS > 0.0)
    then R2:= stat.BSS / stat.TSS; //--> on devrait obtenir exactement le mme R2
   //affichage
   sum:= sum + FCalc.EigVal[i];
   if (tot>0)
    then
     s:= s + HTML_TABLE_COLOR_DATA_GRAY;
     s:= s + format('<TH>%d</TH><TD align=right><font color="blue">%.4f</font></TD><TD align=right><font color="blue">%.4f</font></TD>',
                    [i-1,SQRT(R2),R2]);
     s:= s + format('<TD align=right>%.5f</TD><TD align=right>%.2f</TD><TD align=right>%.2f</TD>',
                    [FCalc.EigVal[i],100.0*FCalc.EigVal[i]/tot,100.0*sum/tot]);
     s:= s + '</TR>';
  end;
 if (tot > 0.0)
  then s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH colspan=3 align="center">Tot.</TH><TD align=right><font color="green"><b>%.5f</b></font></TD><TD align=right>%.2f</TD><TD align=center>-</TD></TR>',[sum,100.0*sum/tot])
  else s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH colspan=3 align="center">Tot.</TH><TD align=right><font color="green"><b>%.5f</b></font></TD><TD align=right>%s</TD><TD align=center>-</TD></TR>',[sum,'ERR']);
 s:= s + '</table>';
 result:= s;
end;

function TOpFactDiscCorrespAnalysis.getHTMLSqDistCentroids: string;
var s: string;
    k,k1,k2,j: integer;
    value: double;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width="150">-</TH>';
 //for k:= 1 to FTarget.nbValues do s:= s + format('<TH width=100 %s>%s</TH>',[HTML_BGCOLOR_DATA_GREEN,FTarget.LstValues.getDescription(k)]);
 for k:= 1 to FResTarget.count do s:= s + format('<TH width=100 %s>%s</TH>',[HTML_BGCOLOR_DATA_GREEN,FResTarget.get(k-1).Description]);
 s:= s + '</TR>';
 for k1:= 1 to FResTarget.count do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TH %s>%s</TH>',[HTML_BGCOLOR_DATA_GREEN, FResTarget.get(k1-1).Description]);
   for k2:= 1 to FResTarget.count do
    begin
     value:= 0;
     if (k1 <> k2)
      then
       begin
        for j:= 1 to FCalc.FNbAxis do
         //value:= value + SQR(FCalc.FLigCoord[k1][j] - FCalc.FLigCoord[k2][j]);
         value:= value + SQR(FResTarget.get(k1-1).Coord[j-1] - FResTarget.get(k2-1).Coord[j-1]);
       end;
     s:= s + format('<TD align="right">%.4f</TD>',[value]);
    end;
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactDiscCorrespAnalysis.getNewForm: TForm;
begin
 result:= TFrmOpViewFactDiscCorresp.CreateFromOperator(self); 
end;

function TOpFactDiscCorrespAnalysis.getResultsFromAFC: boolean;
var k,j: integer;
    i,iIndex: integer;
    moda: TModalite;
    input: TAttribute;
    variable: TVariable;
    seuilCtr: double;
begin
 result:= true;
 TRY
 //*************************
 //descripton pour le target
 FResTarget:= TLstModalite.create(FCalc.FNbAxis,self.FMaxNbFactors);
 for k:= 1 to FCalc.FNbLig do
  begin
   moda:= TModalite.Create(FCalc.Dataset.LstAtts[asTarget].Attribute[0].sValue[k],FCalc.FNbAxis);
   moda.Weight:= FCalc.MatFreq[k][0];
   moda.SqDist:= FCalc.FDistoLig[k];
   moda.Inertia:= moda.Weight * moda.SqDist;
   for j:= 1 to moda.NbAxis do
    begin
     moda.Coord[j-1]:= FCalc.FLigCoord[k][j];
     moda.Ctr[j-1]:= FCalc.LigCtr[k][j];
     moda.Cos2[j-1]:= FCalc.LigCos2[k][j];
    end;
   FResTarget.add(moda);
  end;
 //**************************
 //description pour les input
 FResInputs:= TLstVariable.create(FCalc.FNbAxis,self.FMaxNbFactors);
 iIndex:= 0;
 for k:= 1 to FData.LstAtts[asInput].Count do
  begin
   input:= FData.LstAtts[asInput].Attribute[k-1];
   variable:= TVariable.create(input.Name,FCalc.FNbAxis,input.nbValues);
   for i:= 1 to input.nbValues do
    begin
     variable.Descriptions[i-1]:= FCalc.Dataset.lstAtts[asInput].Attribute[pred(iIndex+i)].Name;
     variable.Weight[i-1]:= FCalc.MatFreq[0][iIndex+i];
     variable.SqDist[i-1]:= FCalc.FDistoCol[iIndex+i];
     variable.Inertia[i-1]:= variable.Weight[i-1] * variable.SqDist[i-1];
     for j:= 1 to variable.NbAxis do
      begin
       variable.Coord[j-1][i-1]:= FCalc.FColCoord[iIndex+i][j];
       variable.Ctr[j-1][i-1]:= FCalc.ColCtr[iIndex+i][j];
       variable.Cos2[j-1][i-1]:= FCalc.ColCos2[iIndex+i][j];
      end;
    end;
   inc(iIndex,input.nbValues);
   FResInputs.add(variable);
  end;
 //********************************************************
 //trier les rsultats pour une prsentation plus lisible ?
 if ((self.PrmOp as TOpPrmFactDiscCorrespAnalysis).SortValues = true)
  then
   begin
    seuilCtr:= 100.0/(1.0 * FCalc.FNbLig) - ZERO_FACT_DCA_POUR_SEUIL;
    FResTarget.SortOnCtr(seuilCtr);
    seuilCtr:= 100.0/(1.0 * FCalc.FNbCol) - ZERO_FACT_DCA_POUR_SEUIL;
    FResInputs.SortOnMaxCtr(seuilCtr);
   end;
 EXCEPT
 result:= false;
 END;
end;

function TOpFactDiscCorrespAnalysis.launchAFC(): boolean;
var ok: boolean;
begin
 TRY
 //crer l'objet
 FCalc:= TOpFactAFC.CreateWithDataset(self.MLOwner,FDataAfc);
 //dterminer le nombre max. d'axes et fixer le nombre d'axes  produire
 FMaxNbFactors:= Math.Min(FTarget.nbValues-1,FDataAfc.LstAtts[asInput].Count-1);
 (FCalc.PrmOp as TOpPrmFactAFC).AskNbAxis:= FMaxNbFactors;
 //apprentissage
 ok:= FCalc.calcEigen();
 ok:= ok and FCalc.calcAdditional();
 ok:= ok and FCalc.calcProjections();
 ok:= ok and FCalc.calcPostProjections();
 EXCEPT
 ok:= false;
 END;
 result:= ok;
end;

function TOpFactDiscCorrespAnalysis.postProcess: boolean;
begin
 result:= true;
 TRY
 //tri des tableaux croiss
 FLstCrossTab.CompareMode:= 2;//comparaison sur le T de Tschuprow
 FLstCrossTab.SortStats();
 EXCEPT
 result:= false;
 END;
end;

function TOpFactDiscCorrespAnalysis.prepareLocalData(): boolean;
var ok: boolean;
begin
 TRY
 FData:= self.WorkData;
 FTarget:= FData.LstAtts[asTarget].Attribute[0];
 //calculer tous les croisements avec la cible
 ok:= self.calcCrossTabs();
 //accoler les tableaux de contingence
 ok:= ok and self.concatenateCrossTabs();
 EXCEPT
 ok:= false;
 END;
 result:= ok;
end;

function TOpFactDiscCorrespAnalysis.setProjection(): boolean;
var k,i,j,jIndex,id: integer;
    newAtt: TAttribute;
    value: double;
begin
 result:= true;
 TRY
 //vider la liste des variables dj produites
 GenAtts.Clear();
 //reconstruire la liste des variables  produire
 for k:= 1 to FCalc.FNbAxis do
  begin
   //crer le nouvelle attribut
   newAtt:= TAttContinue.Create(Format('DCA_%d_Axis_%d',[(MLOwner as TMLCompFactDiscCorrespAnalysis).Number,k]),FData.LstAtts[asAll].Size);
   //remplir pour toutes les observations
   for i:= 1 to newAtt.Size do
    begin
     jIndex:= 0;
     value:= 0;
     for j:= 1 to FData.LstAtts[asInput].Count do
      begin
       //pour cette variable, numro de modalit pour l'individu i
       id:= FData.LstAtts[asInput].Attribute[j-1].dValue[i];
       //tenir compte du dcalage pour l'accs au coefficient
       value:= value + FCoefs[jIndex+id-1,k-1];
       //toujours avec le dcalage, passer  la variable suivante
       inc(jIndex,FData.LstAtts[asInput].Attribute[j-1].nbValues);
      end;
     newAtt.cValue[i]:= value;
    end;
   //ajouter dans les sorties
   GenAtts.Add(newAtt);
  end;
 EXCEPT
 result:= false;
 END;
end;

{ TVariable }

constructor TVariable.create(prmNomVar: string; prmNbAxis, prmNbModa: integer);
begin
 inherited Create();
 NomVar:= prmNomVar;
 Nbaxis:= prmNbAxis;
 NbModalites:= prmNbModa;
 setLength(Descriptions,NbModalites);
 setLength(Coord,NbAxis,NbModalites);
 setLength(Ctr,NbAxis,NbModalites);
 setLength(Cos2,NbAxis,NbModalites);
 setLength(Weight,NbModalites);
 setLength(SqDist,NbModalites);
 setLength(Inertia,NbModalites);
end;

destructor TVariable.destroy;
begin
 Finalize(Descriptions);
 Finalize(Coord);
 Finalize(Ctr);
 Finalize(Cos2);
 Finalize(Weight);
 Finalize(SqDist);
 Finalize(Inertia);
 inherited;
end;

function TVariable.getHTML(seuilCtr, seuilCos2: double; couleurBG: string): string;
var s: string;
    j,a: integer;
    couleur: string;
    cumsum: double;
begin
 s:= '';
 //valeurs
 for j:= 1 to self.NbModalites do
  begin
   s:= s + format('<TR %s>',[couleurBG]);
   s:= s + format('<TD align="left">%s</TD>',[Descriptions[j-1]]);
   //poids, distances au centre de gravit (Sq. Dist) et inerties (inertia)
   s:= s + format('<TD align="right"><font color="teal">%.5f</font></TD>',[Weight[j-1]]);
   s:= s + format('<TD align="right"><font color="teal">%.5f</font></TD>',[SqDist[j-1]]);
   s:= s + format('<TD align="right"><font color="teal">%.5f</font></TD>',[Inertia[j-1]]);
   for a:= 1 to NbAxis do
    begin
     couleur:= couleurBG;
     if (Ctr[a-1][j-1] >= seuilCtr) and (Cos2[a-1][j-1] >= seuilCos2)
      then
       begin
        if (Coord[a-1][j-1] < 0)
         then couleur:= FACT_COLOR_HIGHLIGHT[cthNegLow]
         else couleur:= FACT_COLOR_HIGHLIGHT[cthPosLow];
       end;
     s:= s + format('<TD align="right" %s>%.5f</TD>',[couleur,Coord[a-1][j-1]]);
    end;
   for a:= 1 to NbAxis do s:= s + format('<TD align="right">%.1f</TD>',[Ctr[a-1][j-1]]);
   //for a:= 1 to NbAxis do s:= s + format('<TD align="right">%.3f</TD>',[Cos2[a-1][j-1]]);
   //new - 12/12/2012 - mettre la somme cumule
   cumsum:= 0.0;
   for a:= 1 to NbAxis do
    begin
     cumsum:= cumsum + Cos2[a-1][j-1];
     s:= s + format('<TD align="right">%.2f &nbsp; (%.2f)</TD>',[Cos2[a-1][j-1],cumsum]);
    end;
   //<<
   s:= s + '</TR>';
  end;
 result:= s;
end;

function TVariable.getMaxCtr(axe: integer): double;
begin
 result:= Math.MaxValue(Ctr[axe]);
end;

{ TLstVariable }

procedure TLstVariable.add(variable: TVariable);
begin
 LstVariable.Add(variable);
end;

function TLstVariable.count: integer;
begin
 result:= LstVariable.Count;
end;

constructor TLstVariable.create(prmNbAxis,prmMaxNbAxis: integer);
begin
 inherited create();
 NbAxis:= prmNbAxis;
 MaxNbAxis:= prmMaxNbAxis;
 LstVariable:= TObjectList.Create(true);
end;

destructor TLstVariable.destroy;
begin
 LstVariable.Free();
 inherited;
end;

function TLstVariable.get(index: integer): TVariable;
begin
 result:= LstVariable.Items[index] as TVariable;
end;

const
    //alternance de couleur pour matrialiser les blocs de modalits associs aux variables
    COLOR_LIGNE_VARIABLE : array[0..1] of string = ('BGCOLOR="#E1E6FA"', 'BGCOLOR="#EAF1DD"');//'BGCOLOR="#EFEFEF"');

function TLstVariable.getHTML(nbDummies: integer): string;
var s: string;
    seuilCtr,seuilCos2: double;
    i,j: integer;
begin
 //calculer les seuils
 seuilCtr:= 100.0 / (1.0 * nbDummies) - ZERO_FACT_DCA_POUR_SEUIL;
 seuilCos2:= 1.0 / (1.0 * self.MaxNbAxis) - ZERO_FACT_DCA_POUR_SEUIL;
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH colspan=4>%s</TH>',['Row Characterization']) +
         format('<TH colspan=%d>Coord.</TH>',[NbAxis]) + format('<TH colspan=%d>Contributions (%s)</TH>',[NbAxis,'%']) + format('<TH colspan=%d>COS</TH>',[NbAxis]) + '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width="150">Values</TH><TH width="50">Weight</TH>';
 //mettre les distances au centre de gravit (Sq. Dist) et les inerties (inertia)
 s:= s + '<TH width="50">Sq. Dist.</TH><TH width="50">Inertia</TH>';
 //en-tte des coordones, etc.
 for j:= 1 to NbAxis do s:= s + format('<TD width="50" align="center">coord %d</TD>',[j]);
 for j:= 1 to NbAxis do s:= s + format('<TD width="50" align="center">ctr %d</TD>',[j]);
 for j:= 1 to NbAxis do s:= s + format('<TD width="70" align="center">cos %d</TD>',[j]);
 s:= s + '</TR>';
 for i:= 0 to pred(self.count) do
  begin
   s:= s + self.get(i).getHTML(seuilCtr,seuilCos2,COLOR_LIGNE_VARIABLE[i mod 2])
  end;
 s:= s + '</TABLE>';
 result:= s;
end;

//comparaison de deux variables
function CompareVariable(Item1, Item2: Pointer): integer;
var p1, p2: TVariable;
    value1, value2: double;
begin
 p1:= TVariable(Item1);
 p2:= TVariable(Item2);
 //c'est ici que a va se jouer
 value1:= p1.getMaxCtr(idFactor);
 value2:= p2.getMaxCtr(idFactor);
 //comparaison - tri croissant
 if (value1 < value2)
  then result:= -1
  else
   begin
    if (value1 > value2)
     then result:= +1
     else result:= 0;
   end;
end;

procedure TLstVariable.SortOnMaxCtr(seuil: double);
var output: TObjectList;

  //dpiler tant que non vide et suprieur  seuil
  procedure depiler(limit: double);
  var i,initNb: integer;
      ctrMax: double;
  begin
   initNb:= self.count();
   for i:= pred(initNb) downto 0 do
    begin
     ctrMax:= self.get(i).getMaxCtr(idFactor);
     if (ctrMax >= limit)
      then
       begin
        output.add(self.get(i));
        self.LstVariable.Delete(i);
       end
      else break;
    end;
  end;

begin
 //modifier le statut de la liste interne
 LstVariable.OwnsObjects:= false;
 //crer la liste de rsultats
 output:= TObjectList.Create(false);
 //lancer les tris successifs
 idFactor:= 0;
 while (idFactor < self.NbAxis) and (self.count() > 1) do
  begin
   self.LstVariable.Sort(CompareVariable);
   depiler(seuil);
   inc(idFactor);   
  end;
 //s'il en reste, dpiler compltement
 if (self.count() > 0)
  then
   begin
    //prendre en rfrence le dernier facteur
    idFactor:= pred(self.NbAxis);
    depiler(-1);//-1 pour qu'il prenne tout ce qui reste
   end;
 //assigner et elle redevient propritaire
 LstVariable.Assign(output);
 LstVariable.OwnsObjects:= true;
 output.Free();
end;

initialization
 Classes.RegisterClass(TMLGenCompFactDiscCorrespAnalysis);
end.
