(*************************************************************************)
(* UCompFactAnalysisHarrisCA.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(*************************************************************************)

{
@abstract(Harris Component Analysis)
@author(Ricco)
@created(07/09/2012)

Principal Factor Analysis - Option METHOD = HARRIS
Traduite en Franais avec "Analyse en facteurs communs et spcifiques (Harris Image Matrix)"

}
unit UCompFactAnalysisHarrisCA;

interface

USES
        Forms,Classes,
        IniFiles,
        Matrices,
        UCalcStatDes,
        UCompDefinition,
        UCompFADefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UCompFactAnalysisPFA;

TYPE
        {gnrateur de composant ACP}
        TMLGenCompHarrisCA = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant PFA}
        TMLCompFactHarrisCA = class(TMLCompFactPFA)
        protected
        //obtenir la classe de calcul
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur PFA, on travaille sur les INPUT}
        TOpFactHarrisCA = class(TOpFactPFA)
        protected
        //dtruire les matrices et vecteurs
        procedure   DelPrivateMatrix(); override;
        //construire la matrice  diagonaliser
        procedure   BuildVCV(); override;
        //calculer les loadings
        procedure   CalcLoadings(); override;
        //calculer les coefficients de projection
        //procedure   CalcFactorScores(); override;
        //obtenir la classe de paramtrage
        function    getClassParameter: TClassOperatorParameter; override;
        //prefixe de nom de variable
        function    getPrexifNameForProjection(): string; override;
        public
        //function    getRotatedCoefs(newLoadings, newEgVec: PMatrix; nbFactors: integer): PMatrix; override;
        //affichage des loadings
        function    getHTMLLoadings(loadings: PMatrix; nbFactors: integer; sorted: boolean; loadingsForSorting: PMatrix = nil): string; override;
        //obtenir la classe
        function    getInstanceFA(): TClassOpFactAnalysis; override; 
        end;

        TOpPrmFactHarrisCA = class(TOpPrmFactPFA)
        end;


implementation

uses
    UStringAddBuffered, UConstConfiguration, Sysutils,
    UCompFactAnalysisPCA, Math;

const
    HTML_BGCOLOR_COMMUNALITIES_PRIOR : string = 'BGCOLOR="#E6E296"';
    HTML_BGCOLOR_COMMUNALITIES_FINAL : string = 'BGCOLOR="#EEEEC6"';

{ TMLGenCompHarrisCA }

procedure TMLGenCompHarrisCA.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompHarrisCA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactHarrisCA;
end;

{ TMLCompFactHarrisCA }

function TMLCompFactHarrisCA.getClassOperator: TClassOperator;
begin
 result:= TOpFactHarrisCA;
end;

{ TOpFactHarrisCA }

procedure TOpFactHarrisCA.BuildVCV;
var i,j: integer;
    value, vi,vj: double;
begin
 inherited;
 //effectuer des modifications supplmentaires
 for i:= 1 to FDimAnalysis do
  begin
   vi:= FUniqueness[i];
   for j:= 1 to FDimAnalysis do
    begin
     vj:= FUniqueness[j];
     if (vi > 0.0) and (vj > 0.0)
      then value:= FMatToDiag[i,j]/(sqrt(vi)*sqrt(vj))
      else value:= 0.0;
     FMatToDiag[i,j]:= value;
    end;
  end;
 //la nouvelle trace
 FTrace:= 0.0;
 for i:= 1 to FDimAnalysis do FTrace:= FTrace + FMatToDiag[i,i];
end;

procedure TOpFactHarrisCA.CalcLoadings;
var i,j: integer;
    value: double;
begin
 inherited;
 //et multiplier par les uniqueness
 for i:= 1 to FDimAnalysis do
  begin
   value:= SQRT(FUniqueness[i]);
   for j:= 1 to FNbAxis do
    FLoadings[i,j]:= FLoadings[i,j] * value;
  end;
end;

procedure TOpFactHarrisCA.DelPrivateMatrix;
begin
 //appel de la mthode de l'anctre
 inherited;
 //autres...
end;

function TOpFactHarrisCA.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactHarrisCA;
end;

procedure getColorCommunalities(value: double; var colPrior,colFinal: string);
begin
 if (value > 0.5)
  then
   begin
    colPrior:= HTML_BGCOLOR_COMMUNALITIES_PRIOR;
    colFinal:= HTML_BGCOLOR_COMMUNALITIES_FINAL;
   end
  else
   begin
    colPrior:= HTML_BGCOLOR_DATA_GRAY;
    colFinal:= HTML_BGCOLOR_DATA_GRAY;
   end;
end;

function TOpFactHarrisCA.getHTMLLoadings(loadings: PMatrix; nbFactors: integer; sorted: boolean; loadingsForSorting: PMatrix = nil): string;
var buf: TBufString;
    i,j: integer;
    sum,value: double;
    vExpl, finalCom: array of double;
    strColor: string;
    strPrior,strFinal: string;
    priorCom: double;
    index: TPCAIndexVariables;
begin

 if (sorted = false)
  then index:= self.getVariablesIndex()
  else
   begin
    if (loadingsForSorting = nil)
     then index:= self.sortOnLoadings(loadings,nbFactors)
     else index:= self.sortOnLoadings(loadingsForSorting,nbFactors);
   end;

 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Attribute</TH>');
 buf.AddStr(format('<TH colspan=2 %s>Communality Estimates</TH>',[HTML_BGCOLOR_COMMUNALITIES_PRIOR]));
 setLength(vExpl,succ(nbFactors));
 setLength(finalCom,succ(self.FDimAnalysis));
 for j:= 1 to nbFactors do
  begin
   buf.AddStr(format('<TH align=center colspan=2>Axis_%d</TH>',[j]));
   //en profiter pour calculer la variance explique "non-pondre"
   vExpl[j]:= 0.0;
   for i:= 1 to FDimAnalysis do
    begin
     value:= loadings[i,j] * loadings[i,j];
     vExpl[j]:= vExpl[j] + value;
     finalCom[i]:= finalCom[i] + value;
    end;
  end;
 buf.AddStr('</TR>');

 //rajouter la seconde ligne d'en-tte
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>-</TH>');
 buf.AddStr('<TH width=60>Prior</TH><TH width=60>Final</TH>');
 for j:= 1 to nbFactors do
  buf.AddStr('<TD align=center width=60>Corr.</TD><TD align=center width=70>Sq. (Cumul.)</TD>');
  
 //les valeurs
 for i:= 1 to FDimAnalysis do
  begin
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[Dataset.LstAtts[asInput].Attribute[pred(index[i])].Name]));
   //code couleur pour les communalities - on s'intresse uniquement aux variables lies aux autres
   getColorCommunalities(FPriorCommunality[index[i]],strPrior,strFinal);
   buf.AddStr(format('<TD align=right %s>%.5f</TD><TD align=right %s>%.5f</TD>',[strPrior,FPriorCommunality[index[i]],strFinal,finalCom[index[i]]]));
   //new -- 06/07/2006 -- calcul des communalities
   sum:= 0.0;
   for j:= 1 to nbFactors do
    begin
      TRY
      value:= loadings[index[i],j];//!\ pas besoin de rectification, toujours en analyse norme dans les mthodes associes
      sum:= sum + value * value;
      //mettre en valeur si % de la variable > % de l'axe
      strColor:= codeCouleurHTMLLoadings(value);
      //la ligne  afficher
      buf.AddStr(format('<TD align=right width=60 %s>%.5f</TD><TD align=right width=60><i>%.2f (%.2f)</i></TD>',[strColor,value,value*value,sum]));
      EXCEPT
      END;
    end;
   buf.AddStr('</TR>');
  end;
  
 //pourcentage expliqu par l'axe
 buf.AddStr(HTML_TABLE_COLOR_DATA_BLUE+format('<TH>%s</TH>',['Unweighted Var. Expl.']));
 priorCom:= Math.sum(FPriorCommunality);
 buf.AddStr(format('<TD align=right %s>%.5f</TD>',[HTML_BGCOLOR_COMMUNALITIES_PRIOR,priorCom]));
 buf.AddStr(format('<TD align=right %s>%.5f</TD>',[HTML_BGCOLOR_COMMUNALITIES_FINAL,Math.sum(vExpl)]));
 sum:= 0.0;
 for j:= 1 to nbFactors do
  begin
   sum:= sum + vExpl[j];
   buf.AddStr(format('<TD align=right width=60>%.5f</TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[vExpl[j],100.0*vExpl[j]/priorCom,'%',100.0*sum/priorCom,'%']));
  end;
 buf.AddStr('</TR>');
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
 Finalize(vExpl);
 Finalize(finalCom);
 Finalize(index);
end;

function TOpFactHarrisCA.getInstanceFA: TClassOpFactAnalysis;
begin
 result:= TOpFactHarrisCA;
end;

function TOpFactHarrisCA.getPrexifNameForProjection: string;
begin
 result:= 'Harris';
end;

initialization
 Classes.RegisterClass(TMLGenCompHarrisCA);
end.
