(********************************************************************)
(* UCompFactAnalysisPFA.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Non iterative Principal Factor Analysis)
@author(Ricco)
@created(07/09/2012)

Principal Factor Analysis (non iterative)
Traduite en Franais avec "Analyse en facteurs communs et spcifiques"

}

unit UCompFactAnalysisPFA;

interface

USES
        Forms,Classes,
        IniFiles,
        Matrices,
        UCalcStatDes,
        UCompDefinition,
        UCompFADefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UCompFactAnalysisPCA;

TYPE
        {gnrateur de composant ACP}
        TMLGenCompFactPFA = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant PFA}
        TMLCompFactPFA = class(TMLCompFactPCA)
        protected
        //obtenir la classe de calcul
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur PFA, on travaille sur les INPUT}
        TOpFactPFA = class(TOpFactPCA)
        protected
        //Uniqueness
        FUniqueness: array of double;
        //matrix to diagonalize
        FMatToDiag: PMatrix;
        //dtruire les matrices et vecteurs
        procedure   DelPrivateMatrix(); override;
        //construire la matrice  diagonaliser
        procedure   BuildVCV(); override;
        //calcul des valeurs et des vecteurs propres
        procedure   BuildEigen(); override;
        //calculer les coefficients de projection
        procedure   CalcFactorScores(); override;
        //obtenir la classe de paramtrage
        function    getClassParameter: TClassOperatorParameter; override;
        //pas de Bartlett ici
        procedure   calcBartlett(); override;
        //prfixe de nom de variable en projection
        function    getPrexifNameForProjection(): string; override;
        public
        //new - 07/09/2012 - obtenir les coefficients aprs rotation (varimax ou quartimax)
        function    getRotatedCoefs(newLoadings, newEgVec: PMatrix; nbFactors: integer): PMatrix; override;
        //no bartlett
        function    getHTMLBartlett(): string; override;
        //a na plus lieu d'tre ici - test de significativit des axes
        function    getHTMLSignificance(): string; override;
        //affichage des loadings
        function    getHTMLLoadings(loadings: PMatrix; nbFactors: integer; sorted: boolean; loadingsForSorting: PMatrix = nil): string; override;
        //affichage des coefs, avec les carrs multiples corrlations des variables avec les facteurs
        function    getHTMLRegCoef(factorCoefs: PMatrix; loadings: PMatrix; nbFactors: integer): string; override;
        //obtenir la classe
        function    getInstanceFA(): TClassOpFactAnalysis; override; 
        end;

        TOpPrmFactPFA = class(TOpPrmFactPCA)
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        public
        function    getHTMLParameters(): string; override;
        end;

implementation

uses
    SysUtils, MATH,
    UCalcMatrixToAttributes, UCalcMatrixAdditionalFunctions,
    UConstConfiguration, UStringAddBuffered, UDlgOpPrmFactFA;

const
    HTML_BGCOLOR_COMMUNALITIES_PRIOR : string = 'BGCOLOR="#E6E296"';
    HTML_BGCOLOR_COMMUNALITIES_FINAL : string = 'BGCOLOR="#EEEEC6"';

    HTML_BGCOLOR_SQUARED_CORRELATION : string = 'BGCOLOR="#EEEEC6"';

{ TMLGenCompFactPFA }

procedure TMLGenCompFactPFA.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactPFA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactPFA;
end;

{ TMLCompFactPFA }

function TMLCompFactPFA.getClassOperator: TClassOperator;
begin
 result:= TOpFactPFA;
end;

{ TOpFactPFA }

procedure TOpFactPFA.BuildEigen;
var nb,k: integer;
begin
 //on redfinit compltement la chose
 GetEigensFromSymetricMatrix(FMatToDiag,FDimAnalysis,FEigVal,FEgVec);
 //dtecter le nombre de val. propres positives
 nb:= 0;
 for k:= 1 to FDimAnalysis do
  if (FEigVal^[k] > 0) then inc(nb) else break;
 //redimensionner le nombre d'axes  produire
 FNbAxis:= Math.Min(FNbAxis,nb);
end;

procedure TOpFactPFA.BuildVCV;
var resInv: integer;
    value: double;
    k: integer;
begin
 //on obtient la matrice de corrlation
 inherited;
 //on passe  l'inverse
 DimMatrix(FInvMatCorr,FDimAnalysis,FDimAnalysis);
 resInv:= InvMat(MatCorr,1,FDimAnalysis,FInvMatCorr);
 if (resInv = MAT_SINGUL)
  then
   begin
    //on passe par la SVDInverse
    CopyMatrix(FInvMatCorr,MatCorr,1,1,FDimAnalysis,FDimAnalysis);
    SVDInverse(FInvMatCorr,FDimAnalysis);
   end;
 //si on est l, c'est que tout va bien
 //matrice  diagonaliser
 DimMatrix(FMatToDiag,FDimAnalysis,FDimAnalysis);
 CopyMatrix(FMatToDiag,MatCorr,1,1,FDimAnalysis,FDimAnalysis);
 //les communalits
 setLength(FUniqueness,FDimAnalysis+1);
 setLength(FPriorCommunality,FDimAnalysis+1);
 FTrace:= 0.0;//recalculer la trace
 for k:= 1 to FDimAnalysis do
  begin
   value:= FInvMatCorr^[k]^[k];
   if (abs(value) > 0)
    then FUniqueness[k]:= 1.0/value;
   FPriorCommunality[k]:= 1.0-FUniqueness[k];
   FTrace:= FTrace + FPriorCommunality[k];
   //modifier la matrice  diagonaliser
   FMatToDiag^[k]^[k]:= FPriorCommunality[k];
  end;
 //pour dsactiver le test sur le dterminant
 FDetR:= 1.0;
end;

procedure TOpFactPFA.calcBartlett;
begin
 //none
end;

procedure TOpFactPFA.CalcFactorScores;
begin
 //produit matriciel simple
 MatrixProd(FInvMatCorr,FDimAnalysis,FDimAnalysis,FLoadings,FDimAnalysis,FDimAnalysis,FFactorScores);
end;

procedure TOpFactPFA.DelPrivateMatrix;
begin
 inherited;
 if assigned(FMatToDiag) then delMatrix(FMatToDiag,FDimAnalysis,FDimAnalysis);
 if assigned(FUniqueness) then Finalize(FUniqueness);
end;

function TOpFactPFA.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactPFA;
end;

function TOpFactPFA.getHTMLBartlett: string;
begin
 result:= '';
end;

procedure getColorCommunalities(value: double; var colPrior,colFinal: string);
begin
 if (value > 0.5)
  then
   begin
    colPrior:= HTML_BGCOLOR_COMMUNALITIES_PRIOR;
    colFinal:= HTML_BGCOLOR_COMMUNALITIES_FINAL;
   end
  else
   begin
    colPrior:= HTML_BGCOLOR_DATA_GRAY;
    colFinal:= HTML_BGCOLOR_DATA_GRAY;
   end;
end;

function TOpFactPFA.getHTMLLoadings(loadings: PMatrix; nbFactors: integer; sorted: boolean; loadingsForSorting: PMatrix = nil): string;
var buf: TBufString;
    i,j: integer;
    sum,value: double;
    vExpl, finalCom: array of double;
    strColor: string;
    strPrior,strFinal: string;
    index: TPCAIndexVariables;
begin

 if (sorted = false)
  then index:= self.getVariablesIndex()
  else
   begin
    if (loadingsForSorting = nil)
     then index:= self.sortOnLoadings(loadings,nbFactors)
     else index:= self.sortOnLoadings(loadingsForSorting,nbFactors);
   end;

 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Attribute</TH>');
 buf.AddStr(format('<TH colspan=2 %s>Communality Estimates</TH>',[HTML_BGCOLOR_COMMUNALITIES_PRIOR]));
 setLength(vExpl,succ(nbFactors));
 setLength(finalCom,succ(self.FDimAnalysis));
 for j:= 1 to nbFactors do
  begin
   buf.AddStr(format('<TH align=center colspan=2>Axis_%d</TH>',[j]));
   //en profiter pour calculer la variance explique "non-pondre"
   vExpl[j]:= 0.0;
   for i:= 1 to FDimAnalysis do
    begin
     value:= loadings[i,j] * loadings[i,j];
     vExpl[j]:= vExpl[j] + value;
     finalCom[i]:= finalCom[i] + value;
    end;
  end;
 buf.AddStr('</TR>');
 //rajouter la seconde ligne d'en-tte
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>-</TH>');
 buf.AddStr('<TH width=60>Prior</TH><TH width=60>Final</TH>');
 for j:= 1 to nbFactors do
  buf.AddStr('<TD align=center width=60>Corr.</TD><TD align=center width=70>Sq. (Cumul.)</TD>');

 //les valeurs
 for i:= 1 to FDimAnalysis do
  begin
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[Dataset.LstAtts[asInput].Attribute[pred(index[i])].Name]));
   //code couleur pour les communalities - on s'intresse uniquement aux variables lies aux autres
   getColorCommunalities(FPriorCommunality[index[i]],strPrior,strFinal);
   buf.AddStr(format('<TD align=right %s>%.5f</TD><TD align=right %s>%.5f</TD>',[strPrior,FPriorCommunality[index[i]],strFinal,finalCom[index[i]]]));
   //new -- 06/07/2006 -- calcul des communalities
   sum:= 0.0;
   for j:= 1 to nbFactors do
    begin
      TRY
      value:= loadings[index[i],j];//!\ pas besoin de rectification, toujours en analyse norme dans les mthodes associes
      sum:= sum + value * value;
      //mettre en valeur si % de la variable > % de l'axe
      strColor:= codeCouleurHTMLLoadings(value);
      //la ligne  afficher
      buf.AddStr(format('<TD align=right width=60 %s>%.5f</TD><TD align=right width=60><i>%.2f (%.2f)</i></TD>',[strColor,value,value*value,sum]));
      EXCEPT
      END;
    end;
   buf.AddStr('</TR>');
  end;
  
 //pourcentage expliqu par l'axe
 buf.AddStr(HTML_TABLE_COLOR_DATA_BLUE+format('<TH>%s</TH>',['Var. Expl.']));
 buf.AddStr(format('<TD align=right %s>%.5f</TD>',[HTML_BGCOLOR_COMMUNALITIES_PRIOR,FTrace]));
 sum:= 0.0;
 for j:= 1 to nbFactors do sum:= sum + vExpl[j];
 buf.AddStr(format('<TD align=right %s>%.5f</TD>',[HTML_BGCOLOR_COMMUNALITIES_FINAL,sum]));
 sum:= 0.0;
 for j:= 1 to nbFactors do
  begin
   sum:= sum + vExpl[j];
   buf.AddStr(format('<TD align=right width=60>%.5f</TD><TD align=right width=60><i>%0.0f %s (%0.0f %s)</i></TD>',[vExpl[j],100.0*vExpl[j]/FTrace,'%',100.0*sum/FTrace,'%']));
  end;
 buf.AddStr('</TR>');
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
 Finalize(vExpl);
 Finalize(finalCom);
 Finalize(index);
end;

function TOpFactPFA.getHTMLRegCoef(factorCoefs: PMatrix; loadings: PMatrix; 
  nbFactors: integer): string;
var  buf: TBufString;
     s: string;
     i,j: integer;
     sqMC: array of double;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 //
 buf.AddStr(HTML_HEADER_TABLE_RESULT);
 //les carrs des corrlations
 buf.AddStr(format('<TR %s><TH colspan=3>Squared Multiple Corr. of the Variables with Each Factor</TH>',[HTML_BGCOLOR_SQUARED_CORRELATION]));
 //calculs
 setLength(sqMC,nbFactors+1);
 for j:= 1 to nbFactors do
  for i:= 1 to FDimAnalysis do
   sqMC[j]:= sqMC[j] + factorCoefs[i,j]*loadings[i,j];
 //affichage
 for j:= 1 to nbFactors do
  buf.AddStr(format('<TD align=right>%.7f</TD>',[sqMC[j]]));
 Finalize(sqMC);
 buf.AddStr('</TR>');
 //en-tte pour chaque axe factoriel demand
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=150>Attribute</TH>');
 //new -- 06/07/2006 -- ajouter la moyenne et l'cart-type
 s:= format('<TH width=80 align=right %s>Mean</TH><TH width=80 align=right %s>Std-dev</TH>',[HTML_BGCOLOR_HEADER_GREEN,HTML_BGCOLOR_HEADER_GREEN]);
 for j:= 1 to nbFactors do
  s:= s+format('<TH width=80 align=right>Axis_%d</TH>',[j]);
 s:= s+'</TR>';
 //le contenu
 for i:= 1 to FDimAnalysis do
  begin
   s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[Dataset.LstAtts[asInput].Attribute[pred(i)].Name]);
   //new -- 06/07/2006 -- ajouter la moyenne et l'cart-type
   s:= s + format('<TD align=right %s>%.7f</TD>',[HTML_BGCOLOR_DATA_GREEN,TCalcStatDesContinuous(FLstStats.Stat(pred(i))).Average]);
   //s:= s + format('<TD align=right %s>%.6f</TD>',[HTML_BGCOLOR_DATA_GREEN,TCalcStatDesContinuous(FLstStats.Stat(pred(i))).StdDevPop]);
   //new -- 30/06/2011
   if ((self.PrmOp as TOpPrmFactPCA).Standardize = false)
    then s:= s + format('<TD align=right %s>%.1f</TD>',[HTML_BGCOLOR_DATA_GREEN,1.0])
    else s:= s + format('<TD align=right %s>%.7f</TD>',[HTML_BGCOLOR_DATA_GREEN,TCalcStatDesContinuous(FLstStats.Stat(pred(i))).StdDevPop]);
   //suite...
   for j:= 1 to nbFactors do
    begin
      TRY
      s:= s+format('<TD align=right>%.7f</TD>',[factorCoefs^[i]^[j]]);
      EXCEPT
      END;
    end;
   s:= s+'</TR>';
  end;
 s:= s+'</table>';
 //dans le buffer...
 buf.AddStr(s);
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TOpFactPFA.getHTMLSignificance: string;
begin
 result:= '';
end;

function TOpFactPFA.getInstanceFA: TClassOpFactAnalysis;
begin
 result:= TOpFactPFA;
end;

function TOpFactPFA.getPrexifNameForProjection: string;
begin
 result:= 'PFA';
end;

function TOpFactPFA.getRotatedCoefs(newLoadings, newEgVec: PMatrix;
  nbFactors: integer): PMatrix;
var fScores: PMatrix;
begin
 //produit matriciel simple
 MatrixProd(FInvMatCorr,FDimAnalysis,FDimAnalysis,newLoadings,FDimAnalysis,FDimAnalysis,fScores);
 result:= fScores;
end;

{ TOpPrmFactPFA }

function TOpPrmFactPFA.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFactFA.CreateFromOpPrm(self);
end;

function TOpPrmFactPFA.getHTMLParameters: string;
var s: string;
begin
 s:= '<P><B>Number of asked factors : </B>';
 if (NbAxisCreated<0)
  then s:= s+'all'
  else s:= s+IntToStr(NbAxisCreated);
 s:= s + '<br><B>Correlations and partial correlations : </B>' + IntToStr(ord(ShowCorr));
 s:= s + '<br><B>Reproduced correlations : </B>' + IntToStr(ord(ShowRepCorr));
 s:= s + '<br><B>Sort variables according to loadings : </B>' + IntToStr(ord(SortVarOnLoadings));
 result:= s;
end;

procedure TOpPrmFactPFA.SetDefaultParameters;
begin
 inherited;
 self.Standardize:= true;//non modifiable
 self.Cos2Ctr:= false;//non modifiable non plus
end;

initialization
 RegisterClass(TMLGenCompFactPFA);
end.
