(*****************************************************************)
(* UCompFADefinition.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(*****************************************************************)

{
@abstract(Definition de la classe d'objet "Analyse factorielle")
@author(Ricco)
@created(12/01/2004)
Utilisation intensive des mthodes matricielles. Bcp d'optimisations  prvoir.
}
unit UCompFADefinition;

interface

USES
        UCompManageDataset,
        UOperatorDefinition,
        UDatasetImplementation,
        Matrices;

TYPE
        {Classe gnrique de composant analyse factorielle}
        TMLCompFactAnalysis = class(TMLCompLocalData)
                              protected
                              function    GetLogResultDescription(): string; override; 
                              end;

        //dclaration forward
        TOpFactAnalysis = class;

        //type classe d'oprateur d'analyse factorielle
        TClassOpFactAnalysis = class of TOpFactAnalysis;

        {Classe gnrique d'oprateur Fact Analysis,
        la gestion des donnes gnres localement va tre trs importante,
        le nombre de ces nouvelles variables sera constamment modifie selon les paramtres de l'analyse}
        TOpFactAnalysis = class(TOpLocalData)
                          private
                          //new - 08/06/2012 - reprsente le jeu de donnes
                          FDataset: TMLDataset;
                          protected
                          //new - 08/06/2012 - tableau des valeurs propres
                          FEigVal: PVector;
                          procedure   ReInitialize(); override;
                          procedure   DelPrivateMatrix(); virtual;
                          //new - 08/06/2012 - pour obtenir la dimension du vecteur FEigVal
                          function    getDimEigVal(): integer; virtual; abstract;
                          //new - 08/06/2012 - lire la valeur propre i
                          function    getEigVal(i: integer): double;
                          //new - 08/06/2012 - crire la valeur propre i
                          procedure   setEigVal(i: integer; value: double);
                          //new - 08/06/2012 - excution standardise
                          function    CoreExecute(): boolean; override;
                          public
                          constructor Create(AOwner: TObject); override;
                          constructor CreateWithDataset(AOwner: TObject; prmDataset: TMLDataset); virtual;
                          destructor  Destroy(); override;
                          //new - 08/06/2012 - dcomposer les calculs en 3 tapes
                          function    calcEigen(): boolean; virtual; abstract;
                          function    calcAdditional(): boolean; virtual; abstract;
                          function    calcProjections(): boolean; virtual; abstract;
                          //new -- 24/07/2012 -- calcul post-projection
                          function    calcPostProjections(): boolean; virtual;
                          //new - 08/06/2012 - obtenir une instance d'analyse factorielle
                          function    getInstanceFA(): TClassOpFactAnalysis; virtual; abstract; 
                          //proprits
                          property    EigVal[i: integer]: double read getEigVal write setEigVal;
                          property    EigValCount: integer read getDimEigVal;
                          property    Dataset: TMLDataset read FDataset;
                          end;

        {paramtrage d'une analyse factorielle}
        TOpPrmFactAnalysis = class(TOperatorParameter)
                             public
                             procedure getFactSettings(prmSource: TOpPrmFactAnalysis); virtual;
                             end;

        //new -- 22/11/2012 -- un type numr pour les codes couleurs dans les affichages
        COLOR_TYPE_FOR_HIGHLIGHT = (cthNegHigh,cthNegLow,cthPosLow,cthPosHigh);

const
        //couleurs pour la mise en vidence des valeurs
        FACT_COLOR_HIGHLIGHT : array[COLOR_TYPE_FOR_HIGHLIGHT] of string =
                               ('BGCOLOR="#FBFD75"','BGCOLOR="#F8FAB7"','BGCOLOR="#F7E1B2"','BGCOLOR="#FCC998"'); 


implementation

uses
   Sysutils, Math;

{ TOpFactAnalysis }

function TOpFactAnalysis.calcPostProjections: boolean;
begin
 result:= true;
end;

function TOpFactAnalysis.CoreExecute: boolean;
var ok: boolean;
begin
 ok:= self.calcEigen();
 ok:= ok and self.calcAdditional();
 ok:= ok and self.calcProjections();
 ok:= ok and self.calcPostProjections();
 result:= ok;
end;

constructor TOpFactAnalysis.Create(AOwner: TObject);
begin
 inherited Create(AOwner);
 FDataset:= self.WorkData;
end;

constructor TOpFactAnalysis.CreateWithDataset(AOwner: TObject; prmDataset: TMLDataset);
begin
 inherited create(AOwner);
 FDataset:= prmDataset;
end;

procedure TOpFactAnalysis.DelPrivateMatrix;
begin
 delVector(FEigVal,self.getDimEigVal());
end;

destructor TOpFactAnalysis.Destroy;
begin
 self.DelPrivateMatrix();
 inherited Destroy();
end;

function TOpFactAnalysis.getEigVal(i: integer): double;
begin
 if assigned(FEigVal) and (i > 0) and (i <= self.getDimEigVal())
  then result:= FEigVal^[i]
  else result:= Math.NaN;
end;

procedure TOpFactAnalysis.ReInitialize;
begin
 inherited ReInitialize();
 self.DelPrivateMatrix();
end;

procedure TOpFactAnalysis.setEigVal(i: integer; value: double);
begin
 if (i > 0) and (i <= self.getDimEigVal())
  then FEigVal^[i]:= value;
end;

{ TMLCompFactAnalysis }

function TMLCompFactAnalysis.GetLogResultDescription: string;
var data: TMLDataset;
begin
 data:= (Operator as TOpLocalData).WorkData;
 result:= Format('%d attributes, and %d examples',[data.LstAtts[asInput].Count,data.Examples.Size]);
end;

{ TOpPrmFactAnalysis }

procedure TOpPrmFactAnalysis.getFactSettings(
  prmSource: TOpPrmFactAnalysis);
begin
 //rien...
end;

end.
