(******************************************************************)
(* UCompFactBootstrap.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Bootstrap pour le calcul des intervalles de confiance des valeurs propres d'une analyse factorielle)
@author(Ricco)
@created(10/06/2012)

Bootstrap pour la dtection du nombre de facteurs significatifs en Analyse Factorielle.
Applicable  l'ACP, mais aussi  l'ACM (pas l'AFC qui est trs particulire)

Procde par bootstrap, les dcalages significatifs sont indiqus dans le tableau des rsultats,
uniquement si la borne basse est suprieure  1.

PARAMETRES :

> Le nombre d'essais est paramtrable
> Le niveau de confiance est paramtrable
> Gnrateur de nombre alatoires paramtrable.
> Construction du graphique en plus du tableau de valeurs.
}

unit UCompFactBootstrap;

interface

USES
    Forms,Classes,
    IniFiles,Contnrs,
    UCompDefinition,
    UOperatorDefinition,
    UCompManageDataset,
    UDatasetDefinition,
    UDatasetImplementation,
    UDatasetExamples,
    UCalcRndGenerator,
    UCompFADefinition,
    Matrices,
    UCompFactParallelAnalysis;


TYPE
        {gnrateur de composant Parallel Analysis}
        TMLGenCompFactBootstrap = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant Parallel Analysis}
        TMLCompFactBootstrap = class(TMLCompLocalData)
        protected
        function    getClassOperator: TClassOperator; override;
        public
        //n'est connectable que sur certains composants (ACP, AFC, et ACM pour l'instant)
        class function isConnectableTo(destination: TMLComponent): boolean; override;
        end;

        {oprateur}
        TOpFactBootstrap = class(TOpFactPA)
        private
        //bornes basses
        FRefEigensLo: PVector;
        //valeur de rfrence pour les valeurs propres : moyenne des valeurs
        FMeanEigen: double;
        protected
        //calculer les analyses factorielles, en grand nombre
        function    calcFactorials(): boolean; override;
        //calculer le rcapitulatif des valeurs propres
        function    calcRecaps(): boolean; override;
        //les fonctions hrites
        function    getClassParameter: TClassOperatorParameter; override;
        function    getNewForm(): TForm; override;
        public
        //rcuprer au passage l'oprateur FACT
        destructor  destroy(); override;
        function    getHTMLResultsSummary(): string; override;
        function    getRefEigenLo(i: integer): double;
        end;

        TOpPrmFactBootstrap = class(TOpPrmFactPA)
        protected
        procedure   SetDefaultParameters(); override;
        end;


implementation

uses
    UCompFactAnalysisAFC, UCompFactAnalysisPCA, UCompFactAnalysisMCA,
    ULogFile, SysUtils, StStat, Math, UConstConfiguration,
    UFrmOpViewFactBootstrap, UFrmBaseOperator, UCompFactAnalysisAFDM;

{ TMLGenCompFactBootstrap }

procedure TMLGenCompFactBootstrap.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactBootstrap.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactBootstrap;
end;

{ TMLCompFactBootstrap }

function TMLCompFactBootstrap.getClassOperator: TClassOperator;
begin
 result:= TOpFactBootstrap;
end;

class function TMLCompFactBootstrap.isConnectableTo(
  destination: TMLComponent): boolean;
var ok: boolean;
begin
 ok:= destination is TMLCompFactPCA;
 ok:= ok or (destination is TMLCompFactMCA);
 //new -- 25/07/2012 - AFDM
 ok:= ok or (destination is TMLCompFactAFDM);
 result:= ok;
end;

{ TOpFactBootstrap }

function TOpFactBootstrap.calcFactorials: boolean;
var tmpEx, workEx: TExamples;
    ok: boolean;
    i,k: integer;
    Fact: TOpFactAnalysis;
begin
 //individus
 tmpEx:= TExamples.Create(FOpFact.Dataset.Examples.Size);
 tmpEx.Copy(FOpFact.Dataset.Examples);
 workEx:= TExamples.Create(FOpFact.Dataset.Examples.Size);
 //commencer les itrations
 ok:= true;
 //prparer l'objet de calcul
 Fact:= FOpFact.getInstanceFA().CreateWithDataset(FOpFact.MLOwner,FOpFact.Dataset);
 (Fact.PrmOp as TOpPrmFactAnalysis).getFactSettings(FOpFact.PrmOp as TOpPrmFactAnalysis);
 TRY
 //lancer les itrations
 for i:= 1 to (self.PrmOp as TOpPrmFactPA).NbReplication do
  begin
   //mlanger alatoirement les donnes
   tmpEx.SampleReplicate(workEx,FRnd);
   Fact.Dataset.Examples.Copy(workEx);
   //lancer le calcul des valeurs propres
   ok:= Fact.calcEigen();
   if (ok = true)
    then
     begin
      //recopier les valeurs
      for k:= 1 to Fact.EigValCount do
       FEigens[k-1][i-1]:= Fact.EigVal[k];
     end
    else BREAK;//on arrte l...
  end;
 EXCEPT
 ok:= false;
 END;
 FOpFact.Dataset.Examples.Copy(tmpEx);
 workEx.Free();
 tmpEx.Free();
 result:= ok;
end;

function TOpFactBootstrap.calcRecaps: boolean;
var prm: TOpPrmFactBootstrap;
    k: integer;
    pLo,pHi: double;
begin
 result:= true;
 TRY
 //vecteur pour les bornes basses
 if assigned(FRefEigensLo) then delVector(FRefEigensLo, FOpFact.EigValCount);
 dimVector(FRefEigensLo,FOpFact.EigValCount);
 //paramtres
 prm:= self.PrmOp as TOpPrmFactBootstrap;
 pLo:= (1.0 - prm.Percentile)/2.0;
 pHi:= prm.Percentile + pLo;
 //calculs, on en profite pour calculer la moyenne des valeurs propres observes
 FMeanEigen:= 0.0;
 for k:= 1 to FOpFact.EigValCount do
  begin
   FRefEigensLo^[k]:= StStat.Percentile(FEigens[k-1],pLo);
   FRefEigens^[k]:= StStat.Percentile(FEigens[k-1],pHi);
   FMeanEigen:= FMeanEigen + FOpFact.EigVal[k];
  end;
 If (FOPFact.EigValCount > 0) then FMeanEigen:= FMeanEigen / (1.0 * FOPFact.EigValCount); 
 EXCEPT
 result:= false;
 END;
end;

destructor TOpFactBootstrap.destroy;
begin
 if assigned(FRefEigensLo) then delVector(FRefEigensLo, FOpFact.EigValCount);
 inherited;
end;

function TOpFactBootstrap.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactBootstrap;
end;

CONST
    CODE_COULEUR : array[0..1] of String = (HTML_BGCOLOR_LIGHT_BLUE,HTML_BGCOLOR_LIGHT_GREEN);

function TOpFactBootstrap.getHTMLResultsSummary: string;
var s: string;
    k: integer;
    colorLo,ColorHi: string;
    colorFontLo,colorFontHi: string;
    testColor: boolean;
    pLo,pHi: double;
    prm: TOpPrmFactBootstrap;
begin
 s:= '<H3>Bootstrap Confidence Intervals for Eigenvalues</H3>';
 s:= s + format('<b>Average of observed eigenvalues</b> = %.6g<br><br>',[FMeanEigen]);
 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY;
 prm:= self.PrmOp as TOpPrmFactBootstrap;
 pLo:= (1.0 - prm.Percentile)/2.0;
 pHi:= prm.Percentile + pLo;
 s:= s + format('<TH>Component</TH><TH>(%.3g) Lower bound</TH><TH>(%.3g) Upper bound</TH></TR>',[pLo,pHi]);
 testColor:= true;
 for k:= 1 to FOpFact.EigValCount do
  begin
   colorLo:= HTML_BGCOLOR_DATA_GRAY;
   colorHi:= HTML_BGCOLOR_DATA_GRAY;
   //doit-on encore coloriser
   if (testColor = true)
    then
     begin
      if (k < FOpFact.EigValCount) and (FRefEigensLo^[k] > FRefEigens^[k+1])
       then colorLo:= CODE_COULEUR[k mod 2];
      if (k > 1) and (FRefEigensLo^[k-1] > FRefEigens^[k])
       then colorHi:= CODE_COULEUR[pred(k) mod 2];
     end;
   //si pas de modif, on invalide la gestion de la couleur pour le dcalage
   if (colorLo = HTML_BGCOLOR_DATA_GRAY) AND (colorHi = HTML_BGCOLOR_DATA_GRAY) then testColor:= false;
   //tester si la borne basse est suprieure  1
   if (FRefEigensLo^[k] >= FMeanEigen) then colorFontLo:= '"BLACK"' else colorFontLo:= '"GRAY"';
   if (FRefEigens^[k] >= FMeanEigen) then colorFontHi:= '"BLACK"' else colorFontHi:= '"GRAY"';
   //dessiner la ligne
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TH>%d</TH><TD align=center %s><font color=%s>%.6f</font></TD><TD align=center %s><font color=%s>%.6f</font></TD></TR>',[k,colorLo,colorFontLo,FRefEigensLo^[k],colorHi,colorFontHi,FRefEigens^[k]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TOpFactBootstrap.getNewForm: TForm;
begin
 result:= TFrmOpViewFactBootstrap.CreateFromOperator(self);
end;

function TOpFactBootstrap.getRefEigenLo(i: integer): double;
begin
 result:= FRefEigensLo^[i];
end;

{ TOpPrmFactBootstrap }

procedure TOpPrmFactBootstrap.SetDefaultParameters;
begin
 NbReplication:= 200;
 Percentile:= 0.90;//pour le calcul de l'intervalle de confiance
 SeedRnd:= 1;
end;

initialization
 CLASSES.RegisterClass(TMLGenCompFactBootstrap);
end.
