(*********************************************************************)
(* UCompParallelAnalysis.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(*********************************************************************)

{
@abstract(Parallel Analysis for the determination of the significant factors)
@author(Ricco)
@created(08/06/2012)

Analyse parallle pour la dtection du nombre de facteurs significatifs en Analyse Factorielle.
Applicable  l'ACP, mais aussi  l'ACM (pas l'AFC qui est trs particulire)

Procde par procdure de Monte-Carlo sur les donnes rshuffles.

PARAMETRES :

Le nombre d'essais est paramtrable

Le seuil de significativit est paramtrable.
> Dfaut, quantile 0.95, mais valeur modifiable.
> On peut aussi prendre la moyenne.

Gnrateur de nombre alatoires paramtrable.

Construction du graphique en plus du tableau de valeurs.
}

unit UCompFactParallelAnalysis;

interface

USES
    Forms,Classes,
    IniFiles,Contnrs,
    UCompDefinition,
    UOperatorDefinition,
    UCompManageDataset,
    UDatasetDefinition,
    UDatasetImplementation,
    UDatasetExamples,
    UCalcRndGenerator,
    UCompFADefinition,
    Matrices;


TYPE
        {gnrateur de composant Parallel Analysis}
        TMLGenCompFactPA = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant Parallel Analysis}
        TMLCompFactPA = class(TMLCompLocalData)
        protected
        function    getClassOperator: TClassOperator; override;
        public
        //n'est connectable que sur certains composants (ACP, AFC, et ACM pour l'instant)
        class function isConnectableTo(destination: TMLComponent): boolean; override;
        end;

        //valeurs observes d'une valeur propre
        TTabEigens = array of double;
        //valeurs observes pour l'ensemble des valeurs propres
        TArrayTabEigens = array of TTabEigens;

        {oprateur}
        TOpFactPA = class(TOpLocalData)
        protected
        //l'oprateur sur lequel le composant est branch
        FOpFact: TOpFactAnalysis;
        //tableau des valeurs collectes
        FEigens: TArrayTabEigens;
        //valeurs de rfrences
        FRefEigens: PVector;
        //gnrateur de donnes alatoires
        FRnd: TRndGenerator;
        //cration des structures internes de calcul
        procedure   createStructures();
        //suppression des structures internes
        procedure   destroyStructures();
        //calculer les analyses factorielles, en grand nombre
        function    calcFactorials(): boolean; virtual;
        //calculer le rcapitulatif des valeurs propres
        function    calcRecaps(): boolean; virtual;
        //les fonctions hrites
        function    getClassParameter: TClassOperatorParameter; override;
        function    CoreExecute(): boolean; override;
        function    getNewForm(): TForm; override;
        public
        //rcuprer au passage l'oprateur FACT
        constructor Create(AOwner: TObject); override;
        destructor  destroy(); override;
        function    getHTMLResultsSummary(): string; override;
        function    getDimEigenValue(): integer;
        function    getObservedEigen(i: integer): double;
        function    getRefEigen(i: integer): double;
        end;

        TOpPrmFactPA = class(TOperatorParameter)
        protected
        function    CreateDlgParameters(): TForm; override;
        procedure   SetDefaultParameters(); override;
        public
        NbReplication: integer;//nombre de rplication, df. 100
        Percentile: double;//pourcentage pour le quantile, df. 0.95
        SeedRnd: integer;//valeur de dpart pour le gnrateur de nombres alatoires
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        function    getHTMLParameters(): string; override;
        end;

implementation

uses
    UCompFactAnalysisAFC, UCompFactAnalysisPCA, UCompFactAnalysisMCA,
    ULogFile, SysUtils, StStat, Math, UConstConfiguration,
    UFrmOpViewFactParallelAnalysis, UFrmBaseOperator,
  UDlgOpPrmFactParallelAnalysis, UCompFactAnalysisAFDM;

{ TMLGenCompFactPA }

procedure TMLGenCompFactPA.GenCompInitializations;
begin
 FMLComp:= mlcFactorialAnalysis;
end;

function TMLGenCompFactPA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFactPA;
end;

{ TMLCompFactPCA }

function TMLCompFactPA.getClassOperator: TClassOperator;
begin
 result:= TOpFactPA;
end;

class function TMLCompFactPA.isConnectableTo(
  destination: TMLComponent): boolean;
var ok: boolean;
begin
 ok:= destination is TMLCompFactPCA;
 ok:= ok or (destination is TMLCompFactMCA);
 //new -- 25/07/2012 - AFDM
 ok:= ok or (destination is TMLCompFactAFDM);
 result:= ok;
end;

{ TOpFactPA }

//11/06/2012 - fonctionne pour l'ACP et l'ACM 
function TOpFactPA.calcFactorials(): boolean;
var Fact: TOpFactAnalysis;
    i,k: integer;
    ok: boolean;
    dataset: TMLDataset;
begin
 //dupliquer les donnes de travail
 dataset:= FOpFact.Dataset.duplicate();
 //begin
 ok:= true;
 TRY
 //lancer les itrations
 for i:= 1 to (self.PrmOp as TOpPrmFactPA).NbReplication do
  begin
   //mlanger alatoirement les donnes
   dataset.shuffle(FRnd);
   //prparer l'objet de calcul
   Fact:= FOpFact.getInstanceFA().CreateWithDataset(FOpFact.MLOwner,dataset);
   (Fact.PrmOp as TOpPrmFactAnalysis).getFactSettings(FOpFact.PrmOp as TOpPrmFactAnalysis);
   //lancer le calcul des valeurs propres
   ok:= Fact.calcEigen();
   if (ok = true)
    then
     begin
      //recopier les valeurs
      for k:= 1 to Fact.EigValCount do
       FEigens[k-1][i-1]:= Fact.EigVal[k];
      //contrle
      //TraceLog.WriteToLogFile(format('>>%.6f<<',[FEigens[0][i-1]]));
     end
    else BREAK;//on arrte l...
   //on dtruit compltement l'objet de calcul
   Fact.Free();
  end;
 EXCEPT
 ok:= false;
 END;
 result:= ok;
end;

(*
>> 11/06/2012 << cette solution n'est pas bonne pour l'ACM, les tableaux internes ne sont pas rinitialiss correctement
function TOpFactPA.calcFactorials(): boolean;
var Fact: TOpFactAnalysis;
    i,k: integer;
    ok: boolean;
    //jeu de donnes local
    dataset: TMLDataset;
begin
 //dupliquer les donnes de travail
 dataset:= FOpFact.Dataset.duplicate();
 //commencer les itrations
 ok:= true;
 //prparer l'objet de calcul
 Fact:= FOpFact.getInstanceFA().CreateWithDataset(FOpFact.MLOwner,dataset);
 (Fact.PrmOp as TOpPrmFactAnalysis).getFactSettings(FOpFact.PrmOp as TOpPrmFactAnalysis);
 TRY
 //lancer les itrations
 for i:= 1 to (self.PrmOp as TOpPrmFactPA).NbReplication do
  begin
   //mlanger alatoirement les donnes
   Fact.Dataset.shuffle(FRnd);
   //lancer le calcul des valeurs propres
   ok:= Fact.calcEigen();
   if (ok = true)
    then
     begin
      //recopier les valeurs
      for k:= 1 to Fact.EigValCount do
       FEigens[k-1][i-1]:= Fact.EigVal[k];
      //contrle
      TraceLog.WriteToLogFile(format('>>%.6f<<',[FEigens[0][i-1]]));
     end
    else BREAK;//on arrte l...
  end;
 EXCEPT
 ok:= false;
 END;
 Fact.Free();
 dataset.Free();
 result:= ok;
end;
*)

function TOpFactPA.calcRecaps: boolean;
var prm: TOpPrmFactPA;
    k: integer;
begin
 result:= true;
 TRY
 prm:= self.PrmOp as TOpPrmFactPA;
 //quantile
 for k:= 1 to FOpFact.EigValCount do
  FRefEigens^[k]:= StStat.Percentile(FEigens[k-1],prm.Percentile);
 EXCEPT
 result:= false;
 END;
end;

function TOpFactPA.CoreExecute: boolean;
var ok: boolean;
begin
 //gnrateur de nombre alatoire
 if assigned(FRnd) then FRnd.Free();
 FRnd:= TRndGenerator.Create(seedUser,DEFAULT_SEED_VALUE_1,(self.PrmOp as TOpPrmFactPA).SeedRnd);
 //prparer les structures de collecte
 self.createStructures();
 //le vectteur des valeurs de rfrence
 if assigned(FRefEigens) then delVector(FRefEigens, FOpFact.EigValCount);
 dimVector(FRefEigens,FOpFact.EigValCount);
 //lancer les itrations
 ok:= self.calcFactorials();
 //calculer le rcapitulatif par valeur propre
 ok:= ok and self.calcRecaps();
 //on n'a plus besoin des structures de collecte
 self.destroyStructures();
 //renvoyer le rsultat
 result:= ok;
end;

constructor TOpFactPA.Create(AOwner: TObject);
begin
 inherited;
 FOpFact:= (MLOwner as TMLComponent).Predecessor.Operator as TOpFactAnalysis;
end;

procedure TOpFactPA.createStructures;
var prm: TOpPrmFactPA;
    k: integer;
begin
 prm:= self.PrmOp as TOpPrmFactPA;
 setLength(Feigens,FopFact.EigValCount);
 for k:= 0 to pred(FopFact.EigValCount) do
  setLength(FEigens[k],prm.NbReplication);
end;

destructor TOpFactPA.destroy;
begin
 if assigned(FRefEigens) then delVector(FRefEigens, FOpFact.EigValCount);
 if assigned(FRnd) then FRnd.Free();
 inherited;
end;

procedure TOpFactPA.destroyStructures;
var k: integer;
begin
 for k:= 0 to pred(length(FEigens)) do
  finalize(FEigens[k]);
 finalize(FEigens);
end;

function TOpFactPA.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFactPA;
end;

function TOpFactPA.getDimEigenValue: integer;
begin
 result:= FOpFact.EigValCount;
end;

function TOpFactPA.getHTMLResultsSummary: string;
var s: string;
    k: integer;
    color: string;
begin
 s:= '<H3>Parallel Analysis</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + format('<TH>Component</TH><TH width=120>Eigenvalue</TH><TH width=120>(%.2g) Critical value</TH></TR>',[(self.PrmOp as TOpPrmFactPA).Percentile]);
 for k:= 1 to FOpFact.EigValCount do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   if (FopFact.EigVal[k] > FRefEigens^[k])
    then color:= HTML_BGCOLOR_DATA_RED
    else color:= HTML_BGCOLOR_DATA_GRAY;
   s:= s + format('<TH>%d</TH><TD %s align=center>%.6f</TD><TD align=center>%.6f</TD></TR>',[k,color,FopFact.EigVal[k],FRefEigens^[k]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

{ TOpPrmFactPA }

function TOpPrmFactPA.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFactParallelAnalysis.CreateFromOpPrm(self);
end;

function TOpPrmFactPA.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Option</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>Replications</TD><TD align=right>%d</TD></TR>',[NbReplication]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>Confidence coef.</TD><TD align=right>%.2g</TD></TR>',[Percentile]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>Random Seed</TD><TD align=right>%d</TD></TR>',[SeedRnd]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmFactPA.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 NbReplication:= prmINI.ReadInteger(prmSection,'replication',NbReplication);
 Percentile:= prmINI.ReadFloat(prmSection,'percentile',Percentile);
 SeedRnd:= prmINI.ReadInteger(prmSection,'seed_rnd',SeedRnd);
end;

procedure TOpPrmFactPA.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(NbReplication,sizeof(NbReplication));
 prmStream.ReadBuffer(Percentile,sizeof(Percentile));
 prmStream.ReadBuffer(SeedRnd,sizeof(SeedRnd));
end;

procedure TOpPrmFactPA.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'replication',NbReplication);
 prmINI.WriteFloat(prmSection,'percentile',Percentile);
 prmINI.WriteInteger(prmSection,'seed_rnd',SeedRnd);
end;

procedure TOpPrmFactPA.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(NbReplication,sizeof(NbReplication));
 prmStream.WriteBuffer(Percentile,sizeof(Percentile));
 prmStream.WriteBuffer(SeedRnd,sizeof(SeedRnd));
end;

procedure TOpPrmFactPA.SetDefaultParameters;
begin
 NbReplication:= 200;//nombre de rplication, df. 200
 Percentile:= 0.95;//pourcentage pour le quantile, df. 0.95
 SeedRnd:= 1;//valeur de dpart pour le gnrateur de nombres alatoires
end;

function TOpFactPA.getNewForm: TForm;
begin
 result:= TFrmOpViewFactParallelAnalysis.CreateFromOperator(self);
end;

function TOpFactPA.getObservedEigen(i: integer): double;
begin
 result:= FopFact.EigVal[i];
end;

function TOpFactPA.getRefEigen(i: integer): double;
begin
 result:= FRefEigens^[i];
end;

initialization
 CLASSES.RegisterClass(TMLGenCompFactPA);
end.
