(***********************************************************************************)
(* UCompFCBinarizeDiscreteToContinuous.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(***********************************************************************************)

{
@abstract(Codage disjonctif complet)
@author(Ricco)
@created(25/04/2004)
}
unit UCompFCBinarizeDiscreteToContinuous;

interface

USES
        Forms, Classes, IniFiles,
        UCompDefinition,
        UCompFCDefinition,
        UOperatorDefinition,
        UCalcStatDes, UDatasetDefinition, UDatasetImplementation;

TYPE
        {gnrateur de composant}
        TGenFCBinarizationContinuous = class(TMLGenComp)
                                       protected
                                       procedure   GenCompInitializations(); override;
                                       public
                                       function    GetClassMLComponent: TClassMLComponent; override;
                                       end;

        {composant}
        TMLCompFCBinContinuous  = class(TMLCompFC)
                                  protected
                                  function    getClassOperator: TClassOperator; override;
                                  function    GetLogResultDescription(): string; override;
                                  end;

        {oprateur}
        TOpFCBinContinuous = class(TOperatorFC)
                             protected
                             function  CheckAttributes(): boolean; override;
                             function  getClassParameter: TClassOperatorParameter; override;
                             function    CoreExecute(): boolean; override;
                             public
                             function    getHTMLResultsSummary(): string; override;
                             end;

        {paramtrage}
        TOpPrmFCBinContinuous = class(TOperatorPrmFC)
                                private
                                //new -- 11/08/2006 -- option description complte ou pas (avec les K modalits ou pas) : false par dfaut (K-1)
                                FFullDesc: boolean;
                                protected
                                function    CreateDlgParameters(): TForm; override;
                                procedure   SetDefaultParameters(); override;
                                public
                                function    getHTMLParameters(): string; override;
                                procedure   LoadFromStream(prmStream: TStream); override;
                                procedure   SaveToStream(prmStream: TStream); override;
                                procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
                                procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
                                property    FullDescription: boolean read FFullDesc write FFullDesc;
                                end;

implementation

USES
        Sysutils, UConstConfiguration, UDlgOpPrmFCBinarization;

{ TGenFCBinarizationContinuous }

procedure TGenFCBinarizationContinuous.GenCompInitializations;
begin
 FMLComp:= mlcFeatureConstruction;
end;

function TGenFCBinarizationContinuous.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFCBinContinuous;
end;

{ TMLCompFCBinContinuous }

function TMLCompFCBinContinuous.getClassOperator: TClassOperator;
begin
 result:= TOpFCBinContinuous;
end;

function TMLCompFCBinContinuous.GetLogResultDescription: string;
begin
 result:= format('%d continuous attributes generated',[LocalDataset.Count]);
end;

{ TOpFCBinContinuous }

function TOpFCBinContinuous.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asInput].Count>0) and (self.WorkData.LstAtts[asInput].isAllCategory(caDiscrete));
end;

function TOpFCBinContinuous.CoreExecute: boolean;
var j,i,k,v,nbValues: integer;
    attSource: TAttDiscrete;
    attDest: array of TAttContinue;
begin
 result:= TRUE;
 TRY
  //vider la liste courante
  GenAtts.Clear();
  //pour chaque attribut  coder
  for j:= 0 to pred(self.WorkData.LstAtts[asInput].Count) do
   begin
    attSource:= self.WorkData.LstAtts[asInput].Attribute[j] as TAttDiscrete;
    //nombre de valeurs
    nbValues:= attSource.nbValues;
    //si description complte, les "K" modalits gnrent une variable, sinon on s'en tient aux "K-1"
    if (self.PrmOp as TOpPrmFCBinContinuous).FullDescription then inc(nbValues);
    //crer les variables 0/1, sauf la dernire
    SetLength(attDest,nbValues);
    for k:= 1 to pred(nbValues) do
     begin
      attDest[k]:= TAttContinue.Create(attSource.Name+'_'+attSource.LstValues.getDescription(k)+'_'+IntToStr((self.MLOwner as TMLCompFCBinContinuous).Number),attSource.Size);
      GenAtts.Add(attDest[k]);
     end;
    //coder
    for i:= 1 to attSource.Size do
     begin
      //tous  zro
      for k:= 1 to pred(nbValues) do
       attDest[k].cValue[i]:= 0.0;
      //sauf celui qui correspond
      v:= attSource.dValue[i];
      if (v<nbValues)
       then attDest[v].cValue[i]:= 1.0;
     end;
   end;
  //librer tout de suite -- mme s'il y a un compteur de rfrences (semble-t-il ?)
  SetLength(attDest,0);
 EXCEPT
 result:= FALSE;
 END;
end;

function TOpFCBinContinuous.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFCBinContinuous;
end;

function TOpFCBinContinuous.getHTMLResultsSummary: string;
var s,tmp: string;
    j,k,i: integer;
    decalage: integer;
begin
 //prendre en compte l'option
 if (prmOp as TOpPrmFCBinContinuous).FullDescription
  then decalage:= 1
  else decalage:= 2;
 //dcrire les rsultats
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan="2">Attribute binarization</TH></TR>';
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Source att</TH><TH>New attributes</TH></TR>';
 k:= 0;
 for j:= 0 to pred(self.WorkData.LstAtts[asInput].Count) do
  begin
   //lister
   tmp:= '(';
   for i:= k to k+self.WorkData.LstAtts[asInput].Attribute[j].nbValues-decalage do
    tmp:= tmp+self.GenAtts.Attribute[i].Name+',';
   tmp[length(tmp)]:= ')';
   //dcaler
   inc(k,self.WorkData.LstAtts[asInput].Attribute[j].nbValues-(decalage-1));
   //afficher
   s:= s+ HTML_TABLE_COLOR_DATA_GRAY
       + format('<TD>%s</TD>',[self.WorkData.LstAtts[asInput].Attribute[j].Name])
       + format('<TD>%s</TD>',[tmp])
       + '</TR>';
  end;
 s:= s+'</table>';
 result:= s;
end;

{ TOpPrmFCBinContinuous }

function TOpPrmFCBinContinuous.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFCBinarization.CreateFromOpPrm(self);
end;

function TOpPrmFCBinContinuous.getHTMLParameters: string;
begin
 if FFullDesc
  then result:= 'Used values : K (all)'
  else result:= 'Used values : K-1 (ignore last value)';
end;

procedure TOpPrmFCBinContinuous.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FFullDesc:= prmINI.ReadBool(prmSection,'use_K_values',FFullDesc);
end;

procedure TOpPrmFCBinContinuous.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FFullDesc,sizeof(FFullDesc));
end;

procedure TOpPrmFCBinContinuous.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteBool(prmSection,'use_K_values',FFullDesc);
end;

procedure TOpPrmFCBinContinuous.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FFullDesc,sizeof(FFullDesc));
end;

procedure TOpPrmFCBinContinuous.SetDefaultParameters;
begin
 FFullDesc:= false;
end;

initialization
 RegisterClass(TGenFCBinarizationContinuous);
end.
