(***********************************************************************)
(* UCompFCRandomProjection.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(***********************************************************************)

{
@abstract(Random Projection)
@author(Ricco)
@created(26/11/2005)

Composant de base de construction de variables. Il sert avant tout de modle pour
les autres composants qui pourront tre ajouts par la suite.

Pour l'instant, il prend au hasard "p" fois un couple de variables INPUT, en ralise le produit,
le rsultat devient une nouvelle variable propose en sortie du composant.

Le dtail du calcul est le suivant :
(1) pour deux variables (a) et (b)
(2) la moyenne est calcule >> sur les individus slectionnes << --> on obtient avg(a) et avg(b)
(3) la tierce variable est alors [(a)-avg(a)] x [(b)-avg(b)], ceci sur toutes les observations !

Ce composant fait suite  une discussion avec R.Priam. Il est appel  voluer, soit par modification,
soit par hritage, etc.

}
unit UCompFCRandomProjection;

interface

USES
    //units std delphi
    Classes, IniFiles, Forms, Sysutils,
    //gestion des composants
    UCompDefinition, UOperatorDefinition,
    UCompManageDataset;


TYPE
    //********************************************************************************
    //**** Gnrateur de composant, il permet d'insrer un composant dans le diagramme
    //********************************************************************************
    TMLGenCompFCRndProj = class(TMLGenComp)
    protected
    //spcifier dans quel onglet de la palette doit apparatre l'icne --  surcharger ventuellement dans les hritiers
    procedure   genCompInitializations(); override;
    public
    //spcifier quel est la classe du composant  gnrer sur le diagramme --  surcharger obligatoirement chez les hritiers
    function    getClassMLComponent: TClassMLComponent; override;
    end;

    //**********************************************************
    //**** class du composant graphique insr dans le diagramme
    //**********************************************************
    TMLCompFCRndProj = class(TMLCompLocalData)
    protected
    //spcifier l'oprateur de calcul --  surcharger obligatoirement
    function    getClassOperator: TClassOperator; override;
    end;

    //**************************************************************************************
    //** classe oprateur de calcul : prend en charge le calcul et l'affichage des rsultats
    //**************************************************************************************
    TOpFCRndProj = class(TOpLocalData)
    private
    //tableau des moyennes
    FAvg: array of double;
    protected
    //spcifier la classe de paramtrage -- surcharge obligatoire
    function    getClassParameter: TClassOperatorParameter; override;
    //contrle de cohrence des attributs avant les calculs --  surcharger selon les mthodes
    function    checkAttributes(): boolean; override;
    //lancer effectivement les calculs -- surcharge obligatoire
    function    coreExecute(): boolean; override;
    public
    //libration de la mmoire alloue -- surcharge selon les mthodes, appel de la mthode anctre avec "inherited" obligatoire
    destructor  destroy(); override;
    //affichage des rsultats au format HTML -- surcharge obligatoire
    function    getHTMLResultsSummary(): string; override;
    end;

    //***********************************************************************
    //** classe paramtrage de l'oprateur de calcul : gre la bote de paramtrage et l'E/S des paramtres
    //** si la mthode de requiert pas de paramtres, il suffit de crer une classe vide hritire de "TOperatorParameter"
    //***********************************************************************
    TOpPrmFCRndProj = class(TOperatorParameter)
    protected
    //cration de la bote de dialogue de paramtrage -- surcharge obligatoire (doit renvoyer "nil" si pas de bote  gnrer)
    function    CreateDlgParameters(): TForm; override;
    //spcifier les paramtres par dfaut de la mthode -- surcharge obligatoire
    procedure   SetDefaultParameters(); override;
    public
    //nombre de variables  gnrer -- cette proprit est le paramtre de la mthode
    NbNewVar: integer;
    //affichage HTML des paramtres -- surcharge obligatoire
    function    getHTMLParameters(): string; override;
    //sauvegarde des paramtres dans diffrents formats -- flux de donnes et fichiers INI -- surcharge obligatoire
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    //chargement des paramtres dans diffrents formats -- flux de donnes et fichiers INI -- surcharge obligatoire -- doit tre "mirroir" des save
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    end;

implementation

uses UDatasetDefinition, UDatasetImplementation, UCalcRndGenerator;

{ TMLGenCompFCRndProj }

procedure TMLGenCompFCRndProj.genCompInitializations;
begin
 //le composant est plac dans la palette "Feature Construction"
 FMLComp:= mlcFeatureConstruction;
end;

function TMLGenCompFCRndProj.getClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFCRndProj;
end;

{ TMLCompFCRndProj }

function TMLCompFCRndProj.getClassOperator: TClassOperator;
begin
 result:= TOpFCRndProj;
end;

{ TOpFCRndProj }

function TOpFCRndProj.checkAttributes: boolean;
begin
 //dans ce cas, on vrifie si toutes les variables INPUT sont toutes continues
 //voir dfinition dans les units >> UDatasetDefinition.pas et UDatasetImplementation.pas <<
 result:= self.WorkData.LstAtts[asInput].isAllCategory(caContinue);
end;

function TOpFCRndProj.coreExecute: boolean;
var example,j: integer;
    p: integer;
    attA, attB, newAtt: TAttribute;
    kA, kB: integer;
    rnd: TRndGenerator;
    value: TTypeContinue;
begin
 //****************************************************
 //calculer les moyennes sur les individus slectionns
 //****************************************************
 SetLength(FAvg,self.workdata.LstAtts[asInput].Count);//dimensionner le tableau
 for j:= 0 to self.workdata.LstAtts[asInput].Count-1 do FAvg[j]:= 0.0;//rz du tableau
 for j:= 0 to self.workdata.LstAtts[asInput].Count-1 do //pour chaque variable
  begin
   //somme des valeurs, sur les individus slectionns, attention au d-rfrencement avec la proprit ".Number"
   for example:= 1 to self.WorkData.Examples.Size do
    FAvg[j]:= FAvg[j] + self.WorkData.LstAtts[asInput].Attribute[j].cValue[self.WorkData.Examples.Number[example]];
   FAvg[j]:= FAvg[j]/(1.0 * self.WorkData.Examples.Size);//moyenne
  end;
 //vider la liste de variables gnres
 self.GenAtts.Clear();
 //****************************************************
 //*** gnration de nouvelles variables **************
 //****************************************************
 //initialiser le gnrateur de nombre alatoire, le dpart dpend du numro du composant...
 rnd:= TRndGenerator.Create(seedUser,DEFAULT_SEED_VALUE_1,(MLOwner as TMLComponent).Number);
 //nombre de variables  gnrer, rcupr  partir de l'objet paramtre, ventuellement modifiable par l'utilisateur
 p:= (self.PrmOp as TOpPrmFCRndProj).NbNewVar;
 for j:= 1 to p do
  begin
   //slectionner deux variables au hasard, on peut ventuellement conserver la liste pour l'affichage des rsultats...
   kA:= rnd.IRanMarRange(self.workdata.LstAtts[asInput].Count); attA:= self.WorkData.LstAtts[asInput].Attribute[kA];
   kB:= rnd.IRanMarRange(self.workdata.LstAtts[asInput].Count); attB:= self.WorkData.LstAtts[asInput].Attribute[kB];
   //gnrer la nouvelle variable, attention  la rgle de nommage qui lui permet d'avoir un nom unique
   newAtt:= TAttContinue.Create(format('RP_%dx%d_%d_%d',[kA,kB,j,(MLOwner as TMLComponent).Number]),self.WorkData.LstAtts[asAll].Size);
   //calcul des valeurs de la nouvelle variable --> !!! sur toutes les observations cette fois-ci !!!
   for example:= 1 to self.WorkData.LstAtts[asAll].Size do
    begin
     value:= (attA.cValue[example] - FAvg[kA]) * (attB.cValue[example] - FAvg[kB]);
     newAtt.cValue[example]:= value;
    end;
   //ajouter la variable dans le pool de sortie
   self.GenAtts.Add(newAtt);
  end;
 rnd.Free();//librer le gnrateur
 //tout est ok !??!
 result:= true;
end;

destructor TOpFCRndProj.destroy;
begin
 //librer le tableau des moyennes s'il est allou
 if (FAvg <> nil) then Finalize(FAvg);
 //ne pas oublier...
 inherited;
end;

function TOpFCRndProj.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFCRndProj;
end;

function TOpFCRndProj.getHTMLResultsSummary: string;
var s: string;
    j: integer;
begin
 s:= format('<b>%d</b> new computed attributes<br><br><ul>',[self.GenAtts.Count]);
 for j:= 0 to self.GenAtts.Count-1 do
  s:= s + format('<li> %s',[self.GenAtts.Attribute[j].Name]);
 s:= s + '</ul>';
 //
 result:= s;
end;

{ TOpPrmFCRndProj }

function TOpPrmFCRndProj.CreateDlgParameters: TForm;
begin
 //pas de fiche de paramtrage dispo pour l'instant
 result:= nil;
end;

function TOpPrmFCRndProj.getHTMLParameters: string;
begin
 result:= format('# generated attributes : %d',[NbNewVar]); 
end;

procedure TOpPrmFCRndProj.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 NbNewVar:= prmINI.ReadInteger(prmSection,'nb_new_var',NbNewVar);
end;

procedure TOpPrmFCRndProj.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(NbNewVar,sizeof(NBNewVar));
end;

procedure TOpPrmFCRndProj.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_new_var',NbNewVar);
end;

procedure TOpPrmFCRndProj.SaveToStream(prmStream: TStream);
begin
 //sauver dans le flux le nombre de variables  gnrer
 prmStream.WriteBuffer(NbNewVar,sizeof(NbNewVar));
end;

procedure TOpPrmFCRndProj.SetDefaultParameters;
begin
 //par dfaut, on gnre 2 variables
 NbNewVar:= 2;
end;

initialization
 //enregistrement de la classe pour qu'elle soit reconnue dans
 //le fichier de confuguration XML
 //!\ --> le fichier XML doit tre renseign correctement pour que le dispositif fonctionne
 RegisterClass(TMLGenCompFCRndProj);
end.
