(***************************************************************)
(* UCompFSStepDisc.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(***************************************************************)

{

@abstract(Slectionner les input avec un test conditionnel avec le LAMBDA de WILKS)
@author(Ricco)
@created(23/01/2007)


C'est une variante STEPDISC. Le rsultat est le mme que l'exemple dcrit dans le livre de Tenehaus (pp. 256  265).
Mais les calculs sont tout  fait diffrents, on regarde le Lambda partiel conscutif  l'adjonction (forward) ou le
retrait (backward) d'une variable dans le modle.

Les calculs s'arrtent lorsqu'il n'est plus possible d'ajouter une variable significative (foward) ou de
supprimer une variable non-significative (backward). Le paramtrag repose sur la p-value seuil du lambda partiel de WILKS (
test de Fisher).

Attention, dans le processus, tout ajout (retrait) de variable n'est plus remis en cause.

Paramtrage :
-------------
(0) Type de Rgle d'arrt
0.a -> p-value
0.b -> valeur seuil
0.c -> nombre de variables  retenir

(1) forward ou backward

(3) paramtre d'affichage
1 -> le meilleur rsultat
2 -> les deux meilleurs rsultats
etc.

}
unit UCompFSStepDisc;

interface

USES
    Forms, Classes, IniFiles,
    UCompDefinition,
    UCompFSDefinition,
    UCompFSInputSelection,
    UOperatorDefinition,
    UCalcStatDesConditionnalDesc,
    UCalcStatDesManova,
    UDatasetDefinition,
    Matrices,
    UCalcFSStepDisc;

TYPE
    {gnrateur}
    TMLGenFSStepDisc = class(TMLGenFS)
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    {composant}
    TMLCompFSStepDisc = class(TMLCompFSInputSelection)
    protected
    function    getClassOperator: TClassOperator; override;
    end;

    {oprateur}
    TOpFSStepDisc = class(TOpFSInputSelSpvInputContinuous)
    private
    //objet MANOVA
    FManova: TCalcSDCondManova;
    //matrice T et W -- attention, ils ne sont pas owner du tout... donc  ne pas dtruire
    FMatT, FMatW: PMatrix;
    //target
    FTarget: TAttribute;
    //inputs
    FInputs: TLstAttributes;
    //les solutions
    FSoluce: TGlobalResultat;
    //les candidats
    FCandidats: TLstCandidat;
    //dtecter la solution optimale courante - tester si la rgle d'arrt est active - assurer le passage de IN vers OUT
    function    isStoppingRuleActivatedForward(input,output: TLstCandidat; level: TLevelResultat): boolean;
    //dtecter si rgle d'arrt dclenche en backward -- mj. de la input list
    function    isStoppingRuleActivatedBackward(input: TLstCandidat; level : TLevelResultat): boolean;
    //prparer les objects globaux
    procedure   prepareObjects();
    //lister les variables slectionnes
    function    getHTMLSelectedAtts(): string;
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    //recherche forward
    procedure   forwardSolution();
    //recherche backward
    procedure   backwardSolution();
    function    getNewForm(): TForm; override;    
    public
    constructor Create(AOwner: TObject); override;
    destructor  Destroy(); override;
    function    getHTMLResultsSummary(): string; override;
    function    CoreExecute(): boolean; override;
    property    Soluce: TGlobalResultat read FSoluce;
    end;

    {paramtre d'oprateur}
    TOpPrmFSStepDisc = class(TOpPrmFSInputSelection)
    private
    protected
    procedure   SetDefaultParameters(); override;
    function    CreateDlgParameters(): TForm; override;
    public
    //critre de rgle d'arrt -- 0 : F-Value ; 1 :  Sig Level ; 2 : nombre de variables -- df. 1
    Criteria : integer;
    //Seuil F pour la comparaison -- df. 3.84
    FValue : double;
    //Seuil proba pour la comparaison -- df. 0.05
    SigValue : double;
    //Nombre de variables acceptes -- def. 5
    NbAccepted : integer;
    //sens de la recherche -- true -> forward, false -> backward
    ForwardSearch: boolean;
    //nombre d'lments  afficher dans la rapport -- 0 : tous, 1, 2, etc. le nombre demand -- df. 5
    NbColSoluce: integer;
    //lister les attributs slectionns dans le rapport ? -- df. true
    IsShowSelected: boolean;
    //montrer le dtail du processus ? -- df. true
    IsShowDetailedResults: boolean;
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    end;

implementation

uses
    Sysutils,
    UDatasetImplementation, FMath, ULogFile, UConstConfiguration,
    UDlgOpPrmFSStepdisc, UFrmOpViewFSStepdisc;

{ TMLGenFSStepDisc }

function TMLGenFSStepDisc.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFSStepDisc; 
end;

{ TMLCompFSStepDisc }

function TMLCompFSStepDisc.getClassOperator: TClassOperator;
begin
 result:= TOpFSStepDisc;
end;

{ TOpFSStepDisc }

procedure TOpFSStepDisc.backwardSolution;
var T,W: PMatrix;
    inputList, outputList : TLstCandidat;
    cur: TCandidat;
    jCur: integer;
    i,j: integer;
    ic, jc: TCandidat;
    lastSize: integer;//dernire taille de matrice
    prevLambda, num, denom: double;
    levelRes: TLevelResultat;//rsultat pour un niveau
    res: TResultat;//un rsultat
    K: double;//nombre de classes
    N: double;//effectifs
begin
 K:= 1.0*self.FTarget.nbValues;
 N:= 1.0*self.WorkData.Examples.Size;
 self.prepareObjects();//prparer les objets globaux de calcul
 inputList:= TLstCandidat.create(false); inputList.assign(FCandidats); //liste locale des candidats
 outputList:= TLstCandidat.create(false);//liste vide des slectionns
 lastSize:= FInputs.Count;
 prevLambda:= FManova.Wilks;//WILKS de toutes les variables prsentes
 REPEAT
  //passer au niveau suivant -- une variable en moins
  dec(lastSize);
  //rsultat pour le niveau de recherche
  levelRes:= TLevelResultat.create();
  levelRes.Df1:= TRUNC(K-1);
  levelRes.Df2:= TRUNC(N-K-lastSize);//(la solution prcdente tait J+1, l on teste J)
  //s'il n'y a plus de variables...
  if (lastSize = 0)
   then
    begin
     //que se passe-t-il si on enlve cette dernire variable
     cur:= inputList.getCandidat(0);
     res:= TResultat.Create(cur);    
     res.Lambda:= 1.0;
     res.F:= (N-K-lastSize)/(K-1.0)*(prevLambda/res.Lambda-1.0);
     res.PValue:= PSnedecor(levelRes.Df1,levelRes.Df2,res.F);
     //l'ajouter au niveau
     levelRes.addresult(res);
    end
   else
    begin
     //prparer les matrices
     dimMatrix(T,lastSize,lastSize);
     dimMatrix(W,lastSize,lastSize);
     //pour chaque candidat  la sortie
     for jCur:= 0 to pred(inputList.getCount()) do
      begin
       cur:= inputList.getCandidat(jCur);
       //former la configuration  tester
       outputList.clear();
       for j:= 0 to pred(inputList.getCount()) do
        if (j <> jCur) then outputList.add(inputList.getCandidat(j));
       //remplir la matrice associe
       for i:= 0 to pred(outputList.getCount()) do
        begin
         ic:= outputList.getCandidat(i);
         for j:= 0 to pred(outputList.getCount()) do
          begin
           jc:= outputList.getCandidat(j);
           T^[i+1]^[j+1]:= FManova.matT^[ic.Numero]^[jc.Numero];
           W^[i+1]^[j+1]:= FManova.MatWBrut^[ic.Numero]^[jc.Numero];
          end;
        end;
       //instancier le rsultat
       res:= TResultat.Create(cur);
       //calculer le lambda de wilks
       num:= Det(W,1,lastSize);
       denom:= Det(T,1,lastSize);
       if (denom > 0.0)
        then res.Lambda:= num / denom
        else res.Lambda:= 1.0;
       //test F
       res.F:= (N-K-lastSize)/(K-1.0)*(res.Lambda/prevLambda-1.0);
       res.PValue:= PSnedecor(levelRes.Df1,levelRes.Df2,res.F);
       //ajouter la solution
       levelRes.addresult(res);
      end;
     //librer les matrices
     delMatrix(T,lastSize,lastSize);
     delMatrix(W,lastSize,lastSize);
    end;
  //trier les rsulats et dtecter si une solution existe
  levelRes.sortAscendant();//pour backawrd
  //l'ajouter dans les solutions
  FSoluce.addLevel(levelRes);
  //mettre  jour le lambda quoiqu'il en soit
  if (levelRes.getCountResults() > 0)
   then prevLambda:= levelRes.getResult(0).Lambda;
 //l'ordre du test est important ici, car isStopping... modifie le contenu de inputList
 UNTIL (self.isStoppingRuleActivatedBackward(inputList,levelRes) = TRUE) or (inputList.getCount() = 0);
 //**************************************************************
 //rajouter maintenant dans la sortie les variables slectionnes --> ceux qui sont rests dans la inputList sont les slectionns
 for j:= 0 to pred(inputList.getCount()) do
  self.OutputData.LstAtts[asInput].Add(inputList.getCandidat(j).Variable);
 //***************************************************************
 //librer
 inputList.Free();
 outputList.Free();
end;

function TOpFSStepDisc.CoreExecute: boolean;
var v : double;
    i,j: integer;
begin
 result:= true;
 try
 //target and inputs
 FTarget:= self.WorkData.LstAtts[asTarget].Attribute[0];
 FInputs:= self.WorkData.LstAtts[asInput];
 //MANOVA
 if assigned(FManova) then FManova.Free();
 FManova:= TCalcSDCondManova.create(FTarget,FInputs,self.WorkData.Examples);
 //rcuprer la matrice totale en pointant directement sur celui de la Manova
 FMatT:= FManova.MatT;
 //rcuprer la mvcv-intra brute en pointant sur celui de la manova et en la pondrant -- donc celui de la Manova n'est plus "brut" !!!
 FMatW:= FManova.MatWBrut;
 //pondrer pour avoir un estimateur sans biais
 v:= 1.0 / (1.0 * self.WorkData.Examples.Size);
 for i:= 1 to FInputs.Count do
  for j:= 1 to Finputs.Count do
   FMatW^[i]^[j]:= v * FMatW^[i]^[j];
 //*************************************************
 //calculer maintenant les diffrentes solutions....
 //*************************************************
 case (self.PrmOp as TOpPrmFSStepDisc).ForwardSearch of
  true : forwardSolution();
  else backwardSolution();
 end;
 except
 result:= false;
 end;
end;

constructor TOpFSStepDisc.Create(AOwner: TObject);
begin
  inherited;

end;

destructor TOpFSStepDisc.Destroy;
begin
 if assigned(FSoluce) then FSoluce.Free();
 if assigned(FCandidats) then FCandidats.Free();//aprs SOLUCE !!! car c'est le proprio des objets candidats !!!
 if assigned(FManova) then FManova.Free();
 inherited;
end;

procedure TOpFSStepDisc.forwardSolution;
var T,W: PMatrix;
    inputList, outputList : TLstCandidat;
    cur: TCandidat;
    jCur: integer;
    i,j: integer;
    ic, jc: TCandidat;
    lastSize: integer;//dernire taille de matrice
    prevLambda, num, denom: double;
    levelRes: TLevelResultat;//rsultat pour un niveau
    res: TResultat;//un rsultat
    K: double;//nombre de classes
    N: double;//effectifs
begin
 K:= 1.0*self.FTarget.nbValues;
 N:= 1.0*self.WorkData.Examples.Size;
 self.prepareObjects();//prparer les objets globaux de calcul
 inputList:= TLstCandidat.create(false); inputList.assign(FCandidats); //liste locale des candidats
 outputList:= TLstCandidat.create(false);//liste vide des slectionns
 lastSize:= 0;
 prevLambda:= 1.0;//pas de discrimination puisque pas de variables...
 REPEAT
  //passer au niveau suivant
  inc(lastSize);
  //rsultat pour le niveau de recherche
  levelRes:= TLevelResultat.create();
  levelRes.Df1:= TRUNC(K-1);
  levelRes.Df2:= TRUNC(N-K-(lastSize-1));//cf. formule lastSize-1 = J --> nombre de variables dj dans le modle
  //allons au plus simple -- ici il y a clairement beaucoup de temps  gagner pour les calculs
  dimMatrix(T,lastSize,lastSize);
  dimMatrix(W,lastSize,lastSize);
  //former les matrices de calcul du Lambda pour chaque candidat
  for jCur:= 0 to pred(inputList.getCount()) do
   begin
    cur:= inputList.getCandidat(jCur);
    //**************************************
    //* traitement d'une candidature *******
    //**************************************
    //recopier dj ceux qui sont dans output
    for i:= 0 to pred(outputList.getCount()) do
     begin
      ic:= outputList.getCandidat(i);
      for j:= 0 to pred(outputList.getCount()) do
       begin
        jc:= outputList.getCandidat(j);
        T^[i+1]^[j+1]:= FManova.matT^[ic.Numero]^[jc.Numero];
        W^[i+1]^[j+1]:= FManova.MatWBrut^[ic.Numero]^[jc.Numero];
       end;
     end;
    //recopier la dernire ligne et colonne
    for i:= 0 to pred(outputList.getCount()) do
     begin
      ic:= outputList.getCandidat(i);
      T^[lastSize]^[i+1]:= FManova.matT^[ic.Numero]^[cur.Numero];
      T^[i+1]^[lastSize]:= FManova.matT^[cur.Numero]^[ic.Numero];
      W^[lastSize]^[i+1]:= FManova.MatWBrut^[ic.Numero]^[cur.Numero];
      W^[i+1]^[lastSize]:= FManova.MatWBrut^[cur.Numero]^[ic.Numero];
     end;
    //and la dernire case
    T^[lastSize]^[lastSize]:= FManova.matT^[cur.Numero]^[cur.Numero];
    W^[lastSize]^[lastSize]:= FManova.MatWBrut^[cur.Numero]^[cur.Numero];
    //instancier le rsultat
    res:= TResultat.Create(cur);
    //calculer le lambda de wilks
    num:= Det(W,1,lastSize);
    denom:= Det(T,1,lastSize);
    if (denom > 0.0)
     then res.Lambda:= num / denom
     else res.Lambda:= 1.0;
    //test F
    //new -- 02/02/2007 -- attention  la discrimination parfaite
    if (res.Lambda > 0.0)
     then res.F:= (N-K-(-1.0+lastSize))/(K-1.0)*(prevLambda/res.Lambda-1.0)
     else res.F:= 1e308; 
    res.PValue:= PSnedecor(levelRes.Df1,levelRes.Df2,res.F);
    //l'ajouter
    levelRes.addresult(res);
   end;
  //fin des calculs
  delMatrix(T,lastSize,lastSize);
  delMatrix(W,lastSize,lastSize);
  //trier les rsulats et dtecter si une solution existe
  levelRes.sortDescendant();//pour forward
  //l'ajouter dans les solutions
  FSoluce.addLevel(levelRes);
  //mettre  jour le lambda quoiqu'il en soit
  if (levelRes.getCountResults() > 0)
   then prevLambda:= levelRes.getResult(0).Lambda;
 //l'ordre du test est important ici, car isStopping... modifie le contenu de inputList
 UNTIL (self.isStoppingRuleActivatedForward(inputList,outputList, levelRes) = true) OR (inputList.getCount() = 0);
 //librer
 inputList.Free();
 outputList.Free();
 //**************************************************************
 //rajouter maintenant dans la sortie les variables slectionnes
 for i:= 0 to pred(FSoluce.getCountLevel()) do
  begin
   levelRes:= FSoluce.getLevel(i);
   if (levelRes.Best <> nil)
    then self.OutputData.LstAtts[asInput].Add(levelRes.Best.Candidat.Variable);
  end;
 //***************************************************************
end;

function TOpFSStepDisc.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFSStepDisc;
end;

function TOpFSStepDisc.getHTMLResultsSummary: string;
var s: string;
    prm: TOpPrmFSStepDisc;
begin
 prm:= Self.PrmOp as TOpPrmFSStepDisc;
 //
 s:= '<H3>Selection results</H3>';
 s:= s + Format('<b>[%d]</b> selected attributes on <b>[%d]</b>',[self.OutputData.LstAtts[asInput].Count,FInputs.Count]);
 //lister les attributs slectionns ?
 if (prm.IsShowSelected = true)
  then s:= s + self.getHTMLSelectedAtts();
 if (prm.IsShowDetailedResults = true)
  then
   begin
    s:= s + '<H3>Detailed results</H3>';
    if (prm.NbColSoluce < 1)
     then s:= s + FSoluce.getHTML(FInputs.Count)
     else s:= s + FSoluce.getHTML(prm.NbColSoluce);
   end;
 //
 result:= s;
end;

function TOpFSStepDisc.getHTMLSelectedAtts: string;
var s: string;
    j: integer;
begin
 s:= '<H3>Selected attributes'' subset</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>N</TH><TH>Selected atts</TH></TR>';
 for j:= 0 to pred(self.OutputData.LstAtts[asInput].Count) do
  s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>%d</TD><TD>%s</TD></TR>',[succ(j),self.OutputData.LstAtts[asInput].Attribute[j].Name]);
 s:= s + '</table>';
 result:= s;
end;

function TOpFSStepDisc.getNewForm: TForm;
begin
 result:= TFrmOpViewFSStepdisc.CreateFromOperator(self);
end;

function TOpFSStepDisc.isStoppingRuleActivatedBackward(input: TLstCandidat;
  level: TLevelResultat): boolean;
var best: TResultat;
    prm: TOpPrmFSStepDisc;
begin
 //paramtrage de la mthode
 prm:= self.prmOp as TOpPrmFSStepDisc;
 //le meilleur rsultat sur le niveau
 best:= level.getResult(0);
 //selon le paramtrage activ
 result:= true;
 case prm.Criteria of
  0 : //seuil F
    begin
     if (best.F < prm.FValue)
      then
       begin
        level.Best:= best;
        input.remove(input.indexOf(best.Candidat));
        result:= false;
       end;
    end;
  1 : //p-value
    begin
     if (best.PValue > prm.SigValue)
      then
       begin
        level.Best:= best;
        input.remove(input.indexOf(best.Candidat));
        result:= false;
       end;
    end;
  else
    //nombre
    begin
     if (input.getCount() > prm.NbAccepted)
      then
       begin
        level.Best:= best;
        input.remove(input.indexOf(best.Candidat));
       end;
      //on s'arrte d'enlever si on a atteint le nombre voulu
      result:= (input.getCount() <= prm.NbAccepted);
    end;
 end;
end;

function TOpFSStepDisc.isStoppingRuleActivatedForward(input,output: TLstCandidat; level: TLevelResultat): boolean;
var best: TResultat;
    prm: TOpPrmFSStepDisc;
    ok: boolean;

    procedure exchangeInOut();
    var index: integer;
    begin
     index:= input.indexOf(best.Candidat);
     if (index >= 0)
      then
       begin
        input.remove(index);
        output.add(best.Candidat);
       end
      //catastrophe...
      else TraceLog.WriteToLogFile('[STEPDISC] !!! ERROR !!! critique -- candidat non trouved durant echange');
    end;

begin
 //paramtrage de la mthode
 prm:= self.prmOp as TOpPrmFSStepDisc;
 //le meilleur rsultat sur le niveau
 best:= level.getResult(0);
 //si la rgle d'arrt est le nombre d'lments
 case prm.Criteria of
 0 : //seuil F
   begin
    ok:= not(best.F > prm.FValue);
    if not(ok)
     then
      begin
       level.Best:= best;
       exchangeInOut();
      end;
    result:= ok;
   end;
 1 : //p-value
   begin
    ok:= not(best.PValue < prm.SigValue);
    if not(ok)
     then
      begin
       level.Best:= best;
       exchangeInOut();
      end;
    result:= ok;
   end;
 else
   //on se base sur le nombre de candidats  accepter
   begin
    //valider automatiquement
    level.Best:= best;
    //passer l'input dans output
    exchangeInOut();
    //et tester si on doit s'arrter
    result:= (output.getCount() >= prm.NbAccepted);
   end;
 end;
end;

procedure TOpFSStepDisc.prepareObjects;
var j: integer;
begin
 //solutions
 if assigned(FSoluce) then FSoluce.Free();
 FSoluce:= TGlobalResultat.create();
 //les candidats -- trait aprs soluce...
 if assigned(FCandidats) then FCandidats.Free();
 FCandidats:= TLstCandidat.create(true);
 for j:= 0 to pred(FInputs.Count) do
  FCandidats.add(TCandidat.create(succ(j),FInputs.Attribute[j]));
end;

{ TOpPrmFSStepDisc }

function TOpPrmFSStepDisc.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFSStepdisc.CreateFromOpPrm(self);
end;

function TOpPrmFSStepDisc.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH colspan="2">Parameters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>Stopping rule</TD><TD>%d</TD></TR>',[Criteria]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>F to enter/remove</TD><TD>%.4f</TD></TR>',[FValue]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Sig. level</TD><TD>%.4f</TD></TR>',[SigValue]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Subset size</TD><TD>%d</TD></TR>',[NbAccepted]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TD colspan="2">Search algorith</TD></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Backward (0) or Forward (1)</TD><TD>%d</TD></TR>',[ord(ForwardSearch)]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TD colspan="2">Report</TD></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Show selected subset</TD><TD>%d</TD></TR>',[ord(IsShowSelected)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Show detailed results</TD><TD>%d</TD></TR>',[ord(IsShowDetailedResults)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD># columns for details</TD><TD>%d</TD></TR>',[NbColSoluce]);
 s:= s + '</TABLE>';
 result:= s;
end;

procedure TOpPrmFSStepDisc.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 Criteria:= prmINI.ReadInteger(prmSection,'criteria',1);
 FValue:= prmINI.ReadFloat(prmSection,'f_value',3.84);
 SigValue:= prmINI.ReadFloat(prmSection,'sig_value',0.05);
 NbAccepted:= prmINI.ReadInteger(prmSection,'nb_accepted',5);
 ForwardSearch:= prmINI.ReadBool(prmSection,'forward',true);
 NbColSoluce:= prmINI.ReadInteger(prmSection,'col_soluce',5);
 IsShowSelected:= prmINI.ReadBool(prmSection,'show_selection',true);
 IsShowDetailedResults:= prmINI.ReadBool(prmSection,'show_details',true);
end;

procedure TOpPrmFSStepDisc.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(Criteria,sizeof(Criteria));
 prmStream.ReadBuffer(FValue,sizeof(FValue));
 prmStream.ReadBuffer(SigValue,sizeof(SigValue));
 prmStream.ReadBuffer(NbAccepted,sizeof(NbAccepted));
 prmStream.ReadBuffer(ForwardSearch,sizeof(ForwardSearch));
 prmStream.ReadBuffer(NbColSoluce,sizeof(NbColSoluce));
 prmStream.ReadBuffer(IsShowSelected,sizeof(IsShowSelected));
 prmStream.ReadBuffer(IsShowDetailedResults,sizeof(IsShowDetailedResults));
end;

procedure TOpPrmFSStepDisc.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'criteria',Criteria);
 prmINI.WriteFloat(prmSection,'f_value',FValue);
 prmINI.WriteFloat(prmSection,'sig_value',SigValue);
 prmINI.WriteInteger(prmSection,'nb_accepted',NbAccepted);
 prmINI.WriteBool(prmSection,'forward',ForwardSearch);
 prmINI.WriteInteger(prmSection,'col_soluce',NbColSoluce);
 prmINI.WriteBool(prmSection,'show_selection',IsShowSelected);
 prmINI.WriteBool(prmSection,'show_details',IsShowDetailedResults);
end;

procedure TOpPrmFSStepDisc.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(Criteria,sizeof(Criteria));
 prmStream.WriteBuffer(FValue,sizeof(FValue));
 prmStream.WriteBuffer(SigValue,sizeof(SigValue));
 prmStream.WriteBuffer(NbAccepted,sizeof(NbAccepted));
 prmStream.WriteBuffer(ForwardSearch,sizeof(ForwardSearch));
 prmStream.WriteBuffer(NbColSoluce,sizeof(NbColSoluce));
 prmStream.WriteBuffer(IsShowSelected,sizeof(IsShowSelected));
 prmStream.WriteBuffer(IsShowDetailedResults,sizeof(IsShowDetailedResults));
end;

procedure TOpPrmFSStepDisc.SetDefaultParameters;
begin
 //critre de rgle d'arrt -- 0 : F-Value ; 1 :  Sig Level ; 2 : nombre de variables -- df. 1
 Criteria := 1;
 //Seuil F pour la comparaison -- df. 3.84
 FValue := 3.84;
 //Seuil proba pour la comparaison -- df. 0.05
 SigValue := 0.05;
 //Nombre de variables acceptes -- def. 5
 NbAccepted := 5;
 //sens de la recherche -- true -> forward, false -> backward
 ForwardSearch:= true;
 //nombre de colonnes  afficher dans les rsultats
 NbColSoluce:= 5;
 //lister les attributs slectionns dans le rapport ? -- df. false
 IsShowSelected:= true;
 //montrer le dtail du processus ? -- df. true
 IsShowDetailedResults:= true;
end;

initialization
 RegisterClass(TMLGenFSStepDisc);
end.
