(*************************************************************************)
(* UCompFSLogisticRegForward.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*************************************************************************)

{
@abstract(Slection de variables pour la rgression logistique binaire - Mthode Forward base sur la statistique du Score)
@author(Ricco)
@created(05/12/2007)

Trs simple :

(1) effectuer la rgression avec (p-1) variables dj slectionnes
(2) choisir la variable avec le score le plus lev parmi les non-slectionns
(3) l'inclure si significative i.e. p-value infrieure au seuil
(4) on peut aussi fixer un nombre maximal de variables (0 si cette option est dsactive)
}
unit UCompFSLogisticRegForward;

interface

USES
    Classes, IniFiles, Forms,
    UCompDefinition,
    UCompFSDefinition,
    UCompFSInputSelection,
    UOperatorDefinition,
    UCalcFSStepLogisticReg,
    UDatasetDefinition,
    UCompFSLogisticRegBackward;

TYPE
    ///gnrateur de backawrd selection pour rgression logistique
    TMLGenFSForwardLogReg = class(TMLGenFS)
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    ///composant backward selection pour rgression logistique
    TMLCompFSForwardLogReg = class(TMLCompFSInputSelection)
    protected
    function    getClassOperator: TClassOperator; override;
    end;

    TOpFSForwardLogReg = class(TOpFSBackwardLogReg)
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    procedure   RunSearch(); override;
    function    searchNext(inputList, outputList: TLstCandidat; pSeuil: double): TCandidat;
    end;

    TOpPrmFSRegLogForward = class(TOpPrmFSRegLogBackward)
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    MaxSelected: integer;//0 par dfaut i.e. non activ
    function    getHTMLParameters(): string; override;
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;    
    end;


implementation

uses
    Sysutils,
    UDatasetImplementation, FMath, ULogFile, UConstConfiguration,
    UCompSpvLogisticRegression, UCompManageDataset, d6_matrices,
    UCalcSpvStructScore, UDlgOpPrmFSLogRegForward;

const LAMBDA_FOR_REGULARIZATION = 0.0001;    

{ TMLGenFSForwardLogReg }

function TMLGenFSForwardLogReg.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFSForwardLogReg;
end;

{ TMLCompFSForwardLogReg }

function TMLCompFSForwardLogReg.getClassOperator: TClassOperator;
begin
 result:= TOpFSForwardLogReg;
end;

{ TOpPrmFSRegLogForward }

function TOpPrmFSRegLogForward.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmFSLogRegForward.CreateFromOpPrm(self);
end;

function TOpPrmFSRegLogForward.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan="2">Parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Probability cut</TD><TD align="right">%.4f</TD></TR>',[PValue]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD># sel. upper bound</TD><TD align="right">%d</TD></TR>',[MaxSelected]);
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TD colspan=2>Report</TD>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Show details</TD><TD align="right">%d</TD></TR>',[ord(IsShowDetails)]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD># columns</TD><TD align="right">%d</TD></TR>',[NbSoluceShowed]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpPrmFSRegLogForward.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 MaxSelected:= prmINI.ReadInteger(prmSection,'max_selected',MaxSelected);
end;

procedure TOpPrmFSRegLogForward.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(MaxSelected,sizeof(MaxSelected));
end;

procedure TOpPrmFSRegLogForward.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'max_selected',MaxSelected);
end;

procedure TOpPrmFSRegLogForward.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(MaxSelected,sizeof(MaxSelected));
end;

procedure TOpPrmFSRegLogForward.SetDefaultParameters;
begin
 inherited;
 MaxSelected:= 0;
end;

{ TOpFSForwardLogReg }

function TOpFSForwardLogReg.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFSRegLogForward;
end;

procedure TOpFSForwardLogReg.RunSearch;
var pSeuil: double;
    maxSel,nbSel: integer;
    inputList,outputList: TLstCandidat;
    j: integer;
    candidat: TCandidat;
begin
 pSeuil:= (self.PrmOp as TOpPrmFSRegLogForward).PValue;
 maxSel:= (self.PrmOp as TOpPrmFSRegLogForward).MaxSelected;
 inputList:= TLstCandidat.create(false); inputList.assign(Candidats); //liste locale des candidats
 outputList:= TLstCandidat.create(false); //vide, on va la remplir
 REPEAT
  nbSel:= outputList.getCount();
  //traiter les variables restantes
  if (inputList.getCount() > 0)
   then
    begin
     candidat:= searchNext(inputList,outputList, pSeuil);
     if (candidat <> nil)
      then
       begin
        inputList.remove(inputList.indexOf(candidat));
        outputList.add(candidat);
       end;
    end;
 //on a atteint le max. de sel (si option active) OU aucun ajout n'a t effectu  cette tape
 UNTIL ((maxSel > 0) and (outputList.getCount() >= maxSel)) OR (nbSel = outputList.getCount());
 //rajouter maintenant dans la sortie les variables slectionnes --> ceux qui sont rests dans la inputList sont les slectionns
 for j:= 0 to pred(outputList.getCount()) do
  self.OutputData.LstAtts[asInput].Add(outputList.getCandidat(j).Variable);
 //***************************************************************
 inputList.Free();
 outputList.Free();
end;

function TOpFSForwardLogReg.searchNext(inputList,outputList: TLstCandidat;
  pSeuil: double): TCandidat;
var curAtts: TLstAttributes;
    i,j,example: integer;
    model: TCalcLogisticReg;
    ok: boolean;
    probas: TVector;
    score: TTabScore;
    U,y,v: double;
    //z1,z2: double;
    levelRes: TLevelResultat;//rsultat pour un niveau
    res: TResultat;//un rsultat
    UBound: integer;
    H,VCV: TMatrix;
    j1,j2: integer;
    err: integer;
    //new -- 24/11/2010 -- buffrisation des valeurs
    z: array of double;
begin
 result:= nil;
 curAtts:= TLstAttributes.Create(false,self.WorkData.Examples.Size);//la slection courante
 for j:= 0 to pred(outputList.getCount()) do curAtts.Add(outputList.getCandidat(j).Variable);
 UBound:= curAtts.count+1;
 //raliser la rgression avec les variables dj slectionnes
 model:= TCalcLogisticReg.create(NIL,self.Target,curAtts,self.Inputs);
 ok:= model.learning(self.WorkData.Examples);
 if (ok = true)
  then
   begin
    //rsultat de la rgression pour le niveau
    levelRes:= TLevelResultat.create();//rsultat pour un niveau
    levelRes.AIC:= model.AIC;
    levelRes.CHi2:= model.CHI2;
    levelRes.ddl:= model.ddlCHI2;
    levelres.pValue:= model.pCHI2;
    Soluce.addLevel(levelRes);
    //prparer le terrain en formant le tableau des probas, etc. pour les variables slectionnes
    //Rq. le score des variables actives est forcment gal  zro, pas besoin de le calculer
    score:= TTabScore.Create(self.Target);
    dimVector(probas,self.Workdata.Examples.Size);
    for i:= 1 to self.WorkData.Examples.Size do
     begin
      example:= self.WorkData.Examples.Number[i];
      model.getScore(example,score);
      probas[i]:= score.getValue(1);
     end;
    score.Free();
    //*********************************
    //valuer chaque variable candidate
    //*********************************
    //new -- 24/11/2010 -- prparer le buffer
    setLength(z,curAtts.Count+2); //+2 parce qu'on rajoute une variable supplmentaire plus bas...
    z[0]:= 1.0;
    //
    for j:= 0 to inputList.getCount()-1 do
     begin
      res:= TResultat.create(inputList.getCandidat(j));
      res.Chi2:= 0.0;
      res.PValue:= 1.0;
      levelRes.addresult(res);
      U:= 0.0;
      curAtts.Add(inputList.getCandidat(j).Variable);
      dimMatrix(H,UBound,UBound);
      dimMatrix(VCV,UBound,UBound);
      //calculer son score et la matrice hessienne
      for i:= 1 to self.WorkData.Examples.Size do
       begin
        example:= self.WorkData.Examples.Number[i];
        y:= -1.0 * (self.Target.cValue[example] - 2.0);//1 --> 1 ; 2 --> 0
        v:= inputList.getCandidat(j).Variable.cValue[example];
        U:= U + (y - probas[i]) * v;
        //new -- 24/11/2010 -- remplir le buffer (il y a dj la variable additionnelle ici !)
        for j1:= 1 to curAtts.Count do z[j1]:= curAtts.Attribute[pred(j1)].cValue[example];
        //
        for j1:= 0 to curAtts.Count do
         for j2:= j1 to curAtts.Count do
          begin
           //if (j1 > 0) then z1:= curAtts.Attribute[pred(j1)].cValue[example] else z1:= 1.0;//
           //if (j2 > 0) then z2:= curAtts.Attribute[pred(j2)].cValue[example] else z2:= 1.0;//
           //H[j1,j2]:= H[j1,j2] + z1 * z2 * probas[i] * (1.0 - probas[i]);
           //new -- 24/11/2010
           H[j1,j2]:= H[j1,j2] + z[j1] * z[j2] * probas[i] * (1.0 - probas[i]);
          end;
       end;
      //remplir le triangle infrieur de la matrice hessienne
      for j1:= 0 to pred(curAtts.Count) do for j2:= succ(j1) to curAtts.Count do H[j2,j1]:= H[j1,j2];
      curAtts.delete(inputList.getCandidat(j).Variable);
      //rajouter un terme de rgularisation
      for j1:= 0 to UBound do H[j1,j1]:= H[j1,j1] - LAMBDA_FOR_REGULARIZATION;  
      //inverser la matrice
      err:= invMat(H,0,UBound,VCV);
      if (err = MAT_OK)
       then
        begin
         //former la statistique du chi2 avec la forme quadratique
         res.Chi2:= U * U * VCV[UBound,UBound];
         if (res.Chi2 > 0.0) then res.PValue:= fmath.PKhi2(1,res.Chi2);
        end;
      H:= nil;
      VCV:= nil;
     end;//du for j...
    Finalize(z);
    probas:= nil;
    //trier par ordre dcroissant les rsultats du niveau et vrifier que la variable est significative
    levelRes.sortDescendant();
    if (levelRes.getResult(0).PValue <= pSeuil)
     then
      begin
       levelRes.Best:= levelRes.getResult(0);
       result:= levelRes.Best.Candidat;
      end;
   end;
 model.Free();
 curAtts.Free();
end;

initialization
 RegisterClass(TMLGenFSForwardLogReg);
end.
