(*******************************************************************************)
(* UCompISContinuousAttributeValue.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(*******************************************************************************)

{
@abstract(Slectionner un sous-ensemble d'individus  partir d'une rgle simplifie)
@author(Ricco)
@created(07/08/2006)

Trs simplement, on utilise un repre "Attribut = valeur" pour slectionner les individus actifs.
On utilise ici une variable continue, la comparaison est de type "X < v" ou "X >= v"
}

unit UCompISContinuousAttributeValue;

interface

USES
        Classes, IniFiles, Forms,
        UCompDefinition, UOperatorDefinition,
        UCompISDefinition,
        UCompManageDataset,
        UDatasetDefinition;

TYPE
        {gnrateur}
        TMLGenCompISCAttValue = class(TMLGenComp)
                               protected
                               procedure   GenCompInitializations(); override;
                               public
                               function    GetClassMLComponent: TClassMLComponent; override;
                               end;

        {composant}
        TMLCompISCAttValue = class(TMLCompIS)
                            protected
                            function    getClassOperator: TClassOperator; override;
                            end;

        {oprateur}
        TOpISCAttValue = class(TOpLocalData)
                        protected
                        function    getClassParameter: TClassOperatorParameter; override;
                        function    CoreExecute(): boolean; override;
                        public
                        function    getHTMLResultsSummary(): string; override;
                        end;

        {paramtrage}
        TOpPrmISCAttValue = class(TOperatorParameter)
                           private
                           {nom de l'attribut de rfrence}
                           FAttName: string;
                           {oprateur de comparaison, cf. plus bas "TTypeContinuousComparaison"}
                           FCondOperator: integer;
                           {valeur de comparaison}
                           FValue: TTypeContinue;
                           protected
                           procedure   SetDefaultParameters(); override;
                           function    CreateDlgParameters(): TForm; override;
                           public
                           function    getHTMLParameters(): string; override;
                           //I-O
                           procedure   LoadFromStream(prmStream: TStream); override;
                           procedure   SaveToStream(prmStream: TStream); override;
                           procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
                           procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
                           //properties
                           property    AttName: string read FAttName write FAttName;
                           property    CondOperator: integer read FCondOperator write FCondOperator;
                           property    Value: TTypeContinue read FValue write FValue;
                           end;                        

implementation

USES
    Sysutils,
    UDatasetExamples,
    UDatasetImplementation, UDlgOpPrmISCAttributeValue,
    UStringAddBuffered;

//********************* types locaux pour une meilleure lecture du code **************************************
TYPE
        {type de comparaison : <, <=, = , <>, >=, >}
        TTypeContinuousComparaison = (ccmpL,ccmpLE,ccmpE,ccmpNE,ccmpGE,ccmpG);

CONST
        STR_IS_CONTINUOUS_COMPARAISON : array[TTypeContinuousComparaison] of string = ('<','>=','=','<>','>=','>');
//*************************************************************************************************************        

{ TMLGenCompISCAttValue }

procedure TMLGenCompISCAttValue.GenCompInitializations;
begin
 FMLComp:= mlcInstanceSelection;
end;

function TMLGenCompISCAttValue.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompISCAttValue; 
end;

{ TMLCompISCAttValue }

function TMLCompISCAttValue.getClassOperator: TClassOperator;
begin
 result:= TOpISCAttValue; 
end;

{ TOpISCAttValue }

function TOpISCAttValue.CoreExecute: boolean;
var attRef: TAttribute;
    value: TTypeContinue;
    operator: integer;
    exSource, exOut: TExamples;
    i,example: integer;
begin
 result:= FALSE;
 //rcuprer l'attribut
 attRef:= self.WorkData.LstAtts[asAll].GetFromName((self.PrmOp as TOpPrmISCAttValue).AttName);
 if assigned(attRef)
  then
   begin
    value:= (self.PrmOp as TOpPrmISCAttValue).Value;
    operator:= (self.PrmOp as TOpPrmISCAttValue).CondOperator;
    //brancher les pointeurs
    exSource:= self.WorkData.Examples;
    exOut:= self.OutputData.Examples;
    //filtrer les exemples
    exOut.BeginAdd();
    for i:= 1 to exSource.Size do
     begin
      example:= exSource.Number[i];
      case TTypeContinuousComparaison(operator) of
       ccmpL: if (attRef.cValue[example] < value) then exOut.AddExample(example);
       ccmpLE: if (attRef.cValue[example] <= value) then exOut.AddExample(example);
       ccmpE: if (attRef.cValue[example] = value) then exOut.AddExample(example);
       ccmpNE: if (attRef.cValue[example] <> value) then exOut.AddExample(example);
       ccmpGE: if (attRef.cValue[example] >= value) then exOut.AddExample(example);
       //else du case... (le dernier cas possible est ">")
       else if (attRef.cValue[example] > value) then exOut.AddExample(example);
      end;
     end;
    exOut.EndAdd();
    result:= (exOut.Size > 0);
   end;
end;

function TOpISCAttValue.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmISCAttValue; 
end;

function TOpISCAttValue.getHTMLResultsSummary: string;
begin
 result:= format('<b>%d</b> examples selected from <b>%d</b>',[self.OutputData.Examples.Size,self.workdata.examples.Size]);
end;

{ TOpPrmISCAttValue }

function TOpPrmISCAttValue.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmISCAttValue.CreateFromOpPrm(self);
end;

function TOpPrmISCAttValue.getHTMLParameters: string;
begin
 result:= format('<b>%s %s %.2f</b>',[FAttName,STR_IS_CONTINUOUS_COMPARAISON[TTypeContinuousComparaison(FCondOperator)],FValue])
end;

procedure TOpPrmISCAttValue.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FAttName:= prmINI.ReadString(prmSection,'attribute',FAttName);
 FCondOperator:= prmINI.ReadInteger(prmSection,'cond',FCondOperator);
 FValue:= prmINI.ReadFloat(prmSection,'value',FValue);
end;

procedure TOpPrmISCAttValue.LoadFromStream(prmStream: TStream);
begin
 loadStringFromStream(FAttName,prmStream);
 prmStream.ReadBuffer(FCondOperator,SizeOf(FCondOperator));
 prmStream.readBuffer(FValue,SizeOf(FValue));
end;

procedure TOpPrmISCAttValue.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteString(prmSection,'attribute',FAttName);
 prmINI.WriteInteger(prmSection,'cond',FCondOperator);
 prmINI.WriteFloat(prmSection,'value',FValue);
end;

procedure TOpPrmISCAttValue.SaveToStream(prmStream: TStream);
begin
 saveStringToStream(FAttName,prmStream);
 prmStream.WriteBuffer(FCondOperator,SizeOf(FCondOperator));
 prmStream.WriteBuffer(FValue,SizeOf(FValue));
end;

procedure TOpPrmISCAttValue.SetDefaultParameters;
begin
 FAttName:= '';
 FCondOperator:= 0;
 FValue:= 0;
end;

initialization
 RegisterClass(TMLGenCompISCAttValue);
end.
