(************************************************************)
(* UCompISFirst.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(************************************************************)

{
@abstract(Slection des "m" premires observations parmi "n")
@author(Ricco)
@created(12/05/2012)

Il s'agit de mettre en SELECTED les "m" premires observations.
Les "n-m" deviennent UNSELECTED

}

unit UCompISFirst;

interface

USES
        Classes, IniFiles, Forms,
        UCompDefinition, UOperatorDefinition,
        UCompISDefinition,
        UCompManageDataset;

TYPE
       {gnrateur}
       TMLGenCompISFirst = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
       end;

       {composant}
       TMLCompISFirst = class(TMLCompIS)
        protected
        function    getClassOperator: TClassOperator; override;
       end;

       {oprateur}
       TOpISFirst = class(TOpLocalData)
        private
        FNbInstances: integer;
        procedure   calcNbInstances(n: integer);
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    CoreExecute(): boolean; override;
        public
        function    getHTMLResultsSummary(): string; override;
       end;

       {paramtrage}
       TOpPrmISFirst = class(TOperatorParameter)
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        public
        //nombre d'observations  slectionner
        //si m < 1, c'est une proportion
        m: double;
        function    getHTMLParameters(): string; override;
        //I-O
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
       end;


implementation

USES
        Sysutils,
        UDatasetDefinition, UDatasetImplementation,
        UDatasetExamples, UStringAddBuffered, UDlgOpPrmISFirst;

{ TMLGenCompISFirst }

procedure TMLGenCompISFirst.GenCompInitializations;
begin
 FMLComp:= mlcInstanceSelection;
end;

function TMLGenCompISFirst.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompISFirst;
end;

{ TMLCompISFirst }

function TMLCompISFirst.getClassOperator: TClassOperator;
begin
 result:=TOpISFirst;
end;

{ TOpISFirst }

procedure TOpISFirst.calcNbInstances(n: integer);
var m :double;
begin
 m:= (self.PrmOp as TOpPrmISFirst).m;
 if (m >= 1)
  then FNbInstances:= trunc(m)
  else FNbInstances:= trunc(m * n);
 //viter les dbordements
 if (FNbInstances > n) then FNbInstances:= n;
end;

function TOpISFirst.CoreExecute: boolean;
var exSource, exOut: TExamples;
begin
 exSource:= self.WorkData.Examples;
 exOut:= self.OutputData.Examples;
 //tout copier
 exOut.Copy(exSource);
 //puis retailler le cas chant
 self.calcNbInstances(exSource.Size); 
 if (FNbInstances < exSource.Size)
  then exOut.Size:= FNbInstances;
 //
 result:= true;
end;

function TOpISFirst.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmISFirst;
end;

function TOpISFirst.getHTMLResultsSummary: string;
begin
 result:= Format('<b>%d</b> selected examples from <b>%d</b>',[self.OutputData.Examples.Size,self.workdata.examples.Size]);
end;

{ TOpPrmISFirst }

function TOpPrmISFirst.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmISFirst.CreateFromOpPrm(self);
end;

function TOpPrmISFirst.getHTMLParameters: string;
begin
 result:= format('Sample size : %g<br>',[m]);
end;

procedure TOpPrmISFirst.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 m:= prmINI.ReadFloat(prmSection,'m',m);
end;

procedure TOpPrmISFirst.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(m,sizeof(m));
end;

procedure TOpPrmISFirst.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteFloat(prmSection,'m',m);
end;

procedure TOpPrmISFirst.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(m,sizeof(m));
end;

procedure TOpPrmISFirst.SetDefaultParameters;
begin
 m:= 500;
end;

initialization
 RegisterClass(TMLGenCompISFirst); 
end.
