(**********************************************************************)
(* UCalcPLSRegressionPlus.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Rgression PLS - Version unifie)
@author(Ricco)
@created(16/05/2008)

Version unifie de la Rgression PLS (PLS Regression et PLS Factorial). L'ide surtout
est de produire des sorties qui concordent avec les standards du domaine. La principale
rfrence est SIMCA-P. Dans le tutoriel ddi, nous ferons le parallle avec R, SAS et SPAD.

Pour la doc. des indicateurs, GARSON fait trs fort encore --> http://www2.chass.ncsu.edu/garson/PA765/pls.htm 
}

unit UCalcPLSRegressionPlus;

interface

USES
    JFGCalcPLS;

TYPE
    //hritier de la classe de calcul, seuls les sorties vont tre modifies maintenant
    TCalcPLSPlus = class(TCalcPLS)
    public
    //proportion de variance explique
    function getHTMLPropVarExplained(): string;
    //R2 et redondance
    function getHTMLR2AndRedundancy(): string;
    //X-loadings - lien entre INPUTS et les X-scores (interprtation des X-scores - vecteur Ph)
    function getHTMLXLoadings(): string;
    //X-weights - lien entre les INPUTS et les Y-scores (Wh* est d'ailleurs le coefficient de projection)
    function getHTMLInputsWeights(): string;
    //Y-weights - lien entre les TARGET et les Y-scores (vecteur Ch)
    function getHTMLTargetWeights(): string;
    //VIP - importance des variables dans la projection - c'est le VIP sur le dernier facteur qui est dterminant
    function getHTMLVip(): string;
    //coefficients de prdiction - non standardiss
    function getHTMLUnstandardizedCoef(): string;
    //coefficients de prdiction - standardiss
    function getHTMLStandardizedCoef(): string;
    end;

implementation

uses
    Sysutils,
    UStringAddBuffered, UConstConfiguration;

{ TCalcPLSPlus }

function TCalcPLSPlus.getHTMLInputsWeights: string;
var buf: TBufString;
    s: string;
    h,j: integer;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 s:= '<BR><BR><H4>X-Weights (Model Effect Weights - Vector Wh and Wh*)</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH width=100>-</TH><TH colspan=%d>Wh</TH><TH colspan=%d><font color="Green">Wh*</font></TH></TR>',[FNbAxis,FNbAxis]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Input</TH>';
 for h:=0 to pred(FNbAxis) do s:= s + format('<TH width=60> Factor %d</TH>',[h+1]);
 for h:=0 to pred(FNbAxis) do s:= s + format('<TH width=60><font color="Green">Factor %d</font></TH>',[h+1]);
 s:= s + '</TR>';
 buf.AddStr(s);
 for j:=0 to pred(FCountX) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDataSourceInput.Attribute[j].Name]);
   for h:=0 to pred(FNbAxis) do s:= s + format('<TD align=center>%.4f</TD>',[FWh[j,h]]);
   for h:=0 to pred(FNbAxis) do s:= s + format('<TD align=center><font color="Green">%.4f</font></TD>',[FWhEtoile[j,h]]);
   s:= s+'</TR>';
   buf.AddStr(s);
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;


function TCalcPLSPlus.getHTMLPropVarExplained: string;
var buf: TBufString;
    s: string;
    h: integer;
    sumX,sumY: double;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr('<H4>Proportion of Variance Explained (Percentage)</H4>');
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>-</TH><TH colspan=2>Input variables (X)</TH><TH colspan=2>Target Variables (Y)</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Latent Factor</TH><TH width=80>Current X (%)</TH><TH width=80><font color="blue">Cumulative X (%)</font></TH><TH width=80>Current Y (%)</TH><TH width=80><font color="blue">Cumulative Y (%)</font></TH></TR>';
 buf.AddStr(s);
 sumX:= 0.0;
 sumY:= 0.0;
 for h:= 1 to self.NbAxis do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%d</TH>',[h]);
   sumX:= sumX + FRdXT[h-1];
   sumY:= sumY + FRdYT[h-1];
   s:= s + format('<TD align=center>%.3f</TD><TD align=center><font color="blue">%.3f</font></TD><TD align=center>%.3f</TD><TD align=center><font color="blue">%.3f</font></TD></TR>',
                  [100*FRdXT[h-1],100*sumX,100*FRdYT[h-1],100*sumY]);
   buf.AddStr(s);
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TCalcPLSPlus.getHTMLR2AndRedundancy: string;
var buf: TBufString;
    s: string;
    h,j: integer;
    value,sum: double;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr('<BR><BR><H4>R-squared</H4>');
 //les inputs vs. x-scores
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TH colspan=%d>Input(s) vs. X-Scores</TH></TR>',[FNbAxis*2+1]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>-</TH>' + format('<TH colspan=%d>R-squared</TH><TH colspan=%d><font color="blue">Cumulative R-squared</font></TH></TR>',[FNbAxis,FNbAxis]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=100>Input</TH>';
 for h:=0 to pred(FNbAxis) do s:= s+format('<TH width=60>Factor %d</TH>',[h+1]);
 for h:=0 to pred(FNbAxis) do s:= s+format('<TH width=60><font color="blue">Factor %d</font></TH>',[h+1]);
 s:= s+'</TR>';
 buf.AddStr(s);
 for j:=0 to pred(FCountX) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[FDataSourceInput.Attribute[j].Name]);
   //faire la srie des individuels
   for h:=0 to pred(FNbAxis) do
    begin
     value:=FCorXT[j,h]*FCorXT[j,h];
     s:= s+format('<TD align=center>%.4f</TD>',[value]);
    end;
   //faire la srie des cumulatifs
   sum:= 0.0;
   for h:=0 to pred(FNbAxis) do
    begin
     value:=FCorXT[j,h]*FCorXT[j,h];
     sum:= sum + value;
     s:= s+format('<TD align=center><font color="blue">%.4f</font></TD>',[sum]);
    end;
   s:= s+'</TR>';
   buf.AddStr(s);
  end;
 s:= HTML_TABLE_COLOR_DATA_GREEN + '<TH>Total Exp.</TH>';
 for h:=0 to pred(FNbAxis) do s:= s+format('<TD align=center>%.4f</TD>',[FRdXT[h]]);
 sum:= 0.0;
 for h:=0 to pred(FNbAxis) do
  begin
   sum:= sum + FRdXT[h]; 
   s:= s+format('<TD align=center><font color="blue">%.4f</font></TD>',[sum]);
  end;
 s:= s + '</table>';
 buf.AddStr(s);
 buf.AddStr('<BR><BR>');
 //les targets vs. x-scores
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TH colspan=%d>Target(s) vs. X-Scores</TH></TR>',[FNbAxis*2+1]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>-</TH>' + format('<TH colspan=%d>R-squared</TH><TH colspan=%d>Cumulative R-squared</TH></TR>',[FNbAxis,FNbAxis]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=100>Target</TH>';
 for h:=0 to pred(FNbAxis) do s:= s+format('<TH width=60>Factor %d</TH>',[h+1]);
 for h:=0 to pred(FNbAxis) do s:= s+format('<TH width=60>Factor %d</TH>',[h+1]);
 s:= s+'</TR>';
 buf.AddStr(s);
 for j:=0 to pred(FCountY) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY+format('<TH>%s</TH>',[FDataSourceTarget.Attribute[j].Name]);
   //faire la srie des individuels
   for h:=0 to pred(FNbAxis) do
    begin
     value:=FCorYT[j,h]*FCorYT[j,h];
     s:= s+format('<TD align=center>%.4f</TD>',[value]);
    end;
   //faire la srie des cumulatifs
   sum:= 0.0;
   for h:=0 to pred(FNbAxis) do
    begin
     value:=FCorYT[j,h]*FCorYT[j,h];
     sum:= sum + value;
     s:= s+format('<TD align=center><font color="blue">%.4f</font></TD>',[sum]);
    end;
   s:= s+'</TR>';
   buf.AddStr(s);
  end;
 s:= HTML_TABLE_COLOR_DATA_GREEN + '<TH>Total Exp.</TH>';
 for h:=0 to pred(FNbAxis) do s:= s+format('<TD align=center>%.4f</TD>',[FRdYT[h]]);
 sum:= 0.0;
 for h:=0 to pred(FNbAxis) do
  begin
   sum:= sum + FRdYT[h]; 
   s:= s+format('<TD align=center><font color="blue">%.4f</font></TD>',[sum]);
  end;
 s:= s + '</table>';
 buf.AddStr(s);
 // voir si c'est utile...
 //buf.AddStr('<H5>Target(s) vs. Y-Scores</H5>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TCalcPLSPlus.getHTMLStandardizedCoef: string;
var buf: TBufString;
    s: string;
    j,k: integer;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 s:= '<BR><BR><H4>Standardized Regression Parameters (by target variable)</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + format('<TH width=100>-</TH><TH colspan=%d>Target Variable(s)</TH></TR>',[FCountY+2]);
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + '<TH width=100>-</TH><TH colspan=2>Average</TH>';
 for k:= 0 to pred(FCountY) do s:= s + format('<TD align=right>%.4f</TD>',[StatPLS.FAvgY[k]]);
 s:= s + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GREEN + '<TH width=100>-</TH><TH colspan=2>Std.Deviation</TH>';
 for k:= 0 to pred(FCountY) do s:= s + format('<TD align=right>%.4f</TD>',[StatPLS.FStdY[k]]);
 s:= s + '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Input</TH><TH width=80>Average</TH><TH width=80>Std.Deviation</TH>';
 for k:= 0 to pred(FCountY) do s:= s + format('<TH width=80>%s</TH>',[FDataSourceTarget.Attribute[k].Name]);
 s:= s + '</TR>';
 buf.AddStr(s);
 //
 for j:= 0 to pred(FCountX) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDataSourceInput.Attribute[j].Name]);
   s:= s + format('<TD align=right %s>%.4f</TD><TD align=right %s>%.4f</TD>',[HTML_BGCOLOR_DATA_GREEN,StatPLS.FAvgX[j],HTML_BGCOLOR_DATA_GREEN,StatPLS.FStdX[j]]);
   for k:= 0 to pred(FCountY) do s:= s + format('<TD align=center>%.6f</TD>',[FBStdCoefReg[j,k]]);
   s:= s + '</TR>';
   buf.AddStr(s);
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;


function TCalcPLSPlus.getHTMLTargetWeights: string;
var buf: TBufString;
    s: string;
    h,j: integer;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 s:= '<BR><BR><H4>Y-Weights (Dependent Variable Weights - Vector Ch)</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Target</TH>';
 for h:=0 to pred(FNbAxis) do s:= s + format('<TH width=60>Factor %d</TH>',[h+1]);
 s:= s + '</TR>';
 buf.AddStr(s);
 for j:=0 to pred(FCountY) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDataSourceTarget.Attribute[j].Name]);
   for h:=0 to pred(FNbAxis) do s:= s + format('<TD align=center>%.4f</TD>',[FCh[j,h]]);
   s:= s+'</TR>';
   buf.AddStr(s);
 end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;


function TCalcPLSPlus.getHTMLUnstandardizedCoef: string;
var buf: TBufString;
    s: string;
    j,k: integer;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 s:= '<BR><BR><H4>Unstandardized Regression Parameters (by target variable)</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + format('<TH width=100>-</TH><TH colspan=%d>Target Variable(s)</TH></TR>',[FCountY]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Input</TH>';
 for k:= 0 to pred(FCountY) do s:= s + format('<TH width=100>%s</TH>',[FDataSourceTarget.Attribute[k].Name]);
 s:= s + '</TR>';
 buf.AddStr(s);
 for j:= 0 to pred(FCountX) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDataSourceInput.Attribute[j].Name]);
   for k:= 0 to pred(FCountY) do s:= s + format('<TD align=center>%.6f</TD>',[FBCoefReg[j,k]]);
   s:= s + '</TR>';
   buf.AddStr(s);
  end;
 //ne pas oublier la constante
 s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',['_constant_']);
 for k:= 0 to pred(FCountY) do s:= s + format('<TD align=center>%.6f</TD>',[FBCoefReg[FCountX,k]]);
 s:= s + '</TR>';
 buf.AddStr(s);
 //
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TCalcPLSPlus.getHTMLVip: string;
var buf: TBufString;
    s,sCol: string;
    h,j: integer;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 s:= '<BR><BR><H4>VIP (Variable Importance in Projection)</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + '<TH>Input</TH>';
 for h:=0 to pred(FNbAxis) do s:= s+format('<TH width=60>Factor %d</TH>',[h+1]);
 s:= s+'</TR>';
 buf.AddStr(s);
 for j:=0 to pred(FCountX) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDataSourceInput.Attribute[j].Name]);
   for h:=0 to pred(FNbAxis) do
    if (h = pred(FNbAxis))
     then
      begin
       if (FVIP[j,h] > 1.0) then sCol:= HTML_BGCOLOR_LIGHT_RED else sCol:= HTML_BGCOLOR_DATA_GRAY; 
       s:= s+format('<TD align=center %s><font color="blue">%.4f</font></TD>',[sCol,FVIP[j,h]]);
      end
     else s:= s+format('<TD align=right>%.4f</TD>',[FVIP[j,h]]);
   s:= s+'</TR>';
   buf.AddStr(s);
 end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TCalcPLSPlus.getHTMLXLoadings: string;
var buf: TBufString;
    s: string;
    h,j: integer;
    value: double;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 s:= '<BR><BR><H4>X-loadings (Model Effect Loadings - Vector Ph)</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s+'<TH>Input</TH>';
 for h:=0 to pred(FNbAxis) do s:= s+format('<TH width=60>Factor%d</TH>',[h+1]);
 s:= s+'</TR>';
 buf.AddStr(s);
 for j:=0 to pred(FCountX) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[FDataSourceInput.Attribute[j].Name]);
   for h:=0 to pred(FNbAxis) do
    begin
     value:=FPh[j,h];
     s:= s + format('<TD align=center>%.4f</TD>',[value]);
    end;
   s:= s+'</TR>';
   buf.AddStr(s);
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

end.
