(**************************************************************************)
(* UCalcPLSSelectionStructure.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(**************************************************************************)

{
@abstract(PLS -- Structure de gestion des PLS)
@author(Ricco)
@created(29/08/2006)

Grer les donnes pour la PLS

}

unit UCalcPLSSelectionStructure;

interface

USES
    Forms,Classes,
    IniFiles,Contnrs,
    UCompDefinition,
    UOperatorDefinition,
    UCompManageDataset,
    UDatasetDefinition,
    UDatasetExamples,
    UCalcRndGenerator,
    JFGCompPLS;

TYPE
    {donnes apprentissage-test}
    TDataCVPLS = class
    private
    FTrain, FTest: TExamples;
    public
    constructor create(size: integer);
    destructor  destroy(); override;
    procedure   copyIndexes(train,test: TExamples);
    function    getTrain(): TExamples;
    function    getTest(): TExamples;
    end;

    {liste des donnes}
    TLstDataCVPLS = class
    private
    FLstData: TObjectList;
    FNbCv: integer;
    public
    constructor create(nbCV: integer; size: integer);
    destructor  destroy(); override;
    //partitionner les donnes en autant de app.-test. pour la validation croise
    //seed ici peut prendre que deux valeurs : random ou standard, user ne doit jamais tre appel
    procedure   partitionning(prmExamples: TExamples; rndSeed: TStartSeed);
    //indice 0 correspond  toutes les donnes (dupliques en app-test), de 1  nbCV chaque couple app-test
    function    getData(numCV: integer): TDataCVPLS;
    //renvoyer le nombre de folds
    function    getNbFolds(): integer;
    end;

implementation

uses
    Sysutils,
    UConstConfiguration;

{ TDataCVPLS }

procedure TDataCVPLS.copyIndexes(train, test: TExamples);
begin
 FTrain.Copy(train);
 FTest.Copy(test);
end;

constructor TDataCVPLS.create(size: integer);
begin
 inherited create();
 FTrain:= TExamples.Create(size);
 FTest:= TExamples.Create(size);
end;

destructor TDataCVPLS.destroy;
begin
 FTrain.Free();
 FTest.Free();
 inherited;
end;

function TDataCVPLS.getTest: TExamples;
begin
 result:= FTest;
end;

function TDataCVPLS.getTrain: TExamples;
begin
 result:= FTrain;
end;

{ TLstDataCVPLS }

constructor TLstDataCVPLS.create(nbCV, size: integer);
var cv: integer;
begin
 inherited create();
 FNbCV:= nbCV;
 FLstData:= TObjectList.Create(TRUE);
 for cv:= 1 to succ(nbCV) do
  FLstData.Add(TDataCVPLS.create(size));
end;

destructor TLstDataCVPLS.destroy;
begin
 FLstData.Free();
 inherited;
end;

function TLstDataCVPLS.getData(numCV: integer): TDataCVPLS;
begin
 result:= FLstData.Items[numCV] as TDataCVPLS;
end;

function TLstDataCVPLS.getNbFolds: integer;
begin
 result:= FNbCV;
end;

procedure TLstDataCVPLS.partitionning(prmExamples: TExamples; rndSeed: TStartSeed);
var data: TExamples;
    dataCV: TDataCVPLS;
    train, test: TExamples;
    i,f,bb,bh,foldSize: integer;
begin
 //mlanger les donnes
 data:= TExamples.Create(prmExamples.Size);
 data.Copy(prmExamples);
 data.procRandomizeExamples(rndSeed);
 //les donnes 0 sont tous les exemples
 dataCV:= self.getData(0);
 dataCV.getTrain().Copy(data);
 dataCV.getTest().Copy(data);
 //prparer les index. intermdiaires
 train:= TExamples.Create(prmExamples.Size);
 test:= TExamples.Create(prmExamples.Size);
 //puis partitionner maintenant
 foldSize:= prmExamples.Size div FNbCV;
 for f:= 1 to FNbCV do
  begin
   //construire la liste des individus
   bb:= pred(f)*foldSize;
   bh:= f*foldSize+1;
   train.BeginAdd();
   test.BeginAdd();
   for i:= 1 to data.Size do
    begin
     if (i>bb) and (i<bh)
      then test.AddExample(data.Number[i])
      else train.AddExample(data.Number[i]);
    end;
   train.EndAdd();
   test.EndAdd();
   //puis affecter
   dataCV:= self.getData(f);
   dataCV.getTrain().Copy(train);
   dataCV.getTest().Copy(test);
  end;
 //librer
 train.Free();
 test.Free();
 data.Free();
end;

end.
