(***************************************************************)
(* UCalcFactNIPALS.pas - Copyright (c) 2005 Ricco RAKOTOMALALA *)
(***************************************************************)

{
@abstract(PLS -- classes de calcul)
@author(Ricco)
@created(01/05/2008)

Optimiser la classe de calcul PLS pour l'apprentissage supervis - PLS-DA et PLS-LDA.
Principale nouveaut : le nombre d'axes peut tre dtermin au fur et  mesure.

Deux choix posisbles :
(1) nombre d'axes fix par l'utilisateur, on ne change rien
(2) nombre d'axes = 0, un second seuil (Redondance sur Y) est introduit. Dans ce cas
le nombre d'axes max. explor est 20 (si la rgle d'arrt sur redondance n'a pas agit d'ici l).

Les donnes sont centres et rduites d'office. Idem, on ne demande pas le rapport dtaill.

}
unit UCalcPLSSupervised;

interface

USES
    JFGCalcPLS,
    UDatasetDefinition,
    UDatasetExamples;

TYPE
    ///classe de calcul pour PLS spcialis pour le supervis
    TCalcPLSForSpv = class(TCalcPLS)
    public
    ///nombre d'axes max.
    FMaxNbAxis: integer;
    ///dtection automatique des axes
    FNbAxisAutoDetect: boolean;
    ///seuil pour la redondance
    FSeuilRdY: double;
    ///r-crire le constructeur
    constructor create(sourceInput: TLstAttributes; sourceTarget: TLstAttributes; examples: TExamples; prmNbAxis: integer = 0; prmRdYCut : double = 0.01);
    ///lancer les calculs - c'est rellement une r-criture de la procdure avec moults copier/coller
    procedure   runAnalysis(examples: TExamples); override;
    end;

implementation

USES
    MATH, ULogFile, Sysutils, Classes;

CONST
    ///nombre d'axes max. par dfaut
    DEFAULT_MAX_NB_AXIS : integer = 30;

    ///tolerance pour la convergence
    PLS_TOLERANCE = 1.0e-8;

    ///nombre d'itration maximum
    PLS_MAX_ITERATION = 300;

{ TCalcSpvPLS }

constructor TCalcPLSForSpv.create(sourceInput, sourceTarget: TLstAttributes;
  examples: TExamples; prmNbAxis: integer; prmRdYCut: double);
begin
 FMaxNbAxis:= sourceInput.Count;
 if (prmNbAxis > 0)
  then
   begin
    FMaxNbAxis:= MATH.MIN(prmNbAxis,FMaxNbAxis);
    FNbAxisAutoDetect:= false;
   end
  else
   begin
    FMaxNbAxis:= MATH.MIN(DEFAULT_MAX_NB_AXIS,FMaxNbAxis);
    FNbAxisAutoDetect:= true;
   end;
 //rcup. seuil de redondance quoiqu'il en soit
 FSeuilRdY:= prmRdYCut;
 //on fixe le nombre d'axes par le max. pour l'initialisation des tableaux
 inherited create(SourceInput,SourceTarget,examples,true,FMaxNbAxis,false);
end;

procedure TCalcPLSForSpv.runAnalysis(examples: TExamples);
var h,i,j,k,iteration: integer;
    curWh,prodThi,sumWhi,prodUhi,sumChi: TVectorPLS;
    FXhpred,FYhpred,FWhi,FChi,FXhinit,FYhinit: TMatrixPLS;
    prodTh,sumWh,prodUh,sumCh,sumPh,sumUh,sum: TDataPLS;
    okConvergence: boolean;
    somme: double;
    constante: TDataPLS;
    RdY_is_Inf_Seuil: boolean;

begin
    TRY
        //rcupration des paramtres, calcul des stats ( moyennes, ecarts types)
        FStatData:= TStatPLS.create(FDataSourceInput,FDataSourceTarget,examples);
        FStatData.runComputation(FDataSourceInput,examples,FCountX,FStatData.FAvgX,FStatData.FStdX);
        FStatData.runComputation(FDataSourceTarget,examples,FCountY,FStatData.FAvgY,FStatData.FStdY);
        //TraceLog.WriteToLogFile(format('[PLS] calcul des stats = %d ms.',[getTickCount()-tps]));


        //Allocation des divers vecteurs et matrices utiliss
        setLength(FYhpred,succ(FSize),FCountY);
        setLength(FXhpred,succ(FSize),FCountX);
        setLength(FWhi,succ(FSize),FCountX);
        setLength(FChi,succ(FSize),FCountY);
        setLength(sumWhi,succ(FSize));
        setLength(prodUhi,succ(FSize));
        setLength(prodThi,succ(FSize));
        setLength(sumChi,succ(FSize));

        //tps:= getTickCount();
        //remplir les premires versions de Xh et Yh
        self.setFirstMatrix(examples,FDataSourceInput,FCountX,FVarianceTotaleX,FUsedStdDevX,FStatData.FAvgX,FStatData.FStdX,FXh);
        self.setFirstMatrix(examples,FDataSourceTarget,FCountY,FVarianceTotaleY,FUsedStdDevY,FStatData.FAvgY,FStatData.FStdY,FYh);

        //sauvegarde des matrices initiales pour pouvoir les utiliser  la fin de l'algorithme
        //inutile si pas besoin des calculs dtaills
        //if FDetailedResult
        if TRUE
         then
          begin
            setLength(FYhinit,succ(FSize),FCountY);
            setLength(FXhinit,succ(FSize),FCountX);
            for i:= 1 to FSize do for j:= 0 to pred(FCountX) do FXhinit[i,j]:=FXh[i,j];
            for i:= 1 to FSize do for k:= 0 to pred(FCountY) do FYhinit[i,k]:=FYh[i,k];
          end;
        //TraceLog.WriteToLogFile(format('[PLS] remplissage (initialisation) des matrices = %d ms.',[getTickCount()-tps]));

        //vecteur Wh courant
        setLength(curWh,FCountX);

        //*************************
        //***** production des axes
        //*************************

        //***** pour chaque axe  produire
        //for h:= 0 to pred(FNbAxis) do
        //quelques modifs
        h:= 0;
        if (FNbAxisAutoDetect = true) then FNbAxis:= 1;
        RdY_is_Inf_Seuil:= false;
        //selon FNbAxisAutoDetect, une des conditions sera vrifie....
        while ((FNbAxisAutoDetect = false) and (h < FNbAxis)) OR ((FNbAxisAutoDetect = true) and ((RdY_is_Inf_Seuil = false) and (h < FMaxNbAxis))) do

                begin

                         //copier la premire colonne de Yh-1 dans Uh
                         for i:= 1 to FSize do FUh[i,h]:= FYh[i,0];


                        //initialiser le vecteur Wh courant
                        for j:= 0 to pred(FCountX) do curWh[j]:= 0.0;

                        //*** on rpete jusqu' convergence ***
                        iteration:= 1;

                         REPEAT

                                //calculer Uh'Uh
                                prodUh:= 0.0;

                                for i:= 1 to FSize do

                                begin
                                        prodUh:= prodUh+FUh[i,h]*FUh[i,h];
                                end;

                                //calculer Wh
                                 for j:= 0 to pred(FCountX) do

                                         begin
                                                 sumWh:= 0.0;
                                                for i:= 1 to FSize do sumWh:= sumWh+FXh[i,j]*FUh[i,h];
                                                 //normaliser par Uh'Uh

                                                if (prodUh>0)
                                                then sumWh:= sumWh/prodUh;

                                                //acffecter  Wh
                                                 curWh[j]:= sumWh;

                                         end;

                                //normer Wh  1
                                sumWh:= 0.0;

                                for j:= 0 to pred(FCountX) do sumWh:= sumWh+curWh[j]*curWh[j];

                                if (sumWh>0)
                                then for j:= 0 to pred(FCountX) do curWh[j]:= curWh[j]/SQRT(sumWh);



                                //calculer la projection Th
                                for i:= 1 to FSize do

                                        begin
                                                 prodTh:= 0.0;
                                                 for j:= 0 to pred(FCountX) do prodTh:= prodTh+FXh[i,j]*curWh[j];


                                                //affecter
                                                 FTh[i,h]:= prodTh;
                                         end;

                                //calculer Th'Th
                                prodTh:= 0.0;

                                for i:= 1 to FSize do prodTh:= prodTh+FTh[i,h]*FTh[i,h];

                                //calculer Ch
                                 for k:= 0 to pred(FCountY) do

                                         begin
                                                sumCh:= 0.0;
                                                for i:= 1 to FSize do sumCh:= sumCh+FYh[i,k]*FTh[i,h];
                                                 //normaliser par Th'Th

                                                if (prodTh>0)
                                                then sumCh:= sumCh/prodTh;

                                                //affecter  Ph
                                                 FCh[k,h]:= sumCh;

                                         end;

                                  //calculer Uh
                                for i:= 1 to FSize do

                                        begin
                                                 sumUh:= 0.0;
                                                 for k:= 0 to pred(FCountY) do sumUh:= sumUh+FYh[i,k]*FCh[k,h];


                                                 FUh[i,h]:= sumUh;
                                         end;



                                //tester la convergence
                                okConvergence:= isConvergent(curWh,h);


                                //rcuprer le Wh
                                for j:= 0 to pred(FCountX) do FWh[j,h]:= curWh[j];

                                inc(iteration);

                        UNTIL (iteration>PLS_MAX_ITERATION) or okConvergence;

                         //calculer Th'Th
                        prodTh:= 0.0;
                        for i:= 1 to FSize do prodTh:= prodTh+FTh[i,h]*FTh[i,h];

                        //Calcul de Ph
                                 for j:= 0 to pred(FCountX) do

                                         begin
                                                 sumPh:= 0.0;
                                                for i:= 1 to FSize do sumPh:= sumPh+FXh[i,j]*FTh[i,h];
                                                 //normaliser par Th'Th

                                                if (prodTh>0) then sumPh:= sumPh/prodTh;

                                                //affecter  Ph
                                                 FPh[j,h]:= sumPh;

                                         end;

                        // sauvegarde de l'etat des matrices des rsidus
                        for i:= 1 to FSize do

                         begin
                                for j:= 0 to pred(FCountX) do FXhpred[i,j]:=FXh[i,j];
                                for k:= 0 to pred(FCountY) do FYhpred[i,k]:=FYh[i,k] ;

                         end;



                        //nouvelle version de Xh -- tape 2.4
                         for i:= 1 to FSize do

                                begin
                                        for j:= 0 to pred(FCountX) do FXh[i,j]:= FXh[i,j]-FTh[i,h]*FPh[j,h];

                                 end;


                         //nouvelle version de Yh -- tape 2.5
                         for i:= 1 to FSize do

                                begin
                                        for k:= 0 to pred(FCountY) do FYh[i,k]:= FYh[i,k]-FTh[i,h]*FCh[k,h];

                                 end;




                        for j:= 0 to pred(FCountX) do
                        begin

                                 sum:=0.0;


                                        for k:= 0 to h-1 do
                                        begin

                                        sum:=sum+FWhseconde[j,k]*FPh[j,k]*FWh[j,h];

                                        end;
                                     //if h=0 then  FWhseconde[j,h]:=FWh[j,h];
                                     //if h>0 then  FWhseconde[j,h]:=FWh[j,h]-FWhseconde[j,h-1]*FPh[j,h-1]*FWh[j,h];

                                    FWhseconde[j,h]:=FWh[j,h]-sum;
                                   //FWhseconde[j,h]:=FWh[j,h];
                                   //for k:=0 to h-1 do  FWhseconde[j,h]:=FWhseconde[j,h]-FWhseconde[j,k]*FPh[j,k]*FWh[j,h];


                        end;

                    //new -- 10/03/2006 -- optimisation -- pas besoin de calculer les indicateurs si classement
                    //puisque a ne sera jamais affich

                    //non, finalement on garde... 28/08/2006

                    //if FDetailedResult
                    if TRUE
                     then self.globalGoodness(h,FXhpred,FYhpred,FWhi,FChi,FXhinit,FYhinit);

                    //faut-il poursuivre ?
                    if (FNbAxisAutoDetect = true)
                     then
                      begin
                        RdY_is_Inf_Seuil:= (FRdYT[h] < FSeuilRdY);
                        //si c'est pas satisfaisant et qu'on est en dessous du max. admis, on ajoute
                        if ((RdY_is_Inf_Seuil = false) and (FNbAxis < FMaxNbAxis)) then inc(FNbAxis);
                        //si c'est satisfaisant, le dernier coup n'est plus utile, sauf si c'est le premier
                        if ((RdY_is_Inf_Seuil = true) and (FNbAxis > 1)) then dec(FNbAxis);
                      end;

                    //passer au suivant
                    inc(h);

                end;

          //******************************************************
          //dernier coup - calculer les coefficients de rgression
          //******************************************************

          //tait-ce une auto-dtection et a-t-on une valeur de redondance en dessous du seuil ?
          //if (FNbAxisAutoDetect = true) and (FNbAxis > 1) and (FNbAxis < FMaxNbAxis) then dec(FNbAxis);

          //
          h:= pred(FNbAxis);

          begin

                  for j:= 0 to pred(FCountX) do

                  begin

                         for k:=0 to pred(FCountY) do FCoefReg[j,k]:=FWhseconde[j,h]*FCh[k,h];

                  end;

         end;

         //********************************
         //***** fin de production des axes
         //********************************
         //TraceLog.WriteToLogFile(format('[PLS] production des (%d) axes = %d ms.',[FNbAxis,getTickCount()-tps]));


         //************************************************************
         //R.R. -- new -- 23/10/2005
         //calcul des coefficients de la rgression -- Tenenhaus, p.137
         //B est de dimension [FCountX x FCountY]
         //Nota : j'ai bien l'impression que le code ci-dessus y est ddi
         //mais une erreur doit s'y glisser, notamment pour le calcul de Wh*
         //************************************************************

         //--> attention aux dimensions de FCh et FWh
         //setLength(FCh,FCountY,FNbAxis);
         //setLength(FWh,FCountX,FNbAxis);
         //setLength(FPh,FCountX,FNbAxis);

         //calculer Wh* avec la formule par rcurrence -- p.114
         //tps:= getTickCount();
         //setLength(WhEtoile,FCountX,FNbAxis);
         for j:= 0 to pred(FCountX) do FWhEtoile[j,0]:= FWh[j,0];
         for k:= 1 to pred(FNbAxis) do
          begin
           somme:= 0.0;
           for j:= 0 to pred(FCountX) do
            somme:= somme + FPh[j,k-1]*FWh[j,k];
           for j:= 0 to pred(FCountX) do
            FWhEtoile[j,k]:= FWh[j,k] - FWhEtoile[j,pred(k)] * somme;
          end;

         for j:= 0 to pred(FCountX) do
          for i:= 0 to pred(FCountY) do
           begin
            sum:= 0.0;
            for k:= 0 to pred(FNbAxis) do
             begin
              //appliquer
              sum:= sum + FWhEtoile[j,k] * FCh[i,k];
             end;
            FBCoefReg[j,i]:= sum;
            //new -- R.R. -- conserver ces coefs. dans un autre tableau  toutes fins utiles (ex. calculs des intervalles de confiance, etc.)
            FBStdCoefReg[j,i]:= sum;
           end;

         //new -- R.R. -- 10/03/2006 -- corriger pour avoir les coefficients sur les variables originelles
         //cela permettra aussi de rduire le temps d'excution sur la projection
         //pour chaque Yi
         for i:= 0 to pred(FCountY) do
          begin
           constante:= 0.0;
           for j:= 0 to pred(FCountX) do
            begin
             if (FUsedStdDevX[j] > 0.0)
              then
               begin
                FBCoefReg[j,i]:= FUsedStdDevY[i] * FBCoefReg[j,i] / FUsedStdDevX[j];
                constante := constante + (-1.0) * FBCoefReg[j,i] * FStatData.FAvgX[j];//FBCoefReg a t corrig  la ligne prcdente !
               end;
            end;
           constante:= constante + FStatData.FAvgY[i];
           FBCoefReg[FCountX,i]:= constante;
          end;

         //Finalize(WhEtoile);
         //TraceLog.WriteToLogFile(format('[PLS] calcul des coefs. de projection = %d ms.',[FNbAxis,getTickCount()-tps]));
         //************************************************************


        //vider la mmoire
        setLength(curWh,0);
        setLength(FXh,0,0);
        setLength(FXhpred,0,0);
        setLength(FYh,0,0);
        setLength(FYhpred,0,0);
        setLength(FWhi,0,0);
        setLength(FChi,0,0);
        setLength(sumWhi,0);
        setLength(prodUhi,0);
        setLength(prodThi,0);
        setLength(sumChi,0);

    EXCEPT
        on e: Exception do
            TraceLog.WriteToLogFile(format('[TCalcPLSForSpv.runAnalysis] error -> %s',[e.Message]));
    END;

end;

end.
