(**********************************************************************)
(* UCompPLSRegressionPlus.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Rgression PLS - Version unifie)
@author(Ricco)
@created(16/05/2008)

Version unifie de la Rgression PLS (PLS Regression et PLS Factorial). L'ide surtout
est de produire des sorties qui concordent avec les standards du domaine. La principale
rfrence est SIMCA-P. Dans le tutoriel ddi, nous ferons le parallle avec R, SAS et SPAD.

Pour la doc. des indicateurs, GARSON fait trs fort encore --> http://www2.chass.ncsu.edu/garson/PA765/pls.htm 
}
unit UCompPLSRegressionPlus;

interface

USES
    Classes, Forms,
    IniFiles,
    UCompDefinition,
    UOperatorDefinition,
    UCompManageDataset,
    JFGCompPLS;

TYPE
       //***********************************************************************
       //** gnrateur de composant, son rle est de crer une instance de composant qui est insr dans le diagramme de traitements
       //***********************************************************************
       TMLGenCompPLSRegPlus = class(TMLGenCompPLS)
       public
       //spcifier quel est la classe du composant  gnrer sur le diagramme --  surcharger obligatoirement chez les hritiers
       function    getClassMLComponent: TClassMLComponent; override;
       end;

       //***********************************************************************
       //** classe du composant graphique insr dans le diagramme
       //***********************************************************************
       TMLCompPLSRegPlus = class(TMLCompPLS)
       //spcifier l'oprateur de calcul --  surcharger obligatoirement
       function    getClassOperator: TClassOperator; override;
       end;

       //***********************************************************************
       //** classe oprateur de calcul : prend en charge le calcul et l'affichage des rsultats
       //***********************************************************************
       TOpPLSRegPlus = class(TOpPLS)
       private
       //projections : prdiction et rsidus
       procedure   getPredResidProjections();
       //X-scores
       procedure   getXScores();
       //Y-scores - les u~h pour tre prcis (pp.136) - on prfrera cette formule car elle s'exprime en fonction des Y
       procedure   getYScores();
       protected
       function    getClassParameter: TClassOperatorParameter; override;
       //lancer les calculs
       function    CoreExecute(): boolean; override;
       //trs compliqu car doit s'accorder avec les options maintenant
       procedure   SetProjections(); override;
       //fentre d'affichage avec CHART
       function    getNewForm(): TForm; override;
       public
       //attention, trs riche ici maintenant
       function    getHTMLResultsSummary(): string; override;
       end;

       //***********************************************************************
       //** classe paramtrage de l'oprateur de calcul : gre la bote de paramtrage et l'E/S des paramtres
       //** si la mthode de requiert pas de paramtres, il suffit de crer une classe vide hritire de "TOperatorParameter"
       //***********************************************************************
       TOpPrmPLSRegPlus = class(TOpPrmPLS)
       protected
       function    CreateDlgParameters(): TForm; override;
       procedure   SetDefaultParameters(); override;
       public
       //liste des options -- trs riche maintenant
       //sorties (Output)
       OutXScores, OutYScores, OutPredResid: boolean;
       //affichage (Report)
       RepR2, RepVIP, RepUnstdCoef, RepStdCoef, RepXLoadings, RepXWeigths, RepYWeigths: boolean;
       //description des paramtres
       function    getHTMLParameters(): string; override;
       //sauvegarde
       procedure   LoadFromStream(prmStream: TStream); override;
       procedure   SaveToStream(prmStream: TStream); override;
       procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
       procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
       end;

implementation

uses
    Sysutils, Math,
    UCalcPLSRegressionPlus, UDatasetDefinition, UDatasetImplementation,
  UDlgOpPrmPlsRegPlus, UConstConfiguration, UFrmOpViewPLSRegressionChart;

{ TMLGenCompPLSRegPlus }

function TMLGenCompPLSRegPlus.getClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompPLSRegPlus;
end;

{ TMLCompPLSRegPlus }

function TMLCompPLSRegPlus.getClassOperator: TClassOperator;
begin
 result:= TOpPLSRegPlus;
end;

{ TOpPLSRegPlus }

function TOpPLSRegPlus.CoreExecute: boolean;
begin
 //vider la memoire
 if assigned(FCalc)then FCalc.Free();
 // Execution
 TRY

 //vrifier que le nombre d'axes  produire n'est pas suprieure  la dimension)
 if ((self.PrmOp as TOpPrmPLS).NbAxisCreated>self.WorkData.LstAtts[asInput].Count)
 then FTrueNbAxis:= self.WorkData.LstAtts[asInput].Count
 else FTrueNbAxis:= (self.PrmOp as TOpPrmPLS).NbAxisCreated;

 //calculs proprement dit...avec la classe >>TCalcPLSPlus<<
 FCalc:= TCalcPLSPlus.create(self.WorkData.LstAtts[asInput],self.WorkData.LstAtts[asTarget],self.WorkData.Examples,(self.PrmOp as TOpPrmPLS).Standardize,FTrueNbAxis,(self.PrmOp as TOpPrmPLS).DetailedResult);
 FCalc.runAnalysis(self.WorkData.Examples);

 //raliser les projections
 self.SetProjections();

 //suite...
 result:= TRUE;

 EXCEPT
 result:= FALSE;
 END;
end;

function TOpPLSRegPlus.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmPLSRegPlus;
end;

function TOpPLSRegPlus.getHTMLResultsSummary: string;
var calcPlus:  TCalcPLSPlus;
    prm: TOpPrmPLSRegPlus;
    s: string;
begin
 //transtypons une fois pour toutes...
 calcPlus:= FCalc as TCalcPLSPlus;
 prm:= self.PrmOp as TOpPrmPLSRegPlus;
 //enquiller maintenant
 s:= '<H3>PLS Regression RESULTS</H3>';
 s:= s + calcPlus.getHTMLPropVarExplained();
 //mettre cet ordre dans la bote de paramtrage (cf. Garson)
 if prm.RepR2 then s:= s + calcPlus.getHTMLR2AndRedundancy(); 
 if prm.RepXLoadings then s:= s + calcPlus.getHTMLXLoadings();
 if prm.RepXWeigths then s:= s + calcPlus.getHTMLInputsWeights();
 if prm.RepYWeigths then s:= s + calcPlus.getHTMLTargetWeights();
 if prm.RepVIP then s:= s + calcPlus.getHTMLVip();
 if prm.RepUnstdCoef then s:= s + calcPlus.getHTMLUnstandardizedCoef();
 if prm.RepStdCoef then s:= s + calcPlus.getHTMLStandardizedCoef();
 //and then...
 result:= s;
end;

function TOpPLSRegPlus.getNewForm: TForm;
begin
 result:= TfrmOpViewPlsRegressionChart.CreateFromOperator(self);
end;

procedure TOpPLSRegPlus.getPredResidProjections;
var i,j: integer;
    att: TAttribute;
begin
 //prdiction
 for j:= 1 to self.WorkData.LstAtts[asTarget].Count do
  begin
   att:= TAttContinue.Create(Format('plsr_pred_%s_%d',[self.WorkData.LstAtts[asTarget].Attribute[j-1].name,(MLOwner as TMLComponent).Number]),WorkData.LstAtts[asAll].Size);
   GenAtts.Add(att);
  end;
 FCalc.setProjection(GenAtts);
 //rsidus
 for j:= 1 to self.WorkData.LstAtts[asTarget].Count do
  begin
   att:= TAttContinue.Create(Format('plsr_residual_%s_%d',[self.WorkData.LstAtts[asTarget].Attribute[j-1].name,(MLOwner as TMLComponent).Number]),WorkData.LstAtts[asAll].Size);
   GenAtts.Add(att);
   //calcul  la vole : y - y^
   for i:= 1 to WorkData.LstAtts[asAll].Size do att.cValue[i]:= self.WorkData.LstAtts[asTarget].Attribute[j-1].cValue[i] - GenAtts.Attribute[j-1].cValue[i];
  end;
end;

procedure TOpPLSRegPlus.getXScores;
var j: integer;
    att: TAttribute;
    tmpLst: TLstAttributes;
begin
 tmpLst:= TLstAttributes.Create(FALSE,WorkData.LstAtts[asAll].Size);
 for j:= 1 to MATH.Min(self.WorkData.LstAtts[asInput].Count,(self.PrmOp as TOpPrmPLSRegPlus).NbAxisCreated)  do
  begin
   att:= TAttContinue.Create(Format('plsr_t_%d_%d',[j,(MLOwner as TMLComponent).Number]),WorkData.LstAtts[asAll].Size);
   tmpLst.Add(att);
  end;
 FCalc.setProjectionAxis(tmpLst);
 //remplir la bonne liste
 for j:= 0 to pred(tmpLst.Count) do genAtts.Add(tmpLst.Attribute[j]);
 tmpLst.Free();//no souci, elle n'est pas propritaire
end;

procedure TOpPLSRegPlus.getYScores;
var j: integer;
    att: TAttribute;
    tmpLst: TLstAttributes;
begin
 tmpLst:= TLstAttributes.Create(FALSE,WorkData.LstAtts[asAll].Size);
 for j:= 1 to MATH.Min(self.WorkData.LstAtts[asInput].Count,(self.PrmOp as TOpPrmPLSRegPlus).NbAxisCreated)  do
  begin
   att:= TAttContinue.Create(Format('plsr_utilde_%d_%d',[j,(MLOwner as TMLComponent).Number]),WorkData.LstAtts[asAll].Size);
   tmpLst.Add(att);
  end;
 FCalc.setProjectionYScores(tmpLst);
 //remplir la bonne liste
 for j:= 0 to pred(tmpLst.Count) do genAtts.Add(tmpLst.Attribute[j]);
 tmpLst.Free();//no souci, elle n'est pas propritaire
end;

procedure TOpPLSRegPlus.SetProjections;
var prm: TOpPrmPLSRegPlus;
begin
 prm:= self.PrmOp as TOpPrmPLSRegPlus;
 //vider la liste courante
 genAtts.Clear();
 //demande-t-il les prdictions et les rsidus
 if prm.OutPredResid then self.getPredResidProjections();
 //et les X-scores
 if prm.OutXScores then self.getXScores();
 //et les Y-scores
 if prm.OutYScores then self.getYScores();
end;

{ TOpPrmPLSRegPlus }

function TOpPrmPLSRegPlus.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmPlsRegPlus.CreateFromOpPrm(self);
end;

function TOpPrmPLSRegPlus.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_BLUE+'<TH colspan=2>PLS parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Number of axis</TD><TD align=center width=50>%d</TD></TR>',[NbAxisCreated]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Standardize</TD><TD align=center>%d</TD></TR>',[ord(Standardize)]);
 s:= s+HTML_TABLE_COLOR_HEADER_BLUE+'<TH colspan=2>Output parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Fitted and Residual Values</TD><TD align=center>%d</TD></TR>',[ord(OutPredResid)]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>X - latent factor scores (T)</TD><TD align=center>%d</TD></TR>',[ord(outXScores)]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Y - latent factor scores (U~)</TD><TD align=center>%d</TD></TR>',[ord(OutYScores)]);
 s:= s+HTML_TABLE_COLOR_HEADER_BLUE+'<TH colspan=2>Report parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>R-squared</TD><TD align=center>%d</TD></TR>',[ord(RepR2)]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>V.I.P</TD><TD align=center>%d</TD></TR>',[ord(RepVIP)]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Unstd. Reg. Parameters</TD><TD align=center>%d</TD></TR>',[ord(RepUnstdCoef)]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Std. Reg. Parameters</TD><TD align=center>%d</TD></TR>',[ord(RepStdCoef)]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>X - Loadings</TD><TD align=center>%d</TD></TR>',[ord(RepXLoadings)]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>X - Weights</TD><TD align=center>%d</TD></TR>',[ord(RepXWeigths)]);
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Y - Weights</TD><TD align=center>%d</TD></TR>',[ord(RepYWeigths)]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpPrmPLSRegPlus.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 //
 outXScores:= prmIni.ReadBool(prmSection,'outXScores',outXScores);
 OutYScores:= prmIni.ReadBool(prmSection,'OutYScores',OutYScores);
 OutPredResid:= prmIni.ReadBool(prmSection,'OutPredResid',OutPredResid);
 //
 RepR2:= prmIni.ReadBool(prmSection,'RepR2',RepR2);
 RepVIP:= prmIni.ReadBool(prmSection,'RepVIP',RepVIP);
 RepUnstdCoef:= prmIni.ReadBool(prmSection,'RepUnstdCoef',RepUnstdCoef);
 RepStdCoef:= prmIni.ReadBool(prmSection,'RepStdCoef',RepStdCoef);
 RepXLoadings:= prmIni.ReadBool(prmSection,'RepXLoadings',RepXLoadings);
 RepXWeigths:= prmIni.ReadBool(prmSection,'RepXWeigths',RepXWeigths);
 RepYWeigths:= prmIni.ReadBool(prmSection,'RepYWeigths',RepYWeigths);
end;

procedure TOpPrmPLSRegPlus.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(outXScores,sizeof(outXScores));
 prmStream.ReadBuffer(OutYScores,sizeof(OutYScores));
 prmStream.ReadBuffer(OutPredResid,sizeof(OutPredResid));
 prmStream.ReadBuffer(RepR2,sizeof(RepR2));
 prmStream.ReadBuffer(RepVIP,sizeof(RepVIP));
 prmStream.ReadBuffer(RepUnstdCoef,sizeof(RepUnstdCoef));
 prmStream.ReadBuffer(RepStdCoef,sizeof(RepStdCoef));
 prmStream.ReadBuffer(RepXLoadings,sizeof(RepXLoadings));
 prmStream.ReadBuffer(RepXWeigths,sizeof(RepXWeigths));
 prmStream.ReadBuffer(RepYWeigths,sizeof(RepYWeigths));
end;

procedure TOpPrmPLSRegPlus.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 //
 prmIni.WriteBool(prmSection,'outXScores',outXScores);
 prmIni.WriteBool(prmSection,'OutYScores',OutYScores);
 prmIni.WriteBool(prmSection,'OutPredResid',OutPredResid);
 //
 prmIni.WriteBool(prmSection,'RepR2',RepR2);
 prmIni.WriteBool(prmSection,'RepVIP',RepVIP);
 prmIni.WriteBool(prmSection,'RepUnstdCoef',RepUnstdCoef);
 prmIni.WriteBool(prmSection,'RepStdCoef',RepStdCoef);
 prmIni.WriteBool(prmSection,'RepXLoadings',RepXLoadings);
 prmIni.WriteBool(prmSection,'RepXWeigths',RepXWeigths);
 prmIni.WriteBool(prmSection,'RepYWeigths',RepYWeigths);
end;

procedure TOpPrmPLSRegPlus.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(outXScores,sizeof(outXScores));
 prmStream.WriteBuffer(OutYScores,sizeof(OutYScores));
 prmStream.WriteBuffer(OutPredResid,sizeof(OutPredResid));
 prmStream.WriteBuffer(RepR2,sizeof(RepR2));
 prmStream.WriteBuffer(RepVIP,sizeof(RepVIP));
 prmStream.WriteBuffer(RepUnstdCoef,sizeof(RepUnstdCoef));
 prmStream.WriteBuffer(RepStdCoef,sizeof(RepStdCoef));
 prmStream.WriteBuffer(RepXLoadings,sizeof(RepXLoadings));
 prmStream.WriteBuffer(RepXWeigths,sizeof(RepXWeigths));
 prmStream.WriteBuffer(RepYWeigths,sizeof(RepYWeigths));
end;

procedure TOpPrmPLSRegPlus.SetDefaultParameters;
begin
 //non modifiables dans ce composant
 Standardize:= true;//non modifiable, donnes toujours centres et rduites
 DetailedResult:= false;//non, on va grer finement les affichages
 //nombre d'axes
 NbAxisCreated:= 5;
 //sorties
 OutXScores:= TRUE;
 OutYScores:= FALSE;
 OutPredResid:= FALSE;
 //affichage - par dfaut la variance explique sera systmatiquement affiche
 RepR2:= FALSE;
 RepVIP:= FALSE;
 RepUnstdCoef:= FALSE;
 RepStdCoef:= FALSE;
 RepXLoadings:= TRUE;
 RepXWeigths:= TRUE;
 RepYWeigths:= TRUE;;
end;

initialization
 RegisterClass(TMLGenCompPLSRegPlus);
end.
