(*****************************************************************)
(* UCompPLSSelection.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(*****************************************************************)

{
@abstract(PLS -- Slection par validation croise du nombre d'axes)
@author(Ricco)
@created(29/08/2006)

Utilisation de la validation croise pour slectionner le nombre adquat d'axes en rgression PLS.
S'appuie essentiellement sur le livre de Tenehaus "La rgression PLS"
cf. pages 83 (PLS1) et pages 138 (PLS2).

}

unit UCompPLSSelection;

interface

USES
    Forms,Classes,
    IniFiles,Contnrs,
    UCompDefinition,
    UOperatorDefinition,
    UCompManageDataset,
    UDatasetDefinition,
    UDatasetImplementation,
    UDatasetExamples,
    UCalcRndGenerator,
    JFGCompPLS,
    UCalcPLSSelectionStructure;

TYPE
    {gnrateur}
    TGenClusPLSSelection = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    {composant -- attention, l'anctre n'est pas un clustering}
    TMLCompClusPLSSelection = class(TMLCompLocalData)
    protected
    function    getClassOperator: TClassOperator; override;
    public
    //n'est connectable que sur un composant PLS rgression ou factorial
    class function isConnectableTo(destination: TMLComponent): boolean; override;
    end;

    //dcl. forward.
    TResultsCV_PLS = class;

    {oprateur}
    TOpClusPLSSelection = class(TOpLocalData)
    private
    //oprateur PLS prcdent
    FOpPLS: TOpPLS;
    //donnes de validataion croise
    FData: TLstDataCVPLS;
    //donnes de rsultats
    FResults: TResultsCV_PLS;
    //nombre d'axes slectionns
    FSelectedH: integer;
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    function    CoreExecute(): boolean; override;
    public
    //rcuprer au passage l'oprateur PLS
    constructor Create(AOwner: TObject); override;
    destructor  destroy(); override;
    function    getHTMLResultsSummary(): string; override;
    end;

    {paramtrage}
    TOpPrmClusPLSSelection = class(TOperatorParameter)
    private
    protected
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    {proprits de paramtrage}
    //nombre de folds (2  10 -- df. 5)
    NbFolds: integer;
    //Initialisation du gnrateur de nombre alatoire (standard ou random, df. standard)
    InitRnd: TStartSeed;
    //mode de stopping : 0 -> Q2, 1 -> PRESS
    Stopping: integer;
    //valeur de coupure pour le Q2
    Q2CutValue: double;
    //valeur de coupure pour le PRESS -- en %
    PRESSCutValue: integer;
    //mj du plugged PLS component
    UpdatePluggedPLS: boolean; 
    //mthodes
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    end;

    {structure de rsultat -- RSS(h-1), PRESS(h), Q2(h) et Q2cum(h}
    TInfoPrediction = record
    RSS: single;
    PRESS: single;
    Q2: single;
    Q2cum: single;
    PRESSReduction: single;
    end;

    {Tableau de rsultats pour la liste de Y}
    TArrayInfoPrediction = array of TInfoPrediction;

    {structure de rsultats}
    TResultsCV_PLS = class
    private
    //liste des variables
    FLstY: TLstAttributes;
    //tableau des rsultats par axes
    FAxisResult: array of TArrayInfoPrediction;
    //tableau des Q2h
    FQ2h,FQ2cum: array of single;
    //paramtrage pour la valeur de coupure
    FPrm: TOpPrmClusPLSSelection;
    public
    constructor create(lstY: TLstAttributes; prm: TOpPrmClusPLSSelection);
    destructor  destroy(); override;
    procedure   addResult(infos: TArrayInfoPrediction);
    function    getNbAxis(): integer;
    //voir si l'axe h est relavant selon les rgles d'arrt 1 et 2 -- cf. Tenenhaus pp. 138
    function    isRelevantAxis(h: integer): boolean;
    function    getNbYAttributes(): integer;
    function    getHTML(): string;
    end;
    


implementation

USES
    Sysutils,
    JFGCalcPLS, UCalcStatDes, UDlgOpPrmPlsSelection, UConstConfiguration,
    ULogFile;

{ TGenClusPLSSelection }

procedure TGenClusPLSSelection.GenCompInitializations;
begin
 FMLComp:= mlcPLS;
end;

function TGenClusPLSSelection.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompClusPLSSelection; 
end;

{ TMLCompClusPLSSelection }

function TMLCompClusPLSSelection.getClassOperator: TClassOperator;
begin
 result:= TOpClusPLSSelection;
end;

class function TMLCompClusPLSSelection.isConnectableTo(
  destination: TMLComponent): boolean;
begin
 result:= destination.InheritsFrom(TMLCompPLS);
end;

{ TOpClusPLSSelection }

function TOpClusPLSSelection.CoreExecute: boolean;
var nbFolds: integer;
    nbExamples: integer;
    h,i: integer;
    calc: TCalcPLS;
    info: TArrayInfoPrediction;
    inputs, targets: TLstAttributes;
    Err: TVectorPLS;
    f: integer;
    stat: TCalcStatDesContinuous;
    examples: TExamples;
    tabPRESS: array of single;
begin
  inputs:= self.WorkData.LstAtts[asInput];
  targets:= self.WorkData.LstAtts[asTarget];
  examples:= self.WorkData.Examples;
  setLength(Err,targets.Count);
  try
  nbExamples:= self.WorkData.Examples.Size;
  nbFolds:= (self.PrmOp as TOpPrmClusPLSSelection).NbFolds;
  if (nbFolds > nbExamples) then nbFolds:= nbExamples;
  //crer les donnes
  if assigned(FData) then FData.Free();
  FData:= TLstDataCVPLS.create(nbFolds,nbExamples);
  FData.partitionning(examples,(self.PrmOp as TOpPrmClusPLSSelection).InitRnd);
  //crer la structure de collecte des rsultats
  if assigned(FResults) then FResults.Free();
  FResults:= TResultsCV_PLS.create(targets, self.PrmOp as TOpPrmClusPLSSelection);
  //**********************************************
  //lancer l'exploration des diffrentes solutions
  //pour l'instant c'est vraiment pas optimal, on n'exploite pas les rsultats des axes prcdents
  //mais au moins on gre aussi parcimonieusement que possible la mmoire, c'est un moindre mal....
  //>>
  //prparation du tableau de rsultat
  setLength(info,targets.Count);
  setLength(tabPRESS,targets.Count);
  //zoo...
  h:= 0;
  REPEAT
   //comme cela, on est sr de tester au moins le 1er axe
   h:= h + 1;
   //RSS(h-1)
   if (h = 1)
    then
     begin
      //premire tape sans axes, la meilleure prdiction c'est la moyenne...
      for i:= 0 to pred(FResults.getNbYAttributes()) do
       begin
       stat:= TCalcStatDesContinuous.Create(targets.Attribute[i],FData.getData(0).getTrain());
       info[i].RSS:= stat.TSS;
       tabPRESS[i]:= info[i].RSS;
       stat.Free();
      end;
     end
    else
     begin
      calc:= TCalcPLS.create(inputs,targets,FData.getData(0).getTrain(),true,pred(h),false);
      calc.runAnalysis(FData.getData(0).getTrain());
      calc.getErrPred(FData.getData(0).getTrain(),inputs,err);
      calc.Free();
      for i:= 0 to pred(FResults.getNbYAttributes()) do info[i].RSS:= Err[i];
     end;
   //PRESS(h)
   for i:= 0 to pred(FResults.getNbYAttributes()) do info[i].PRESS:= 0.0;
   for f:= 1 to FData.getNbFolds() do
    begin
     calc:= TCalcPLS.create(inputs,targets,FData.getData(f).getTrain(),true,h,false);
     calc.runAnalysis(FData.getData(f).getTrain());
     calc.getErrPred(FData.getData(f).getTest(),inputs,err);
     calc.Free();
     for i:= 0 to pred(FResults.getNbYAttributes()) do
      info[i].PRESS:= info[i].PRESS + Err[i];
    end;
   //Q2h
   for i:= 0 to pred(FResults.getNbYAttributes()) do
    begin
     if (info[i].RSS > 0.0)
      then info[i].Q2:= 1.0 - info[i].PRESS / info[i].RSS
      else info[i].Q2:= 0.0;
     //le R2, pourquoi pas...?
     info[i].PRESSReduction:= 100.0 * (TabPRESS[i] - info[i].PRESS)/TabPRESS[i];
     TabPRESS[i]:= info[i].PRESS;
     //TraceLog.WriteToLogFile(format('[PLS-SELECTION] --> PRESS = %.4f, R2 = %.4f',[info[i].PRESS,info[i].R2Adj]));
    end;
   //ajouter les rsultats
   FResults.addResult(info);
  UNTIL (h >= self.WorkData.LstAtts[asInput].Count) OR not(FResults.isRelevantAxis(h));
  finalize(TabPRESS);
  //dterminer le bon nombre d'axes
  //le dernier tait pertinent ?
  if not(FResults.isRelevantAxis(h))
   then FSelectedH:= pred(h)
   else FSelectedH:= h;
  //
  finalize(Err);
  //alors, une mise  jour est demande ??? et le nombre d'axes est diffrent ?
  if (self.PrmOp as TOpPrmClusPLSSelection).UpdatePluggedPLS and (FSelectedH <> (FOpPls.PrmOp as TOpPrmPLS).NbAxisCreated)
   then
    begin
     (FOpPls.PrmOp as TOpPrmPLS).NbAxisCreated:= FSelectedH;
     (FOpPls.MLOwner as TMLComponent).Execute(true);
    end;
  //**********************************************
  //and then...
  result:= true;
  except
  result:= false;
  end;
end;

constructor TOpClusPLSSelection.Create(AOwner: TObject);
begin
 inherited;
 FOpPLS:= (MLOwner as TMLComponent).Predecessor.Operator as TOpPLS;
end;

destructor TOpClusPLSSelection.destroy;
begin
 if assigned(FData) then FData.Free();
 if assigned(FResults) then FResults.Free();
 inherited;
end;

function TOpClusPLSSelection.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmClusPLSSelection;
end;

function TOpClusPLSSelection.getHTMLResultsSummary: string;
var s: string;
begin
 s:= '<H3>Component selection results</H3>';
 s:= s + format('<H5>Number of components = %d</H5>',[FSelectedH]);
 s:= s + '<H5>Detailed results</H5>';
 //au cas o il y aurait plantage ou que sais-je encore...
 if assigned(FResults)
  then s:= s + FResults.getHTML();
 result:= s;
end;

{ TOpPrmClusPLSSelection }

function TOpPrmClusPLSSelection.CreateDlgParameters: TForm;
begin
 result:= TDlgOpPrmPLSSelection.CreateFromOpPrm(self);
end;

function TOpPrmClusPLSSelection.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Parameter</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD># folds</TD><TD>%d</TD></TR>',[NbFolds]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Rnd</TD><TD>%d</TD></TR>',[ord(InitRnd)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Stopping rule</TD><TD>%d</TD></TR>',[Stopping]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Q2 cut value</TD><TD>%.4f</TD></TR>',[Q2CutValue]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>PRESS Reduction cut (%s)</TD><TD>%d</TD></TR>',['%',PRESSCutValue]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Update plugged component</TD><TD>%d</TD></TR>',[ord(UpdatePluggedPLS)]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmClusPLSSelection.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 NbFolds:= prmINI.ReadInteger(prmSection,'nb_folds',NbFolds);
 InitRnd:= TStartSeed(prmINI.ReadInteger(prmSection,'rnd',ord(InitRnd)));
 stopping:= prmINI.ReadInteger(prmSection,'stopping',stopping);
 Q2CutValue:= prmINI.ReadFloat(prmSection,'q2_cut_value',Q2CutValue);
 PRESSCutValue:= prmINI.ReadInteger(prmSection,'press_cut_value',PRESSCutValue);
 UpdatePluggedPLS:= prmINI.ReadBool(prmSection,'update',UpdatePluggedPLS);
end;

procedure TOpPrmClusPLSSelection.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(NbFolds,sizeof(NbFolds));
 prmStream.ReadBuffer(InitRnd,sizeof(InitRnd));
 prmStream.ReadBuffer(Stopping,sizeof(Stopping));
 prmStream.ReadBuffer(Q2CutValue,sizeof(Q2CutValue));
 prmStream.ReadBuffer(PRESSCutValue,sizeof(PRESSCutValue));
 prmStream.ReadBuffer(UpdatePluggedPLS,sizeof(UpdatePluggedPLS));
end;

procedure TOpPrmClusPLSSelection.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_folds',NbFolds);
 prmINI.WriteInteger(prmSection,'rnd',ord(InitRnd));
 prmINI.WriteInteger(prmSection,'stopping',stopping);
 prmINI.WriteFloat(prmSection,'q2_cut_value',Q2CutValue);
 prmINI.WriteInteger(prmSection,'press_cut_value',PRESSCutValue);
 prmINI.WriteBool(prmSection,'update',UpdatePluggedPLS);
end;

procedure TOpPrmClusPLSSelection.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(NbFolds,sizeof(NbFolds));
 prmStream.WriteBuffer(InitRnd,sizeof(InitRnd));
 prmStream.WriteBuffer(Stopping,sizeof(Stopping));
 prmStream.WriteBuffer(Q2CutValue,sizeof(Q2CutValue));
 prmStream.WriteBuffer(PRESSCutValue,sizeof(PRESSCutValue));
 prmStream.WriteBuffer(UpdatePluggedPLS,sizeof(UpdatePluggedPLS));
end;

procedure TOpPrmClusPLSSelection.SetDefaultParameters;
begin
 NbFolds:= 5;
 InitRnd:= seedStandard;
 Stopping:= 1;
 Q2CutValue:= 0.05;
 PRESSCutValue:= 20;
 UpdatePluggedPLS:= TRUE;
end;

{ TResultsCV_PLS }

procedure TResultsCV_PLS.addResult(infos: TArrayInfoPrediction);
var oldAxis,i,j: integer;
    num,denom: single;
begin
 oldAxis:= length(FQ2h);
 //rsultats pour chaque Y
 setLength(FAxisResult,succ(oldAxis));
 setLength(FAxisResult[oldAxis],FLstY.Count);
 for i:= 0 to pred(FLstY.Count) do
  begin
   FAxisResult[oldAxis][i].RSS:= infos[i].RSS;
   FAxisResult[oldAxis][i].PRESS:= infos[i].PRESS;
   FAxisResult[oldAxis][i].Q2:= infos[i].Q2;
   FAxisResult[oldAxis][i].PRESSReduction:= infos[i].PRESSReduction;
   //FAxisResult[oldAxis][i].Q2cum:= infos[i].Q2cum;
   //doit se recalculer
   FAxisResult[oldAxis][i].Q2cum:= 1.0;
   for j:= 0 to oldAxis do
    if (FAxisResult[j][i].RSS > 0.0)
     then FAxisResult[oldAxis][i].Q2cum:= FAxisResult[oldAxis][i].Q2cum * (FAxisResult[j][i].PRESS / FAxisResult[j][i].RSS);
   FAxisResult[oldAxis][i].Q2cum:= 1.0 - FAxisResult[oldAxis][i].Q2cum;
  end;
 //le rsultat "global" Q2h -- formule pp.138
 setlength(FQ2h,succ(oldAxis));
 num:= 0.0;
 denom:= 0.0;
 for i:= 0 to pred(FLstY.Count) do
  begin
   num:= num + infos[i].PRESS;
   denom:= denom + infos[i].RSS;
  end;
 if (denom > 0.0)
  then FQ2h[oldAxis]:= 1.0 - num / denom
  else FQ2h[oldAxis]:= 0.0;
 //
 setLength(FQ2cum,succ(oldAxis));
 FQ2cum[oldAxis]:= 1.0;
 for j:= 0 to oldAxis do
  FQ2cum[oldAxis]:= FQ2cum[oldAxis] * (1.0 - FQ2h[j]);
 FQ2cum[oldAxis]:= 1.0 - FQ2cum[oldAxis];
end;

constructor TResultsCV_PLS.create(lstY: TLstAttributes; prm: TOpPrmClusPLSSelection);
begin
 inherited create();
 FLstY:= lstY;
 FPrm:= prm;
end;

destructor TResultsCV_PLS.destroy;
begin
 finalize(FQ2h);
 finalize(FQ2cum);
 finalize(FAxisResult);
 inherited;
end;

function TResultsCV_PLS.getHTML: string;
var s: string;
    i,h: integer;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=3>-</TH>';
 for i:= 0 to pred(self.getNbYAttributes) do
  s:= s + format('<TH colspan=3>%s</TH>',[self.FLstY.Attribute[i].Name]);
 s:= s + '</TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>h</TH><TH>Q2</TH><TH>Q2cum</TH>';
 for i:= 0 to pred(self.getNbYAttributes) do
  s:= s + '<TH>Q2</TH><TH>PRESS</TH><TH>D(PRESS)</TH>';
 s:= s + '</TR>';
 h:= 0;
 while (h < self.getNbAxis()) do //and self.isRelevantAxis(succ(h)) do
  begin
   //couleur d'affichage de la ligne
   if self.isRelevantAxis(succ(h))
    then s:= s + HTML_TABLE_COLOR_DATA_GREEN
    else s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   //la ligne
   s:= s + format('<TD>%d</TD><TD align="right">%.3f</TD><TD align="right">%.3f</TD>',[succ(h),FQ2h[h],FQ2cum[h]]);
   for i:= 0 to pred(self.getNbYAttributes) do
    s:= s + format('<TD align="right">%.3f</TD><TD align="right">%.3f</TD><TD align="right">%.1f %s</TD>',
                   [FAxisResult[h][i].Q2,FAxisResult[h][i].PRESS,FAxisResult[h][i].PRESSReduction,'%']);
   s:= s + '</TR>';
   inc(h);
  end;
 s:= s + '</table>';
 //
 result:= s;
end;

function TResultsCV_PLS.getNbAxis: integer;
begin
 result:= length(FQ2h);
end;

function TResultsCV_PLS.getNbYAttributes: integer;
begin
 result:= self.FLstY.Count;
end;

function TResultsCV_PLS.isRelevantAxis(h: integer): boolean;
var okRelevant: boolean;
    i: integer;
begin
 //le tableau commence  zro
 dec(h);
 case FPrm.Stopping of
 //rgle du Q2
 0: begin
     //rgle 1 -- pp.138
     okRelevant:= (FQ2h[h] > FPrm.Q2CutValue);
     //rgle 2 -- pp. 138
     for i:= 0 to pred(FLstY.Count) do
      okRelevant:= okRelevant OR (FAxisResult[h][i].Q2 > FPrm.Q2CutValue);
    end;
  //rgle du PRESS
 else
    begin
     okRelevant:= false;
     for i:= 0 to pred(FLstY.Count) do
      okRelevant:= okRelevant OR (FAxisResult[h][i].PRESSReduction > FPrm.PRESSCutValue);
    end;
 end;
 //et donc...
 result:= okRelevant;
end;


initialization
 RegisterClass(TGenClusPLSSelection);
end.
