(************************************************************************)
(* UCompRegOutlierDetection.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(************************************************************************)

{
@abstract(Regression multiple. Dtection des points aberrants et influents)
@author(Ricco)
@created(05/05/2007)

Reprend les principales formules implmentes dans REGRESS32. Rajoute quand mme le rsidu
standardis pour tre en adquation avec le cours le L3-IDS.

Attention, ce composant ne calcule les indicateurs que sur les donnes ayant servi  la
construction du modle. Ca n'a pas trop de les calculer sur les autres observations. Ainsi,
aucune nouvelle variable n'est construite et exporte par le composant. Cela explique aussi sa filiation
dans la hirarchie des classes.

New -- 07/05/2007 -- Enlever pour l'instant Mahalanobis et Wilks
New -- 07/05/2007 -- Mettre une borne standard  2 pour RSTUDENT
}

unit UCompRegOutlierDetection;

interface

USES
    Forms,Classes,IniFiles, Math,
    UCompDefinition,
    UOperatorDefinition,
    UCompRegDefinition,
    UDatasetDefinition,
    UCalcRegLinMultiple,
    UCompRegMultiple,
    UCompManageDataset;

TYPE
    //gnrateur
    TMLGenRegOutlier = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLCompRegOutlier = class(TCompRegMulAfter)
    protected
    function    getClassOperator: TClassOperator; override;
    end;

    //TEnumOutlierDetection = (odrLeverage, odrDFFITS, odrStandardizedRes, odrStudentizedRes, odrCovratio,
    //                         odrWilks, odrMahalanobis, odrCook);
    TEnumOutlierDetection = (odrLeverage, odrStandardizedRes, odrStudentizedRes,
                             odrDFFITS, odrCook, odrCovratio);

CONST
    STR_OUTLIER : array[TEnumOutlierDetection] of string = ('Leverage','RStandard','RStudent',
                                                            'DFFITS','Cook''s D','|COVRATIO|');

    MIN_VALUE_OUTLIER : single = -MATH.MaxSingle;
    MAX_VALUE_OUTLIER : single = +MATH.MaxSingle;

TYPE

    //tableau de valeurs
    TTabOutlierValue = array[TEnumOutlierDetection] of single;

    //tableau de dcompte
    TTabOutlierCount = array[TEnumOutlierDetection] of integer;

    //oprateur
    TOpRegOutlier = class(TOpLocalData)
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    //prparer les bornes
    procedure   setBornes();
    //calculer les valeurs
    procedure   computeOutliers();
    //tester si une valeur dborde des bornes et incrmente le tableau de comptage le cas chant
    procedure   testValue(odr: TEnumOutlierDetection; value: single);
    //outil spcifique de visualisation
    function    getNewForm(): TForm; override;
    public
    //oprateur de calcul de la rgression linaire multiple
    opCalc: TOpRegMultiple;
    //tableau des bornes basses
    TabBBasse: TTabOutlierValue;
    //tableau des bornes hautes
    TabBHaute: TTabOutlierValue;
    //tableau des donnes
    TabValues: array of TTabOutlierValue;
    //nombre d'lments dtects
    TabDetected: TTabOutlierCount;
    {dtruire les tableaux}
    destructor  Destroy; override;
    {excution effective}
    function    CoreExecute(): boolean; override;
    {rsum des rsultats}
    function    getHTMLResultsSummary(): string; override;
    //tester si une valeur dborde des bornes
    function    isOutOfBounds(odr: TEnumOutlierDetection; value: single): boolean;
    //description sous forme de chane de caractre d'une borne
    function    descBorne(value: single): string;    
    end;

    //paramtrage
    TPrmOpRegOutlier = class(TOperatorParameter)
    protected
    procedure   SetDefaultParameters(); override;
    function    CreateDlgParameters(): TForm; override;
    public
    //sig. level pour certaines mthodes de dtection
    sigLevel: double;
    function    getHTMLParameters(): string; override;
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    end;

implementation

uses FMath, LOISTAT, UConstConfiguration, SysUtils, MATRIXMM,
  UFrmOpViewRegOutlier, UDlgOpPrmRegOutlier;


{ TMLGenRegOutlier }

procedure TMLGenRegOutlier.GenCompInitializations;
begin
 FMLComp:= mlcRegression;
end;

function TMLGenRegOutlier.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompRegOutlier; 
end;

{ TMLCompRegOutlier }

function TMLCompRegOutlier.getClassOperator: TClassOperator;
begin
 result:= TOpRegOutlier; 
end;

{ TPrmOpRegOutlier }

function TPrmOpRegOutlier.CreateDlgParameters: TForm;
begin
 //result:= TDlgOpPrmRegOutlier.CreateFromOpPrm(self);
 //new -- 07/05/2007 -- pas de paramtrage "visible donc"
 result:= NIL;
end;

procedure TPrmOpRegOutlier.SetDefaultParameters;
begin
 sigLevel:= 0.05;
end;


procedure TPrmOpRegOutlier.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 sigLevel:= prmINI.ReadFloat(prmSection,'sigLevel',sigLevel);
end;

procedure TPrmOpRegOutlier.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(sigLevel,sizeof(sigLevel));
end;

procedure TPrmOpRegOutlier.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteFloat(prmSection,'sigLevel',sigLevel);
end;

procedure TPrmOpRegOutlier.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(sigLevel,sizeof(sigLevel));
end;


function TPrmOpRegOutlier.getHTMLParameters: string;
var s: string;
begin
 s:= '<P><B>Parameters</B><BR>'+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_DATA_BLUE+'<TD width=120>Sig. Level</TD>'+format('<TD>%.4f</TD></TR>',[sigLevel]);
 s:= s+'</table>';
 result:= s;
end;

{ TOpRegOutlier }

procedure TOpRegOutlier.computeOutliers;
var n: integer;
    odr: TEnumOutlierDetection; 
    //reprise de REGRESS
    H: PManipMatMM;
    TMP2,XT,TMP1: PManipMatMM;
    i: Integer;
    si,dffitsi,standardisei,studentisei,covratioi,cooki: single;
    VEps,VH,P: single;
begin
 n:= self.WorkData.Examples.Size;
 p:= opCalc.LstExogenous.Count;
 //s'il y a une constante, a devient p+1
 if ((opCalc.PrmOp as TOpPrmRegMultiple).Constant = true) then p:= p + 1; 
 //rz du tableau de comptage
 FillChar(TabDetected[odrLeverage],sizeof(TabDetected),0);
 //dimensionner le tableau des rsultats
 setlength(TabValues,n);
 {Matrice H}
 MultiplicationMM(opCalc.CalcReg.X,opCalc.CalcReg.XPXM1,TMP1);
 TranspositionMM(opCalc.CalcReg.X,XT);
 MultiplicationMM(TMP1,XT,TMP2);
 XT.Free;
 TMP1.Free;
 H:= PManipMatMM.Create(n,1);
 For i:= 1 To n Do
  H.Place(i,1,TMP2.Prendre(i,i));
 TMP2.Free;
 for i:= 1 to n do TabValues[i-1,odrLeverage]:= H[i,1];//recopie trs simple...
 {Les indicateurs}
 For i:= 1 To H.MatMM.Lig Do
  Begin
   VEps:= opCalc.CalcReg.E.MatMM.PtrData^[i]^[1];
   VH  := H.MatMM.PtrData^[i]^[1];
   {sigma_i}
   si:= SQRT((1.0/(N-P-1))*((N-P)*opCalc.CalcReg.SigmaC2-VEps*VEps/(1.0-VH)));
   {dffits_i}
   dffitsi:= SQRT(VH/(1-VH))*(VEps/(si*SQRT(1-VH)));
   TabValues[i-1,odrDFFITS]:= dffitsi;
   {standardiss}
   standardisei:= VEps/(SQRT(opCalc.CalcReg.SigmaC2)*SQRT(1.0-VH));
   TabValues[i-1,odrStandardizedRes]:= standardisei;
   {studentiss}
   studentisei:= VEps/(si*SQRT(1.0-VH));
   TabValues[i-1,odrStudentizedRes]:= studentisei;
   {Covratioi}
   covratioi:= 1.0/(POWER((N-P-1)/(N-P)+(studentisei*studentisei)/(N-P),P)*(1.0-VH));
   TabValues[i-1,odrCovratio]:= ABS(covratioi);
   {Wilks}
   //wilksi:= (N/(N-1.0))*(1.0-VH);
   //TabValues[i-1,odrWilks]:= wilksi;
   {Mahalanobis}
   //mahalanobisi:= (N*(N-2.0)/(N-1.0))*((1-wilksi)/wilksi);
   //TabValues[i-1,odrMahalanobis]:= mahalanobisi;
   {Cook}
   cooki:= (standardisei*standardisei)*(VH/(1.0-VH))*(1.0/P);
   TabValues[i-1,odrCook]:= cooki;
  End;
 H.Free;
 //comtabiliser maintenant les valeurs "aberrantes"
 for i:= 1 to n do
  for odr:= low(TEnumOutlierDetection) to high(TEnumOutlierDetection) do
   self.testValue(odr,TabValues[i-1,odr]);
end;

function TOpRegOutlier.CoreExecute: boolean;
begin
 //se brancher sur l'oprateur -- ouahhh, faut le chercher celui l
 opCalc:= ((self.MLOwner as TMLCompRegOutlier).Predecessor as TCompRegMultiple).Operator as TOpRegMultiple;
 TRY
 //initialiser les bornes
 self.setBornes();
 //construire la matrice des points
 self.computeOutliers();
 //tout est bon  ce stade...
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

function TOpRegOutlier.descBorne(value: single): string;
begin
 if (value = MIN_VALUE_OUTLIER) or (value = MAX_VALUE_OUTLIER)
  then result:= '-'
  else result:= format('%.4f',[value]);
end;

destructor TOpRegOutlier.Destroy;
begin
 finalize(TabValues);
 inherited;
end;

function TOpRegOutlier.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpRegOutlier;
end;

function TOpRegOutlier.getHTMLResultsSummary: string;
var s,sBB,sBH: string;
    odr: TEnumOutlierDetection;
begin
 s:= '<H3>Outliers and influantial points detection for regression</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Statistic</TH><TH>Lower bound</TH><TH>Upper bound</TH><TH># detected</TH></TR>';
 for odr:= low(TEnumOutlierDetection) to high(TEnumOutlierDetection) do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[STR_OUTLIER[odr]]);
   sBB:= format('<TD align="center">%s</TD>',[self.descBorne(TabBBasse[odr])]);
   sBH:= format('<TD align="center">%s</TD>',[self.descBorne(TabBHaute[odr])]);
   s:= s + format('%s%s<TD align="center">%d</TD>',[sBB,sBH,TabDetected[odr]]) + '</TR>';
  end;
 s:= s + '</table>';

 result:= s;
end;

function TOpRegOutlier.getNewForm: TForm;
begin
 result:= TfrmOpViewRegOutlier.CreateFromOperator(self);
end;

function TOpRegOutlier.isOutOfBounds(odr: TEnumOutlierDetection;
  value: single): boolean;
begin
 result:= (value < TabBBasse[odr]) or (value > TabBHaute[odr]);
end;

procedure TOpRegOutlier.setBornes;
var n,p: single;
    //sigLevel: double;
begin
 n:= self.WorkData.Examples.Size;
 p:= opCalc.LstExogenous.Count;
 //s'il y a une constante, a devient p+1 i.e. le nombre de paramtres estims
 if ((opCalc.PrmOp as TOpPrmRegMultiple).Constant = true) then p:= p + 1;
 //niveau de signification
 //sigLevel:= (self.PrmOp as TPrmOpRegOutlier).sigLevel;
 //leverage - Hi
 TabBBasse[odrLeverage]:= MIN_VALUE_OUTLIER;
 TabBHaute[odrLeverage]:= 2.0 * p / n;
 //DFFITS
 TabBBasse[odrDFFITS]:= -2.0 * SQRT(p / n);
 TabBHaute[odrDFFITS]:= 2.0 * SQRT(p / n);
 //Rsidu standardis
 TabBBasse[odrStandardizedRes]:= MIN_VALUE_OUTLIER;
 TabBHaute[odrStandardizedRes]:= MAX_VALUE_OUTLIER;
 //Rsidu studentis
 //TabBBasse[odrStudentizedRes]:= ILS_STUDENT(sigLevel / 2.0, n - p - 1);
 //>> new -- 07/05/2007 -- mettre une borne "standard"  2
 TabBBasse[odrStudentizedRes]:= -2.0;
 TabBHaute[odrStudentizedRes]:= -1.0 * TabBBasse[odrStudentizedRes];
 //CovRatio
 TabBBasse[odrCovratio]:= 1.0 - 3.0 * p / n;
 TabBHaute[odrCovratio]:= 1.0 + 3.0 * p / n;
 //Wilks
 //TabBBasse[odrWilks]:= (n / (n - 1.0)) * (1.0 - 2.0 * p / n);
 //TabBHaute[odrWilks]:= MAX_VALUE_OUTLIER;
 //Mahalanobis
 //TabBBasse[odrMahalanobis]:= MIN_VALUE_OUTLIER;
 //TabBHaute[odrMahalanobis]:= (n * (n - 2.0)) / (n - 1.0) * ((2.0 * p - 1.0)/(n - 2.0 * p));
 //Cook
 TabBBasse[odrCook]:= MIN_VALUE_OUTLIER;
 //TabBHaute[odrCook]:= ILS_FISCHER(1.0 - sigLevel, p , n - p);
 TabBHaute[odrCook]:= 4.0 / (n - p);//!\ cf. http://www.ats.ucla.edu/stat/stata/webbooks/reg/chapter2/statareg2.htm et transparents Gueguen...
end;

(*
  >>> FROM REGRESS32...                           
   With StgValCoupure Do
    Begin
     VCoupure[1]:= 2*P/N;
     Cells[1,0]:= '>'+FormatFloat('0.0000',VCoupure[1]);
     VCoupure[2]:= 2*SQRT(P/N);
     Cells[2,0]:= '>'+FormatFloat('0.0000',VCoupure[2]);
     //VCoupure[3]:= 2;
     //new -- 04/08/2005 --  la place d'une valeur seuil fixe, on utilise la loi de Student -- cf. Tenenhaus, p. 95
     //test bilatral pour un risque  5%
     VCoupure[3]:= ILS_STUDENT(0.975,N - Prm.NbExo - 2);
     Cells[3,0]:= '>'+FormatFloat('0.0000',VCoupure[3]);
     VCoupure[4]:= 1-3*P/N;
     HauteCovR:= 1+3*P/N;
     Cells[4,0]:= '<>['+FormatFloat('0.00',VCoupure[4])+','
                  +FormatFloat('0.00',HauteCovR)+']';
     VCoupure[5]:= (N/(N-1.0))*(1.0-2.0*P/N);
     Cells[5,0]:= '<'+FormatFloat('0.0000',VCoupure[5]);
     VCoupure[6]:= (N*(N-2.0))/(N-1)*((2*P-1)/(N-2*P));
     Cells[6,0]:= '>'+FormatFloat('0.0000',VCoupure[6]);
     VCoupure[7]:= ILS_FISCHER(0.95,P,N-P);
     Cells[7,0]:= '>'+FormatFloat('0.0000',VCoupure[7]);
    End;
*)

procedure TOpRegOutlier.testValue(odr: TEnumOutlierDetection;
  value: single);
begin
 if self.isOutOfBounds(odr,value)
  then TabDetected[odr]:= TabDetected[odr] + 1;
end;

initialization
 RegisterClass(TMLGenRegOutlier);
end.
