(************************************************************************)
(* UCompRegOutlierDetection.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(************************************************************************)

{
@abstract(Regression multiple. Dtection des points aberrants et influents  l'aide des DFBETAS)
@author(Ricco)
@created(08/05/2007)

Reprend les ides de REGRESS. La principale astuce ici est qu'il n'est pas ncessaire de
recalculer explicitement les rgressions jackknife. Le temps de calcul est en fortement amlior.

On suspecte un point "influent" i sur le paramtre j lorsque : ABS(DFBETAS)_j,(i) > 2.0/racine(n)

Attention, ce composant ne calcule les indicateurs que sur les donnes ayant servi  la
construction du modle. Ca n'a pas trop de les calculer sur les autres observations. Ainsi,
aucune nouvelle variable n'est construite et exporte par le composant. Cela explique aussi sa filiation
dans la hirarchie des classes.
}
unit UCompRegOutlierDfBetas;

interface

USES
    Forms,Classes,IniFiles, Math,
    UCompDefinition,
    UOperatorDefinition,
    UCompRegDefinition,
    UDatasetDefinition,
    UCalcRegLinMultiple,
    UCompRegMultiple,
    UCompManageDataset;

TYPE
    //gnrateur
    TMLGenRegDFBetas = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLCompRegDFBetas = class(TCompRegMulAfter)
    protected
    function    getClassOperator: TClassOperator; override;
    end;

    //tableau des valeurs pour chaque variable -- dim = nombre de variables (+ 1 si rgression avec constante, en dernire position...)
    TTabDfBetas = array of single;

    //tableau des valeurs d'individus -- i = 1..n
    TValDfBetas = array of TTabDfBetas;

    //tableau de comptage
    TTabCountDfBetas = array of integer;

    //oprateur
    TOpRegDFBetas = class(TOpLocalData)
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    //initaliser les tables
    procedure   initTables();
    //dtruire les tables
    procedure   destroyTables();
    //calculer les points atypiques
    procedure   computeDfBetas();
    //outil spcifique de visualisation
    function    getNewForm(): TForm; override;    
    public
    //valeur de coupure --> 2.0/racine(n)
    CutValue: double;
    //tableau des valeurs
    TabValues: TValDfBetas;
    //tableau de comptage
    TabCount : TTabCountDfBetas;
    //oprateur de calcul de la rgression linaire multiple
    opCalc: TOpRegMultiple;
    {dtruire les tableaux}
    destructor  Destroy; override;
    {excution effective}
    function    CoreExecute(): boolean; override;
    {rsum des rsultats}
    function    getHTMLResultsSummary(): string; override;
    end;

    //paramtrage
    TPrmOpRegDFBetas = class(TOperatorParameter)
    protected
    procedure   SetDefaultParameters(); override;
    function    CreateDlgParameters(): TForm; override;
    end;


implementation

uses FMath, UConstConfiguration, SysUtils, MATRIXMM, UFrmOpViewRegDfBetas;


{ TMLGenRegDFBetas }

procedure TMLGenRegDFBetas.GenCompInitializations;
begin
 FMLComp:= mlcRegression;
end;

function TMLGenRegDFBetas.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompRegDFBetas; 
end;

{ TMLCompRegDFBetas }

function TMLCompRegDFBetas.getClassOperator: TClassOperator;
begin
 result:= TOpRegDFBetas;
end;

{ TPrmOpRegDFBetas }

function TPrmOpRegDFBetas.CreateDlgParameters: TForm;
begin
 result:= nil;
end;

procedure TPrmOpRegDFBetas.SetDefaultParameters;
begin
 //rien...
end;

{ TOpRegDFBetas }

procedure TOpRegDFBetas.computeDfBetas;
var n: integer;
    //reprise de REGRESS
    H: PManipMatMM;
    TMP2,XT,TMP1: PManipMatMM;
    i,j: Integer;
    si,studentisei: single;
    VEps,VH: single;
    p: integer;
    num,denom: single;
begin
 n:= self.WorkData.Examples.Size;
 p:= opCalc.LstExogenous.Count;
 //s'il y a une constante, a devient p+1
 if ((opCalc.PrmOp as TOpPrmRegMultiple).Constant = true) then p:= p + 1;
 //calculs matriciels
 MultiplicationMM(opCalc.CalcReg.X,opCalc.CalcReg.XPXM1,TMP1);
 TranspositionMM(opCalc.CalcReg.X,XT);
 MultiplicationMM(TMP1,XT,TMP2);
 TMP1.Free;
 //calcul de la nouvelle matrice au numrateur --> cf. Jensen & Ramirez, "Bringing Order to Outlier Diagnostics in Regression Models"
 MultiplicationMM(opCalc.CalcReg.XPXM1,XT,TMP1);
 XT.Free();
 //suite calcul du h_i
 H:= PManipMatMM.Create(n,1);
 For i:= 1 To n Do
  H.Place(i,1,TMP2.Prendre(i,i));
 TMP2.Free;
 {Les indicateurs}
 For i:= 1 To H.MatMM.Lig Do
  Begin
   VEps:= opCalc.CalcReg.E.MatMM.PtrData^[i]^[1];
   VH  := H.MatMM.PtrData^[i]^[1];
   {sigma_i}
   si:= SQRT((1.0/(N-P-1))*((N-P)*opCalc.CalcReg.SigmaC2-VEps*VEps/(1.0-VH)));
   {studentiss}
   studentisei:= VEps/(si*SQRT(1.0-VH));
   for j:= 0 to p-1 do
    begin
     num:= TMP1[j+1,i];
     denom:= opCalc.CalcReg.XPXM1[j+1,j+1] * (1.0  - H[i,1]);
     if (denom > 0.0)
      then TabValues[i-1][j]:= studentisei * num / SQRT(denom)
      else TabValues[i-1][j]:= 0.0;
     //comptage  la vole
     if (ABS(TabValues[i-1][j]) > CutValue) then inc(TabCount[j]);
    end;
  end;
 H.Free();
 TMP1.Free();
end;

function TOpRegDFBetas.CoreExecute: boolean;
begin
 //se brancher sur l'oprateur -- ouahhh, faut le chercher celui l
 opCalc:= ((self.MLOwner as TMLCompRegDFBetas).Predecessor as TCompRegMultiple).Operator as TOpRegMultiple;
 TRY
 CutValue:= 2.0 / SQRT(1.0 * opCalc.WorkData.Examples.Size);//valeur de coupure utilis tout le long
 self.initTables();//crer les tableaux
 self.computeDfBetas();//calcul proprement dit
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

destructor TOpRegDFBetas.Destroy;
begin
 self.destroyTables();
 inherited;
end;

procedure TOpRegDFBetas.destroyTables;
var i,n: integer;
begin
 finalize(TabCount);
 n:= self.opCalc.WorkData.Examples.Size;
 for i:= 0 to pred(n) do finalize(TabValues[i]);
 finalize(TabValues);
end;

function TOpRegDFBetas.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpRegDFBetas;
end;

function TOpRegDFBetas.getHTMLResultsSummary: string;
var s: string;
    j,p: integer;
begin
 p:= length(TabCount);
 s:= '<H3>DFBETAS for Regression Diagnostics</H3>';
 s:= s + format('<H5>Cut value = %.4f</H5>',[CutValue]);
 s:= s + '<H5># detected according to coefficients</H5>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Coefficient</TH><TH># detected</TH></TR>';
 if ((opCalc.PrmOp as TOpPrmRegMultiple).Constant = true)
  then
   begin
    s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD align="left">Intercept</TD>' + format('<TD align="right">%d</TD>',[TabCount[p-1]]) + '</TR>';
    p:= p - 1;
   end;
 for j:= 0 to pred(p) do
  s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align="left">%s</TD><TD align="right">%d</TD>',[opCalc.LstExogenous.Attribute[j].Name,TabCount[j]]) + '</TR>';
 s:= s + '</TABLE>';
 result:= s;
end;

function TOpRegDFBetas.getNewForm: TForm;
begin
 result:= TfrmOpViewRegDfBetas.CreateFromOperator(self);
end;

procedure TOpRegDFBetas.initTables;
var p: integer;
    i,n: integer;
begin
 p:= self.opCalc.LstExogenous.Count;
 n:= self.opCalc.WorkData.Examples.Size;
 if ((opCalc.PrmOp as TOpPrmRegMultiple).Constant = true) then p:= p + 1;
 //tableau de comptage
 setLength(TabCount,p);
 fillchar(TabCount[0],p*sizeof(integer),0);
 //tableau des valeurs
 setLength(TabValues,n);
 for i:= 0 to pred(n) do setLength(TabValues[i],p);
end;

initialization
 RegisterClass(TMLGenRegDfBetas);
end.
