(***********************************************************)
(* UCalcRegSvm.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(***********************************************************)

{
@abstract(Classe de calcul - Mthodes de rgression bases sur les SVM)
@author(Ricco)
@created(07/04/2009)

Utilisation de la bibliothque LIBSVM (http://www.csie.ntu.edu.tw/~cjlin/libsvm/)
importe sous forme de DLL.

}

unit UCalcRegSvm;

interface

USES
    UDatasetDefinition,
    UDatasetExamples,
    UOperatorDefinition,
    ULibSvmImport, UCalcSpvSMOAttTransformation;

TYPE
    //****************************************
    //classe de base - rgression avec les svm
    //****************************************
    TCalcRegSvm = class
    private
    //tableau interne pour prparer les prdictions futures, viter ainsi les allocations successives
    FTmpNodePrediction: array_svm_node;
    //mthode de transformation des variables -- ici, ce sera toujours [0,1] au dpart
    FAttTransSVM: TAttTrans;
    protected
    //variable  prdire
    FY: TAttribute;
    //variables prdictives
    FX: TLstAttributes;
    //tous les attributs
    FAll: TLstAttributes;
    //mettre les paramtres par dfaut
    procedure  setDefaultParameter();
    //rcuprer les paramtres de l'utilisateur
    procedure  getUserParameter();
    public
    //paramtrage de l'oprateur
    OpPrm: TOperatorParameter;
    //paramtres de calcul
    SvmParam: p_svm_parameter;
    //Problme  rsoudre
    SvmProblem: p_svm_problem;
    //modle
    SvmModel: p_svm_model;
    //tableau interne pour prparer les prdictions futures, viter ainsi les allocations successives
    TmpNodePrediction: array_svm_node;
    //constructeur (3 : epsilon-SVR ; 4 : nu-SVR)
    constructor create(prmY: TAttribute; prmX: TLstAttributes; prmOperator: TOperatorParameter);
    //destructeur
    destructor destroy(); override;
    //lancer l'excution
    function    execute(prmExamples: TExamples): boolean;
    //effectuer la prdiction et mesurer l'erreur
    procedure Predict(prmPred, prmErr: TAttribute);
    //les rsultats au format HTML
    function    GetHTMLResult(): string;
    end;

implementation

USES
    Sysutils, UCompRegSvm, UConstConfiguration, ULogFile;

{ TCalcRegSvm }

constructor TCalcRegSvm.create(prmY: TAttribute; prmX: TLstAttributes;
  prmOperator: TOperatorParameter);
var j: integer;
begin
 inherited create();
 FY:= prmY;
 FX:= prmX;
 OpPrm:= prmOperator;
 //gestion des paramtres
 self.setDefaultParameter();
 self.getUserParameter();
 //construire la liste complte des attributs
 FAll:= TLstAttributes.Create(FALSE,FX.Size);
 for j:= 0 to pred(FX.Count) do FAll.Add(FX.Attribute[j]);
 FAll.Add(FY);
end;

destructor TCalcRegSvm.destroy;
begin
 if assigned(SvmParam) then svm_destroy_param(SvmParam);
 //if assigned(SvmModel) then svm_destroy_model(SvmModel);
 //new - 05/05/2012 - modif libsvm 3.12
 if assigned(SvmModel) then svm_free_model_content(SvmModel);
 if assigned(SvmProblem) then dispose(SvmProblem);
 if assigned(FAttTransSVM) then FreeAndNil(FAttTransSVM);
 Finalize(FTmpNodePrediction);
 if assigned(FAll) then FreeAndNil(FAll);
 inherited destroy();
end;

function TCalcRegSvm.execute(prmExamples: TExamples): boolean;
var ok: boolean;
    i,j: integer;
    chaine: PChar;
begin
 //paramtres de normalisation -- [0,1] --> attention, l'endogne est insre en dernire position pour qu'il soit aussi normalis
 //FAttTransSVM:= TAttTransNormalize.create(FAll,prmExamples);
 //centrer et rduire les donnes
 FAttTransSVM:= TAttTransStandardize.create(FAll,prmExamples);
 //prparer les donnes en respectant le formalisme LIBSVM -- cf. le code SVMTrain (le code Java est autrement plus clair)
 ok:= true;
 TRY
 //prparer l'espace mmoire
 new(SvmProblem);
 setLength(SvmProblem^.y,prmExamples.Size);
 setLength(SvmProblem^.x,prmExamples.Size);
 //copier les donnes
 for i:= 1 to prmExamples.Size do
  begin
   //SvmProblem^.y[pred(i)]:= self.FY.cValue[prmExamples.Number[i]]; //!\ l'endogne transforme est en dernire position 
   //SvmProblem^.y[pred(i)]:= 2.0*FAttTransSVM.cValue[FX.Count,prmExamples.Number[i]]-1.0;//pour faire varier de -1  +1
   SvmProblem^.y[pred(i)]:= FAttTransSVM.cValue[FX.Count,prmExamples.Number[i]];
   //!\ ouah, le +1 est fondamental
   setLength(SvmProblem^.x[pred(i)],self.FX.Count+1);//+1 pour rserver le dernier au bornage des colonnes !!!
   for j:= 0 to pred(self.FX.Count) do
    begin
    SvmProblem^.x[pred(i)][j].index:= succ(j);
    //donnes transformes
    //SvmProblem^.x[pred(i)][j].value:= 2.0*FAttTransSVM.cValue[j,prmExamples.Number[i]]-1.0;//pour faire varier de -1  +1
    SvmProblem^.x[pred(i)][j].value:= FAttTransSVM.cValue[j,prmExamples.Number[i]];
    end;
    //!\\ warning --> parce que sinon, y sait pas  quelle colonne s'arrter, il n'y a pas d'indicateur de nombre de colonnes
    SvmProblem^.x[pred(i)][self.FX.Count].index:= -1;
  end;
 //qqs paramtres  rgler
 SvmProblem^.l:= prmExamples.Size;
 if (SvmParam^.gamma = 0) then SvmParam^.gamma:= 1.0/(1.0*self.FX.Count);
 EXCEPT
 TraceLog.WriteToLogFile('[SVR -- LIBSVM] //!\\ exception during data preparation');
 ok:= false;
 END;
 //tester les paramtres
 chaine:= '_test_';
 if ok
  then
   begin
   TRY
   chaine:= svm_check_parameter(SvmProblem,SvmParam);
   EXCEPT
   TraceLog.WriteToLogFile('[SVR -- LIBSVM] //!\\ exception during check parameters');
   ok:= false;
   END;
   end;
 if (chaine<>nil) and (strlen(chaine)>0)
  then
   begin
   TraceLog.WriteToLogFile(format('[C-SVC -- LIBSVM] //!\\ not passed check parameters == %s',[chaine]));
   ok:= false;
   end;
 //lancer les oprations en faisant appel  la procdure de la bibliothque
 if ok
  then
   begin
     TRY
     TraceLog.WriteToLogFile('[C-SVC -- LIBSVM] ok check parameters, run training');
     SvmModel:= svm_train(SvmProblem,SvmParam);
     EXCEPT
     on e: exception do
      begin
      TraceLog.WriteToLogFile(format('[C-SVC -- LIBSVM] //!\\ exception during model training == %s',[e.Message]));
      ok:= false;
      end;
     END;
   end;
 //prparer le tableau de noeud intermdiaire pour les prdictions
 if ok
  then
   begin
    //tableau pour les prdictions
    setlength(FTmpNodePrediction,self.FX.Count+1);
    for j:= 0 to pred(self.FX.Count) do
     FTmpNodePrediction[j].index:= succ(j);
    FTmpNodePrediction[self.FX.Count].Index:= -1;//don't forget my dear !!!
   end;
 result:= ok;
end;

function TCalcRegSvm.GetHTMLResult: string;
var s: string;
begin
 s:= '<H3>SVM characteristics</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Characteristic</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right>%d</TD></TR>',['# support vectors',SvmModel^.l]);
 s:= s + '</table>';
 //and then...
 result:= s;
end;

procedure TCalcRegSvm.getUserParameter();
var prm: TOpPrmRegressionSvm;
begin
 prm:= OpPrm as TOpPrmRegressionSvm;
 SvmParam^.kernel_type := prm._kernel_type; //>>type de noyau
 SvmParam^.degree := prm.degree; //>>degr du kernel
 SvmParam^.gamma := prm.gamma;	// 1/k >>gamma
 SvmParam^.coef0 := prm.coef0; //>>coef. 0 dans le kernel
 SvmParam^.C := prm.C;//>>penalit
 SvmParam^.eps := prm.eps;//>>tolerance
 SvmParam^.shrinking := prm.shrinking;//>>use shrinking heuristics
 SvmParam^.p:= prm.SvmTypeEpsilon;//epsilon-svr
 SvmParam^.nu:= prm.SvmTypeNu;//nu-svr
end;

procedure TCalcRegSvm.Predict(prmPred, prmErr: TAttribute);
var j,example: integer;
    prediction: double;
begin
 for example:= 1 to FY.Size do
  begin
   //remplir le tableau temporaire
   for j:= 0 to pred(self.FX.Count) do
    //FTmpNodePrediction[j].value:= 2.0*FAttTransSVM.cValue[j,example]-1.0;//pour faire varier de -1  +1
    FTmpNodePrediction[j].value:= FAttTransSVM.cValue[j,example];
   //rcuprer la prdiction
   prediction:= svm_predict(SvmModel,FTmpNodePrediction);
   //correction par d-normalisation
   //prediction:= (prediction + 1.0) * 0.5;
   prediction:= FAttTransSVM.inverseTransform(FX.Count,prediction);
   //affectation
   prmPred.cValue[example]:= prediction;//Y^
   prmErr.cValue[example]:= FY.cValue[example] - prediction;//Y - Y^
  end;
end;

procedure TCalcRegSvm.setDefaultParameter;
begin
  new(SvmParam);
  SvmParam^.svm_type := (OpPrm as TOpPrmRegressionSvm).SvmType;//-- non --
  SvmParam^.kernel_type := ord(LINEAR); //>>type de noyau
  SvmParam^.degree := 1; //>>degr du kernel
  SvmParam^.gamma := 0;	// 1/k >>gamma
  SvmParam^.coef0 := 0; //>>coef. 0 dans le kernel
  SvmParam^.nu := 0.5; //
  SvmParam^.cache_size := 100;//-- non --
  SvmParam^.C := 1;//>>penalit
  SvmParam^.eps := 1e-3;//>>tolerance
  SvmParam^.p := 0.1;//
  SvmParam^.shrinking := 1;//>>use shrinking heuristics
  SvmParam^.probability := 0;//>>use probability estimate
  SvmParam^.nr_weight := 0;//--non--
  SvmParam^.weight_label := nil;//--non--
  SvmParam^.weight := nil;//--non--
end;

end.
