(***********************************************************)
(* UCompRegSvm.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(***********************************************************)

{
@abstract(Classe de composant - Mthodes de rgression bases sur les SVM)
@author(Ricco)
@created(07/04/2009)

Utilisation de la bibliothque LIBSVM (http://www.csie.ntu.edu.tw/~cjlin/libsvm/)
importe sous forme de DLL.

}

unit UCompRegSvm;

interface

USES
    UOperatorDefinition, UCompDefinition, UCompRegDefinition,
    UDatasetDefinition, UCalcStatDesContinuousPlus, UDatasetExamples,
    UCalcRegSvm,
    Classes, IniFiles;

TYPE

    //***********************
    //gnrateur de composant
    //***********************
    TGenRegSvr = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    end;

    //*********
    //composant
    //*********
    TMLRegSvr = class(TMLCompRegression)
    end;

    //**********************************
    //oprateur rgression selon les SVM
    //**********************************
    TOpRegSvr = class(TOperatorRegression)
    private
    {liste des exognes}
    FLstExogenous: TLstAttributes;
    {classe de calcul importe}
    FCalcReg: TCalcRegSvm;
    {Stats sur les rsidus}
    FEpsilonStat: TCalcStatDesContPlus;
    {Quelques indications sur la qualit de la rgression}
    FSCT, FSCR, FRatio: double;
    {calculer quelques indicateurs}
    procedure   Evaluate(examples: TExamples);
    protected
    {dsalloue le calculateur}
    procedure   ReInitialize(); override;
    {tester la validit des attributs}
    function    CheckAttributes(): boolean; override;
    {description des rsidus en HTML}
    function    getResidualsHTML(): string;
    {la description de la rgression en HTML}
    function    getCoefsRegressionHTML(): string;
    public
    {dfinir une liste d'exognes vide}
    constructor Create(AOwner: TObject); override;
    {dtruire la liste d'exogne}
    destructor  Destroy; override;
    {excution effective}
    function    CoreExecute(): boolean; override;
    {rsum des rsultats}
    function    getHTMLResultsSummary(): string; override;
    {liste des exognes}
    property    LstExogenous: TLstAttributes read FLstExogenous;
    {accs  la classe de calcul}
    property    CalcReg: TCalcRegSvm read FCalcReg;
    end;

    //****************************************
    //paramtrage des rgression selon les SVM
    //****************************************
    TOpPrmRegressionSvm = class(TOperatorParameter)
    protected
    //description du paramtre interne de la mthode
    function    getHtmlInternalParameter(): string; virtual; abstract;
    //mettre les paramtres par dfaut
    procedure   SetDefaultParameters(); override;
    public
    //type de SVM (3: epsilon SVR ; 4: nu SVR)-  dfinir imprativement chez les hritiers
    SvmType: integer;
    //paramtres propres au type de SVR (selon le cas, masques ou pas chez les hritiers... ah l l...
    SvmTypeEpsilon: double; //0.1
    SvmTypeNu: double; //0.5
    //copier-coller simple  partir des paramtres de la LIBSVM
    _kernel_type: integer;//kernel type
    degree: double;	//* for poly */
    gamma: double;	//* for poly/rbf/sigmoid */
    coef0: double;	//* for poly/sigmoid */
    eps: double;	//* stopping criteria */
    C: double;	//* for C_SVC, EPSILON_SVR and NU_SVR */
    shrinking: integer;	//* use the shrinking heuristics */
    //affichage
    function    getHTMLParameters(): string; override;
    //I-O
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    end;

implementation

uses
    UDatasetImplementation,
    UCalcStatDes,
    ULibSVMImport, UConstConfiguration, SysUtils;

const
    STR_SVM_TYPE : array[3..4] of string = ('Epsilon-SVR','Nu-SVR');    

{ TOpPrmRegressionSvm }

procedure TOpPrmRegressionSvm.SetDefaultParameters;
begin
  _kernel_type := ord(LINEAR); //>>type de noyau
  degree := 3; //>>degr du kernel (si polynomial)
  gamma := 0;	// 1/k >>gamma
  coef0 := 0; //>>coef. 0 dans le kernel
  C := 1;//>>penalit
  eps := 0.001;//>>tolerance
  shrinking := 1;//>>use shrinking heuristics
  //paramtres spcifiques aux mthodes
  SvmTypeEpsilon:= 0.1;//pour epsilon-SVR
  SvmTypeNu:= 0.5;//pour nu-SVR
end;

procedure TOpPrmRegressionSvm.LoadFromINI(prmSection: string; prmINI: TMemIniFile);
begin
  SvmType:= prmINI.ReadInteger(prmSection,'svm_type',SvmType);
  _kernel_type:= prmINI.readInteger(prmSection,'kernel_type',_kernel_type);//kernel type
  degree:= prmINI.readFloat(prmSection,'degree',degree);	//* for poly */
  gamma:= prmINI.readFloat(prmSection,'gamma',gamma);	//* for poly/rbf/sigmoid */
  coef0:= prmINI.readFloat(prmSection,'coef0',coef0);	//* for poly/sigmoid */
  eps:= prmINI.readFloat(prmSection,'eps',eps);	//* stopping criteria */
  C:= prmINI.readFloat(prmSection,'C',C);	//* for C_SVC, EPSILON_SVR and NU_SVR */
  shrinking:= prmINI.readInteger(prmSection,'shrinking',shrinking);	//* use the shrinking heuristics */
  SvmTypeEpsilon:= prmINI.ReadFloat(prmSection,'epsilon_svr',SvmTypeEpsilon);
  SvmTypeNu:= prmINI.ReadFloat(prmSection,'nu_svr',SvmTypeNu);
end;

procedure TOpPrmRegressionSvm.LoadFromStream(prmStream: TStream);
begin
  prmStream.ReadBuffer(SvmType,sizeof(SvmType));
  prmStream.ReadBuffer(_kernel_type,sizeof(_kernel_type));//kernel type
  prmStream.ReadBuffer(degree,sizeof(degree));
  prmStream.ReadBuffer(gamma,sizeof(gamma));
  prmStream.ReadBuffer(coef0,sizeof(coef0));
  prmStream.ReadBuffer(eps,sizeof(eps));
  prmStream.ReadBuffer(C,sizeof(C));
  prmStream.ReadBuffer(shrinking,sizeof(shrinking));
  prmStream.ReadBuffer(SvmTypeEpsilon,sizeof(SvmTypeEpsilon));
  prmStream.ReadBuffer(SvmTypeNu,sizeof(SvmTypeNu));
end;

procedure TOpPrmRegressionSvm.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
  prmINI.WriteInteger(prmSection,'svm_type',SvmType);
  prmINI.WriteInteger(prmSection,'kernel_type',_kernel_type);//kernel type
  prmINI.WriteFloat(prmSection,'degree',degree);	//* for poly */
  prmINI.WriteFloat(prmSection,'gamma',gamma);	//* for poly/rbf/sigmoid */
  prmINI.WriteFloat(prmSection,'coef0',coef0);	//* for poly/sigmoid */
  prmINI.WriteFloat(prmSection,'eps',eps);	//* stopping criteria */
  prmINI.WriteFloat(prmSection,'C',C);	//* for C_SVC, EPSILON_SVR and NU_SVR */
  prmINI.WriteInteger(prmSection,'shrinking',shrinking);	//* use the shrinking heuristics */
  prmINI.WriteFloat(prmSection,'epsilon_svr',SvmTypeEpsilon);
  prmINI.WriteFloat(prmSection,'nu_svr',SvmTypeNu);
end;

procedure TOpPrmRegressionSvm.SaveToStream(prmStream: TStream);
begin
  prmStream.WriteBuffer(SvmType,sizeof(SvmType));
  prmStream.WriteBuffer(_kernel_type,sizeof(_kernel_type));//kernel type
  prmStream.WriteBuffer(degree,sizeof(degree));
  prmStream.WriteBuffer(gamma,sizeof(gamma));
  prmStream.WriteBuffer(coef0,sizeof(coef0));
  prmStream.WriteBuffer(eps,sizeof(eps));
  prmStream.WriteBuffer(C,sizeof(C));
  prmStream.WriteBuffer(shrinking,sizeof(shrinking));
  prmStream.WriteBuffer(SvmTypeEpsilon,sizeof(SvmTypeEpsilon));
  prmStream.WriteBuffer(SvmTypeNu,sizeof(SvmTypeNu));
end;

function TOpPrmRegressionSvm.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Parameters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Type</TD><TD align=right>%s</TD></TR>',[STR_SVM_TYPE[SvmType]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + self.getHtmlInternalParameter();
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Kernel type</TD><TD align=right>%s</TD></TR>',[str_kernel_type[kernel_type(_kernel_type)]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Degree (poly)</TD><TD align=right>%.2g</TD></TR>',[degree]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Gamma in kernel function (poly/rbf/sigmoid)</TD><TD align=right>%.2g</TD></TR>',[gamma]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Coef0 in kernel function(poly/sigmoid)</TD><TD align=right>%.2g</TD></TR>',[Coef0]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Tolerance of termination criteria</TD><TD align=right>%.4g</TD></TR>',[eps]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>C (Complexity Cost)</TD><TD align=right>%.2g</TD></TR>',[C]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Use shrinking heuristics</TD><TD align=right>%d</TD></TR>',[shrinking]);
 s:= s + '</table>';
 result:= s;
end;

{ TGenRegSvr }

procedure TGenRegSvr.GenCompInitializations;
begin
 FMLComp:= mlcRegression;
end;

{ TOpRegSvr }

function TOpRegSvr.CheckAttributes: boolean;
var ok: boolean;
begin
 ok:= inherited CheckAttributes();
 //vrifier que les exognes sont tous continus
 ok:= ok and (workdata.LstAtts[asInput].isAllCategory(caContinue));
 //rcuprer dans les champs internes
 if ok
  then
   begin
    FLstExogenous.Clear();
    FLstExogenous.Assign(workdata.LstAtts[asInput]);
   end;
 result:= ok;
end;

function TOpRegSvr.CoreExecute: boolean;
begin
 //init.
 result:= FALSE;
 //vider
 if assigned(FCalcReg) then FreeAndNil(FCalcReg);
 if assigned(FEpsilonStat) then FreeAndNil(FEpsilonStat);
 //puis crer et excuter l'objet de calcul en passant les donnes en paramtres
 FCalcReg:= TCalcRegSvm.create(self.Endogenous,self.LstExogenous,self.PrmOp);
 if FCalcReg.execute(self.WorkData.Examples)
  then
   begin
    //projection et calcul des rsidus -- la taille des attributs est suppose correcte -- cf. "procedure TMLCompRegression.InitializeDataset;"
    FCalcReg.Predict(self.EndoPred,self.ErrPred);
    //calculer les stats sur les rsidus
    FEpsilonStat:= TCalcStatDesContPlus.Create(self.ErrPred,self.WorkData.Examples);
    //calculer quelques stats d'valuation de la rgression
    self.Evaluate(self.WorkData.Examples);
    //et donc, tout est bien
    result:= TRUE;
   end;
end;

constructor TOpRegSvr.Create(AOwner: TObject);
begin
 inherited Create(AOwner);
 FLstExogenous:= TLstAttributes.Create(FALSE,Workdata.LstAtts[asAll].Size);
end;

destructor TOpRegSvr.Destroy;
begin
 FLstExogenous.Free;
 inherited;
end;

procedure TOpRegSvr.Evaluate(examples: TExamples);
var statEndo, statResidu: TCalcStatDesContinuous;
begin
 statEndo:= TCalcStatDesContinuous.Create(self.Endogenous,examples);
 statResidu:= TCalcStatDesContinuous.Create(self.ErrPred,examples);
 if (statEndo.TSS > 0.0)
  then FRatio:= statResidu.TSS / statEndo.TSS
  else FRatio:= 0.0;
 FSCT:= statEndo.TSS;
 FSCR:= statResidu.TSS;
 FreeAndNil(statEndo);
 FreeAndNil(statResidu);
end;

function TOpRegSvr.getCoefsRegressionHTML: string;
begin
 if assigned(FCalcReg)
  then result:= FCalcReg.GetHTMLResult()
  else result:= '';
end;

function TOpRegSvr.getHTMLResultsSummary: string;
var s: string;
begin
 s:= self.getCoefsRegressionHTML();
 s:= s + '<H3>Analysis of variance</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Sum of squares</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Total</TD><TD align=right>%.4f</TD></TR>',[FSCT]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Error</TD><TD align=right>%.4f</TD></TR>',[FSCR]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Pseudo-R2 (1-Error/Total)</TD><TD align=right>%.4f</TD></TR>',[1.0-FRatio]);
 s:= s + '</table>';
 result:= s + self.getResidualsHTML();
end;

function TOpRegSvr.getResidualsHTML: string;
var s: string;
begin
 if assigned(FEpsilonStat)
  then
   begin
    //les stats sur les rsidus
    s:= '<H3>Residual Analysis</H3>';
    s:= s+format('<TABLE class="BodyStyle" cellspacing=2 cellpadding=2 %s>',[HTML_BGCOLOR_DATA_GRAY]);
    s:= s+format('%s<TH>Att. name</TH><TH width="300">Full statistics</TH><TH>Histogram</TH></TR>',[HTML_TABLE_COLOR_HEADER_BLUE]);
    s:= s+FEpsilonStat.getHTMLResult();
    s:= s+'</table>';
    result:= s;
   end
  else result:= '';
end;

procedure TOpRegSvr.ReInitialize;
begin
 inherited ReInitialize();
 if assigned(FCalcReg) then FreeAndNil(FCalcReg);
end;

end.
