(****************************************************************)
(* UCalcRegTreeCART.pas - Copyright (c) 2010 Ricco RAKOTOMALALA *)
(****************************************************************)

{
@abstract(Arbre de Rgression -- Largement inspir par CART)
@author(Ricco)
@created(06/01/2010)

Utilisation du cadre CART pour construire un arbre de rgression.

Le point de dpart est toujours le clustering tree, mais avec une seule variable  prdire.

}
unit UCalcRegTreeCART;

interface

USES
    Contnrs,
    UCalcClusteringTree,
    UCalcStatDes, UDatasetdefinition, UDatasetImplementation,
    UOperatorDefinition,
    UDatasetExamples,
    UCalcRulesDefinition,
    UCalcTreeStructureDefinition,
    UCalcDistribution,
    UCompClusteringDefinition;

TYPE
        {type de donnes  valuer pour CTP}
        TEnumCTPDataSet = (ctpGrowingSet,ctpPruningSet);

        {enregistrements pour vrifier les squences d'arbres}
        TRecTreeSequenceClusCART = record
        {numro de sous-arbre}
        NumTree: integer;
        {nombre de feuilles}
        NbLeaves: integer;
        {mse}
        mseGrow: double;
        msePrune: double;
        {se(mse)}
        seprune: double;
        {theta pour la x-Se Rule}
        theta: double;
        end;

        {tableau des enregistrements de rsultats}
        TTabRecTreeSequenceClusCART = array of TRecTreeSequenceClusCART;

        {feuille pour CTP}
        TSplitLeafClusCART = class(TSplitLeafClus)
        end;

        {le split d'un attribut en supervis pour CART}
        TSplitAttributClusCART = class(TSplitAttributClus)
        protected
        function    getClassSplitLeaf(): TClassSplitLeaf; override;
        end;

        {une liste d'attributs candidats pour un sommet}
        TLstSplitAttClusCART = class(TLstSplitAttClus)
        protected
        function  getClassSplitAttribut(): TClassSplitAttribut; override;
        end;

        {un noeud de l'arbre CART}
        TMLTreeNodeClusCART = class(TMLTreeNodeClus)
        private
        //stats sur le pruning -- erreur quadratique
        FSum2Pruning: double;
        //pour le calcul de l'ecart-type de l'erreur quadratique
        FSum4Pruning: double;
        {numro d'appartenance dans les sous-arbres}
        FCurNsa: integer;
        protected
        function  getClassLstSplitAttributes(): TClassLstSplitAttributes; override;
        public
        //distance par rapport au centre de classe
        function  distToMean(example: integer): double;
        //
        destructor destroy; override;
        {test si un frre existe}
        function  nextBrotherExists(): boolean;
        {obtenir le frre suivant}
        function  getNextBrother(): TMLTreeNodeClusCART;
        {la liste des sous-arbres corresp.  un numro}
        procedure getSubTree(prmNsa: integer; lst: TObjectList);
        {cot de mauvais classement}
        function  localCost(prmSet: TEnumCTPDataSet): double;
        {laguer  partir du sommet courant}
        procedure pruneFromHere();
        end;

        //sommet d'un arbre de rgression CART
        TMLTreeNodeRegTreeCART = class(TMLTreeNodeClusCART)
        public
        function   getHTMLLeafInfos(): string; override;
        function   average(): double;
        function   stddev(): double;
        end;

        //cumul des valeurs calcules sur les noeuds
        TRecInfoNode = record
        growCost: double;
        pruneS4: double;
        end;

        {la structure de l'arbre}
        TMLTreeStructureClusCART = class(TMLTreeStructureClus)
        private
        {l'arbre optimal}
        FOptimalSubTree: integer;
        {la solution slectionne}
        FSelectedSubTree: integer;
        {squences d'arbres utiliss}
        FTabTreeSequence: TTabRecTreeSequenceClusCART;
        {***** recherche des sous-arbres *****}
        {liste des squences d'arbres}
        FLstSeqTree : TObjectList;
        {r-initialisation des numros de squences d'arbres}
        procedure  initialiserCurNsa();
        {numroter les squences d'arbres}
        procedure  numeroterCurNsa(var curNsa: integer);
        {marquer les sommets}
        function  mark_nodes(prmNsa: integer; prmAlpha: double): boolean;
        {calculer l'alpha suivante}
        function  computeNextAlpha(prmNsa: integer): double;
        {calculer alpha  partir d'un noeud}
        function  computeAlphaNode(prmSommet: TMLTreeNodeClusCART; prmNsa: Integer): Double;
        {tester si l'lagage est possible  partir d'un sommet}
        function  isPruningPossible(prmNode: TMLTreeNodeClusCART; prmNsa: Integer; prmAlpha: Double): Boolean;
        {marquer un sous-arbre  partir d'un sommet}
        procedure mark_subTree(prmNode: TMLTreeNodeClusCART; prmNsa: integer);
        {additionner les cots d'un sous-arbre}
        function  computeSubTreeCost(lst: TObjectList; prmSet: TEnumCTPDataSet): double;
        {calculer le cot en pruning d'un sous-arbre candidat}
        function  computePrunedSubTreeCosts(prmNsa: integer; var nbLeaves: integer; var info: TRecInfoNode): double;
        {rechercher le sous-arbre pour un pruning}
        procedure explorePrunedSubTreeCosts(node: TMLTreeNodeClusCART; prmNsa: integer; var costs: double; var nbLeaves: integer; var info: TRecInfoNode);
        {numrer les sous-arbres et leurs cots -- et renvoyer l'arbre optimal}
        function enumerateSubTrees(nbTrees: integer): integer;
        {laguer effectivement  partir du n de sous-arbre}
        procedure pruneSubTree(prmNsa: integer);
        {laguer effectivement  partir d'un sommet et du n de sous-arbre}
        procedure pruneSubTreeFromNode(prmNode: TMLTreeNodeClusCART; prmNsa: integer);
        {laguer  partir d'un sommet, quel que soit son numro}
        procedure pruningFromNode(prmNode: TMLTreeNodeClusCART);
        {*************************************}
        {insrer les individus en pruning}
        procedure applyPruningSet(prmExamples: TExamples);
        {isrer 1 individu en test}
        procedure applyPruningExample(node: TMLTreeNode; example: integer);
        protected
        FPruningSet: TExamples;
        //dcomposer app-pruning le cas chant, surcharg chez CART par ex.
        procedure  defineTrainingSet(prmExamples: TExamples); override;
        {type du noeud de l'arbre}
        function   getClassMLTreeNode(): TClassMLTreeNode; override;
        {ajouter le gestionnaire de liste de squences d'arbres}
        procedure  prepareInternalStructures(); override;
        {dtruire le gestionnaire de gestionnaire de squences d'arbres}
        procedure  destroyInternalStructures(); override;
        public
        destructor destroy(); override;
        {post-lagage  la CART}
        procedure   PostPruning(); override;
        {taille growing}
        function getGrowingSetSize(): integer;
        {taille pruning}
        function getPruningSetSize(): integer;
        {description de la squence d'arbres}
        property TabTreeSequence: TTabRecTreeSequenceClusCART read FTabTreeSequence;
        {numro de l'arbre slection}
        property SelectedSubTree: integer read FSelectedSubTree;
        {numro de l'arbre optimal}
        property OptimalSubTree: integer read FOptimalSubTree;
        end;

        //classe de structure d'arbre
        TMLTreeStrucRegTreeCART = class(TMLTreeStructureClusCART)
        protected
        function   getClassMLTreeNode(): TClassMLTreeNode; override;
        end;

        //oprateur de calcul
        TCalcClusTreeCART = class(TCalcClusTree)
        protected 
        function  getClassTreeStructure(): TClassMLTreeStrucClus; override;
        function  getHTMLSpecificDescription(): string; override; 
        end;

        //oprateur de calcul pour la rgression par arbres
        //driv du culstering (predictif) par arbres
        TCalcRegTreeCART = class(TCalcClusTreeCART)
        protected
        function  getClassTreeStructure(): TClassMLTreeStrucClus; override;
        public
        procedure Predict(prmPred, prmErr: TAttribute);
        end;

implementation

uses
    Sysutils, Math,
    UCompClusteringTreeCART, ULogFile, UConstConfiguration,
  UCompRegTreePredictive;

const
    //epsilon  prendre en compte dans les comparaisons lors du pruning
    CTP_PRUNING_EPSILON = 0.0001;

    //nombre de squences  montrer
    SEUIL_NB_SEQUENCES = 15;

{ TSplitAttributClusCART }

function TSplitAttributClusCART.getClassSplitLeaf: TClassSplitLeaf;
begin
 result:= TSplitLeafClusCART;
end;

{ TLstSplitAttClusCART }

function TLstSplitAttClusCART.getClassSplitAttribut: TClassSplitAttribut;
begin
 result:= TSplitAttributClusCART;
end;

{ TMLTreeNodeClusCART }

destructor TMLTreeNodeClusCART.destroy;
begin
 inherited;
end;

function TMLTreeNodeClusCART.distToMean(example: integer): double;
var norm: TTabVarianceNormalisation;
    j: integer;
    value,dist: double;
begin
 norm:= (self.TreeStructure as TMLTreeStructureClusCART).Normalization;
 dist:= 0.0;
 for j:= 0 to pred(self.FStats.Count) do
  begin
   value:= SQR(self.FStats.Stat(j).Attribute.cValue[example] - (self.FStats.Stat(j) as TCalcStatDesContinuous).Average);
   value:= value / norm[j];
   dist:= dist + value;
  end;
 result:= dist;
end;

function TMLTreeNodeClusCART.getClassLstSplitAttributes: TClassLstSplitAttributes;
begin
 result:= TLstSplitAttClusCART;
end;

function TMLTreeNodeClusCART.getNextBrother: TMLTreeNodeClusCART;
var p: Integer;
begin
 RESULT:= NIL;
 {$B-}
 if assigned(self.Predecessor) and (self.Predecessor.getCountSuccessors()>0)
  Then
   Begin
    p:= self.Predecessor.getSuccessor(self);
    //il existe, et il reste de la marge pour passer au suivant --> d'o le (-1) !!!
    if (p>=0) and (p<(self.Predecessor.getCountSuccessors()-1))
     then result:= self.Predecessor.getSuccessor(succ(p)) as TMLTreeNodeClusCART;
   End;
end;

procedure TMLTreeNodeClusCART.getSubTree(prmNsa: integer; lst: TObjectList);
var fini: Boolean;
    sommet: TMLTreeNodeClusCART;
begin
 //vider la liste
 lst.Clear();

 //initialiser
 fini:= FALSE;
 sommet:= self;//point de dpart, soi-mme !!!

 //parcourir
 while not(fini) do
  Begin
   //traiter le sommet courant
   if sommet.isLeaf() OR ((sommet.FCurNsa>0) and (sommet.FCurNsa<prmNsa))
    Then
     Begin
      if (sommet<>self)
       Then lst.Add(sommet);
     End;

   //passer au sommet suivant si possible
   if sommet.isLeaf() OR ((sommet.FCurNsa>0) and (sommet.FCurNsa<prmNsa))
    Then
     //passer au frre
     Begin
      //remonte
      while (sommet<>self) and not(sommet.nextBrotherExists()) do
       sommet:= sommet.Predecessor as TMLTreeNodeClusCART;

      //passage au frre si ce n'est pas la racine
      if (sommet<>self)
       Then sommet:= sommet.getNextBrother()
       Else FINI:= TRUE;//la recherche en arbre est termine
       
     End
    Else
     Begin
      sommet:= sommet.getSuccessor(0) as TMLTreeNodeClusCART;
     End;
  End;
end;

function TMLTreeNodeClusCART.localCost(prmSet: TEnumCTPDataSet): double;
begin
 case (prmSet) of
  ctpGrowingSet: result:= FStats.inertie((self.TreeStructure as TMLTreeStructureClus).Normalization)
  //pruning
  else result:= FSum2Pruning;
 end;
end;

function TMLTreeNodeClusCART.nextBrotherExists: boolean;
var p: Integer;
begin
 RESULT:= FALSE;
 {$B-}
 if assigned(self.Predecessor) and (self.Predecessor.getCountSuccessors()>0)
  Then
   Begin
    p:= self.Predecessor.getSuccessor(self);
    //il existe, et il reste de la marge pour passer au suivant --> d'o le (-1) !!!
    RESULT:= (p>=0) and (p<(self.Predecessor.getCountSuccessors()-1));
   End;
end;

procedure TMLTreeNodeClusCART.pruneFromHere;
var curNode: TMLTreeNode;
    i: integer;
begin
 if not(self.isLeaf())
  then
   begin
    for i:= 0 to pred(self.getCountSuccessors()) do
     begin
      curNode:= self.getSuccessor(i);
      //en rcursif, se vider de son sous-arbre
      (curNode as TMLTreeNodeClusCART).pruneFromHere();
      //si on est l c'est que le boulot a t fait
      //la suppression entrane la destruction de l'objet !!!
      self.TreeStructure.deleteNode(curNode);
     end;
    //vider la liste des successeurs -- rien  librer explicitement
    self.clearSuccessors();
   end;
end;

procedure TMLTreeStructureClusCART.applyPruningExample(node: TMLTreeNode;
  example: integer);
var curNode: TMLTreeNodeClusCART;
    nextNode: TMLTreeNode; 
    i: integer;
    dist: double;
begin
 curNode:= node as TMLTreeNodeClusCART;
 //ajouter l'exemple dans le comptage -- distance par rapport au centre de classe
 dist:= curNode.distToMean(example);
 curNode.FSum2Pruning:= curNode.FSum2Pruning + dist;
 curNode.FSum4Pruning:= curNode.FSum4Pruning + dist * dist;
 //passer au suivant en dtectant la feuille concerne
 if not(curNode.isLeaf)
  then
   begin
    //chercher le sommet  appeler
    nextNode:= NIL;
    for i:= 0 to pred(curNode.getCountSuccessors()) do
     begin
      if curNode.getSuccessor(i).Condition.TestExample(example)
       then
        begin
         nextNode:= curNode.getSuccessor(i);
         BREAK;
        end;
     end;
    //alors, alors, s'il n'y a rien, on est mal parce que c'est une valeur non-rfrence !!!
    if assigned(nextNode)
     then self.applyPruningExample(nextNode,example)
     else TraceLog.WriteToLogFile(format('[CTP] !!!! ERREUR !!!, valeur non rfrence dans PRUNING, example n %d',[example]));
   end;
end;

procedure TMLTreeStructureClusCART.applyPruningSet(prmExamples: TExamples);
var i,iNode: integer;
begin
 //initaliser tous les cots sur les noeuds
 for iNode:= 0 to pred(self.CountNodes) do
  begin
   (self.Node[iNode] as TMLTreeNodeClusCART).FSum2Pruning:= 0.0;
   (self.Node[iNode] as TMLTreeNodeClusCART).FSum4Pruning:= 0.0;
  end;
 //puis lancer les exemples pour le comptage des erreurs
 for i:= 1 to prmExamples.Size do
  applyPruningExample(self.RootNode,prmExamples.Number[i]);
end;

function TMLTreeStructureClusCART.computeAlphaNode(
  prmSommet: TMLTreeNodeClusCART; prmNsa: Integer): Double;
var cSommet,cFeuilles,tmp: Double;
begin
 cSommet:= prmSommet.localCost(ctpGrowingSet);
 //rcuprer les feuilles
 prmSommet.getSubTree(prmNsa,self.FLstSeqTree);
 //le cot des feuilles
 cFeuilles:= self.computeSubTreeCost(self.FLstSeqTree,ctpGrowingSet);
 //l'opration cabalistique
 tmp:= (cSommet-cFeuilles)/(1.0*self.FLstSeqTree.Count-1.0);
 //renvoyer le tout
 RESULT:= tmp;
end;

function TMLTreeStructureClusCART.computeNextAlpha(prmNsa: integer): double;
var sommet: TMLTreeNodeClusCART;
    alpha,minAlpha: double;
    i: Integer;    
begin
 minAlpha:= 1.0e308;
 //pour tous les sommets non-marqus
 For i:= 0 To pred(self.CountNodes) do
  Begin
   sommet:= self.Node[i] as TMLTreeNodeClusCART;
   //le sommet n'a pas encore t explor ?
   //ce ne doit pas tre non plus une feuille
   If (sommet.FCurNsa=0) and NOT(sommet.isLeaf)
    Then
     Begin
      alpha:= computeAlphaNode(sommet,prmNsa);
      If (alpha<minAlpha)
       Then minAlpha:= alpha;
     End;
  End;
 //TraceLog.WriteToLogFile(format('[CTP] alpha = %.4f',[minAlpha]));
 RESULT:= minAlpha;
end;

function TMLTreeStructureClusCART.computePrunedSubTreeCosts(prmNsa: integer;
  var nbLeaves: integer; var info: TRecInfoNode): double;
var cost: double;
begin
 //calculer le cout de l'arbre prun de niveau prmNsa
 cost:= 0;
 info.growCost:= 0.0;
 info.pruneS4:= 0.0;
 nbLeaves:= 0;
 self.explorePrunedSubTreeCosts(self.RootNode as TMLTreeNodeClusCART,prmNsa,cost,nbLeaves,info);
 RESULT:= cost;
end;

function TMLTreeStructureClusCART.computeSubTreeCost(lst: TObjectList;
  prmSet: TEnumCTPDataSet): double;
var i: integer;
    cost: double;
begin
 cost:= 0;
 for i:= 0 to pred(lst.Count) do
  cost:= cost+(lst.Items[i] as TMLTreeNodeClusCART).localCost(prmSet);
 result:= cost;
end;

procedure TMLTreeStructureClusCART.defineTrainingSet(prmExamples: TExamples);
var proportion: single;
begin
 FTrainingSet:= TExamples.Create(prmExamples.Size);
 FPruningSet:= TExamples.Create(prmExamples.Size);
 //subdiviser : proportion est la taille relative du train ici et "PruningSetSize" est en pourcentage
 //new -- 24/08/2006 -- suggestion Rainer Block, taille de pruning gal  zro...
 if ((self.PrmMethod as TPrmOpRegTreeCART).PruningSetSize > 0)
  then
   begin
    proportion:= 1.0-1.0*(self.PrmMethod as TPrmOpRegTreeCART).PruningSetSize/100.0;
    prmExamples.SamplingSplitting(proportion,FTrainingSet,FPruningSet,(self.PrmMethod as TPrmOpRegTreeCART).ModeRndGenerator);
   end
  else
   begin
    FTrainingSet.Copy(prmExamples);
    FPruningSet.Size:= 0;
   end;
end;

destructor TMLTreeStructureClusCART.destroy();
begin
 FreeAndNil(FTrainingSet);
 FreeAndNil(FPruningSet);
 inherited;
end;

procedure TMLTreeStructureClusCART.destroyInternalStructures;
begin
 inherited;
 FLstSeqTree.Free();
 SetLength(FTabTreeSequence,0);
end;

function TMLTreeStructureClusCART.enumerateSubTrees(
  nbTrees: integer): integer;
var subTree,nbLeaves,optSubTree: integer;
    pruneCost,growCost,minCost,thresold: double;
    info,minInfo: TRecInfoNode;
    seMinCost: double;
    nPrim: double;
begin
 nPrim:= 1.0 * FPruningSet.Size;
 //
 SetLength(FTabTreeSequence,nbTrees);
 //lister les squences avec leurs perfs. respectives
 //en profiter pour dtecter l'arbre optimal
 FOptimalSubTree:= -1;
 minCost:= MATH.MaxDouble;
 for subTree:= 1 to nbTrees do
  begin
   pruneCost:= self.computePrunedSubTreeCosts(subTree,nbLeaves,info);
   pruneCost:= pruneCost/nPrim;
   growCost:= info.growCost/(1.0*self.FTrainingSet.Size);
   //affecter les valeurs
   FTabTreeSequence[pred(subTree)].NumTree:= subTree;
   FTabTreeSequence[pred(subTree)].NbLeaves:= nbLeaves;
   FTabTreeSequence[pred(subTree)].mseGrow:= growCost;
   FTabTreeSequence[pred(subTree)].msePrune:= pruneCost;
   FTabTreeSequence[pred(subTree)].sePrune:= 1.0/sqrt(nPrim)*Math.Power(1.0/nPrim*info.pruneS4-SQR(pruneCost),0.5);//Breiman, pp.226
   FTabTreeSequence[pred(subTree)].theta:= -1.0;
   //tester
   if (pruneCost <= minCost)
    then
     begin
      minCost:= pruneCost;
      minInfo:= info;
      FOptimalSubTree:= subTree;
     end;
  end;
 //erreur type du mse de l'arbre minimum
 seMinCost:= FTabTreeSequence[pred(FOptimalSubTree)].sePrune;
 //pour tous les arbres, plus petits que FOptimalSubtree - calculer la valeur de theta pour l'obtenir
 if (seMinCost > 0.0)
  then for subTree:= nbTrees downto FOptimalSubTree do FTabTreeSequence[pred(subTree)].Theta:= (FTabTreeSequence[pred(subTree)].msePrune - minCost) / seMinCost;
 //appliquer la x-SE RULE pour trouver le bon seuil
 thresold:= minCost+ (self.PrmMethod as TPrmOpRegTreeCART).SERule * seMinCost;//rgle de l'cart-type
 //infos. sur l'erreur min et son ecart-type
 TraceLog.WriteToLogFile(format('[REGTREE-CART] min err pruning (se) = %.4f (%.4f) --> threshold = %.4f',[minCost,seMinCost,thresold]));
 //rechercher le "bon"  partir des moins complexes vers les plus complexes
 optSubTree:= -1;
 for subTree:= nbTrees downto 1 do
  begin
   if (FTabTreeSequence[pred(subTree)].msePrune <= thresold)
    then
     begin
      optSubTree:= subTree;
      BREAK;
     end;
  end;
 TraceLog.WriteToLogFile(format('[REGTREE-CART] the selected tree is n%2d, with err = %.4f',[optSubTree,FTabTreeSequence[pred(optSubTree)].msePrune]));
 //renvoyer la bonne rponse
 result:= optSubTree;
end;

procedure TMLTreeStructureClusCART.explorePrunedSubTreeCosts(
  node: TMLTreeNodeClusCART; prmNsa: integer; var costs: double;
  var nbLeaves: integer; var info: TRecInfoNode);
var i: Integer;
begin
 if (node.FCurNsa > prmNsa) and NOT(node.isLeaf())
  Then
   Begin
    For i:= 0 To pred(node.getCountSuccessors) Do
     self.explorePrunedSubTreeCosts(node.getSuccessor(i) as TMLTreeNodeClusCART,prmNsa,costs,nbLeaves,info);
   End
  //ds le premier trouv, on calcule et on arrte les frais
  Else
   begin
    costs:= costs + node.localCost(ctpPruningSet);
    //passer les autres infos
    info.growCost:= info.growCost + node.localCost(ctpGrowingSet);
    info.pruneS4:= info.pruneS4 + node.FSum4Pruning;
    //
    inc(nbLeaves);
    (*
    //-- affichage de vrification --
    sTmp:= node.DistClass.TabFreq.getStringResult();
    sTmp:= sTmp+format(' -- growErr = %.2f, pruneErr (on %d examples) = %.2f',[node.localCost(cartGrowingSet),node.FPruningCount,node.localCost(cartPruningSet)]);
    TraceLog.WriteToLogFile(sTmp);
    //--
    *)
   end;
end;

function TMLTreeStructureClusCART.getClassMLTreeNode: TClassMLTreeNode;
begin
 result:= TMLTreeNodeClusCART; 
end;

function TMLTreeStructureClusCART.getGrowingSetSize: integer;
begin
 result:= self.FTrainingSet.Size;
end;

function TMLTreeStructureClusCART.getPruningSetSize: integer;
begin
 result:= self.FPruningSet.Size;
end;

procedure TMLTreeStructureClusCART.initialiserCurNsa;
var i: integer;
begin
 self.FLstSeqTree.Clear();
 for i:= 0 to pred(self.CountNodes) do
  (self.Node[i] as TMLTreeNodeClusCART).FCurNsa:= 0;
end;

function TMLTreeStructureClusCART.isPruningPossible(
  prmNode: TMLTreeNodeClusCART; prmNsa: Integer; prmAlpha: Double): Boolean;
var cSommet,cFeuille: Double;
begin
 cSommet:= prmNode.localCost(ctpGrowingSet)+prmAlpha;
 //rcuprer les feuilles
 prmNode.getSubTree(prmNsa,self.FLstSeqTree);
 //le cot des feuilles
 cFeuille:= self.computeSubTreeCost(self.FLstSeqTree,ctpGrowingSet);
 //ajouter la complexit
 cFeuille:= cFeuille+prmAlpha*self.FLstSeqTree.Count;
 //alors...
 RESULT:= (cSommet <= cFeuille + CTP_PRUNING_EPSILON);
end;

function TMLTreeStructureClusCART.mark_nodes(prmNsa: integer;
  prmAlpha: double): boolean;
var the_end: Boolean;
    node: TMLTreeNodeClusCART;
    okPruning: Boolean;
begin
 RESULT:= FALSE;

 //initialiser
 the_end:= FALSE;
 node:= self.RootNode as TMLTreeNodeClusCART;

 //la boucle de recherche
 //elle mule une fonction rcursive
 While NOT(the_end) Do
  Begin
   //doit-on traiter le sommet ?
   //c'est le cas si le sommet n'est pas une feuille et qu'elle est non numrote
   if Not(node.isLeaf) and (node.FCurNsa=0)
    Then
     Begin
      //tester si'l est possible d'laguer  partir du sommet courant
      //le test repose sur la recherche des feuilles du sous-arbre
      okPruning:= isPruningPossible(node,prmNsa,prmAlpha);

      //si ok, alors on insre le numro de squence d'arbre sous le sommet, y compris lui mme
      if okPruning
       Then self.mark_subTree(node,prmNsa);

      //si l'on est  la racine et que l'lagage a t accept
      //il n'y a plus rien  faire
      if (node = self.RootNode) and okPruning
       Then
        Begin
         RESULT:= TRUE;
         BREAK;
        End;
      
     End;

   //on passe au sommet suivant

   //tester si c'est une feuille OU si les sommets en dessous sont inintrssants
   if node.isLeaf OR ((node.FCurNsa>0) and (Node.FCurNsa<prmNsa))
    Then
     //on passe au frre
     Begin
      //remonte
      while (node<>self.RootNode) and not(node.nextBrotherExists()) do
       node:= node.Predecessor as TMLTreeNodeClusCART;

      //passage au frre si l'on n'est pas sur la racine
      if (node<>self.RootNode)
       Then node:= node.getNextBrother()
       Else the_end:= TRUE;
       
     End
    Else
     //on descend sur l'enfant
     Begin
      node:= node.getSuccessor(0) as TMLTreeNodeClusCART;
     End;
  End;

end;

procedure TMLTreeStructureClusCART.mark_subTree(prmNode: TMLTreeNodeClusCART;
  prmNsa: integer);
Var i: Integer;
begin
 If (prmNode.FCurNsa=0)
  Then
   //le sommet n'a pas t pralablement numrot
   Begin
    prmNode.FCurNsa:= prmNsa;
    If NOT(prmNode.isLeaf())
     Then
      Begin
       //appel rcursif vers les feuilles
       For i:= 0 To pred(prmNode.getCountSuccessors()) Do
        self.mark_subTree(prmNode.getSuccessor(i) as TMLTreeNodeClusCART,prmNsa);
      End;
   End;
end;

procedure TMLTreeStructureClusCART.numeroterCurNsa(var curNsa: integer);
var curAlpha: double;
begin
 //tout remettre  zro
 self.initialiserCurNsa();
 //dfinir la squence
 curNsa:= 1;
 curAlpha:= 0.0;
 if (self.CountNodes=1)
  //un seul sommet : la racine
  then (self.RootNode as TMLTreeNodeClusCART).FCurNsa:= curNsa
  else
   //chercher les squences
   begin
    while not(self.mark_nodes(curNsa,curAlpha)) do
     begin
      inc(curNsa);
      curAlpha:= computeNextAlpha(curNsa);
     end;
   end;
end;

procedure TMLTreeStructureClusCART.PostPruning;
var nbSubTrees: integer;
begin
 //-- appliquer le pruning-set
 if (self.FPruningSet.Size>0)
  then
   begin
     self.applyPruningSet(self.FPruningSet);
     //-- numroter les squences de sous-arbres --
     self.numeroterCurNsa(nbSubTrees);
     //dtecter celle correspondant  l'optimum
     FSelectedSubTree:= self.EnumerateSubTrees(nbSubTrees);
     //effectuer le post-lagage physique
     if (FSelectedSubTree>0)
      then self.pruneSubTree(FSelectedSubTree);
   end
  else TraceLog.WriteToLogFile('[CTP] mode SANS PRUNING asked ???');
end;

procedure TMLTreeStructureClusCART.prepareInternalStructures;
begin
 inherited;
 //liste non propritaire...
 FLstSeqTree:= TObjectList.Create(FALSE);
end;

procedure TMLTreeStructureClusCART.pruneSubTree(prmNsa: integer);
begin
 self.pruneSubTreeFromNode(self.RootNode as TMLTreeNodeClusCART,prmNsa);
end;

procedure TMLTreeStructureClusCART.pruneSubTreeFromNode(
  prmNode: TMLTreeNodeClusCART; prmNsa: integer);
Var i: Integer;
begin
 if (prmNode.FCurNsa > prmNsa)
  Then
   Begin
    if NOT(prmNode.isLeaf())
     Then
      Begin
       For i:= 0 To pred(prmNode.getCountSuccessors()) do
        self.pruneSubTreeFromNode(prmNode.getSuccessor(i) as TMLTreeNodeClusCART,prmNsa);
      End;
   End
  //et ses enfants sont limins par la mme occasion
  Else self.pruningFromNode(prmNode);
end;

procedure TMLTreeStructureClusCART.pruningFromNode(
  prmNode: TMLTreeNodeClusCART);
begin
 if not(prmNode.isLeaf())
  then
   begin
    prmNode.pruneFromHere;
    //et ne pas oublier de s'ajouter dans la liste des feuilles
    self.addLeaveReference(prmNode);
   end;
end;

{ TCalcClusTreeCTP }

function TCalcClusTreeCART.getClassTreeStructure: TClassMLTreeStrucClus;
begin
 result:= TMLTreeStructureClusCART;
end;

function TCalcClusTreeCART.getHTMLSpecificDescription: string;
var i: integer;
    arbre: TMLTreeStructureClusCART;
    tmp,colorLine: string;
    ok_NB_SEQ,curOk: boolean;
begin
 arbre:= FTree as TMLTreeStructureClusCART;
 ok_Nb_SEQ:= (Length(arbre.TabTreeSequence)<=SEUIL_NB_SEQUENCES);
 tmp:= format('<H3>Trees sequence (# %d) -- Mean Squared Error</H3>',[Length(arbre.TabTreeSequence)]);;
 tmp:= tmp+HTML_HEADER_TABLE_RESULT;
 tmp:= tmp+HTML_TABLE_COLOR_HEADER_GRAY+'<TH>N</TH><TH># Leaves</TH><TH>mse (growing)</TH><TH>mse (pruning)</TH><TH width=60>se (mse)</TH><TH width=60>x</TH></TR>';
 for i:= high(arbre.TabTreeSequence) downto low(arbre.TabTreeSequence) do
  begin
   curOk:= ok_Nb_SEQ OR (self.PrmCalc as TPrmOpRegTreeCART).ShowAllTreeSeq;
   //si a passe pas, un des cas suivant peut faire passer
   curOk:= curOk OR
           ((i=low(arbre.TabTreeSequence))
             or (i=high(arbre.TabTreeSequence))
             or (arbre.TabTreeSequence[i].NumTree = arbre.OptimalSubTree)
             or (arbre.TabTreeSequence[i].NumTree = arbre.SelectedSubTree)
           );
   //suite normale... si ok...
   if curOk
    then
     begin
       colorLine:= HTML_TABLE_COLOR_DATA_GRAY;
       if (arbre.TabTreeSequence[i].NumTree = arbre.OptimalSubTree)
        then colorLine:= HTML_TABLE_COLOR_DATA_GREEN;
       if (arbre.TabTreeSequence[i].NumTree = arbre.SelectedSubTree)
        then colorLine:= HTML_TABLE_COLOR_DATA_RED;
       tmp:= tmp + colorLine +
       format('<TD align="right">%d</TD><TD align="right">%d</TD><TD align="right">%.4f</TD><TD align="right">%.4f</TD><TD align="right">%.4f</TD>',
                                 [arbre.TabTreeSequence[i].NumTree,arbre.TabTreeSequence[i].NbLeaves,arbre.TabTreeSequence[i].mseGrow,arbre.TabTreeSequence[i].msePrune,arbre.TabTreeSequence[i].sePrune]);
       if (arbre.TabTreeSequence[i].theta < 0.0)
        then tmp:= tmp + '<TD align=center>-</TD>'
        else tmp:= tmp + format('<TD align=right>%.4f</TD>',[arbre.TabTreeSequence[i].theta]);
       tmp:= tmp + '</TR>';
     end;
  end;
 tmp:= tmp+'</table>';
 result:= tmp;
end;

{ TMLTreeNodeRegTreeCART }

function TMLTreeNodeRegTreeCART.average: double;
begin
 result:= (FStats.Stat(0) as TCalcStatDesContinuous).average;
end;

function TMLTreeNodeRegTreeCART.getHTMLLeafInfos: string;
begin
 result:= format(' then <b>avg(%s) = %.4f</b> (std-dev = %.4f, with %d examples [%.2f%s])',
                [FStats.Stat(0).Attribute.Name,
                 self.average(),self.stddev(),
                 FStats.NbExamples,100.0*FStats.NbExamples/(1.0*(self.TreeStructure as TMLTreeStrucRegTreeCART).FStats.NbExamples),'%']);
end;

function TMLTreeNodeRegTreeCART.stddev: double;
begin
 result:= (FStats.Stat(0) as TCalcStatDesContinuous).StdDev;
end;

{ TMLTreeStrucRegTreeCART }

function TMLTreeStrucRegTreeCART.getClassMLTreeNode: TClassMLTreeNode;
begin
 result:= TMLTreeNodeRegTreeCART;
end;

{ TCalcRegTreeCTP }

function TCalcRegTreeCART.getClassTreeStructure: TClassMLTreeStrucClus;
begin
 result:= TMLTreeStrucRegTreeCART;
end;

procedure TCalcRegTreeCART.Predict(prmPred, prmErr: TAttribute);
var i: integer;
    endo: TAttribute;
    leaf: TMLTreeNodeRegTreeCART;
    value: TTypeContinue; 
begin
 endo:= self.Targets.Attribute[0];
 //construire la prdiction pour chaque exemple et calculer l'erreur
 for i:= 1 to endo.Size do
  begin
   //calcul de la prdiction
   leaf:= self.SetLeafExample(i) as TMLTreeNodeRegTreeCART;
   if assigned(leaf)
    then value:= leaf.average()
    //si pas couvert (?), on prend la moyenne globale
    else value:= ((self.FTree as TMLTreeStrucRegTreeCART).RootNode as TMLTreeNodeRegTreeCART).average();
   //assignation
   prmPred.cValue[i]:= value;
   prmErr.cValue[i]:= endo.cValue[i]-value;
  end;
end;

end.
