(**********************************************************************)
(* UCompRegTreePredictive.pas - Copyright (c) 2010 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Arbre de Rgression -- Implmentation de CART)
@author(Ricco)
@created(06/01/2010)

Utilisation du cadre CART pour construire un arbre de rgression
>> chapiter 8, en particulier la formule pour le post-lagage est en pp.226

}

unit UCompRegTreePredictive;

interface

uses
    Forms, Classes, IniFiles,
    UCompDefinition,
    UCompRegDefinition,
    UOperatorDefinition,
    UCompClusteringTree,
    UDatasetDefinition,
    UDatasetImplementation,
    UDatasetExamples,
    UCalcRegTreeCART;

TYPE

    //gnrateur
    TMLGenRegTreeCART = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //composant
    TMLCompRegTreeCART = class(TMLCompRegression)
    protected
    function    getClassOperator: TClassOperator; override;
    function    getGenericAttName(): string; override;
    end;

    //oprateur -- attention, la grosse astuce, on passe par la classe de clustering, mais on force la slection d'une seule variable TARGET !
    TOpRegTreeCART = class(TOperatorRegression)
    private
    //classe de calcul effectif, c'est en fait un clustering tree qui est en dessous
    FCalc: TCalcRegTreeCART;
    //variable dont les modalits sont les clusters correspondant aux feuilles de l'arbre
    FClusters: TAttDiscrete;
    //coefficient de dtermination
    R2: double;
    protected
    function    getClassParameter: TClassOperatorParameter; override;
    function    CoreExecute(): boolean; override;
    procedure   evaluation(examples: TExamples);
    public
    function    getHTMLResultsSummary(): string; override;
    destructor  destroy(); override;
    function    getNewForm(): TForm; override;
    property    Calc: TCalcRegTreeCART read FCalc;
    end;

    //paramtrage
    TPrmOpRegTreeCART = class(TOpPrmClusTree)
    protected
    FPruningSetSize: integer;
    FSERule: double;
    function    CreateDlgParameters(): TForm; override;
    procedure   SetDefaultParameters(); override;
    public
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    //proprits
    property PruningSetSize: integer read FPruningSetSize write FPruningSetSize;
    property SERule: double read FSERule write FSERule;
    property ShowAllTreeSeq: boolean read FShowAllTreeSeq write FShowAllTreeSeq;
    end;


implementation

uses
    Sysutils, UCompClusteringDefinition, UDlgOpPrmRegressionTree,
    UDlgBaseOperatorParameter, UConstConfiguration, UCalcStatDes,
  UDlgOpPrmRegTreeCART, UFrmOpViewRegTreeCART;

{ TMLGenRegTree }

procedure TMLGenRegTreeCART.GenCompInitializations;
begin
 FMLComp:= mlcRegression;
end;

function TMLGenRegTreeCART.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompRegTreeCART; 
end;

{ TMLCompRegTree }

function TMLCompRegTreeCART.getClassOperator: TClassOperator;
begin
 result:= TOpRegTreeCART;
end;

function TMLCompRegTreeCART.getGenericAttName: string;
begin
 result:= 'CRTRegTree';
end;

{ TOpRegTree }

function TOpRegTreeCART.CoreExecute: boolean;
begin
 if assigned(FCalc) then FCalc.Free();
 if assigned(FClusters) then FClusters.Free();
 result:= true;
 TRY
 FClusters:= TAttDiscrete.Create('_clus_reg_tree',self.WorkData.LstAtts[asTarget].Size);
 FCalc:= TCalcRegTreeCART.Create(self.WorkData.LstAtts[asTarget],self.WorkData.LstAtts[asInput],self.PrmOp as TPrmOpRegTreeCART,FClusters);
 FCalc.BuildClusters(self.WorkData.Examples);
 //construire la prdiction
 FCalc.Predict(self.EndoPred,self.ErrPred);
 //valuer la rgression
 self.Evaluation(self.WorkData.Examples);
 EXCEPT
 result:= false;
 END;
end;

destructor TOpRegTreeCART.destroy;
begin
 if assigned(FCalc) then FCalc.Free();
 if assigned(FClusters) then FClusters.Free();
 inherited destroy();
end;

procedure TOpRegTreeCART.evaluation(examples: TExamples);
var statEndo, statResidu: TCalcStatDesContinuous;
begin
 statEndo:= TCalcStatDesContinuous.Create(self.Endogenous,examples);
 statResidu:= TCalcStatDesContinuous.Create(self.ErrPred,examples);
 if (statEndo.TSS > 0.0)
  then R2:= 1.0 - statResidu.TSS / statEndo.TSS
  else R2:= 0.0;
 FreeAndNil(statEndo);
 FreeAndNil(statResidu);
end;

function TOpRegTreeCART.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpRegTreeCART; 
end;

function TOpRegTreeCART.getHTMLResultsSummary: string;
var s: string;
begin
 //rsultats gnraux
 s:= '<H3>Global results</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + format('%s<TD width="150">Endogenous attribute</TD><TD align="right" width="150"><B>%s</B></TD></TR>',[HTML_TABLE_COLOR_DATA_BLUE,Endogenous.Name]);
 s:= s + format('%s<TD>Examples</TD><TD align="right">%d</TD></TR>',[HTML_TABLE_COLOR_DATA_BLUE,self.WorkData.Examples.Size]);
 s:= s + format('%s<TD>R</TD><TD align="right">%.4f</TD></TR>',[HTML_TABLE_COLOR_DATA_BLUE,R2]);
 s:= s + '</table>';
 //suite, description de l'arbre
 result:= s + FCalc.GetHTMLResult();
end;

{ TPrmOpRegTree }

function TPrmOpRegTreeCART.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmRegTreeCART.CreateFromOpPrm(self);
end;

function TPrmOpRegTreeCART.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Tree Parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Max Number of leaves</TD><TD align="right">%d</TD></TR>',[FMaxNbClusters]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Min. size for split</TD><TD align="right">%d</TD></TR>',[FSizeBeforeSplit]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Max. depth</TD><TD align="right">%d</TD></TR>',[FMaxDepth]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>x-SE Rule</TD><TD align="right">%.4f</TD></TR>',[FSeRule]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Pruning set size</TD><TD align="right">%d %s</TD></TR>',[FPruningSetSize,'%']);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Show all tree sequence</TD><TD align="right">%d</TD></TR>',[ord(FShowAllTreeSeq)]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Rnd generator</TD><TD width=40 align="right">%d</TD></TR>',[ord(ModeRndGenerator)]);
 s:= s+'</table>';
 result:= s;
end;

procedure TPrmOpRegTreeCART.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 FSERule:= prmINI.ReadFloat(prmSection,'se_rule',FSERule);
 FPruningSetSize:= prmINI.ReadInteger(prmSection,'pruning_size',FPruningSetSize);
 FShowAllTreeSeq:= prmINI.ReadBool(prmSection,'show_all_sequence',FShowAllTreeSeq); 
end;

procedure TPrmOpRegTreeCART.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(FSERule,sizeof(FSERule));
 prmStream.ReadBuffer(FPruningSetSize,sizeof(FPruningSetSize));
 prmStream.ReadBuffer(FShowAllTreeSeq,sizeof(FShowAllTreeSeq));
end;

procedure TPrmOpRegTreeCART.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteFloat(prmSection,'se_rule',FSERule);
 prmINI.WriteInteger(prmSection,'pruning_size',FPruningSetSize);
 prmINI.WriteBool(prmSection,'show_all_sequence',FShowAllTreeSeq);
end;

procedure TPrmOpRegTreeCART.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(FSERule,sizeof(FSERule));
 prmStream.WriteBuffer(FPruningSetSize,sizeof(FPruningSetSize));
 prmStream.WriteBuffer(FShowAllTreeSeq,sizeof(FShowAllTreeSeq));
end;

procedure TPrmOpRegTreeCART.SetDefaultParameters;
begin
  inherited;
  //non modifiables
  FNormalization:= 0;//pas de normalisation de l'endodne
  FLeavesSizeAfterSplit:= 1;//refus si une feuille a moins de 1 individu
  FGoodnessThreshold:= 0.0;//construction "hurdling"
  //modifiables
  FMaxNbClusters:= 100;//nom max. de feuilles
  FSizeBeforeSplit:= 5;//moins de 5 individus, pas de segmentation
  FMaxDepth:= 10;//profondeur max.
  FShowAllTreeSeq:= FALSE;//pas tout montrer au dpart
  //nouveaux
  FSERule:= 1.0;
  FPruningSetSize:= 33;
end;

function TOpRegTreeCART.getNewForm: TForm;
begin
 result:= TfrmOpViewRegTreeCART.CreateFromOperator(self); 
end;

initialization
 Classes.RegisterClass(TMLGenRegTreeCART);
end.

