(********************************************************************)
(* UCompRegSimultaneous.pas - Copyright (c) 2011 Ricco RAKOTOMALALA *)
(********************************************************************)

{
@abstract(Rgressions simultanes - Rgressions croises)
@author(Ricco)
@created(01/07/2011)

La mthode est dcrite dans l'ouvrage "Pratique de la rgression linaire multiple" (section 3.6)
http://eric.univ-lyon2.fr/~ricco/cours/cours/La_regression_dans_la_pratique.pdf

}

unit UCompRegSimultaneous;

interface

USES
       Forms,Classes,IniFiles,
       UCompDefinition,
       UOperatorDefinition,
       UCompRegDefinition,
       UDatasetDefinition,
       UCalcStatDesContinuousPlus,
       UCalcRegLinMultiple,
       UCompManageDataset,
       Matrices, UCalcStatDes, UDatasetExamples;

TYPE

       //***********************************************************************
       //** gnrateur de composant, son rle est de crer une instance de composant qui est insr dans le diagramme de traitements
       //***********************************************************************
       TMLGenCompSimReg = class(TMLGenComp)
       protected
       //spcifier dans quel onglet de la palette doit apparatre l'icne -- ne pas surcharger dans les hritiers
       procedure   genCompInitializations(); override;
       public
       //spcifier quel est la classe du composant  gnrer sur le diagramme --  surcharger obligatoirement chez les hritiers
       function    getClassMLComponent: TClassMLComponent; override;
       end;

       //***********************************************************************
       //** classe du composant graphique insr dans le diagramme
       //***********************************************************************
       TMLCompSimReg = class(TMLCompLocalData)
       protected
       //spcifier l'oprateur de calcul --  surcharger obligatoirement
       function    getClassOperator: TClassOperator; override;
       end;

       //dclaration forward
       TCalcSimReg = class;

       //***********************************************************************
       //** classe oprateur de calcul : prend en charge le calcul et l'affichage des rsultats
       //***********************************************************************
       TOpSimReg = class(TOpLocalData)
       private
       //variables locales
       FCalc: TCalcSimReg;
       protected
       //spcifier la classe de paramtrage -- surcharge obligatoire
       function    getClassParameter: TClassOperatorParameter; override;
       //contrle de cohrence des attributs avant les calculs --  surcharger selon les mthodes
       function    checkAttributes(): boolean; override;
       //lancer effectivement les calculs -- surcharge obligatoire
       function    coreExecute(): boolean; override;
       public
       //libration de la mmoire alloue -- surcharge selon les mthodes, appel de la mthode anctre avec "inherited" obligatoire
       destructor  destroy(); override;
       //affichage des rsultats au format HTML -- surcharge obligatoire
       function    getHTMLResultsSummary(): string; override;
       end;

       //***********************************************************************
       //** classe paramtrage de l'oprateur de calcul : gre la bote de paramtrage et l'E/S des paramtres
       //** si la mthode de requiert pas de paramtres, il suffit de crer une classe vide hritire de "TOperatorParameter"
       //***********************************************************************
       TOpPrmSimReg = class(TOperatorParameter)
       private
       protected
       //cration de la bote de dialogue de paramtrage -- surcharge obligatoire (doit renvoyer "nil" si pas de bote  gnrer)
       function    CreateDlgParameters(): TForm; override;
       //spcifier les paramtres par dfaut de la mthode -- surcharge obligatoire
       procedure   SetDefaultParameters(); override;
       public
       end;

       TTabCoefs = array of array of double;

       //***********************************************************************
       //** classe de calcul
       //***********************************************************************
       TCalcSimReg = class
       private
       //affichage des matrices
       function   getHTMLMatrix(mat: PMatrix): string;
       //affichage des caractristiques globales
       function   getHTMLGlobalCar(): string;
       //affichage des coefficients
       function   getHTMLCoefficients(const tab: TTabCoefs; includingConst: boolean): string;
       //calculs des caractristiques des rgressions
       procedure  calcRegCarac();
       //calcul des coefs. standardiss
       procedure  calcBetaCoef();
       //calcul des coefs. non-standardiss
       procedure  calcCoef();
       //corrlations partielles
       procedure calcPartialCorrelation();
       public
       //nombre de variables
       p: integer;
       //nombre d'observations
       n: integer;
       //variables
       LstVar: TLstAttributes;
       //matrice de corrlation
       MatCor: PMatrix;
       //inverse de la matrice
       InvMatCor: PMatrix;
       //Matrice des corrlations partielles
       PartMatCor: PMatrix;
       //stats
       Stats: TLstCalcStatDesContinuous;
       //R, F,  p-value et sigma-error
       TabR2,TabF,TabPValue,TabSigErr: array of double;
       //tableau des coefs. standardiss
       TabBetaCoef: TTabCoefs;
       //tableau des coefs.
       TabCoef: TTabCoefs;
       //initialisation
       constructor create(prmAtts: TLstAttributes);
       //destruction
       destructor destroy(); override;
       //calculs - renvoie true si tout est ok
       function  calc(prmIndiv: TExamples): boolean;
       //affichage
       function   getHTML(): string;
       //vrifie si la matrice de corrlation a t inversible
       function   isNotSingular(): boolean;
       end;


implementation

USES
    UDatasetImplementation, UCalcMatrixToAttributes, UConstConfiguration,
    SYSUTILS, UStringAddBuffered, FMATH;


{ TMLGenCompSimReg }

procedure TMLGenCompSimReg.genCompInitializations;
begin
 FMLComp:= mlcRegression;
end;

function TMLGenCompSimReg.getClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSimReg;
end;

{ TMLCompSimReg }

function TMLCompSimReg.getClassOperator: TClassOperator;
begin
 result:= TOpSimReg;
end;

{ TOpSimReg }

function TOpSimReg.checkAttributes: boolean;
var ok: boolean;
begin
 ok:= (self.WorkData.LstAtts[asInput].Count > 1);
 ok:= ok and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue) = true);
 result:= ok;
end;

function TOpSimReg.coreExecute: boolean;
begin
 if assigned(FCalc) then FCalc.Free();
 FCalc:= TCalcSimReg.create(self.WorkData.LstAtts[asInput]);
 TRY
 result:= FCalc.calc(self.WorkData.Examples);
 EXCEPT
 FreeAndNil(FCalc);
 result:= false;
 END;
end;

destructor TOpSimReg.destroy;
begin
 if assigned(FCalc) then FreeAndNil(FCalc);
 inherited;
end;

function TOpSimReg.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSimReg;
end;

function TOpSimReg.getHTMLResultsSummary: string;
begin
 result:= '';
 if assigned(FCalc) then result:= FCalc.getHTML();
end;

{ TOpPrmSimReg }

function TOpPrmSimReg.CreateDlgParameters: TForm;
begin
 result:= nil;
end;

procedure TOpPrmSimReg.SetDefaultParameters;
begin
 //
end;

{ TCalcSimReg }

function TCalcSimReg.calc(prmIndiv: TExamples): boolean;
begin
 n:= prmIndiv.Size;
 result:= true;
 TRY
 Stats:= TLstCalcStatDesContinuous.Create(LstVar,prmIndiv);
 MatCor:= BuildMatVCV(prmIndiv,LstVar,vcvNormNormalized,Stats);
 Matrices.DimMatrix(InvMatCor,p,p);
 if (Matrices.InvMat(MatCor,1,p,InvMatCor) = MAT_OK)
  then
   Begin
    //suite des calculs...
    //matrice des corrlations partielles
    self.calcPartialCorrelation();
    //calcul des caractristiques de la rgression
    self.calcRegCarac();
    //calcul des coefficients standardiss
    self.calcBetaCoef();
    //calcul des coeffcients non standardiss
    self.calcCoef();
   end
  else
   begin
    DelMatrix(InvMatCor,p,p);
    InvMatCor:= nil;
   end;
 EXCEPT
 result:= false;
 END;
end;

procedure TCalcSimReg.calcBetaCoef;
var k,l: integer;
begin
 for l:= 1 to p do
  for k:= 1 to p do
   if (InvMatCor^[l]^[l] > 0.0) then TabBetaCoef[k,l]:= -1.0 * InvMatCor^[k]^[l] / InvMatCor^[l]^[l]; 
end;

procedure TCalcSimReg.calcCoef;
var k,l: integer;
    sk,sl: double;
    sum: double;
begin
 for l:= 1 to p do
  begin
   sl:= (Stats.Stat(pred(l)) as TCalcStatDesContinuous).StdDev;
   sum:= 0.0;
   for k:= 1 to p do
    begin
     sk:= (Stats.Stat(pred(k)) as TCalcStatDesContinuous).StdDev;
     if (sk > 0.0)
      then TabCoef[k,l]:= TabBetaCoef[k,l] * sl / sk
      else TabCoef[k,l]:= 0.0;
     //pour la constante
     if (k <> l)
      then sum:= sum + TabCoef[k,l] * (Stats.Stat(pred(k)) as TCalcStatDesContinuous).Average;
    end;
   TabCoef[0,l]:= (Stats.Stat(pred(l)) as TCalcStatDesContinuous).Average - sum;
  end;
end;

procedure TCalcSimReg.calcPartialCorrelation;
var i,j: integer;
begin
 Matrices.DimMatrix(PartMatCor,p,p);
 for i:= 1 to p do
  for j:= 1 to p do
   begin
    if (InvMatCor^[i]^[i] > 0.0) and (InvMatCor^[j]^[j] > 0.0)
     then PartMatCor^[i]^[j]:= -1.0 * InvMatCor^[i]^[j] / sqrt(InvMatCor^[i]^[i] * InvMatCor^[j]^[j])
     else PartMatCor^[i]^[j]:= 0.0;
   end;
end;

procedure TCalcSimReg.calcRegCarac;
var j: integer;
    pm1,nmp,nm1: double;
begin
 //R
 for j:= 1 to p do
  if (InvMatCor^[j]^[j] > 0) then TabR2[j]:= 1.0 - 1.0 / InvMatCor^[j]^[j] else TabR2[j]:= 0.0;
 //valeurs intermdiaires
 pm1:= -1.0 + p;
 nm1:= -1.0 + n;
 nmp:= 1.0 * (n - p);
 //F
 if (pm1 > 0.0) and (nmp > 0.0)
  then
   begin
    for j:= 1 to p do TabF[j]:= (TabR2[j]/pm1)/((1.0 - TabR2[j])/nmp);
    //p-value
    for j:= 1 to p do TabPValue[j]:= PSnedecor(p - 1, n - p, TabF[j]);
   end;
 //sig.err
 if (nmp > 0.0)
  then for j:= 1 to p do
   if (InvMatCor[j,j] > 0.0) then TabSigErr[j]:= SQRT((nm1 * (Stats.Stat(pred(j)) as TCalcStatDesContinuous).Variance / InvMatCor[j,j]) / nmp);
end;

constructor TCalcSimReg.create(prmAtts: TLstAttributes);
begin
  inherited create();
  //connexion
  p:= prmAtts.Count;
  LstVar:= prmAtts;
  //tableaux
  //R, F,  p-value et sigma-error
  setlength(TabR2,p+1);
  setlength(TabF,p+1);
  setlength(TabPValue,p+1);
  setlength(TabSigErr,p+1);
  //tableau des coefs. standardiss
  setLength(TabBetaCoef,p+1,p+1);
  //tableau des coefs.
  setLength(TabCoef,p+1,p+1);//la ligne 0 sera utilise pour la constante
end;

destructor TCalcSimReg.destroy;
begin
  if assigned(Stats) then FreeAndNil(Stats);
  if assigned(MatCor) then DelMatrix(MatCor,p,p);
  if assigned(InvMatCor) then DelMatrix(InvMatCor,p,p);
  if assigned(PartMatCor) then DelMatrix(PartMatCor,p,p);
  //R, F,  p-value et sigma-error
  setlength(TabR2,0);
  setlength(TabF,0);
  setlength(TabF,0);
  setlength(TabSigErr,0);
  //tableau des coefs. standardiss
  setLength(TabBetaCoef,0,0);
  //tableau des coefs.
  setLength(TabCoef,0,0);
  inherited;
end;

function TCalcSimReg.getHTMLCoefficients(const tab: TTabCoefs;
  includingConst: boolean): string;
var k,l: integer;
    buf: TBufString;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT);
 buf.AddStr(HTML_TABLE_COLOR_DATA_BLUE + format('<TH %s width=120>&nbsp;</TH>',[HTML_BGCOLOR_HEADER_GRAY]) + format('<TH colspan=%d>Dependent variables</TH></TR>',[p]));
 buf.AddStr(HTML_TABLE_COLOR_DATA_BLUE + format('<TH %s>Indep. variables</TH>',[HTML_BGCOLOR_HEADER_GRAY]));
 for l:= 1 to p do buf.AddStr(format('<TH width=70>%s</TH>',[LstVar.Attribute[pred(l)].Name]));
 buf.AddStr('</TR>');
 //la constante ???
 if (includingConst = true)
  then
   begin
    buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY + format('<TH %s>Intercept</TH>',[HTML_BGCOLOR_HEADER_GRAY]));
    for l:= 1 to p do buf.AddStr(format('<TD align=right>%.6f</TD>',[tab[0,l]]));
    buf.AddStr('</TR>');
   end;
 //les autres coefficients...
 for k:= 1 to p do
  begin
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY + format('<TH %s>%s</TH>',[HTML_BGCOLOR_HEADER_GRAY,LstVar.Attribute[pred(k)].Name]));
   for l:= 1 to p do
    begin
     if (l = k)
      then buf.AddStr('<TD align=right>-</TD>')
      else buf.AddStr(format('<TD align=right>%.6f</TD>',[tab[k,l]]))
    end;
   buf.AddStr('</TR>');
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TCalcSimReg.getHTML: string;
var s: string;
begin
 s:= '';
 if (self.isNotSingular() = false)
  then
   begin
    s:= s + '<H3>Correlation matrix is singular - Calculation canceled</H3>';
    s:= s + '<H4>Correlation matrix</H4>';
    s:= s + self.getHTMLMatrix(MatCor);
   end
  else
   begin
    s:= s + '<H3>Regressions characteristics</H3>';
    s:= s + '<H4>Global characteristics</H4>';
    s:= s + self.getHTMLGlobalCar();
    s:= s + '<H4>Standardized coefficients</H4>';
    s:= s + self.getHTMLCoefficients(TabBetaCoef,false);
    s:= s + '<H4>Unstandardized coefficients</H4>';
    s:= s + self.getHTMLCoefficients(TabCoef,true);
    s:= s + '<H3>Matrices</H3>';
    s:= s + '<H4>Correlation matrix</H4>';
    s:= s + self.getHTMLMatrix(MatCor);
    s:= s + '<H4>Partial Correlation matrix</H4>';
    s:= s + self.getHTMLMatrix(PartMatCor);
    s:= s + '<H4>Inverse of the Correlation matrix</H4>';
    s:= s + self.getHTMLMatrix(InvMatCor);
   end;
 result:= s;
end;

function TCalcSimReg.getHTMLGlobalCar: string;
var j: integer;
    s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TH %s width=120>&nbsp;</TH>',[HTML_BGCOLOR_HEADER_GRAY]) + format('<TH colspan=%d>Dependent variables</TH></TR>',[p]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TH %s>Characteristics</TH>',[HTML_BGCOLOR_HEADER_GRAY]);
 for j:= 1 to p do s:= s + format('<TH width=70>%s</TH>',[LstVar.Attribute[pred(j)].Name]);
 s:= s + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TH>R</TH>';
 for j:= 1 to p do s:= s + format('<TD align=right>%.4f</TD>',[sqrt(TabR2[j])]);
 s:= s + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TH>R</TH>';
 for j:= 1 to p do s:= s + format('<TD align=right>%.4f</TD>',[TabR2[j]]);
 s:= s + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TH>F</TH>';
 for j:= 1 to p do s:= s + format('<TD align=right>%.4f</TD>',[TabF[j]]);
 s:= s + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH>Pr(>F)(DF1=%d,DF2=%d)</TH>',[p-1,n-p]);
 for j:= 1 to p do s:= s + format('<TD align=right %s>%.6f</TD>',[colorCodeForPValue(TabPValue[j]),TabPValue[j]]);
 s:= s + '</TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TH>Std. Err. Estimate</TH>';
 for j:= 1 to p do s:= s + format('<TD align=right>%.4f</TD>',[TabSigErr[j]]);
 s:= s + '</TR>';
 s:= s + '</table>';
 result:= s; 
end;

function TCalcSimReg.getHTMLMatrix(mat: PMatrix): string;
var buf: TBufString;
    i,j: integer;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(HTML_HEADER_TABLE_RESULT+HTML_TABLE_COLOR_HEADER_GRAY+'<TD width=120>Attribute</TD>');
 for j:= 1 to p do buf.AddStr(format('<TD width=70  align=right>%s</TD>',[LstVar.Attribute[pred(j)].Name]));
 buf.AddStr('</TR>');
 for i:= 1 to p do
  begin
   buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY + format('<TD %s>%s</TD>',[HTML_BGCOLOR_HEADER_GRAY,LstVar.Attribute[pred(i)].Name]));
   for j:= 1 to p do buf.AddStr(format('<TD align=right>%.6f</TD>',[mat^[i]^[j]]));
   buf.AddStr('</TR>');
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TCalcSimReg.isNotSingular: boolean;
begin
 result:= (InvMatCor <> nil);
end;

initialization
 Classes.RegisterClass(TMLGenCompSimReg);
end.
