(******************************************************************)
(* UCompRegAssessment.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Comparaison valeurs observes et prdites pour la rgression)
@author(Ricco)
@created(07/04/2009)

Le calcul est assez simple, on peut choisir le sous-ensemble d'observations sur
lequel nous effectuons la comparaison

}

unit UCompRegAssessment;

interface

USES
    Contnrs, Forms, Classes, IniFiles,
    UCompDefinition, UOperatorDefinition, UCompManageDataset,
    UDatasetExamples, UDatasetDefinition,
    UCompSpvAssesTestSet;

TYPE
    //
    TMLGenRegAssessment = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    public
    function    GetClassMLComponent: TClassMLComponent; override;
    end;

    //
    TMLCompRegAssessment = class(TMLCompLocalData)
    protected
    function    getClassOperator: TClassOperator; override;
    end;

    //
    TOpRegAssessment = class(TOpLocalData)
    private
    //les individus utiliss pour l'valuation
    FUsedExamples: TExamples;
    //liste des comparaisons
    FLstEvaluation: TObjectList;
    protected
    //classe de paramtres assoicie
    function    getClassParameter: TClassOperatorParameter; override;
    //checker les attributs  utiliser
    function    CheckAttributes(): boolean; override;
    //checker les exemples  utiliser -- les rcuprer au passage
    function   CheckExamples(): boolean; override;
    //lancer le calcul des matrices de confusion
    function    CoreExecute(): boolean; override;
    public
    //prdicteur par dfaut (la moyenne de l'endogne)
    DefPredictor: double;
    //constructeur
    constructor Create(AOwner: TObject); override;
    //destructeur
    destructor  destroy(); override;
    //envoyer le rsultat des calculs
    function    getHTMLResultsSummary(): string; override;
    end;

    //
    TOpParameterRegAssessment = class(TOperatorParameter)
    protected
    procedure   SetDefaultParameters(); override;
    function    CreateDlgParameters(): TForm; override;
    public
    //utiliser les observations TRAIN ou TEST (si elles existent !!!)
    SelExamples: TEnumEvaluationSet;
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    function    getHTMLParameters(): string; override;
    end;

    //classe de calcul - une valuation
    TOneRegAssessment = class(TObject)
    public
    //les indicateurs
    SCT, SCR, Ratio: double;
    //les variables
    FYObs, FYPred: TAttribute;
    //Oprateur de calcul
    FOperator: TOpRegAssessment;
    //constructeur
    constructor create(prmOperator: TOpRegAssessment; prmYObs,prmYPred: TAttribute; examples: TExamples = nil);
    //calculer
    procedure   compute(examples: TExamples);
    //affichage HTML
    function    getDescriptionHTML(): string;
    end;

implementation

USES
    Windows, Sysutils,
    ULogFile,
    UDatasetImplementation, UCalcStatDes, UConstConfiguration,
  UDlgOpPrmRegAssessment;

CONST
   //en-dessous, on refuse d'valuer
   SPV_ASSES_MIN_EVALUATION_SET_SIZE = 5;


{ TMLGenRegAssessment }

procedure TMLGenRegAssessment.GenCompInitializations;
begin
 FMLComp:= mlcRegression;
end;

function TMLGenRegAssessment.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompRegAssessment;
end;

{ TMLCompRegAssessment }

function TMLCompRegAssessment.getClassOperator: TClassOperator;
begin
 result:= TOpRegAssessment;
end;

{ TOpRegAssessment }

function TOpRegAssessment.CheckAttributes: boolean;
var ok: boolean;
begin
 //un seul target et il doit tre discret
 ok:= (self.WorkData.LstAtts[asTarget].Count = 1) and (self.WorkData.LstAtts[asTarget].Attribute[0].isCategory(caContinue));
 //et tous les input doivent tre discrets -- le test de conformit sera ralis lors du calcul
 ok:= ok and ((self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue)));
 //and then...
 result:= ok;
end;

function TOpRegAssessment.CheckExamples: boolean;
var tps: cardinal;
begin
 if assigned(FUsedExamples) then FUsedExamples.Free();
 tps:= GetTickCount();
 TraceLog.WriteToLogFile('[ASSES-TEST-REG] begin check examples');
 //dterminer les individus  utiliser
 case (self.PrmOp as TOpParameterRegAssessment).SelExamples of
  evalSetTest:
   begin
    //complmentaire des individus slectionn  la taille de la base --> c'est bien les individus non-slectionns
    FUsedExamples:= self.WorkData.Examples.getComplementaire(self.WorkData.LstAtts[asAll].Size);
   end
  else
   begin
    FUsedExamples:= TExamples.Create(self.WorkData.Examples.Size);
    FUsedExamples.Copy(self.WorkData.Examples);
   end;
 end;
 tps:= GetTickCount() - tps;
 TraceLog.WriteToLogFile(format('[ASSES-TEST-REG] end check examples = %d ms.',[tps]));
 //tester si la taille convient
 result:= (FUsedExamples.Size >= SPV_ASSES_MIN_EVALUATION_SET_SIZE);
end;

function TOpRegAssessment.CoreExecute: boolean;
var Y,X: TAttribute;
    j: integer;
    st: TCalcStatDesContinuous;
begin
 result:= true;
 FLstEvaluation.Clear();
 Y:= self.WorkData.LstAtts[asTarget].Attribute[0];
 //attention, calcul de la moyenne sur les individus "selected" toujours
 st:= TCalcStatDesContinuous.Create(Y,self.WorkData.Examples);
 DefPredictor:= st.Average;
 st.Free;
 //pour chaque variable INPUT cd. les prdictions  valuer
 TRY
 for j:= 0 to pred(self.WorkData.LstAtts[asInput].Count) do
  begin
   X:= self.WorkData.LstAtts[asInput].Attribute[j];
   FLstEvaluation.Add(TOneRegAssessment.create(self,Y,X,FUsedExamples));
  end;
 EXCEPT
 result:= false;
 END;
end;

constructor TOpRegAssessment.Create(AOwner: TObject);
begin
 inherited;
 FLstEvaluation:= TObjectList.Create(TRUE);
end;

destructor TOpRegAssessment.destroy;
begin
 FLstEvaluation.Free();
 if assigned(FUsedExamples) then FUsedExamples.Free();
 inherited;
end;

function TOpRegAssessment.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpParameterRegAssessment;
end;

function TOpRegAssessment.getHTMLResultsSummary: string;
var s: string;
    k: integer;
begin
 s:= format('<br>Dataset size : %d<br><br>',[FUsedExamples.Size]);
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + '<TH colspan=2>&nbsp;</TH><TH colspan=2>Error Sum of Squares<TH width=80>&nbsp;</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + '<TH>Observed</TH><TH>Predicted</TH><TH width=80>Default (Mean)</TH><TH width=80>Model</TH><TH width=80>Pseudo-R2</TH></TR>';
 for k:= 0 to pred(FLstEvaluation.Count) do
  s:= s + (FLstEvaluation.Items[k] as TOneRegAssessment).getDescriptionHTML();
 s:= s + '</Table>';
 result:= s;
end;

{ TOpParameterRegAssessment }

function TOpParameterRegAssessment.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmRegAssessment.CreateFromOpPrm(self);
end;

function TOpParameterRegAssessment.getHTMLParameters: string;
begin
 result:= format('<b>Used dataset : </b>%s examples',[SPV_ASSES_EVALUATION_SET[SelExamples]]);
end;

procedure TOpParameterRegAssessment.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 SelExamples:= TEnumEvaluationSet(prmINI.ReadInteger(prmSection,'examples',ord(SelExamples)));
end;

procedure TOpParameterRegAssessment.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(SelExamples,sizeof(SelExamples));
end;

procedure TOpParameterRegAssessment.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'examples',ord(SelExamples));
end;

procedure TOpParameterRegAssessment.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(SelExamples,sizeof(SelExamples));
end;

procedure TOpParameterRegAssessment.SetDefaultParameters;
begin
 //par dfaut, on prend comme "train set" les "selected"
 SelExamples:= evalSetLearning;
end;

{ TOneRegAssessment }

procedure TOneRegAssessment.compute(examples: TExamples);
var i,example: integer;
    scd,scm: double;
    defPred: double;
begin
 //rcuprer la prdiction par dfaut
 defPred:= FOperator.DefPredictor;
 //calcul
 scd:= 0.0;
 scm:= 0.0;
 for i:= 1 to examples.Size do
  begin
   example:= examples.Number[i];
   scm:= scm + SQR(FYObs.cValue[example]-FYPred.cValue[example]);
   scd:= scd + SQR(FYObs.cValue[example]-defPred);
  end;
 //rcuprer les valeurs
 SCT:= scd;
 SCR:= scm;
 //
 if (scd > 0.0)
  then Ratio:= scm / scd
  else Ratio:= 1.0;
end;

constructor TOneRegAssessment.create(prmOperator: TOpRegAssessment; prmYObs, prmYPred: TAttribute;
  examples: TExamples);
begin
 inherited create();
 FYObs:= prmYObs;
 FYPred:= prmYPred;
 FOperator:= prmOperator;
 if (examples <> nil) then self.compute(examples);
end;

function TOneRegAssessment.getDescriptionHTML: string;
begin
 result:= HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD>%s</TD><TD align="center">%.4f</TD><TD align="center">%.4f</TD><TD align="center">%.4f</TD></TR>',
          [FYObs.Name,FYPred.Name,SCT,SCR,1.0-Ratio]);  
end;

initialization
 Classes.RegisterClass(TMLGenRegAssessment);
end.
