(**********************************************************************)
(* UCompSpvPosteriorProba.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Classe de calcul des scores pour toutes les modalits de l'attribut classe)
@author(Ricco)
@created(11/12/2008)

Calcule la valeur du score (proba d'affectation) d'un individu relativement  toutes les valeurs de l'attribut classe.

(1) Prdcesseur : forcment (!) un composant apprentissage supervis
(2) Construit des variables qui correspondent au score d'affectation, calcul sur tous les individus et l'ajoute  l'ensemble des donnes
(3) Successeur : n'importe quel composant

On l'intgre dans la catgorie de composants le SCORING.

}

unit UCompSpvPosteriorProba;

interface

USES
        Forms, Contnrs, Classes,
        IniFiles,
        UCompDefinition,
        UCompManageDataset,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UCompSpvLDefinition;

TYPE
        {gnrateur de composant}
        TMLGenComputePosterior = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant -- gnre un nouveau composant  la vole}
        TMLComputePosterior = class(TMLCompLocalData)
        protected
        function    getClassOperator: TClassOperator; override;
        public
        //verifier  qui on peut le connecter
        class function isConnectableTo(destination: TMLComponent): boolean; override;
        end;

        {oprateur}
        TOpComputePosterior = class(TOpLocalData)
        private
        //composant meta supervised prcdent -- qui va fournir le Posterior
        FCompMetaSpv: TMLCompMetaSpvLearning;
        protected
        //classe de paramtres assoicie
        function    getClassParameter: TClassOperatorParameter; override;
        {lancer le rafrachissement des valeurs de la variable}
        function    CoreExecute(): boolean; override;
        public
        constructor Create(AOwner: TObject); override;
        {envoyer le rsultat des calculs}
        function    getHTMLResultsSummary(): string; override;
        end;

        {paramtrage de l'oprateur}
        TOpPrmComputePosterior = class(TOperatorParameter)
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        end;

implementation

USES
        Sysutils, UCalcSpvStructScore;

{ TMLGenComputePosterior }

procedure TMLGenComputePosterior.GenCompInitializations;
begin
 FMLComp:= mlcSpvScoring;
end;

function TMLGenComputePosterior.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLComputePosterior;
end;

{ TMLComputePosterior }

function TMLComputePosterior.getClassOperator: TClassOperator;
begin
 result:= TOpComputePosterior;
end;

class function TMLComputePosterior.isConnectableTo(
  destination: TMLComponent): boolean;
begin
 //connectable seulement si le prdcesseur est une MetaSpvLearning
 result:= (destination is TMLCompMetaSpvLearning);
end;

{ TOpComputePosterior }

function TOpComputePosterior.CoreExecute: boolean;
var k,i: integer;
    classAttribute: TAttDiscrete;
    opMetaSpv: TOpMetaSpvLearning;
    attName: string;
    newAtt: TAttribute;
    postProba: TTabScore;  
begin
 result:= false;
 if (FCompMetaSpv.OutputData.LstAtts[asTarget].Count = 1)
  then
   begin
    TRY
    classAttribute:= FCompMetaSpv.OutputData.LstAtts[asTarget].Attribute[0] as TAttDiscrete;
    //vider la liste des attributs gnrs puis les re-gnrer
    GenAtts.Clear();
    for k:= 1 to classAttribute.nbValues do
     begin
      attName:= 'PP_'+classAttribute.LstValues.getDescription(k)+'_'+IntToStr((self.MLOwner as TMLComponent).Number);
      newAtt:= TAttContinue.Create(attName,classAttribute.Size);
      GenAtts.Add(newAtt);
     end;
    //pour chaque observation, calculer le score
    opMetaSpv:= FCompMetaSpv.Operator as TOpMetaSpvLearning;
    postProba:= TTabScore.create(classAttribute);
    for i:= 1 to classAttribute.Size do
     begin
      //calculer les probas a posteriori, puis les affecter aux variables en sortie
      opMetaSpv.getScore(i,postProba);
      for k:= 1 to classAttribute.nbValues do GenAtts.Attribute[k-1].cValue[i]:= postProba[k];
     end;
    postProba.Free();
    //si tout a bien fonctionn
    result:= true;
    EXCEPT
    result:= false;
    END;
   end;
end;

constructor TOpComputePosterior.Create(AOwner: TObject);
begin
 inherited;
 //le composant meta supervis, si le cast est mauvais c'est qu'il y a problme -- le prdesseur doit tre un apprentissage supervis
 FCompMetaSpv:= (MLOwner as TMLComponent).Predecessor as TMLCompMetaSpvLearning;
end;

function TOpComputePosterior.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmComputePosterior; 
end;

function TOpComputePosterior.getHTMLResultsSummary: string;
var s: string;
    k: integer;    
begin
 s:= '<b>Generated attributes :</b><br>';
 for k:= 0 to GenAtts.Count-1 do
  s:= s + GenAtts.Attribute[k].Name + '</br>';
 result:= s;
end;

{ TOpPrmComputePosterior }

function TOpPrmComputePosterior.CreateDlgParameters: TForm;
begin
 result:= nil;
end;

procedure TOpPrmComputePosterior.SetDefaultParameters;
begin
 //rien...
end;

initialization
 RegisterClass(TMLGenComputePosterior);
end.
