(******************************************************************************)
(* UCompSpvScoringPrecRecallCurve.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(******************************************************************************)

{
@abstract(Classe de calcul d'une courbe Precision-Recall)
@author(Ricco)
@created(25/09/2009)

Dans la catgorie des SCORING.

Calcul des courbes Precision-Recall  partir de 2 types d'informations :

Dans ** DEFINE STATUS **
TARGET : l'attribut de rfrence
INPUT  : les attributs SCORE qui permettent d'effectuer le classement des individus

** Comme paramtre de la mthode **
POSITIVE CLASS VALUE : la modalit de la classe qu'il faut considrer comme "les positifs"

N.B: Il peut y avoir plus variables dans INPUT, on peut ainsi construire simultanment
plusieurs courbes Precision-Recall et effectuer des comparaisons.

Par attribut de ciblage (scoring), on conserve les infos
(a) la valeur du recall
(b) la valeur de la prcision

}

unit UCompSpvScoringPrecRecallCurve;

interface

USES
   Forms,
   UCompDefinition,
   UCompManageDataset,
   UOperatorDefinition,
   UDatasetDefinition,
   UDatasetImplementation,
   UCompSpvLDefinition,
   UCompSpvScoring,
   UCompSpvScoringLift,
   UDatasetExamples;

TYPE
   {gnrateur de composant}
   TMLGenSpvScoringPrecRecall = class(TMLGenComp)
   protected
   procedure   GenCompInitializations(); override;
   public
   function    GetClassMLComponent: TClassMLComponent; override;
   end;

   {composant}
   TMLSpvScoringPrecRecall = class(TMLCompLocalData)
   protected
   function    getClassOperator: TClassOperator; override;
   end;

   {oprateur}
   TOpSpvScoringPrecRecall = class(TOpSpvScoringLIFT)
   protected
   //nombre d'exemples
   FNbExamples: integer;
   //tableau des precision
   FTabPrecision: TTabStoreScoringValues;
   //classe de paramtrage
   function    getClassParameter: TClassOperatorParameter; override;
   //tableaux
   procedure   prepareTabs(); override;
   //dtruire les tableaux de calculs
   procedure   destroyTabs(); override;
   //calculer la courbe pour l'attribut nj
   procedure   computeCurve(numAtt: integer; examples: TExamples); override;
   //obtenir la nouvelle fentre d'affichage
   function    getNewForm(): TForm; override;
   public
   //envoyer le rsultat des calculs
   function    getHTMLResultsSummary(): string; override;
   //proprits
   property    TabPrecision: TTabStoreScoringValues read FTabPrecision;
   property    NbExamples: integer read FNbExamples;
   end;

   {paramtrage}
   TOpPrmSpvScoringPrecRecall = class(TOpPrmSpvScoringLIFT)
   end;


implementation

USES
  Classes, SysUtils, UConstConfiguration, UExperiments, UFrmSpvScoringRoc,
  UFrmBaseOperator, UFrmSpvScoringPrecRecall;


{ TMLGenSpvScoringPrecRecall }

procedure TMLGenSpvScoringPrecRecall.GenCompInitializations;
begin
 FMLComp:= mlcSpvScoring;
end;

function TMLGenSpvScoringPrecRecall.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLSpvScoringPrecRecall;
end;

{ TMLSpvScoringPrecRecall }

function TMLSpvScoringPrecRecall.getClassOperator: TClassOperator;
begin
 result:= TOpSpvScoringPrecRecall;
end;

{ TOpSpvScoringPrecRecall }

procedure TOpSpvScoringPrecRecall.computeCurve(numAtt: integer;
  examples: TExamples);
var srtEx: TExamples;
    attScore: TAttribute;
    i,curObsPos: integer;
    nbAdd: integer;//grer la frquence d'ajout dans le graphique
begin
 attScore:= self.WorkData.LstAtts[asInput].Attribute[numAtt];
 //rcuprer les individus et les trier selon la variable de score
 srtEx:= TExamples.Create(examples.Size);
 srtEx.Copy(examples);
 srtEx.QuickSortBy(attScore);//attention le tri est ascendant, il faut prendre dans l'ordre inverse !!!
 //enquiller les valeurs
 curObsPos:= 0;
 nbAdd:= 0;
 for i:= srtEx.Size downto 1 do
  begin
   inc(nbAdd);
   //est-ce un positif ?
   if (FClassAttribute.dValue[srtEx.Number[i]] = FIdPositiveClass)
    //oui, c'est la classe que l'on traite comme "positif"
    then inc(curObsPos);
   //rcuprer les valeurs
   FTabTVP[i,numAtt]:= (1.0 * curObsPos) / (1.0 * FNbAllPos);
   FTabPrecision[i,numAtt]:= (1.0 *  curObsPos) / (1.0 * nbAdd);
  end;

 //new -- 09/12/2006 -- se donner la possibilit de sauver des rsultats dan sun fichier  part
 {$IFDEF WRITE_EXPE_RESULTS}
 writelnExpeResult(FTabAUC[numAtt]);
 {$ENDIF}

 //vider
 srtEx.Free();
end;

procedure TOpSpvScoringPrecRecall.destroyTabs;
begin
 if (FTabTVP <> nil) then Finalize(FTabTVP);
 if (FTabPrecision <> nil) then Finalize(FTabPrecision);
end;

function TOpSpvScoringPrecRecall.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSpvScoringPrecRecall;
end;

function TOpSpvScoringPrecRecall.getHTMLResultsSummary: string;
var s: string;
begin
 s:= '<H3>Precision-Recall Curve</H3>';
 //infos sur les effectifs
 s:= s+format('<b>Sample size : </b> %d<br>',[FUsedExamples.Size]);
 s:= s+format('<b>Positive examples : </b> %d<br>',[FNbAllPos]);
 //and then...
 result:= s;
end;

function TOpSpvScoringPrecRecall.getNewForm: TForm;
begin
 result:= TfrmOpViewSpvScoringPrecRecall.CreateFromOperator(self);
end;

procedure TOpSpvScoringPrecRecall.prepareTabs;
begin
 self.destroyTabs();
 //laborer les tableaux, de la mme taille que le nombre d'observations
 FNbExamples:= FUsedExamples.Size;
 setLength(FTabTVP,succ(NbExamples),self.WorkData.LstAtts[asInput].Count);
 setLength(FTabPrecision,succ(NbExamples),self.WorkData.LstAtts[asInput].Count);
end;

initialization
 RegisterClass(TMLGenSpvScoringPrecRecall);
end.
