(*********************************************************************************)
(* UCompSpvScoringReliabilityDiagram.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(*********************************************************************************)

{
@abstract(Classe de calcul du diagramme de fiabilit)
@author(Ricco)
@created(27/09/2009)

Dans la catgorie des SCORING.

Calcule des courbes LIFT  partir de 2 types d'informations :

Dans ** DEFINE STATUS **
TARGET : l'attribut de rfrence
INPUT  : les attributs SCORE qui permettent d'effectuer le classement des individus

** Comme paramtre de la mthode **
POSITIVE CLASS VALUE : la modalit de la classe qu'il faut considrer comme "les positifs"

N.B: Il peut y avoir plus variables dans INPUT, on peut ainsi construire simultanment
plusieurs courbes diagramme de fiabilit et effectuer des comparaisons.

Par attribut de ciblage (scoring), on conserve 2 types d'infos

(a) la valeur du seuil du score qui dfinit le groupe
(b) la proportion de positifs dans le groupe
(c) la moyenne des scores dans le groupe

}


unit UCompSpvScoringReliabilityDiagram;

interface

USES
        Forms,
        UCompDefinition,
        UCompManageDataset,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UCompSpvLDefinition,
        UCompSpvScoring,
        UCompSpvScoringLift,
        UDatasetExamples;

TYPE
   {gnrateur de composant}
   TMLGenSpvScoringRelDiagram = class(TMLGenComp)
   protected
   procedure   GenCompInitializations(); override;
   public
   function    GetClassMLComponent: TClassMLComponent; override;
   end;

   {composant}
   TMLSpvScoringRelDiagram = class(TMLCompLocalData)
   protected
   function    getClassOperator: TClassOperator; override;
   end;

   {oprateur}
   TOpSpvScoringRelDiagram = class(TOpSpvScoringLIFT)
   protected
   //nombre de groupes
   FNbGroups: integer;
   //tableau des proportions des positifs
   FTabPPos: TTabStoreScoringValues;
   //tableau des moyennes des scores
   FTabMoyScore: TTabStoreScoringValues;
   //classe de paramtrage
   function    getClassParameter: TClassOperatorParameter; override;
   //tableaux
   procedure   prepareTabs(); override;
   //dtruire les tableaux de calculs
   procedure   destroyTabs(); override;
   //calculer la courbe pour l'attribut nj
   procedure   computeCurve(numAtt: integer; examples: TExamples); override;
   //obtenir la nouvelle fentre d'affichage
   function    getNewForm(): TForm; override;
   public
   //envoyer le rsultat des calculs
   function    getHTMLResultsSummary(): string; override;
   //proprits
   property    TabPPos: TTabStoreScoringValues read FTabPPos;
   property    TabMoyScore: TTabStoreScoringValues read FTabMoyScore;
   property    NbGroups: integer read FNbGroups;
   end;

   {paramtrage}
   TOpPrmSpvScoringRelDiagram = class(TOpPrmSpvScoringLIFT)
   end;

implementation

uses
    Math, Classes, Sysutils, UConstConfiguration, UFrmBaseOperator,
  UFrmSpvScoringRelDiagram;

{ TMLGenSpvScoringRelDiagram }

procedure TMLGenSpvScoringRelDiagram.GenCompInitializations;
begin
 FMLComp:= mlcSpvScoring;
end;

function TMLGenSpvScoringRelDiagram.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLSpvScoringRelDiagram;
end;

{ TMLSpvScoringRelDiagram }

function TMLSpvScoringRelDiagram.getClassOperator: TClassOperator;
begin
 result:= TOpSpvScoringRelDiagram;
end;

{ TOpSpvScoringRelDiagram }

procedure TOpSpvScoringRelDiagram.computeCurve(numAtt: integer;
  examples: TExamples);
var srtEx: TExamples;
    attScore: TAttribute;
    i,k,example: integer;
    v,mins,maxs,ranges,curBorne: double;
    sumPos, nbEx, sumScore: double;

    procedure closeGroup();
    begin
     FTabScore[k+1][numAtt]:= curBorne;
     if (nbEx > 0.0) then FTabPPos[k][numAtt]:= sumPos/nbEx else FTabPPos[k][numAtt]:= 0.0;
     if (nbEx > 0.0) then FTabMoyScore[k][numAtt]:= sumScore/nbEx else FTabMoyScore[k][numAtt]:= 0.0;
     //r-init.
     nbEx:= 0.0;
     sumPos:= 0.0;
     sumScore:= 0.0;
    end;

begin
 attScore:= self.WorkData.LstAtts[asInput].Attribute[numAtt];
 //rcuprer les individus et les trier selon la variable de score
 srtEx:= TExamples.Create(examples.Size);
 srtEx.Copy(examples);
 srtEx.QuickSortBy(attScore);//attention le tri est ascendant !!!
 //calculer le min et le max, puis la largeur de chaque groupe
 mins:= attScore.cValue[srtEx.Number[1]];
 FTabScore[0][numAtt]:= mins;
 maxs:= attScore.cValue[srtEx.Number[srtEx.Size]];
 ranges:= (maxs - mins) / (1.0 * NbGroups);
 curBorne:= mins + ranges;
 k:= 0;
 sumPos:= 0.0; nbEx:= 0.0; sumScore:= 0.0;
 for i:= 1 to srtEx.Size do
  begin
   example:= srtEx.Number[i];
   v:= attScore.cValue[example];
   if (k < pred(NbGroups)) and (v > curBorne) // si c'est le dernier, pas besoin de rentrer dedans
    then
     begin
      closeGroup();
      inc(k); curBorne:= curBorne + ranges;
     end;
    nbEx:= nbEx + 1.0;
    if (FClassAttribute.dValue[srtEx.Number[i]] = FIdPositiveClass) then sumPos:= sumPos + 1.0;
    sumScore:= sumScore + v;
  end;
 //last
 closeGroup();
 //vider
 srtEx.Free();
end;

procedure TOpSpvScoringRelDiagram.destroyTabs;
begin
 if (FTabPPos <> nil) then FTabPPos:= nil;
 if (FTabMoyScore <> nil) then FTabMoyScore:= nil;
 if (FTabScore <> nil) then FTabScore:= nil;
end;

function TOpSpvScoringRelDiagram.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSpvScoringRelDiagram;
end;

function TOpSpvScoringRelDiagram.getHTMLResultsSummary: string;
var s: string;
    j,k: integer;
begin
 s:= '<H3>Reliability Diagram</H3>';
 for j:= 0 to pred(self.WorkData.LstAtts[asInput].Count) do
  begin
   s:= s + format('<H5>Reliability diagram for "%s"</H5>',[self.WorkData.LstAtts[asInput].Attribute[j].Name]);
   s:= s + HTML_HEADER_TABLE_RESULT;
   s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=100>Scores</TH><TH width=100>Avg(score)</TH><TH width=100>Empirical Prob.</TH></TR>';
   for k:= 1 to NbGroups do
    begin
     s:= s + HTML_TABLE_COLOR_DATA_GRAY + Format('<TD align=center>%.4f - %.4f</TD><TD align=center>%.4f</TD><TD align=center>%.4f</TD></TR>',
                                                [FTabScore[k-1][j],FTabScore[k][j],FTabMoyScore[k-1][j],FTabPPos[k-1][j]]);
    end;
   s:= s + '</table>';
  end;

 result:= s;
end;

function TOpSpvScoringRelDiagram.getNewForm: TForm;
begin
 result:= TfrmSpvScoringRelDiagram.CreateFromOperator(self);
end;

procedure TOpSpvScoringRelDiagram.prepareTabs;
begin
 self.destroyTabs();
 //calculer le nombre de groupes
 FNbGroups:= Math.Max(3,TRUNC(LN(FUsedExamples.Size)));
 //prparer les tableaux
 setLength(FTabScore,succ(NbGroups),self.WorkData.LstAtts[asInput].Count);
 setLength(FTabPPos,NbGroups,self.WorkData.LstAtts[asInput].Count);
 setLength(FTabMoyScore,NbGroups,self.WorkData.LstAtts[asInput].Count);
end;

initialization
 RegisterClass(TMLGenSpvScoringRelDiagram);
end.
