(**********************************************************************)
(* UCompMetaSpvCSCommitte.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Classe Abstraite - Agrgation d'Apprentissage tenant compte des cots de mauvais classement)
@author(Ricco)
@created(29/12/2008)

La correction est ralise uniquement sur la rgle d'affectation pour chaque classifieur.
Aprs on les combine d'une manire qui reste  dfinir selon les mthodes.

}

unit UCompMetaSpvCSCommitte;

interface

USES
        Contnrs,
        Forms,Classes,IniFiles,
        UDatasetExamples,
        UCompSpvLDefinition,
        UOperatorDefinition,
        UDatasetDefinition,UCalcDistribution,
        UCalcSpvStructScore,
        UCalcRndGenerator,
        UCompMetaSpvGenericCostSensitive;

TYPE
        {Oprateur associ}
        TOpMetaSpvCSCommitte = class(TOpMetaSpvCostSensitive)
        private
        {liste des clasiffieurs}
        FClassifiers: TObjectList;
        {tableau des taux d'erreur}
        FTabErrors: array of double;
        //procdure d'affectation pour un classifieur
        procedure    ApplyOneClassifier(numero: integer; example: integer; var response: TTypeDiscrete);
        protected
        //gnrateur de nombre alatoire
        FRnd: TRndGenerator;
        {individus sur lesquels est appliqu l'apprentissage}
        FInternalExamples: TExamples;
        {sauf cas contraire, tous les committees ont le mme type de paramtrage}
        function    getClassParameter: TClassOperatorParameter; override;
        {lancer les apprentissages successifs}
        procedure   RunLearning(); override;
        {vider la liste des classifieurs}
        procedure   ReInitialize(); override;
        {supprimer la liste}
        procedure   destroyClassifiers(); override;
        {prparer les donnes internes}
        procedure   prepareCalcData(); virtual;
        {supprimer les donnes internes}
        procedure   destroyCalcData(); virtual;
        {prparer les individus d'chantillonnage}
        procedure   prepareSample(); virtual; abstract;
        {rcuprer le cout de mauvais d'un classifieur - calcul rapide - on en profite pour mettre  jour les poids le cas chant}
        function    getCost(prmExamples: TExamples): double; virtual;
        {renvoyer les scores d'affectation pour un individu}
        procedure   resetScore(example: integer; var postProba: TTabScore); override;
        {vote  la majorit simple au dpart}
        procedure   ClassifyExample(example: integer; var response: TTypeDiscrete); override;
        public
        {rcupration des donnes de travail, les donnes en entre}
        constructor Create(AOwner: TObject); override;
        {destroy}
        destructor  Destroy(); override;
        //sortie HTML
        function    getHTMLResultsSummary(): string; override;
        {pointeur sur la liste des classifieurs}
        property Classifiers: TObjectList read FClassifiers;
        {individus de l'apprentissage courant}
        property InternalExamples: TExamples read FInternalExamples;
        end;

        {paramtre de l'oprateur - ils ont en commun le nombre d'instances  produire}
        TOpPrmMetaSpvCSCommitte = class(TOpPrmMetaSpvCostSensitive)
        private
        {nombre d'instances}
        FNbInstances: integer;
        {gnrateur de nombre alatoire, 0 : Random, 1: Standard}
        FRndMode: integer;
        protected
        function  createDlgParameters(): TForm; override;
        function  getCommitteParameters(): string; virtual;
        function  getClassifierParameters(): string; virtual;
        procedure SetDefaultParameters(); override;
        public
        function    getHTMLParameters(): string; override;
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        {nombre d'instances}
        property NbInstances: integer read FNbInstances write FNbInstances;
        property RndMode: integer read FRndMode write FRndMode;
        end;


implementation

USES
        Math,
        Sysutils, UConstConfiguration, ULogFile, UDlgOpPrmCommitteeCostSensitive;

{ TOpPrmMetaSpvCSCommitte }

function TOpPrmMetaSpvCSCommitte.createDlgParameters: TForm;
begin
 result:= TdlgOpPrmMetaSpvCSCommittee.CreateFromOpPrm(self);
end;

function TOpPrmMetaSpvCSCommitte.getClassifierParameters: string;
begin
 result:= (self.Operator as TOpMetaSpvCostSensitive).OpMLSpv.getHTMLParameters();
end;

function TOpPrmMetaSpvCSCommitte.getCommitteParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Number of classifier(s)</TD><TD width=30 align=right>%d</TD></TR>',[FNbInstances]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Rnd mode</TD><TD width=30 align=right>%s</TD></TR>',[StartSeedDescription[TStartSeed(FRndMode)]]);
 s:= s+'</table>';
 result:= s;
end;

function TOpPrmMetaSpvCSCommitte.getHTMLParameters: string;
var s: string;
begin
 s:= self.MatCost.getHTML(); 
 s:= s+'<H4>Committee parameters</H4>';
 s:= s+self.getCommitteParameters();
 s:= s+'<H4>Learning algorithm parameters</H4>';
 s:= s+self.getClassifierParameters();
 result:= s;
end;

procedure TOpPrmMetaSpvCSCommitte.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 FNbInstances:= prmINI.ReadInteger(prmSection,'nb_classifiers',FNbInstances);
 FRndMode:= prmINI.ReadInteger(prmSection,'rnd_mode',FRndMode);
end;

procedure TOpPrmMetaSpvCSCommitte.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(FNbInstances,sizeof(FNbInstances));
 prmStream.ReadBuffer(FRndMode,sizeof(FRndMode));
end;

procedure TOpPrmMetaSpvCSCommitte.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'nb_classifiers',FNbInstances);
 prmINI.WriteInteger(prmSection,'rnd_mode',FRndMode);
end;

procedure TOpPrmMetaSpvCSCommitte.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(FNbInstances,sizeof(FNbInstances));
 prmStream.WriteBuffer(FRndMode,sizeof(FRndMode));
end;

procedure TOpPrmMetaSpvCSCommitte.SetDefaultParameters;
begin
 inherited;
 FNbInstances:= 25;
 FRndMode:= 1;
end;

{ TOpMetaSpvCSCommitte }

procedure TOpMetaSpvCSCommitte.ApplyOneClassifier(numero, example: integer;
  var response: TTypeDiscrete);
var classifier: TCalcSpvLearning;
    k,kMin,j,nbValues: integer;
    cout,minCout: double;
    matCost: TCostMatrix;
    tmpProba: TTabScore;
begin
 classifier:= FClassifiers.Items[numero] as TCalcSpvLearning;

 //calculer les probas. d'affectation
 tmpProba:= TTabScore.create(classifier.ClassAttribute);
 classifier.getScore(example,tmpProba);
 //recherche du cout minimum
 nbValues:= self.OpMLSpv.ClassAttribute.nbValues;
 matCost:= (self.PrmOp as TOpPrmMetaSpvCostSensitive).MatCost;
 //calculer le cot associ  chaque prdiction
 minCout:= Math.MaxDouble;
 kMin:= 0;
 for k:= 1 to nbValues do
  begin
   cout:= 0.0;
   for j:= 1 to nbValues do cout:= cout + tmpProba[j]*matCost[j,k];
   if (cout < minCout)
    then
     begin
      minCout:= cout;
      kMin:= k;
     end;
  end;
 //et  la sortie, on slectionne le plus petit, s'il a t dtect
 if (kMin > 0)
  then response:= kMin
  //sinon, on prend l'affectation  la proba, sans tenir compte des cots --> a ne devrait jamais arriver !!!
  else response:= tmpProba.getIndexMax();
 //vider
 tmpProba.Free();

 //les classifieurs classent sans tenir compte des cots ? hum, pas terrible du tout semble-t-il...
 //classifier.classification(example,response);
end;

procedure TOpMetaSpvCSCommitte.ClassifyExample(example: integer;
    var response: TTypeDiscrete);

var k,kMin,j,nbValues: integer;
    cout,minCout: double;
    matCost: TCostMatrix;
begin
 //calculer les probas d'affectation
 self.resetScore(example,FPostProba);
 //on peut se baser sur le vote  la majorit simple, tout btement
 //et renvoyer l'indice du max. de la proba. conditionnelle
 //response:= FPostProba.getIndexMax();

 //cette seconde solution semble meilleure
 //on utilise les probas. conditionnelles
 //pour produire une affectation sensible aux cots
 //les cots interviennent  2 niveaux : (1) individuellement, pour chaque classifieur, (2) globalement sur l'ensemble des classifieurs
 
 //>>>recherche du cout minimum
 nbValues:= self.OpMLSpv.ClassAttribute.nbValues;
 matCost:= (self.PrmOp as TOpPrmMetaSpvCostSensitive).MatCost;
 //calculer le cot associ  chaque prdiction
 minCout:= Math.MaxDouble;
 kMin:= 0;
 for k:= 1 to nbValues do
  begin
   cout:= 0.0;
   for j:= 1 to nbValues do cout:= cout + FPostProba[j]*matCost[j,k];
   if (cout < minCout)
    then
     begin
      minCout:= cout;
      kMin:= k;
     end;
  end;
 //et  la sortie, on slectionne le plus petit, s'il a t dtect
 if (kMin > 0)
  then response:= kMin
  //sinon, on prend l'affectation  la proba, sans tenir compte des cots --> a ne devrait jamais arriver !!!
  else response:= FPostProba.getIndexMax();
 
end;

constructor TOpMetaSpvCSCommitte.Create(AOwner: TObject);
begin
 inherited;
 FRnd:= TRndGenerator.Create(TStartSeed((self.PrmOp as TOpPrmMetaSpvCSCommitte).FRndMode),DEFAULT_SEED_VALUE_1,DEFAULT_SEED_VALUE_2);
end;

destructor TOpMetaSpvCSCommitte.Destroy;
begin
 FRnd.Free();
 inherited;
end;

procedure TOpMetaSpvCSCommitte.destroyCalcData;
begin
 FreeAndNil(FInternalExamples);
end;

procedure TOpMetaSpvCSCommitte.destroyClassifiers;
begin
 inherited;
 if assigned(FClassifiers) then FreeAndNil(FClassifiers);
 setLength(FTabErrors,0);
end;

function TOpMetaSpvCSCommitte.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmMetaSpvCSCommitte;
end;

function TOpMetaSpvCSCommitte.getCost(prmExamples: TExamples): double;
var i: integer;
    example: integer;
    err: double;
    kActual,kPred: TTypeDiscrete;
    matCost: TCostMatrix;
begin
 //rcuprer la matrice de cot
 matCost:= (self.PrmOp as TOpPrmMetaSpvCSCommitte).MatCost;
 //on l'value
 err:= 0.0;
 for i:= 1 to prmExamples.Size do
  begin
   example:= prmExamples.Number[i];
   kActual:= OpMLSpv.ClassAttribute.dValue[example];
   self.ApplyOneClassifier(pred(FClassifiers.Count),example,kPred);
   //cot du classement
   err:= err+matCost[kActual,kPred];
  end;
 err:= err/(1.0*prmExamples.Size);
 result:= err;
end;

function TOpMetaSpvCSCommitte.getHTMLResultsSummary: string;
var s: string;
    i: integer;
begin
 s:= inherited getHTMLResultsSummary();
 //tableau des taux d'erreur
 s:= s+'<h4>Classifiers performances</h4>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Classifiers</TH><TH>Error rate</TH></TR>';
 for i:= 1 to high(FTabErrors) do
  s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD align=center>%d</TD><TD align=right>%.4f</TD></TR>',[i,FTabErrors[i]]);
 s:= s+'</table>';
 result:= s;
end;

procedure TOpMetaSpvCSCommitte.prepareCalcData;
begin
 FInternalExamples:= TExamples.Create(self.WorkData.Examples.Size);
end;

procedure TOpMetaSpvCSCommitte.ReInitialize;
begin
 inherited ReInitialize();
 //vider l'ensemble des classifieurs
 if assigned(FClassifiers) then FClassifiers.Clear;
end;

procedure TOpMetaSpvCSCommitte.resetScore(example: integer;
  var postProba: TTabScore);
var i: integer;
    kPred: TTypeDiscrete;
begin
 //vider le tableau d'affectation
 postProba.raz();
 //enquiller les diffrentes rponses
 //attention, on fait voter sur la base des conclusions corriges
 for i:= 0 to pred(FClassifiers.Count) do
  begin
   self.ApplyOneClassifier(i,example,kPred);
   postProba[kPred]:= postProba[kPred]+1.0;
  end;
end;

procedure TOpMetaSpvCSCommitte.RunLearning;
var i: integer;
    classifier: TCalcSpvLearning;
    err: double;
begin
 //tout vider
 if assigned(FClassifiers) then FreeAndNil(FClassifiers);
 if assigned(FRnd) then FRnd.Free();
 FRnd:= TRndGenerator.Create(TStartSeed((self.PrmOp as TOpPrmMetaSpvCSCommitte).FRndMode),DEFAULT_SEED_VALUE_1,DEFAULT_SEED_VALUE_2);
 setLength(FTabErrors,succ((self.PrmOp as TOpPrmMetaSpvCSCommitte).NbInstances));
 //construire
 FClassifiers:= TObjectList.Create(TRUE);
 //prparer les donnes internes
 self.prepareCalcData();
 //lancer les apprentissage successifs
 for i:= 1 to (self.PrmOp as TOpPrmMetaSpvCSCommitte).NbInstances do
  begin
   TRY
   //chantillonner les individus
   self.prepareSample();
   //lancer l'apprentissage
   classifier:= OpMLSpv.getInstanceSpvLearning();
   classifier.learning(FInternalExamples);
   //ajouter le classifieur dans la liste - trs important de le faire avant le calcul du taux d'erreur
   FClassifiers.Add(classifier);
   //calculer le cot de mauvais classement
   err:= self.getCost(self.WorkData.Examples);
   FTabErrors[i]:= err;
   //if (i mod 25 = 0) then TraceLog.WriteToLogFile(format('[CS-COMMITTEE] iteration numero = %d',[i]));
   EXCEPT
   //si a a plant...
   On e: Exception do
    begin
     TraceLog.WriteToLogFile(format('[%s] ERROR during learning (%d-th) classifier >> %s',[Self.ClassName,i,e.Message]));
     Raise Exception.Create(e.Message);
    end;
   END;
  end;
 //supprimer les donnes internes
 self.destroyCalcData();
 //nombre de classifieurs construits
 TraceLog.WriteToLogFile(format('[%s] number of generated classifiers = %d',[Self.ClassName,FClassifiers.Count]));
end;


end.
