(**********************************************************************)
(* UCompMetaSpvCSMultiCost.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(MultiCost)
@author(Ricco)
@created(01/01/2009)

L'ide est trs simple :

(1) Bagging avec intervention des cots
(2) A partir des distributions produites par le bags, construire
une prdiction sensible aux cots
(3) Apprentissage avec comme variable  prdire, la prdiction  l'tape prcdente.

La mthode est inspire de MetaCost de Domingos (KDD-1999), la diffrence est dans l'intoduction des cots ds le Bagging.

}

unit UCompMetaSpvCSMultiCost;

interface

USES
        Forms,
        UCompDefinition,
        UCompSpvLDefinition,
        UOperatorDefinition,
        UCalcSpvStructScore,
        UCompMetaSpvGenericCostSensitive,
        UCompMetaSpvCSOneInstance,
        UCompMetaSpvCSCommitte,
        UDatasetDefinition;

TYPE
        {gnrateur de composant meta}
        TMLGCompCSMultiCost = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant meta, singe un single learning}
        TMLCompCSMultiCost = class(TMLCompMetaSpvCostSensitive)
        protected
        function    ShortMLCompName(): string; override;
        function    getClassOperator: TClassOperator; override;
        end;

        TTabAffect = array of array of single;

        {l'oprateur meta}
        TOpCSMultiCost = class(TOpCSOneInstance)
        private
        //attribut - prdiction du bagging
        FBagPred: TAttribute;
        //Vecteur de proba -- en global pour acclrer les calculs
        FPostProba: TTabScore;
        //lancer la session bagging
        function    runBagging(): boolean;
        // partir des tableaux des probas, laborer la variable FBagPred
        procedure   createBagPred(var tab: TTabAffect);
        //appliquer un classifieur en tenant compte de la matrice de cot
        procedure   applyClassifier(classifier: TCalcSpvLearning; i: integer; var response: TTypeDiscrete);
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        procedure   RunLearning(); override;
        procedure   ReInitialize(); override;
        procedure   destroyClassifiers(); override;
        //ajouter la classe interne en sortie
        procedure   SetProjections(); override;
        end;

        {paramtre d'oprateur meta -- la grosse astuce}
        TOpPrmCSMultiCost = class(TOpPrmMetaSpvCSCommitte)
        end;


implementation

uses
    Classes, Sysutils, Math,
    UCompMetaSpvBagging, UCompMetaSpvCommitte, UDatasetImplementation,
    ULogFile, UCalcRndGenerator, UDatasetExamples;

{ TMLGCompCSMultiCost }

procedure TMLGCompCSMultiCost.GenCompInitializations;
begin
 FMLComp:= mlcMetaSpvLearning;
end;

function TMLGCompCSMultiCost.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompCSMultiCost;
end;

{ TMLCompCSMultiCost }

function TMLCompCSMultiCost.getClassOperator: TClassOperator;
begin
 result:= TOpCSMultiCost;
end;

function TMLCompCSMultiCost.ShortMLCompName: string;
begin
 result:= 'MultiCost';
end;

{ TOpCSMultiCost }

procedure TOpCSMultiCost.applyClassifier(classifier: TCalcSpvLearning;
  i: integer; var response: TTypeDiscrete);
var j,nbValues,k,kMin: integer;
    costMatrix: TCostMatrix;
    minCout,cout: double;
begin
 //version sans tenir compte des cots -- ce serait le MetaCost originel
 //classifier.classification(i,response);

 //version avec prise en compte des cots
 nbValues:= self.OpMLSpv.ClassAttribute.nbValues;
 //rcuprer et vrifier la matrice de cot
 costMatrix:= (self.PrmOp as TOpPrmCSMultiCost).MatCost;
 //tableau pour les probas
 classifier.getScore(i,FPostProba);
 //calculer le cot associ  chaque prdiction
 minCout:= Math.MaxDouble;
 kMin:= 0;
 for k:= 1 to nbValues do
  begin
   cout:= 0.0;
   for j:= 1 to nbValues do cout:= cout + FPostProba[j]*costMatrix[j,k];
   if (cout < minCout)
    then
     begin
      minCout:= cout;
      kMin:= k;
     end;
  end;
 //et  la sortie, on slectionne le plus petit, s'il a t dtect
 //le else ne devrait jamais arriver, sauf plantage...
 if (kMin > 0) then response:= kMin else response:= succ(random(nbValues));

end;

procedure TOpCSMultiCost.createBagPred(var tab: TTabAffect);
var j,n,i,nbValues,k,kMin: integer;
    costMatrix: TCostMatrix;
    minCout,cout: double;
begin
 //quelques infos importantes
 n:= self.WorkData.LstAtts[asAll].Attribute[0].Size;
 nbValues:= self.OpMLSpv.ClassAttribute.nbValues;
 //rcuprer et vrifier la matrice de cot
 costMatrix:= (self.PrmOp as TOpPrmCSMultiCost).MatCost;
 //crer la variable
 FBagPred:= TAttDiscrete.Create('_bag_pred_',n);
 FBagPred.LstValues.assign(self.OpMLSpv.ClassAttribute.LstValues);
 //pour chaque observation
 for i:= 1 to n do
  begin
   //calculer le cot associ  chaque prdiction
   minCout:= Math.MaxDouble;
   kMin:= 0;
   for k:= 1 to nbValues do
    begin
     cout:= 0.0;
     for j:= 1 to nbValues do cout:= cout + tab[i-1,j-1]*costMatrix[j,k];
     if (cout < minCout)
      then
       begin
        minCout:= cout;
        kMin:= k;
       end;
    end;
   //et  la sortie, on slectionne le plus petit, s'il a t dtect
   //le else ne devrait jamais arriver, sauf plantage...
   if (kMin > 0) then FBagPred.dValue[i]:= kMin else FBagPred.dValue[i]:= succ(random(nbValues));
  end;
end;

procedure TOpCSMultiCost.destroyClassifiers;
begin
 inherited;
 if assigned(FBagPred) then FreeAndNil(FBagPred);
end;

function TOpCSMultiCost.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmCSMultiCost;
end;

procedure TOpCSMultiCost.ReInitialize;
begin
 inherited;
 if assigned(FBagPred) then FreeAndNil(FBagPred);
end;

function TOpCSMultiCost.runBagging(): boolean;
var rnd: TRndGenerator;//gnrateur de nombre alatoire
    classifier: TCalcSpvLearning;//apprentissage
    r,repetition,i,n,nbValues,k: integer;
    kPred: TTypeDiscrete;
    examples: TExamples;
    tabAffect: TTabAffect;
begin
 nbValues:= self.OpMLSpv.ClassAttribute.nbValues;
 //gnrateur de nombre alatoire
 rnd:= TRndGenerator.Create(TStartSeed((self.PrmOp as TOpPrmCSMultiCost).RndMode),DEFAULT_SEED_VALUE_1,DEFAULT_SEED_VALUE_2);
 //tableau des affectations
 repetition:= (self.PrmOp as TOpPrmCSMultiCost).NbInstances;
 n:= self.WorkData.LstAtts[asAll].Attribute[0].Size;
 TRY
 setLength(tabAffect,n,nbValues);
 for i:= 0 to n-1 do for k:= 0 to nbValues-1 do tabAffect[i,k]:= 0.0;
 //index des individus
 examples:= TExamples.Create(self.WorkData.Examples.Size);
 //pour chaque replication
 for r:= 1 to repetition do
  begin
   self.WorkData.Examples.SampleReplicate(examples,rnd);
   classifier:= OpMlSpv.getInstanceSpvLearning();
   classifier.learning(examples);
   //classer chaque individu
   for i:= 1 to n do
    begin
     self.applyClassifier(classifier,i,kPred);
     //aller directement au comptage
     tabAffect[i-1,kPred-1]:= tabAffect[i-1,kPred-1] + 1.0;
    end;
   //vider le classifieur
   classifier.Free();
   //infos.
   if (r mod 100 = 0) then TraceLog.WriteToLogFile(format('[%s] iteration number = %d',[Self.ClassName,r]));
  end;
 examples.Free();
 rnd.Free();
 // partir du tableau d'affectation, construire la variable "bagpred"
 self.createBagPred(tabAffect);
 //relcher le tableau maintenant
 setLength(tabAffect,0,0);
 //si tout s'est bien pass...
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

procedure TOpCSMultiCost.RunLearning;
var tmpData: TMLDataset;
    ok: boolean;
    nbValues: integer;
begin
 if assigned(FClassifier) then FreeAndNil(FClassifier);
 if assigned(FBagPred) then FreeAndNil(FBagPred);
 if assigned(FPostProba) then FreeAndNil(FPostProba);
 //vrifier la matrice de cot
 nbValues:= self.OpMLSpv.ClassAttribute.nbValues;
 if ((self.PrmOp as TOpPrmCSMultiCost).MatCost.Size <> nbValues) then (self.PrmOp as TOpPrmCSMultiCost).MatCost.reset(nbValues);
 //lancer la sessions Bagging
 FPostProba:= TTabScore.create(self.OpMLSpv.ClassAttribute);
 ok:= runBagging();
 FreeAndNil(FPostProba);
 //si c'est bon, on lance l'apprentissage sur la variable prdite
 if ok
  then
   begin
    //TraceLog.WriteToLogFile(Format('[BAG_PRED] i = %d, value = %d',[1,FBagPred.dValue[1]]));
    //prparer les donnes intermdiaires
    tmpData:= TMLDataset.CreateFromMLDataset(self.WorkData);
    tmpData.LstAtts[asTarget].Clear();
    tmpData.LstAtts[asTarget].Add(FBagPred);
    //apprentissage en modifiant la variable  prdire
    TRY
    opMLSpv.ConnectAttributes(tmpData);//!\ trs sensible
    FClassifier:= OpMlSpv.getInstanceSpvLearning();
    ok:= FClassifier.learning(self.WorkData.Examples);
    if not(ok) then  TraceLog.WriteToLogFile(format('collapse creation during %s learning',[FClassifier.ClassName]));
    FINALLY
    opMLSpv.ConnectAttributes(self.WorkData);//!\obligatoire!!!
    //librer le dataset intermdiaire
    tmpData.Free();
    END;
    if not(ok) then raise Exception.Create(format('collapse creation during %s learning',[FClassifier.ClassName]));
   end
  //envoyer une exception
  else Raise Exception.Create('Error during bagging');
end;

procedure TOpCSMultiCost.SetProjections;
var tmp: TAttribute;
begin
 inherited;
 //et ajouter la variable temporaire qui a servi de classe  prdire par la suite
 tmp:= FBagPred.duplicate('tmp_MultiCost_class_'+IntToStr((self.MLOwner as TMLComponent).Number));
 self.GenAtts.Add(tmp);
end;

initialization
 RegisterClass(TMLGCompCSMultiCost);
end.
