(*************************************************************************)
(* UCompMetaSpvCSOneInstance.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(*************************************************************************)

{
@abstract(Classe Apprentissage tenant compte des cots de mauvais classement)
@author(Ricco)
@created(26/12/2008)

La correction est ralise uniquement sur la rgle d'affectation

}

unit UCompMetaSpvCSOneInstance;

interface

USES
        Forms,
        UCompDefinition,
        UCompSpvLDefinition,
        UOperatorDefinition,
        UCalcSpvStructScore,
        UCompMetaSpvGenericCostSensitive;

TYPE
        {gnrateur de composant meta}
        TMLGCompCSOneInstance = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant meta, singe un single learning}
        TMLCompCSOneInstance = class(TMLCompMetaSpvCostSensitive)
        protected
        function    ShortMLCompName(): string; override;
        function    getClassOperator: TClassOperator; override;
        function    GetLogResultDescription(): string; override;
        end;

        {l'oprateur meta}
        TOpCSOneInstance = class(TOpMetaSpvCostSensitive)
        private
        protected
        {un seul classifieur}
        FClassifier: TCalcSpvLearning;
        function    getNewForm(): TForm; override;
        function    getClassParameter: TClassOperatorParameter; override;
        procedure   RunLearning(); override;
        {il dtruit carrment l'instance de calcul ici}
        procedure   ReInitialize(); override;
        //rcuprer les scores d'affectation
        procedure   resetScore(example: integer; var postProba: TTabScore); override;
        {dtruire les classifieurs internalement gnrs}
        procedure   destroyClassifiers(); override;
        public
        {reprendre tel quel le rapport gnr par l'oprateur d'apprentissage}
        function    getHTMLResultsSummary(): string; override;
        {le classifieur}
        property    Classifier: TCalcSpvLearning read FClassifier;
        end;

        {paramtre d'oprateur meta}
        TOpPrmCSOneInstance = class(TOpPrmMetaSpvCostSensitive)
        public
        function    getHTMLParameters(): string; override;
        end;


implementation

uses
    Classes,
    SysUtils, ULogFile, UFrmOpSingleSpvLearning;

{ TMLCompCSOneInstance }

function TMLCompCSOneInstance.getClassOperator: TClassOperator;
begin
 result:= TOpCSOneInstance;
end;

function TMLCompCSOneInstance.GetLogResultDescription: string;
begin
 result:= format('one instance of %s is generated',[MLCompSupervised.Description]);
end;

function TMLCompCSOneInstance.ShortMLCompName: string;
begin
 result:= 'CSOneInstance';
end;

{ TMLGCompCSOneInstance }

procedure TMLGCompCSOneInstance.GenCompInitializations;
begin
 FMLComp:= mlcMetaSpvLearning;
end;

function TMLGCompCSOneInstance.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompCSOneInstance;
end;

{ TOpCSOneInstance }

procedure TOpCSOneInstance.destroyClassifiers;
begin
 if assigned(FClassifier)
  then FreeAndNil(FClassifier);
end;

function TOpCSOneInstance.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmCSOneInstance; 
end;

function TOpCSOneInstance.getHTMLResultsSummary: string;
var s: string;
begin
 s:= inherited getHTMLResultsSummary();
 result:= s+FClassifier.getHTMLResults();
end;

function TOpCSOneInstance.getNewForm: TForm;
begin
 if (FClassifier.hasOwnViewForm() = FALSE)
  then result:= TFrmOpViewSingleSpvLearning.CreateFromOperator(self)
  else result:= FClassifier.getClassViewForm().CreateFromOperator(self);
end;

procedure TOpCSOneInstance.ReInitialize;
begin
 inherited ReInitialize();
 if assigned(FClassifier)
  then FreeAndNil(FClassifier);
end;


procedure TOpCSOneInstance.resetScore(example: integer;
  var postProba: TTabScore);
begin
 FClassifier.getScore(example,postProba);
end;

procedure TOpCSOneInstance.RunLearning;
begin
 if assigned(FClassifier) then FClassifier.Free;
 //cration de l'instance
 FClassifier:= OpMLSpv.getInstanceSpvLearning();
 //apprentissage en envoyant les bons exemples
 if not(FClassifier.learning(workdata.Examples))
  then
   begin
    TraceLog.WriteToLogFile(format('collapse creation during %s learning',[FClassifier.ClassName]));
    raise Exception.Create(format('collapse creation during %s learning',[FClassifier.ClassName]));
   end;
end;


{ TOpPrmCSOneInstance }

function TOpPrmCSOneInstance.getHTMLParameters: string;
begin
 result:= '<H4>Learning Algorithm Parameter</H4>';
 result:= result + (Operator as TOpCSOneInstance).Classifier.OpPrmSpv.getHTMLParameters();
 result:= result + inherited getHTMLParameters();
end;

initialization
 Classes.RegisterClass(TMLGCompCSOneInstance);
end.
