(********************************************************************************)
(* UCompMetaSpvGenericCostSensitive.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(********************************************************************************)

{
@abstract(Classe gnrique - Apprentissage tenant compte des cots de mauvais classement)
@author(Ricco)
@created(26/12/2008)

Hritier de MetaSpvLearning, le plus gros travail sera sur le paramtrage.

(1) On a besoin qu'une matrice de cots soit dclare.
(2) Il faut que ses dimensions soient cohrents avec le nombre de modalits de la variable  prdire

}

unit UCompMetaSpvGenericCostSensitive;

interface

USES
    Grids, Forms, Classes, IniFiles,
    UOperatorDefinition,
    UCompSpvLDefinition,
    UDatasetDefinition;

TYPE
    //Matrice  2 dim.
    TMatCout = array of array of double;

    //classe matrice de cot
    TCostMatrix = class
    private
    //dimension (matrice Size x Size)
    FSize: integer;
    //la matrice
    FMat: TMatCout;
    //modifier la dimension
    procedure setSize(newSize: integer);
    //lire les valeurs, attention indices dcals d'une unit
    function  getCost(i,j: integer): double;
    //crire les valeurs, indices dcals d'une unit
    procedure setCost(i,j: integer; value: double);
    //initialiser une grille TStringGrid
    procedure initializeGrid(var grid: TStringGrid);
    public
    //constructeur - destructeur
    constructor create(prmSize: integer);
    destructor  destroy(); override;
    //crire et lire dans une TSTringGrid
    procedure setCostsToGrid(var grid: TStringGrid);
    procedure getCostsFromGrid(grid: TStringGrid);
    //r-initialization en cas de divergence
    procedure reset(newSize: integer);
    //description HTML de la matrice
    function  getHTML(): string;
    //crire et lire dans un stream
    procedure   LoadFromStream(prmStream: TStream); virtual;
    procedure   SaveToStream(prmStream: TStream); virtual;
    {sauvegarde dans un fichier INI}
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); virtual;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); virtual;
    //crire et lire dans un INIFile
    property Size: integer read FSize write setSize;
    property Cost[i,j: integer]: double read getCost write SetCost; default;
    end;

    //classe de composant
    TMLCompMetaSpvCostSensitive = class(TMLCompMetaSpvLearning)
    end;

    //classe de calcul
    TOpMetaSpvCostSensitive = class(TOpMetaSpvLearning)
    protected
    //classer un exemple  partir des probas et des cots
    procedure   ClassifyExample(example: integer; var response: TTypeDiscrete); override;
    //ajouter une vrification pour la matrice de cot
    function    CoreExecute(): boolean; override;
    end;

    //classe de paramtrage
    TOpPrmMetaSpvCostSensitive = class(TOpPrmMetaSpvLearning)
    protected
    //matrice de cots
    FMatCost: TCostMatrix;
    //un des onglets est dvolu  la matrice de cots
    function    CreateDlgParameters(): TForm; override;
    //paramtre par dfaut, matrice unitaire
    procedure   SetDefaultParameters(); override;
    public
    //constructeur et destructeur
    constructor Create(prmOp: TOperator); override;
    destructor  destroy(); override;
    //afficher la matrice de cots en HTML
    function    getHTMLParameters(): string; override;
    {chargement  partir d'un flux}
    procedure   LoadFromStream(prmStream: TStream); override;
    {chargement  partir d'un fichier INI}
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    {sauvegarde dans un flux}
    procedure   SaveToStream(prmStream: TStream); override;
    {sauvegarde dans un fichier INI}
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    //proprit
    property MatCost: TCostMatrix read FMatCost;
    end;


implementation

USES
    Math,
    Sysutils, UConstConfiguration, UDlgOpPrmBaseCostSensitive;

{ TCostMatrix }

constructor TCostMatrix.create(prmSize: integer);
begin
 inherited create();
 self.reset(prmSize);
end;

destructor TCostMatrix.destroy;
begin
  Finalize(FMat);
  inherited;
end;

function TCostMatrix.getCost(i, j: integer): double;
begin
 result:= FMat[pred(i),pred(j)];
end;

procedure TCostMatrix.getCostsFromGrid(grid: TStringGrid);
var i,j: integer;
begin
 //tester
 if ((grid.ColCount-1) <> FSize) then self.setSize(grid.ColCount-1);
 //rcuprer
 for i:= 1 to FSize do
  for j:= 1 to FSize do
   begin
    //scurisons....
    TRY
    Cost[i,j]:= StrToFloat(grid.Cells[j,i]);
    EXCEPT
    Cost[i,j]:= 0.0;
    END;
   end;
end;

function TCostMatrix.getHTML: string;
var s: string;
    i,j: integer;
begin
 s:= '<H4>Misclassification Cost Matrix</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH>&nbsp;</TH><TH colspan=%d>Observed vs. Predicted</TH></TR>',[FSize]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TH>Cost(i,j)</TH>';
 for j:= 1 to FSize do s:= s + Format('<TH>%d</TH>',[j]);
 s:= s + '</TR>';
 for i:= 1 to FSize do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%d</TH>',[i]);
   for j:= 1 to FSize do
    s:= s + format('<TD align=center>%.0f</TD>',[Cost[i,j]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

procedure TCostMatrix.initializeGrid(var grid: TStringGrid);
var i: integer;
begin
 //dimension
 grid.ColCount:= FSize+1;
 grid.RowCount:= FSize+1;
 //marge
 for i:= 1 to pred(grid.ColCount) do
  begin
   grid.Cells[i,0]:= IntToStr(i);
   grid.Cells[0,i]:= IntToStr(i);
  end;
end;

procedure TCostMatrix.LoadFromINI(prmSection: string; prmINI: TMemIniFile);
var i,j,newSize: integer;
    tmp: string;
    value: double;
begin
 newSize:= prmINI.ReadInteger(prmSection,'Cost_Matrix_Size',2);
 self.setSize(newSize);
 for i:= 1 to newSize do
  For j:= 1 to newSize do
   begin
    tmp:= Format('cost_%d_%d',[i,j]);
    value:= prmINI.ReadFloat(prmSection,tmp,0);
    Cost[i,j]:= value;
   end;
end;

procedure TCostMatrix.LoadFromStream(prmStream: TStream);
var i,j,newSize: integer;
    value: double;
begin
 prmStream.ReadBuffer(newSize,sizeof(FSize));
 self.setSize(newSize);
 for i:= 1 to newSize do
  For j:= 1 to newSize do
   begin
    prmStream.ReadBuffer(value,sizeof(value));
    Cost[i,j]:= value;
   end;
end;

procedure TCostMatrix.reset(newSize: integer);
var i,j: integer;
begin
 self.setSize(newSize);
 for i:= 1 to FSize do
  for j:= 1 to FSize do
   if (i = j)
    then Cost[i,j]:= 0.0
    else Cost[i,j]:= 1.0;
end;

procedure TCostMatrix.SaveToINI(prmSection: string; prmINI: TMemIniFile);
var i,j: integer;
    tmp: string;
begin
 prmINI.WriteInteger(prmSection,'Cost_Matrix_Size',FSize);
 for i:= 1 to FSize do
  For j:= 1 to FSize do
   begin
    tmp:= Format('cost_%d_%d',[i,j]);
    prmINI.WriteFloat(prmSection,tmp,Cost[i,j]);
   end;
end;

procedure TCostMatrix.SaveToStream(prmStream: TStream);
var i,j: integer;
    value: double;
begin
 prmStream.WriteBuffer(FSize,sizeof(FSize));
 for i:= 1 to FSize do
  For j:= 1 to FSize do
   begin
    value:= Cost[i,j];
    prmStream.WriteBuffer(value,sizeof(value));
   end;
end;

procedure TCostMatrix.setCost(i, j: integer; value: double);
begin
 FMat[pred(i),pred(j)]:= value;
end;

procedure TCostMatrix.setCostsToGrid(var grid: TStringGrid);
var i,j: integer;
begin
 self.initializeGrid(grid);
 for i:= 1 to FSize do
  for j:= 1 to FSize do
   grid.Cells[j,i]:= FloatToStr(self.Cost[i,j]);
end;

procedure TCostMatrix.setSize(newSize: integer);
begin
 if (FSize <> newSize)
  then
   begin
    FSize:= newSize;
    SetLength(FMat,FSize,FSize);
   end;
end;

{ TOpPrmMetaSpvCostSensitive }

constructor TOpPrmMetaSpvCostSensitive.Create(prmOp: TOperator);
begin
 inherited;
 //par dfaut, on table sur un problme  2 modalits
 FMatCost:= TCostMatrix.create(2);
end;

function TOpPrmMetaSpvCostSensitive.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmMetaSpvCostSensitiveBase.CreateFromOpPrm(self);
end;

destructor TOpPrmMetaSpvCostSensitive.destroy;
begin
 FMatCost.Free();
 inherited;
end;

function TOpPrmMetaSpvCostSensitive.getHTMLParameters: string;
begin
 result:= FMatCost.getHTML();
end;

procedure TOpPrmMetaSpvCostSensitive.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FMatCost.LoadFromINI(prmSection,prmINI);
end;

procedure TOpPrmMetaSpvCostSensitive.LoadFromStream(prmStream: TStream);
begin
 FMatCost.LoadFromStream(prmStream);
end;

procedure TOpPrmMetaSpvCostSensitive.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FMatCost.SaveToINI(prmSection,prmINI);
end;

procedure TOpPrmMetaSpvCostSensitive.SaveToStream(prmStream: TStream);
begin
 FMatCost.SaveToStream(prmStream);
end;

procedure TOpPrmMetaSpvCostSensitive.SetDefaultParameters;
begin
 //rien...
end;

{ TOpMetaSpvCostSensitive }

procedure TOpMetaSpvCostSensitive.ClassifyExample(example: integer;
  var response: TTypeDiscrete);
var k,kMin,j,nbValues: integer;
    cout,minCout: double;
    matCost: TCostMatrix;
begin
 nbValues:= self.OpMLSpv.ClassAttribute.nbValues;
 matCost:= (self.PrmOp as TOpPrmMetaSpvCostSensitive).MatCost;
 //calculer les probas. d'affectation
 self.resetScore(example,FPostProba);
 //calculer le cot associ  chaque prdiction
 minCout:= Math.MaxDouble;
 kMin:= 0;
 for k:= 1 to nbValues do
  begin
   cout:= 0.0;
   for j:= 1 to nbValues do cout:= cout + FPostProba[j]*matCost[j,k];
   if (cout < minCout)
    then
     begin
      minCout:= cout;
      kMin:= k;
     end;
  end;
 //et  la sortie, on slectionne le plus petit, s'il a t dtect
 if (kMin > 0)
  then response:= kMin
  //sinon, on prend l'affectation  la proba, sans tenir compte des cots --> a ne devrait jamais arriver !!!
  else response:= FPostProba.getIndexMax();
end;

function TOpMetaSpvCostSensitive.CoreExecute: boolean;
var costMatrix: TCostMatrix;
begin
 costMatrix:= (PrmOp as TOpPrmMetaSpvCostSensitive).MatCost;
 //si incohrence, on rinitialise d'autorit la matrice de cot
 //c'tait a ou envoyer un message d'erreur et bloquer l'excution
 if (costMatrix.Size <> self.OpMLSpv.ClassAttribute.nbValues)
  then costMatrix.reset(self.OpMLSpv.ClassAttribute.nbValues);
 //suite...
 result:= inherited CoreExecute(); 
end;

end.
