(*****************************************************************)
(* UCompSpvCSTreeC45.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(*****************************************************************)

{
@abstract(Cost-sensitive decision tree - algo C4.5)
@author(Ricco)
@created(30/12/2008)

Arbre dcision mthode C4.5 - Introduction des cots de mauvais classement.
Voir Bardford, Kunz, Kohavi, Brunk et Brodley - ECML-98

NOTE :
La mthode est trop sensible au paramtrage.
On prfrera CART qui est autrement plus performant et surtout STABLE dans ce contexte.

}

unit UCompSpvCSTreeC45;

interface

USES
        Forms, Classes, IniFiles,
        UCalcTreeStructureDefinition,
        UCompDefinition,
        UCompSpvLDefinition,
        UOperatorDefinition,
        UCompSpvTreeID3,
        UCalcRndGenerator,
        UFrmBaseOperator,
        UCompSpvTreeCART,
        UCompMetaSpvGenericCostSensitive,
        UDatasetExamples,
        UDatasetdefinition, UDatasetImplementation,
        UCalcSpvTreeC45;

TYPE
        {gnrateur de composant C4.5}
        TMLGCompSpvCSTreeC45 = class(TMLGCompSpvTree)
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant SpvTree C4.5}
        TMLCompSpvCSTreeC45 = class(TMLCompSpvTree)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {l'oprateur C4.5}
        TOpSpvCSTreeC45 = class(TOpSpvTree)
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    getClassSpvLearning(): TClassCalcSpvLearning; override;
        end;

        {paramtre de l'algo CART}
        TOpPrmSpvCSTreeC45 = class(TOpPrmSpvTree)
        private
        //taille minimale pour les feuilles
        FMinSizeLeaf: integer;
        //lambda pour l'estimation laplacienne des distributions
        FLambda: double;
        //matrice de cot
        FCostMatrix: TCostMatrix;
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        public
        {son propritaire est pass en paramtre : l'oprateur}
        constructor Create(prmOp: TOperator); override;
        destructor  Destroy(); override;
        function    getHTMLParameters(): string; override;
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        property    CostMatrix: TCostMatrix read FCostMatrix;
        property    Lambda: double read FLambda write FLambda;
        property    MinSizeLeaf: integer read FMinSizeLeaf write FMinSizeLeaf;
        end;

        //classe de calcul interne
        TCalcSpvCSTreeC45 = class(TCalcSpvTree)
        protected
        function    getClassTreeStructureSpv(): TClassMLTreeStructure; override;
        function    coreLearning(examples: TExamples): boolean; override;
        end;

        //classe de noeud d'arbre
        TMLTreeNodeSpvCSC45 = class(TMLTreeNodeSpvC45)
        protected
        procedure AssignConclusion(); override;
        function  getClassLstSplitAttributes(): TClassLstSplitAttributes; override;
        public
        //on dsactive la construction "hurdling"
        function  isSatured(): boolean; override;
        end;

        //classe de structure d'arbre
        TMLTreeStructureSpvCSC45 = class(TMLTreeStructureSpvC45)
        protected
        //matrice de cot rcupre du paramtrage de la mthode
        FCostMatrix: TCostMatrix;
        function   getClassMLTreeNode(): TClassMLTreeNode; override;
        public
        constructor create(paramMethod: TOperatorParameter; prmTarget,prmInput: TLstAttributes; prmExamples: TExamples); override;
        procedure   PostPruning(); override;
        property    CostMatrix: TCostMatrix read FCostMatrix;
        end;

        //
        TLstSplitAttSpvCSC45 = class(TLstSplitAttSpvC45)
        protected
        function  getClassSplitAttribut(): TClassSplitAttribut; override;
        end;

        //
        TSplitAttributSpvCSC45 = class(TSplitAttributSpvC45)
        protected
        function    ComputeAcceptSplit(): boolean; override;
        //discrtisation -- introduire une valeur de coupure simplifie
        //suggre dans la version C4.5 Release 8, et que l'on peut rapprocher  celui de MDLPC 
        procedure   OptimizeContinuousLeaves(); override;
        function    getClassSplitLeaf(): TClassSplitLeaf; override;
        end;

        //
        TSplitLeafSpvCSC45 = class(TSplitLeafSpvC45)
        end;

implementation

USES
        Math, Contnrs, Sysutils, UCalcRulesDefinition,
        ULogFile, UDlgOpPrmSpvTreeCSC45, UConstConfiguration;

{ TMLGCompSpvCSTreeC45 }

function TMLGCompSpvCSTreeC45.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvCSTreeC45;
end;

{ TMLCompSpvCSTreeC45 }

function TMLCompSpvCSTreeC45.getClassOperator: TClassOperator;
begin
 result:= TOpSpvCSTreeC45;
end;

{ TOpSpvCSTreeC45 }

function TOpSpvCSTreeC45.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSpvCSTreeC45;
end;

function TOpSpvCSTreeC45.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcSpvCSTreeC45; 
end;

{ TOpPrmSpvCSTreeC45 }

constructor TOpPrmSpvCSTreeC45.Create(prmOp: TOperator);
begin
 inherited;
 FCostMatrix:= TCostMatrix.create(2);
end;

function TOpPrmSpvCSTreeC45.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSpvTreeCSC45.CreateFromOpPrm(self);
end;

destructor TOpPrmSpvCSTreeC45.Destroy;
begin
 FCostMatrix.Free();
 inherited;
end;

function TOpPrmSpvCSTreeC45.getHTMLParameters: string;
var s: string;
begin
 s:= '<H4>Algorithm parameters</H4>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Cost sensitive C4.5 parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Min size of leaves</TD><TD align=right>%d</TD></TR>',[FMinSizeLeaf]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Lambda for laplacian</TD><TD align=right>%.2f</TD></TR>',[FLambda]);
 s:= s+'</table>';
 //matrice de cot
 s:= s + FCostMatrix.getHTML();
 //
 result:= s;
end;

procedure TOpPrmSpvCSTreeC45.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FMinSizeLeaf:= prmINI.ReadInteger(prmSection,'MinSize',FMinSizeLeaf);
 FLambda:= prmINI.ReadFloat(prmSection,'Lambda',FLambda);
 FCostMatrix.LoadFromINI(prmSection,prmINI);
end;

procedure TOpPrmSpvCSTreeC45.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FMinSizeLeaf,sizeof(FMinSizeLeaf));
 prmStream.ReadBuffer(FLambda,sizeof(FLambda));
 FCostMatrix.LoadFromStream(prmStream);
end;

procedure TOpPrmSpvCSTreeC45.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'MinSize',FMinSizeLeaf);
 prmINI.WriteFloat(prmSection,'Lambda',FLambda);
 FCostMatrix.SaveToINI(prmSection,prmINI);
end;

procedure TOpPrmSpvCSTreeC45.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FMinSizeLeaf,sizeof(FMinSizeLeaf));
 prmStream.WriteBuffer(FLambda,sizeof(FLambda));
 FCostMatrix.SaveToStream(prmStream);
end;

procedure TOpPrmSpvCSTreeC45.SetDefaultParameters;
begin
 //valeurs par dfaut...
 FMinSizeLeaf:= 5;
 FLambda:= 3;
end;

{ TMLTreeStructureSpvCSC45 }

constructor TMLTreeStructureSpvCSC45.create(
  paramMethod: TOperatorParameter; prmTarget, prmInput: TLstAttributes;
  prmExamples: TExamples);
begin
 inherited;
 FCostMatrix:= (paramMethod as TOpPrmSpvCSTreeC45).CostMatrix;
end;

function TMLTreeStructureSpvCSC45.getClassMLTreeNode: TClassMLTreeNode;
begin
 result:= TMLTreeNodeSpvCSC45;
end;

//comparaison -- pour le tri selon invers selon le niveau
function compareNodeAsDepth(Item1, Item2: Pointer): Integer;
begin
 if (TMLTreeNodeSpvCSC45(Item1).Depth > TMLTreeNodeSpvCSC45(Item2).Depth)
  then result:= +1
  else
   begin
    if (TMLTreeNodeSpvCSC45(Item1).Depth < TMLTreeNodeSpvCSC45(Item2).Depth)
     then result:= -1
     else result:= 0;
   end;
end;
//-----------------------------------------------------------

procedure TMLTreeStructureSpvCSC45.PostPruning;
var srtLeaves: TObjectList;
    i: integer;
    leaf,predecessor,successor: TMLTreeNodeSpvCSC45;
    errRef,errLeaves,tmp: double;
    memeConclusion: boolean;
begin
 //TraceLog.WriteToLogFile(format('[CS-C4.5] number of leaves >> before << pruning = %d',[self.CountLeaves]));
 srtLeaves:= TObjectList.Create(FALSE);
 //rcuprer les feuilles
 for i:= 0 to pred(self.CountLeaves) do
  srtLeaves.Add(self.Leaf[i]);
 //trier selon le niveau
 srtLeaves.Sort(compareNodeAsDepth);
 //lancer le post-pruning -- s'il n'y a qu'une feuille, c'est la racine !
 while (srtLeaves.Count > 1) do
  begin
   //rcuprer le premier noeud, qui est en fait le plus profond
   leaf:= srtLeaves.Items[srtLeaves.Count-1] as TMLTreeNodeSpvCSC45;
   //si c'est pas la racine
   if (leaf.Predecessor <> NIL)
    then
     begin
      predecessor:= leaf.Predecessor as TMLTreeNodeSpvCSC45;
      if predecessor.isPrunable()
       then
        begin
         //calculer l'erreur pessimiste
         errRef:= predecessor.pessimistic;
         //pour chaque enfant
         errLeaves:= 0.0;
         memeConclusion:= TRUE;
         for i:= 0 to pred(predecessor.getCountSuccessors()) do
          begin
           successor:= predecessor.getSuccessor(i) as TMLTreeNodeSpvCSC45;
           //scurisons avant de raliser les calculs : si non affect, on prend celui du pre
           tmp:= successor.pessimistic;
           if (tmp = Math.MaxDouble) then tmp:= predecessor.pessimistic;
           //on prend la moyenne pondre ici puisque les cots sont normaliss par les effectifs
           errLeaves:= errLeaves + (1.0*successor.DistClass.NbExamples)/(1.0*predecessor.DistClass.NbExamples) * tmp;
           //quelle que soit le rsultat, les successeurs seront exclus du calcul au tour suivant
           srtLeaves.Extract(successor);
           //vrifier la conclusion
           memeConclusion:= memeConclusion and (predecessor.NodeConclusion = successor.NodeConclusion);
          end;
         //comparer pour savoir s'il faut laguer ou pas
         if (errRef <= errLeaves) or memeConclusion
          then
           begin
            predecessor.pruneFromHere();
            //le prdecesseur devient une feuille -- candidat  la suppression donc
            srtLeaves.Add(predecessor);
            srtLeaves.Sort(compareNodeAsDepth);
           end;
        end
       //else srtLeaves.Extract(leaf);
       else srtLeaves.delete(srtLeaves.Count-1);//c'est plus rapide !!!
     end;
  end;
 srtLeaves.Free();
 //TraceLog.WriteToLogFile(format('[CS-C4.5] number of leaves >> after << pruning = %d',[self.CountLeaves]));
end;

{ TCalcSpvCSTreeC45 }

function TCalcSpvCSTreeC45.coreLearning(examples: TExamples): boolean;
var MatCost: TCostMatrix;
begin
 MatCost:= (self.OpPrmSpv as TOpPrmSpvCSTreeC45).CostMatrix;
 //vrifier la matrice de cot et assurer le coup le cas chant
 if (MatCost.Size <> self.ClassAttribute.nbValues)
  then MatCost.reset(self.ClassAttribute.nbValues);
 //poursuivre normalement
 result:= inherited coreLearning(examples);
end;

function TCalcSpvCSTreeC45.getClassTreeStructureSpv: TClassMLTreeStructure;
begin
 result:= TMLTreeStructureSpvCSC45;
end;

{ TMLTreeNodeSpvCSC45 }

procedure TMLTreeNodeSpvCSC45.AssignConclusion;
var tree: TMLTreeStructureSpvCSC45;
    k,kMin,i: integer;
    sum,min: double;
    Lambda: double;
    marge: double;
begin
 tree:= self.TreeStructure as TMLTreeStructureSpvCSC45;
 Lambda:= (self.PrmMethod as TOpPrmSpvCSTreeC45).Lambda;
 min:= Math.MaxDouble;
 kMin:= -1;
 if (DistClass.NbExamples > 0)
  then
   begin
    //la marge corrige par le Lambda
    marge:= DistClass.NbExamples + Lambda * DistClass.Attribute.nbValues;
    //rechercher la prdiction qui minimise le cot de mauvaise affectation
    for k:= 1 to DistClass.Attribute.nbValues do
     begin
      sum:= 0.0;
      for i:= 1 to DistClass.Attribute.nbValues do
       //utiliser la distribution laplacienne
       sum:= sum + (Lambda + DistClass.TabFreq.Value[i]) / marge * tree.CostMatrix[i,k];
      //si plus petit
      if (sum < min)
       then
        begin
         min:= sum;
         kMin:= k;
        end;
     end;
   end;
 //affectation de la conclusion
 if (kMin > 0)
  then self.FNodeConclusion:= kMin
  //a ne devrait jamais arriver....sauf si le sommet est vide...
  else self.FNodeConclusion:= succ(random(DistClass.Attribute.nbValues));
 //calculer le cout de mauvais classement associ au noeud
 //en utilisant le lambda pour l'estimation des probabilits
 //>> c'est forcment "min" puisque c'est celle correspondant  la conclusion
 FPessimisticErr:= min
end;

function TMLTreeNodeSpvCSC45.getClassLstSplitAttributes: TClassLstSplitAttributes;
begin
 //pour la partie growing, rien ne change...
 result:= TLstSplitAttSpvCSC45; 
end;

function TMLTreeNodeSpvCSC45.isSatured: boolean;
begin
 //si le gain est nul, on ne poursuit pas...
 // la diffrence de C4.5, pas de hurdling ici...
 //si gain nul, on s'arrte
 result:= inherited isSatured() or (self.GoodnessSplit <= 0.0);
end;

{ TLstSplitAttSpvCSC45 }

function TLstSplitAttSpvCSC45.getClassSplitAttribut: TClassSplitAttribut;
begin
 result:= TSplitAttributSpvCSC45;
end;

{ TSplitAttributSpvCSC45 }

function TSplitAttributSpvCSC45.ComputeAcceptSplit: boolean;
var j: integer;
    nb: integer;
    minSize: integer;
    ok: boolean;
begin
 minSize:= (self.PrmMethod as TOpPrmSpvCSTreeC45).MinSizeLeaf;
 //
 ok:= (self.Count >= 2);
 //si ok = true, on sait qu'il y a au moins 2 feuilles dans ce split, sinon a ne sert  rien de tester le reste
 if ok
  then
   begin
    //chercher si toutes les feuilles sont de taille suprieure ou gale  min
    nb:= 0;
    for j:= 0 to pred(self.Count) do
     nb:= nb + ORD(self.SplitLeaf[j].NbExamples >= minSize);
    ok:= (nb = Self.Count);
   end;
 //
 result:= ok;
end;

function TSplitAttributSpvCSC45.getClassSplitLeaf: TClassSplitLeaf;
begin
 result:= TSplitLeafSpvCSC45;
end;

procedure TSplitAttributSpvCSC45.OptimizeContinuousLeaves;
var ls,rs: TSplitLeafSpvCSC45;
    exSorted: TExamples;
    i,example,numAtt: integer;
    v,vMax: double;
    m,mMax: TTypeContinue;
    n,vt: double;
    vcoupure: double;//valeur de coupure suggre par Quinlan (C4.5 - Release 8)
begin
 //toujours binaire donc
 ls:= self.SplitLeaf[0] as TSplitLeafSpvCSC45;
 rs:= self.SplitLeaf[1] as TSplitLeafSpvCSC45;
 //prendre la liste trie pour l'instant,  voir les optimisations plus tard

   //rcuprer celui envoy par le sommet - ouh l c'est pas beau
   numAtt:= Node.InputAttributes.GetIndex(Attribute);
   exSorted:= nil;
   if (numAtt<Node.InputSortedExamples.Count)
    then exSorted:= Node.InputSortedExamples.Items[numAtt] as TExamples;
   if not(assigned(exSorted)) OR (exSorted.Size<>Node.Examples.Size)
    //c'est la cata
    then
     begin
      TraceLog.WriteToLogFile('mauvaise rcupration de la liste trie');
      Raise Exception.Create('mauvaise rcupration de la liste trie');
     end;

 ls.Dist.BeginUpdate();//il n'y a personne sur cette portion pour l'instant
 rs.Dist.RefreshStat(Node.Examples);//il y a tout le monde sur cette portion pour l'instant
 mMax:= self.Attribute.cValue[exSorted.Number[1]];
 vMax:= 0.0;
 vt:= (Node as TMLTreeNodeSpvCSC45).DistClass.Shannon();
 n:= 1.0*(Node as TMLTreeNodeSpvCSC45).DistClass.TabFreq.Value[0];
 //pour chaque individu
 for i:= 1 to pred(exSorted.Size) do
  begin
   example:= exSorted.Number[i];
   ls.Dist.AddValue(example);
   rs.Dist.RemoveValue(example);
   //pas de tests  effectuer sur les ex-aequos
   if (self.Attribute.cValue[example]<self.Attribute.cValue[exSorted.Number[succ(i)]])
    then
     begin
      //
      m:= 0.5*(self.Attribute.cValue[example]+self.Attribute.cValue[exSorted.Number[succ(i)]]);
      //v:= self.ComputeGoodness();
      //plus rapide, plein de choses ne sont plus recalcules
      v:= -1.0*(vt-((1.0*ls.Dist.TabFreq.Value[0]/n)*ls.Dist.Shannon()+(1.0*rs.Dist.TabFreq.Value[0]/n)*rs.Dist.Shannon()));
      //suite standard
      if (v>vMax)
       then
        begin
         vMax:= v;
         mMax:= m;
        end;
     end;
  end;
 //calculer la valeur de coupure
 //if (compte > 0) and (n > 0) then vcoupure:= math.Log2(compte)/n else vcoupure:= 0.0;
 if (n > 1) then vcoupure:= math.log2(n-1)/n else vcoupure:= 0.0;
 //si a passe pas, on invalide en fixant la borne  la valeur minimale de la colonne
 if (vMax < vCoupure) then mMax:= self.Attribute.cValue[exSorted.Number[1]];
 //fichier log
 //TraceLog.WriteToLogFile(format('discretize %s at %.2f >> split value %.4f',[self.Attribute.Name,mMax,vMax]));
 //rafrachir la liste des enfants, on peut se passer du tri mais pas de cette partie....
 (ls.Condition as TRuleCondContinue).SetNewThresold(mMax);
 (rs.Condition as TRuleCondContinue).SetNewThresold(mMax);
 self.BeginUpdate();
 for i:= 1 to self.Node.Examples.Size do
  begin
   example:= self.Node.Examples.Number[i];
   m:= self.Attribute.cValue[example];
   //si ce n'est toi, c'est donc ton frre
   if ls.Condition.TestValue(m)
    then ls.AddExample(example)
    else rs.AddExample(example);
  end;
 self.EndUpdate();
end;

initialization
 RegisterClass(TMLGCompSpvCSTreeC45);
end.
