(******************************************************************)
(* UCompSpvCSTreeCART.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Cost-sensitive decision tree - algo CART)
@author(Ricco)
@created(29/12/2008)

Arbre dcision mthode CART - Introduction des cots de mauvais classement.

}

unit UCompSpvCSTreeCART;

interface

USES
        Forms, Classes, IniFiles,
        UCalcTreeStructureDefinition,
        UCompDefinition,
        UCompSpvLDefinition,
        UOperatorDefinition,
        UCompSpvTreeID3,
        UCalcRndGenerator,
        UFrmBaseOperator,
        UCompSpvTreeCART,
        UCompMetaSpvGenericCostSensitive,
        UCalcSpvTreeCART,
        UDatasetExamples,
        UDatasetdefinition, UDatasetImplementation;

TYPE
        {gnrateur de composant CART}
        TMLGCompSpvCSTreeCART = class(TMLGCompSpvTree)
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant SpvTree CART}
        TMLCompSpvCSTreeCART = class(TMLCompSpvTree)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {l'oprateur CART}
        TOpSpvCSTreeCART = class(TOpSpvTreeCART)
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    getClassSpvLearning(): TClassCalcSpvLearning; override;
        end;

        {paramtre de l'algo CART}
        TOpPrmSpvCSTreeCART = class(TOpPrmSpvTreeCART)
        private
        //matrice de cot
        FCostMatrix: TCostMatrix;
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        public
        {son propritaire est pass en paramtre : l'oprateur}
        constructor Create(prmOp: TOperator); override;
        destructor  Destroy(); override;
        function    getHTMLParameters(): string; override;
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        property    CostMatrix: TCostMatrix read FCostMatrix;
        end;

        {algo d'arbre de dcision supervise}
        TCalcSpvCSTreeCART = class(TCalcSpvTreeCART)
        protected
        function    getClassTreeStructureSpv(): TClassMLTreeStructure; override;
        function    coreLearning(examples: TExamples): boolean; override;
        function    getHTMLHeaderSequence(): string; override;
        end;

        //structure de noeud - arbre CART avec gestion des cots
        TMLTreeNodeSpvCSCART = class(TMLTreeNodeSpvCART)
        protected
        //cout en resubstitution
        FGrowErr: double;
        //assigner la conclusion
        procedure AssignConclusion(); override;
        public
        function  localCost(prmSet: TEnumCartDataSet): double; override;
        end;

        //structure d'arbre CART avec gestion des cots
        TMLTreeStructureSpvCSCART = class(TMLTreeStructureSpvCART)
        private
        //matrice de cot, pas propritaire
        FCostMatrix: TCostMatrix;
        protected
        function   getClassMLTreeNode(): TClassMLTreeNode; override;
        procedure  applyPruningExample(node: TMLTreeNode; example: integer); override;
        function   enumerateSubTrees(nbTrees: integer): integer; override;
        //calculer l'cart quadratique
        function   getQuadratic(prmNsa: integer; meanRef: double): double;
        //calculer le cot pour un individu associ  un sous-arbre
        function   getCost(example: integer; prmNsa: integer): double;
        public
        constructor create(paramMethod: TOperatorParameter; prmTarget,prmInput: TLstAttributes; prmExamples: TExamples); override;
        property   CostMatrix: TCostMatrix read FCostMatrix;
        end;

implementation

USES
        Math, UCalcStatDes, ULogFile, Sysutils, UDlgOpPrmSpvTreeCSCART,
  UConstConfiguration;

{ TMLGCompSpvCSTreeCART }

function TMLGCompSpvCSTreeCART.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvCSTreeCART;
end;

{ TMLCompSpvCSTreeCART }

function TMLCompSpvCSTreeCART.getClassOperator: TClassOperator;
begin
 result:= TOpSpvCSTreeCART;
end;

{ TOpSpvCSTreeCART }

function TOpSpvCSTreeCART.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSpvCSTreeCART;
end;

function TOpSpvCSTreeCART.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcSpvCSTreeCART;
end;

{ TOpPrmSpvCSTreeCART }

constructor TOpPrmSpvCSTreeCART.Create(prmOp: TOperator);
begin
 inherited;
 FCostMatrix:= TCostMatrix.create(2);//par dfaut, mais modifiable facilement
end;

function TOpPrmSpvCSTreeCART.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSpvTreeCSCART.CreateFromOpPrm(self);
end;

destructor TOpPrmSpvCSTreeCART.Destroy;
begin
 FCostMatrix.Free();
 inherited;
end;

function TOpPrmSpvCSTreeCART.getHTMLParameters: string;
begin
 result:= inherited getHTMLParameters();
 result:= result + FCostMatrix.getHTML();
end;

procedure TOpPrmSpvCSTreeCART.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 FCostMatrix.LoadFromINI(prmSection,prmINI);
end;

procedure TOpPrmSpvCSTreeCART.LoadFromStream(prmStream: TStream);
begin
 inherited;
 FCostMatrix.LoadFromStream(prmStream);
end;

procedure TOpPrmSpvCSTreeCART.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 inherited;
 FCostMatrix.SaveToINI(prmSection,prmINI);
end;

procedure TOpPrmSpvCSTreeCART.SaveToStream(prmStream: TStream);
begin
 inherited;
 FCostMatrix.SaveToStream(prmStream);
end;

procedure TOpPrmSpvCSTreeCART.SetDefaultParameters;
begin
 inherited;
end;

{ TCalcSpvCSTreeCART }

function TCalcSpvCSTreeCART.coreLearning(examples: TExamples): boolean;
var MatCost: TCostMatrix;
begin
 MatCost:= (self.OpPrmSpv as TOpPrmSpvCSTreeCART).CostMatrix;
 //vrifier la matrice de cot et assurer le coup le cas chant
 if (MatCost.Size <> self.ClassAttribute.nbValues)
  then MatCost.reset(self.ClassAttribute.nbValues);
 //poursuivre normalement
 result:= inherited coreLearning(examples);
end;

function TCalcSpvCSTreeCART.getClassTreeStructureSpv: TClassMLTreeStructure;
begin
 result:= TMLTreeStructureSpvCSCART;
end;

function TCalcSpvCSTreeCART.getHTMLHeaderSequence: string;
begin
 result:= HTML_TABLE_COLOR_HEADER_GRAY +
         '<TH>N</TH><TH># Leaves</TH><TH>Cost (growing set)</TH><TH>Cost (pruning set)</TH><TH>SE (pruning set)</TH><TH width=60>x</TH></TR>';
end;

{ TMLTreeStructureSpvCSCART }

procedure TMLTreeStructureSpvCSCART.applyPruningExample(node: TMLTreeNode;
  example: integer);
var curNode: TMLTreeNodeSpvCART;
    nextNode: TMLTreeNode; 
    i: integer;
begin
 curNode:= node as TMLTreeNodeSpvCSCART;
 //ajouter l'exemple dans le comptage
 curNode.PruningCount:= curNode.PruningCount + 1; 
 //ajouter le cot de mauvais classement
 curNode.PruningErr:= curNode.PruningErr + self.CostMatrix[(self as TMLTreeStructureSpvCART).ClassAttribute.dValue[example],curNode.NodeConclusion];
 //passer au suivant en dtectant la feuille concerne
 if not(curNode.isLeaf)
  then
   begin
    //chercher le sommet  appeler
    nextNode:= NIL;
    for i:= 0 to pred(curNode.getCountSuccessors()) do
     begin
      if curNode.getSuccessor(i).Condition.TestExample(example)
       then
        begin
         nextNode:= curNode.getSuccessor(i);
         BREAK;
        end;
     end;
    //alors, alors, s'il n'y a rien, on est mal parce que c'est une valeur non-rfrence !!!
    if assigned(nextNode)
     then self.applyPruningExample(nextNode,example)
     else TraceLog.WriteToLogFile(format('[CART] !!!! ERREUR !!!, valeur non rfrence dans PRUNING, example n %d',[example]));
   end;
end;

constructor TMLTreeStructureSpvCSCART.create(
  paramMethod: TOperatorParameter; prmTarget, prmInput: TLstAttributes;
  prmExamples: TExamples);
begin
 inherited;
 //brancher seulement
 FCostMatrix:= (paramMethod as TOpPrmSpvCSTreeCART).CostMatrix;
end;

function TMLTreeStructureSpvCSCART.getClassMLTreeNode: TClassMLTreeNode;
begin
 result:= TMLTreeNodeSpvCSCART;
end;

function TMLTreeStructureSpvCSCART.enumerateSubTrees(nbTrees: integer): integer;
var subTree,nbLeaves,optSubTree: integer;
    pruneCost,growCost,minCost,thresold,seMinCost,sePrune,seMin: double;
begin
 SetLength(FTabTreeSequence,nbTrees);
 //lister les squences avec leurs perfs. respectives
 //en profiter pour dtecter l'arbre optimal
 FOptimalSubTree:= -1;
 minCost:= Math.MaxDouble;
 seMin:= 0.0;
 for subTree:= 1 to nbTrees do
  begin
   pruneCost:= self.computePrunedSubTreeCosts(subTree,nbLeaves,growCost);
   pruneCost:= pruneCost/(1.0*self.FPruningSet.Size);
   growCost:= growCost/(1.0*self.FTrainingSet.Size);
   //nouveaut, calculer la somme des carts quadratiques
   sePrune:= self.getQuadratic(subTree,pruneCost);
   sePrune:= SQRT(sePrune/power(1.0*self.FPruningSet.Size,2.0));
   //affecter les valeurs
   FTabTreeSequence[pred(subTree)].NumTree:= subTree;
   FTabTreeSequence[pred(subTree)].NbLeaves:= nbLeaves;
   FTabTreeSequence[pred(subTree)].ErrPrune:= pruneCost;
   FTabTreeSequence[pred(subTree)].ErrGrow:= growCost;
   FTabTreeSequence[pred(subTree)].StdDevPrune:= sePrune;
   //new -- 02/01/2010 -- la valeur de theta
   FTabTreeSequence[pred(subTree)].Theta:= -1.0;
   //tester
   if (pruneCost<=minCost)
    then
     begin
      minCost:= pruneCost;
      seMin:= sePrune;
      FOptimalSubTree:= subTree;
     end;
  end;
 //ok, on le bon numro
 //rcuprer l'cart-type du plus petit
 seMinCost:= seMin;
 //appliquer la xxx-SE RULE pour trouver le bon seuil
 thresold:= minCost+(self.PrmMethod as TOpPrmSpvCSTreeCART).SERule*seMinCost;
 //new --02/01/2010
 //pour tous les arbres, plus petits que FOptimalSubtree - calculer la valeur de theta pour l'obtenir
 if (seMinCost > 0.0)
  then for subTree:= nbTrees downto FOptimalSubTree do FTabTreeSequence[pred(subTree)].Theta:= (FTabTreeSequence[pred(subTree)].ErrPrune - minCost) / seMinCost;
 //suite...
 //rechercher le "bon"  partir des moins complexes vers les plus complexes
 optSubTree:= -1;
 for subTree:= nbTrees downto 1 do
  begin
   if (FTabTreeSequence[pred(subTree)].ErrPrune <= thresold)
    then
     begin
      optSubTree:= subTree;
      BREAK;
     end;
  end;
 //TraceLog.WriteToLogFile(format('[CART] the selected tree is n%2d, with err = %.4f',[optSubTree,FTabTreeSequence[pred(optSubTree)].ErrPrune]));
 //renvoyer la bonne rponse
 result:= optSubTree;
end;

function TMLTreeStructureSpvCSCART.getCost(example,
  prmNsa: integer): double;
var node: TMLTreeNodeSpvCSCART;
    successor: TMLTreeNode;
    i: integer;
begin
 node:= self.RootNode as TMLTreeNodeSpvCSCART;
 while (node.CurNsa > prmNsa) and not(node.isLeaf) do
  begin
   //dtecter le noeud suivant  dclencher
   i:= 0;
   repeat
    successor:= node.getSuccessor(i);
    inc(i);
   until (successor.Condition.TestValue(successor.Condition.Attribute.cValue[example]));
   //passer au suivant donc - mieux vaut ne trouver ici sinon on est mal
   node:= successor as TMLTreeNodeSpvCSCART;
  end;
 //c'est une feuille, on prend la conclusion -- 24/04/2005 -- on peut le faire sans calcul du score !
 result:= self.CostMatrix[self.ClassAttribute.dValue[example],node.NodeConclusion];
end;

function TMLTreeStructureSpvCSCART.getQuadratic(prmNsa: integer;
  meanRef: double): double;
var i: integer;
    quad,cost: double;
begin
 quad:= 0.0;
 //pour chaque individu de l'chantillon test
 for i:= 1 to self.FPruningSet.Size do
  begin
   cost:= getCost(self.FPruningSet.Number[i],prmNsa);
   quad:= quad + SQR(cost-meanRef);
  end;
 result:= quad;
end;

{ TMLTreeNodeSpvCSCART }

procedure TMLTreeNodeSpvCSCART.AssignConclusion;
var tree: TMLTreeStructureSpvCSCART;
    k,kMin,i: integer;
    sum,min: double;
begin
 tree:= self.TreeStructure as TMLTreeStructureSpvCSCART;
 min:= Math.MaxDouble;
 kMin:= -1;
 if (DistClass.NbExamples > 0)
  then
   begin
    for k:= 1 to DistClass.Attribute.nbValues do
     begin
      sum:= 0.0;
      for i:= 1 to DistClass.Attribute.nbValues do
       sum:= sum + DistClass.TabFreq.Value[i]*tree.CostMatrix[i,k];
      //si plus petit
      if (sum < min)
       then
        begin
         min:= sum;
         kMin:= k;
        end;
     end;
   end;
 //affectation de la conclusion
 if (kMin > 0)
  then self.FNodeConclusion:= kMin
  //a ne devrait jamais arriver....sauf si le sommet est vide...
  else self.FNodeConclusion:= succ(random(DistClass.Attribute.nbValues));
 //calculer le cout de mauvais classement associ au noeud
 FGrowErr:= 0.0;
 for i:= 1 to DistClass.Attribute.nbValues do
  FGrowErr:= FGrowErr + DistClass.TabFreq.Value[i]*tree.CostMatrix[i,self.FNodeConclusion];
end;

function TMLTreeNodeSpvCSCART.localCost(prmSet: TEnumCartDataSet): double;
begin
 if (prmSet = cartPruningSet)
  then result:= inherited localCost(prmSet)
  else result:= self.FGrowErr;
end;

initialization
 RegisterClass(TMLGCompSpvCSTreeCART);
end.
