(**************************************************************)
(* UCompSpvTrirls.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(**************************************************************)

{
@abstract(Utilisation de la rgression logistique de la bibliothque LR-TRIRLS)
@author(Ricco)
@created(18/02/2006)

}
unit UCompSpvTrirls;

interface

USES
       Classes, Forms, IniFiles,
       UCompDefinition,
       UCompSpvLDefinition,
       UOperatorDefinition,
       UDatasetImplementation;

TYPE
       TMLGCompSpvTRIRLS = class(TMLGenCompSpvLearning)
                         public
                         function    GetClassMLComponent: TClassMLComponent; override;
                         end;

       TMLCompSpvTRIRLS = class(TMLCompSpvLearning)
                        protected
                        function    getClassOperator: TClassOperator; override;
                        function    GetLogResultDescription(): string; override;
                        end;

       TOpSpvTRIRLS = class(TOpSpvLearningContinuous)
                    protected
                    function    getClassParameter: TClassOperatorParameter; override;
                    function    getClassSpvLearning(): TClassCalcSpvLearning; override;
                    function    ConnectDescriptors(prmData: TMLDataset): boolean; override;
                    end;

       TOpPrmTRIRLS = class(TOpPrmSpvLearning)
                    private
                    protected
                    function    CreateDlgParameters(): TForm; override;
                    procedure   SetDefaultParameters(); override;
                    public
                    rrlambda: double;
                    cgdeveps: double;
                    cgmax: integer;
                    lreps: double;
                    lrmax: integer;
                    function    getHTMLParameters(): string; override;
                    procedure   LoadFromStream(prmStream: TStream); override;
                    procedure   SaveToStream(prmStream: TStream); override;
                    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
                    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
                    end;

implementation

uses
    Sysutils, UDatasetDefinition, UCalcSpvTrirls, UImportTRIRLS,
    UConstConfiguration, UDlgOpPrmSpvTRIRLS;

{ TMLGCompSpvTRIRLS }

function TMLGCompSpvTRIRLS.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvTRIRLS;
end;

{ TMLCompSpvTRIRLS }

function TMLCompSpvTRIRLS.getClassOperator: TClassOperator;
begin
 result:= TOpSpvTRIRLS; 
end;

function TMLCompSpvTRIRLS.GetLogResultDescription: string;
begin
 result:= '[IRLS] from LR-TRIRLS generated';
end;

{ TOpSpvTRIRLS }

function TOpSpvTRIRLS.ConnectDescriptors(prmData: TMLDataset): boolean;
var ok: boolean;
begin
 //accepter les prdicteurs quasi-continus, i.e. continus ou discrets binaires
 ok:= (prmData.LstAtts[asInput].Count > 0) AND (prmData.LstAtts[asInput].isAllCategory(caQuasiContinue));
 if ok then FDescriptorsAtt:= prmData.LstAtts[asInput];//branchement direct, pas de recopie locale
 //a marche que pour le binaire
 ok:= ok and (self.ClassAttribute.nbValues = 2);
 //and then...
 result:= ok;
end;

function TOpSpvTRIRLS.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmTRIRLS;
end;

function TOpSpvTRIRLS.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcSpvTrirls;
end;

{ TOpPrmTRIRLS }

function TOpPrmTRIRLS.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSpvTRIRLS.CreateFromOpPrm(self);
end;

function TOpPrmTRIRLS.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Parameters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Ridge regression parameter (lambda)</TD><TD align=right>%g</TD></TR>',[rrlambda]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Conjugate gradient iterations (cgdeveps)</TD><TD align=right>%g</TD></TR>',[cgdeveps]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Conjugate gradient iterations (cgmax)</TD><TD align=right>%d</TD></TR>',[cgmax]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>IRLS iterations (lreps)</TD><TD align=right>%g</TD></TR>',[lreps]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>IRLS iterations (lrmax)</TD><TD align=right>%d</TD></TR>',[lrmax]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmTRIRLS.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 rrlambda:= prmINI.ReadFloat(prmSection,'rrlambda',rrlambda);
 cgdeveps:= prmINI.ReadFloat(prmSection,'cgdeveps',cgdeveps);
 cgmax:= prmINI.ReadInteger(prmSection,'cgmax',cgmax);
 lreps:= prmINI.ReadFloat(prmSection,'lreps',lreps);
 lrmax:= prmINI.ReadInteger(prmSection,'lrmax',lrmax);
end;

procedure TOpPrmTRIRLS.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(rrlambda, sizeof(rrlambda));
 prmStream.ReadBuffer(cgdeveps, sizeof(cgdeveps));
 prmStream.ReadBuffer(cgmax, sizeof(cgmax));
 prmStream.ReadBuffer(lreps, sizeof(lreps));
 prmStream.ReadBuffer(lrmax, sizeof(lrmax));
end;

procedure TOpPrmTRIRLS.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteFloat(prmSection,'rrlambda',rrlambda);
 prmINI.WriteFloat(prmSection,'cgdeveps',cgdeveps);
 prmINI.WriteInteger(prmSection,'cgmax',cgmax);
 prmINI.WriteFloat(prmSection,'lreps',lreps);
 prmINI.WriteInteger(prmSection,'lrmax',lrmax);
end;

procedure TOpPrmTRIRLS.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(rrlambda, sizeof(rrlambda));
 prmStream.WriteBuffer(cgdeveps, sizeof(cgdeveps));
 prmStream.WriteBuffer(cgmax, sizeof(cgmax));
 prmStream.WriteBuffer(lreps, sizeof(lreps));
 prmStream.WriteBuffer(lrmax, sizeof(lrmax));
end;

procedure TOpPrmTRIRLS.SetDefaultParameters;
begin
  rrlambda:= 1;
  cgdeveps:= 1.0e-6;
  cgmax:= 200;
  lreps:= 1.0e-6;
  lrmax:= 30;
end;

initialization
 RegisterClass(TMLGCompSpvTRIRLS);
end.
