(****************************************************************)
(* UCompSpvLibcmBVM.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(****************************************************************)

{
@abstract(Utilisation de la mthode BVM de la bibliothque LIBCVM)
@author(Ricco)
@created(06/05/2012)

Intgration de la mthode BVM (Ball Vector Machine) de la bibliothque LIBCVM
http://c2inet.sce.ntu.edu.sg/ivor/cvm.html

}
unit UCompSpvLibcmBVM;

interface

USES
       Classes, Forms, IniFiles,
       UCompDefinition,
       UCompSpvLDefinition,
       UOperatorDefinition,
       UDatasetImplementation,
       UCompSpvLibcvmCVM,
       UCalcLibcvmCVM;

TYPE

      TCalcLibcvmBvm = class(TCalcLibcvmCvm)
      protected
      //rcuprer les paramtres en provenance de Tanagra --> en virtuel, surcharg pour BVM
      procedure getUserParameters(); override;
      end;

      TMLGCompSpvBVM = class(TMLGenCompSpvLearning)
      public
      function    GetClassMLComponent: TClassMLComponent; override;
      end;

      TMLCompSpvBVM = class(TMLCompSpvLearning)
      protected
      function    getClassOperator: TClassOperator; override;
      function    GetLogResultDescription(): string; override;
      end;

      TOpSpvBVM = class(TOpSpvCVM)
      protected
      function    getClassParameter: TClassOperatorParameter; override;
      function    getClassSpvLearning(): TClassCalcSpvLearning; override;
      end;

      TOpPrmBVM = class(TOpPrmCVM)
      protected
      function    CreateDlgParameters(): TForm; override;
      procedure   SetDefaultParameters(); override;
      public
      sample_size: integer;
      function    getHTMLTableParameters(): string; override;
      procedure   LoadFromStream(prmStream: TStream); override;
      procedure   SaveToStream(prmStream: TStream); override;
      procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
      procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
      end;


implementation

uses
     Sysutils,
     UConstConfiguration, ULibCVMImport,
     UDlgOpPrmSpvBVM;

{ TCalcLibcvmBvm }

procedure TCalcLibcvmBvm.getUserParameters;
begin
 inherited;
 FParam^.svm_type := ord(BVM);//<-- type de mthode
 FParam^.sample_size:= (self.OpPrmSpv as TOpPrmBVM).sample_size;//taille de l'chantillon
end;

{ TMLGCompSpvBVM }

function TMLGCompSpvBVM.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvBVM;
end;

{ TMLCompSpvBVM }

function TMLCompSpvBVM.getClassOperator: TClassOperator;
begin
 result:= TOpSpvBVM;
end;

function TMLCompSpvBVM.GetLogResultDescription: string;
begin
  result:= '[BVM - Ball Vector Machine] from LIBCVM generated';
end;

{ TOpSpvBVM }

function TOpSpvBVM.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmBVM;
end;

function TOpSpvBVM.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcLibcvmBvm;
end;

{ TOpPrmBVM }

function TOpPrmBVM.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSpvBvm.CreateFromOpPrm(self); 
end;

function TOpPrmBVM.getHTMLTableParameters: string;
begin
 result:= inherited getHTMLTableParameters();
 result:= result + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Sample size for probabilistic sampling</TD><TD align=right>%d</TD></TR>',[sample_size]);
end;

procedure TOpPrmBVM.LoadFromINI(prmSection: string; prmINI: TMemIniFile);
begin
 inherited;
 sample_size:= prmINI.ReadInteger(prmSection,'sample_size',sample_size);
end;

procedure TOpPrmBVM.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(sample_size,sizeof(sample_size));
end;

procedure TOpPrmBVM.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'sample_size',sample_size);
end;

procedure TOpPrmBVM.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(sample_size,sizeof(sample_size));
end;

procedure TOpPrmBVM.SetDefaultParameters;
begin
 inherited;
 _kernel_type:= ord(RBF);//attention, seuls certains types sont autoriss
 sample_size:= 60;
end;

initialization
 RegisterClass(TMLGCompSpvBVM);
end.
