(*****************************************************************)
(* UCompSpvLibcvmCVM.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(*****************************************************************)

{
@abstract(Utilisation de la mthode CVM de la bibliothque LIBCVM)
@author(Ricco)
@created(06/05/2012)

Intgration de la mthode CVM (Core Vector Machine) de la bibliothque LIBCVM
http://c2inet.sce.ntu.edu.sg/ivor/cvm.html

}

unit UCompSpvLibcvmCVM;

interface

USES
       Classes, Forms, IniFiles,
       UCompDefinition,
       UCompSpvLDefinition,
       UOperatorDefinition,
       UDatasetImplementation,
       UCompSpvLibSvmCsvc;

TYPE

      TMLGCompSpvCVM = class(TMLGenCompSpvLearning)
      public
      function    GetClassMLComponent: TClassMLComponent; override;
      end;

      TMLCompSpvCVM = class(TMLCompSpvLearning)
      protected
      function    getClassOperator: TClassOperator; override;
      function    GetLogResultDescription(): string; override;
      end;

      TOpSpvCVM = class(TOpSpvCSVC)
      protected
      function    getClassParameter: TClassOperatorParameter; override;
      function    getClassSpvLearning(): TClassCalcSpvLearning; override;
      end;

      TOpPrmCVM = class(TOpPrmCSVC)
      protected
      function    CreateDlgParameters(): TForm; override;
      procedure   SetDefaultParameters(); override;
      public
      num_basis: integer; //nombre max. de core vectors
      function    getHTMLTableParameters(): string; override;
      procedure   LoadFromStream(prmStream: TStream); override;
      procedure   SaveToStream(prmStream: TStream); override;
      procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
      procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
      end;


implementation

uses
     Sysutils,   
     UConstConfiguration,
     UCalcLibcvmCVM,
     UDlgOpPrmSpvCVM, ULibCVMImport;

{ TMLGCompSpvCVM }

function TMLGCompSpvCVM.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvCVM;
end;

{ TMLCompSpvCVM }

function TMLCompSpvCVM.getClassOperator: TClassOperator;
begin
 result:= TOpSpvCVM;
end;

function TMLCompSpvCVM.GetLogResultDescription: string;
begin
 result:= '[CVM - Core Vector Machine] from LIBCVM generated';
end;

{ TOpSpvCVM }

function TOpSpvCVM.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmCVM;
end;

function TOpSpvCVM.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcLibcvmCvm;
end;

{ TOpPrmCVM }

function TOpPrmCVM.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSpvCvm.CreateFromOpPrm(self);
end;

function TOpPrmCVM.getHTMLTableParameters: string;
var s: string;
begin
 s:= '';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Kernel type</TD><TD align=right>%s</TD></TR>',[str_kernel_type[kernel_type(_kernel_type)]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Degree (poly/norm poly)</TD><TD align=right>%.2f</TD></TR>',[degree]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Gamma (poly/rbf/sigmoid/laplacian/norm poly/inv dist/inv sq dist)</TD><TD align=right>%.2g</TD></TR>',[gamma]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Coef0 (poly/sigmoid/norm poly)</TD><TD align=right>%.4g</TD></TR>',[Coef0]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Tolerance of termination criteria (eps)</TD><TD align=right>%.4g</TD></TR>',[eps]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>C (Complexity Cost)</TD><TD align=right>%.2g</TD></TR>',[C]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Compute probability estimates</TD><TD align=right>%d</TD></TR>',[probability]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Use shrinking heuristics</TD><TD align=right>%d</TD></TR>',[shrinking]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Data normalization</TD><TD align=right>%d</TD></TR>',[normalization]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Max # of core vectors</TD><TD align=right>%d</TD></TR>',[num_basis]);
 result:= s;
end;

procedure TOpPrmCVM.LoadFromINI(prmSection: string; prmINI: TMemIniFile);
begin
 inherited;
 num_basis:= prmINI.ReadInteger(prmSection,'num_basis',num_basis);
end;

procedure TOpPrmCVM.LoadFromStream(prmStream: TStream);
begin
 inherited;
 prmStream.ReadBuffer(num_basis,sizeof(num_basis));
end;

procedure TOpPrmCVM.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 inherited;
 prmINI.WriteInteger(prmSection,'num_basis',num_basis);
end;

procedure TOpPrmCVM.SaveToStream(prmStream: TStream);
begin
 inherited;
 prmStream.WriteBuffer(num_basis,sizeof(num_basis));
end;

procedure TOpPrmCVM.SetDefaultParameters;
begin
 inherited;
 //cf. http://c2inet.sce.ntu.edu.sg/ivor/cvm.html pour CVM/BVM
 self._kernel_type:= ord(RBF);
 self.C:= 1;
 self.gamma:= 1;
 self.eps:= 1e-4;//1e-4 (cf. article Tsang et al., 2007)
 self.num_basis:= 50000;//50000 (jamais mentionn dans les articles ?)
end;

initialization
 RegisterClass(TMLGCompSpvCVM);
end.
