(*************************************************************************)
(* UCompSpvLogisticRegHLTest.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(*************************************************************************)

{
@abstract(Rgression logistique, Test de Hosmer et Lemeshow pour la rgression logistique)
@author(Ricco)
@created(25/09/2009)

Tester de Hosmer &Lemeshow,  plugger juste aprs la rgression logistique one instance.
On ne peut rien mettre en sortie.

}


unit UCompSpvLogisticRegHLTest;

interface

USES
        Forms, Contnrs,
        UCompDefinition,
        UOperatorDefinition,
        UDatasetImplementation,
        UCompSpvLogisticRegression,
        UCompManageDataset;

TYPE
        {gnrateur de composant}
        TMLGenHosmerLemeshowTest = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {Composant}
        TMLHosmerLemeshowTest = class(TMLCompLogisticRegAfter)
        protected
        function    getClassOperator: TClassOperator; override;
        public
        //non, il n'accepte rien en aval
        function    isConnectableBy(prmGen: TMLGenComp): boolean; override;
        end;

        {Oprateur}
        TOpHosmerLemeshowTest = class(TOpLocalData)
        private
        //classifieur branch
        FClassifier: TCalcLogisticReg;
        //liste des lignes du tableau de HL
        FLstRow: TObjectList;
        //statistique
        FChi2, FPValue: double;
        FDf: integer;
        //calculer les scores pour tous les individus
        procedure   computeScores(var score: TAttContinue);
        //calculer la statistique
        procedure   computeStatistics();
        protected
        function    CoreExecute(): boolean; override;
        function    getClassParameter: TClassOperatorParameter; override;
        public
        destructor  Destroy; override;
        function    getHTMLResultsSummary(): string; override;
        end;

        {Paramtrage}
        TOpPrmHosmerLemeshowTest = class(TOperatorParameter)
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        end;

        //une ligne du tableau de HL
        TRowHLTest = class
        //valeur du dcile
        decile: double;
        //effectif
        nbObs: double;
        //observs
        obsPos, obsNeg: double;
        //ngatifs
        theoPos, theoNeg: double;
        //
        constructor create(seuil: double);
        end;



implementation

uses
    Classes, Sysutils,
    UDatasetDefinition, UDatasetExamples,
    UCompMetaSpvOneInstance, ULogFile, UConstConfiguration, fmath;

{ TMLGenHosmerLemeshowTest }

procedure TMLGenHosmerLemeshowTest.GenCompInitializations;
begin
 FMLComp:= mlcSpvAssessment;
end;

function TMLGenHosmerLemeshowTest.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLHosmerLemeshowTest;
end;

{ TMLHosmerLemeshowTest }

function TMLHosmerLemeshowTest.getClassOperator: TClassOperator;
begin
 result:= TOpHosmerLemeshowTest;
end;

function TMLHosmerLemeshowTest.isConnectableBy(
  prmGen: TMLGenComp): boolean;
begin
 result:= FALSE;
end;

{ TOpPrmHosmerLemeshowTest }

function TOpPrmHosmerLemeshowTest.CreateDlgParameters: TForm;
begin
 result:= nil;
end;

procedure TOpPrmHosmerLemeshowTest.SetDefaultParameters;
begin
 //none
end;

{ TOpHosmerLemeshowTest }

procedure TOpHosmerLemeshowTest.computeScores(var score: TAttContinue);
var i,j,n: integer;
    p: double;
begin
 n:= FClassifier.ClassAttribute.Size;
 score:= TAttContinue.Create('_score_',n);
 for i:= 1 to n do
  begin
   p:= FClassifier.Coef[0];
   for j:= 0 to pred(FClassifier.Descriptors.Count) do
    p:= p + FClassifier.Coef[j+1] * FClassifier.Descriptors.Attribute[j].cValue[i];
   p:= 1.0 / (1.0 + exp(-1.0 * p));
   score.cValue[i]:= p;
  end;
end;

procedure TOpHosmerLemeshowTest.computeStatistics;
var k: integer;
    row: TRowHLTest;
    v: double;
begin
 FDf:= FLstRow.Count - 2;
 FChi2:= 0.0;
 for k:= 0 to pred(FLstRow.Count) do
  begin
   row:= FLstRow.Items[k] as TRowHLTest;
   v:= 0.0;
   if (row.theoPos > 0.0) then v:= v + SQR(row.obsPos - row.theoPos)/row.theoPos;
   if (row.theoNeg > 0.0) then v:= v + SQR(row.obsNeg - row.theoNeg)/row.theoNeg;
   FChi2:= FChi2 + v;
  end;
 FPValue:= PKhi2(FDf,FChi2); 
end;

function TOpHosmerLemeshowTest.CoreExecute: boolean;
var score: TAttContinue;
    exSort: TExamples;
    saut: integer;
    predSeuil,seuil: double;
    row: TRowHLTest;
    k,i,example: integer;
    v: double;
begin
 //brancher le classifieur
 FClassifier:= (((MLOwner as TMLHosmerLemeshowTest).Predecessor as TMLCompOneInstance).Operator as TOpOneInstance).Classifier as TCalcLogisticReg;
 //lancer les calculs
 result:= true;
 exSort:= nil;
 TRY
 //obtenir les scores de "+" de tous les individus
 self.computeScores(score);
 //trier l'index selon le score croissant
 exSort:= TExamples.Create(FClassifier.CurExamples.Size);
 exSort.Copy(FClassifier.CurExamples);
 exSort.QuickSortBy(score);
 //dcouper en dciles puis crer les lignes du tableau de HL
 if (FLstRow <> nil) then FLstRow.Free();
 FLstRow:= TObjectList.Create(TRUE);
 saut:= round(1.0 * exSort.Size / 10.0);
 //crer les seuils successifs
 predSeuil:= score.cValue[exSort.Number[saut]];
 row:= TRowHLTest.create(predSeuil);
 FLstRow.Add(row);
 for k:= 2 to 9 do
  begin
   seuil:= score.cValue[exSort.Number[k*saut]];
   if (seuil > predSeuil)
    then
     begin
      row:= TRowHLTest.create(seuil);
      FLstRow.Add(row);
      predSeuil:= seuil;
     end;
  end;
 //dernier seuil
 seuil:= score.cValue[exSort.Number[exSort.Size]];
 if (seuil > predSeuil)
  then
   begin
    row:= TRowHLTest.create(seuil);
    FLstRow.Add(row);
   end;
 //remplir maintenant
 k:= 0;
 row:= FLstRow.Items[k] as TRowHLTest;
 for i:= 1 to exSort.Size do
  begin
   example:= exSort.Number[i];
   //le score
   v:= score.cValue[example];
   //dans le bon dcile ?
   if (v > row.decile)
    then
     begin
      inc(k);
      if (k < FLstRow.Count)
       then row:= FLstRow.Items[k] as TRowHLTest;
     end;
   //rajouter
   row.nbObs:= row.nbObs + 1.0;
   if (FClassifier.ClassAttribute.dValue[example] = 1)
    then row.obsPos:= row.obsPos + 1.0;
   row.theoPos:= row.theoPos + v;
  end;
 //recap.
 for k:= 0 to pred(FLstRow.Count) do
  begin
   row:= FLstRow.Items[k] as TRowHLTest;
   row.obsNeg:= row.nbObs - row.obsPos;
   row.theoNeg:= row.nbObs - row.theoPos;
  end;
 //calculer la statistique
 self.computeStatistics();
 EXCEPT
 result:= false;
 END;
 if (score <> nil) then score.Free();
 if (exSort <> nil) then exSort.Free();
end;

destructor TOpHosmerLemeshowTest.Destroy;
begin
  if (FLstRow <> nil) then FreeAndNil(FLstRow);
  inherited;
end;

function TOpHosmerLemeshowTest.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmHosmerLemeshowTest;
end;

function TOpHosmerLemeshowTest.getHTMLResultsSummary: string;
var s: string;
    k: integer;
    row: TRowHLTest;
begin
 s:= '<H3>Hosmer Lemeshow Goodness-of-Fit Test</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY +
     '<TH colspan=2>&nbsp;</TH><TH colspan=2>Positive</TH><TH colspan=2>Negative</TH><TH>&nbsp;</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY +
     '<TH width=60>Decile</TH><TH width=60>Prob.</TH><TH>Observed</TH><TH>Expected</TH><TH>Observed</TH><TH>Expected</TH><TH width=60>Total</TH></TR>';
 for k:= 0 to pred(FLstRow.Count) do
  begin
   row:= FLstRow.Items[k] as TRowHLTest;
   s:= s + HTML_TABLE_COLOR_DATA_GRAY +
       format('<TD align="center">%d</TD><TD align=center>%.3f</TD><TD align=right>%.3f</TD><TD align=right>%.3f</TD><TD align=right>%.3f</TD><TD align=right>%.3f</TD><TD align=right>%.3f</TD></TR>',
       [succ(k),row.decile,row.obsPos,row.theoPos,row.obsNeg,row.theoNeg,row.nbObs]);
  end;
 s:= s + '</table>';
 s:= s + '<H4>Hosmer Lemeshow Statistic</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>&nbsp;</TH><TH>Chi-Square</TH><TH align=right>d.f.</TH><TH>Significance</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Goodness-of-fit test</TD><TD align=right>%.4f</TD><TD align=right>%d</TD><TD align=right>%.4f</TD></TR>',
         [FChi2, FDf, FPValue]);
 s:= s + '</table>';
 result:= s;
end;

{ TRowHLTest }

constructor TRowHLTest.create(seuil: double);
begin
 inherited Create();
 decile:= seuil;
 nbObs:= 0;
 obsPos:= 0;
 obsNeg:= 0;
 theoPos:= 0;
 theoNeg:= 0;
end;

initialization
 RegisterClass(TMLGenHosmerLemeshowTest);
end.
