(*************************************************************************)
(* UCompSpvLogisticRegHLTest.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(*************************************************************************)

{
@abstract(Rgression logistique, Analyse des rsidus)
@author(Ricco)
@created(28/09/2009)

Calcul des rsidus de la rgression logistique
}

unit UCompSpvLogisticRegResiduals;

interface

USES
        Forms, Contnrs,
        UCompDefinition, UDatasetDefinition,
        UOperatorDefinition,
        UCompSpvLogisticRegression,
        UCompManageDataset;

TYPE
        TLogRegResiduals = (lrHat, lrPearson, lrStdPearson, lrDifChisq, lrDeviance, lrStdDeviance, lrDifDev, lrCook, lrDfBeta, lrDfBetas);

const
        STR_LOGISTIC_RESIDUALS : array[TLogRegResiduals] of string =
        ('Hat value', 'Pearson residual', 'Standardized Pearson Residual', 'DIFCHISQ', 'Deviance', 'Standardized Deviance', 'DIFDEV', 'Cook Distance', 'DFBETA', 'DFBETAS');

TYPE
        //Tableau des seuils de coupure
        TTabLogRegResValues = array[TLogRegResiduals] of double;

TYPE
        {gnrateur de composant}
        TMLGenSpvLogRegResiduals = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {Composant}
        TMLSpvLogRegResiduals = class(TMLCompLogisticRegAfter)
        protected
        function    getClassOperator: TClassOperator; override;
        public
        end;

        {Oprateur}
        TOpSpvLogRegResiduals = class(TOpLocalData)
        private
        //classifieur branch
        FClassifier: TCalcLogisticReg;
        //liste des rsidus calculs
        FLstRes: TObjectList;
        //hat values
        FHat: TAttribute;
        //Seuil min.
        FMinCut, FMaxCut: TTabLogRegResValues;
        //Comptage des hors zones
        FCount: TTabLogRegResValues;
        //obtenir le postfixe pour  le nom des variables
        function    getPostFixeAttName(): string;
        //obtenir la proba d'un individu
        function    getProba(example: integer): double;
        //obtenir la valeur d'un individu
        function    getClassValue(example: integer): double;
        //calcul de la Hat matrix
        procedure   HatValues();
        //calcul des autres indicateurs
        procedure   OtherResidualsInfluences();
        //dfBeta et DfBetas
        procedure   CalcDfBetas();
        //ajouter les variables supplmentaires
        procedure   AddGeneratedToDataset();
        //reset des seuils
        procedure   ResetCutValues();
        protected
        function    CoreExecute(): boolean; override;
        function    getClassParameter: TClassOperatorParameter; override;
        //outil spcifique de visualisation
        function    getNewForm(): TForm; override;
        public
        constructor Create(AOwner: TObject); override;
        destructor  Destroy(); override;
        function    getHTMLResultsSummary(): string; override;
        function    isOutside(lrType: TLogRegResiduals; ACol: integer; example: integer): boolean;
        //
        property    Classifier: TCalcLogisticReg read FClassifier;
        property    LstRes: TObjectList read FLstRes;
        end;

        {Paramtrage}
        TOpPrmSpvLogRegResiduals = class(TOperatorParameter)
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        end;

implementation

uses
    Math,
    UDatasetImplementation,
    Classes, UCompMetaSpvOneInstance, SysUtils, UConstConfiguration,
  UFrmOpViewSpvLogRegResiduals;

Const
    MIN_CUT_VALUE : double = - Math.MaxDouble;
    MAX_CUT_VALUE : double = Math.MaxDouble;

{ TMLGenSpvLogRegResiduals }

procedure TMLGenSpvLogRegResiduals.GenCompInitializations;
begin
 FMLComp:= mlcSpvAssessment;

end;

function TMLGenSpvLogRegResiduals.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLSpvLogRegResiduals; 
end;

{ TMLSpvLogRegResiduals }

function TMLSpvLogRegResiduals.getClassOperator: TClassOperator;
begin
 result:= TOpSpvLogRegResiduals;
end;

{ TOpSpvLogRegResiduals }

procedure TOpSpvLogRegResiduals.AddGeneratedToDataset;
var j: integer;
begin
 self.GenAtts.Clear();
 for j:= 0 to pred(FLstRes.Count) do
  self.GenAtts.Add(FLstRes.Items[j] as TAttribute);
end;

function TOpSpvLogRegResiduals.CoreExecute: boolean;
begin
 //brancher le classifieur
 FClassifier:= (((MLOwner as TMLSpvLogRegResiduals).Predecessor as TMLCompOneInstance).Operator as TOpOneInstance).Classifier as TCalcLogisticReg;
 //zoo...
 result:= true;
 FLstRes.Clear();//les variables sont libres
 //vider le tableau des cutValues
 self.ResetCutValues();
 //calculs successifs
 self.HatValues();//hat
 //les autres indicateurs
 self.OtherResidualsInfluences();
 //DfBeta et DfBetas
 self.CalcDfBetas();
 //et envoyer tout le monde en sortie
 self.AddGeneratedToDataset();
 TRY
 EXCEPT
 result:= false;
 END;
end;

constructor TOpSpvLogRegResiduals.Create(AOwner: TObject);
begin
 inherited create(AOwner);
 FLstRes:= TObjectList.Create(TRUE);
end;

destructor TOpSpvLogRegResiduals.Destroy;
begin
 FLstRes.Free();
 inherited;
end;

function TOpSpvLogRegResiduals.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmSpvLogRegResiduals;
end;

function TOpSpvLogRegResiduals.getClassValue(example: integer): double;
begin
 result:= 2.0 - FClassifier.ClassAttribute.cValue[example];
end;

procedure TOpSpvLogRegResiduals.HatValues;
var i,k,j: integer;
    p,z,v,w: double;
begin
 //crer
 FHat:= TAttContinue.Create('hat'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size);
 FLstRes.Add(FHat);
 //calculer
 for i:= 1 to FHat.Size do
  begin
   v:= 0.0;
   for k:= 0 to FClassifier.Descriptors.Count do
    begin
     w:= 0.0;
     for j:= 0 to FClassifier.Descriptors.Count do
      begin
       //very fruste, mais c'est le plus scure pour l'instant
       if (j = 0) then z:= 1.0 else z:= FClassifier.Descriptors.Attribute[j-1].cValue[i];
       //produit
       w:= w + z * FClassifier.MatVCOV[j,k];
      end;
     if (k = 0) then z:= 1.0 else z:= FClassifier.Descriptors.Attribute[k-1].cValue[i];
     v:= v + w * z;
    end;
   p:= self.getProba(i);
   FHat.cValue[i]:= p * (1.0 - p) * v;
   if (FHat.cValue[i] < FMinCut[lrHat]) or (FHat.cValue[i] > FMaxCut[lrHat])
    then FCount[lrHat]:= FCount[lrHat] + 1.0; 
  end;
end;

function TOpSpvLogRegResiduals.getHTMLResultsSummary: string;
var s: string;
begin
 s:= '<H3>Residuals and Inlfuential Points - Logistic Regression</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY +
     '<TH>&nbsp;</TH><TH colspan=2>Cut-off values</TH><TH>&nbsp;</TH></TR>';
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY +
     '<TH>Criterion</TH><TH>Lower</TH><TH>Upper</TH><TH>Count</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Hat (Leverage)</TD><TD align=center>-</TD><TD align=right>%.4f</TD><TD align=right>%.0f</TD></TR>',
        [FMaxCut[lrHat],FCount[lrHat]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>DIFCHISQ</TD><TD align=center>-</TD><TD align=right>%.4f</TD><TD align=right>%.0f</TD></TR>',
        [FMaxCut[lrDifChisq],FCount[lrDifChisq]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>DIFDEV</TD><TD align=center>-</TD><TD align=right>%.4f</TD><TD align=right>%.0f</TD></TR>',
        [FMaxCut[lrDifDev],FCount[lrDifDev]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Cook''s D</TD><TD align=center>-</TD><TD align=right>%.4f</TD><TD align=right>%.0f</TD></TR>',
        [FMaxCut[lrCook],FCount[lrCook]]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>DFBETAS</TD><TD align=center>%.4f</TD><TD align=right>%.4f</TD><TD align=right>%.0f</TD></TR>',
        [FMinCut[lrDfBetas],FMaxCut[lrDfBetas],FCount[lrDfBetas]]);
 s:= s + '</table>';
 result:= s;
end;

function TOpSpvLogRegResiduals.getPostFixeAttName: string;
var predML: TMLComponent;
begin
 predML:= (MLOwner as TMLSpvLogRegResiduals).Predecessor as TMLCompOneInstance;
 result:= format('_logReg_%d',[predML.Number]);
end;

function TOpSpvLogRegResiduals.getProba(example: integer): double;
var p: double;
    j: integer;
begin
 p:= FClassifier.Coef[0];
 for j:= 0 to pred(FClassifier.Descriptors.Count) do
  p:= p + FClassifier.Coef[j+1] * FClassifier.Descriptors.Attribute[j].cValue[example];
 p:= 1.0 / (1.0 + exp(-1.0 * p));
 result:= p;
end;

procedure TOpSpvLogRegResiduals.OtherResidualsInfluences();
var r,rstd,difchisq: TAttribute;
    dev,stdDev,difdev: TAttribute;
    cook: TAttribute;
    i: integer;
    v,p: double;
begin
 r:= TAttContinue.Create('pearson'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size);
 FLstRes.Add(r);
 rstd:= TAttContinue.Create('std_pearson'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size);
 FLstRes.Add(rstd);
 difChisq:= TAttContinue.Create('difchisq'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size);
 FLstRes.Add(difChisq);
 dev:= TAttContinue.Create('deviance'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size);
 FLstRes.Add(dev);
 stdDev:= TAttContinue.Create('std_deviance'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size);
 FLstRes.Add(stdDev);
 difdev:= TAttContinue.Create('difdev'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size);
 FLstRes.Add(difdev);
 cook:= TAttContinue.Create('cook'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size);
 FLstRes.Add(cook);
 for i:= 1 to FClassifier.ClassAttribute.Size do
  begin
   p:= self.getProba(i);
   //Pearson
   if (p > 0.0) and (p < 1.0)
    then v:= (self.getClassValue(i) - p)/sqrt(p * (1.0 - p))
    else v:= 0.0;
   r.cValue[i]:= v;
   rstd.cValue[i]:= v/sqrt(1.0 - FHat.cValue[i]);
   difchisq.cValue[i]:= SQR(rstd.cValue[i]);
   if (difchisq.cValue[i] < FMinCut[lrDifChisq]) or (difchisq.cValue[i] > FMaxCut[lrDifChisq])
    then FCount[lrDifChisq]:= FCount[lrDifChisq] + 1.0;
   //dviance
   if (self.getClassValue(i) > 0.5) //pour savoir si y = 1
    then v:= +sqrt(2 * abs(ln(self.getProba(i))))
    else v:= -sqrt(2 * abs(ln(1.0 - self.getProba(i))));
   dev.cValue[i]:= v;
   stdDev.cValue[i]:= v/sqrt(1.0 - FHat.cValue[i]);
   difdev.cValue[i]:= SQR(dev.cValue[i]) + SQR(r.cValue[i]) * FHat.cValue[i]/(1.0 - FHat.cValue[i]);
   if (difdev.cValue[i] < FMinCut[lrDifDev]) or (difdev.cValue[i] > FMaxCut[lrDifDev]) then FCount[lrDifDev]:= FCount[lrDifDev] + 1.0;
   cook.cValue[i]:= SQR(rstd.cValue[i])/(1.0 + self.FClassifier.Descriptors.Count)*(FHat.cValue[i] / (1.0 - FHat.cValue[i]));
   if (cook.cValue[i] < FMinCut[lrCook]) or (cook.cValue[i] > FMaxCut[lrCook]) then FCount[lrCook]:= FCount[lrCook] + 1.0;
  end;
end;

procedure TOpSpvLogRegResiduals.CalcDfBetas;
var dfB,dfBs: TObjectList;
    j,k,i: integer;
    z,v,r: double;
begin
 //dfbeta
 dfb:= TObjectList.Create(FALSE);
 dfb.Add(TAttContinue.Create('dfbeta_1'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size));
 for j:= 0 to pred(FClassifier.Descriptors.Count) do
  dfb.Add(TAttContinue.Create('dfbeta_'+FClassifier.Descriptors.Attribute[j].Name+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size));
 //dfbetas
 dfbs:= TObjectList.Create(FALSE);
 dfbs.Add(TAttContinue.Create('dfbetas_1'+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size));
 for j:= 0 to pred(FClassifier.Descriptors.Count) do
  dfbs.Add(TAttContinue.Create('dfbetas_'+FClassifier.Descriptors.Attribute[j].Name+self.getPostFixeAttName(),FClassifier.ClassAttribute.Size));
 //calculer
 for i:= 1 to FClassifier.ClassAttribute.Size do
  begin
   r:= (self.getClassValue(i)-self.getProba(i))/(1.0 - FHat.cValue[i]);
   for k:= 0 to FClassifier.Descriptors.Count do
    begin
     v:= 0.0;
     for j:= 0 to FClassifier.Descriptors.Count do
      begin
       if (j = 0) then z:= 1.0 else z:= FClassifier.Descriptors.Attribute[j-1].cValue[i];
       v:= v + FClassifier.MatVCOV[k,j] * z;
      end;
     v:= v * r;
     (dfB.Items[k] as TAttribute).cValue[i]:= v;
     if (FClassifier.MatVCOV[k,k] > 0) then v:= v /sqrt(FClassifier.MatVCOV[k,k]) else v:= 0.0;
     if (v < FMinCut[lrDfBetas]) or (v > FMaxCut[lrDfBetas]) then FCount[lrDfBetas]:= FCount[lrDfBetas] + 1.0;
     (dfBs.Items[k] as TAttribute).cValue[i]:= v;
    end;
  end;
 //fusionner les listes
 for j:= 0 to pred(dfb.Count) do FLstRes.Add(dfb.Items[j]);
 for j:= 0 to pred(dfbs.Count) do FLstRes.Add(dfbs.Items[j]);
 //virer les conteneurs
 dfb.Free();
 dfbs.Free();
end;

procedure TOpSpvLogRegResiduals.ResetCutValues;
var lr: TLogRegResiduals;
begin
 //rinitialiser
 for lr:= low(TLogRegResiduals) to high(TLogRegResiduals) do
  begin
   FMinCut[lr]:= MIN_CUT_VALUE;
   FMaxCut[lr]:= MAX_CUT_VALUE;
  end;
 //mettre les bonnes valeurs
 //(lrHat, lrPearson, lrStdPearson, lrDifChisq, lrDeviance, lrStdDeviance, lrDifDev, lrCook, lrDfBeta, lrDfBetas);
 FMaxCut[lrHat]:= 2.0 * (1.0 + self.FClassifier.Descriptors.Count)/(1.0 * self.FClassifier.CurExamples.Size);
 FMaxCut[lrDifChisq]:= +3.84;
 FMaxCut[lrDifDev]:= +3.84;
 FMaxCut[lrCook]:= 4.0 / (1.0 * self.FClassifier.CurExamples.Size - 1.0 * self.FClassifier.Descriptors.Count - 1.0);
 FMinCut[lrDfBetas]:= -2.0/sqrt(1.0 * self.FClassifier.CurExamples.Size);
 FMaxCut[lrDfBetas]:= 2.0/sqrt(1.0 * self.FClassifier.CurExamples.Size);
end;

function TOpSpvLogRegResiduals.getNewForm: TForm;
begin
 result:= TFrmOpViewSpvLogRegResisuals.CreateFromOperator(self);
end;

function TOpSpvLogRegResiduals.isOutside(lrType: TLogRegResiduals; ACol: integer;
  example: integer): boolean;
var att: TAttribute;
begin
 att:= FLstRes.Items[ACol-1] as TAttribute;
 result:= (att.CValue[example] < FMinCut[lrType]) or (att.CValue[example] > FMaxCut[lrType]); 
end;

{ TOpPrmSpvLogRegResiduals }

function TOpPrmSpvLogRegResiduals.CreateDlgParameters: TForm;
begin
 result:= nil;//pour l'instant
end;

procedure TOpPrmSpvLogRegResiduals.SetDefaultParameters;
begin
 //none
end;

initialization
 RegisterClass(TMLGenSpvLogRegResiduals);
end.
