(***************************************************************************)
(* UCompSpvMultiCategoryLogReg.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(***************************************************************************)

{
@abstract(Rgression logistique multinomiale)
@author(Ricco)
@created(01/12/2007)

Cette procdure est en comptition avec la rgression qui venait de WEKA. Il y a
vraiment un souci dans le calcul de la matrice Hessienne...

Les classes proposes ici rutilisent au max. celles de la rgression logistique binaire.

new -- 07/12/2007 -- dans la matrice hessienne, soustraction de lambda pour rgulariser, utile surtout lorsque quasi-sparation parfaite
new -- 26/05/2011 -- criture de la fonction directFuncGrad pour bnficier du L_BFGS pralable

}

unit UCompSpvMultiCategoryLogReg;

interface

USES
        Forms, Contnrs,
        UCompDefinition,
        UCompSpvLDefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcDistribution,
        Stat,
        d6_matrices, d6_fmath,
        UCalcSpvStructScore,
        UCalcStatDes,
        UCompSpvLogisticRegression,
        UFrmBaseOperator;


TYPE
        {le gnrateur de composant Supervised}
        TMLGCompMCLogisticReg = class(TMLGenCompSpvLearning)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant Spv}
        TMLCompMCLogisticReg = class(TMLCompSpvLearning)
        protected
        function    getClassOperator: TClassOperator; override;
        function    GetLogResultDescription(): string; override;
        end;

        {l'oprateur}
        TOpMCLogisticReg = class(TOpSpvLearningContinuous)
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    getClassSpvLearning(): TClassCalcSpvLearning; override;
        end;

        {paramtrage de l'oprateur}
        TOpPrmMCLogisticReg = class(TOpPrmSpvLearning)
        protected
        function    CreateDlgParameters(): TForm; override;
        end;

        TIndividualProba = array of double;

        {classe de calcul}
        TCalcMCLogisticReg = class(TCalcLogisticReg)
        private
        //index du coefficient de la variable j pour la classe c
        function    index(c: integer; j: integer): integer;
        protected
        FOWald: TVector;//valeur de la statistique de Wald pour les effets globaux
        FPWald: TVector;//p-value associe
        FStdCoef: TVector;//coefficients standardiss
        //crer les structures de calcul
        procedure   createStructures(); override;
        procedure   destroyStructures(); override;
        procedure   initCoefs(); override;
        procedure   computeProbaOnStandardizedData(X: TVector; i: integer; var tab: TIndividualProba);
        procedure   CorrectComputedCoefs(); override;
        function    Log_Likelihood(X: TVector): float; override;
        procedure   directHessGrad(Func : TFuncNVar; const X: TVector; Lbound, Ubound: Integer; const G: TVector; const H: TMatrix); override;
        procedure   directGrad(Func : TFuncNVar; const X: TVector; Lbound, Ubound: Integer; const G: TVector); override;
        {new -- 26/05/2011 -- deviance et vecteur gradient pour LBFGS}
        function    directFuncGrad(const X: TVector; LBound, UBound: integer; const G: TVector): double; override;
        procedure   computeInvHessian();
        function    computeDevianceInterceptOnly(): double; override;
        procedure   computeSpecificStats(); override;
        procedure   RunOptimization(); override;
        function    getHTMLCoefficients(): string; override;
        function    getHTMLPostResults(): string; override;
        function    getHTMLRefValue(): string; override;
        procedure   computeVariableOverallEffect();
        public
        procedure   getScore(example: integer; var postProba: TTabScore); override;
        function    hasOwnViewForm(): boolean; override;
        function    getClassViewForm(): TClassFrmBaseOperator; override;
        end;

implementation

USES
        Windows, Sysutils, Classes, Math,
        fmath,
        d6_optim,
        UConstConfiguration, UStringsResources, ULogFile;

CONST
        MAX_ITER_OPTMIZATION_LOGISTIC = 50;
        TOLERANCE_OPTIMIZATION_LOGISTIC = 1.0e-10;
        MIN_VALUE_FOR_STDDEV = 1.0e-6;

        LAMBDA_FOR_REGULARIZATION = 0.0001;

{ TMLGCompMCLogisticReg }

procedure TMLGCompMCLogisticReg.GenCompInitializations;
begin
 FMLComp:= mlcSpvLearning;
end;

function TMLGCompMCLogisticReg.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompMCLogisticReg;
end;

{ TMLCompMCLogisticReg }

function TMLCompMCLogisticReg.getClassOperator: TClassOperator;
begin
 result:= TOpMCLogisticReg; 
end;

function TMLCompMCLogisticReg.GetLogResultDescription: string;
begin
 result:= format('Multicategory Logistic regression result generated [%s]',[self.Description])
end;

{ TOpMCLogisticReg }

function TOpMCLogisticReg.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmMCLogisticReg;
end;

function TOpMCLogisticReg.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcMCLogisticReg;
end;

{ TOpPrmMCLogisticReg }

function TOpPrmMCLogisticReg.CreateDlgParameters: TForm;
begin
 result:= nil;
end;

{ TCalcMCLogisticReg }

procedure TCalcMCLogisticReg.CorrectComputedCoefs;
var j,c: integer;
begin

 //garder une copie des coefficients standardiss pour les tests ultrieurs
 CopyVector(FStdCoef,FCoef,0,NbParam-1);
 
 //correction des coefficients
 for c:= 1 to NbClasses-1 do
  begin
   for j:= 1 to self.Descriptors.Count do
    begin
     //variable considre comme constante ?
     if (FReduction[j] < MIN_VALUE_FOR_STDDEV)
      then FCoef[(c-1)*(NbVar+1)+j]:= 0.0
      else
       begin
        FCoef[index(c,j)]:= FCoef[index(c,j)] / FReduction[j];
        FCoef[index(c,0)]:= FCoef[index(c,0)] - FCoef[index(c,j)] * FCentrage[j];
       end;
    end;
  end;

end;

procedure TCalcMCLogisticReg.createStructures;
begin
 NbClasses:= self.ClassAttribute.nbValues;
 NbVar:= self.Descriptors.Count;
 NbParam:= (NbClasses-1)*(NbVar+1);//+1  cause de la constante
 //
 dimVector(FCoef,NbParam-1);//-1 car 1er indice  zro
 dimMatrix(FHInv,NbParam-1,NbParam-1);
 dimVector(FStdDev,NbParam-1);
 dimVector(FWald,NbParam-1);
 dimVector(FPValueWald,NbParam-1);
 dimVector(FOWald,NbVar);
 dimVector(FPWald,NbVar);
 dimvector(FStdCoef,NbParam-1);
end;

procedure TCalcMCLogisticReg.destroyStructures;
begin
 inherited;
 //LRegProba:= nil;//vive le garbage collector...
  FOWald:= nil;
  FPWald:= nil;
  FStdCoef:= nil;
end;

procedure TCalcMCLogisticReg.directGrad(Func: TFuncNVar; const X: TVector;
  Lbound, Ubound: Integer; const G: TVector);
var i,j,c,example,y: integer;
    v,z: double;
    t: TIndividualProba;
begin

 TRY

 setLength(t,NbClasses-1);
 for c:= LBound to UBound do G[c]:= 0;

 for i:= 1 to FCurExamples.Size do
  begin
   example:= FCurExamples.Number[i];
   y:= ClassAttribute.dValue[example];
   self.computeProbaOnStandardizedData(X,i,t);

   //****************
   //vecteur gradient
   for c:= 1 to NbClasses-1 do
    begin
     if (y = c) then z:= 1.0 else z:= 0.0;
     G[index(c,0)]:= G[index(c,0)] + (z - t[c-1]) * 1.0;
     for j:= 1 to NbVar do
      begin
       //v:= self.getCorrectedValue(pred(j),example);
       v:= FXData[i,j];
       G[index(c,j)]:= G[index(c,j)] + (z - t[c-1]) * v;
      end;
    end;
   //***************

  end;

 finalize(t);

 EXCEPT
  on e: Exception do
   begin
   TraceLog.WriteToLogFile(format('[MULTINOMIAL LOGISTIC REGRESSION] error DIRECT.GRAD = %s',[e.Message]));
   Raise e;
   end;
 END;

end;

procedure TCalcMCLogisticReg.directHessGrad(Func: TFuncNVar; const X: TVector;
  Lbound, Ubound: Integer; const G: TVector; const H: TMatrix);
var i,j,c,c1,c2,example,y,j1,j2: integer;
    v,z,pc1,pc2,v1,v2: double;
    t: TIndividualProba;
begin
 TRY
 setLength(t,NbClasses-1);
 for c:= LBound to UBound do G[c]:= 0;
 for c1:= LBound to UBound do for c2:= LBound to UBound do H[c1,c2]:= 0.0;

 for i:= 1 to FCurExamples.Size do
  begin
   example:= FCurExamples.Number[i];
   y:= ClassAttribute.dValue[example];
   self.computeProbaOnStandardizedData(X,i,t);

   //****************
   //vecteur gradient
   for c:= 1 to NbClasses-1 do
    begin
     if (y = c) then z:= 1.0 else z:= 0.0;
     G[index(c,0)]:= G[index(c,0)] + (z - t[c-1]) * 1.0;
     for j:= 1 to NbVar do
      begin
       //v:= self.getCorrectedValue(pred(j),example);
       v:= FXData[i,j];
       G[index(c,j)]:= G[index(c,j)] + (z - t[c-1]) * v;
      end;
    end;
   //***************


   //*****************
   //Matrice hessienne
   for c1:= 1 to NbClasses-1 do
    begin
     for c2:= c1 to NbClasses-1 do
      begin
       pc1:= t[c1-1];
       pc2:= t[c2-1];
       if (c2 = c1) then z:= 1.0 else z:= 0.0;
       for j1:= 0 to NbVar do
        for j2:= j1 to NbVar do
         begin
          //if (j1 > 0) then v1:= self.getCorrectedValue(pred(j1),example) else v1:= 1.0;
          v1:= FXData[i,j1];
          //if (j2 > 0) then v2:= self.getCorrectedValue(pred(j2),example) else v2:= 1.0;
          v2:= FXData[i,j2];
          v:= pc1 * (z - pc2) * v1 * v2;
          H[index(c1,j1),index(c2,j2)]:= H[index(c1,j1),index(c2,j2)] + v;
         end;
      end;
    end;
   //*****************

  end;

 //complter la partie symtrique dans chaque bloc de la matrice Hessienne
 for c1:= 1 to NbClasses-1 do
  for c2:= c1 to NbClasses-1 do
   begin
    for j1:= 0 to nbVar do
     for j2:= j1+1 to NbVar do
      H[index(c1,j2),index(c2,j1)]:= H[index(c1,j1),index(c2,j2)];
   end;

 //complter les blocs complets symtriques de la matrice Hessienne
 for c1:= 1 to NbClasses-1 do
  for c2:= c1+1 to NbClasses-1 do
   begin
    for j1:= 0 to nbVar do
     for j2:= 0 to NbVar do
      H[index(c2,j1),index(c1,j2)]:= H[index(c1,j1),index(c2,j2)];
   end;

 //R.R.-- 07/12/2007 --- ?? -- rgularisation en soustrayant lambda sur la diagonale
 for c:= 1 to NbClasses-1 do
  for j:= 0 to NbVar do
   H[index(c,j),index(c,j)]:= H[index(c,j),index(c,j)] - LAMBDA_FOR_REGULARIZATION; 

 finalize(t);
 EXCEPT
  on e: Exception do
   begin
   TraceLog.WriteToLogFile(format('[MULTINOMIAL LOGISTIC REGRESSION] error DIRECT.HESS.GRAD = %s',[e.Message]));
   Raise e;
   end;
 END;
end;

procedure TCalcMCLogisticReg.computeProbaOnStandardizedData(X: TVector; i: integer;
  var tab: TIndividualProba);
var c,j: integer;
    v,sum,value: extended;
begin
 sum:= 0.0;
 for c:= 1 to NbClasses-1 do
  begin
   //combinaison linaire
   v:= X[index(c,0)];
   for j:= 1 to self.Descriptors.Count do
    begin
     //value:= self.getCorrectedValue(pred(j), self.FCurExamples.Number[i]) * X[index(c,j)];
     value:= FXData[i,j] * X[index(c,j)];
     v:= v + value;
    end;
   v:= exp(v); tab[c-1]:= v;
   sum:= sum + v;
  end;
 //corriger maintenant
 for c:= 0 to length(tab)-1 do tab[c]:= tab[c]/(1.0 + sum);
end;

procedure TCalcMCLogisticReg.getScore(example: integer;
  var postProba: TTabScore);
var sum,cusum,v: double;
    c,j: integer;
begin
 //combinaison linaire
 sum:= 0.0;
 for c:= 1 to NbClasses-1 do
  begin
   v:= FCoef[(c-1)*(NbVar+1)];
   for j:= 1 to descriptors.Count do
    v:= v + FCoef[(c-1)*(NbVar+1)+j]*descriptors.Attribute[pred(j)].cValue[example];
   //la proba
   if (v < 708)
    then v:= exp(v)
    else v:= 1e308;
   sum:= sum + v;
   postProba[c]:= v;
  end;
 //
 cusum:= 0.0;
 for c:= 1 to NbClasses-1 do
  begin
   postProba[c]:= postProba[c]/(1.0 + sum);
   cusum:= cusum + postProba[c]; 
  end;
 postProba[nbClasses]:= 1.0 - cusum;
 postProba[0]:= 1.0;
end;

procedure TCalcMCLogisticReg.initCoefs;
var j,c: integer;
    vLast, vCur: double;
begin
 for j:= low(FCoef) to high(FCoef) do FCoef[j]:= 0.0;
 //prendre maintenant les rapports de probas par rapport  la classe de rfrence
 vLast:= self.StatClassAtt.TabFreq.Frequence[NbClasses];
 for c:= 1 to NbClasses - 1 do
  begin
   vCur:= self.StatClassAtt.TabFreq.Frequence[c];
   FCoef[index(c,0)]:= LN(vCur/vLast);
  end;
end;

function TCalcMCLogisticReg.Log_Likelihood(X: TVector): float;
var i,y: integer;
    s,v: extended;//pour prvenir les dbordements de capacit...
    t: TIndividualProba;
begin
 setLength(t,NbClasses-1);
 TRY
 s:= 0.0;
 //pour chaque observation
 for i:= 1 to FCurExamples.Size do
  begin
   self.computeProbaOnStandardizedData(X,i,t);
   y:= ClassAttribute.dValue[FCurExamples.Number[i]];
   if (y = NbClasses)
    then v:= 1.0 - MATH.sum(t)
    else v:= t[y-1];
   //pour viter l'erreur de log.
   if (v > 0) then  s:= s + ln(v);
  end;
 //pour la minimisation, on passe au ngatif
 result:= -2.0*s;
 EXCEPT
   on e: exception do
   begin
    TraceLog.WriteToLogFile(Format('[LOGREG] error LL function == %s',[e.Message]));
    result:= 0.0001*MATH.MaxDouble;
   end;
 END;
 finalize(t);
end;

procedure TCalcMCLogisticReg.RunOptimization;
begin
 //FErrCodeOptimization:= d6_optim.BFGS_OBJECT(Log_Likelihood,directGrad,FCoef,0,NbParam-1,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
 //FErrCodeOptimization:= d6_optim.MarquardtObject(Log_Likelihood,directHessGrad,FCoef,0,NbParam-1,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
 //FErrCodeOptimization:= d6_optim.BFGS_MARQUARDT_OBJECT_PLUS_PLUS(Log_Likelihood,directGrad,directHessGrad,FCoef,0,NbParam-1,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
 //new -- R.R. -- 03/12/2007 -- BFGS + NEWTON-RAPHSON (au vu des perfs. de BFGS, on peut se contenter d'une moindre exigeance aprs)
 FErrCodeOptimization:= self.optimizationBFGS_NewtonRaphson(Log_Likelihood,directGrad,directHessGrad,FCoef,0,NbParam-1,MAX_ITER_OPTMIZATION_LOGISTIC,TOLERANCE_OPTIMIZATION_LOGISTIC,FMinLogLikelihood,FHInv);
end;

procedure TCalcMCLogisticReg.computeInvHessian;
var i,c1,c2,j1,j2: integer;
    v,pc1,pc2,v1,v2: double;
    z: double;
    t: TIndividualProba;
    H: TMatrix;
begin
 setLength(t,NbClasses-1);
 DimMatrix(H,NbParam-1,NbParam-1);

 for i:= 1 to FCurExamples.Size do
  begin
   self.computeProbaOnStandardizedData(FCoef,i,t);

   //*****************
   //Matrice hessienne
   for c1:= 1 to NbClasses-1 do
    begin
     for c2:= c1 to NbClasses-1 do
      begin
       pc1:= t[c1-1];
       pc2:= t[c2-1];
       if (c2 = c1) then z:= 1.0 else z:= 0.0;
       for j1:= 0 to NbVar do
        for j2:= j1 to NbVar do
         begin
          //if (j1 > 0) then v1:= self.getCorrectedValue(pred(j1),example) else v1:= 1.0;
          v1:= FXData[i,j1];
          //if (j2 > 0) then v2:= self.getCorrectedValue(pred(j2),example) else v2:= 1.0;
          v2:= FXData[i,j2];
          v:= pc1 * (z - pc2) * v1 * v2;
          H[index(c1,j1),index(c2,j2)]:= H[index(c1,j1),index(c2,j2)] + v;
         end;
      end;
    end;
   //*****************

  end;//de for chaque individu - for i

 //complter la partie symtrique dans chaque bloc
 for c1:= 1 to NbClasses-1 do
  for c2:= c1 to NbClasses-1 do
   begin
    for j1:= 0 to nbVar do
     for j2:= j1+1 to NbVar do
      H[index(c1,j2),index(c2,j1)]:= H[index(c1,j1),index(c2,j2)];
   end;

 //complter les blocs complets symtriques
 for c1:= 1 to NbClasses-1 do
  for c2:= c1+1 to NbClasses-1 do
   begin
    for j1:= 0 to nbVar do
     for j2:= 0 to NbVar do
      H[index(c2,j1),index(c1,j2)]:= H[index(c1,j1),index(c2,j2)];
   end;

 (*
 //affichage de contrle
 TraceLog.WriteToLogFile('>>');
 for j1:= 0 to nbParam-1 do
  begin
   s:= 'HESS = [';
   for j2:= 0 to nbParam-1 do
    s:= s + format('%.4f,',[H[j1,j2]]);
   s:= copy(s,1,length(s)-1)+']';
   TraceLog.WriteToLogFile(s);
  end;
 TraceLog.WriteToLogFile('<<');
 *)

 //inversion pour obtenir les carts-type
 invMat(H,0,NbParam-1,FHInv);

 (*
 //affichage de contrle
 TraceLog.WriteToLogFile('>>');
 for j1:= 0 to nbParam-1 do
  begin
   s:= 'INV = [';
   for j2:= 0 to nbParam-1 do
    s:= s + format('%.4f,',[FHInv[j1,j2]]);
   s:= copy(s,1,length(s)-1)+']';
   TraceLog.WriteToLogFile(s);
  end;
 TraceLog.WriteToLogFile('<<');
 *)

 H:= nil;
 finalize(t);
end;

function TCalcMCLogisticReg.index(c: integer; j: integer): integer;
begin
 result:= (c - 1) * (NbVar + 1) + j;
end;

procedure TCalcMCLogisticReg.computeSpecificStats;
var c,j: integer;
begin
 //ramener les coefs. dans l'espace originel
 for c:= 1 to NbClasses-1 do
  begin
   for j:= 0 to descriptors.Count do
    if (FHInv[index(c,j)][index(c,j)] <= 0.0) or (FReduction[j] < MIN_VALUE_FOR_STDDEV) or (j = 0)
     then FStdDev[index(c,j)]:= 0.0
     else FStdDev[index(c,j)]:= SQRT(1.0*FHInv[index(c,j)][index(c,j)]) / FReduction[j];
  end;
 //Wald et p-value associe
 for j:= 0 to length(FCoef)-1 do
  begin
   FWald[j]:= 0.0;
   FPValueWald[j]:= 1.0;
   if (FStdDev[j] > 0.0)
    then
     begin
      FWald[j]:= SQR(FCoef[j]/FStdDev[j]);
      if (FWald[j] > 0.0) then FPValueWald[j]:= fmath.PKHI2(1,FWald[j]);
     end;
  end;
 //calcul de l'effet global pour chaque variable
 self.computeVariableOverallEffect();
end;


function TCalcMCLogisticReg.getHTMLCoefficients: string;
var c,j: integer;
    s: string;
begin
 //les coefficients
 s:= '<P><H3>Attributes in the equation</H3><BR>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Class.Value</TH>';
 for c:= 1 to NbClasses-1 do
  s:= s + format('<TH colspan=4>%s</TH>',[ClassAttribute.LstValues.getDescription(c)]);
 s:= s + '</TR>';
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH width=100>Pred.Att.</TH>';
 for c:= 1 to NbClasses-1 do
  s:= s + '<TH width=70>Coef.</TH><TH width=60>Std.Err</TH><TH width=60>Wald</TH><TH width=60>p-value</TH>';
 s:= s + '</TR>';
 for j:= 0 to descriptors.count do
  begin
   s:= s+HTML_TABLE_COLOR_DATA_GRAY;
   if (j=0)
    then s:= s+'<TD>constant</TD>'
    else s:= s+'<TD>'+descriptors.Attribute[pred(j)].Name+'</TD>';
   for c:= 1 to NbClasses-1 do
    begin
     //statistique de wald
     if (j = 0)
      then s:= s + format('<TD align=right>'+STR_FORMAT_VIEW_STAT_ACCURACY_HIGH+'</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD>',[FCoef[(c-1)*(NbVar+1)]])
      else
       s:= s+format('<TD align=right>%.6f</TD>'+
                    '<TD align=right>%.4g</TD>'+
                    '<TD align=right>%.4g</TD>'+
                    '<TD align=right %s>%.4f</TD>',
                    [FCoef[index(c,j)],FStdDev[index(c,j)],FWald[index(c,j)],colorCodeForPValue(FPValueWald[index(c,j)]),FPValueWald[index(c,j)]]);
    end;
   s:= s+'</TR>';
  end;
 s:= s+'</table>';
 result:= s;
end;

function TCalcMCLogisticReg.getHTMLPostResults: string;
var s: string;
    j: integer;
begin
 s:= '<p><h3>Overall Effect</h3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=100>Attribute</TH><TH width=70>d.f.</TH><TH width=70>Chi-2 Wald</TH><TH width=70>p-value</TH></TR>';
 for j:= 1 to NbVar do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<td>%s</td><td align=right>%d</td><td align=right>%.3f</td><td align=right %s>%.4f</td>',
                  [descriptors.Attribute[pred(j)].Name,NbClasses-1,FOWald[j],colorCodeForPValue(FPWald[j]),FPWald[j]]);
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TCalcMCLogisticReg.getHTMLRefValue: string;
begin
 result:= HTML_TABLE_COLOR_DATA_BLUE+format('<TD>Ref. value</TD><TD colspan=2 align=right>%s</TD></TR>',[ClassAttribute.LstValues.getDescription(ClassAttribute.nbValues)]);
end;

function TCalcMCLogisticReg.computeDevianceInterceptOnly: double;
var j,c: integer;
    vLast, vCur: double;
    tmp: TVector;
begin
 dimvector(tmp,NbParam-1);
 for j:= low(tmp) to high(tmp) do tmp[j]:= 0.0;
 //prendre maintenant les rapports de probas par rapport  la classe de rfrence
 vLast:= self.StatClassAtt.TabFreq.Frequence[NbClasses];
 for c:= 1 to NbClasses - 1 do
  begin
   vCur:= self.StatClassAtt.TabFreq.Frequence[c];
   tmp[index(c,0)]:= LN(vCur/vLast);
  end;
 result:= self.Log_Likelihood(tmp);
 tmp:= nil;
end;

procedure TCalcMCLogisticReg.computeVariableOverallEffect;
var coef: TVector;
    M,MInv: TMatrix;
    k,c,j: integer;
    chi2,sum: double;
    res: integer;
begin
 DimVector(coef,NbClasses-1);
 DimMatrix(M,NbClasses-1,NbClasses-1);
 DimMatrix(Minv,NbClasses-1,NbClasses-1);
 //pour chaque variable
 for j:= 1 to NbVar do
  begin
   //copier les donnes locales
   for c:= 1 to NbClasses-1 do coef[c]:= FStdCoef[index(c,j)];
   for c:= 1 to NbClasses-1 do for k:= 1 to NbClasses-1 do M[c,k]:= FHInv[index(c,j),index(k,j)];
   //inverser la matrice
   res:= invMat(M,1,NbClasses-1,MInv);
   FOWald[j]:= 0.0;
   FPWald[j]:= 1.0;
   if (res = MAT_OK)
    then
     begin
      chi2:= 0.0;
      for c:= 1 to NbClasses-1 do
       begin
        sum:= 0.0;
        for k:= 1 to NbClasses-1 do
         sum:= sum + coef[k]*MInv[k,c];
        chi2:= chi2 + sum * coef[c];
       end;
      FOWald[j]:= chi2;
      if (chi2 > 0.0) then FPWald[j]:= fmath.PKhi2(NbClasses-1,chi2);
     end;
  end;
 coef:= nil; M:= nil; MInv:= nil;
end;

function TCalcMCLogisticReg.getClassViewForm: TClassFrmBaseOperator;
begin
 result:= nil;
end;

function TCalcMCLogisticReg.hasOwnViewForm: boolean;
begin
 result:= FALSE;
end;

function TCalcMCLogisticReg.directFuncGrad(const X: TVector; LBound,
  UBound: integer; const G: TVector): double;
var i,j,c,example,y: integer;
    v,z: double;
    t: TIndividualProba;
    s: double;//pour la deviance - 26/05/2011
begin

 s:= 0.0;

 TRY

 setLength(t,NbClasses-1);
 for c:= LBound to UBound do G[c]:= 0;

 for i:= 1 to FCurExamples.Size do
  begin
   example:= FCurExamples.Number[i];
   y:= ClassAttribute.dValue[example];
   self.computeProbaOnStandardizedData(X,i,t);

   //ajout - 26/05/2011 - calcul de la dviance
   if (y = NbClasses)
    then v:= 1.0 - MATH.sum(t)
    else v:= t[y-1];
    
   if (v > 0) then s:= s + ln(v);
   //<<


   //****************
   //vecteur gradient
   for c:= 1 to NbClasses-1 do
    begin
     if (y = c) then z:= 1.0 else z:= 0.0;
     G[index(c,0)]:= G[index(c,0)] + (z - t[c-1]) * 1.0;
     for j:= 1 to NbVar do
      begin
       //v:= self.getCorrectedValue(pred(j),example);
       v:= FXData[i,j];
       G[index(c,j)]:= G[index(c,j)] + (z - t[c-1]) * v;
      end;
    end;
   //***************

  end;

 finalize(t);

 //renvoyer la deviance
 result:= -2.0 * s;

 //<<

 EXCEPT
  on e: Exception do
   begin
   TraceLog.WriteToLogFile(format('[MULTINOMIAL LOGISTIC REGRESSION] error DIRECT.FUNC.GRAD = %s',[e.Message]));
   Raise e;
   end;
 END;

end;

initialization
 CLASSES.RegisterClass(TMLGCompMCLogisticReg);
end.
