(******************************************************************)
(* UCompSpvNaiveBayes.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(Un apprentissage bayesien naf - descripteurs continus - gaussian assumption)
@author(Ricco)
@created(18/10/2010)
}

unit UCompSpvNaiveBayesContinuous;

interface

USES
        Forms, IniFiles, Classes,
        UCompDefinition,
        UCompSpvLDefinition,
        UOperatorDefinition,
        UDatasetDefinition,
        UDatasetImplementation,
        UDatasetExamples,
        UCalcStatDesCrossTab,
        UCalcDistribution,
        UCalcSpvStructScore,
        UCalcStatDesConditionnalDesc;

TYPE
        {le gnrateur de composant Supervised}
        TMLGCompNaiveBayesContinuous = class(TMLGenCompSpvLearning)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {le composant Spv}
        TMLCompNaiveBayesContinuous = class(TMLCompSpvLearning)
        protected
        function    getClassOperator: TClassOperator; override;
        function    GetLogResultDescription(): string; override;
        end;

        {l'oprateur}
        TOpNaiveBayesContinuous = class(TOpSpvLearningContinuous)
        protected
        function    getClassParameter: TClassOperatorParameter; override;
        function    getClassSpvLearning(): TClassCalcSpvLearning; override;
        end;

        {paramtrage de l'oprateur}
        TOpPrmNaiveBayesContinuous = class(TOpPrmSpvLearning)
        protected
        //la valeur du lambda, par dfaut 0.0 - ne peut pas tre ngatif
        FLambdaLaplacian: double;
        //modle homoscdastique ou pas - df. homoscdastique
        FHomoscedastic: boolean;
        function    CreateDlgParameters(): TForm; override;
        procedure   SetDefaultParameters(); override;
        public
        function    getHTMLParameters(): string; override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        property    lambdaLaplacian: double read FLambdaLaplacian write FLambdaLaplacian;
        property    homoscedastic: boolean read FHomoscedastic write FHomoscedastic;
        end;

        TVectorCoef = array of double;
        TMatrixCoef = array of TVectorCoef;

        TCalcNaiveBayesContinuous = class(TCalcSpvLearning)
        private
        //lambda
        FLambda: double;
        //linear or quadratic model
        FHomoscedastic: boolean;
        //liste des statistiques ANOVA
        FLstStat: TLstStatDesCondANOVA;
        //distribution des classes
        FDistClass: TTabFrequence;
        //vecteur des stats. de test et des p-value
        FVecF, FPValue: TVectorCoef;
        //matrice de coefs. linaire
        FCoefLin: TMatrixCoef;
        //matrice de coefs. quadratiques
        FCoefQuad:TMatrixCoef;
        //degrs de libert
        Fddl1, Fddl2: integer;
        //obtenir les coefs. dans le cadre homoscdastique
        procedure   getCoefHomoscedastic();
        //obtenir les coefs. dans le cadre htroscdastique
        procedure   getCoefHeteroscedastic();
        //obtenir la description du modle quadratique
        function    getHtmlQuadratic(): string;
        //obtenir la description du modle linaire
        function    getHtmlLinear(): string;
        protected
        {crer la liste}
        procedure   createStructures(); override;
        {vider la liste}
        procedure   destroyStructures(); override;
        {apprentissage proprement dit}
        function    coreLearning(examples: TExamples): boolean; override;
        public
        //rcuprer les paramtres de lambda et de lapacien
        constructor create(prmOpSpv: TOpPrmSpvLearning; prmClass: TAttribute; prmDescriptors: TLstAttributes; prmAllAttributes: TLstAttributes); override;
        procedure   getScore(example: integer; var postProba: TTabScore); override;
        //on le maintient ici car getScore peut tre sujet  des dbordements souvent lorsque le nombre de variables est lev (cf. fichier debugfile.txt)
        procedure   classification(example: integer; var response: TTypeDiscrete); override;
        function    getHTMLResults(): string; override;
        end;

implementation

uses
        Math,
        SysUtils, ULogFile, UCalcCrossTab, UStringsResources,
        UConstConfiguration, UStringAddBuffered,
  UDlgOpPrmSpvNaiveBayesContinuous;

const
        //seuil en dea duquel on considre que la variance est nulle
        SEUIL_MIN_VARIANCE : double = 1.0e-08;

{ TMLGCompNaiveBayesContinuous }

procedure TMLGCompNaiveBayesContinuous.GenCompInitializations;
begin
 FMLComp:= mlcSpvLearning;
end;

function TMLGCompNaiveBayesContinuous.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompNaiveBayesContinuous;
end;

{ TMLCompNaiveBayesContinuous }

function TMLCompNaiveBayesContinuous.getClassOperator: TClassOperator;
begin
 result:= TOpNaiveBayesContinuous;
end;

function TMLCompNaiveBayesContinuous.GetLogResultDescription: string;
begin
 result:= format('MLCOMP >> one instance of %s running',[self.ClassName]);
end;

{ TOpNaiveBayesContinuous }

function TOpNaiveBayesContinuous.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmNaiveBayesContinuous;
end;

function TOpNaiveBayesContinuous.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcNaiveBayesContinuous;
end;

{ TOpPrmNaiveBayesContinuous }

function TOpPrmNaiveBayesContinuous.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSpvNaiveBayesCont.CreateFromOpPrm(self);
end;

function TOpPrmNaiveBayesContinuous.getHTMLParameters: string;
var s: string;
begin
 s:= '';
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Parameters</TH></TR>';
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Lambda for laplacian</TD><TD align="right">%.1f</TD></TR>',[FLambdaLaplacian]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Homoscedasticity assumption</TD><TD align="right">%d</TD></TR>',[ord(FHomoscedastic)]);
 s:= s+'</table>';
 //and then...
 result:= s;
end;

procedure TOpPrmNaiveBayesContinuous.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FLambdaLaplacian:= prmINI.ReadFloat(prmSection,'lambda_laplacian',FLambdaLaplacian);
 FHomoscedastic:= prmINI.ReadBool(prmSection,'homoscedasticity',FHomoscedastic);
end;

procedure TOpPrmNaiveBayesContinuous.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FLambdaLaplacian,sizeof(FLambdaLaplacian));
 prmStream.ReadBuffer(FHomoscedastic,sizeof(FHomoscedastic));
end;

procedure TOpPrmNaiveBayesContinuous.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteFloat(prmSection,'lambda_laplacian',FLambdaLaplacian);
 prmINI.WriteBool(prmSection,'homoscedasticity',FHomoscedastic);
end;

procedure TOpPrmNaiveBayesContinuous.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FLambdaLaplacian,sizeof(FLambdaLaplacian));
 prmStream.WriteBuffer(FHomoscedastic,sizeof(FHomoscedastic));
end;

procedure TOpPrmNaiveBayesContinuous.SetDefaultParameters;
begin
 FLambdaLaplacian:= 0.0;
 FHomoscedastic:= true;
end;

{ TCalcNaiveBayesContinuous }

procedure TCalcNaiveBayesContinuous.classification(example: integer;
  var response: TTypeDiscrete);
var s, smax, x: double;
    k, kmax: TTypeDiscrete;
    j: integer;
begin
 smax:= -1.0 * Math.MaxDouble;
 kmax:= 0;
 //calculer le score pour chaque modalit
 for k:= 0 to pred(self.ClassAttribute.nbValues) do
  begin
   s:= FCoefLin[0][k];//constante
   for j:= 0 to pred(self.Descriptors.Count) do
    begin
     x:= self.Descriptors.Attribute[j].cValue[example];
     //si modle quadratique
     if (FCoefQuad <> nil) then s:= s + FCoefQuad[j+1][k] * Math.Power(x, 2.0);
     //la partie linaire est toujours ajoute
     s:= s + FCoefLin[j+1][k] * x;
    end;
   //alors ?
   if (s > smax)
    then
     begin
      smax:= s;
      kmax:= k+1;
     end;
  end;
 //and then...
 if (kmax = 0)
  then kmax:= self.FDistClass.getIndexMaxValue();
 response:= kmax;
end;

function TCalcNaiveBayesContinuous.coreLearning(
  examples: TExamples): boolean;
var j: integer;
    stat: TCalcSDCondDescANOVA;
begin
 result:= true;
 TRY
 //degrs de libert
 Fddl1:= self.ClassAttribute.nbValues - 1;
 Fddl2:= examples.Size - self.ClassAttribute.nbValues;
 //calculs des paramtres
 FDistClass.Refresh(examples);
 FLstStat.RefreshStat(examples);
 //rcupration des stats.
 setLength(FVecF,succ(self.Descriptors.Count)); FVecF[0]:= 0.0;
 setLength(FPValue,succ(self.Descriptors.Count)); FPValue[0]:= 1.0;
 for j:= 0 to pred(FLstStat.Count) do
  begin
   stat:= FLstStat.Stat(j) as TCalcSDCondDescANOVA;
   FVecF[j+1]:= stat.Fisher;
   FPValue[j+1]:= stat.ProbaFisher;
  end;
 //construction des coefficients
 if (self.FHomoscedastic = true)
  then self.getCoefHomoscedastic()
  else self.getCoefHeteroscedastic();
 EXCEPT
 result:= false;
 END;
end;

constructor TCalcNaiveBayesContinuous.create(prmOpSpv: TOpPrmSpvLearning;
  prmClass: TAttribute; prmDescriptors, prmAllAttributes: TLstAttributes);
begin
 inherited;
 //rcupration des paramtres
 FLambda:= (prmOpSpv as TOpPrmNaiveBayesContinuous).lambdaLaplacian;
 if (FLambda < 0.0) then FLambda:= 0.0;//pour viter les btises
 FHomoscedastic:= (prmOpSpv as TOpPrmNaiveBayesContinuous).homoscedastic;
end;

procedure TCalcNaiveBayesContinuous.createStructures;
var stat: TCalcSDCondDescANOVA;
    j: integer;
    y,x: TAttribute;
begin
 //distribution de l'endogne
 y:= ClassAttribute;
 FDistClass:= TTabFrequence.CreateFromAtt(y,NIL);
 //structure de calcul pour l'ANOVA
 FLstStat:= TLstStatDesCondANOVA.Create(nil,nil);
 for j:= 0 to pred(Descriptors.Count) do
  begin
   x:= Descriptors.Attribute[j];
   stat:= TCalcSDCondDescANOVA.Create(x,y,nil);
   FLstStat.AddStat(stat);
  end;
end;

procedure TCalcNaiveBayesContinuous.destroyStructures;
begin
 FLstStat.Free();
 FDistClass.Free();
 //vider (ventuellement) tous les tableaux intermdiaires
 if assigned(FVecF) then finalize(FVecF);
 if assigned(FPValue) then finalize(FPValue);
 if assigned(FCoefLin) then finalize(FCoefLin);
 if assigned(FCoefQuad) then finalize(FCoefQuad);
end;

procedure TCalcNaiveBayesContinuous.getCoefHeteroscedastic;
var k,j: integer;
    mu,v,pk,constante: double;
    stat: TCalcSDCondDescANOVA;
begin
 setLength(FCoefQuad,succ(self.Descriptors.Count),self.ClassAttribute.nbValues);
 setLength(FCoefLin,succ(self.Descriptors.Count),self.ClassAttribute.nbValues);
 for k:= 0 to pred(self.ClassAttribute.nbValues) do
  begin
   pk:= (1.0 * FDistClass.Value[k+1]+FLambda)/(1.0 * FDistClass.Value[0]+FLambda*self.ClassAttribute.nbValues);
   constante:= ln(pk);
   for j:= 0 to pred(self.Descriptors.Count) do
    begin
     stat:= FLstStat.Stat(j) as TCalcSDCondDescANOVA;
     mu:= stat.StatCond[k].Average;
     v:= stat.StatCond[k].Variance;//!\very important here...
     if (v >= SEUIL_MIN_VARIANCE)
      then
       begin
        FCoefQuad[j+1][k]:= FCoefQuad[j+1][k] - 1.0 / (2.0 * v);
        FCoefLin[j+1][k]:= FCoefLin[j+1][k] + mu/v;
        constante:= constante - ((mu*mu)/(2.0*v) + ln(sqrt(v)));
       end;
    end;
   FCoefLin[0][k]:= constante;
  end;
end;

procedure TCalcNaiveBayesContinuous.getCoefHomoscedastic;
var k,j: integer;
    mu,v,pk,constante: double;
    stat: TCalcSDCondDescANOVA;
begin
 FCoefQuad:= nil;
 setLength(FCoefLin,succ(self.Descriptors.Count),self.ClassAttribute.nbValues);
 for k:= 0 to pred(self.ClassAttribute.nbValues) do
  begin
   pk:= (1.0 * FDistClass.Value[k+1]+FLambda)/(1.0 * FDistClass.Value[0]+FLambda*self.ClassAttribute.nbValues);
   constante:= ln(pk);
   for j:= 0 to pred(self.Descriptors.Count) do
    begin
     stat:= FLstStat.Stat(j) as TCalcSDCondDescANOVA;
     mu:= stat.StatCond[k].Average;
     v:= stat.VarianceIntra;
     if (v >= SEUIL_MIN_VARIANCE)
      then
       begin
        FCoefLin[j+1][k]:= FCoefLin[j+1][k] + mu/v;
        constante:= constante - (mu*mu)/(2.0*v);
       end;
    end;
   FCoefLin[0][k]:= constante;
  end;
end;

function TCalcNaiveBayesContinuous.getHtmlLinear: string;
var buf: TBufString;
    s: string;
    k,j: integer;
begin
 buf:= TBufSTring.Create();
 buf.BeginUpdate();
 buf.AddStr('<H4>Linear Model</H4>');
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY+format('<TH>&nbsp;</TH><TH colspan=%d>Classification functions</TH></TR>',[self.ClassAttribute.nbValues+2]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE+'<TH>Descriptors</TH>';
 for k:= 1 to self.ClassAttribute.nbValues do
  s:= s + format('<TH>%s</TH>',[self.ClassAttribute.LstValues.getDescription(k)]);
 s:= s + format('<TH>F(%d,%d)</TH><TH>p-value</TH>',[Fddl1,Fddl2]);
 s:= s + '</TR>';
 buf.AddStr(s);
 //la constante
 s:= HTML_TABLE_COLOR_DATA_GRAY + '<TD align="left">Intercept</TD>';
 for k:= 0 to pred(self.ClassAttribute.nbValues) do
  s:= s + format('<TD align="right">%.6f</TD>',[FCoefLin[0][k]]);
 s:= s + '<TD align="center">-</TD><TD align="center">-</TD>';
 s:= s + '</TR>';
 buf.AddStr(s);
 //les coefficients et l'valuation
 for j:= 0 to pred(self.Descriptors.Count) do
  begin
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TD align="left">%s</TD>',[self.Descriptors.Attribute[j].Name]);
   for k:= 0 to pred(self.ClassAttribute.nbValues) do
    s:= s + format('<TD align="right">%.6f</TD>',[FCoefLin[j+1][k]]);
   //les stats.
   s:= s + format('<TD align=right>%.6f</TD><TD align="right" %s>%.6f</TD>',[FVecF[j+1],colorCodeForPValue(FPValue[j+1]),FPValue[j+1]]);
   s:= s + '</TR>';
   buf.AddStr(s);
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TCalcNaiveBayesContinuous.getHtmlQuadratic: string;
var buf: TBufString;
    s: string;
    k,j: integer;
begin
 buf:= TBufSTring.Create();
 buf.BeginUpdate();
 buf.AddStr('<H4>Quadratic Model</H4>');
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY+format('<TH>&nbsp;</TH><TH colspan=%d>Classification functions</TH></TR>',[self.ClassAttribute.nbValues]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE+'<TH>Descriptors</TH>';
 for k:= 1 to self.ClassAttribute.nbValues do
  s:= s + format('<TH>%s</TH>',[self.ClassAttribute.LstValues.getDescription(k)]);
 s:= s + '</TR>';
 buf.AddStr(s);
 //la constante
 s:= HTML_TABLE_COLOR_DATA_GRAY + '<TD align="left">Intercept</TD>';
 for k:= 0 to pred(self.ClassAttribute.nbValues) do
  s:= s + format('<TD align="right">%.6f</TD>',[FCoefLin[0][k]]);
 s:= s + '</TR>';
 buf.AddStr(s);
 //les coefficients et l'valuation
 for j:= 0 to pred(self.Descriptors.Count) do
  begin
   //coef. quadratiques
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TD align="left">(%s)^2</TD>',[self.Descriptors.Attribute[j].Name]);
   for k:= 0 to pred(self.ClassAttribute.nbValues) do
    s:= s + format('<TD align="right">%.6f</TD>',[FCoefQuad[j+1][k]]);
   s:= s + '</TR>';
   buf.AddStr(s);
   //coef. linaires
   s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TD align="left">%s</TD>',[self.Descriptors.Attribute[j].Name]);
   for k:= 0 to pred(self.ClassAttribute.nbValues) do
    s:= s + format('<TD align="right">%.6f</TD>',[FCoefLin[j+1][k]]);
   s:= s + '</TR>';
   //add...
   buf.AddStr(s);
  end;
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

function TCalcNaiveBayesContinuous.getHTMLResults: string;
begin
 if (self.FHomoscedastic = true)
  then result:= self.getHtmlLinear()
  else result:= self.getHtmlQuadratic();
end;

procedure TCalcNaiveBayesContinuous.getScore(example: integer;
  var postProba: TTabScore);
var s, x: double;
    k: TTypeDiscrete;
    j: integer;
begin
 postProba[0]:= 0.0;
 //calculer le score pour chaque modalit
 for k:= 0 to pred(self.ClassAttribute.nbValues) do
  begin
   s:= FCoefLin[0][k];//constante
   for j:= 0 to pred(self.Descriptors.Count) do
    begin
     x:= self.Descriptors.Attribute[j].cValue[example];
     //si modle quadratique
     if (FCoefQuad <> nil) then s:= s + FCoefQuad[j+1][k] * Math.Power(x, 2.0);
     //la partie linaire est toujours ajoute
     s:= s + FCoefLin[j+1][k] * x;
    end;
   //thsauriser
   postProba[k+1]:= s;
  end;
 //normalisation !!! trs important !!!
 postProba.normalize();
end;

initialization
 RegisterClass(TMLGCompNaiveBayesContinuous);
end.
