(************************************************************)
(* UCompSpvCPLS.pas - Copyright (c) 2006 Ricco RAKOTOMALALA *)
(************************************************************)

{
@abstract(PLS pour la discrimination)
@author(Ricco)
@created(08/03/2006)

La rgression PLS est dfinie pour la rgression, on peut en driver facilement,
tout comme la rgression linaire multiple d'ailleurs, un algorithme d'apprentissage
adapt pour la discrimination.

Utilisation de la classe de calcul PLS implmente par Jean-Franois Grange (JFGCalcPLS.pas)

06/05/2008 -- on introduit galement le mcanisme de slection des axes

}

unit UCompSpvCPLS;

interface

USES
       Classes, Forms, IniFiles,
       UCompDefinition,
       UCompSpvLDefinition,
       UOperatorDefinition,
       UDatasetImplementation,
       UDatasetExamples, UCalcSpvStructScore, UDatasetDefinition,
       JFGCalcPLS;

TYPE
       TMLGCompSpvCPLS = class(TMLGenCompSpvLearning)
       public
       function    GetClassMLComponent: TClassMLComponent; override;
       end;

       TMLCompSpvCPLS = class(TMLCompSpvLearning)
       protected
       function    getClassOperator: TClassOperator; override;
       function    GetLogResultDescription(): string; override;
       end;

       TOpSpvCPLS = class(TOpSpvLearningContinuous)
       protected
       function    getClassParameter: TClassOperatorParameter; override;
       function    getClassSpvLearning(): TClassCalcSpvLearning; override;
       function    ConnectDescriptors(prmData: TMLDataset): boolean; override;
       end;

       TOpPrmCPLS = class(TOpPrmSpvLearning)
       private
       ///nombre d'axes  retenir, si 0 c.--d. choix automatique
       FNbAxis: integer;
       ///seuil pour la redondance, 0.025 par dfaut
       FRdYCutValue: double;
       protected
       function    CreateDlgParameters(): TForm; override;
       procedure   SetDefaultParameters(); override;
       public
       function    getHTMLParameters(): string; override;
       procedure   LoadFromStream(prmStream: TStream); override;
       procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
       procedure   SaveToStream(prmStream: TStream); override;
       procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
       property    NbAxis: integer read FNbAxis write FNbAxis;
       property    RdYCutValue: double read FRdYCutValue write FRdYCutValue;
       end;

       //classe de calcul, son rle est d'appeler la classe de JFG
       TCalcSpvPLS = class(TCalcSpvLearning)
       private
       //oprateur de calcul
       FCalc: TCalcPLS;
       //liste factice pour l'attribut classe
       FLstClassAtt: TLstAttributes;
       //tableaux intermdiaires pour le classement
       FTabY: TVectorPLS;
       //vecteur des VIP globaux - cf. tutoriel
       FGlobVIP: TVectorPLS;
       protected
       procedure   createStructures(); override;
       procedure   destroyStructures(); override;
       procedure   postLearning();
       public
       //surcharge normale dans la hirachie
       function    coreLearning(examples: TExamples): boolean; override;
       //prospective totale -- utiliser une fonction logistique sur la sortie
       procedure   getScore(example: integer; var postProba: TTabScore); override;
       //comparer au seuil 0 (si output <= 0 alors c = 1 sinon c = 2
       procedure   classification(example: integer; var response: TTypeDiscrete); override;
       //description des rsultats
       function    getHTMLResults(): string; override;
       end;



implementation

USES
    Math,
    Sysutils, UConstConfiguration, UDlgOpPrmSVM, UDlgOpPrmSpvCPLS, ULogFile,
  UCalcPLSSupervised, UStringAddBuffered, UDlgOpPrmSpvPLSDA;

{ TMLGCompSpvCPLS }

function TMLGCompSpvCPLS.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvCPLS;
end;

{ TMLCompSpvCPLS }

function TMLCompSpvCPLS.getClassOperator: TClassOperator;
begin
 result:= TOpSpvCPLS;
end;

function TMLCompSpvCPLS.GetLogResultDescription: string;
begin
 result:= format('CPLS result generated [%s]',[self.Description]);
end;

{ TOpSpvCPLS }

function TOpSpvCPLS.ConnectDescriptors(prmData: TMLDataset): boolean;
var ok: boolean;
begin
 TRY
 //accepter les prdicteurs quasi-continus, i.e. continus ou discrets binaires
 ok:= (prmData.LstAtts[asInput].Count > 0) AND (prmData.LstAtts[asInput].isAllCategory(caQuasiContinue));
 if ok then FDescriptorsAtt:= prmData.LstAtts[asInput];//branchement direct, pas de recopie locale
 //uniquement pbm binaire rsolu !
 ok:= ok and (self.ClassAttribute.nbValues = 2);
 EXCEPT
 ok:= FALSE;
 END;
 //and then...
 result:= ok;
end;

function TOpSpvCPLS.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmCPLS;
end;

function TOpSpvCPLS.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcSpvPLS;
end;

{ TOpPrmCPLS }

function TOpPrmCPLS.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSpvPlsDa.CreateFromOpPrm(self);
end;

function TOpPrmCPLS.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Parameters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Required axes</TD><TD align=right>%d</TD></TR>',[NbAxis]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Redundancy cut value</TD><TD align=right>%.4f</TD></TR>',[RdYCutValue]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmCPLS.LoadFromINI(prmSection: string; prmINI: TMemIniFile);
begin
 NbAxis:= prmINI.ReadInteger(prmSection,'nb_axis',NbAxis);
 RdYCutValue:= prmINI.ReadFloat(prmSection,'rdy_cut',RdYCutValue);
end;

procedure TOpPrmCPLS.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FNbAxis,sizeof(FNbAxis));
 prmStream.ReadBuffer(FRdYCutValue,sizeof(FRdYCutValue));
end;

procedure TOpPrmCPLS.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'nb_axis',NbAxis);
 prmINI.WriteFloat(prmSection,'rdy_cut',RdYCutValue);
end;

procedure TOpPrmCPLS.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FNbAxis,sizeof(FNbAxis));
 prmStream.WriteBuffer(FRdYCutValue,sizeof(FRdYCutValue));
end;

procedure TOpPrmCPLS.SetDefaultParameters;
begin
 FNbAxis:= 0;
 FRdYCutValue:= 0.025;
end;

{ TCalcSpvPLS }

procedure TCalcSpvPLS.classification(example: integer;
  var response: TTypeDiscrete);
begin
 //new -- 10/03/2006 -- passer directement les valeurs de la liste de variables
 FCalc.projectionOnLstX(example,self.Descriptors,FTabY);
 //grer la rponse maintenant -- "0" car il n'y a qu'une seule sortie
 response:= 1 + ord(FTabY[0] > FCalc.StatPLS.AvgY[0]);
end;

function TCalcSpvPLS.coreLearning(examples: TExamples): boolean;
begin
 if assigned(FCalc) then FreeAndNil(FCalc);
 TRY
 //lancer l'affaire
 //FCalc:= TCalcPLS.create(self.Descriptors,FLstClassAtt,examples,true,MATH.MIN((self.OpPrmSpv as TOpPrmCPLS).FNbAxis,self.descriptors.count),FALSE);
 //new -- 06/05/2008 -- avec la technique de choix auto du nombre d'axes
 FCalc:= TCalcPLSForSpv.create(self.Descriptors,FLstClassAtt,examples,(self.OpPrmSpv as TOpPrmCPLS).FNbAxis,(self.OpPrmSpv as TOpPrmCPLS).RdYCutValue);
 FCalc.runAnalysis(examples);
 TraceLog.WriteToLogFile(format('[C-PLS] used axis for PLS = %d',[FCalc.NbAxis]));
 //post-calcul - les VIP ici
 postLearning();
 //and then...
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

procedure TCalcSpvPLS.createStructures;
begin
 //petite feinte pour lui faire croire qu'on est dans le cadre de la rgression
 FLstClassAtt:= TLstAttributes.Create(false,self.ClassAttribute.Size);
 FLstClassAtt.Add(self.ClassAttribute);
 //tableaux intermdiaires pour le classement
 setLength(FTabY,FLstClassAtt.Count);//"1" quoi
end;

procedure TCalcSpvPLS.destroyStructures;
begin
 finalize(FTabY);
 setLength(FGlobVIP,0);
 if assigned(FCalc) then FreeAndNil(FCalc);
 if assigned(FLstClassAtt) then FreeAndNil(FLstClassAtt); 
end;

function TCalcSpvPLS.getHTMLResults: string;
var s : string;
    buf: TBufString;
    j: integer;
    sCol: string;
begin
  buf:= TBufString.Create();
  buf.BeginUpdate();
  s:= format('<BR><BR><H4>Number of PLS axis used = %d</H4>',[FCalc.NbAxis]);
  s:= s + format('<H4>Positive class value = %s</H4>',[self.ClassAttribute.LstValues.getDescription(2)]);
  s:= s + '<H3>Regression coefficients</H3>';
  s:= s + HTML_HEADER_TABLE_RESULT;
  s:= s + HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Attribute</TH>';
  s:= s + format('<TH>%s = %s</TH>',[self.ClassAttribute.Name,self.ClassAttribute.LstValues.getDescription(2)]);
  s:= s + '<TH width=70>VIP</TH>';
  s:= s + '</TR>';
  buf.AddStr(s);
  
  //les coeffs. pour chaque X
  for j:= 0 to pred(self.Descriptors.Count) do
   begin
    s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',[Self.Descriptors.Attribute[j].Name]);
    s:= s + format('<TD align="right">%.6f</TD>',[FCalc.BCoefReg[j,0]]);
    if (FGlobVIP[j] > 1.0) then sCol:= HTML_BGCOLOR_DATA_RED else sCol:= HTML_BGCOLOR_DATA_GRAY;
    s:= s + format('<TD align="right" %s>%.4f</TD>',[sCol,FGlobVIP[j]]);
    s:= s + '</TR>';
    buf.AddStr(s);
   end;

  //new -- 10/03/2006 -- dernire ligne pour la constante
  s:= HTML_TABLE_COLOR_DATA_GRAY + format('<TH>%s</TH>',['constant']);
  //attention, correction pour que la rgle "Si val.prdiction > 0 alors Pred. = Modalit 2"
  s:= s + format('<TD align="right">%.6f</TD>',[FCalc.BCoefReg[self.Descriptors.Count,0]-FCalc.StatPLS.AvgY[0]]);
  s:= s + '<TD align=center>-</TD>';
  s:= s + '</TR>';
  s:= s + '</table>';
  buf.AddStr(s);
  buf.EndUpdate();
  result:= buf.BufS;
  buf.Free();

end;

procedure TCalcSpvPLS.getScore(example: integer; var postProba: TTabScore);
var value: double;
begin
  //new -- 10/03/2006 -- passer directement les valeurs de la liste de variables
 FCalc.projectionOnLstX(example,self.Descriptors,FTabY);
 //calculer l'cart  la moyenne
 value:= FTabY[0] - FCalc.StatPLS.AvgY[0];
 //et appliquer une fonction logistique (hum, hum, mais faute de mieux...)
 postProba[0]:= 1.0;
 postProba[2]:= 1.0 / (1.0 + exp(-1.0*value));
 postProba[1]:= 1.0 - postProba[2];
end;

procedure TCalcSpvPLS.postLearning;
var j: integer;
begin
 setLength(FGlobVIP,self.Descriptors.Count);
 //la mesure est dj cumulative, on ne prend que le dernier
 for j:= 0 to pred(self.Descriptors.Count) do FGlobVIP[j]:= FCalc.VIP[j,pred(FCalc.NbAxis)];
end;

initialization
 CLASSES.RegisterClass(TMLGCompSpvCPLS);
end.
