(*************************************************************)
(* UCompSpvPLSDA.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(*************************************************************)

{
@abstract(PLS pour le classement)
@author(Ricco)
@created(02/05/2008)

Adaptation de la rgression PLS pour le classement multi-classes. Mthode
appele PLS-DA.

Voir les rfs.
M. Baker, W. Rayens, "Partial Least Squares for Discrimination", in J. Chemometrics,
17:166-173, 2003.
D. Bertrand, P. Courcoux, C. Camps, "La discrimination par rgression PLS sur indicatrices",
Chimiomtrie, 108-111, 2004.
et surtout
> S. Chevallier, D. Bertrand, A. Kohler, P. Courcoux, "Application of PLS-DA in multivariate image analysis",
in J. Chemometrics, 20: 221-229, 2006.

L'ide est simple : la classe  prdire est transforme en tableau d'indicatrices,
puis on lance un PLS-2 classique. Pour classer un nouvel individu, on calcule la prdiction
pour chaque indicatrice, et on prdit la modalit qui prsente la valeur la plus leve.

L'implmentation retenue dans Tanagra introduit une ide supplmentaire, on peut dtecter automatiquement
le "bon" nombre d'axes en se rfrant simplement  la redondance sur Y. Ds qu'elle est en dessous d'un seuil, on
arrte de gnrer de nouveaux axes. 

}

unit UCompSpvPLSDA;

interface

USES
       Classes, Forms, IniFiles,
       UCompDefinition,
       UCompSpvLDefinition,
       UOperatorDefinition,
       UDatasetImplementation,
       UDatasetExamples, UCalcSpvStructScore, UDatasetDefinition,
       UCalcPLSSupervised, JFGCalcPLS, UCompSpvCPLS;

TYPE
       TMLGCompSpvPLSDA = class(TMLGenCompSpvLearning)
       public
       function    GetClassMLComponent: TClassMLComponent; override;
       end;

       TMLCompSpvPLSDA = class(TMLCompSpvLearning)
       protected
       function    getClassOperator: TClassOperator; override;
       function    GetLogResultDescription(): string; override;
       end;

       TOpSpvPLSDA = class(TOpSpvLearningContinuous)
       protected
       function    getClassParameter: TClassOperatorParameter; override;
       function    getClassSpvLearning(): TClassCalcSpvLearning; override;
       function    ConnectDescriptors(prmData: TMLDataset): boolean; override;
       end;

       TOpPrmPLSDA = class(TOpPrmCPLS)
       end;

       //classe de calcul, son rle est d'appeler la classe de JFG
       TCalcSpvPLSDA = class(TCalcSpvLearning)
       protected
       //oprateur de calcul
       FCalc: TCalcPLSForSpv;
       //liste intermdiaire pour l'attribut classe
       FLstClassAtt: TLstAttributes;
       //tableaux intermdiaires pour le classement
       FTabY: TVectorPLS;
       //vecteur des VIP globaux - cf. tutoriel
       FGlobVIP: TVectorPLS;
       protected
       procedure   createStructures(); override;
       procedure   destroyStructures(); override;
       //insrer les valeurs dans les indicatrices
       procedure   populateValues(examples: TExamples);
       //calcul aprs apprentissage
       procedure   postLearning();
       public
       //surcharge normale dans la hirachie
       function    coreLearning(examples: TExamples): boolean; override;
       //prospective totale -- utiliser une fonction logistique sur la sortie
       procedure   getScore(example: integer; var postProba: TTabScore); override;
       //rechercher le score maximum...
       procedure   classification(example: integer; var response: TTypeDiscrete); override;
       //description des rsultats
       function    getHTMLResults(): string; override;
       end;


implementation

USES
    Sysutils, MATH,
    ULogFile, UConstConfiguration, UDlgOpPrmSpvPLSDA, UStringAddBuffered;

{ TMLGCompSpvPLSDA }

function TMLGCompSpvPLSDA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvPLSDA; 
end;

{ TMLCompSpvPLSDA }

function TMLCompSpvPLSDA.getClassOperator: TClassOperator;
begin
 result:= TOpSpvPLSDA; 
end;

function TMLCompSpvPLSDA.GetLogResultDescription: string;
begin
 result:= format('PLS-DA result generated [%s]',[self.Description]);
end;

{ TOpSpvPLSDA }

function TOpSpvPLSDA.ConnectDescriptors(prmData: TMLDataset): boolean;
var ok: boolean;
begin
 TRY
 //accepter les prdicteurs quasi-continus, i.e. continus ou discrets binaires
 ok:= (prmData.LstAtts[asInput].Count > 0) AND (prmData.LstAtts[asInput].isAllCategory(caQuasiContinue));
 if ok then FDescriptorsAtt:= prmData.LstAtts[asInput];//branchement direct, pas de recopie locale
 //traite les problmes multi-classes
 ok:= ok and (self.ClassAttribute.nbValues >= 2);
 EXCEPT
 ok:= FALSE;
 END;
 //and then...
 result:= ok;
end;

function TOpSpvPLSDA.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmPLSDA; 
end;

function TOpSpvPLSDA.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcSpvPLSDA; 
end;

{ TCalcSpvPLSDA }

procedure TCalcSpvPLSDA.classification(example: integer;
  var response: TTypeDiscrete);
var k,kMax: integer;
    vMax: double;
begin
 FCalc.projectionOnLstX(example,self.Descriptors,FTabY);
 //chercher le max.
 vMax:= -1.0 * Math.MaxDouble;
 kMax:= 0;
 for k:= 0 to length(FTabY)-1 do
  begin
   if (FTabY[k] > vMax)
    then
     begin
      vMax:= FTabY[k];
      kMax:= k;
     end;
  end;
 //renvoyer celui qui correspond au max., tout bonnement
 response:= succ(kMax);
end;

function TCalcSpvPLSDA.coreLearning(examples: TExamples): boolean;
var prm: TOpPrmPLSDA;
begin
 prm:= self.OpPrmSpv as TOpPrmPLSDA;
 if assigned(FCalc) then FreeAndNil(FCalc);
 TRY
 self.populateValues(examples);
 FCalc:= TCalcPLSForSpv.create(self.Descriptors,FLstClassAtt,examples,prm.NbAxis,prm.RdYCutValue);
 FCalc.runAnalysis(examples);
 TraceLog.WriteToLogFile(format('[PLS-DA] used axis for PLS = %d',[FCalc.NbAxis]));
 self.postLearning();
 result:= true;
 EXCEPT
 result:= false;
 END;
end;

procedure TCalcSpvPLSDA.createStructures;
var k: integer;
begin
 //la liste est propritaire cette fois-ci, pour crer les dummy variables
 FLstClassAtt:= TLstAttributes.Create(true,self.ClassAttribute.Size);
 //pour chaque modalit de Y
 for k:= 1 to self.ClassAttribute.NbValues do
  FLstClassAtt.Add(TAttContinue.Create(self.ClassAttribute.LstValues.getDescription(k),self.ClassAttribute.Size));
 //tableaux intermdiaires pour le classement
 setLength(FTabY,FLstClassAtt.Count);
end;

procedure TCalcSpvPLSDA.destroyStructures;
begin
 finalize(FTabY);
 setLength(FGlobVIP,0);
 if assigned(FCalc) then FreeAndNil(FCalc);
 if assigned(FLstClassAtt) then FreeAndNil(FLstClassAtt); //ce qui dtruit les objets dans la liste puisqu'elle est propritaire
end;

function TCalcSpvPLSDA.getHTMLResults: string;
var s,tmp,sCol: string;
    buf: TBufString;
    j,c: integer;
begin
 s:= format('<H4>Number of PLS axis used = %d</H4>',[FCalc.NbAxis]);
 s:= s + '<H3>Classification functions</H3>';
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(s);
 tmp:= HTML_HEADER_TABLE_RESULT;
 tmp:= tmp + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Attribute</TH>';
 for c:= 1 to self.ClassAttribute.nbValues do tmp:= tmp + format('<TH width="70">%s</TH>',[self.ClassAttribute.LstValues.getDescription(c)]);
 tmp:= tmp + '<TH width="70">VIP</TH>';
 tmp:= tmp + '</TR>';
 buf.AddStr(tmp);
 for j:= 1 to self.Descriptors.Count do
  begin
   tmp:= HTML_TABLE_COLOR_DATA_GRAY;
   tmp:= tmp + format('<TH %s>%s</TH>',[HTML_BGCOLOR_HEADER_GRAY,self.Descriptors.Attribute[j-1].Name]);
   for c:= 0 to self.ClassAttribute.nbValues-1 do
    tmp:= tmp + format('<TD align="right">%.6f</TD>',[FCalc.BCoefReg[j-1,c]]);
   if (FGlobVIP[j-1] > 1.0) then sCol:= HTML_BGCOLOR_DATA_RED else sCol:= HTML_BGCOLOR_DATA_GRAY;
   tmp:= tmp + format('<TD align="right" %s>%.4f</TD>',[sCol,FGlobVIP[j-1]]);
   tmp:= tmp + '</TR>';
   buf.AddStr(tmp);
  end;
 //la constante
 tmp:= HTML_TABLE_COLOR_DATA_GRAY;
 tmp:= tmp + format('<TH %s>%s</TH>',[HTML_BGCOLOR_HEADER_GRAY,'constant']);
 for c:= 0 to self.ClassAttribute.nbValues-1 do
  tmp:= tmp + format('<TD align="right">%.6f</TD>',[FCalc.BCoefReg[self.Descriptors.Count,c]]);
 tmp:= tmp + '<TD align=center>-</TD>';
 tmp:= tmp + '</TR>';
 buf.AddStr(tmp);
 //and then...
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

procedure TCalcSpvPLSDA.getScore(example: integer;
  var postProba: TTabScore);
var value,sum: double;
    k: integer;
begin
  //new -- 10/03/2006 -- passer directement les valeurs de la liste de variables
 FCalc.projectionOnLstX(example,self.Descriptors,FTabY);
 //affecter et faire la somme - transformation sigmode trs bte
 sum:= 0.0;
 for k:= 0 to length(FtabY)-1 do
  begin
   value:= FTabY[k];
   postProba[k+1]:= 1.0 / (1.0 + exp(-1.0*value));
   sum:= sum + postProba[k+1];
  end;
 //normalisation  1
 postProba[0]:= sum;
 for k:= 1 to self.ClassAttribute.nbValues do
  postProba[k]:= postProba[k] / sum;
 //TraceLog.WriteToLogFile(format('[PLS-DA] score[2] = %.4f ; sum des scores = %.4f',[postProba[2],sum]));
end;

(*
var k,kMax: integer;
    vMax: double;
begin
 FCalc.projectionOnLstX(example,self.Descriptors,FTabY);
 //chercher le max.
 vMax:= -1.0 * Math.MaxDouble;
 kMax:= 0;
 for k:= 0 to length(FTabY)-1 do
  begin
   if (FTabY[k] > vMax)
    then
     begin
      vMax:= FTabY[k];
      kMax:= k;
     end;
  end;
 //renvoyer celui qui correspond au max., tout bonnement
 response:= succ(kMax);
end;
*)

procedure TCalcSpvPLSDA.populateValues(examples: TExamples);
var i,example: integer;
    k: integer;
begin
 for i:= 1 to examples.Size do
  begin
   example:= examples.Number[i];
   for k:= 0 to pred(FLstClassAtt.Count) do FLstClassAtt.Attribute[k].cValue[example]:= 0.0;//rz
   k:= self.ClassAttribute.dValue[example];//rcuprer la modalit
   FLstClassAtt.Attribute[pred(k)].cValue[example]:= 1.0;//affecter le "1"
  end;
end;

procedure TCalcSpvPLSDA.postLearning;
var j: integer;
begin
 setLength(FGlobVIP,self.Descriptors.Count);
 //la mesure est dj cumulative, on ne prend que le dernier
 for j:= 0 to pred(self.Descriptors.Count) do FGlobVIP[j]:= FCalc.VIP[j,pred(FCalc.NbAxis)];
end;

initialization
 RegisterClass(TMLGCompSpvPLSDA);
end.
