(*************************************************************)
(* UCompSpvPLSDA.pas - Copyright (c) 2008 Ricco RAKOTOMALALA *)
(*************************************************************)

{
@abstract(PLS pour le classement)
@author(Ricco)
@created(03/05/2008)

Adaptation de la rgression PLS pour le classement multi-classes. Mthode
appele PLS-LDA.

Toujours la mme ref.
 S. Chevallier, D. Bertrand, A. Kohler, P. Courcoux, "Application of PLS-DA in multivariate image analysis",
in J. Chemometrics, 20: 221-229, 2006.

L'ide est simple : la classe  prdire est transforme en tableau d'indicatrices,
puis on lance un PLS-2 classique. Puis dans un deuxime temps on procdre  une LDA sur les axes
factoriels (appels "scores" dans la littrature). Une autre technique de rgularisation en quelques sorte.

L'implmentation retenue dans Tanagra introduit une ide supplmentaire, on peut dtecter automatiquement
le "bon" nombre d'axes en se rfrant simplement  la redondance sur Y. Ds qu'elle est en dessous d'un seuil, on
arrte de gnrer de nouveaux axes. 

}

unit UCompSpvPLSLDA;

interface

USES
       Classes, Forms, IniFiles,
       UCompDefinition,
       UCompSpvLDefinition,
       UOperatorDefinition,
       UDatasetImplementation,
       UDatasetExamples, UCalcSpvStructScore, UDatasetDefinition,
       UCalcPLSSupervised, JFGCalcPLS, UCompSpvPLSDA, UCompSpvDiscriminantAnalysis;

TYPE
       TMLGCompSpvPLSLDA = class(TMLGCompSpvPLSDA)
       public
       function    GetClassMLComponent: TClassMLComponent; override;       
       end;

       TMLCompSpvPLSLDA = class(TMLCompSpvPLSDA)
       protected
       function    getClassOperator: TClassOperator; override;
       function    GetLogResultDescription(): string; override;
       end;

       TOpSpvPLSLDA = class(TOpSpvPLSDA)
       protected
       function    getClassParameter: TClassOperatorParameter; override;
       function    getClassSpvLearning(): TClassCalcSpvLearning; override;
       end;

       TOpPrmPLSLDA = class(TOpPrmPLSDA)
       end;

       //classe de calcul, son rle est d'appeler la classe de JFG
       TCalcSpvPLSLDA = class(TCalcSpvPLSDA)
       protected
       //liste intermrdiaire pour les axes factoriels
       FLstScores: TLstAttributes;
       //paramtre pour l'analyse discriminante
       FPrmLDA: TOpPrmLDiscAnalysis;
       //calcul LDA
       FCalcLDA: TCalcLDiscAnalysis;
       //coef. de classement
       FCoef: TMatrixPLS;
       procedure   createStructures(); override;
       procedure   destroyStructures(); override;
       procedure   computeCoefficients();
       public
       //surcharge normale dans la hirachie
       function    coreLearning(examples: TExamples): boolean; override;
       //prospective totale -- utiliser une fonction logistique sur la sortie
       procedure   getScore(example: integer; var postProba: TTabScore); override;
       //rechercher le score maximum...
       procedure   classification(example: integer; var response: TTypeDiscrete); override;
       //description des rsultats
       function    getHTMLResults(): string; override;
       end;


implementation

USES
    Sysutils, MATH,
    ULogFile, UStringAddBuffered, UConstConfiguration;

{ TMLCompSpvPLSLDA }

function TMLCompSpvPLSLDA.getClassOperator: TClassOperator;
begin
 result:= TOpSpvPLSLDA;
end;

function TMLCompSpvPLSLDA.GetLogResultDescription: string;
begin
  result:= format('PLS-LDA result generated [%s]',[self.Description]);
end;

{ TOpSpvPLSLDA }

function TOpSpvPLSLDA.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmPLSLDA;
end;

function TOpSpvPLSLDA.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcSpvPLSLDA;
end;

{ TCalcSpvPLSLDA }

procedure TCalcSpvPLSLDA.classification(example: integer;
  var response: TTypeDiscrete);
begin
 FCalcLDA.classification(example,response);
end;

procedure TCalcSpvPLSLDA.computeCoefficients;
var coefScores: TMatrixPLS;
    c,j,k: integer;
begin
 //construire d'abord le tableau des coefs. des scores (axes factoriels)
 setLength(coefScores,self.Descriptors.Count+1,FCalc.NbAxis);
 for k:= 0 to FCalc.NbAxis-1 do
  begin
   //la constante
   coefScores[0,k]:= 0.0;
   for j:= 0 to self.Descriptors.Count-1 do
    if (FCalc.UsedStdDevX[j] > 0.0) then coefScores[0,k]:= coefScores[0,k] - FCalc.WhEtoile[j,k] * FCalc.StatPLS.FAvgX[j] / FCalc.UsedStdDevX[j];
   //les coefs. dnormaliss pour chaque descripteur
   for j:= 1 to self.Descriptors.Count do
    if (FCalc.UsedStdDevX[j-1] > 0.0) then coefScores[j,k]:= FCalc.WhEtoile[j-1,k] / FCalc.UsedStdDevX[j-1];
  end;
 //construire maintenant la combinaison avec les coefs. de l'analyse discriminante
 setLength(FCoef,self.Descriptors.Count+1,self.ClassAttribute.nbValues);
 for c:= 0 to self.ClassAttribute.nbValues-1 do
  begin
   //la constante
   FCoef[0,c]:= FCalcLDA.MatScores^[0]^[c+1];
   for k:= 0 to FCalc.NbAxis-1 do FCoef[0,c]:= FCoef[0,c] + FCalcLDA.MatScores^[k+1]^[c+1] * CoefScores[0,k];
   //les autres coefs.
   for j:= 1 to self.Descriptors.Count do
    begin
     FCoef[j,c]:= 0.0;
     for k:= 0 to FCalc.NbAxis-1 do FCoef[j,c]:= FCoef[j,c] + FCalcLDA.MatScores^[k+1]^[c+1] * CoefScores[j,k];
    end;
  end;
 //librer le tableau intermdiaire
 setLength(coefScores,0,0);
end;

function TCalcSpvPLSLDA.coreLearning(examples: TExamples): boolean;
var j: integer;
    prm: TOpPrmPLSLDA;
begin
 prm:= self.OpPrmSpv as TOpPrmPLSLDA;
 if assigned(FCalc) then FreeAndNil(FCalc);
 if assigned(FCalcLDA) then FreeAndNil(FCalcLDA);
 TRY
 self.populateValues(examples);
 FCalc:= TCalcPLSForSpv.create(self.Descriptors,FLstClassAtt,examples,prm.NbAxis,prm.RdYCutValue);
 FCalc.runAnalysis(examples);
 TraceLog.WriteToLogFile(format('[PLS-LDA] used axis for PLS = %d',[FCalc.NbAxis]));
 //puis dans un deuxime temps, construire les axes factoriels
 FLstScores.Clear();
 for j:= 1 to FCalc.NbAxis do FLstScores.Add(TAttContinue.Create('',self.ClassAttribute.Size));
 FCalc.setProjectionAxis(FLstScores);
 //et enfin, lancer l'analyse discriminante sur les axes factoriels...
 FCalcLDA:= TCalcLDiscAnalysis.create(FPrmLDA,Self.ClassAttribute,FLstScores,nil);//pas besoin de tous les attributs...
 result:= FCalcLDA.learning(examples);
 //calculer les coefficients de classement
 if (result = true) then self.computeCoefficients();
 EXCEPT
 result:= false;
 END;
end;

procedure TCalcSpvPLSLDA.createStructures;
begin
 inherited;
 FLstScores:= TLstAttributes.Create(TRUE,self.ClassAttribute.Size);
 FPrmLDA:= TOpPrmLDiscAnalysis.Create(nil);
 FPrmLDA.MatInversion:= 0;//mthode exacte..., voir si ridge est mieux...
end;

procedure TCalcSpvPLSLDA.destroyStructures;
begin
 inherited;
 if assigned(FLstScores) then FreeAndNil(FLstScores);
 setLength(FCoef,0,0);
end;

function TCalcSpvPLSLDA.getHTMLResults: string;
var s,tmp: string;
    buf: TBufString;
    j,c: integer;
begin
 s:= format('<H4>Number of PLS axis used = %d</H4>',[FCalc.NbAxis]);
 s:= s + '<H3>Classification functions</H3>';
 buf:= TBufString.Create();
 buf.BeginUpdate();
 tmp:= HTML_HEADER_TABLE_RESULT;
 tmp:= tmp + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Attribute</TH>';
 for c:= 1 to self.ClassAttribute.nbValues do tmp:= tmp + format('<TH width="70">%s</TH>',[self.ClassAttribute.LstValues.getDescription(c)]);
 tmp:= tmp + '</TR>';
 buf.AddStr(tmp);
 for j:= 1 to self.Descriptors.Count do
  begin
   tmp:= HTML_TABLE_COLOR_DATA_GRAY;
   tmp:= tmp + format('<TH %s>%s</TH>',[HTML_BGCOLOR_HEADER_GRAY,self.Descriptors.Attribute[j-1].Name]);
   for c:= 0 to self.ClassAttribute.nbValues-1 do
    tmp:= tmp + format('<TD align="right">%.6f</TD>',[FCoef[j,c]]); 
   tmp:= tmp + '</TR>';
   buf.AddStr(tmp);
  end;
 //la constante
 tmp:= HTML_TABLE_COLOR_DATA_GRAY;
 tmp:= tmp + format('<TH %s>%s</TH>',[HTML_BGCOLOR_HEADER_GRAY,'constant']);
 for c:= 0 to self.ClassAttribute.nbValues-1 do
  tmp:= tmp + format('<TD align="right">%.6f</TD>',[FCoef[0,c]]);
 tmp:= tmp + '</TR>';
 buf.AddStr(tmp);
 //and then...
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= s + buf.BufS;
 buf.Free();
end;

procedure TCalcSpvPLSLDA.getScore(example: integer;
  var postProba: TTabScore);
begin
 FCalcLDA.getScore(example,postProba);
end;

{ TMLGCompSpvPLSLDA }

function TMLGCompSpvPLSLDA.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvPLSLDA;
end;

initialization
 RegisterClass(TMLGCompSpvPLSLDA);
end.
