(*********************************************************************)
(* UCompSpvRuleInduction.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(*********************************************************************)

{
@abstract(Induction de rgles)
@author(Ricco)
@created(11/11/2009)

Inspir de la seconde version de CN2 (Article originel de CLARK et Boswell., 1991)
}
unit UCompSpvRuleInduction;

interface

USES
       Classes, Forms, IniFiles,
       UCompDefinition,
       UCompSpvLDefinition,
       UOperatorDefinition,
       UDatasetImplementation,
       UCalcSpvRuleInduction;

TYPE
       TMLGCompSpvRuleInduction = class(TMLGenCompSpvLearning)
       public
       function    GetClassMLComponent: TClassMLComponent; override;
       end;

       TMLCompSpvRuleInduction = class(TMLCompSpvLearning)
       protected
       function    getClassOperator: TClassOperator; override;
       end;

       TOpSpvRuleInduction = class(TOpSpvLearningDiscrete)
       protected
       function    getClassParameter: TClassOperatorParameter; override;
       function    getClassSpvLearning(): TClassCalcSpvLearning; override;
       end;

       TOpPrmRuleInduction = class(TOpPrmSpvLearning)
       private
       //mesure
       FMeasurement: TEnumMeasureRuleInduction;
       //Risque critique pour l'arrt -- significance level
       FSignifLevel: double;
       //Support (premisse) minimum de la rgle
       FSupportMin: integer;
       //Poids accord  la puret dans l'asymtrique J-Measure
       FWeight: double;
       protected
       function    CreateDlgParameters(): TForm; override;
       procedure   SetDefaultParameters(); override;
       public
       //affichage
       function    getHTMLParameters(): string; override;
       //I-O
       procedure   LoadFromStream(prmStream: TStream); override;
       procedure   SaveToStream(prmStream: TStream); override;
       procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
       procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
       //
       property SignifLevel: double read FSignifLevel write FSignifLevel;
       property SupportMin: integer read FSupportMin write FSupportMin;
       property Measurement: TEnumMeasureRuleInduction read FMeasurement write FMeasurement;
       property Weight: Double read FWeight write FWeight;
       end;

implementation

uses
    UConstConfiguration, Sysutils, UDlgOpPrmSpvRuleInduction;

{ TMLGCompSpvRuleInduction }

function TMLGCompSpvRuleInduction.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompSpvRuleInduction;
end;

{ TMLCompSpvRuleInduction }

function TMLCompSpvRuleInduction.getClassOperator: TClassOperator;
begin
 result:= TOpSpvRuleInduction;
end;

{ TOpSpvRuleInduction }

function TOpSpvRuleInduction.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmRuleInduction;
end;

function TOpSpvRuleInduction.getClassSpvLearning: TClassCalcSpvLearning;
begin
 result:= TCalcSpvRuleInduction;
end;

{ TOpPrmRuleInduction }

function TOpPrmRuleInduction.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmSpvRuleInduction.CreateFromOpPrm(self);
end;

function TOpPrmRuleInduction.getHTMLParameters: string;
var s: string;
begin
 s:= '';
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_GRAY+'<TH colspan=2>Parameters</TH></TR>';
 //>>
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Rule evaluation measure</TD><TD align="right">%s</TD></TR>',[STR_MEASURE_RULE_INDUCTION[FMeasurement]]);
 s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Min support of rule</TD><TD align="right">%d</TD></TR>',[FSupportMin]);
 if (FMeasurement = mriAsymJMeasure)
  then s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Purity preference</TD><TD align="right">%.1f</TD></TR>',[FWeight])
  else s:= s+HTML_TABLE_COLOR_DATA_GRAY+format('<TD>Significance level</TD><TD align="right">%.4f</TD></TR>',[FSignifLevel]);
 //<<
 s:= s+'</table>';
 //and then...
 result:= s;
end;

procedure TOpPrmRuleInduction.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 FSignifLevel:= prmINI.ReadFloat(prmSection,'sig_level',FSignifLevel);
 FSupportMin:= prmINI.ReadInteger(prmSection,'support',FSupportMin);
 FMeasurement:= TEnumMeasureRuleInduction(prmINI.ReadInteger(prmSection,'rule_measure',ord(FMeasurement)));
 FWeight:= prmINI.ReadFloat(prmSection,'weight',FWeight);
end;

procedure TOpPrmRuleInduction.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(FSignifLevel,sizeof(FSignifLevel));
 prmStream.ReadBuffer(FSupportMin,sizeof(FSupportMin));
 prmStream.ReadBuffer(FMeasurement,sizeof(FMeasurement));
 prmStream.ReadBuffer(FWeight,sizeof(FWeight));
end;

procedure TOpPrmRuleInduction.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteFloat(prmSection,'sig_level',FSignifLevel);
 prmINI.WriteInteger(prmSection,'support',FSupportMin);
 prmINI.WriteInteger(prmSection,'rule_measure',ord(FMeasurement));
 prmINI.WriteFloat(prmSection,'weight',FWeight);
end;

procedure TOpPrmRuleInduction.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(FSignifLevel,sizeof(FSignifLevel));
 prmStream.WriteBuffer(FSupportMin,sizeof(FSupportMin));
 prmStream.WriteBuffer(FMeasurement,sizeof(FMeasurement));
 prmStream.WriteBuffer(FWeight,sizeof(FWeight));
end;

procedure TOpPrmRuleInduction.SetDefaultParameters;
begin
 //mesure
 FMeasurement:= mriMisclassification;
 //Risque critique pour l'arrt -- significance level
 FSignifLevel:= 0.05;
 //Support (premisse) minimum de la rgle
 FSupportMin:= 10;
 //Poids accord  la puret dans l'asymtrique J-Measure
 FWeight:= 4.0;
end;

initialization
 RegisterClass(TMLGCompSpvRuleInduction);
end.
