(*********************************************************************)
(* UCalcSpvSMOPolyKernel.pas - Copyright (c) 2005 Ricco RAKOTOMALALA *)
(*********************************************************************)

{
@abstract(Support vector machine)
@author(Ricco)
@created(25/09/2008)

Copie de la classe CahedKernel de WEKA, Revision: 1.7
Tentative de cache plus performant, mais abandonne parce que plantage plus que de raison.
De toute manire, la diffrence se fait surtout au niveau des dotProd...
}
unit UCalcSpvSMOCachedKernel;

interface

USES
    UDatasetExamples,
    UCalcSpvSVMBinarySMO;

TYPE
    TCachedKernel = class(TKernel)
    protected
    //** Counts the number of kernel evaluations. */
    m_kernelEvals: integer;

    //** Counts the number of kernel cache hits. */
    m_cacheHits: integer;

    //** The size of the cache (a prime number) */
    m_cacheSize: integer;

    //** Kernel cache */
    m_storage: array of double;
    m_keys: array of longint;

    //** The kernel matrix if full cache is used (i.e. size is set to 0) */
    //protected double[][] m_kernelMatrix;
    m_kernelMatrix: array of array of double;

    //** The number of instance in the dataset */
    m_numInsts: integer;

    //** number of cache slots in an entry */
    m_cacheSlots: integer;

    public

    //constructeur
    constructor create(operator: TBinarySMO; instances: TExamples);

    //buildKernel
    procedure   buildKernel(operator: TBinarySMO; instances: TExamples); virtual;

    //initVars
    procedure  initVars(operator: TBinarySMO; instances: TExamples); virtual;

    function evaluate(id1, id2: integer; inst1: integer): double; virtual; abstract;

    //eval
    function eval(id1, id2: integer; inst1: integer): double; override;

    //
    function numEvals(): integer; override;

    //
    function numCacheHits(): integer; override;

    //
    function dotProd(inst1, inst2: integer): double;

    end;

    TPolyKernel = class(TCachedKernel)
    protected
    //** Use lower-order terms? -- private boolean m_lowerOrder = false;
    m_lowerOrder: boolean;
    //** The exponent for the polynomial kernel. -- private double m_exponent = 1.0;
    m_exponent: double;
    public
    constructor create(operator: TBinarySMO; instances: TExamples; cacheSize: integer; exponent: double; lowerOrder: boolean);
    procedure clean(); override;
    function evaluate(id1, id2: integer; inst1: integer): double; override;
    end;

 TNormalizedPolyKernel = class(TPolyKernel)
 public
 function eval(id1, id2: integer; inst1: integer): double; override;
 end;

implementation

uses
    Math;

{ TCachedKernel }

procedure TCachedKernel.buildKernel(operator: TBinarySMO; instances: TExamples);
begin
 //initaliser le cache
 initVars(operator,instances);
end;

constructor TCachedKernel.create(operator: TBinarySMO; instances: TExamples);
begin
 inherited create(operator);
 m_cacheSize:= 250007;
 m_cacheSlots:= 4;
 //crer le kernel
 buildKernel(operator, instances);
end;

function TCachedKernel.dotProd(inst1, inst2: integer): double;
var somme: double;
    //p1, p2: integer;
    j: integer;
begin
 //formule maison -- autrement plus puissant -- DOTPROD :: produit scalaire
 somme:= 0.0;
 for j:= 0 to pred(FOperator.Descriptors.Count) do
  //somme:= somme+m_operator.Descriptors.Attribute[j].cValue[inst1]*m_operator.Descriptors.Attribute[j].cValue[inst2];
  somme:= somme+FOperator.attTransSVM.cValue[j,inst1]*FOperator.attTransSVM.cValue[j,inst2];
 //and then...
 result:= somme;
end;

function TCachedKernel.eval(id1, id2, inst1: integer): double;
var tmps: double;
    key,thiskey: longint;
    location,loc,i,j: integer;
begin
    //result:= 0;
    key:= -1;
    location:= -1;

    // we can only cache if we know the indexes and caching is not
    // disbled (m_cacheSize == -1)
    if ( (id1 >= 0) and (m_cacheSize <> -1) ) then begin

    // Use full cache?
    if (m_cacheSize = 0) then begin
     //si pas cr, on doit le faire...
	 if (m_kernelMatrix = nil) then begin
      setLength(m_KernelMatrix,m_numInsts);
	  for i := 0 to pred(m_numInsts) do
       begin
	    setLength(m_kernelMatrix[i],i+1);
	    for j := 0 to i do begin
	      inc(m_kernelEvals);
	      m_kernelMatrix[i][j] := evaluate(i, j, m_data.Number[succ(i)]);
	    end;
       end;
	  end;
	inc(m_cacheHits);
	if (id1 > id2) then result:= m_kernelMatrix[id1][id2] else result:= m_kernelMatrix[id2][id1];
	exit;
    end;

    // Use LRU cache
    if (id1 > id2) then
	key := (id1 + (id2 * m_numInsts))
      else
	key := (id2 + ( id1 * m_numInsts));

      location := (key mod m_cacheSize) * m_cacheSlots;
      loc := location;
      for i := 0 to pred(m_cacheSlots) do begin
	   thiskey := m_keys[loc];
	   if (thiskey = 0) then break; // empty slot, so break out of loop early
	   if (thiskey = (key + 1)) then begin
	    inc(m_cacheHits);

	   // move entry to front of cache (LRU) by swapping
	   // only if it's not already at the front of cache
	   if (i > 0) then begin
	    tmps := m_storage[loc];
	    m_storage[loc] := m_storage[location];
	    m_keys[loc] := m_keys[location];
	    m_storage[location] := tmps;
	    m_keys[location] := thiskey;
	    result:= tmps;
        exit;
	    end else
        begin
        result:= m_storage[loc];
        exit;
        end;
	   end;
	  inc(loc);
      end;
    end;

    result := evaluate(id1, id2, inst1);

    inc(m_kernelEvals);

    // store result in cache
    if ( (key <> -1) and (m_cacheSize <> -1) ) then begin
      // move all cache slots forward one array index
      // to make room for the new entry
      //System.arraycopy(m_keys, location, m_keys, location + 1, m_cacheSlots - 1);
      //for i:= pred(m_cacheSlots) downto 1 do m_keys[location+i]:= m_keys[location+i-1];
      System.Move(m_Keys[location],m_Keys[location+1],(m_cacheSlots-1)*sizeof(longint));
      //System.arraycopy(m_storage, location, m_storage, location + 1, m_cacheSlots - 1);
      //for i:= pred(m_cacheSlots) downto 1 do m_storage[location+i]:= m_storage[location+i-1];
      System.Move(m_storage[location],m_storage[location+1],(m_cacheSlots-1)*sizeof(integer));
      //
      m_storage[location] := result;
      m_keys[location] := (key + 1);
    end;
    //return result;
    //on a dj la valeur de result

end;

procedure TCachedKernel.initVars(operator: TBinarySMO; instances: TExamples);
begin
    m_kernelEvals := 0;
    m_cacheHits   := 0;
    m_data        := instances;
    m_numInsts    := m_data.Size;

    if (m_cacheSize > 0) then
    begin
      // Use LRU cache
      setLength(m_storage,m_cacheSize * m_cacheSlots);
      setLength(m_keys, m_cacheSize * m_cacheSlots);
    end
    else
    begin
      m_storage      := nil;
      m_keys         := nil;
      m_kernelMatrix := nil;
    end;
end;

function TCachedKernel.numCacheHits: integer;
begin
 result:= m_CacheHits;
end;

function TCachedKernel.numEvals: integer;
begin
 result:= m_KernelEvals;
end;

{ TPolyKernel }

procedure TPolyKernel.clean;
begin
 setLength(m_storage,0);
 setLength(m_keys,0);
 m_KernelMatrix:= nil;
end;

constructor TPolyKernel.create(operator: TBinarySMO; instances: TExamples;
  cacheSize: integer; exponent: double; lowerOrder: boolean);
begin
 inherited create(operator,instances);
 //init.
 m_exponent:= exponent;
 m_lowerOrder:= lowerOrder;
end;

function TPolyKernel.evaluate(id1, id2, inst1: integer): double;
begin

   if (id1 = id2)
    then result:= dotProd(inst1,inst1)
    else result:= dotProd(inst1,m_data.Number[succ(id2)]);

    // Use lower order terms?
    if (m_lowerOrder) then
      result:= result + 1.0;

    if (m_exponent <> 1.0) then
      result := Math.power(result, m_exponent);

    //return result;
end;

{ TNormalizedPolyKernel }

function TNormalizedPolyKernel.eval(id1, id2, inst1: integer): double;
var division: double;
begin
 division:= SQRT(inherited eval(id1,id1,inst1) * inherited eval(id2,id2,m_data.Number[succ(id2)]));
 
 if (division <> 0.0)
  then result:= inherited eval(id1,id2,inst1) / division
  else result:= 0.0;


end;

end.
