(****************************************************************)
(* UFrmBaseOperator.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(****************************************************************)

{
@abstract(Fiche de base pour la visualisation d'un oprateur)
@author(Ricco)
@created(12/01/2004)
RQ: 01/08/03 on ajoute un autre pointeur sur le composant

20/03/2012 -- Mettre le format HTML lors de la copie du contenu du composant dans le presse-papier
Ainsi, la copie des rsultats (HTML) dans Libre Office et Open Office fonctionne galement.
}
unit UFrmBaseOperator;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, UFrmBase, UOperatorDefinition, UCompDefinition,
  ExtCtrls, ComCtrls, ActnList, Menus, HTMLLite, Chart;

type
  TFrmBaseOperator = class(TFrmBase)
    pnlFond: TPanel;
    pgctrlFond: TPageControl;
    tabHTML: TTabSheet;
    ActListFrmOperator: TActionList;
    actCopy: TAction;
    MenuOperator: TMainMenu;
    menuComponent: TMenuItem;
    menuCopy: TMenuItem;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure TimerBusyTimer(Sender: TObject);
    procedure actCopyExecute(Sender: TObject);
    private
    FOperator: TOperator;
    FMLComponent: TMLComponent;
    FHTML: string;
    LiteHTML: ThtmlLite;
    procedure InsertViewHTML();
    public
    {crer et connecter  l'oprateur}
    constructor CreateFromOperator(prmOp: TOperator); virtual;
    destructor  Destroy; override;
    {afficher les infos prvues}
    procedure   PrepareView(); virtual;
    //new -- 16/10/2010 - envoie les ventuelles infos supplmentaires pour le rapport
    function    getAdditionalInfoForReport(): string; virtual;
    //new -- 16/10/2010 -- sauver l'image d'un chart dans un fichier jpeg
    procedure   saveChartAsJpeg(chart: TChart; nomfichier: String);
    //new -- ajouter un chart dans un rapport
    //new -- 07/09/2012 -- ajouter un indicateur de numero pour viter les doublons
    function    addChartToReport(chart: TChart; jpegWidth: integer = 550; numero: integer = 0): string;
    {la proprit oprateur associe}
    property    Operator: TOperator read FOperator;
    property    MLComponent: TMLComponent read FMLComponent;
  end;

  //new -- 08/03/2008 -- meta-classe des fentres de visualisation, cela permet de mieux grer les configurations complexes
  TClassFrmBaseOperator = class of TFrmBaseOperator;

var
  FrmBaseOperator: TFrmBaseOperator;

implementation

USES
        Clipbrd, ULogFile, UBatchExecution, JPeg, UConstConfiguration;

CONST
        TIMER_CYCLE_ATTENTE = 200;


{$R *.dfm}

{ TFrmBaseOperator }

constructor TFrmBaseOperator.CreateFromOperator(prmOp: TOperator);
begin
 inherited Create(NIL);
 FOperator:= prmOp;
 FMLComponent:= prmOp.MLOwner as TMLComponent;
 self.Caption:= FMLComponent.Description;
 //new -- 05/04/2006 -- si BATCH MODE, pas d'affichage dans les fentres de visualisation
 if not(BATCH_MODE_EXECUTION)
  then
   begin
     //insrer le visualisateur HTMLLite
     self.InsertViewHTML();
     //envoyer l'affichage
     self.PrepareView();
   end;
end;

procedure TFrmBaseOperator.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
 FOperator.Form:= NIL;
 Action:= caFree;
end;

procedure TFrmBaseOperator.PrepareView;
var tps: cardinal;
begin
 tps:= GetTickCount();
 FHTML:= MLComponent.getHTMLDescriptionContents();
 tps:= GetTickCount() - tps;
 TraceLog.WriteToLogFile(format('FRMBASEOPERATOR >> duree formation du string2html %d',[tps]));
 
 //solution HtmlLite
 tps:= GetTickCount();
 LiteHTML.LoadFromString(FHTML,'');
 tps:= GetTickCount() - tps;
 TraceLog.WriteToLogFile(format('FRMBASEOPERATOR >> duree chargement dans le browser %d',[tps]));
end;

procedure TFrmBaseOperator.TimerBusyTimer(Sender: TObject);
begin
 (*
 //attendre tant que le navigateur n'est pas prt
 if not(viewHTML.Busy)
  then
   begin
    self.TimerBusy.Enabled:= FALSE;

    //aprs avoir stopp le timer, charger effectivement le document
    viewHTML.OleObject.Document.Close;
    viewHTML.OleObject.Document.Write(FHTML);
   end;
  *)
end;

//new - 20/03/2012 -- http://www.swissdelphicenter.ch/torry/showcode.php?id=1391
(*
function FormatHTMLClipboardHeader(HTMLText: string): string;
const
  CrLf = #13#10;
begin
  Result := 'Version:0.9' + CrLf;
  Result := Result + 'StartHTML:-1' + CrLf;
  Result := Result + 'EndHTML:-1' + CrLf;
  Result := Result + 'StartFragment:000081' + CrLf;
  Result := Result + 'EndFragment:' + CrLf;
  Result := Result + HTMLText + CrLf;
  Result := StringReplace(Result, '', Format('%.6d', [Length(Result)]), []);
end;
*)

//http://www.tmg.it/articoli-tecnici/sviluppo-e-sistema/61-delphi-copy-unicode-html-to-clipboard
function FormatHTMLClipboardHeader(HTMLText: string): string;
const
  CrLf = #13#10;
begin
  Result := 'Version:0.9' + CrLf;
  Result := Result + 'StartHTML:-1' + CrLf;
  Result := Result + 'EndHTML:-1' + CrLf;
  Result := Result + 'StartFragment:^^^^^^' + CrLf;
  Result := Result + 'EndFragment:' + CrLf;
  //longueur de l'en-tte
  Result := StringReplace(Result, '^^^^^^', Format('%.6d', [Length(Result)*SizeOf(Char)]), []);
  //plus la longueur du texte  passer
  Result := Result + HTMLText + CrLf;
  Result := StringReplace(Result, '', Format('%.6d', [Length(Result)*SizeOf(Char)]), []);
end;

//The second parameter is optional and is put into the clipboard as CF_HTML.
//Function can be used standalone or in conjunction with the VCL clipboard so long as
//you use the USEVCLCLIPBOARD conditional define
//($define USEVCLCLIPBOARD}
//(and clipboard.open, clipboard.close).
//Code from http://www.lorriman.com
//new - 20/03/2012 -- http://www.swissdelphicenter.ch/torry/showcode.php?id=1391
procedure CopyHTMLToClipBoard(const str: string; const htmlStr: string = '');
var
  gMem: HGLOBAL;
  lp: PChar;
  Strings: array[0..1] of string;
  Formats: array[0..1] of UINT;
  i: Integer;
begin
  gMem := 0;
  {$IFNDEF USEVCLCLIPBOARD}
  Win32Check(OpenClipBoard(0));
  {$ENDIF}
  try
    //most descriptive first as per api docs
    Strings[0] := FormatHTMLClipboardHeader(htmlStr);
    Strings[1] := str;
    Formats[0] := RegisterClipboardFormat('HTML Format');
    Formats[1] := CF_TEXT;
    {$IFNDEF USEVCLCLIPBOARD}
    Win32Check(EmptyClipBoard);
    {$ENDIF}
    for i := 0 to High(Strings) do
    begin
      if Strings[i] = '' then Continue;
      //an extra "1" for the null terminator
      gMem := GlobalAlloc(GMEM_DDESHARE + GMEM_MOVEABLE, Length(Strings[i]) + 1);
      {Succeeded, now read the stream contents into the memory the pointer points at}
      try
        Win32Check(gmem <> 0);
        lp := GlobalLock(gMem);
        Win32Check(lp <> nil);
        CopyMemory(lp, PChar(Strings[i]), Length(Strings[i]) + 1);
      finally
        GlobalUnlock(gMem);
      end;
      Win32Check(gmem <> 0);
      SetClipboardData(Formats[i], gMEm);
      Win32Check(gmem <> 0);
      gmem := 0;
    end;
  finally
    {$IFNDEF USEVCLCLIPBOARD}
    Win32Check(CloseClipBoard);
    {$ENDIF}
  end;
end;

(*
procedure TFrmBaseOperator.actCopyExecute(Sender: TObject);
var curClipBrd: TClipBoard;
begin
 //envoyer la chane de rsultats dans le clipboard
 curClipBrd:= ClipBoard();
 TRY
 curClipBrd.Open();
 curClipbrd.AsText:= FHTML;
 FINALLY
 curClipBrd.Close();
 END;
end;
*)

//new -- 20/03/2012 -- faire passer le format HTML dans la copie vers le presse-papier
procedure TFrmBaseOperator.actCopyExecute(Sender: TObject);
var
  curClipBrd: TClipBoard;
  useOld: boolean;
begin
 TRY
  useOld:= FALSE;
  CopyHTMLToClipBoard(FHTML, FHTML);
 EXCEPT
  //retour  l'ancienne procdure en cas de soucis
  useOld:= TRUE;
 END;
 //problme ?
 if (useOld = TRUE)
  then
   begin
     curClipBrd:= ClipBoard();
     TRY
     curClipBrd.Open();
     curClipbrd.AsText:= FHTML;
     FINALLY
     curClipBrd.Close();
     END;
   end;
end;


destructor TFrmBaseOperator.Destroy;
begin
 if assigned(LiteHTML)
  then FreeAndNil(LiteHTML);
 inherited Destroy();
end;

procedure TFrmBaseOperator.InsertViewHTML;
begin

 LiteHTML:= THTMLLite.Create(tabHTML);
 LiteHTML.Parent:= tabHTML;

 LiteHTML.Align:= alClient;
 LiteHTML.DefBackground:= clWhite;

 //new -- 27/05/2012 -- taille par dfaut de la police
 LiteHTML.DefFontSize:= HTML_FONT_SIZE;

end;

function TFrmBaseOperator.getAdditionalInfoForReport: string;
begin
 result:= '';
end;

procedure TFrmBaseOperator.saveChartAsJpeg(chart: TChart;
  nomfichier: String);
var wmfGraph: TMetaFile;
    jpegGraph: TJPegImage;
    bmpGraph: Graphics.TBitmap;
begin
 wmfGraph:= chart.TeeCreateMetafile(true,chart.GetRectangle());
 //ouah la magouille, on est oblig de passer par un bitmap...
 bmpGraph:= Graphics.TBitmap.Create();
 bmpGraph.Transparent:= true;
 bmpGraph.TransparentMode:= tmAuto;
 bmpGraph.Height:= wmfGraph.Height;
 bmpGraph.Width:= wmfGraph.Width;
 bmpGraph.Canvas.Draw(0,0,wmfGraph);
 //enfin in cre le JPEG
 jpegGraph:= TJPegImage.Create();
 jpegGraph.Transparent:= true;
 jPegGraph.Assign(bmpGraph);
 jPegGraph.SaveToFile(nomFichier);
 //vider tout le monde
 jpegGraph.Free();
 bmpGraph.Free();
 wmfGraph.Free();
end;

function TFrmBaseOperator.addChartToReport(chart: TChart; jpegWidth: integer = 550; numero: integer = 0): string;
var s: string;
    nomFichier: string;
    p: integer;
begin
 //former le nom du fichier
 nomFichier:= (self.Operator.MLOwner as TMLComponent).getHTMLDescriptionName();
 //new - 07/09/2012 - ajouter un numro le cas chant
 if (numero > 0)
  then
   begin
    p:= pos('.',nomFichier);
    nomFichier:= copy(nomFichier,1,p-1);
    nomFichier:= nomFichier + '_' + intToStr(numero);
   end;
 //suite normale...
 nomFichier:= ChangeFileExt(nomFichier,'.jpg');
 //sauver le fichier
 self.saveChartAsJpeg(chart,nomFichier);
 //integrer le graphique dans la page HTML - new 07/09/2012 - mettre en rectangulaire avec "height"
 s:= format('<center><img src="%s" width="%d" height="%d" alt="Chart" border=1></center><br>',[ExtractFileName(nomFichier),jpegWidth,jpegWidth]);
 //and then...
 result:= s;
end;

end.
