(****************************************************************)
(* UCalcVarClusBase.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(****************************************************************)

{
@abstract(Structures de base pour le clustering de variables)
@author(Ricco)
@created(19/02/2007)

}

unit UCalcVarClusBase;

interface

USES
    UDatasetDefinition, UDatasetImplementation,
    UOperatorDefinition,
    UCalcStatDes,
    UDatasetExamples,
    Matrices,
    Contnrs,
    UCalcRndGenerator;

TYPE

    //classe de calcul -- forward
    TCalcVarClus = class;

    //classe de classe de calcul
    TClassCalcVarClus = class of TCalcVarClus;

    //une variable
    TAttVarClus = class
    public
    //attribut, pas propritaire
    Att: TAttribute;
    //numro pour la retrouver dans les calculs, attention le premier numero est  1
    Numero: integer;
    //stat-descriptives, pas propritaire
    StatDes: TCalcStatDesContinuous;
    //constructeur
    constructor create(v: TAttribute; num: integer; stat: TCalcStatDesContinuous; examples: TExamples = nil);
    //destructeur, pas de destruction de l'attribut !!!
    destructor  destroy(); override;
    //calculs statistiques
    procedure   refreshStat(examples: TExamples);
    //calculer la corrlation avec une variable quelconque, calcul demand
    function    getCorrelation(other: TAttribute; examples: TExamples): double; overload;
    //calculer la corrlation avec une autre variable de calcul de cluster, la matrice corrlation est fournie
    function    getCorrelation(other: TAttVarClus; calc: TCalcVarClus): double; overload;
    end;

    //liste de variables
    TLstAttVarClus = class
    public
    //liste, propritaire ou non, selon le cas
    Lst: TObjectList;
    //constructeur, qui dfinit si la liste est propritaire ou non
    constructor create(owner: boolean);
    //desctructeur
    destructor  destroy(); override;
    //ajouter une variable
    procedure   add(v: TAttVarClus);
    //vider
    procedure   clear();
    //rcuprer
    function    getVarClus(index: integer): TAttVarClus;
    //nombre d'lments
    function    getCount(): integer;
    //savoir s'il est propritaire de sa liste
    function    isOwner(): boolean;
    //assigner
    procedure   assign(source: TLstAttVarClus);
    end;

    //un cluster
    TVarCluster = class
    public
    //tableau des corrlations, pass au carr c'est le Own Cluster
    TabOwnCorrelation: array of single;
    //tableau de toutes les corrlations
    TabAllCorrelation: array of single;
    //tableau des corrlations - Next closest
    TabClosestRSquared: array of single;
    //liste des variables associs, non propritaire
    Lst: TLstAttVarClus;
    //variable reprsentant le cluster
    VarClus: TAttribute;
    //valeur propre associe
    EigVal: double;
    //taille du vecteur propre du cluster
    EigVecSize: integer;
    //vecteur propre du cluster
    EigVector: PVector;
    //classe de calcul, pour accder  la matrice de corrlation notamment
    ClusCalc: TCalcVarClus;
    //constructeur
    constructor create(ExampleSize: integer; calc: TCalcVarClus);
    //dupliquer en passant l'oprateur  utiliser...hum...
    function    duplicate(calc: TCalcVarClus): TVarCluster;
    //destructeur
    destructor  destroy(); override;
    //recalculer la variable de classe  partir des variables dispo. i.e. recalculer le centre de gravit en gros...
    procedure   refreshClusVar(examples: TExamples); virtual;
    //effectuer une projection  partir du vecteur propre
    procedure   setProjectionFromEigVector();
    //ajouter une nouvelle variable dans le cluster
    procedure   addVar(attVar: TAttVarClus);
    //nombre de variables dans le cluster
    function    getCount(): integer;
    //vider la liste de variables de cluster
    procedure   clear();
    //obtenir la proportion de variance explique
    function    getPropotionVarExplained(): double;
    //rtrouver la variable la plus proche de la variable de classe
    function    getNearestClusVar(examples: TExamples): TAttribute;
    //rcuprer un reprsentant, UseClusVar = true -> on utilise le reprsentant de classe, sinon on cherche le plus proche
    function    getRepresentant(prmUseClusVar: boolean; examples: TExamples): TAttribute;
    //calcul des corrlations
    procedure   majCorrelation(examples: TExamples);
    //fusion de deux clusters, renvoie la distance d'agrgation
    class function fusion(c1,c2: TVarCluster; examples: TExamples; var dist: double): TVarCluster;
    //calcul de la corrlation entre deux cluster sur la base des variables latentes
    class function correlation(c1,c2: TVarCluster; examples: TExamples): double;
    //rcuprer la matrice de corrlation associe
    function    getCorrMatrix(): PMatrix; 
    end;

    //une partition, une srie de clusters
    TStrucVarCluster = class
    public
    //liste de clusters, propritaire
    LstClus: TObjectList;
    //classe de calcul
    ClusCalc: TCalcVarClus;
    //variance explique
    VarExplained: double;
    //constructeur
    constructor create(calc: TCalcVarClus);
    //destructeur
    destructor  destroy(); override;
    //ajouter un cluster
    procedure   add(cluster: TVarCluster);
    //rcuprer un cluster
    function    get(index: integer): TVarCluster;
    //nombre de clusters
    function    getCount(): integer;
    //rcuprer la variance explique par la partition
    function    computeVarExplained(): double;
    //rafrachir les variables de cluster
    procedure   refreshClusVar(examples: TExamples);
    //vider les listes
    procedure   clearClusters();
    //ajouter une variable au cluster le plus proche (le plus corrl)
    procedure   addVarToNearest(attVar: TAttVarClus; examples: TExamples);
    //mj des next closest
    procedure   majRSquaredClosest(examples: TExamples);
    //mj des correlations et les Own
    procedure   majCorrelation(examples: TExamples);
    //description globale des rsultats -- tab.9.4, p.222, Nakache
    function    getHTMLClusterSummary(): string;
    //description de chaque cluster -- tab.9.5, p.223, Nakache
    function    getHTMLClusterTab(): string;
    //description des correlations -- tab.9.7, p.223, Nakache
    function    getHTMLCorrelations(): string;
    //description des coefficients de rgression -- tab.9.6, p.223, Nakache
    function    getHTMLRegressionCoefficients(): string; 
    end;

    //tableau de variance expliques
    TTabVarExplained = array of single;

    //tableau de paliers pour la dtection du bon dcoupage
    TTabPalier = array of array of single;

    //classe de calcul
    TCalcVarClus = class(TObject)
    private
    //propritaire de la liste des variables compute ou pas ?
    OwnerLstVariables: boolean;
    protected
    //calculer les corrlations et stat. des
    procedure   computeBasicStatsAndCorr(examples: TExamples);
    //calculer les clusters
    procedure   computeClusters(examples: TExamples); virtual; abstract;
    //obtenir les rsultats dtaills
    function    getHTMLDetailedResults(): string; virtual;
    //autres initialisations communes
    procedure   commonInitialisations(); virtual;
    public
    //liste des inputs, pas proprio
    Inputs: TLstAttributes;
    //liste des statdes. -- propritaire
    LstStatDes: TLstCalcStatDesContinuous;
    //liste des variables, propritaire
    LstVariables: TLstAttVarClus;
    //matrice des corrlations
    MatCorr: PMatrix;
    //paramtre de l'oprateur
    OpParameter: TOperatorParameter;
    //gnrateur de nombre alatoire
    RndGenerator: TRndGenerator;
    //structure de clusters solution
    StrucClusters: TStrucVarCluster;
    //constructeur
    constructor create(prmInputs: TLstAttributes; param: TOperatorParameter); overload; virtual;
    //autre constructeur
    constructor create(otherCalc: TCalcVarClus; param: TOperatorParameter); overload; virtual; 
    //destructeur
    destructor  destroy(); override;
    //lancer les calculs
    procedure   execute(examples: TExamples);
    //description des rsultats
    function getHTMLResult(): string;
    //calculer les paliers et le nombre optimal de clusters, renvoie aussi le seuil permettant de dire  partir de quelle valeur le moving average des angles est intrssant
    function getOptimalNbClusters(const TabVarExplained: TTabVarExplained; var TabPalier: TTabPalier; var seuilMV: double): integer;
    //afficher une sortie HTML de l'exploration des solutions
    function getHTMLVarExplained(const TabVarExplained: TTabVarExplained; const TabPalier: TTabPalier; nbCluster: integer; seuilMV: double): string; 
    end;    

implementation

uses
    Math, Sysutils,
    UCalcMatrixAdditionalFunctions, UCalcMatrixToAttributes,
    UConstConfiguration, UCompVarClusBase;


{ TAttVarClus }

constructor TAttVarClus.create(v: TAttribute; num: integer; stat: TCalcStatDesContinuous; examples: TExamples = nil);
begin
 inherited create();
 Att:= v;
 Numero:= num;
 StatDes:= stat;
 if (examples <> nil) then StatDes.RefreshStat(examples);
end;

destructor TAttVarClus.destroy;
begin
 //
 inherited;
end;

function TAttVarClus.getCorrelation(other: TAttribute; examples: TExamples): double;
var i,example: integer;
    num,denom,mY,mX,sY,s2Y,sX,s2X,sYX,n: double;
    x,y: double;
begin
 n:= 1.0 * examples.Size;
 sY:= 0.0;
 sX:= 0.0;
 s2Y:= 0.0;
 s2X:= 0.0;
 sYX:= 0.0;
 //additionner
 for i:= 1 to examples.Size do
  begin
   example:= examples.Number[i];
   y:= att.cValue[example];
   x:= other.cValue[example];
   //y
   sY:= sY + y;
   s2Y:= s2Y + y * y;
   //x
   sX:= sX + x;
   s2X:= s2X + x * x;
   //crois
   sYX:= sYX + y * x;
  end;
 //calculer
 mY:= sY/n;
 mX:= sX/n;
 num:= sYX - n * mY * mX;
 s2Y:= s2Y - n * mY * mY;
 s2X:= s2X - n * mX * mX;
 denom:= s2Y * s2X;
 if (denom>0)
  then
   begin
    denom:= SQRT(denom);
    result:= num / denom;
   end
  else result:= 0.0;
end;

function TAttVarClus.getCorrelation(other: TAttVarClus;
  calc: TCalcVarClus): double;
var i,j: integer;
begin
 //rcuprer les deux adresses
 i:= self.Numero;
 j:= other.Numero;
 result:= calc.MatCorr^[i]^[j];
end;

procedure TAttVarClus.refreshStat(examples: TExamples);
begin
 StatDes.RefreshStat(examples);
end;

{ TLstAttVarClus }

procedure TLstAttVarClus.add(v: TAttVarClus);
begin
 Lst.Add(v);
end;

procedure TLstAttVarClus.assign(source: TLstAttVarClus);
begin
 self.Lst.Assign(source.Lst);
end;

procedure TLstAttVarClus.clear;
begin
 Lst.Clear();
end;

constructor TLstAttVarClus.create(owner: boolean);
begin
 inherited create();
 Lst:= TObjectList.Create(owner);
end;

destructor TLstAttVarClus.destroy;
begin
 Lst.Free();
 inherited;
end;

function TLstAttVarClus.getCount: integer;
begin
 result:= Lst.Count;
end;

function TLstAttVarClus.getVarClus(index: integer): TAttVarClus;
begin
 result:= Lst[index] as TAttVarClus;
end;

function TLstAttVarClus.isOwner: boolean;
begin
 result:= Lst.OwnsObjects;
end;

{ TVarCluster }

procedure TVarCluster.addVar(attVar: TAttVarClus);
begin
 self.Lst.add(attVar);
end;

procedure TVarCluster.clear;
begin
 self.Lst.clear();
end;

class function TVarCluster.correlation(c1, c2: TVarCluster;
  examples: TExamples): double;
var i,example: integer;
    num,denom,mY,mX,sY,s2Y,sX,s2X,sYX,n: double;
    x,y: double;
begin
 n:= 1.0 * examples.Size;
 sY:= 0.0;
 sX:= 0.0;
 s2Y:= 0.0;
 s2X:= 0.0;
 sYX:= 0.0;
 //additionner
 for i:= 1 to examples.Size do
  begin
   example:= examples.Number[i];
   y:= c1.VarClus.cValue[example];
   x:= c2.VarClus.cValue[example];
   //y
   sY:= sY + y;
   s2Y:= s2Y + y * y;
   //x
   sX:= sX + x;
   s2X:= s2X + x * x;
   //crois
   sYX:= sYX + y * x;
  end;
 //calculer
 mY:= sY/n;
 mX:= sX/n;
 num:= sYX - n * mY * mX;
 s2Y:= s2Y - n * mY * mY;
 s2X:= s2X - n * mX * mX;
 denom:= s2Y * s2X;
 if (denom>0)
  then
   begin
    denom:= SQRT(denom);
    result:= num / denom;
   end
  else result:= 0.0;
end;

constructor TVarCluster.create(ExampleSize: integer; calc: TCalcVarClus);
begin
 inherited create();
 ClusCalc:= calc;
 EigVecSize:= 0;
 Lst:= TLstAttVarClus.create(false);//no propertary
 VarClus:= TAttContinue.Create('clus',ExampleSize);//variable de classe
end;

destructor TVarCluster.destroy;
begin
 Finalize(TabAllCorrelation);
 Finalize(TabOwnCorrelation);
 Finalize(TabClosestRSquared);
 VarClus.Free;
 Lst.Free;//add by GongYu avoid hugh memory leak
 if assigned(EigVector) then delVector(EigVector,EigVecSize);
 inherited;
end;

function TVarCluster.duplicate(calc: TCalcVarClus): TVarCluster;
var output: TVarCluster;
    i: integer;
begin
 output:= TVarCluster.create(self.VarClus.Size,calc);
 output.Lst.assign(self.Lst);
 for i:= 1 to output.VarClus.Size do output.VarClus.cValue[i]:= self.VarClus.cValue[i];
 if assigned(self.TabAllCorrelation) then output.TabAllCorrelation:= copy(self.TabAllCorrelation,0,length(self.TabAllCorrelation));
 if assigned(self.TabOwnCorrelation) then output.TabOwnCorrelation:= copy(self.TabOwnCorrelation,0,length(self.TabOwnCorrelation));
 if assigned(self.TabClosestRSquared) then output.TabClosestRSquared:= copy(self.TabClosestRSquared,0,length(self.TabClosestRSquared));
 output.EigVal:= self.EigVal;
 output.EigVecSize:= self.EigVecSize;
 if (EigVecSize > 0)
  then
   begin
    dimVector(output.EigVector,output.EigVecSize);
    copyVector(output.EigVector,self.EigVector,1,output.EigVecSize);
   end;
 result:= output;
end;

class function TVarCluster.fusion(c1, c2: TVarCluster; examples: TExamples; var dist: double): TVarCluster;
var output: TVarCluster;
    j: integer;
    v: TAttVarClus;
begin
 output:= TVarCluster.create(c1.VarClus.Size,c1.ClusCalc);
 //rajouter les listes de variables
 output.Lst.assign(c1.Lst);
 for j:= 0 to pred(c2.Lst.getCount()) do
  begin
   v:= c2.Lst.getVarClus(j);
   output.Lst.add(v);
  end;
 //recalculer le "barycentre"
 output.refreshClusVar(examples);
 //renvoyer la distance d'agrgation, p.231 livre Nakache
 dist:= c1.EigVal + c2.EigVal - output.EigVal;
 //adn then...
 result:= output;
end;

function TVarCluster.getCorrMatrix: PMatrix;
var MCorr: PMatrix;
    j,k: integer;
    a,b: integer;
    c: single;
begin
 //calculer la matrice de corrlation locale
 dimMatrix(MCorr,EigVecSize,EigVecSize);
 for j:= 1 to EigVecSize do
  begin
   MCorr^[j]^[j]:= 1.0;
   a:= self.Lst.getVarClus(pred(j)).Numero;
   for k:= succ(j) to EigVecSize do
    begin
     b:= self.Lst.getVarClus(pred(k)).Numero;
     c:= ClusCalc.MatCorr^[a]^[b];
     MCorr^[j]^[k]:= c;
     MCorr^[k]^[j]:= c;
    end;
  end;
 //
 result:= MCorr;
end;

function TVarCluster.getCount: integer;
begin
 result:= self.Lst.getCount();
end;

function TVarCluster.getNearestClusVar(examples: TExamples): TAttribute;
var d,dMax: double;
    attVar: TAttVarClus;
    j,jMax: integer;
begin
 if (self.Lst.getCount() = 0)
  then result:= nil
  else
   begin
    dMax:= -1e308;
    jMax:= -1;
    for j:= 0 to pred(self.Lst.getCount()) do
     begin
      attVar:= self.Lst.getVarClus(j);
      d:= MATH.POWER(attVar.getCorrelation(self.VarClus,examples),2);
      if (d > dMax)
       then
        begin
         jMax:= j;
         dMax:= d; 
        end;
     end;
    result:= self.Lst.getVarClus(jMax).Att;
   end;
end;

function TVarCluster.getPropotionVarExplained: double;
begin
 if (self.getCount() = 0)
  then result:= 0.0
  else result:= self.EigVal / (1.0 * self.getCount());
end;

function TVarCluster.getRepresentant(prmUseClusVar: boolean;
  examples: TExamples): TAttribute;
begin
 if (prmuseClusVar = true) or (self.getCount() = 0)
  then result:= VarClus
  else result:= self.getNearestClusVar(examples);
end;

procedure TVarCluster.majCorrelation(examples: TExamples);
var j: integer;
begin
 //calculer toutes les corrlations avec la variable moyenne
 setLength(TabAllCorrelation,ClusCalc.Inputs.Count);
 for j:= 0 to pred(ClusCalc.Inputs.Count) do
  TabAllCorrelation[j]:= ClusCalc.LstVariables.getVarClus(j).getCorrelation(self.VarClus,examples);
 //calculer les corrlations Own
 setLength(TabOwnCorrelation,EigVecSize);
 for j:= 1 to EigVecSize do
  TabOwnCorrelation[pred(j)]:= TabAllCorrelation[pred(self.Lst.getVarClus(pred(j)).Numero)];
end;

procedure TVarCluster.refreshClusVar;
var i: integer;
    MCorr: PMatrix;
begin
 if (EigVEcSize > 0) and assigned(EigVector) then delVector(EigVector,EigVecSize);
 //nouvelle taille
 EigVecSize:= self.Lst.getCount();
 if (EigVecSize > 0)
  then
   begin
    MCorr:= self.getCorrMatrix();
    //rcuprer la valeur et le vecteur propre
    EigVal:= getFirstEigSymetricMatrix(MCorr,EigVecSize,EigVector);
    //dtruire la matrice
    delMatrix(MCorr,EigVecSize,EigVecSize);
    //projection  partir du nouveau vecteur propre... c'est du propre...
    self.setProjectionFromEigVector();
   end
  else
   begin
    EigVal:= 0.0;
    //rz
    for i:= 1 to self.VarClus.Size do self.VarClus.cValue[i]:= 0.0;
   end;
end;

procedure TVarCluster.setProjectionFromEigVector;
var sum, value: double;
    i,j: integer;
    attV: TAttVarClus;
begin
 //effectuer la projection pour chaque individu i
 for i:= 1 to self.VarClus.Size do
  begin
   sum:= 0.0;
   for j:= 1 to EigVecSize do
    begin
     attV:= self.Lst.getVarClus(pred(j));
     value:= (attV.Att.cValue[i] - attV.StatDes.Average) / attV.StatDes.StdDevPop;
     sum:= sum + EigVector^[j] * value;
    end;
   self.VarClus.cValue[i]:= sum;
  end;
end;

{ TCalcVarClus }

procedure TCalcVarClus.commonInitialisations;
begin
 RndGenerator:= TRndGenerator.Create(seedStandard,DEFAULT_SEED_VALUE_1,DEFAULT_SEED_VALUE_2);
 StrucClusters:= TStrucVarCluster.create(self);
end;

procedure TCalcVarClus.computeBasicStatsAndCorr(examples: TExamples);
var attVar: TAttVarClus;
    att: TAttribute;
    stat: TCalcStatDesContinuous;
    j: integer;
begin
 //matrice des corrlations
 MatCorr:= BuildMatVCV(examples,inputs,vcvNormNormalized,LstStatDes);
 //crer la liste des variables
 if OwnerLstVariables
  then
   begin
    //liste des stats
    LstStatDes:= TLstCalcStatDesContinuous.Create(inputs,examples);
    //liste des variables de clustering
    LstVariables:= TLstAttVarClus.create(true);
    for j:= 0 to pred(Inputs.Count) do
     begin
      att:= Inputs.Attribute[j];
      stat:= LstStatDes.Stat(j) as TCalcStatDesContinuous;
      attVar:= TAttVarClus.create(att,succ(j),stat);
      LstVariables.add(attVar);
     end;
   end;
end;

constructor TCalcVarClus.create(prmInputs: TLstAttributes; param: TOperatorParameter);
begin
 inherited create();
 Inputs:= prmInputs;
 OpParameter:= param;
 OwnerLstVariables:= true;
 self.commonInitialisations();
end;


constructor TCalcVarClus.create(otherCalc: TCalcVarClus;
  param: TOperatorParameter);
begin
 inherited create();
 OwnerLstVariables:= false;
 OpParameter:= param;
 Inputs:= OtherCalc.Inputs;
 LstVariables:= otherCalc.LstVariables;
 LstStatDes:= otherCalc.LstStatDes;
 self.commonInitialisations();
end;

destructor TCalcVarClus.destroy;
begin
 RndGenerator.Free();
 DelMatrix(MatCorr,Inputs.Count,Inputs.Count);
 if OwnerLstVariables
  then
   begin
    LstVariables.Free();
    LstStatDes.Free();
   end;
 if assigned(StrucClusters) then StrucClusters.Free();
 inherited;
end;

procedure TCalcVarClus.execute(examples: TExamples);
begin
 //prparer les stats.
 self.computeBasicStatsAndCorr(examples);
 //puis demander le calcul des clusters
 self.computeClusters(examples);
end;

function TCalcVarClus.getHTMLDetailedResults: string;
begin
 result:= '';
end;

function TCalcVarClus.getHTMLResult: string;
var s: string;
begin
 s:= '<H3>Cluster characterization</H3>';
 s:= s + StrucClusters.getHTMLClusterSummary();
 if (OpParameter as TOpPrmVarClus).ShowRSquare then s:= s + StrucClusters.getHTMLClusterTab();
 if (OpParameter as TOpPrmVarClus).ShowCorrelation then s:= s + StrucClusters.getHTMLCorrelations();
 if (OpParameter as TOpPrmVarClus).ShowRegCoef then s:= s + StrucClusters.getHTMLRegressionCoefficients();
 //les rsultats dtaills
 s:= s + '<H3>Detailed Results</H3>';
 s:= s +  self.getHTMLDetailedResults();
 result:= s;
end;

function detectSeuilPalier(const tab: TTabPalier; col: integer; n: integer): double;
LABEL 99;
VAR
   l,j,ir,i: integer;
   rra: single;
   t: array of single;
BEGIN
   //copier les donnes
   setLength(t,succ(n));
   for i:= 1 to n do t[i]:= tab[i-1,col];
   
   //trier maintenant -- HeapSort
   l := (n DIV 2)+1;
   ir := n;
   WHILE true DO BEGIN
      IF (l > 1) THEN BEGIN
         l := l-1;
         rra := t[l]
      END ELSE BEGIN
         rra := t[ir];
         t[ir] := t[1];
         ir := ir-1;
         IF (ir = 1) THEN BEGIN
            t[1] := rra;
            GOTO 99
         END
      END;
      i := l;
      j := l+l;
      WHILE (j <= ir) DO BEGIN
         IF (j < ir) THEN
            IF (t[j] < t[j+1]) THEN j := j+1;
         IF (rra < t[j]) THEN BEGIN
            t[i] := t[j];
            i := j;
            j := j+j
         END ELSE
            j := ir+1
      END;
      t[i] := rra
   END;
99:

  //suite rcuprer la 3me plus grande valeur
  result:= t[n-2];
  Finalize(t); 
END;

function TCalcVarClus.getHTMLVarExplained(
  const TabVarExplained: TTabVarExplained; const TabPalier: TTabPalier;
  nbCluster: integer; seuilMV: double): string;
var s,sTRColor: string;
    j: integer;
begin
 s:= '<H4>Variation Explained and Determination of the Number of Clusters</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH># clusters</TH><TH width=60>Var.Expl.</TH><TH width=60>Dif.</TH><TH width=60>Cos</TH><TH width=60>Angle</TH><TH width=60>Moving Avg.</TH></TR>';
 for j:= 0 to pred(length(TabVarExplained)) do
  begin
   //solution intressante ?
   if (TabPalier[j,3] >= seuilMV)
    then sTRColor:= HTML_TABLE_COLOR_HEADER_GRAY
    else sTRColor:= HTML_TABLE_COLOR_DATA_GRAY;
   //la solution slectionne ?
   if (j = pred(nbCluster)) then sTRColor:= HTML_TABLE_COLOR_DATA_GREEN;
   //afficher
   s:= s + sTRColor +
      format('<TD>%d</TD><TD align=right>%.4f</TD><TD align=right>%.4f</TD><TD align=right>%.4f</TD><TD align=right>%.4f</TD><TD align=right>%.4f</TD></TR>',
             [succ(j),TabVarExplained[j],TabPalier[j,0],TabPalier[j,1],TabPalier[j,2],TabPalier[j,3]]);
  end;
 s:= s + '</table>';
 result:= s;
end;

function TCalcVarClus.getOptimalNbClusters(
  const TabVarExplained: TTabVarExplained; var TabPalier: TTabPalier;
  var seuilMV: double): integer;
var k: integer;
    maxMoy,moy: double;
    idNbCluster: integer;
begin
 //initaliser  la mme taille que la variance explique
 //0 -> dif. premires, 1 -> cos, 2 -> arccos, 3 -> moving average des angles
 setLength(TabPalier,length(TabVarExplained),4);
 for k:= 1 to pred(length(TabVarExplained)) do
  TabPalier[k,0]:= TabVarExplained[k]-TabVarExplained[pred(k)];
 //calcul du cosinus et donc de l'angle associ
 for k:= 1 to length(TabVarExplained)-2 do
  begin
   TabPalier[k,1]:= (1.0 + TabPalier[k,0]*TabPalier[k+1,0])/(SQRT(1.0+SQR(TabPalier[k,0]))*SQRT(1.0+SQR(TabPalier[k+1,0])));
   TabPalier[k,2]:= MATH.ArcCos(TabPalier[k,1]);
  end;
 //dtecter la solution, ide: trouver une zone  fort angle de courbure
 //solution : on utilise une moyenne mobile sur 3 points
 idNbCluster:= 1;
 maxMoy:= -1.0*MATH.MaxDouble;
 for k:= 1 to length(TabVarExplained)-2 do
  begin
   moy:= MATH.Mean([TabPalier[k-1,2],TabPalier[k,2],TabPalier[k+1,2]]);
   TabPalier[k,3]:= moy;//rcuprer au passage la valeur
   if (moy > maxMoy)
    then
     begin
      maxMoy:= moy;
      idNbCluster:= k+1;
     end;
  end;
 //la valeur seuil ? si trop peu de valeurs, ne pas s'embter
 if (length(TabVarExplained) > 5)
  then seuilMV:= detectSeuilPalier(TabPalier,3,length(TabVarExplained))
  else seuilMV:= maxMoy;
 //and then
 result:= idNbCluster;
end;

{ TStrucVarCluster }

procedure TStrucVarCluster.add(cluster: TVarCluster);
begin
 LstClus.Add(cluster);
end;

procedure TStrucVarCluster.addVarToNearest(attVar: TAttVarClus; examples: TExamples);
var j,jMax: integer;
    d,dMax: double;
    clus: TVarCluster;
begin
 //chercher le cluster le plus proche
 jMax:= -1;
 dMax:= -1e308;
 for j:= 0 to pred(self.getCount()) do
  begin
   clus:= self.get(j);
   d:= MATH.POWER(attVar.getCorrelation(clus.VarClus,examples),2.0);
   if (d > dMax)
    then
     begin
      dMax:= d;
      jMax:= j;
     end;
  end;
 //and then
 clus:= self.get(jMax);
 clus.addVar(attVar);
end;

procedure TStrucVarCluster.clearClusters;
var j: integer;
begin
 for j:= 0 to pred(self.getCount()) do self.get(j).clear();
end;

constructor TStrucVarCluster.create(calc: TCalcVarClus);
begin
 inherited create();
 LstClus:= TObjectList.Create(true);
 ClusCalc:= calc;
end;

destructor TStrucVarCluster.destroy;
begin
 LstClus.Free();
 inherited;
end;

function TStrucVarCluster.get(index: integer): TVarCluster;
begin
 result:= LstClus[index] as TVarCluster;
end;

function TStrucVarCluster.getCount: integer;
begin
 result:= LstClus.Count;
end;

function TStrucVarCluster.getHTMLClusterSummary: string;
var s: string;
    j: integer;
    clus: TVarCluster;
    sum: double;
begin
 sum:= 0.0;
 s:= '<H4>Cluster summary</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Cluster</TH><TH># Members</TH><TH width=60>Variation<br>Explained</TH><TH width=60>Proportion<br>Explained</TH></TR>';
 for j:= 0 to pred(self.getCount()) do
  begin
   clus:= self.get(j);
   sum:= sum + clus.EigVal;
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD align=center>%d</TD><TD align=center>%d</TD><TD align=right>%.4f</TD>',[succ(j),clus.getCount(),clus.EigVal]);
   if (clus.getCount() > 0)
    then s:= s + format('<TD align=right>%.4f</TD>',[clus.EigVal/(1.0 * clus.getCount())])
    else s:= s + '<TD align=center>-</TD>';
   s:= s + '</TR>';
  end;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + '<TD align=center colspan=2>Total</TD>';
 s:= s + format('<TD align=right>%.4f</TD><TD align=right>%.4f</TD></TR>',[sum,sum/(1.0*self.ClusCalc.Inputs.Count)]);
 s:= s + '</table>';
 result:= s;
end;

const
    TAB_HTML_ROW_COLOR : array[0..1] of string = (HTML_TABLE_COLOR_DATA_GRAY,HTML_TABLE_COLOR_HEADER_GRAY);
    TAB_TD_HTML_RATIO_COLOR: array[0..1] of string = (HTML_BGCOLOR_DATA_GRAY,HTML_BGCOLOR_HEADER_GRAY);
    THRESHOLD_GOOD_RATIO : double = 1.0;

function TStrucVarCluster.getHTMLClusterTab: string;
var s: string;
    j,k,nbLig: integer;
    clus: TVarCluster;
    ratio: double;
    TDHighlight: string;
begin
 s:= '<H4>Cluster members and R-square values</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Cluster</TH><TH>Members</TH><TH width=60>Own<br>Cluster</TH><TH width=60>Next<br>Closest</TH><TH width=60>1-R ratio</TH></TR>';
 for j:= 0 to pred(self.getCount()) do
  begin
   clus:= self.get(j);
   s:= s + TAB_HTML_ROW_COLOR[j mod 2];
   nbLig:= MATH.Max(1,clus.getCount()); 
   s:= s + format('<TD align=center valign=top rowspan=%d>%d</TD>',[nbLig,succ(j)]);
   if (clus.getCount() = 0)
    then s:= s + '<TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD></TR>'
    else
     begin
      for k:= 0 to pred(clus.getCount()) do
       begin
        if (k > 0) then s:= s + TAB_HTML_ROW_COLOR[j mod 2];
        //ratio
        ratio:= (1.0-SQR(clus.TabOwnCorrelation[k]))/(1.0-clus.TabClosestRSquared[k]);
        if (ratio >= THRESHOLD_GOOD_RATIO)
         then TDHighlight:= HTML_BGCOLOR_DATA_RED
         else TDHighlight:= TAB_TD_HTML_RATIO_COLOR[j mod 2];
        //
        s:= s + format('<TD align=left>%s</TD><TD align=right>%.4f</TD><TD align=right>%.4f</TD><TD align=right %s>%.4f</TD></TR>',
                       [clus.Lst.getVarClus(k).Att.Name,SQR(clus.TabOwnCorrelation[k]),clus.TabClosestRSquared[k],TDHighlight,ratio]);
       end;
     end;
  end;
 s:= s + '</table>';
 result:= s;
end;

function TStrucVarCluster.computeVarExplained: double;
var j: integer;
    sum: double;
begin
 sum:= 0.0;
 for j:= 0 to pred(self.getCount()) do
  sum:= sum + self.get(j).EigVal;
 result:= sum;
end;

function getRSquared(att1, att2: TAttribute; examples: TExamples): double;
var i,example: integer;
    num,denom,mY,mX,sY,s2Y,sX,s2X,sYX,n: double;
    x,y: double;
begin
 n:= 1.0 * examples.Size;
 sY:= 0.0;
 sX:= 0.0;
 s2Y:= 0.0;
 s2X:= 0.0;
 sYX:= 0.0;
 //additionner
 for i:= 1 to examples.Size do
  begin
   example:= examples.Number[i];
   y:= att1.cValue[example];
   x:= att2.cValue[example];
   //y
   sY:= sY + y;
   s2Y:= s2Y + y * y;
   //x
   sX:= sX + x;
   s2X:= s2X + x * x;
   //crois
   sYX:= sYX + y * x;
  end;
 //calculer
 mY:= sY/n;
 mX:= sX/n;
 num:= sYX - n * mY * mX;
 s2Y:= s2Y - n * mY * mY;
 s2X:= s2X - n * mX * mX;
 denom:= s2Y * s2X;
 if (denom>0)
  then
   begin
    denom:= SQRT(denom);
    result:= num / denom;
   end
  else result:= 0.0;
 result:= MATH.POWER(result,2.0);
end;

procedure TStrucVarCluster.majRSquaredClosest(examples: TExamples);
var j,k,kMax: integer;
    d,dMax: double;
    clusRef,clusOther: TVarCluster;
begin
 //trouver le voisin le plus proche de j
 for j:= 0 to pred(self.getCount()) do
  begin
   clusRef:= self.get(j);
   dMax:= -1e308;
   kMax:= -1;
   for k:= 0 to pred(self.getCount()) do
    begin
     if (k <> j)
      then
       begin
        d:= getRSquared(clusRef.VarClus,self.get(k).VarClus,examples);
        if (d > dMax)
         then
          begin
           dMax:= d;
           kMax:= k;
          end;
       end;
    end;
   if (kMax >= 0)
    then
     begin
      clusOther:= self.get(kMax);
      //mj des corrlations
      setLength(clusRef.TabClosestRSquared,clusRef.getCount());
      for k:= 0 to pred(clusRef.getCount()) do
       clusRef.TabClosestRSquared[k]:= MATH.Power(clusRef.Lst.getVarClus(k).getCorrelation(clusOther.VarClus,examples),2.0);
     end
    else setLength(clusRef.TabClosestRSquared,clusRef.getCount());//avec des valeurs  zro tout simplement...
  end;
end;

procedure TStrucVarCluster.refreshClusVar(examples: TExamples);
var j: integer;
begin
 for j:= 0 to pred(self.getCount()) do
  self.get(j).refreshClusVar(examples);
 //calculer la variance explique
 VarExplained:= self.computeVarExplained();
end;

procedure TStrucVarCluster.majCorrelation(examples: TExamples);
var j: integer;
begin
 //correlations internes
 for j:= 0 to pred(self.getCount()) do
  self.get(j).majCorrelation(examples);
 //correlations avec les plus proches
 self.majRSquaredClosest(examples);
end;

const
    BGCOLOR_SAUMON : string = 'BGCOLOR="#FCC998"';
    BGCOLOR_YELLOW : string = 'BGCOLOR="#FBFD75"';

function TStrucVarCluster.getHTMLCorrelations: string;
var s: string;
    k,j: integer;
    sColor: string;
    seuil: double;
    correlation: single;
    memberShip: integer;
begin
 seuil:= (self.ClusCalc.OpParameter as TOpPrmVarClus).CorrThreshold;
 s:= '<H4>Cluster correlations -- Structure</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 //
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Attribute</TH><TH># membership</TH>';
 for k:= 0 to pred(self.getCount()) do
  s:= s + format('<TH width=60>Cluster %d</TH>',[succ(k)]);
 s:= s + '</TR>';
 //
 for j:= 0 to pred(ClusCalc.LstVariables.getCount()) do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align=left>%s</TD>',[ClusCalc.LstVariables.getVarClus(j).Att.Name]);
   //calcul de l'appartenance  partir des valeurs
   memberShip:= 0;
   for k:= 0 to pred(self.getCount()) do
    begin
     correlation:= self.get(k).TabAllCorrelation[j];
     if (correlation >= seuil) then inc(memberShip);
     if (correlation <= -1.0*seuil) then inc(membership);
    end;
   s:= s + format('<TD align=center>%d</TD>',[memberShip]);
   //affichage des valeurs, on ne peut pas le faire en une passe malheuresement...
   for k:= 0 to pred(self.getCount()) do
    begin
     correlation:= self.get(k).TabAllCorrelation[j];
     sColor:= HTML_BGCOLOR_DATA_GRAY;
     if (correlation >= seuil) then sColor:= BGCOLOR_SAUMON;
     if (correlation <= -1.0*seuil) then sColor:= BGCOLOR_YELLOW;
    s:= s + format('<TD align=right %s>%.4f</TD>',[sColor,correlation]);
    end;
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TStrucVarCluster.getHTMLRegressionCoefficients: string;
var s: string;
    k,j: integer;
    tab: array of array of single;
    clus: TVarCluster; 
begin
 //recopier le tableau dans une structure plus avenante
 setLength(tab,ClusCalc.LstVariables.getCount(),self.getCount());
 for k:= 0 to pred(self.getCount()) do
  begin
   clus:= self.get(k);
   for j:= 0 to pred(clus.getCount()) do
    tab[clus.Lst.getVarClus(j).Numero-1,k]:= clus.EigVector^[j+1];
  end;
 //
 s:= '<H4>Standardized Regression coefficients</H4>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Attribute</TH><TH width=60>Avg.</TH><TH width=60>StdDev</TH>';
 for k:= 0 to pred(self.getCount()) do
  s:= s + format('<TH width=60>Cluster %d</TH>',[succ(k)]);
 s:= s + '</TR>';
 //
 for j:= 0 to pred(ClusCalc.LstVariables.getCount()) do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align=left>%s</TD>',[ClusCalc.LstVariables.getVarClus(j).Att.Name]);
   s:= s + format('<TD align=right %s>%.4f</TD><TD align=right %s>%.4f</TD>',[HTML_BGCOLOR_HEADER_GRAY,ClusCalc.LstVariables.getVarClus(j).StatDes.Average,HTML_BGCOLOR_HEADER_GRAY,ClusCalc.LstVariables.getVarClus(j).StatDes.StdDevPop]);
   for k:= 0 to pred(self.getCount()) do
    s:= s + format('<TD align=right>%.4f</TD>',[tab[j,k]]);
   s:= s + '</TR>';
  end;
 //
 s:= s + '</table>';
 result:= s;
 //
 finalize(tab);
end;

end.
