(******************************************************************)
(* UCompCatVarClusHAC.pas - Copyright (c) 2013 Ricco RAKOTOMALALA *)
(******************************************************************)

{
@abstract(CAH sur les modalits de variables actives qualitatives.)
@author(Ricco)
@created(10/12/2013)

Mthode base sur : H. Abdallah, G. Saporta, "Mesures de distance entre modalits
de variables qualitatives ; application  la classification", in Revue de Statistique
Applique, Tome 51, n2, pp. 75-90, 2003.
--> http://cedric.cnam.fr/fichiers/RC470.pdf

Possibilit de prise en compte de variables supplmentaires qualitatives.
Pour cette raison, seules les mthodes single-linkage, complete-linkage et average-linkage sont proposes.
La mthode de Lance & Williams est utilise pour le calcul de proche en proche des distances.

Utilisation des diffrences secondes pour obtenir le point d'inflexion max. dans la courbe d'volution de la hauteur.

//new - 21/12/2013 -- intgration de la mthode de Ward

}

unit UCompCatVarClusHAC;

interface

USES
        Contnrs, Forms, Classes, IniFiles,
        UCompDefinition,
        UDatasetDefinition, UDatasetExamples,
        UDatasetImplementation,
        UOperatorDefinition,
        UCompManageDataset;
TYPE

       {type d'agrgation}
       TEnumCatVarClusHAC = (cvhSingle, cvhComplete, cvhAverage, cvhWard);

TYPE

       {gnrateur de composant}
        TGenCatVarClusHAC = class(TMLGenComp)
        protected
        procedure   GenCompInitializations(); override;
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant CAH}
        TMLCompCatVarClusHAC = class(TMLCompLocalData)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        TCalcHAC = class;

        {oprateur CAH}
        TOpCatVarClusHCA = class(TOpLocalData)
        protected
        FInputs: TLstAttributes;
        FIllus: TLstAttributes;
        FCalc: TCalcHAC;
        function    getClassParameter: TClassOperatorParameter; override;
        function    CheckAttributes(): boolean; override;
        public
        function    CoreExecute(): boolean; override;
        function    getNewForm(): TForm; override;
        function    getHTMLResultsSummary(): string; override;
        property    Calc: TCalcHAC read FCalc;
        end;

        {paramtrage}
        TPrmOpCatVarClusHCA = class(TOperatorParameter)
        protected
        function    CreateDlgParameters(): TForm; override;
        procedure   SetDefaultParameters(); override;
        public
        //0 : single, 1 : complete, 2 : average
        LinkageType: integer;
        {Mode de dtermination de la coupure optimale : 0->automatique, 1->manuel}
        ModeBestSelection: integer;
        {si choix manuel, nombre de clusters}
        NbClusters: integer;
        {normalisation pour le calcul des distances}
        Normalization: integer;
        {montrer la structure hirarchique}
        ShowTreeStructure: boolean;
        {montrer les carts dans les indices de niveau}
        ShowIndexSelection: boolean;
        {max de niveau  montrer}
        MaxIndexLevel: integer;
        //I-O
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        //HTML
        function    getHTMLParameters(): string; override;
        end;

        {un couple variable = modalit}
        TDummy = class
        public
        //variable
        FVar: TAttribute;
        //modalit
        FValue: TTypeDiscrete;
        //indice de dice avec le voisin interne le plus loign
        FMaxIndexOwn: double;
        //indice de dice avec le voision externe le plus proche
        FMinIndexNext: double;
        //constructeur
        constructor create(aVar: TAttribute; aValue: TTypeDiscrete);
        //destructeur
        destructor  destroy(); override;
        //lire une valeur en codage disjonctif 0/1
        function    getValue(example: integer): double;
        //indice de Dice - Abdallah & Saporta, page 79
        function    diceIndex(other: TDummy; prmExamples: TExamples; linkageType: TEnumCatVarClusHAC): double;
        //description
        function    description(): string;
        end;

        TListDummies = class
        protected
        //liste interne
        FLst: TObjectList;
        public
        //constructor - propritaire ou non
        constructor create(aOwner: boolean); overload;
        //construction  partir d'une liste de variables
        constructor create(aOwner: boolean; aLstVar: TLstAttributes); overload;
        //destructeur
        destructor  destroy(); override;
        //ajouter un dummy
        procedure   add(aDummy: TDummy);
        //ajouter les dummies d'une variable
        procedure   addDummies(aVar: TAttribute);
        //ajouter une liste de dummies
        procedure   assign(source: TListDummies);
        //indice de dice de celui qui est le plus proche
        function    singleLinkageDice(other: TDummy; prmExamples: TExamples): double;
        //indice de dice de celui qui est le plus loign
        function    completeLinkageDice(other: TDummy; prmExamples: TExamples): double;
        //moyenne des indices de dice
        function    averageLinkageDice(other: TDummy; prmExamples: TExamples): double;
        //mthode de Ward - distance carr au centre de classe pondre
        function    wardLinkageDice(other: TDummy; prmExamples: TExamples): double;
        //indice selon le paramtre de linkage
        function     LinkageDice(linkage: TEnumCatVarClusHAC; other: TDummy; prmExamples: TExamples): double; 
        //nombre d'lments
        function    getCount(): integer;
        //lment numro i
        function    get(i: integer): TDummy;
        //description
        function    destription(): string;
        //trier selon la description
        procedure   sortAccordingDescription();
        //obtenir la longueur maximale de description
        function    getMaxLengthDescription(): integer;
        end;

        TDistVec = array of double;

        //matrice de distance - forcment symtrique
        //diagonale inf.
        TDistMatrix = class
        Size: integer;
        Tab: array of TDistVec;
        constructor create(aSize: integer);
        destructor  destroy(); override;
        function    getValue(i,j: integer): double;
        procedure   setValue(i,j: integer ; value: double);
        procedure   DecrementSize();
        end;

        {Un sommet de CAH}
        TNodeHAC = class(TObject)
        private
        {oprateur de calcul}
        FCalc: TCalcHAC;
        {pre}
        FDad: TNodeHAC;
        {enfants gauche et droite}
        FChildL,FChildR: TNodeHAC;
        {numro de sommet pour suivre les fusions}
        FNumNode: integer;
        {Distance d'agrgation}
        FDist: double;
        {coordonne sur l'axe des X, pour le dessin du dendrogramme}
        FXPos: double;
        {indique si le noeud est une solution}
        FIsCluster: boolean;
        //dummies for the node
        FDummies: TListDummies;
        public
        {construire un noeud}
        constructor  Create(prmCalc: TCalcHAC; aDummy: TDummy = nil);
        {dtruire}
        destructor   Destroy(); override;
        {calculer les distances d'agrgation avec un autre noeud - valable seulement initialement}
        function     MergeDistance(other: TNodeHAC; prmExamples:  TExamples; linkageType: TEnumCatVarClusHAC): double;
        {tester si le noeud est une feuille}
        function     isLeaf(): boolean;
        {rcuprer le label qui est l'effectif du sommet en ralit}
        {new -- 12/10/2006 -- et l'tiquette du cluster de base si c'est une feuille}
        function     getLabel(): string;
        {nombre de modalits prsentes}
        function     getNbInstances(): integer;
        //proprits
        property     XPos: double read FXPos;
        property     YPos: double read FDist;
        property     ChildL: TNodeHAC read FChildL;
        property     ChildR: TNodeHAC read FChildR;
        property     isCluster: boolean read FIsCluster;
        end;

        {Structure complte de CAH}
        TSetNodeHAC = class(TObject)
        private
        {Calculateur HAC}
        FCalcHAC: TCalcHAC;
        {racine de l'arbre}
        FRootNode: TNodeHAC;
        {tous les noeuds - liste propritaire}
        FLstAllNodes: TObjectList;
        {les noeuds  fusionner - les noeuds orphelins - liste non propritaire}
        FLstMergeNodes: TObjectList;
        {les feuilles de l'arbre - liste non propritaire}
        FLstLeaves: TObjectList;
        {Indicateur global du dernier numro de noeud ajout}
        FLastNumNode: integer;
        {Tableau de suivi des fusions ascendantes}
        FLstNewNodes: TObjectList;
        protected
        {ajouter un noeud dans la liste initiale}
        procedure   AddLeaf(node: TNodeHAC);
        public
        {cration de la structure}
        constructor Create(prmCalc: TCalcHAC);
        {dtruire la structure}
        destructor  Destroy(); override;
        {initialiser les noeuds}
        procedure   InitializeNodes(prmExamples: TExamples);
        {rechercher le min}
        procedure   SearchMin(var i_min, j_min: integer; var min_d: double);
        {fusionner deux noeuds et l'ajouter dans la structure}
        procedure   MergeNodes(i_min,j_min: integer; prmDist: double);
        {calculer les coordonnes des X pour les feuilles}
        procedure   setLeavesCoords(node: TNodeHAC; var curCoord: double);
        {calculer les coordonnes des noeuds intrieurs de l'arbre}
        procedure   setNodesCoords(node: TNodeHAC);
        {rcuprer le nombre de noeuds}
        function    getCountNodes(): integer;
        {rcuprer le noeud numro index}
        function    getNode(index: integer): TNodeHAC;
        //proprits
        property    RootNode: TNodeHAC read FRootNode;
        end;

        TTabIndex = array of double;
        TTabDistIllusDummies = array of TTabIndex;

        TCalcHAC = class(TObject)
        private
        //paramtre de calcul
        FPrm: TPrmOpCatVarClusHCA;
        //inputs et illustratives
        FInputs, FIllus: TLstAttributes;
        //dummies actives
        FInputDummies: TListDummies;
        //dummies illustratives
        FIllusDummies: TListDummies;
        {Indice du palier max.}
        FIdMaxPalier: integer;
        {tableau des hauteurs des noeuds de l'arbre}
        FTabIndex: TTabIndex;
        {tableau des diffrences premires pour la dtection du nombre de clusters}
        FTabDifIndex: TTabIndex;
        {tableau des diffrences secondes pour la dtection du nombre de clusters}
        FTabDifSecond: TTabIndex;
        {structure d'arbre}
        FSetNodeHAC:TSetNodeHAC;
        {les noeuds de la solution adopte}
        FLstNodesSolution: TObjectList;
        {Matrice de distances}
        FDist, FTmpDist: TDistMatrix;
        {Matrice des distances aux groupes des variables illustratives}
        FTabDistIllusDummies: TTabDistIllusDummies;
        {longueur maximale de description}
        FMaxLengthDescription: integer;
        {initialiser les noeuds}
        procedure   InitializeNodes(prmExamples: TExamples);
        {calculer les coordonnes pour le dessin du dendrogramme}
        procedure   computeNodeCoords();
        protected
        function    getHTMLDescription(): string;
        function    getHTMLSolution(): string;
        function    getHTMLStructure(): string;
        function    getHTMLIndexSelection(): string;
        function    getHTMLResult(): string;
        function    getHTMLDummies(): string;
        procedure   buildBestNodes(examples: TExamples);
        public
        constructor Create(prmInputs,prmIllus: TLstAttributes; OpPrm: TPrmOpCatVarClusHCA);
        destructor  Destroy(); override;
        procedure   BuildClusters(prmExamples: TExamples);
        procedure   DistOfIllutratives(prmExamples: TExamples);
        //new - 07/12/2013 - calculer toutes les mesures - initalisation - n : taille d'chantillon
        procedure   calcAllMeasures(examples: TExamples);
        //mj de la matrice de distances - formule de Lance & Williams
        procedure   refreshMeasures(iEx,jEx: integer; nodeI, nodeJ: TNodeHAC);
        //mj selon le type d'agrgation
        function    refresh(d_ik, d_jk, d_ij: double; nodeI, nodeJ, nodeK: TNodeHAC; linkage: TEnumCatVarClusHAC): double;
        //dtection du plaier
        procedure   detectPalier();
        //
        property    SetNodeHAC: TSetNodeHAC read FSetNodeHAC;
        property    TabIndex: TTabIndex read FTabIndex;
        property    IdMaxPalier: integer read FIdMaxPalier;
        end;


implementation

uses
    Math, Windows, SysUtils,
    ULogFile, UConstConfiguration, UDlgOpPrmCatVarClusHAC,
    UFrmOpViewCatVarClusHAC, StrUtils;

const
        //dcalage de coordonnes des feuilles
        DECALAGE_COORD_LEAF : double = 10.0;


{ TGenCatVarClusHAC }

procedure TGenCatVarClusHAC.GenCompInitializations;
begin
 FMLComp:= mlcClustering;
end;

function TGenCatVarClusHAC.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompCatVarClusHAC;
end;

{ TMLCompCatVarClusHAC }

function TMLCompCatVarClusHAC.getClassOperator: TClassOperator;
begin
 result:= TOpCatVarClusHCA;
end;

{ TOpCatVarClusHCA }

function TOpCatVarClusHCA.CheckAttributes: boolean;
var ok: boolean;
begin
 //rcupration des variables
 FInputs:= workdata.LstAtts[asInput];
 FIllus:= workdata.LstAtts[asIllus];
 //vrification des input
 ok:= (FInputs.Count > 1) and (FInputs.isAllCategory(caDiscrete) = true);
 //vrification des illustratives
 if (FIllus.Count > 0)
  then ok:= ok and (FIllus.isAllCategory(caDiscrete) = true);
 result:= ok;
end;

function TOpCatVarClusHCA.CoreExecute: boolean;
begin
 result:= true;
 TRY
  if assigned(FCalc) then FCalc.Free();
  FCalc:= TCalcHAC.Create(FInputs,FIllus,self.PrmOp as TPrmOpCatVarClusHCA);
  FCalc.BuildClusters(self.WorkData.Examples);
 EXCEPT
  on e: Exception do
   begin
    Result:= false;
    TraceLog.WriteToLogFile(format('[CATVARHAC] erreur : %s',[e.Message]));
   end;
 END;
end;

function TOpCatVarClusHCA.getClassParameter: TClassOperatorParameter;
begin
 result:= TPrmOpCatVarClusHCA;
end;

function TOpCatVarClusHCA.getHTMLResultsSummary: string;
begin
 result:= FCalc.getHTMLResult();
end;

(*
function TOpCatVarClusHCA.getNewForm: TForm;
begin
 result:= nil;
end;
*)

function TOpCatVarClusHCA.getNewForm: TForm;
begin
 result:= TfrmOpViewCatVarClusHAC.CreateFromOperator(self);
end;

{ TPrmOpCatVarClusHCA }

function TPrmOpCatVarClusHCA.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrCatVarClusHAC.CreateFromOpPrm(self);
end;

const
    STR_HAC_CLUS_DET : array[0..1] of string = ('Automatic','User defined');
    STRCatVarClusHAC : array[TEnumCatVarClusHAC] of string = ('Single linkage', 'Complete linkage', 'Average linkage', 'Ward');



function TPrmOpCatVarClusHCA.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2># clusters</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Detection</TD><TD align=right>%s</TD></TR>',[STR_HAC_CLUS_DET[ModeBestSelection]]);
 if (ModeBestSelection = 1) then s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD># clusters</TD><TD align=right>%d</TD></TR>',[NbClusters]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Linkage criterion</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Algorithm</TD><TD align=right>%s</TD></TR>',[STRCatVarClusHAC[TEnumCatVarClusHAC(LinkageType)]]);
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH colspan=2>Visualization</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Index selection</TD><TD align=right>%d</TD></TR>',[ord(ShowIndexSelection)]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>Tree structure</TD><TD align=right>%d</TD></TR>',[ord(ShowTreeStructure)]);
 s:= s + '</table>';
 result:= s;
end;

procedure TPrmOpCatVarClusHCA.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 ModeBestSelection:= prmINI.ReadInteger(prmSection,'clus_selection',ModeBestSelection);
 NbClusters:= prmINI.ReadInteger(prmSection,'user_nb_clus',NbClusters);
 ShowTreeStructure:= prmINI.ReadBool(prmSection,'show_tree_structure',ShowTreeStructure);
 ShowIndexSelection:= prmINI.ReadBool(prmSection,'show_index_selection',ShowIndexSelection);
 MaxIndexLevel:= prmINI.ReadInteger(prmSection,'max_index_level',MaxIndexLevel);
 LinkageType:= prmINI.ReadInteger(prmSection,'linkage_type',LinkageType);
end;

procedure TPrmOpCatVarClusHCA.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(ModeBestSelection,sizeof(ModeBestSelection));
 prmStream.ReadBuffer(NbClusters,sizeof(NbClusters));
 prmStream.ReadBuffer(ShowTreeStructure,sizeof(ShowTreeStructure));
 prmStream.ReadBuffer(ShowIndexSelection,sizeof(ShowIndexSelection));
 prmStream.ReadBuffer(MaxIndexLevel,sizeof(MaxIndexLevel));
 prmStream.ReadBuffer(LinkageType,sizeof(LinkageType));
end;

procedure TPrmOpCatVarClusHCA.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'clus_selection',ModeBestSelection);
 prmINI.WriteInteger(prmSection,'user_nb_clus',NbClusters);
 prmINI.WriteInteger(prmSection,'normalization',Normalization);
 prmINI.WriteBool(prmSection,'show_tree_structure',ShowTreeStructure);
 prmINI.WriteBool(prmSection,'show_index_selection',ShowIndexSelection);
 prmINI.WriteInteger(prmSection,'max_index_level',MaxIndexLevel);
 prmINI.WriteInteger(prmSection,'linkage_type',LinkageType);
end;

procedure TPrmOpCatVarClusHCA.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(ModeBestSelection,sizeof(ModeBestSelection));
 prmStream.WriteBuffer(NbClusters,sizeof(NbClusters));
 prmStream.WriteBuffer(ShowTreeStructure,sizeof(ShowTreeStructure));
 prmStream.WriteBuffer(ShowIndexSelection,sizeof(ShowIndexSelection));
 prmStream.WriteBuffer(MaxIndexLevel,sizeof(MaxIndexLevel));
 prmStream.WriteBuffer(LinkageType,sizeof(LinkageType));
end;

procedure TPrmOpCatVarClusHCA.SetDefaultParameters;
begin
 self.LinkageType:= 2;//Average linkage par dfaut
 self.ShowTreeStructure:= false;
 self.ShowIndexSelection:= true;
 self.ModeBestSelection:= 0;
 self.NbClusters:= 3;
 self.MaxIndexLevel:= 0;
end;

{ TValueAttribute }

constructor TDummy.create(aVar: TAttribute;
  aValue: TTypeDiscrete);
begin
 inherited create();
 FVar:= aVar;
 FValue:= aValue;
end;

function TDummy.description: string;
begin
 result:= FVar.Name + ' = ' + FVar.LstValues.getDescription(FValue);
end;

destructor TDummy.destroy;
begin
 //rien pour l'instant
 inherited;
end;

function TDummy.diceIndex(other: TDummy; prmExamples: TExamples; linkageType: TEnumCatVarClusHAC): double;
var s: double;
    k_ij,k_ijprim: double;
    i: integer;
begin
 s:= 0.0;
 for i:= 1 to prmExamples.Size do
  begin
   k_ij:= self.getValue(prmExamples[i]);
   k_ijprim:= other.getValue(prmExamples[i]);
   s:= s + k_ij * (1.0 - k_ijprim) + k_ijprim * (1.0 - k_ij);
  end;
 //result:= 0.5 * s;
 s:= 0.5 * s;
 //new -- 21/12/2013 -- elle est au carr l, il faut prendre la racine carre (sauf pour Ward)
 //selon le mode de d'agrgation
 case linkageType of
    cvhWard:  result:= 0.5 * s; {(1*1)/(1+1) * s}
    else result:= SQRT(s);
 end;
end;

function TDummy.getValue(example: integer): double;
begin
 result:= integer(FVar.dValue[example] = FValue);
end;

{ TListDummies }

procedure TListDummies.add(aDummy: TDummy);
begin
 FLst.Add(aDummy);
end;

procedure TListDummies.addDummies(aVar: TAttribute);
var k: TTypeDiscrete;
begin
 if (aVar.isCategory(caDiscrete) = true)
  then
   begin
    for k:= 1 to aVar.nbValues do
     self.add(TDummy.create(aVar,k));
   end;
end;

procedure TListDummies.assign(source: TListDummies);
var i: integer;
begin
 for i:= 0 to pred(source.getCount()) do
  FLst.Add(source.get(i));
end;

function TListDummies.averageLinkageDice(other: TDummy; prmExamples: TExamples): double;
var s: double;
    i: integer;
begin
 result:= Math.MaxDouble;
 if (self.getCount() > 0)
  then
   begin
    s:= 0.0;
    for i:= 0 to pred(self.getCount()) do
     s:= s + self.get(i).diceIndex(other,prmExamples,cvhAverage);
    s:= s/(1.0 * self.getCount());
    result:= s;
   end;
end;

function TListDummies.completeLinkageDice(other: TDummy; prmExamples: TExamples): double;
var s,value: double;
    i: integer;
begin
 result:= Math.MaxDouble;
 if (self.getCount() > 0)
  then
   begin
    s:= -1.0 * Math.MaxDouble;
    for i:= 0 to pred(self.getCount()) do
     begin
      value:= self.get(i).diceIndex(other,prmExamples,cvhComplete);
      if (value > s)
       then s:= value;
     end;
    result:= s;
   end;
end;

constructor TListDummies.create(aOwner: boolean);
begin
 inherited create();
 FLst:= TObjectList.Create(aOwner);
end;

constructor TListDummies.create(aOwner: boolean; aLstVar: TLstAttributes);
var i: integer;
begin
 inherited create();
 FLst:= TObjectList.Create(aOwner);
 for i:= 0 to pred(aLstVar.Count) do
  self.addDummies(aLstVar.Attribute[i]); 
end;

function TListDummies.destription: string;
var i: integer;
    s: string;
begin
 s:= '';
 for i:= 0 to pred(self.getCount()) do
  s:= s + self.get(i).description() + #13#10;
 if (length(s) > 0)
  then result:= copy(s,1,length(s)-2)
  else result:= '';
end;

destructor TListDummies.destroy;
begin
 FLst.Free();
 inherited;
end;

function TListDummies.get(i: integer): TDummy;
begin
 result:= FLst.Items[i] as TDummy;
end;

function TListDummies.getCount: integer;
begin
 result:= FLst.Count;
end;

function TListDummies.getMaxLengthDescription: integer;
var max,i: integer;
begin
 max:= 0;
 for i:= 0 to pred(self.getCount()) do
  if (length(self.get(i).description()) > max)
   then max:= length(self.get(i).description());
 result:= max;
end;

function TListDummies.LinkageDice(linkage: TEnumCatVarClusHAC;
  other: TDummy; prmExamples: TExamples): double;
begin
 case linkage of
  cvhSingle: result:= self.singleLinkageDice(other,prmExamples);
  cvhComplete: result:= self.completeLinkageDice(other,prmExamples);
  cvhWard: result:= self.wardLinkageDice(other,prmExamples);
  else result:= self.averageLinkageDice(other,prmExamples); 
 end;
end;

function TListDummies.singleLinkageDice(other: TDummy; prmExamples: TExamples): double;
var s,value: double;
    i: integer;
begin
 result:= Math.MaxDouble;
 if (self.getCount() > 0)
  then
   begin
    s:= +1.0 * Math.MaxDouble;
    for i:= 0 to pred(self.getCount()) do
     begin
      value:= self.get(i).diceIndex(other,prmExamples,cvhSingle);
      if (value < s)
       then s:= value;
     end;
    result:= s;
   end;
end;

constructor TDistMatrix.create(aSize: integer);
var i: integer;
begin
 inherited Create();
 Size:= aSize;
 setLength(Tab,Size);
 for i:= 0 to pred(Size) do
  setLength(Tab[i],i);
end;

procedure TDistMatrix.DecrementSize;
begin
 Finalize(Tab[Size-1]);
 setLength(Tab,Size-1);
 dec(Size);
end;

destructor TDistMatrix.destroy;
var i: integer;
begin
 for i:= 0 to pred(length(Tab)) do
  Finalize(Tab[i]);
 Finalize(Tab);
 inherited;
end;

function TDistMatrix.getValue(i, j: integer): double;
begin
 if (i > j)
  then result:= Tab[i,j]
  else result:= Tab[j,i];
end;

procedure TDistMatrix.setValue(i, j: integer; value: double);
begin
 if (i > j)
  then Tab[i,j]:= value
  else Tab[j,i]:= value;
end;

//comparer selon la description
function CompareDescription(Item1, Item2: Pointer): Integer;
begin
 if (TDummy(Item1).description() < TDummy(Item2).description())
  then result:= -1
  else
   if (TDummy(Item1).description() > TDummy(Item2).description())
    then result:= +1
    else result:= 0;
end;

procedure TListDummies.sortAccordingDescription;
begin
 self.FLst.Sort(CompareDescription);
end;

function TListDummies.wardLinkageDice(other: TDummy;
  prmExamples: TExamples): double;
var s,m: double;
    i,j: integer;
begin
 result:= Math.MaxDouble;
 if (self.getCount() > 0)
  then
   begin
    s:= 0.0;
    for i:= 1 to prmExamples.Size do
     begin
      //calculer la moyenne pour l'individu
      m:= 0;
      for j:= 0 to pred(self.getCount()) do
       m:= m + self.get(j).getValue(prmExamples[i]);
      m:= m / (1.0 * self.getCount());
      //faire la somme d'cart au carr
      s:= s + SQR(m - other.getValue(prmExamples[i]));
     end;
    //indice de Dice (qui est dj au carr)
    s:= 0.5 * s;
    //indice de Ward
    s:= (1.0 * self.getCount())/(1.0 + self.getCount()) * s;
    //and then...
    result:= s;
   end;
end;

{ TNodeHAC }

constructor TNodeHAC.Create(prmCalc: TCalcHAC; aDummy: TDummy = nil);
begin
 inherited Create();
 FCalc:= prmCalc;
 FDad:= NIL;
 FChildL:= NIL;
 FChildR:= NIL;
 FXPos:= -1.0;
 FIsCluster:= false;
 FDummies:= TListDummies.create(FALSE);
 if assigned(aDummy)
  then FDummies.add(aDummy);
end;

destructor TNodeHAC.Destroy;
begin
 FDummies.Free();
 inherited Destroy();
end;

function TNodeHAC.getLabel: string;
var s: string;
    l: integer;
begin
 l:= self.FCalc.FMaxLengthDescription + 4;
 s:= Format('Members (%d) - Height : %3f',[FDummies.getCount(),self.YPos]) + #13#10;
 l:= Math.Max(l,length(s));
 s:= s + DupeString('-',l) + #13#10;
 s:= s + FDummies.destription();
 result:= s;
end;

function TNodeHAC.getNbInstances: integer;
begin
 result:= FDummies.getCount();
end;

function TNodeHAC.isLeaf: boolean;
begin
 result:= (self.FChildL = nil) and (self.FChildR = nil);
end;

function TNodeHAC.MergeDistance(other: TNodeHAC; prmExamples:  TExamples; linkageType: TEnumCatVarClusHAC): double;
begin
 result:= self.FDummies.get(0).diceIndex(other.FDummies.get(0),prmExamples, linkageType);
end;

{ TSetNodeHAC }

procedure TSetNodeHAC.AddLeaf(node: TNodeHAC);
begin
 FLstAllNodes.Add(node);
 FLstMergeNodes.Add(node);
 FLstLeaves.Add(node);
 node.FNumNode:= FLstLeaves.Count;
 node.FDist:= 0.0;
end;

constructor TSetNodeHAC.Create(prmCalc: TCalcHAC);
begin
 inherited Create();
 FCalcHAC:= prmCalc;
 //les listes internes
 FLstAllNodes:= TObjectList.Create(TRUE);
 FLstMergeNodes:= TObjectList.Create(FALSE);
 FLstLeaves:= TObjectList.Create(FALSE);
 FLstNewNodes:= TObjectList.Create(FALSE);
end;

destructor TSetNodeHAC.Destroy;
begin
 FLstNewNodes.Free;
 FLstLeaves.Free;
 FLstMergeNodes.Free;
 FLstAllNodes.Free;
 inherited Destroy();
end;

function TSetNodeHAC.getCountNodes: integer;
begin
 result:= self.FLstAllNodes.Count;
end;

function TSetNodeHAC.getNode(index: integer): TNodeHAC;
begin
 result:= self.FLstAllNodes[index] as TNodeHAC;
end;

procedure TSetNodeHAC.InitializeNodes(prmExamples: TExamples);
var i: integer;
begin
 for i:= 0 to pred(FCalcHAC.FInputDummies.getCount()) do
  self.AddLeaf(TNodeHAC.Create(self.FCalcHAC,FCalcHAC.FInputDummies.get(i)));
end;

procedure TSetNodeHAC.MergeNodes(i_min, j_min: integer; prmDist: double);
var node: TNodeHAC;
    nodeL,nodeR: TNodeHAC;
begin
 //noeuds  dtruire
 nodeR:= self.FLstMergeNodes.Items[i_min] as TNodeHAC;
 nodeL:= self.FLstMergeNodes.Items[j_min] as TNodeHAC;
 //noeud cr
 node:= TNodeHAC.Create(self.FCalcHAC);
 //Numro du dernier sommet ajout et sa hauteur
 inc(self.FLastNumNode);
 node.FNumNode:= self.FLastNumNode;
 //diviser par n
 node.FDist:= prmDist;
 //affecter les donnes simples pour chaque variable j
 node.FDummies.assign(nodeL.FDummies);
 node.FDummies.assign(nodeR.FDummies);
 //mj des listes internes et des branchements
 node.FChildL:= nodeL;
 node.FChildR:= nodeR;
 nodeL.FDad:= node;
 nodeR.FDad:= node;
 self.FLstAllNodes.Add(node);
 self.FLstNewNodes.Add(node);
 self.FLstMergeNodes.Remove(nodeL);
 self.FLstMergeNodes.Remove(nodeR);
 //mise  jour des distances
 self.FCalcHAC.refreshMeasures(i_min,j_min,nodeR,nodeL);
 //ajouter node dans la liste des " traiter"
 self.FLstMergeNodes.Add(node);
end;

procedure TSetNodeHAC.SearchMin(var i_min, j_min: integer;
  var min_d: double);
var i,j: integer;
    d: double;
begin
 i_min:= 0;
 j_min:= 0;
 min_d:= Math.MaxDouble;
 For i:= 1 to pred(self.FLstMergeNodes.Count) do
  for j:= 0 to pred(i) do
   begin
    d:= FCalcHAC.FDist.getValue(i,j);
    //"<=" et non "<" pour tre raccord avec hclust de R !!! spcial a quand mme...
    if (d <= min_d)
     then
      begin
       i_min:= i;
       j_min:= j;
       min_d:= d;
      end;
   end;
end;

procedure TSetNodeHAC.setLeavesCoords(node: TNodeHAC; var curCoord: double);
begin
 if not(node.isLeaf())
  then
   begin
    self.setLeavesCoords(node.FChildL,curCoord);
    self.setLeavesCoords(node.FChildR,curCoord);
   end
  else
   begin
    node.FXPos:= curCoord;
    curCoord:= curCoord + DECALAGE_COORD_LEAF;
   end;
end;

procedure TSetNodeHAC.setNodesCoords(node: TNodeHAC);
begin
 //doit-on en calculer les coordonnes ?
 if (node.FXPos < 0.0)
  then
   begin
    //tester si fils gauche n'est pas bien plac
    if (node.FChildL.FXPos < 0.0)
     then setNodesCoords(node.FChildL);
    //tester  droite
    if (node.FChildR.FXPos < 0.0)
     then setNodesCoords(node.FChildR);
    //se positionner au milieu de ses enfants
    node.FXPos:= 0.5 * (node.FChildL.FXPos + node.FChildR.FXPos);
   end;
end;

{ TCalcHAC }

procedure TCalcHAC.buildBestNodes(examples: TExamples);
var i,j,k: integer;
    node,otherNode: TNodeHAC;
    dummy: TDummy;
    min,d: double;
begin
 //ide est trs simple - on part de la racine, et on descend en se calant sur la hauteur
 //on ajoute et on extrait itrativement les sommets jusqu' ce que le nombre de solutions soit gal  celui demand !
 FLstNodesSolution:= TObjectList.Create(FALSE);
 FLstNodesSolution.Add(FSetNodeHAC.FRootNode);
 for i:= 1 to FIdMaxPalier+1 do
  begin
   node:= FSetNodeHAC.FLstNewNodes.Items[FSetNodeHAC.FLstNewNodes.Count-i] as TNodeHAC;
   FLstNodesSolution.Extract(node);
   FLstNodesSolution.Add(node.FChildL);
   FLstNodesSolution.Add(node.FChildR);
  end;
 //new -- indiquer au noeud qu'il est un cluster
 for i:= 0 to pred(FLstNodesSolution.Count) do
  begin
   node:= FLstNodesSolution.Items[i] as TNodeHAC;
   node.FIsCluster:= TRUE;
   //son plus proche voisin loign et son plus lointain coloc
   for j:= 0 to pred(node.FDummies.getCount) do
    begin
     dummy:= node.FDummies.get(j);
     dummy.FMaxIndexOwn:= node.FDummies.LinkageDice(TEnumCatVarClusHAC(FPrm.LinkageType),dummy,examples);
     min:= 1.0 * Math.MaxDouble;
     for k:= 0 to pred(FLstNodesSolution.Count) do
      begin
       otherNode:= FLstNodesSolution.Items[k] as TNodeHAC;
       if (node <> otherNode)
        then
         begin
          d:= otherNode.FDummies.LinkageDice(TEnumCatVarClusHAC(FPrm.LinkageType),dummy,examples);
          if (d < min)
           then min:= d;
         end;
      end;
     dummy.FMinIndexNext:= min;
    end;
   
   //trier les dummies selon la description
   node.FDummies.sortAccordingDescription();
  end;
end;

procedure TCalcHAC.BuildClusters(prmExamples: TExamples);
var tps: cardinal;
    i_min,j_min: integer;
    min_d: double;
begin
 //initialiser les noeuds de la structure
 TRY
 self.InitializeNodes(prmExamples);
 //new - 07/12/2013 - calculer toutes les mesures d'agrgation
 tps:= GetTickCount();
 FDist:= TDistMatrix.create(FSetNodeHAC.FLstMergeNodes.Count);
 FTmpDist:= TDistMatrix.create(FSetNodeHAC.FLstMergeNodes.Count);
 self.calcAllMeasures(prmExamples);
 tps:= GetTickCount()-tps;
 TraceLog.WriteToLogFile(format('[HAC] TObjectList, calcul distances = %d ms.',[tps]));
 //fusionner en remontant jusqu' la racine de l'arbre
 tps:= GetTickCount();
 while (FSetNodeHAC.FLstMergeNodes.Count > 1) do
  begin
   //recherche du minimum
   FSetNodeHAC.SearchMin(i_min,j_min,min_d);
   //fusion
   FSetNodeHAC.MergeNodes(i_min,j_min,min_d);
  end;
 //librer la matrice de distance
 FTmpDist.Free();
 FDist.Free();
 tps:= GetTickCount()-tps;
 TraceLog.WriteToLogFile(format('[CATVARHAC] TObjectList, construction hierarchie = %d ms.',[tps]));
 //le premier sommet est la racine du dendrogramme
 FSetNodeHAC.FRootNode:= FSetNodeHAC.FLstMergeNodes.Items[0] as TNodeHAC;
 TraceLog.WriteToLogFile(format('[CATVARHAC] -> dummies sur la racine : %d, distance agregation = %.4f',[FSetNodeHAC.FRootNode.FDummies.getCount(),FSetNodeHAC.FRootNode.FDist]));
 //structure arbre ok - rechercher maintenant le meilleur cluster
 self.detectPalier();
 //forcer la solution si demande par l'utilisateur
 if (self.FPrm.ModeBestSelection = 1)
  then self.FIdMaxPalier:= self.FPrm.NbClusters-2;
 //dtecter les noeuds du clustering dfinitif
 self.buildBestNodes(prmExamples);
 //calculer les distances des dummies illustratives avec les clusters
 self.DistOfIllutratives(prmExamples);
 //new -- 17/06/2006 -- calculer les coordonnes des noeuds pour dessiner le dendrogramme
 self.computeNodeCoords();
 EXCEPT
  on e: exception do
   begin
    TraceLog.WriteToLogFile(format('[HAC] :: BuildClusters = %s',[e.Message]));
    raise Exception.Create(e.Message);
   end;
 END;
end;

procedure TCalcHAC.calcAllMeasures(examples: TExamples);
var i,j: integer;
    nodeI,nodeJ: TNodeHAC;
    d: double;
begin
 //tester le croisement
 for i:= 1 to FSetNodeHAC.FLstMergeNodes.Count-1 do
  begin
   nodeI:= FSetNodeHAC.FLstMergeNodes.Items[i] as TNodeHAC;
   for j:= 0 to pred(i) do
    begin
     nodeJ:= FSetNodeHAC.FLstMergeNodes.Items[j] as TNodeHAC;
     //distance euclidienne = indice de Dice pour les indicatrices
     d:= nodeI.MergeDistance(nodeJ, examples,TEnumCatVarClusHAC(self.FPrm.LinkageType));
     //affectation
     FDist.setValue(i,j,d);
    end;
  end;
end;

procedure TCalcHAC.computeNodeCoords;
var curCoord: double;
begin
 curCoord:= 0.0;
 //numroter les feuilles
 self.FSetNodeHAC.setLeavesCoords(self.SetNodeHAC.FRootNode,curCoord);
 //numroter les noeuds
 self.FSetNodeHAC.setNodesCoords(self.SetNodeHAC.FRootNode);
end;

constructor TCalcHAC.Create(prmInputs,prmIllus: TLstAttributes; OpPrm: TPrmOpCatVarClusHCA);
begin
 inherited Create();
 FPrm:= OpPrm as TPrmOpCatVarClusHCA;
 FInputs:= prmInputs;
 FIllus:= prmIllus;
 //cration des dummies
 FInputDummies:= TListDummies.create(TRUE,prmInputs);
 FIllusDummies:= TListDummies.create(TRUE,prmIllus);
 //longueur max de description
 self.FMaxLengthDescription:= FInputDummies.getMaxLengthDescription();
end;

destructor TCalcHAC.Destroy;
begin
 SetLength(FTabIndex,0);
 SetLength(FTabDifIndex,0);
 SetLength(FTabDifSecond,0);
 setLength(FTabDistIllusDummies,0,0);
 if assigned(FLstNodesSolution)
  then FLstNodesSolution.Free;
 if assigned(FSetNodeHAC)
  then FSetNodeHAC.Free;
 if assigned(FInputDummies)
  then FInputDummies.Free();
 if assigned(FIllusDummies)
  then FIllusDummies.Free();
 inherited;
end;

procedure TCalcHAC.detectPalier;
var i: integer;
    node: TNodeHAC;
    max: double;
begin
 setLength(FTabIndex,FSetNodeHAC.FLstNewNodes.Count+1);
 setLength(FTabDifIndex,FSetNodeHAC.FLstNewNodes.Count+1);
 setLength(FTabDifSecond,FSetNodeHAC.FLstNewNodes.Count+1);
 //au premier sommet, pas de partitionnement
 FTabIndex[0]:= 0;
 FTabDifIndex[0]:= 0;
 //tableau des distances d'agrgation
 for i:= 0 to pred(FSetNodeHAC.FLstNewNodes.Count) do
  begin
   node:= FSetNodeHAC.FLstNewNodes.Items[pred(FSetNodeHAC.FLstNewNodes.Count)-i] as TNodeHAC;
   FTabIndex[i+1]:= node.FDist;
  end;
 //calcul des diffrences 1res
 for i:= 1 to high(FTabDifIndex)-1 do
  FTabDifIndex[i]:= FTabIndex[i]-FTabIndex[succ(i)];
 //ne pas oublier le premier saut
 FTabDifIndex[high(FtabDifIndex)]:= (FSetNodeHAC.FLstNewNodes.Items[0] as TNodeHAC).FDist;
 //calcul des diffrences secondes
 for i:= 1 to high(FTabDifSecond)-2 do
  FTabDifSecond[i]:= FTabDifIndex[i]-FTabDifIndex[succ(i)];
 //dtecter automatiquement le palier
 FIdMaxPalier:= -1;
 max:= -1.0 * math.MaxDouble;
 //for i:= low(FTabDifIndex) to high(FTabDifIndex)-1 do
 for i:= 1 to high(FTabDifSecond)-2 do
  begin
   if (FTabDifSecond[i] > max)
    then
     begin
      max:= FTabDifSecond[i];
      FIdMaxPalier:= i;
     end;
  end;
 FIdMaxPalier:= FIdMaxPalier - 1;
 TraceLog.WriteToLogFile(Format('[HAC] id max palier = %d',[FIdMaxPalier]));
end;

const
    TAB_HTML_ROW_COLOR : array[0..1] of string = (HTML_TABLE_COLOR_DATA_GRAY,HTML_TABLE_COLOR_HEADER_GRAY);
    TAB_TD_HTML_RATIO_COLOR: array[0..1] of string = (HTML_BGCOLOR_DATA_GRAY,HTML_BGCOLOR_HEADER_GRAY);
    THRESHOLD_GOOD_RATIO : double = 1.0;

procedure TCalcHAC.DistOfIllutratives(prmExamples: TExamples);
var i,j: integer;
    dummy: TDummy;
    node: TNodeHAC;
begin
 setLength(FTabDistIllusDummies,FIllusDummies.getCount(),self.FLstNodesSolution.Count);
 for i:= 0 to pred(FIllusDummies.getCount()) do
  begin
   dummy:= FIllusDummies.get(i);
   for j:= 0 to pred(self.FLstNodesSolution.Count) do
    begin
     node:= self.FLstNodesSolution.Items[j] as TNodeHAC;
     FTabDistIllusDummies[i,j]:= node.FDummies.LinkageDice(TEnumCatVarClusHAC(FPrm.LinkageType),dummy,prmExamples);
    end;
  end;
end;

function TCalcHAC.getHTMLDescription: string;
var j: integer;
    s: string;
    node: TNodeHAC;
begin
 s:= '<P><H3>Clusters'' characteristics</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Cluster number</TH><TH width=100>Cluster size</TH></TR>';
 for j:= 0 to pred(self.FLstNodesSolution.Count) do
  begin
   node:= self.FLstNodesSolution.Items[j] as TNodeHAC;
   s:= s + TAB_HTML_ROW_COLOR[j mod 2];
   s:= s + format('<TD align="center">%d</TD><TD align="center">%d</TD></TR>',[succ(j),node.FDummies.getCount()]);
  end;
 s:= s + '</table>';
 result:= s;
end;

const
    BG_HIGHLIGHT : string = 'BGCOLOR="#FCC998"';
    BG_NORMAL : string = HTML_BGCOLOR_DATA_GRAY;

function TCalcHAC.getHTMLDummies: string;
var i,j: integer;
    s: string;
    min: double;
    dummy: TDummy;
    bgColor: string;
begin
 s:= '<P><H3>Distance to clusters - Supplementary variables</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Variable = level</TH>';;
 for j:= 0 to pred(self.FLstNodesSolution.Count) do
  s:= s + format('<TH width=70>Cluster %d</TH>',[succ(j)]);
 s:= s + '</TR>';
 for i:= 0 to pred(FIllusDummies.getCount()) do
  begin
   dummy:= FIllusDummies.get(i);
   min:= Math.MinValue(FTabDistIllusDummies[i]);
   s:= s + HTML_TABLE_COLOR_DATA_GRAY;
   s:= s + format('<TD>%s</TD>',[dummy.description()]);
   for j:= 0 to pred(self.FLstNodesSolution.Count) do
    begin
     if (FTabDistIllusDummies[i,j] = min)
      then bgColor:= BG_HIGHLIGHT
      else bgColor:= BG_NORMAL;
     s:= s + format('<TD align="right" %s>%.3f</TD>',[bgColor,FTabDistIllusDummies[i,j]]);
    end;
   s:= s + '</TR>';
  end;
 s:= s + '</table>';
 result:= s;
end;

function TCalcHAC.getHTMLIndexSelection: string;
var s: string;
    sIdMax: string;
    i: integer;
    maxLevel: integer;
    sValueSecond: string;
begin
 s:= '<P><H3>Best cluster selection</H3><BR>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+format('%s <th width=60>Clusters</th><th width=60>Height</th><th width=60>DIF.1</th><th width=60>DIF.2</th></tr>',[HTML_TABLE_COLOR_HEADER_GRAY]);
 maxLevel:= FSetNodeHAC.FLstNewNodes.Count;
 //premire ligne, pas de fusion de toute manire
 s:= s + format('%s <td>%d</td><td align="right">%6.4f</td><td align="center">%s</td><td align="center">%s</td></tr>',
                [HTML_TABLE_COLOR_DATA_GRAY,1,FTabIndex[1],'-','-']);
 for i:= 2 to maxLevel do
  begin
   //indiquer la bonne solution
   if (i = FIdMaxPalier+2)
    then sIdMax:= HTML_TABLE_COLOR_DATA_GREEN
    else sIdMax:= HTML_TABLE_COLOR_DATA_GRAY;
   //value diff. second
   if (i = high(FTabDifSecond))
    then sValueSecond:= '-'
    else sValueSecond:= format('%6.4f',[FTabDifSecond[i-1]]);
   //envoyer la sauce
   s:= s+format('%s <td>%d</td><td align="right">%6.4f</td><td align="right">%6.4f</td><td align="right">%s</td></tr>',
                [sIdMax,i,FTabIndex[i],FTabDifIndex[i-1],sValueSecond]);
  end;
 s:= s+'</table>';
 result:= s;
end;

function TCalcHAC.getHTMLResult: string;
var s: string;
begin
 s:= '';
 //description des effectifs par groupe
 s:= s + self.getHTMLDescription();
 //rattachement de chaque dummy  un cluster - qualit du rattachement
 s:= s + self.getHTMLSolution();
 //positionnement des modalits des variables illustratives
 if (self.FIllusDummies.getCount() > 0)
  then s:= s + self.getHTMLDummies(); 
 //affichage de la slection du meilleur cluster ?
 if FPrm.ShowIndexSelection
  then s:= s + self.getHTMLIndexSelection();
 //affichage de la structure de l'arbre ?
 if FPrm.ShowTreeStructure
  then s:= s + self.getHTMLStructure();
 //zoo...
 result:= s;
end;

function TCalcHAC.getHTMLSolution: string;
var j,k: integer;
    s: string;
    node: TNodeHAC;
    dummy: TDummy;
    nbLig: integer;
    ratio: double;
    TDHighlight: string;
begin
 s:= '<P><H3>Clusters'' members</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + '<TH>Cluster</TH><TH>Members</TH><TH width=80>Distance<br>Own Cluster</TH><TH width=80>Distance<br>Next Closest</TH><TH width=80>Ratio<br>(Own / Next)</TH></TR>';
 for j:= 0 to pred(self.FLstNodesSolution.Count) do
  begin
   node:= self.FLstNodesSolution.Items[j] as TNodeHAC;
   s:= s + TAB_HTML_ROW_COLOR[j mod 2];
   nbLig:= MATH.Max(1,node.FDummies.getCount());
   s:= s + format('<TD align=center valign=top rowspan=%d>%d<br>(Size = %d)</TD>',[nbLig,succ(j),node.FDummies.getCount()]);
   if (node.FDummies.getCount() = 0)
    then s:= s + '<TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD></TR>'
    else
     begin
      for k:= 0 to pred(node.FDummies.getCount()) do
       begin
        dummy:= node.FDummies.get(k);
        if (k > 0) then s:= s + TAB_HTML_ROW_COLOR[j mod 2];
        //ratio
        ratio:= 0;
        if (dummy.FMinIndexNext > 0)
         then ratio:= dummy.FMaxIndexOwn / dummy.FMinIndexNext;
        if (ratio >= THRESHOLD_GOOD_RATIO)
         then TDHighlight:= HTML_BGCOLOR_DATA_RED
         else TDHighlight:= TAB_TD_HTML_RATIO_COLOR[j mod 2];
        //
        s:= s + format('<TD align=left>%s</TD><TD align=right>%.2f</TD><TD align=right>%.2f</TD><TD align=right %s>%.4f</TD></TR>',
                       [dummy.description(),dummy.FMaxIndexOwn,dummy.FMinIndexNext,TDHighlight,ratio]);
       end;
     end;
  end;
 s:= s + '</table>';
 result:= s;
end;

function TCalcHAC.getHTMLStructure: string;
var s: string;
    i: integer;
    node: TNodeHAC;
begin
 s:= '<P><H3>Tree structure</H3><BR>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+format('%s <th colspan="4">First level clusters</th></tr>',[HTML_TABLE_COLOR_HEADER_GRAY]);
 s:= s+format('%s <td align="left">Tree leaves</td><td align="right" colspan="3">%d</td></tr>',[HTML_TABLE_COLOR_DATA_GRAY,FSetNodeHAC.FLstLeaves.Count]);
 s:= s+format('%s <th>Node</th><th>Size</th><th>Childs</th><th>Height</th></tr>',[HTML_TABLE_COLOR_HEADER_GRAY]);
 for i:= 0 to pred(FSetNodeHAC.FLstLeaves.Count) do
  begin
   node:= FSetNodeHAC.FLstLeaves.Items[i] as TNodeHAC;
   s:= s+format('%s <td align="center">%d</td><td align="right">%d</td><td align="center">-</td><td align="center">-</td></tr>',[HTML_TABLE_COLOR_DATA_GRAY,node.FNumNode,node.getNbInstances()]);
  end;
 for i:= 0 to pred(FSetNodeHAC.FLstNewNodes.Count) do
  begin
   node:= FSetNodeHAC.FLstNewNodes.Items[i] as TNodeHAC;
   s:= s+format('%s <td align="center">%d</td><td align="right">%d</td><td align="right">(%d,%d)</td><td align="right">%8.4f</td></tr>',[HTML_TABLE_COLOR_DATA_GRAY,node.FNumNode,node.getNbInstances(),node.FChildL.FNumNode,node.FChildR.FNumNode,node.FDist]);
  end;
 s:= s+'</table>';
 result:= s;
end;

procedure TCalcHAC.InitializeNodes(prmExamples: TExamples);
begin
 //construire le 1er niveau de la structure d'arbre
 FSetNodeHAC:= TSetNodeHAC.Create(self);
 FSetNodeHAC.InitializeNodes(prmExamples);
 FSetNodeHAC.FLastNumNode:= FSetNodeHAC.FLstLeaves.Count;
end;


function TCalcHAC.refresh(d_ik, d_jk, d_ij: double; nodeI,
  nodeJ, nodeK: TNodeHAC; linkage: TEnumCatVarClusHAC): double;
var a_i, a_j, b, c: double;
    denom: double;
begin
 //pondration selon le type d'agrgation
 case linkage of
  cvhSingle:
   begin
    a_i:= 0.5;
    a_j:= 0.5;
    b:= 0.0;
    c:= -0.5;
   end;
  cvhComplete:
   begin
    a_i:= 0.5;
    a_j:= 0.5;
    b:= 0.0;
    c:= +0.5;
   end;
  cvhWard:
   begin
   denom:= 1.0*(nodeK.getNbInstances+nodeI.getNbInstances+nodeJ.getNbInstances);
   a_i:= 1.0*(nodeK.getNbInstances + nodeI.getNbInstances)/denom;
   a_j:= 1.0*(nodeK.getNbInstances + nodeJ.getNbInstances)/denom;
   b:= -1.0*(nodeK.getNbInstances)/denom;
   c:= 0.0;
   end;
  else
   //par dfaut, average linkage
   begin
    a_i:= 1.0*nodeI.getNbInstances()/(1.0*(nodeI.getNbInstances()+nodeJ.getNbInstances()));
    a_j:= 1.0*nodeJ.getNbInstances()/(1.0*(nodeI.getNbInstances()+nodeJ.getNbInstances()));
    b:= 0.0;
    c:= 0.0;
   end;
 end;
 //formule de Lance & Williams
 result:= a_i * d_ik + a_j * d_jk + b * d_ij + c * ABS(d_ik - d_jk);
end;

procedure TCalcHAC.refreshMeasures(iEx, jEx: integer; nodeI, nodeJ: TNodeHAC);
var idNode: array of integer;
    k,i,j: integer;
    nodeK: TNodeHAC;
    d,d_ik,d_jk,d_ij: double;
    exchange: TDistMatrix;
begin
 //les indices mesures  reporter
 setLength(idNode,FDist.Size-2);
 k:= 0;
 for i:= 0 to pred(FDist.Size-2) do
  begin
   if (k = jEx)
    then inc(k);
   if (k = iEx)
    then inc(k);
   idNode[i]:= k;
   inc(k);
  end;
 //reporter les mesures
 FTmpDist.DecrementSize();
 for i:= 1 to FTmpDist.Size-2 do
  for j:= 0 to pred(i) do
   FTmpDist.setValue(i,j,FDist.getValue(idNode[i],idNode[j]));
 //calculer la distance par rapport au nouveau cluster
 //coefficients - formule de Lance & Williams
 d_ij:= FDist.getValue(iEx,jEx);
 for k:= 0 to length(idNode)-1 do
  begin
   nodeK:= self.FSetNodeHAC.FLstMergeNodes.Items[k] as TNodeHAC;
   d_ik:= FDist.getValue(iEx,idNode[k]);
   d_jk:= FDist.getValue(jEx,idNode[k]);
   d:= self.refresh(d_ik,d_jk,d_ij,nodeI,nodeJ,nodeK,TEnumCatVarClusHAC(FPrm.LinkageType));
   FTmpDist.setValue(FTmpDist.Size-1,k,d);
  end;
 exchange:= FDist;
 FDist:= FTmpDist;
 FTmpDist:= exchange;
 FTmpDist.DecrementSize();
 Finalize(idNode);
end;

initialization
 Classes.RegisterClass(TGenCatVarClusHAC);
end.
