(****************************************************************)
(* UCompVarClusBase.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(****************************************************************)

{
@abstract(Composant de base pour le clustering de variables)
@author(Ricco)
@created(19/02/2007)

Pour la classification de variables, on mettra au point au moins 3 approches :
- VarKMeans
- VarClus Bottom-Up i.e. VarHCA
- VarClus Top-Down i.e. VarClus, algorithme descendant

Ils ont en commun :
- vrification des Input : tous continus
- production d'une srie de nouvelles variables, autant qu'il y a de clusters
- option variable reprsentation du cluster, soit l'axe factoriel associ au cluster, soit la variable la plus proche
- la description des clusters,  la sauce SAS


}
unit UCompVarClusBase;

interface

uses
    UCompDefinition,
    UCompManageDataset,
    UDatasetDefinition,
    UDatasetImplementation,
    UOperatorDefinition,
    UDatasetExamples,
    UCalcVarClusBase,
    Classes, IniFiles;

TYPE
    //gnrateur de composant
    TMLGenVarClus = class(TMLGenComp)
    protected
    procedure   GenCompInitializations(); override;
    end;

    //composant
    TMLCompVarClusBase = class(TMLCompLocalData)
    protected
    //nom gnrique des variables produites
    function getGenericAttName(): string; virtual; abstract;
    end;

    //oprateur
    TOpVarClusBase = class(TOpLocalData)
    protected
    //les inputs
    FInputs: TLstAttributes;
    //rcuprer l'objet de calcul
    FCalc: TCalcVarClus;
    //demander la classe de calcul
    function getClassCalc(): TClassCalcVarClus; virtual; abstract;
    //vrification des attributs en entre, tous continus, au moins deux
    function CheckAttributes(): boolean; override;
    public
    //destructeur
    destructor  destroy(); override;
    //lancer les calculs
    function    CoreExecute(): boolean; override;
    //raliser les projections
    procedure   setProjections();
    //rapport
    function    getHTMLResultsSummary(): string; override;
    //
    property    Calc: TCalcVarClus read FCalc;
    end;

    //paramtrage
    TOpPrmVarClus = class(TOperatorParameter)
    protected
    procedure   SetDefaultParameters(); override;
    public
    //paramtres d'affichage des rapports
    ShowRSquare: boolean;//df. true
    ShowCorrelation: boolean;//df. false
    ShowRegCoef: boolean;//df. false
    CorrThreshold: double;//seuil pour mise en vidence des corrlations, df. 0.7
    //I-O
    procedure   LoadFromStream(prmStream: TStream); override;
    procedure   SaveToStream(prmStream: TStream); override;
    procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
    procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
    //HTML
    function    getHTMLParameters(): string; override;
    end;

implementation

USES
    Sysutils;

{ TOpVarClusBase }

function TOpVarClusBase.CheckAttributes: boolean;
begin
 result:= (self.WorkData.LstAtts[asInput].Count > 1) and self.WorkData.LstAtts[asInput].isAllCategory(caContinue); 
end;

function TOpVarClusBase.CoreExecute: boolean;
begin
 FInputs:= self.WorkData.LstAtts[asInput];
 TRY
 result:= true;
 if assigned(FCalc) then FCalc.Free();
 FCalc:= getClassCalc().create(FInputs,self.PrmOp);
 FCalc.execute(self.WorkData.Examples);
 //projection si tout va bien
 self.setProjections();
 EXCEPT
 result:= false;
 END;
end;

destructor TOpVarClusBase.destroy;
begin
 if assigned(FCalc) then FCalc.Free();
 inherited;
end;

function TOpVarClusBase.getHTMLResultsSummary: string;
begin
 result:= FCalc.getHTMLResult();
end;

procedure TOpVarClusBase.setProjections;
var j: integer;
begin
 GenAtts.Clear();
 //dupliquer les variables pour chaque cluster
 for j:= 0 to pred(self.FCalc.StrucClusters.getCount()) do
  GenAtts.Add(self.FCalc.StrucClusters.get(j).VarClus.duplicate((self.MLOwner as TMLCompVarClusBase).getGenericAttName()+'_'+IntToStr((self.MLOwner as TMLCompVarClusBase).Number)+'_'+IntToStr(succ(j))));
end;

{ TMLGenVarClus }

procedure TMLGenVarClus.GenCompInitializations;
begin
 FMLComp:= mlcClustering;
end;

{ TOpPrmVarClus }

function TOpPrmVarClus.getHTMLParameters: string;
var s: string;
begin
 s:= '<H5>Report Parameters</H5>';
 s:= s + format('Show R-Square : %d<br>',[ord(ShowRSquare)]);
 s:= s + format('Show Correlation with the cluster''s latent variables : %d<br>',[ord(ShowCorrelation)]);
 s:= s + format('Correlation Threshold for Highlights and counting : %.2f<br>',[CorrThreshold]);
 s:= s + format('Show Regression Coefficients : %d<br>',[ord(ShowRegCoef)]);
 s:= s + '<H5>Algorithm Parameters</H5>';
 result:= s;
end;

procedure TOpPrmVarClus.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 ShowRSquare:= prmINI.ReadBool(prmSection,'show_r_square',ShowRSquare);
 ShowCorrelation:= prmINI.ReadBool(prmSection,'show_correlation',ShowCorrelation);
 ShowRegCoef:= prmINI.ReadBool(prmSection,'show_reg_coef',ShowRegCoef);
 CorrThreshold:= prmINI.ReadFloat(prmSection,'coef_threshold',CorrThreshold);
end;

procedure TOpPrmVarClus.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(ShowRSquare,sizeof(ShowRSquare));
 prmStream.ReadBuffer(ShowCorrelation,sizeof(ShowCorrelation));
 prmStream.ReadBuffer(ShowRegCoef,sizeof(ShowRegCoef));
 prmStream.ReadBuffer(CorrThreshold,sizeof(CorrThreshold));
end;

procedure TOpPrmVarClus.SaveToINI(prmSection: string; prmINI: TMemIniFile);
begin
 prmINI.WriteBool(prmSection,'show_r_square',ShowRSquare);
 prmINI.WriteBool(prmSection,'show_correlation',ShowCorrelation);
 prmINI.WriteBool(prmSection,'show_reg_coef',ShowRegCoef);
 prmINI.WriteFloat(prmSection,'coef_threshold',CorrThreshold);
end;

procedure TOpPrmVarClus.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(ShowRSquare,sizeof(ShowRSquare));
 prmStream.WriteBuffer(ShowCorrelation,sizeof(ShowCorrelation));
 prmStream.WriteBuffer(ShowRegCoef,sizeof(ShowRegCoef));
 prmStream.WriteBuffer(CorrThreshold,sizeof(CorrThreshold));
end;

procedure TOpPrmVarClus.SetDefaultParameters;
begin
 ShowRSquare:= true;
 ShowCorrelation:= true;
 ShowRegCoef:= false;
 CorrThreshold:= 0.7;
end;

end.
