(*****************************************************************)
(* UFrmOpScatterplot.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(*****************************************************************)

unit UFrmOpScatterplot;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, UFrmBaseOperator, Menus, ActnList, ExtCtrls,
  ComCtrls, TeeProcs, TeEngine, Chart, Series, StdCtrls, UDatasetImplementation;

type
  TFrmOpViewDataScatterplot = class(TFrmBaseOperator)
    tabChart: TTabSheet;
    panelChart: TPanel;
    chart: TChart;
    Series1: TPointSeries;
    cmbX2: TComboBox;
    cmbX1: TComboBox;
    cmbBy: TComboBox;
    actModifySeries: TAction;
    Copycharttoclipboard1: TMenuItem;
    actCopyChartToClipboard: TAction;
    N1: TMenuItem;
    Increasepointsize1: TMenuItem;
    Decreasepointsize1: TMenuItem;
    actIncPointSize: TAction;
    actDecresePointSize: TAction;
    N2: TMenuItem;
    Jitter1: TMenuItem;
    procedure actModifySeriesExecute(Sender: TObject);
    procedure actCopyChartToClipboardExecute(Sender: TObject);
    procedure actIncPointSizeExecute(Sender: TObject);
    procedure actDecresePointSizeExecute(Sender: TObject);
    procedure Jitter1Click(Sender: TObject);
  private
  {donnes  visualiser}
  FData: TMLDataset;
  {accepter le rafrachissement  la vole}
  FOkRefresh: boolean;
  //new - 04/06/2012 - le min et max en abcisse et ordonne
  x1Min, x1Max, x2Min, x2Max: double;
  {rafrachir l'affichage}
  procedure   RefreshChart();
  //rafraichir trs simplement -- new - 04/06/2012 - rajouter JITTER
  procedure   RefreshWithoutBy(jitter: boolean = false);
  //rafraichir avec des couleurs diffrentes -- new - 04/06/2012 - rajouter JITTER
  procedure   RefreshWithBy(jitter: boolean = false);
  {crer un scatterplot  nos spcifications}
  function    CreateScatterplot(): TPointSeries;
  public
  procedure   PrepareView(); override;
  function    getAdditionalInfoForReport(): string; override;
  //new - 04/06/2012 - reset le min et le max
  procedure   resetMinMax();
  //new - 04/06/2012 - rcuprer le min et le max. dans le mme temps
  procedure   setScatterValue(scatter: TPointSeries; index: integer; x1, x2: double; jitter: boolean = false);
  end;

var
  FrmOpViewDataScatterplot: TFrmOpViewDataScatterplot;

implementation

uses
        UCompViewScatterPlot, UDatasetDefinition, UCalcDistribution,
        Contnrs, UOperatorDefinition, UCompDefinition, Math;

const
        NB_MOTIFS_GRAPHIQUES = 6;
        MOTIFS_GRAPHIQUES : array[0..pred(NB_MOTIFS_GRAPHIQUES)] of TSeriesPointerStyle
                          = (psCircle,psTriangle,psRectangle,psCross,psDiagCross,psStar);

        MAX_POINT_SIZE = 10;
        MIN_POINT_SIZE = 2;

        JITTER_FACTOR : double = 0.04;

var
        CUR_POINT_SIZE : integer = 2;//!\ ah bon ? on peut faire a en PASCAL maintenant ???

{$R *.dfm}

{ TFrmOpViewDataScatterplot }

function TFrmOpViewDataScatterplot.CreateScatterplot: TPointSeries;
var sc: TPointSeries;
begin
 sc:= TPointSeries.Create(self);
 sc.Pointer.HorizSize:= CUR_POINT_SIZE;
 sc.Pointer.VertSize:= CUR_POINT_SIZE;
 sc.Pointer.Style:= psStar;
 sc.SeriesColor:= clTeeColor;
 //renvoyer le pointeur
 result:= sc;
end;

procedure TFrmOpViewDataScatterplot.PrepareView;
var att: TAttribute;
    i: integer;
begin
 //activer la bonne page
 self.pgctrlFond.ActivePage:= tabChart;
 //connecter les donnes
 FData:= (Operator as TOpScatterPlot).Workdata;
 //la premire slection vide pour le "BY"
 self.cmbBy.Items.AddObject('',nil);
 //remplir les listes
 for i:= 0 to pred(FData.LstAtts[asAll].Count) do
  begin
   att:= FData.LstAtts[asAll].Attribute[i];
   if att.isCategory(caContinue)
    then
     begin
      cmbX1.Items.AddObject(att.Name,att);
      cmbX2.Items.AddObject(att.Name,att);
     end
    else cmbBy.Items.AddObject(att.Name,att);
  end;
 FOkRefresh:= FALSE;
 //provoquer la premire slection
 cmbX1.ItemIndex:= 0;
 cmbX2.ItemIndex:= 0;
 FOkRefresh:= TRUE;
 self.RefreshChart();
end;

procedure TFrmOpViewDataScatterplot.RefreshChart;
begin
 //add by GongYu
 if (cmbx1.Items.count<=0) or (cmbx2.Items.count<=0) then
  begin
   exit;
  end;
 //add end
 //mj min/max
 self.resetMinMax();
 //remplir
 if (cmbBY.ItemIndex>0)
  then RefreshWithBy(false)
  else RefreshWithoutBy(false);
end;

procedure TFrmOpViewDataScatterplot.RefreshWithBy(jitter: boolean = false);
var x1,x2,y: TAttribute;
    scatter: TPointSeries;
    dist: TTabFrequence;
    j,i,p,example: integer;
    lstChart: TObjectLIst;
begin
 chart.SeriesList.Clear;
 //rcuprer les attributs
 x1:= cmbX1.Items.Objects[cmbX1.ItemIndex] as TAttribute;
 x2:= cmbX2.Items.Objects[cmbX2.ItemIndex] as TAttribute;
 y:= cmbBy.Items.Objects[cmbBy.ItemIndex] as TAttribute;
 if assigned(y)
  then
   begin
     //rcuprer le titre
     chart.Title.Text.Text:= format('(X1) %s vs. (X2) %s by (Y) %s',[x1.Name,x2.Name,y.Name]);
     //trs difficile de travailler sans faire deux passages malheureusement
     //construire la distribution des donnes
     dist:= TTabFrequence.CreateFromAtt(y,FData.Examples);
     lstChart:= TObjectList.Create(FALSE);
     //initialiser les scatters
     for j:= 1 to dist.Size do
      begin
       scatter:= self.CreateScatterplot();
       //mj du style pour les copies en noir et blanc -- de 0  NB_MOTIFS_GRAPHIQUES
       scatter.Pointer.Style:= MOTIFS_GRAPHIQUES[pred(j) mod pred(NB_MOTIFS_GRAPHIQUES)];
       scatter.Title:= y.LstValues.getDescription(j);
       //tester si la liste est vide !!!
       if (dist.Value[j]>0)
        then scatter.FillSampleValues(dist.Value[j]);// tout a pour a !!!
       lstChart.Add(scatter);
      end;
     //dist va maintenant servir de compteur d'ajout
     dist.ReInitialization();
     //remplir les scatters
     for i:= 1 to FData.Examples.Size do
      begin
       //rcuprer le bon scatter
       example:= FData.Examples.Number[i];
       p:= y.dValue[example];
       scatter:= lstChart.Items[pred(p)] as TPointSeries;
       //remplir
       //scatter.XValue[dist.Value[p]]:= x1.cValue[example];
       //scatter.YValue[dist.Value[p]]:= x2.cValue[example];
       self.setScatterValue(scatter,dist.Value[p],x1.cValue[example],x2.cValue[example],jitter);
       dist.IncrementCell(p);
      end;
     //insrer les sries dans le CHART
     for j:= 0 to pred(lstChart.Count) do
      begin
       scatter:= lstChart.Items[j] as TPointSeries;
       chart.AddSeries(scatter);
      end;
     dist.Free;
     lstChart.Free;
     chart.Legend.Visible:= TRUE;
   end;
end;

procedure TFrmOpViewDataScatterplot.RefreshWithoutBy(jitter: boolean = false);
var x1,x2: TAttribute;
    scatter: TPointSeries;
    i: integer;
begin
 chart.Legend.Visible:= FALSE;
 //vider le graphique
 chart.SeriesList.Clear;
 //rcuprer les attributs
 x1:= cmbX1.Items.Objects[cmbX1.ItemIndex] as TAttribute;
 x2:= cmbX2.Items.Objects[cmbX2.ItemIndex] as TAttribute;
 //rcuprer le titre
 chart.Title.Text.Text:= format('(X1) %s vs. (X2) %s',[x1.Name,x2.Name]);
 //crer la srie
 scatter:= self.CreateScatterplot();
 scatter.FillSampleValues(FData.Examples.Size);
 //puis remplir les points
 for i:= 1 to FData.Examples.Size do
  begin
   //scatter.XValue[pred(i)]:= x1.cValue[FData.Examples.Number[i]];
   //scatter.YValue[pred(i)]:= x2.cValue[FData.Examples.Number[i]];
   self.setScatterValue(scatter,pred(i),x1.cValue[FData.Examples.Number[i]],x2.cValue[FData.Examples.Number[i]],jitter);
  end;
 chart.AddSeries(scatter);
end;

procedure TFrmOpViewDataScatterplot.actModifySeriesExecute(
  Sender: TObject);
begin
 //modifier les sries slectionnes
 self.RefreshChart();
end;

procedure TFrmOpViewDataScatterplot.actCopyChartToClipboardExecute(
  Sender: TObject);
begin
 self.chart.CopyToClipboardMetafile(TRUE);
end;

procedure TFrmOpViewDataScatterplot.actIncPointSizeExecute(
  Sender: TObject);
var k: integer;
begin
 if (CUR_POINT_SIZE < MAX_POINT_SIZE)
  then
   begin
     for k:= 0 to pred(self.chart.SeriesCount) do
      begin
       (self.Chart.Series[k] as TPointSeries).Pointer.HorizSize:= succ(CUR_POINT_SIZE);
       (self.Chart.Series[k] as TPointSeries).Pointer.VertSize := succ(CUR_POINT_SIZE);
      end;
     inc(CUR_POINT_SIZE);
   end;
end;

procedure TFrmOpViewDataScatterplot.actDecresePointSizeExecute(
  Sender: TObject);
var k: integer;
begin
 if (CUR_POINT_SIZE > MIN_POINT_SIZE)
  then
   begin
     for k:= 0 to pred(self.chart.SeriesCount) do
      begin
       (self.Chart.Series[k] as TPointSeries).Pointer.HorizSize:= pred(CUR_POINT_SIZE);
       (self.Chart.Series[k] as TPointSeries).Pointer.VertSize := pred(CUR_POINT_SIZE);
      end;
     dec(CUR_POINT_SIZE);
   end;
end;

function TFrmOpViewDataScatterplot.getAdditionalInfoForReport: string;
var s: string;
begin
 //titre
 s:= '<H3>Scatter Plot</H3>';
 s:= s + self.addChartToReport(chart);
 result:= s;
end;

procedure TFrmOpViewDataScatterplot.resetMinMax;
begin
 self.x1Min:= Math.MaxDouble;
 self.x1Max:= -1.0 * Math.MaxDouble;
 self.x2Min:= Math.MaxDouble;
 self.x2Max:= -1.0 * Math.MaxDouble;
end;

procedure TFrmOpViewDataScatterplot.setScatterValue(scatter: TPointSeries;
  index: integer; x1, x2: double; jitter: boolean = false);
begin
 //premier passage
 if (jitter = false)
  then
   begin
    x1Min:= Min(self.x1Min,x1);
    x2Min:= Min(self.x2Min,x2);
    x1Max:= Max(self.x1Max,x1);
    x2Max:= Max(self.x2Max,x2);
    scatter.XValue[index]:= x1;
    scatter.YValue[index]:= x2;
   end
  else
   begin
    scatter.XValue[index]:= x1 + (0.5 - random)*JITTER_FACTOR*(self.x1Max - self.x1Min);
    scatter.YValue[index]:= x2 + (0.5 - random)*JITTER_FACTOR*(self.x2Max - self.x2Min);
   end;
end;

procedure TFrmOpViewDataScatterplot.Jitter1Click(Sender: TObject);
begin
 //mme appel, mais avec bruitage
 if (cmbBY.ItemIndex>0)
  then RefreshWithBy(true)
  else RefreshWithoutBy(true);
end;

end.
