(*********************************************************************)
(* UCompFCResidualScores.pas - Copyright (c) 2007 Ricco RAKOTOMALALA *)
(*********************************************************************)

{
@abstract(Residual Scores)
@author(Ricco)
@created(03/03/2007)

Calcul du rsidu de la rgression, en srie pour une liste de TARGET.
L'ide est de produire des variables intermdiaires qui permettent de produire
des analyses partielles (corrlation partielles essentiellement).

Pas mal par ex., l'ide d'enlever l'effet de taille dans une ACP pour une meilleure
exploitation des informations disponibles.

}
unit UCompFCResidualScores;

interface

USES
    //units std delphi
    Classes, IniFiles, Forms, Sysutils,
    //gestion des composants
    UCompDefinition, UOperatorDefinition,
    UCompManageDataset,
    UDatasetDefinition, UDatasetImplementation,
    UCompFCDefinition, UDatasetExamples,
    MATRIXMM;


TYPE
    //********************************************************************************
    //**** Gnrateur de composant, il permet d'insrer un composant dans le diagramme
    //********************************************************************************
    TMLGenCompFCResidScores = class(TMLGenComp)
    protected
    //spcifier dans quel onglet de la palette doit apparatre l'icne --  surcharger ventuellement dans les hritiers
    procedure   genCompInitializations(); override;
    public
    //spcifier quel est la classe du composant  gnrer sur le diagramme --  surcharger obligatoirement chez les hritiers
    function    getClassMLComponent: TClassMLComponent; override;
    end;

    //**********************************************************
    //**** class du composant graphique insr dans le diagramme
    //**********************************************************
    TMLCompFCResidScores = class(TMLCompFC)
    protected
    //spcifier l'oprateur de calcul --  surcharger obligatoirement
    function    getClassOperator: TClassOperator; override;
    end;

    //**************************************************************************************
    //** classe oprateur de calcul : prend en charge le calcul et l'affichage des rsultats
    //**************************************************************************************
    TOpFCResidScores = class(TOperatorFC)
    private
    //Target et Input
    FTarget, FInput: TLstAttributes;
    //matrice [INV(X'X)]X'
    FXPXM1XP: PManipMatMM;
    //tableau des matrices de projection (les coefs. a)
    FTabCoef: array of PManipMatMM;
    //tableau des R (variance explique)
    FTabR2: array of double; 
    //dtruire les matrices internes
    procedure   disposeMatrices();
    //calculer la matrice de projection
    procedure   computeProjMatrix(examples: TExamples);
    //projection des variables de score
    procedure   setProjection();
    protected
    //spcifier la classe de paramtrage -- surcharge obligatoire
    function    getClassParameter: TClassOperatorParameter; override;
    //contrle de cohrence des attributs avant les calculs --  surcharger selon les mthodes
    function    checkAttributes(): boolean; override;
    //lancer effectivement les calculs -- surcharge obligatoire
    function    coreExecute(): boolean; override;
    public
    //libration de la mmoire alloue -- surcharge selon les mthodes, appel de la mthode anctre avec "inherited" obligatoire
    destructor  destroy(); override;
    //affichage des rsultats au format HTML -- surcharge obligatoire
    function    getHTMLResultsSummary(): string; override;
    end;

    //***********************************************************************
    //** classe paramtrage de l'oprateur de calcul : gre la bote de paramtrage et l'E/S des paramtres
    //** si la mthode de requiert pas de paramtres, il suffit de crer une classe vide hritire de "TOperatorParameter"
    //***********************************************************************
    TOpPrmFCResidScores = class(TOperatorPrmFC)
    protected
    //cration de la bote de dialogue de paramtrage -- surcharge obligatoire (doit renvoyer "nil" si pas de bote  gnrer)
    function    CreateDlgParameters(): TForm; override;
    //
    procedure   SetDefaultParameters(); override;
    end;

implementation

uses UConstConfiguration;


{ TMLGenCompFCResidScores }

procedure TMLGenCompFCResidScores.genCompInitializations;
begin
 //le composant est plac dans la palette "Feature Construction"
 FMLComp:= mlcFeatureConstruction;
end;

function TMLGenCompFCResidScores.getClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFCResidScores; 
end;

{ TMLCompFCResidScores }

function TMLCompFCResidScores.getClassOperator: TClassOperator;
begin
 result:= TOpFCResidScores;
end;

{ TOpFCResidScores }

function TOpFCResidScores.checkAttributes: boolean;
var ok: boolean;
begin
 ok:= (self.WorkData.LstAtts[asTarget].Count > 0) and (self.WorkData.LstAtts[asTarget].isAllCategory(caContinue) = true);
 ok:= ok AND ((self.WorkData.LstAtts[asInput].Count > 0) and (self.WorkData.LstAtts[asInput].isAllCategory(caContinue) = true));
 //and then...
 result:= ok;
end;

procedure TOpFCResidScores.computeProjMatrix(examples: TExamples);
var X,XP,XPX,XPXM1: PManipMatMM;
    i,j,example,NbJ,k: integer;
    Y: PManipMatMM;
    YC,E: PManipMatMM;
    att: TAttribute;
    mY: TData;
    SCR,SCT,R2: double;
begin
 FTarget:= self.WorkData.LstAtts[asTarget];
 FInput:= self.WorkData.LstAtts[asInput];
 NbJ:= FInput.Count;
 //construire la matrice de donnes
 Y:= PManipMatMM.Create(examples.Size,1);//initialis une seule fois, mais utilis pour chaque TARGET
 X:= PManipMatMM.Create(examples.Size,FInput.Count+1);
 TRY
 for i:= 1 to examples.Size do
  begin
   example:= examples.Number[i];
   for j:= 1 to NbJ do
    X[i,j]:= FInput.Attribute[pred(j)].cValue[example];
   X[i,NbJ+1]:= 1.0;
  end;
 //****************************************************************
 //transposition, multiplier, inverser et re-multiplier --> XPXM1XP
 //****************************************************************
 TranspositionMM(X,XP);
 MultiplicationMM(XP,X,XPX);
 if (InversionMM(XPX,XPXM1)<>ErrMM_NoError)
  then Raise Exception.Create('SINGULAR MATRIX');
 MultiplicationMM(XPXM1,XP,FXPXM1XP);
 //***********************************************************************
 //calculer les coefs. de la rgression, et les indicateurs de qualit, R
 //***********************************************************************
 SetLength(FTabCoef,FTarget.Count);
 SetLength(FTabR2,FTarget.Count);
 //pour chaque endogne
 for k:= 0 to pred(FTarget.Count) do
  begin
   att:= FTarget.Attribute[k];
   //charger les valeurs de Y
   for i:= 1 to examples.Size do Y[i,1]:= att.cValue[examples.Number[i]];
   //moyenne des Y
   mY:= MoyenneMM(Y,1);
   //calculer les coefs.
   MultiplicationMM(FXPXM1XP,Y,FTabCoef[k]);
   //projection
   MultiplicationMM(X,FTabCoef[k],YC);
   //vecteur des erreurs
   SoustractionMM(Y,YC,E);
   //calcul du R2
   SCT:= 0.0;
   SCR:= 0.0;
   for i:= 1 to Y.MatMM.Lig do
    begin
     SCT:= SCT + SQR(Y[i,1]-mY);
     SCR:= SCR + SQR(E[i,1]);//la moyenne est zro forcment
    end;
   if (SCT > 0.0)
    then R2:= 1.0 - SCR / SCT
    else R2:= 0.0;
   FTabR2[k]:= R2;
   //dtruire pour le coup d'aprs
   YC.Free();
   E.Free();
  end;
 FINALLY
 if assigned(X) then X.Free();
 if assigned(XP) then XP.Free();
 if assigned(XPX) then XPX.Free();
 if assigned(XPXM1) then XPXM1.Free();
 if assigned(Y) then Y.Free();
 END;
end;

function TOpFCResidScores.coreExecute: boolean;
begin
 result:= true;
 TRY
 //librer le cas chant
 if assigned(FXPXM1XP) then self.disposeMatrices();
 //charger les donnes, crer la matrice XPXM1XP et les coefs. de rgression
 self.computeProjMatrix(self.WorkData.Examples);
 //crer les nouvelles variables rsiduelles
 self.setProjection();
 EXCEPT
 result:= false;
 END;
end;

destructor TOpFCResidScores.destroy;
begin
 if assigned(FXPXM1XP) then self.disposeMatrices();
 inherited;
end;

procedure TOpFCResidScores.disposeMatrices;
var j: integer;
begin
 FreeAndNil(FXPXM1XP);
 for j:= 0 to pred(length(FTabCoef)) do
  FTabCoef[j].Free();
 Finalize(FTabCoef);
 Finalize(FTabR2);
end;

function TOpFCResidScores.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFCResidScores; 
end;

function TOpFCResidScores.getHTMLResultsSummary: string;
var s: string;
    k,j: integer;
begin
 s:= '<H3>Residual Scores</H3>';
 s:= s +  '<H5>Regression coefficients and Explained Variance</H5>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 //en-tte
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TH width=100 align=center>Target</TH><TH width=60 %s>Expl. Var. (%s)</TH>',[HTML_BGCOLOR_HEADER_GREEN,'%']);
 for j:= 0 to pred(FInput.Count) do
  s:= s + format('<TH width=60>%s</TH>',[FInput.Attribute[j].Name]);
 s:= s + '<TH width=60>constant</TH></TR>';
 //pour chaque target
 for k:= 0 to pred(FTarget.Count) do
  begin
   s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right %s>%.1f</TD>',[FTarget.Attribute[k].Name,HTML_BGCOLOR_DATA_GREEN,100.0*FTabR2[k]]);
   for j:= 0 to pred(FInput.Count) do
    s:= s + format('<TD align=right>%.4f</TD>',[FTabCoef[k].Valeur[succ(j),1]]);
   s:= s + format('<TD align=right>%.4f</TD></TR>',[FTabCoef[k].Valeur[succ(FInput.Count),1]]);
  end;
 s:= s + '</table>';
 result:= s;
end;

procedure TOpFCResidScores.setProjection;
var numComp: integer;
    k,i,j: integer;
    att,newAtt: TAttribute;
    residual,value: single;
begin
 //vider
 GenAtts.Clear();
 //rcuprr le numro de composant
 numComp:= (self.MLOwner as TMLComponent).Number;
 //pour chaque target
 for k:= 0 to pred(FTarget.Count) do
  begin
   att:= FTarget.Attribute[k];
   //crer et ajouter le correspondant
   newAtt:= TAttContinue.Create(format('rs_%s_%d',[att.Name,numComp]),att.Size);
   GenAtts.Add(newAtt);
   //pour chaque observation
   for i:= 1 to att.Size do
    begin
     //projection
     value:= FTabCoef[k].Valeur[FInput.Count+1,1];
     for j:= 1 to FInput.Count do
      value:= value + FTabCoef[k].Valeur[j,1] * FInput.Attribute[pred(j)].cValue[i];
     //rsidu et affectation
     residual:= att.cValue[i] - value;
     newAtt.cValue[i]:= residual;
    end;
  end;
end;

{ TOpPrmFCResidScores }

function TOpPrmFCResidScores.CreateDlgParameters: TForm;
begin
 result:= NIL;
end;

procedure TOpPrmFCResidScores.SetDefaultParameters;
begin
 //
end;

initialization
 RegisterClass(TMLGenCompFCResidScores);
end.
