(*************************************************************)
(* ULibCVMImport.pas - Copyright (c) 2012 Ricco RAKOTOMALALA *)
(*************************************************************)

{
@abstract(Unit d'importation de LIBCVM)
@author(Ricco)
@created(06/05/2012)

Importation implicite des fonctions de la bibliothque LIBCVM
(http://c2inet.sce.ntu.edu.sg/ivor/cvm.html)

Base sur LIBSVM 2.85, les structures de donnes et le fonctions sont les mmes.
Pas de souci donc, on se cale sur l'unit d'import de LIBSVM

}

unit ULibCVMImport;

interface

TYPE
//struct svm_node
svm_node = record
           index: integer;
           value: double;
end;
p_svm_node = ^svm_node;

//tableaux de noeuds
array_svm_node = array of svm_node;
array_array_svm_node = array of array_svm_node;

//tableaux de rels
array_double = array of double;
array_array_double = array of array_double;

//tableaux d'entiers
array_integer = array of integer;

//struct svm_problem
svm_problem = record
              l: integer;
              u: integer;//CVM !
              y: array_double;
              x: array_array_svm_node;
              graph: pointer; //structure de graphe - on n'exploitera pas !!! --> pointeur "libre", hum....
end;
p_svm_problem = ^svm_problem;

//enum { C_SVC, NU_SVC, ONE_CLASS, EPSILON_SVR, NU_SVR, CVDD, CVM, CVM_LS, CVR, BVM };	/* svm_type */
svm_type = (C_SVC, NU_SVC, ONE_CLASS, EPSILON_SVR, NU_SVR, CVDD, CVM, CVM_LS, CVR, BVM);

CONST
str_svm_type : array[svm_type] of string = ('C-SVC','NU-SVC','ONE CLASS','EPSILON-SVR','NU-SVR', 'CVDD', 'CVM', 'CVM-LS', 'CVR', 'BVM');

TYPE
//enum { LINEAR, POLY, RBF, SIGMOID, PRECOMPUTED, EXP, NORMAL_POLY, INV_DIST, INV_SQDIST };   /* kernel_type */
kernel_type = (LINEAR, POLY, RBF, SIGMOID, PRECOMPUTED, EXP, NORMAL_POLY, INV_DIST, INV_SQDIST);

CONST
str_kernel_type : array[kernel_type] of string = ('LINEAR','POLY','RBF','SIGMOID', 'PRECOMPUTED', 'LAPLACIAN', 'NORMALIZED POLY', 'INVERSE DISTANCE', 'INVERSE SQUARE DISTANCE');

TYPE
//struct svm_parameter
svm_parameter = record
	svm_type: integer;
	kernel_type: integer;
	degree: double;	//* for poly */
	gamma: double;	//* for poly/rbf/sigmoid */
	coef0: double;	//* for poly/sigmoid */

	//* these are for training only */
	cache_size: double; //* in MB */
	eps: double;	//* stopping criteria */
	C: double;	//* for C_SVC, EPSILON_SVR and NU_SVR */
	nr_weight: integer;		//* for C_SVC */
	weight_label: array_integer;	//* for C_SVC */
	weight: array_double;		//* for C_SVC */
	nu: double;	//* for NU_SVC, ONE_CLASS, and NU_SVR */
        mu: double; //for CVR
	p: double;	//* for EPSILON_SVR */
        mc_type: integer;        //* for multiclass SVM/CVM/BVM */
	shrinking: integer;	//* use the shrinking heuristics */
	probability: integer; //* do probability estimates */
        //new - rr - pour CVM - 06/05/2012
	sample_size: integer;    //* size of probabilistic sampling in BVM */
	num_basis: integer;
	knn: integer;
	weight_type: integer;
end;
p_svm_parameter = ^svm_parameter;

//
// svm_model
//
//struct svm_model
svm_model = record
	param: svm_parameter;	// parameter
	nr_class: integer;		// number of classes, = 2 in regression/one class svm
	l: integer;			// total #SV
        u: integer;
	SV: array_array_svm_node;		// SVs (SV[l])
	sv_coef: array_array_double;	// coefficients for SVs in decision functions (sv_coef[n-1][l])
	rho: array_double;		// constants in decision functions (rho[n*(n-1)/2])
        cNorm: array_double;          // center Norm of decison functions (rho[n*(n-1)/2])
	probA: array_double;          // pariwise probability information
	probB: array_double;
	// for classification only
	_label: array_integer;		// label of each class (label[n])
	nSV: array_integer;		// number of SVs for each class (nSV[n])
				// nSV[0] + nSV[1] + ... + nSV[n-1] = l
	// XXX
	free_sv: integer;		// 1 if svm_model is created by svm_load_model
				            // 0 if svm_model is created by svm_train
end;
p_svm_model = ^svm_model;


function  svm_train(const prob: p_svm_problem ; const param: p_svm_parameter): p_svm_model; stdcall; external '.\dll\libcvm.dll';
procedure svm_cross_validation(const prob: p_svm_problem ; const param: p_svm_parameter ; nr_fold: integer ; target: array_double); stdcall; external '.\dll\libcvm.dll';

//R.R. -- inutile pour l'instant les fonctions d'I/O, on ne sauve pas les modles.
//__declspec(dllexport) int __stdcall svm_save_model(const char *model_file_name, const struct svm_model *model); stdcall; external '.\dll\libcvm.dll';
//__declspec(dllexport) struct svm_model* __stdcall svm_load_model(const char *model_file_name); stdcall; external '.\dll\libcvm.dll';

function  svm_get_svm_type(const model: p_svm_model): integer; stdcall; external '.\dll\libcvm.dll';
function  svm_get_nr_class(const model: p_svm_model): integer; stdcall; external '.\dll\libcvm.dll';
procedure svm_get_labels(const model: p_svm_model ; _label: array_integer); stdcall; external '.\dll\libcvm.dll';
function  svm_get_svr_probability(const model: p_svm_model): double; stdcall; external '.\dll\libcvm.dll';

procedure svm_predict_values(const model: p_svm_model ; const x: array_svm_node ; var dec_values: array_double); stdcall; external '.\dll\libcvm.dll';
function  svm_predict(const model: p_svm_model ; const x: array_svm_node): double; stdcall; external '.\dll\libcvm.dll';
function  svm_predict_probability(const model: p_svm_model ; const x: array_svm_node ; const prob_estimates: array_double): double; stdcall; external '.\dll\libcvm.dll';

procedure svm_destroy_model(model: p_svm_model); stdcall; external '.\dll\libcvm.dll';
procedure svm_destroy_param(param: p_svm_parameter); stdcall; external '.\dll\libcvm.dll';

function  svm_check_parameter(const prob: p_svm_problem ; const param: p_svm_parameter): PChar; stdcall; external '.\dll\libcvm.dll';
function  svm_check_probability_model(const prob: p_svm_problem): integer; stdcall; external '.\dll\libcvm.dll';

implementation

end.
